Package: singleRcapture
Type: Package
Title: Single-Source Capture-Recapture Models
Version: 1.1.0
Authors@R: 
    c(person(given = "Piotr",
             family = "Chlebicki",
             role = c("aut", "ctb"),
             email = "piotrek.pecet@gmail.com",
             comment = c(ORCID = "0009-0006-4867-7434")),
      person(given = "Maciej",
             family = "Beręsewicz",
             role = c("aut", "cre"),
             email = "maciej.beresewicz@ue.poznan.pl",
             comment = c(ORCID = "0000-0002-8281-4301")),
      person(given = "Cyprian",
             family = "Jurkowski",
             role = "ctb"))
Maintainer: Maciej Beręsewicz <maciej.beresewicz@ue.poznan.pl>
Description: Implementation of single-source capture-recapture methods for population size estimation using zero-truncated, zero-one truncated and zero-truncated one-inflated Poisson, Geometric and Negative Binomial regression as well as Zelterman's, Chao's and ratio-regression estimators. Package includes point and interval estimators for the population size with variances estimated using analytical or bootstrap method. Details can be found in: van der Heijden et all. (2003) <doi:10.1191/1471082X03st057oa>, Böhning and van der Heijden (2019) <doi:10.1214/18-AOAS1232>, Böhning et al. (2020) Capture-Recapture Methods for the Social and Medical Sciences or Böhning and Friedl (2021) <doi:10.1007/s10260-021-00556-8>.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: yes
RdMacros: mathjaxr
BuildManual: TRUE
VignetteBuilder: knitr
Depends: R (>= 3.5.0)
RoxygenNote: 7.3.3
URL: https://github.com/ncn-foreigners/singleRcapture
BugReports: https://github.com/ncn-foreigners/singleRcapture/issues
Imports: stats, lamW, mathjaxr, sandwich, doParallel, foreach, parallel
Suggests: rmarkdown, knitr, tinytest, covr, VGAM, lmtest
NeedsCompilation: no
Packaged: 2026-05-13 07:34:02 UTC; berenz
Author: Piotr Chlebicki [aut, ctb] (ORCID:
    <https://orcid.org/0009-0006-4867-7434>),
  Maciej Beręsewicz [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8281-4301>),
  Cyprian Jurkowski [ctb]
Repository: CRAN
Date/Publication: 2026-05-13 08:30:02 UTC
Built: R 4.5.2; ; 2026-05-13 11:18:14 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(add1,singleRStaticCountData)
S3method(anova,singleRStaticCountData)
S3method(bread,singleRStaticCountData)
S3method(coef,singleRRatioReg)
S3method(confint,singleRRatioReg)
S3method(confint,singleRStaticCountData)
S3method(cooks.distance,singleRStaticCountData)
S3method(df.residual,singleRStaticCountData)
S3method(dfbeta,singleRStaticCountData)
S3method(dfpopsize,singleRStaticCountData)
S3method(drop1,singleRStaticCountData)
S3method(estfun,singleRStaticCountData)
S3method(estimatePopsize,default)
S3method(extractAIC,singleRRatioReg)
S3method(extractAIC,singleRStaticCountData)
S3method(family,singleRStaticCountData)
S3method(fitted,singleRStaticCountData)
S3method(hatvalues,singleRStaticCountData)
S3method(influence,singleRStaticCountData)
S3method(logLik,singleRRatioReg)
S3method(logLik,singleRStaticCountData)
S3method(model.frame,singleRStaticCountData)
S3method(model.matrix,singleRStaticCountData)
S3method(nobs,singleRRatioReg)
S3method(nobs,singleRStaticCountData)
S3method(plot,singleRRatioReg)
S3method(plot,singleRStaticCountData)
S3method(popSizeEst,singleRRatioReg)
S3method(popSizeEst,singleRStaticCountData)
S3method(predict,singleRStaticCountData)
S3method(print,popSizeEstResults)
S3method(print,singleRRatioReg)
S3method(print,singleRStaticCountData)
S3method(print,singleRfamily)
S3method(print,summarysingleRRatioReg)
S3method(print,summarysingleRStaticCountData)
S3method(print,summarysingleRmargin)
S3method(profile,singleRStaticCountData)
S3method(redoPopEstimation,singleRStaticCountData)
S3method(residuals,singleRStaticCountData)
S3method(rstandard,singleRStaticCountData)
S3method(rstudent,singleRStaticCountData)
S3method(sigma,singleRStaticCountData)
S3method(simulate,singleRStaticCountData)
S3method(simulate,singleRfamily)
S3method(stratifyPopsize,singleRStaticCountData)
S3method(summary,singleRRatioReg)
S3method(summary,singleRStaticCountData)
S3method(summary,singleRmargin)
S3method(vcov,singleRStaticCountData)
S3method(vcovHC,singleRStaticCountData)
export(Hurdleztgeom)
export(Hurdleztnegbin)
export(Hurdleztpoisson)
export(chao)
export(controlMethod)
export(controlModel)
export(controlPopVar)
export(dfpopsize)
export(estimatePopsize)
export(estimatePopsizeFit)
export(marginalFreq)
export(oichao)
export(oiztgeom)
export(oiztnegbin)
export(oiztpoisson)
export(popSizeEst)
export(ratioReg)
export(redoPopEstimation)
export(stratifyPopsize)
export(zelterman)
export(zotgeom)
export(zotnegbin)
export(zotpoisson)
export(ztHurdlegeom)
export(ztHurdlenegbin)
export(ztHurdlepoisson)
export(ztgeom)
export(ztnegbin)
export(ztoigeom)
export(ztoinegbin)
export(ztoipoisson)
export(ztpoisson)
import(mathjaxr)
importFrom(doParallel,registerDoParallel)
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(graphics,abline)
importFrom(graphics,arrows)
importFrom(graphics,axis)
importFrom(graphics,barplot)
importFrom(graphics,boxplot)
importFrom(graphics,hist)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,matplot)
importFrom(graphics,panel.smooth)
importFrom(graphics,par)
importFrom(graphics,points)
importFrom(graphics,text)
importFrom(lamW,lambertW0)
importFrom(parallel,makeCluster)
importFrom(parallel,stopCluster)
importFrom(sandwich,bread)
importFrom(sandwich,estfun)
importFrom(sandwich,vcovHC)
importFrom(stats,AIC)
importFrom(stats,BIC)
importFrom(stats,add1)
importFrom(stats,anova)
importFrom(stats,coef)
importFrom(stats,confint)
importFrom(stats,contrasts)
importFrom(stats,cooks.distance)
importFrom(stats,dbinom)
importFrom(stats,density)
importFrom(stats,df.residual)
importFrom(stats,dfbeta)
importFrom(stats,dlnorm)
importFrom(stats,dnbinom)
importFrom(stats,dnorm)
importFrom(stats,drop1)
importFrom(stats,extractAIC)
importFrom(stats,family)
importFrom(stats,fitted)
importFrom(stats,glm)
importFrom(stats,hatvalues)
importFrom(stats,influence)
importFrom(stats,lm)
importFrom(stats,logLik)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.response)
importFrom(stats,nobs)
importFrom(stats,optim)
importFrom(stats,pnorm)
importFrom(stats,ppoints)
importFrom(stats,predict)
importFrom(stats,printCoefmat)
importFrom(stats,profile)
importFrom(stats,pt)
importFrom(stats,qnorm)
importFrom(stats,qqline)
importFrom(stats,qqnorm)
importFrom(stats,quantile)
importFrom(stats,rbinom)
importFrom(stats,reformulate)
importFrom(stats,residuals)
importFrom(stats,rstandard)
importFrom(stats,rstudent)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,sigma)
importFrom(stats,simulate)
importFrom(stats,terms)
importFrom(stats,uniroot)
importFrom(stats,update)
importFrom(stats,var)
importFrom(stats,vcov)
importFrom(stats,weighted.mean)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# `singleRcapture`: An R Package for Single-Source Capture-Recapture Models <img src="man/figures/logo.png" align="right" width="150"/>

<!-- badges: start -->

[![R-CMD-check](https://github.com/ncn-foreigners/singleRcapture/workflows/R-CMD-check/badge.svg)](https://github.com/ncn-foreigners/singleRcapture/actions)
[![Codecov test
coverage](https://codecov.io/gh/ncn-foreigners/singleRcapture/branch/main/graph/badge.svg)](https://app.codecov.io/gh/ncn-foreigners/singleRcapture?branch=main)
[![CRAN
status](https://www.r-pkg.org/badges/version/singleRcapture)](https://CRAN.R-project.org/package=singleRcapture)
[![CRAN total
downloads](https://cranlogs.r-pkg.org/badges/grand-total/singleRcapture)](https://cran.r-project.org/package=singleRcapture)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/singleRcapture)](https://cran.r-project.org/package=singleRcapture)
[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.8436043.svg)](https://doi.org/10.5281/zenodo.8436043)
<!-- badges: end -->

Capture-recapture type experiments are used to estimate the total
population size in situations when observing only a part of such
population is feasible. In recent years these types of experiments have
seen more interest.

Single-source models are distinct from other capture-recapture models
because we cannot estimate the population size based on how many units
were observed in two or three sources which is the standard approach.

Instead in single-source models we utilize count data regression models
on positive distributions (i.e. on counts greater than 0) where the
dependent variable is the number of times a particular unit was observed
in source data.

This package aims to implement already existing and introduce new
methods of estimating population size from single source to simplify the
research process.

Currently, we have implemented most of the frequentist approaches used
in literature such as:

- Zero-truncated Poisson, geometric and negative binomial regression.
- Zero-truncated one-inflated and one-inflated zero-truncated Poisson,
  geometric and negative binomial model.
- Zero-one-truncated Poisson, geometric and negative binomial models.
- Generalized Chao’s and Zelterman’s models based on logistic
  regression.
- Three types of bootstrap parametric, semi-parametric and
  nonparametric.
- And a wide range of additional functionalities associated with
  (vector) generalized linear models relevant to the topic.

For more details we see the
`singleRcapture: An R Package for Single-Source Capture-Recapture Models`
vignette on [CRAN](https://cran.r-project.org/package=singleRcapture) or
[pkgdown
website](https://ncn-foreigners.ue.poznan.pl/singleRcapture/articles/singleRcapture.html).

## Installation

You can install the current version of the `singleRcapture` package from
main branch [GitHub](https://github.com/ncn-foreigners/singleRcapture)
with:

``` r
# install.packages("devtools")
remotes::install_github("ncn-foreigners/singleRcapture")
```

or install the stable version from
[CRAN](https://cran.r-project.org/package=singleRcapture) with:

``` r
install.packages(singleRcapture)
```

### Examples

The main function of this package is `estimatePopsize` which fitts
regression on specified distribution and then uses fitted regression to
estimate the population size.

Lets look at a model from 2003 publication (Van Der Heijden, P. G.,
Bustami, R., Cruyff, M. J., Engbersen, G., & Van Houwelingen, H. C.
(2003). Point and interval estimation of the population size using the
truncated Poisson regression model. Statistical Modelling, 3(4),
305-322.). The call to `estimatePopsize` will look very similar to
anyone who used the `stats::glm` function:

``` r
library(singleRcapture)
model <- estimatePopsize(
  formula = capture ~ gender + age + nation, # specify formula
  data = netherlandsimmigrant,
  popVar = "analytic", # specify 
  model = "ztpoisson", # distribution used
  method = "IRLS", # fitting method one of three currently supported
  controlMethod = controlMethod(silent = TRUE) # ignore convergence at half step warning
)
summary(model) # a summary method for singleR class with standard glm-like output and population size estimation resutls
#> 
#> Call:
#> estimatePopsize.default(formula = capture ~ gender + age + nation, 
#>     data = netherlandsimmigrant, model = "ztpoisson", method = "IRLS", 
#>     popVar = "analytic", controlMethod = controlMethod(silent = TRUE))
#> 
#> Pearson Residuals:
#>      Min.   1st Qu.    Median      Mean   3rd Qu.      Max. 
#> -0.486442 -0.486442 -0.298080  0.002093 -0.209444 13.910844 
#> 
#> Coefficients:
#> -----------------------
#> For linear predictors associated with: lambda 
#>                      Estimate Std. Error z value  P(>|z|)    
#> (Intercept)           -1.3411     0.2149  -6.241 4.35e-10 ***
#> gendermale             0.3972     0.1630   2.436 0.014832 *  
#> age>40yrs             -0.9746     0.4082  -2.387 0.016972 *  
#> nationAsia            -1.0926     0.3016  -3.622 0.000292 ***
#> nationNorth Africa     0.1900     0.1940   0.979 0.327398    
#> nationRest of Africa  -0.9106     0.3008  -3.027 0.002468 ** 
#> nationSurinam         -2.3364     1.0136  -2.305 0.021159 *  
#> nationTurkey          -1.6754     0.6028  -2.779 0.005445 ** 
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> AIC: 1712.901
#> BIC: 1757.213
#> Residual deviance: 1128.553
#> 
#> Log-likelihood: -848.4504 on 1872 Degrees of freedom 
#> Number of iterations: 8
#> -----------------------
#> Population size estimation results: 
#> Point estimate 12690.35
#> Observed proportion: 14.8% (N obs = 1880)
#> Std. Error 2808.165
#> 95% CI for the population size:
#>           lowerBound upperBound
#> normal      7186.449   18194.25
#> logNormal   8431.277   19718.31
#> 95% CI for the share of observed population:
#>           lowerBound upperBound
#> normal     10.332933   26.16035
#> logNormal   9.534288   22.29793
```

We implemented a method for `plot` function to visualise the model fit
and other useful diagnostic information. One of which is `rootogram`, a
type of plot that compares fitted and observed marginal frequencies:

``` r
plot(model, plotType = "rootogram")
```

<img src="man/figures/README-plot-1.png" alt="" width="75%" />

The possible values for `plotType` argument are:

- `qq` - the normal quantile-quantile plot for pearson residuals
  (default),
- `marginal` - a `matplot` comparing fitted and observed marginal
  frequencies,
- `fitresid` - plot of linear predictor values contrasted with pearson
  residuals,
- `bootHist` - histogram of bootstrap sample,
- `rootogram` - rootogram, example presented above,
- `dfpopContr` - contrasting two deletion effects to identify presence
  of influential observations,
- `dfpopBox` - boxplot of results from `dfpopsize` function see its
  documentation,
- `scaleLoc` - scale-location plot,
- `cooks` - plot of `cooks.values` for distributions for which it is
  defined,
- `hatplot` - plot of `hatvalues`,
- `strata` - plot of confidence intervals for selected such populations.

User can also pass arguments to specify additional information such as
plot title, subtitle etc. similar to calling `plot` on some data. For
more info check `plot.singleR` method documentation.

As we have seen there are some significant differences between fitted
and observed marginal frequencies. To check our intuition let’s perform
goodness of fit test between fitted and observed marginal frequencies.

To do it we call a `summary` function of `marginalFreq` function which
computes marginal frequencies for the fitted `singleR` class object:

``` r
summary(marginalFreq(model), df = 2, dropl5 = "group")
#> Test for Goodness of fit of a regression model:
#> 
#>                  Test statistics df P(>X^2)
#> Chi-squared test           50.06  2 1.3e-11
#> G-test                     34.31  2 3.6e-08
#> 
#> -------------------------------------------------------------- 
#> Cells with fitted frequencies of < 5 have been grouped 
#> Names of cells used in calculating test(s) statistic: 1 2 3
```

Finally let us check if we have any influential observations. We will do
this by comparing the deletion effect of every observation on population
size estimate by removing it entirely from the model (from population
size estimate and regression) and by only omitting it in pop size
estimation (this is what is called the contribution of an observation).
If observation is not influential these two actions should have the
approximately the same effect:

``` r
plot(model, plotType = "dfpopContr")
```

<img src="man/figures/README-plot-example-1.png" alt="" width="75%" />

it is easy to deduce from the plot above that we have influential
observations in our dataset (one in particular).

Lastly `singleRcapture` offers some post-hoc procedures for example a
function `stratifyPopsize` that estimates sizes of user specified sub
populations and returns them in a `data.frame`:

``` r
stratifyPopsize(model, alpha = c(.01, .02, .03, .05), # different significance level for each sub population
    strata = list(
    "Females from Surinam" = netherlandsimmigrant$gender == "female" & netherlandsimmigrant$nation == "Surinam",
    "Males from Turkey" = netherlandsimmigrant$gender == "male" & netherlandsimmigrant$nation == "Turkey",
    "Younger males" = netherlandsimmigrant$gender == "male" & netherlandsimmigrant$age == "<40yrs",
    "Older males" = netherlandsimmigrant$gender == "male" & netherlandsimmigrant$age == ">40yrs"
))
#>                   name Observed Estimated ObservedPercentage  StdError
#> 1 Females from Surinam       20  931.4677           2.147149  955.0657
#> 2    Males from Turkey       78 1291.2513           6.040652  741.0066
#> 3        Younger males     1391 7337.0708          18.958520 1282.1402
#> 4          Older males       91 1542.1886           5.900705  781.4747
#>   normalLowerBound normalUpperBound logNormalLowerBound logNormalUpperBound
#> 1      -1528.61853         3391.554            119.2661            8389.158
#> 2       -432.58790         3015.090            405.4127            4573.791
#> 3       4554.71057        10119.431           5134.8122           10834.785
#> 4         10.52637         3073.851            630.7551            3992.674
#>   confLevel
#> 1      0.01
#> 2      0.02
#> 3      0.03
#> 4      0.05
```

`strata` argument may be specified in various ways for example:

``` r
stratifyPopsize(model, strata = ~ gender / age)
#>                     name Observed Estimated ObservedPercentage  StdError
#> 1         gender==female      398 3811.0911          10.443203 1153.9733
#> 2           gender==male     1482 8879.2594          16.690581 1812.0790
#> 3 genderfemale:age<40yrs      378 3169.8263          11.924944  880.9478
#> 4   gendermale:age<40yrs     1391 7337.0708          18.958520 1282.1402
#> 5 genderfemale:age>40yrs       20  641.2648           3.118836  407.5264
#> 6   gendermale:age>40yrs       91 1542.1886           5.900705  781.4747
#>   normalLowerBound normalUpperBound logNormalLowerBound logNormalUpperBound
#> 1       1549.34513         6072.837           2189.0443            6902.133
#> 2       5327.64991        12430.869           6090.7762           13354.880
#> 3       1443.20030         4896.452           1904.3126            5484.617
#> 4       4824.12208         9850.019           5306.3306           10421.082
#> 5       -157.47223         1440.002            212.3382            2026.726
#> 6         10.52637         3073.851            630.7551            3992.674
#>   confLevel
#> 1      0.05
#> 2      0.05
#> 3      0.05
#> 4      0.05
#> 5      0.05
#> 6      0.05
```

The package was designed with convenience in mind, for example it is
possible to specify that weights provided on call are to be interpreted
as number of occurrences of units in each row:

``` r
df <- netherlandsimmigrant[, c(1:3,5)]
df$ww <- 0
### this is dplyr::count without dependencies
df <- aggregate(ww ~ ., df, FUN = length)
summary(estimatePopsize(
  formula = capture ~ nation + age + gender, 
  data = df, 
  model = ztpoisson, 
  weights = df$ww,
  controlModel = controlModel(weightsAsCounts = TRUE)
))
#> 
#> Call:
#> estimatePopsize.default(formula = capture ~ nation + age + gender, 
#>     data = df, model = ztpoisson, weights = df$ww, controlModel = controlModel(weightsAsCounts = TRUE))
#> 
#> Pearson Residuals:
#>      Min.   1st Qu.    Median      Mean   3rd Qu.      Max. 
#> -317.6467   -2.4060    3.7702    0.0803   13.4920  183.2108 
#> 
#> Coefficients:
#> -----------------------
#> For linear predictors associated with: lambda 
#>                      Estimate Std. Error z value  P(>|z|)    
#> (Intercept)           -1.3411     0.2149  -6.241 4.35e-10 ***
#> nationAsia            -1.0926     0.3016  -3.622 0.000292 ***
#> nationNorth Africa     0.1900     0.1940   0.979 0.327398    
#> nationRest of Africa  -0.9106     0.3008  -3.027 0.002468 ** 
#> nationSurinam         -2.3364     1.0136  -2.305 0.021159 *  
#> nationTurkey          -1.6754     0.6028  -2.779 0.005445 ** 
#> age>40yrs             -0.9746     0.4082  -2.387 0.016972 *  
#> gendermale             0.3972     0.1630   2.436 0.014832 *  
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> AIC: 1712.901
#> BIC: 1757.213
#> Residual deviance: 1128.553
#> 
#> Log-likelihood: -848.4504 on 1872 Degrees of freedom 
#> Number of iterations: 8
#> -----------------------
#> Population size estimation results: 
#> Point estimate 12690.35
#> Observed proportion: 14.8% (N obs = 1880)
#> Std. Error 2808.169
#> 95% CI for the population size:
#>           lowerBound upperBound
#> normal      7186.444   18194.26
#> logNormal   8431.275   19718.32
#> 95% CI for the share of observed population:
#>           lowerBound upperBound
#> normal     10.332927   26.16037
#> logNormal   9.534281   22.29793
```

Methods such as regression diagnostics will be adjusted (values of
weights will be reduced instead of rows being removed etc.)

We also included option to use common non standard argument such as
significance levels different from usual 5%:

``` r
set.seed(123)
modelInflated <- estimatePopsize(
    formula = capture ~ gender + age,
    data = netherlandsimmigrant,
    model = "ztoigeom",
    method = "IRLS",
    # control parameters for population size estimation check documentation of controlPopVar
    controlPopVar = controlPopVar(
        alpha = .01, # significance level 
    )
)
summary(modelInflated)
#> 
#> Call:
#> estimatePopsize.default(formula = capture ~ gender + age, data = netherlandsimmigrant, 
#>     model = "ztoigeom", method = "IRLS", controlPopVar = controlPopVar(alpha = 0.01, 
#>         ))
#> 
#> Pearson Residuals:
#>      Min.   1st Qu.    Median      Mean   3rd Qu.      Max. 
#> -0.357193 -0.357193 -0.357193  0.000343 -0.287637 10.233608 
#> 
#> Coefficients:
#> -----------------------
#> For linear predictors associated with: lambda 
#>             Estimate Std. Error z value P(>|z|)    
#> (Intercept)  -1.5346     0.1846  -8.312 < 2e-16 ***
#> gendermale    0.3863     0.1380   2.800 0.00512 ** 
#> age>40yrs    -0.7788     0.2942  -2.648 0.00810 ** 
#> -----------------------
#> For linear predictors associated with: omega 
#>             Estimate Std. Error z value  P(>|z|)    
#> (Intercept)  -1.7591     0.3765  -4.673 2.97e-06 ***
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> AIC: 1736.854
#> BIC: 1759.01
#> Residual deviance: 1011.271
#> 
#> Log-likelihood: -864.4272 on 3756 Degrees of freedom 
#> Number of iterations: 6
#> -----------------------
#> Population size estimation results: 
#> Point estimate 5661.522
#> Observed proportion: 33.2% (N obs = 1880)
#> Std. Error 963.9024
#> 99% CI for the population size:
#>           lowerBound upperBound
#> normal      3178.674   8144.370
#> logNormal   3861.508   9096.681
#> 99% CI for the share of observed population:
#>           lowerBound upperBound
#> normal      23.08343   59.14416
#> logNormal   20.66688   48.68564
```

and the option to estimate standard error of population size estimate by
bootstrap, models with more than one distribution parameter being
dependent on covariates and some non standard link functions for
example:

``` r
modelInflated2 <- estimatePopsize(
    formula = capture ~ age,
    data = netherlandsimmigrant,
    popVar = "bootstrap",
    model = ztoigeom(omegaLink = "cloglog"),
    method = "IRLS",
    controlPopVar = controlPopVar(
        B = 500,# number of boostrap samples
        alpha = .01, # significance level 
        # type of bootstrap see documentation for estimatePopsize
        bootType = "semiparametric",
        # control regression fitting on bootstrap samples
        bootstrapFitcontrol = controlMethod(
          epsilon = .Machine$double.eps, 
          silent = TRUE, 
          stepsize = 2
        )
    ),
    controlModel = controlModel(omegaFormula = ~ gender) # put covariates on omega i.e. the inflation parameter
)
#> Warning in estimatePopsize.default(formula = capture ~ age, data = netherlandsimmigrant, : The (analytically computed) hessian of the score function is not negative define.
#> NOTE: Second derivative test failing does not 
#>         necessarily mean that the maximum of score function that was found 
#>         numericaly is invalid since R^k is not a bounded space.
#> Additionally in one inflated and hurdle models second derivative test often fails even on valid arguments.
#> Warning in estimatePopsize.default(formula = capture ~ age, data =
#> netherlandsimmigrant, : Switching from observed information matrix to Fisher
#> information matrix because hessian of log-likelihood is not negative define.
popSizeEst(modelInflated2)
#> Point estimate: 5496.376
#> Variance: 1217986
#> 99% confidence intervals:
#> lowerBound upperBound 
#>   4002.763  10231.716
```

the results are significantly different (the warning issued concerns the
second derivative test for existence of local minimum, here it was
inconclusive but we manually checked that fitting process found the
optimal regression coefficients it’s here to provide more information to
the user):

``` r
plot(modelInflated2, 
     plotType = "bootHist", 
     labels = TRUE, 
     ylim = c(0, 175),
     breaks = 15)
```

<img src="man/figures/README-plot-inflated-1.png" alt="" width="75%" />

and information criteria support the second model:

    #>  First model: AIC = 1736.854 BIC = 1759.01
    #> Second model: AIC = 1734.803 BIC = 1756.959

## Funding

Work on this package is supported by the the National Science Centre,
OPUS 20 grant no. 2020/39/B/HS4/00941.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# `singleRcapture`: An R Package for Single-Source Capture-Recapture Models <img src="man/figures/logo.png" align="right" width="150"/>

<!-- badges: start -->

[![R-CMD-check](https://github.com/ncn-foreigners/singleRcapture/workflows/R-CMD-check/badge.svg)](https://github.com/ncn-foreigners/singleRcapture/actions)
[![Codecov test
coverage](https://codecov.io/gh/ncn-foreigners/singleRcapture/branch/main/graph/badge.svg)](https://app.codecov.io/gh/ncn-foreigners/singleRcapture?branch=main)
[![CRAN
status](https://www.r-pkg.org/badges/version/singleRcapture)](https://CRAN.R-project.org/package=singleRcapture)
[![CRAN total
downloads](https://cranlogs.r-pkg.org/badges/grand-total/singleRcapture)](https://cran.r-project.org/package=singleRcapture)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/singleRcapture)](https://cran.r-project.org/package=singleRcapture)
[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.8436043.svg)](https://doi.org/10.5281/zenodo.8436043)
<!-- badges: end -->

Capture-recapture type experiments are used to estimate the total
population size in situations when observing only a part of such
population is feasible. In recent years these types of experiments have
seen more interest.

Single-source models are distinct from other capture-recapture models
because we cannot estimate the population size based on how many units
were observed in two or three sources which is the standard approach.

Instead in single-source models we utilize count data regression models
on positive distributions (i.e. on counts greater than 0) where the
dependent variable is the number of times a particular unit was observed
in source data.

This package aims to implement already existing and introduce new
methods of estimating population size from single source to simplify the
research process.

Currently, we have implemented most of the frequentist approaches used
in literature such as:

- Zero-truncated Poisson, geometric and negative binomial regression.
- Zero-truncated one-inflated and one-inflated zero-truncated Poisson,
  geometric and negative binomial model.
- Zero-one-truncated Poisson, geometric and negative binomial models.
- Generalized Chao’s and Zelterman’s models based on logistic
  regression.
- Modified Chao’s one-inflation-robust `oichao()` model based on counts
  2 and 3.
- Standalone `ratioReg()` support for ratio regression with optional
  one-inflation model selection.
- Three types of bootstrap parametric, semi-parametric and
  nonparametric.
- And a wide range of additional functionalities associated with
  (vector) generalized linear models relevant to the topic.

For more details we see the
`singleRcapture: An R Package for Single-Source Capture-Recapture Models`
vignette on [CRAN](https://cran.r-project.org/package=singleRcapture) or
[pkgdown
website](https://ncn-foreigners.ue.poznan.pl/singleRcapture/articles/singleRcapture.html).

## Installation

You can install the current version of the `singleRcapture` package from
main branch [GitHub](https://github.com/ncn-foreigners/singleRcapture)
with:

``` r
# install.packages("devtools")
remotes::install_github("ncn-foreigners/singleRcapture")
```

or install the stable version from
[CRAN](https://cran.r-project.org/package=singleRcapture) with:

``` r
install.packages("singleRcapture")
```

### Examples

The main function of this package is `estimatePopsize` which fits
regression on specified distribution and then uses fitted regression to
estimate the population size.

Let’s look at a model from 2003 publication (Van Der Heijden, P. G.,
Bustami, R., Cruyff, M. J., Engbersen, G., & Van Houwelingen, H. C.
(2003). Point and interval estimation of the population size using the
truncated Poisson regression model. Statistical Modelling, 3(4),
305-322.). The call to `estimatePopsize` will look very similar to
anyone who used the `stats::glm` function:

``` r
library(singleRcapture)
model <- estimatePopsize(
  formula = capture ~ gender + age + nation, # specify formula
  data = netherlandsimmigrant,
  popVar = "analytic", # specify 
  model = "ztpoisson", # distribution used
  method = "IRLS", # fitting method one of three currently supported
  controlMethod = controlMethod(silent = TRUE) # ignore convergence at half step warning
)
summary(model) # a summary method for singleR class with standard glm-like output and population size estimation results
#> 
#> Call:
#> estimatePopsize.default(formula = capture ~ gender + age + nation, 
#>     data = netherlandsimmigrant, model = "ztpoisson", method = "IRLS", 
#>     popVar = "analytic", controlMethod = controlMethod(silent = TRUE))
#> 
#> Pearson Residuals:
#>      Min.   1st Qu.    Median      Mean   3rd Qu.      Max. 
#> -0.486442 -0.486442 -0.298080  0.002093 -0.209444 13.910844 
#> 
#> Coefficients:
#> -----------------------
#> For linear predictors associated with: lambda 
#>                      Estimate Std. Error z value  P(>|z|)    
#> (Intercept)           -1.3411     0.2149  -6.241 4.35e-10 ***
#> gendermale             0.3972     0.1630   2.436 0.014832 *  
#> age>40yrs             -0.9746     0.4082  -2.387 0.016972 *  
#> nationAsia            -1.0926     0.3016  -3.622 0.000292 ***
#> nationNorth Africa     0.1900     0.1940   0.979 0.327398    
#> nationRest of Africa  -0.9106     0.3008  -3.027 0.002468 ** 
#> nationSurinam         -2.3364     1.0136  -2.305 0.021159 *  
#> nationTurkey          -1.6754     0.6028  -2.779 0.005445 ** 
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> AIC: 1712.901
#> BIC: 1757.213
#> Residual deviance: 1128.553
#> 
#> Log-likelihood: -848.4504 on 1872 Degrees of freedom 
#> Number of iterations: 8
#> -----------------------
#> Population size estimation results: 
#> Point estimate 12690.35
#> Observed proportion: 14.8% (N obs = 1880)
#> Std. Error 2808.165
#> 95% CI for the population size:
#>           lowerBound upperBound
#> normal      7186.449   18194.25
#> logNormal   8431.277   19718.31
#> 95% CI for the share of observed population:
#>           lowerBound upperBound
#> normal     10.332933   26.16035
#> logNormal   9.534288   22.29793
```

We implemented a method for `plot` function to visualise the model fit
and other useful diagnostic information. One of which is `rootogram`, a
type of plot that compares fitted and observed marginal frequencies:

``` r
plot(model, plotType = "rootogram")
```

<img src="man/figures/README-plot-1.png" alt="" width="75%" />

The possible values for `plotType` argument are:

- `qq` - the normal quantile-quantile plot for pearson residuals
  (default),
- `marginal` - a `matplot` comparing fitted and observed marginal
  frequencies,
- `fitresid` - plot of linear predictor values contrasted with pearson
  residuals,
- `bootHist` - histogram of bootstrap sample,
- `rootogram` - rootogram, example presented above,
- `dfpopContr` - contrasting two deletion effects to identify presence
  of influential observations,
- `dfpopBox` - boxplot of results from `dfpopsize` function see its
  documentation,
- `scaleLoc` - scale-location plot,
- `cooks` - plot of `cooks.values` for distributions for which it is
  defined,
- `hatplot` - plot of `hatvalues`,
- `strata` - plot of confidence intervals for selected such populations.

User can also pass arguments to specify additional information such as
plot title, subtitle etc. similar to calling `plot` on some data. For
more info check `plot.singleR` method documentation.

The package also includes a standalone `ratioReg()` function for the
ratio-regression approach. It is separate from `singleRmodels()` because
it works on neighbouring marginal-frequency ratios rather than on
unit-level count likelihoods:

``` r
toyRatio <- data.frame(
  observed = c(rep(1, 50), rep(2, 30), rep(3, 15), rep(4, 6), rep(5, 2))
)

ratioModel <- ratioReg(
  observed ~ 1,
  data = toyRatio,
  model = "auto",
  estimator = "SM",
  confint = "none"
)

summary(ratioModel)
#> 
#> Call:
#> ratioReg(formula = observed ~ 1, data = toyRatio, model = "auto", 
#>     estimator = "SM", confint = "none")
#> 
#> Selected ratio regression model: M1 
#> Selection criterion: AIC 
#> Primary population estimator: SM 
#> Observed population size: 103 
#> 
#> Model selection table:
#>    model available logLik     AIC    BIC selected
#> M0    M0      TRUE   7.85  -9.701 -11.54    FALSE
#> M1    M1      TRUE  20.14 -32.284 -34.74     TRUE
#> 
#> Coefficients for M0 :
#> (Intercept)  log(k + 1) 
#>     -0.1035     -0.5724 
#> 
#> Coefficients for M1 :
#>  (Intercept)   log(k + 1) oneInflation 
#>       0.1714      -0.7865      -0.1371 
#> 
#> Population size estimates:
#>  estimator pointEstimate variance lowerBound upperBound
#>         HT      139.7316       NA         NA         NA
#>         SM      139.7308       NA         NA         NA
```

The default ratio-regression plot shows the observed log-ratios together
with the fitted `M0` and `M1` curves:

``` r
plot(ratioModel)
```

<img src="man/figures/README-ratioReg-plot-1.png" alt="" width="75%" />

As we have seen there are some significant differences between fitted
and observed marginal frequencies. To check our intuition let’s perform
goodness of fit test between fitted and observed marginal frequencies.

To do it we call a `summary` function of `marginalFreq` function which
computes marginal frequencies for the fitted `singleR` class object:

``` r
summary(marginalFreq(model), df = 2, dropl5 = "group")
#> Test for Goodness of fit of a regression model:
#> 
#>                  Test statistics df P(>X^2)
#> Chi-squared test           50.06  2 1.3e-11
#> G-test                     34.31  2 3.6e-08
#> 
#> -------------------------------------------------------------- 
#> Cells with fitted frequencies of < 5 have been grouped 
#> Names of cells used in calculating test(s) statistic: 1 2 3
```

Finally let us check if we have any influential observations. We will do
this by comparing the deletion effect of every observation on population
size estimate by removing it entirely from the model (from population
size estimate and regression) and by only omitting it in pop size
estimation (this is what is called the contribution of an observation).
If observation is not influential these two actions should have the
approximately the same effect:

``` r
plot(model, plotType = "dfpopContr")
```

<img src="man/figures/README-plot-example-1.png" alt="" width="75%" />

it is easy to deduce from the plot above that we have influential
observations in our dataset (one in particular).

Lastly `singleRcapture` offers some post-hoc procedures for example a
function `stratifyPopsize` that estimates sizes of user specified sub
populations and returns them in a `data.frame`:

``` r
stratifyPopsize(model, alpha = c(.01, .02, .03, .05), # different significance level for each sub population
    strata = list(
    "Females from Surinam" = netherlandsimmigrant$gender == "female" & netherlandsimmigrant$nation == "Surinam",
    "Males from Turkey" = netherlandsimmigrant$gender == "male" & netherlandsimmigrant$nation == "Turkey",
    "Younger males" = netherlandsimmigrant$gender == "male" & netherlandsimmigrant$age == "<40yrs",
    "Older males" = netherlandsimmigrant$gender == "male" & netherlandsimmigrant$age == ">40yrs"
))
#>                   name Observed Estimated ObservedPercentage  StdError
#> 1 Females from Surinam       20  931.4677           2.147149  955.0657
#> 2    Males from Turkey       78 1291.2513           6.040652  741.0066
#> 3        Younger males     1391 7337.0708          18.958520 1282.1402
#> 4          Older males       91 1542.1886           5.900705  781.4747
#>   normalLowerBound normalUpperBound logNormalLowerBound logNormalUpperBound
#> 1      -1528.61853         3391.554            119.2661            8389.158
#> 2       -432.58790         3015.090            405.4127            4573.791
#> 3       4554.71057        10119.431           5134.8122           10834.785
#> 4         10.52637         3073.851            630.7551            3992.674
#>   confLevel
#> 1      0.01
#> 2      0.02
#> 3      0.03
#> 4      0.05
```

`strata` argument may be specified in various ways for example:

``` r
stratifyPopsize(model, strata = ~ gender / age)
#>                     name Observed Estimated ObservedPercentage  StdError
#> 1         gender==female      398 3811.0911          10.443203 1153.9733
#> 2           gender==male     1482 8879.2594          16.690581 1812.0790
#> 3 genderfemale:age<40yrs      378 3169.8263          11.924944  880.9478
#> 4   gendermale:age<40yrs     1391 7337.0708          18.958520 1282.1402
#> 5 genderfemale:age>40yrs       20  641.2648           3.118836  407.5264
#> 6   gendermale:age>40yrs       91 1542.1886           5.900705  781.4747
#>   normalLowerBound normalUpperBound logNormalLowerBound logNormalUpperBound
#> 1       1549.34513         6072.837           2189.0443            6902.133
#> 2       5327.64991        12430.869           6090.7762           13354.880
#> 3       1443.20030         4896.452           1904.3126            5484.617
#> 4       4824.12208         9850.019           5306.3306           10421.082
#> 5       -157.47223         1440.002            212.3382            2026.726
#> 6         10.52637         3073.851            630.7551            3992.674
#>   confLevel
#> 1      0.05
#> 2      0.05
#> 3      0.05
#> 4      0.05
#> 5      0.05
#> 6      0.05
```

The package was designed with convenience in mind, for example it is
possible to specify that weights provided on call are to be interpreted
as number of occurrences of units in each row:

``` r
df <- netherlandsimmigrant[, c(1:3,5)]
df$ww <- 0
### this is dplyr::count without dependencies
df <- aggregate(ww ~ ., df, FUN = length)
summary(estimatePopsize(
  formula = capture ~ nation + age + gender, 
  data = df, 
  model = ztpoisson, 
  weights = df$ww,
  controlModel = controlModel(weightsAsCounts = TRUE)
))
#> 
#> Call:
#> estimatePopsize.default(formula = capture ~ nation + age + gender, 
#>     data = df, model = ztpoisson, weights = df$ww, controlModel = controlModel(weightsAsCounts = TRUE))
#> 
#> Pearson Residuals:
#>      Min.   1st Qu.    Median      Mean   3rd Qu.      Max. 
#> -317.6467   -2.4060    3.7702    0.0803   13.4920  183.2108 
#> 
#> Coefficients:
#> -----------------------
#> For linear predictors associated with: lambda 
#>                      Estimate Std. Error z value  P(>|z|)    
#> (Intercept)           -1.3411     0.2149  -6.241 4.35e-10 ***
#> nationAsia            -1.0926     0.3016  -3.622 0.000292 ***
#> nationNorth Africa     0.1900     0.1940   0.979 0.327398    
#> nationRest of Africa  -0.9106     0.3008  -3.027 0.002468 ** 
#> nationSurinam         -2.3364     1.0136  -2.305 0.021159 *  
#> nationTurkey          -1.6754     0.6028  -2.779 0.005445 ** 
#> age>40yrs             -0.9746     0.4082  -2.387 0.016972 *  
#> gendermale             0.3972     0.1630   2.436 0.014832 *  
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> AIC: 1712.901
#> BIC: 1757.213
#> Residual deviance: 1128.553
#> 
#> Log-likelihood: -848.4504 on 1872 Degrees of freedom 
#> Number of iterations: 8
#> -----------------------
#> Population size estimation results: 
#> Point estimate 12690.35
#> Observed proportion: 14.8% (N obs = 1880)
#> Std. Error 2808.169
#> 95% CI for the population size:
#>           lowerBound upperBound
#> normal      7186.444   18194.26
#> logNormal   8431.275   19718.32
#> 95% CI for the share of observed population:
#>           lowerBound upperBound
#> normal     10.332927   26.16037
#> logNormal   9.534281   22.29793
```

Methods such as regression diagnostics will be adjusted (values of
weights will be reduced instead of rows being removed etc.)

We also included option to use common non standard argument such as
significance levels different from usual 5%:

``` r
set.seed(123)
modelInflated <- estimatePopsize(
    formula = capture ~ gender + age,
    data = netherlandsimmigrant,
    model = "ztoigeom",
    method = "IRLS",
    # control parameters for population size estimation check documentation of controlPopVar
    controlPopVar = controlPopVar(
        alpha = .01, # significance level 
    )
)
summary(modelInflated)
#> 
#> Call:
#> estimatePopsize.default(formula = capture ~ gender + age, data = netherlandsimmigrant, 
#>     model = "ztoigeom", method = "IRLS", controlPopVar = controlPopVar(alpha = 0.01, 
#>         ))
#> 
#> Pearson Residuals:
#>      Min.   1st Qu.    Median      Mean   3rd Qu.      Max. 
#> -0.357193 -0.357193 -0.357193  0.000343 -0.287637 10.233608 
#> 
#> Coefficients:
#> -----------------------
#> For linear predictors associated with: lambda 
#>             Estimate Std. Error z value P(>|z|)    
#> (Intercept)  -1.5346     0.1846  -8.312 < 2e-16 ***
#> gendermale    0.3863     0.1380   2.800 0.00512 ** 
#> age>40yrs    -0.7788     0.2942  -2.648 0.00810 ** 
#> -----------------------
#> For linear predictors associated with: omega 
#>             Estimate Std. Error z value  P(>|z|)    
#> (Intercept)  -1.7591     0.3765  -4.673 2.97e-06 ***
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> AIC: 1736.854
#> BIC: 1759.01
#> Residual deviance: 1011.271
#> 
#> Log-likelihood: -864.4272 on 3756 Degrees of freedom 
#> Number of iterations: 6
#> -----------------------
#> Population size estimation results: 
#> Point estimate 5661.522
#> Observed proportion: 33.2% (N obs = 1880)
#> Std. Error 963.9024
#> 99% CI for the population size:
#>           lowerBound upperBound
#> normal      3178.674   8144.370
#> logNormal   3861.508   9096.681
#> 99% CI for the share of observed population:
#>           lowerBound upperBound
#> normal      23.08343   59.14416
#> logNormal   20.66688   48.68564
```

and the option to estimate standard error of population size estimate by
bootstrap, models with more than one distribution parameter being
dependent on covariates and some non standard link functions for
example:

``` r
modelInflated2 <- estimatePopsize(
    formula = capture ~ age,
    data = netherlandsimmigrant,
    popVar = "bootstrap",
    model = ztoigeom(omegaLink = "cloglog"),
    method = "IRLS",
    controlPopVar = controlPopVar(
        B = 500,# number of boostrap samples
        alpha = .01, # significance level 
        # type of bootstrap see documentation for estimatePopsize
        bootType = "semiparametric",
        # control regression fitting on bootstrap samples
        bootstrapFitcontrol = controlMethod(
          epsilon = .Machine$double.eps, 
          silent = TRUE, 
          stepsize = 2
        )
    ),
    controlModel = controlModel(omegaFormula = ~ gender) # put covariates on omega i.e. the inflation parameter
)
#> Warning in estimatePopsize.default(formula = capture ~ age, data = netherlandsimmigrant, : The (analytically computed) hessian of the score function is not negative define.
#> NOTE: Second derivative test failing does not 
#>         necessarily mean that the maximum of score function that was found 
#>         numericaly is invalid since R^k is not a bounded space.
#> Additionally in one inflated and hurdle models second derivative test often fails even on valid arguments.
#> Warning in estimatePopsize.default(formula = capture ~ age, data =
#> netherlandsimmigrant, : Switching from observed information matrix to Fisher
#> information matrix because hessian of log-likelihood is not negative define.
popSizeEst(modelInflated2)
#> Point estimate: 5496.376
#> Variance: 1217986
#> 99% confidence intervals:
#> lowerBound upperBound 
#>   4002.763  10231.716
```

the results are significantly different (the warning issued concerns the
second derivative test for existence of local minimum, here it was
inconclusive but we manually checked that fitting process found the
optimal regression coefficients it’s here to provide more information to
the user):

``` r
plot(modelInflated2, 
     plotType = "bootHist", 
     labels = TRUE, 
     ylim = c(0, 175),
     breaks = 15)
```

<img src="man/figures/README-plot-inflated-1.png" alt="" width="75%" />

and information criteria support the second model:

    #>  First model: AIC = 1736.854 BIC = 1759.01
    #> Second model: AIC = 1734.803 BIC = 1756.959

## Funding

Work on this package is supported by the National Science Centre through
OPUS 20 grant no. 2020/39/B/HS4/00941.

## Citation

Chlebicki, P., & Beręsewicz, M. (2026). singleRcapture: An R Package for
Single-Source Capture-Recapture Models. Journal of Statistical Software,
115(1), 1–30. <https://doi.org/10.18637/jss.v115.i01>

``` bibtex
 @Article{,
    title = {{singleRcapture}: An {R} Package for Single-Source Capture-Recapture Models},
    author = {Piotr Chlebicki and Maciej Ber\c{e}sewicz},
    journal = {Journal of Statistical Software},
    year = {2025},
    volume = {115},
    number = {1},
    pages = {1--30},
    doi = {10.18637/jss.v115.i01},
  }
```

developer_tests_enabled <- function() {
  isTRUE(tolower(Sys.getenv(
    "TEST_SINGLERCAPTURE_MULTICORE_DEVELOPER",
    unset = "FALSE"
  )) == "true")
}

load_inflated_test_data <- function() {
  list(
    N = 500,
    x1 = rep(0:3, c(99, 228, 137, 36)),
    test_inflated = read.csv("test_inflated.csv")
  )
}

load_zerotruncated_test_data <- function() {
  set.seed(123)

  test_zerotruncated <- read.csv("test_zerotruncated.csv")

  list(
    eps = if (capabilities("long.double")) sqrt(.Machine$double.eps) else 0.01,
    N = 1000,
    test_zerotruncated = test_zerotruncated,
    x1 = test_zerotruncated$x1,
    x2 = test_zerotruncated$x2,
    x3 = test_zerotruncated$x3
  )
}

source("helper_families.R")
if (!developer_tests_enabled()) exit_file("developer tests disabled")

inflated_test_data <- load_inflated_test_data()
N <- inflated_test_data$N
x1 <- inflated_test_data$x1
test_inflated <- inflated_test_data$test_inflated

# ztoinegbin
coefficients_ztoinegbin <- c(.6, .2, -1.25, .1, -1, -.3)
#eta <- cbind(coefficients_ztoinegbin[1] + coefficients_ztoinegbin[2] * x1, coefficients_ztoinegbin[3] + coefficients_ztoinegbin[4] * x1, coefficients_ztoinegbin[5] + coefficients_ztoinegbin[6] * x1)

model_ztoinegbin <- ztoinegbin(
  lambdaLink = "log",
  alphaLink  = "log",
  omegaLink  = "logit",
)

#y <- model_ztoinegbin$simulate(n = N, eta = eta, lower = -1)
observed_ztoinegbin <- test_inflated$ztoinegbin1

data_ztoinegbin <- data.frame(
  x = x1[observed_ztoinegbin > 0],
  y = observed_ztoinegbin[observed_ztoinegbin > 0]
)

fit_ztoinegbin <- estimatePopsize(
  formula = y ~ x,
  model   = model_ztoinegbin,
  data    = data_ztoinegbin,
  method = "IRLS",
  controlModel = controlModel(
    omegaFormula = ~ x,
    alphaFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

expect_silent(summary(fit_ztoinegbin))

expect_silent(print(summary(fit_ztoinegbin)))

# don't change this
expect_true(
  sum(abs(coef(fit_ztoinegbin) - coefficients_ztoinegbin)) < 8
)

expect_silent(
  population_ztoinegbin <- popSizeEst(fit_ztoinegbin)
)

expect_equivalent(
  population_ztoinegbin$pointEstimate,
  N,
  tolerance = .3
)

expect_true(
  (population_ztoinegbin$confidenceInterval[1, 1] < N) &
  (N < population_ztoinegbin$confidenceInterval[1, 2])
)

expect_true(
  (population_ztoinegbin$confidenceInterval[2, 1] < N) &
  (N < population_ztoinegbin$confidenceInterval[2, 2])
)

expect_true(
  all(predict(
    fit_ztoinegbin,
    type = "response",
    se.fit = TRUE
  )[, 4:6] > 0)
)

expect_true(
  all(predict(
    fit_ztoinegbin,
    type = "link",
    se.fit = TRUE
  )[, 4:6]> 0)
)

expect_true(
  all(predict(
    fit_ztoinegbin,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_ztoinegbin,
    type = "contr"
  )),
  fit_ztoinegbin$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztoinegbin,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_ztoinegbin,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 4:6] > 0)
)

expect_true(
  all(predict(
    fit_ztoinegbin,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[, 4:6]> 0)
)

expect_true(
  all(predict(
    fit_ztoinegbin,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztoinegbin,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_ztoinegbin,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

source("helper_families.R")
if (!developer_tests_enabled()) exit_file("developer tests disabled")

inflated_test_data <- load_inflated_test_data()
N <- inflated_test_data$N
x1 <- inflated_test_data$x1
test_inflated <- inflated_test_data$test_inflated

# negbin
coefficients_Hurdleztnegbin <- c(.4, .4, -.4, .2, -1.1, .37)
#eta <- cbind(coefficients_Hurdleztnegbin[1] + coefficients_Hurdleztnegbin[2] * x1, coefficients_Hurdleztnegbin[3] + coefficients_Hurdleztnegbin[4] * x1, coefficients_Hurdleztnegbin[5] + coefficients_Hurdleztnegbin[6] * x1)

model_Hurdleztnegbin <- Hurdleztnegbin(
  lambdaLink = "log",
  alphaLink  = "log",
  piLink  = "logit",
  eimStep = 30
)

#y <- model_Hurdleztnegbin$simulate(n = N, eta = eta, lower = -1)
observed_Hurdleztnegbin <- test_inflated$Hurdleztnegbin1

data_Hurdleztnegbin <- data.frame(
  x = x1[observed_Hurdleztnegbin > 0],
  y = observed_Hurdleztnegbin[observed_Hurdleztnegbin > 0]
)

fit_Hurdleztnegbin <- estimatePopsize(
  formula = y ~ x,
  model   = model_Hurdleztnegbin,
  data    = data_Hurdleztnegbin,
  method = "IRLS",
  controlModel = controlModel(
    piFormula = ~ x,
    alphaFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

expect_silent(summary(fit_Hurdleztnegbin, confint = TRUE, correlation = TRUE))

expect_silent(print(summary(fit_Hurdleztnegbin, confint = TRUE, correlation = TRUE)))

expect_equivalent(
  coef(fit_Hurdleztnegbin),
  coefficients_Hurdleztnegbin,
  tolerance = .7
)

expect_silent(
  population_Hurdleztnegbin <- popSizeEst(fit_Hurdleztnegbin)
)

expect_equivalent(
  population_Hurdleztnegbin$pointEstimate,
  N,
  tolerance = .35
)

expect_true(
  (population_Hurdleztnegbin$confidenceInterval[2, 1] < N) &
  (N < population_Hurdleztnegbin$confidenceInterval[2, 2])
)

expect_true(
  all(predict(
    fit_Hurdleztnegbin,
    type = "response",
    se.fit = TRUE
  )[, 4:6] > 0)
)

expect_true(
  all(predict(
    fit_Hurdleztnegbin,
    type = "link",
    se.fit = TRUE
  )[, 4:6]> 0)
)

expect_true(
  all(predict(
    fit_Hurdleztnegbin,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_Hurdleztnegbin,
    type = "contr"
  )),
  fit_Hurdleztnegbin$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_Hurdleztnegbin,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_Hurdleztnegbin,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 4:6] > 0)
)

expect_true(
  all(predict(
    fit_Hurdleztnegbin,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[, 4:6]> 0)
)

expect_true(
  all(predict(
    fit_Hurdleztnegbin,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_Hurdleztnegbin,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_Hurdleztnegbin,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

source("helper_families.R")
if (!developer_tests_enabled()) exit_file("developer tests disabled")

inflated_test_data <- load_inflated_test_data()
N <- inflated_test_data$N
x1 <- inflated_test_data$x1
test_inflated <- inflated_test_data$test_inflated

# Hurdlezt ####
# poisson
coefficients_Hurdleztpoisson <- c(.6, -.3, -.5, .1)
#eta <- cbind(coefficients_Hurdleztpoisson[1] + coefficients_Hurdleztpoisson[2] * x1, coefficients_Hurdleztpoisson[3] + coefficients_Hurdleztpoisson[4] * x1)

model_Hurdleztpoisson <- Hurdleztpoisson(
  lambdaLink = "log",
  piLink  = "logit",
)

#y <- model_Hurdleztpoisson$simulate(n = N, eta = eta, lower = -1)
observed_Hurdleztpoisson <- test_inflated$Hurdleztpoisson1

data_Hurdleztpoisson <- data.frame(
  x = x1[observed_Hurdleztpoisson > 0],
  y = observed_Hurdleztpoisson[observed_Hurdleztpoisson > 0]
)

fit_Hurdleztpoisson <- estimatePopsize(
  formula = y ~ x,
  model   = model_Hurdleztpoisson,
  data    = data_Hurdleztpoisson,
  method = "IRLS",
  controlModel = controlModel(
    piFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

expect_silent(summary(fit_Hurdleztpoisson))

expect_silent(print(summary(fit_Hurdleztpoisson)))

expect_equivalent(
  coef(fit_Hurdleztpoisson),
  coefficients_Hurdleztpoisson,
  tolerance = .6
)

expect_silent(
  population_Hurdleztpoisson <- popSizeEst(fit_Hurdleztpoisson)
)

expect_equivalent(
  population_Hurdleztpoisson$pointEstimate,
  N,
  tolerance = .25
)

expect_true(
  (population_Hurdleztpoisson$confidenceInterval[1, 1] < N) &
  (N < population_Hurdleztpoisson$confidenceInterval[1, 2])
)

expect_true(
  (population_Hurdleztpoisson$confidenceInterval[2, 1] < N) &
  (N < population_Hurdleztpoisson$confidenceInterval[2, 2])
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_Hurdleztpoisson,
    type = "contr"
  )),
  fit_Hurdleztpoisson$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_Hurdleztpoisson,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_Hurdleztpoisson,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_Hurdleztpoisson,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

# same with different link
model_Hurdleztpoisson <- Hurdleztpoisson(
  lambdaLink = "neglog",
  piLink  = "probit",
)

#y <- model_Hurdleztpoisson$simulate(n = N, eta = eta, lower = -1)
observed_Hurdleztpoisson <- test_inflated$Hurdleztpoisson2

data_Hurdleztpoisson <- data.frame(
  x = x1[observed_Hurdleztpoisson > 0],
  y = observed_Hurdleztpoisson[observed_Hurdleztpoisson > 0]
)

fit_Hurdleztpoisson <- estimatePopsize(
  formula = y ~ x,
  model   = model_Hurdleztpoisson,
  data    = data_Hurdleztpoisson,
  method = "IRLS",
  controlModel = controlModel(
    piFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE, stepsize = .3)
)

expect_silent(summary(fit_Hurdleztpoisson))

expect_silent(print(summary(fit_Hurdleztpoisson)))

expect_equivalent(
  coef(fit_Hurdleztpoisson),
  coefficients_Hurdleztpoisson,
  tolerance = .6
)

expect_silent(
  population_Hurdleztpoisson <- popSizeEst(fit_Hurdleztpoisson)
)

expect_equivalent(
  population_Hurdleztpoisson$pointEstimate,
  N,
  tolerance = .3
)

expect_true(
  (population_Hurdleztpoisson$confidenceInterval[1, 1] < N) &
  (N < population_Hurdleztpoisson$confidenceInterval[1, 2])
)

expect_true(
  (population_Hurdleztpoisson$confidenceInterval[2, 1] < N) &
  (N < population_Hurdleztpoisson$confidenceInterval[2, 2])
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_Hurdleztpoisson,
    type = "contr"
  )),
  fit_Hurdleztpoisson$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_Hurdleztpoisson,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_Hurdleztpoisson,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_Hurdleztpoisson,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_Hurdleztpoisson,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

source("helper_families.R")
if (!developer_tests_enabled()) exit_file("developer tests disabled")

inflated_test_data <- load_inflated_test_data()
N <- inflated_test_data$N
x1 <- inflated_test_data$x1
test_inflated <- inflated_test_data$test_inflated

# Ztoi ####

# ztoipoisson
coefficients_ztoipoisson <- c(.6, .2, -1.25, .1)
#eta <- cbind(coefficients_ztoipoisson[1] + coefficients_ztoipoisson[2] * x1, coefficients_ztoipoisson[3] + coefficients_ztoipoisson[4] * x1)

model_ztoipoisson <- ztoipoisson(
  lambdaLink = "log",
  omegaLink  = "logit",
)

#y <- model_ztoipoisson$simulate(n = N, eta = eta, lower = -1)
observed_ztoipoisson <- test_inflated$ztoipoisson1

data_ztoipoisson <- data.frame(
  x = x1[observed_ztoipoisson > 0],
  y = observed_ztoipoisson[observed_ztoipoisson > 0]
)

fit_ztoipoisson <- estimatePopsize(
  formula = y ~ x,
  model   = model_ztoipoisson,
  data    = data_ztoipoisson,
  method = "IRLS",
  controlModel = controlModel(
    omegaFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

expect_silent(summary(fit_ztoipoisson))

expect_silent(print(summary(fit_ztoipoisson)))

expect_equivalent(
  coef(fit_ztoipoisson),
  coefficients_ztoipoisson, tol = .3
)

expect_silent(
  population_ztoipoisson <- popSizeEst(fit_ztoipoisson)
)

expect_equivalent(
  population_ztoipoisson$pointEstimate,
  N, tol = .1
)

expect_true(
  (population_ztoipoisson$confidenceInterval[1, 1] < N) &
  (N < population_ztoipoisson$confidenceInterval[1, 2])
)

expect_true(
  (population_ztoipoisson$confidenceInterval[2, 1] < N) &
  (N < population_ztoipoisson$confidenceInterval[2, 2])
)

# test whether estimation of subpopulation sizes are 'good'
strata_ztoipoisson <- stratifyPopsize(fit_ztoipoisson, strata = ~ x)
strata_order_ztoipoisson <- 1 + as.numeric(substr(strata_ztoipoisson$name, start = 4, stop = 4))

expect_equivalent(
  strata_ztoipoisson$Estimated,
  as.numeric(table(x1))[strata_order_ztoipoisson],
  tolerance = .15
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_ztoipoisson,
    type = "contr"
  )),
  fit_ztoipoisson$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztoipoisson,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztoipoisson,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_ztoipoisson,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

# same with different link
model_ztoipoisson <- ztoipoisson(
  lambdaLink = "neglog",
  omegaLink  = "cloglog",
)

#y <- model_ztoipoisson$simulate(n = N, eta = eta, lower = -1)
observed_ztoipoisson <- test_inflated$ztoipoisson2

data_ztoipoisson <- data.frame(
  x = x1[observed_ztoipoisson > 0],
  y = observed_ztoipoisson[observed_ztoipoisson > 0]
)

fit_ztoipoisson <- estimatePopsize(
  formula = y ~ x,
  model   = model_ztoipoisson,
  data    = data_ztoipoisson,
  method = "IRLS",
  controlModel = controlModel(
    omegaFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

expect_silent(summary(fit_ztoipoisson))

expect_silent(print(summary(fit_ztoipoisson)))

expect_equivalent(
  coef(fit_ztoipoisson),
  coefficients_ztoipoisson,
  tolerance = .7
)

expect_silent(
  population_ztoipoisson <- popSizeEst(fit_ztoipoisson)
)

# expect_equivalent(
#   population_ztoipoisson$pointEstimate, N,
#   tolerance = .25
# )

expect_true(
  (population_ztoipoisson$confidenceInterval[1, 1] < N) &
  (N < population_ztoipoisson$confidenceInterval[1, 2])
)

expect_true(
  (population_ztoipoisson$confidenceInterval[2, 1] < N) &
  (N < population_ztoipoisson$confidenceInterval[2, 2])
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_silent(
  predict(
    fit_ztoipoisson,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztoipoisson,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztoipoisson,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_ztoipoisson,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

source("helper_families.R")
if (!developer_tests_enabled()) exit_file("developer tests disabled")

inflated_test_data <- load_inflated_test_data()
N <- inflated_test_data$N
x1 <- inflated_test_data$x1
test_inflated <- inflated_test_data$test_inflated

# ztHurdle ####

# poisson
coefficients_ztHurdlepoisson <- c(.6, -.3, -.5, .1)
#eta <- cbind(coefficients_ztHurdlepoisson[1] + coefficients_ztHurdlepoisson[2] * x1, coefficients_ztHurdlepoisson[3] + coefficients_ztHurdlepoisson[4] * x1)

model_ztHurdlepoisson <- ztHurdlepoisson(
  lambdaLink = "log",
  piLink  = "logit",
)

#y <- model_ztHurdlepoisson$simulate(n = N, eta = eta, lower = -1)
observed_ztHurdlepoisson <- test_inflated$ztHurdlepoisson1

data_ztHurdlepoisson <- data.frame(
  x = x1[observed_ztHurdlepoisson > 0],
  y = observed_ztHurdlepoisson[observed_ztHurdlepoisson > 0]
)

fit_ztHurdlepoisson <- estimatePopsize(
  formula = y ~ x,
  model   = model_ztHurdlepoisson,
  data    = data_ztHurdlepoisson,
  method = "IRLS",
  controlModel = controlModel(
    piFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

expect_silent(summary(fit_ztHurdlepoisson))

expect_silent(print(summary(fit_ztHurdlepoisson)))

expect_equivalent(
  coef(fit_ztHurdlepoisson),
  coefficients_ztHurdlepoisson,
  tolerance = .7
)

expect_silent(
  population_ztHurdlepoisson <- popSizeEst(fit_ztHurdlepoisson)
)

expect_equivalent(
  population_ztHurdlepoisson$pointEstimate,
  N,
  tolerance = .5
)

expect_true(
  (population_ztHurdlepoisson$confidenceInterval[1, 1] < N) &
  (N < population_ztHurdlepoisson$confidenceInterval[1, 2])
)

expect_true(
  (population_ztHurdlepoisson$confidenceInterval[2, 1] < N) &
  (N < population_ztHurdlepoisson$confidenceInterval[2, 2])
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_ztHurdlepoisson,
    type = "contr"
  )),
  fit_ztHurdlepoisson$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztHurdlepoisson,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztHurdlepoisson,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_ztHurdlepoisson,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

# same with different link
model_ztHurdlepoisson <- ztHurdlepoisson(
  lambdaLink = "neglog",
  piLink  = "probit",
)

#y <- model_ztHurdlepoisson$simulate(n = N, eta = eta, lower = -1)
observed_ztHurdlepoisson <- test_inflated$ztHurdlepoisson2

data_ztHurdlepoisson <- data.frame(
  x = x1[observed_ztHurdlepoisson > 0],
  y = observed_ztHurdlepoisson[observed_ztHurdlepoisson > 0]
)

fit_ztHurdlepoisson <- estimatePopsize(
  formula = y ~ x,
  model   = model_ztHurdlepoisson,
  data    = data_ztHurdlepoisson,
  method = "IRLS",
  controlModel = controlModel(
    piFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

expect_silent(summary(fit_ztHurdlepoisson))

expect_silent(print(summary(fit_ztHurdlepoisson)))

expect_equivalent(
  coef(fit_ztHurdlepoisson),
  coefficients_ztHurdlepoisson,
  tolerance = .4
)

expect_silent(
  population_ztHurdlepoisson <- popSizeEst(fit_ztHurdlepoisson)
)

expect_equivalent(
  population_ztHurdlepoisson$pointEstimate,
  N,
  tolerance = .4
)

expect_true(
  (population_ztHurdlepoisson$confidenceInterval[1, 1] < N) &
  (N < population_ztHurdlepoisson$confidenceInterval[1, 2])
)

expect_true(
  (population_ztHurdlepoisson$confidenceInterval[2, 1] < N) &
  (N < population_ztHurdlepoisson$confidenceInterval[2, 2])
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_ztHurdlepoisson,
    type = "contr"
  )),
  fit_ztHurdlepoisson$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztHurdlepoisson,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztHurdlepoisson,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztHurdlepoisson,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_ztHurdlepoisson,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

source("helper_families.R")
if (!developer_tests_enabled()) exit_file("developer tests disabled")

inflated_test_data <- load_inflated_test_data()
N <- inflated_test_data$N
x1 <- inflated_test_data$x1
test_inflated <- inflated_test_data$test_inflated

# ztoigeom
coefficients_ztoigeom <- c(.6, .2, -1.25, .1)
#eta <- cbind(coefficients_ztoigeom[1] + coefficients_ztoigeom[2] * x1, coefficients_ztoigeom[3] + coefficients_ztoigeom[4] * x1)

model_ztoigeom <- ztoigeom(
  lambdaLink = "log",
  omegaLink  = "logit",
)

#y <- model_ztoigeom$simulate(n = N, eta = eta, lower = -1)
observed_ztoigeom <- test_inflated$ztoigeom1

data_ztoigeom <- data.frame(
  x = x1[observed_ztoigeom > 0],
  y = observed_ztoigeom[observed_ztoigeom > 0]
)

fit_ztoigeom <- estimatePopsize(
  formula = y ~ x,
  model   = model_ztoigeom,
  data    = data_ztoigeom,
  method = "IRLS",
  controlModel = controlModel(
    omegaFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

expect_silent(summary(fit_ztoigeom))

expect_silent(print(summary(fit_ztoigeom)))

expect_equivalent(
  coef(fit_ztoigeom),
  coefficients_ztoigeom,
  tolerance = .7
)

expect_silent(
  population_ztoigeom <- popSizeEst(fit_ztoigeom)
)

expect_equivalent(
  population_ztoigeom$pointEstimate,
  N,
  tolerance = .2
)

expect_true(
  (population_ztoigeom$confidenceInterval[1, 1] < N) &
  (N < population_ztoigeom$confidenceInterval[1, 2])
)

expect_true(
  (population_ztoigeom$confidenceInterval[2, 1] < N) &
  (N < population_ztoigeom$confidenceInterval[2, 2])
)

expect_true(
  all(predict(
    fit_ztoigeom,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztoigeom,
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztoigeom,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_ztoigeom,
    type = "contr"
  )),
  fit_ztoigeom$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztoigeom,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_ztoigeom,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztoigeom,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztoigeom,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztoigeom,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_ztoigeom,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

# same for different link
model_ztoigeom <- ztoigeom(
  lambdaLink = "neglog",
  omegaLink  = "cloglog",
)

#y <- model_ztoigeom$simulate(n = N, eta = eta, lower = -1)
observed_ztoigeom <- test_inflated$ztoigeom2

data_ztoigeom <- data.frame(
  x = x1[observed_ztoigeom > 0],
  y = observed_ztoigeom[observed_ztoigeom > 0]
)

fit_ztoigeom <- estimatePopsize(
  formula = y ~ x,
  model   = model_ztoigeom,
  data    = data_ztoigeom,
  method = "IRLS",
  controlModel = controlModel(
    omegaFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

# TODO: re-enable after investigating platform/coverage-specific NaN warnings.
# expect_silent(summary(fit_ztoigeom))
#
# expect_silent(print(summary(fit_ztoigeom)))

expect_silent(
  population_ztoigeom <- popSizeEst(fit_ztoigeom)
)

expect_equivalent(
  population_ztoigeom$pointEstimate,
  N,
  tolerance = .2
)

expect_true(
  (population_ztoigeom$confidenceInterval[1, 1] < N) &
  (N < population_ztoigeom$confidenceInterval[1, 2])
)

expect_true(
  (population_ztoigeom$confidenceInterval[2, 1] < N) &
  (N < population_ztoigeom$confidenceInterval[2, 2])
)

expect_true(
  all(predict(
    fit_ztoigeom,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztoigeom,
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztoigeom,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_ztoigeom,
    type = "contr"
  )),
  fit_ztoigeom$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztoigeom,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_ztoigeom,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztoigeom,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztoigeom,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztoigeom,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_ztoigeom,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

source("helper_families.R")
if (!developer_tests_enabled()) exit_file("developer tests disabled")

inflated_test_data <- load_inflated_test_data()
N <- inflated_test_data$N
x1 <- inflated_test_data$x1
test_inflated <- inflated_test_data$test_inflated

# oizt ####

# poisson
coefficients_oiztpoisson <- c(.6, -.3, -.5, .1)
#eta <- cbind(coefficients_oiztpoisson[1] + coefficients_oiztpoisson[2] * x1, coefficients_oiztpoisson[3] + coefficients_oiztpoisson[4] * x1)

model_oiztpoisson <- oiztpoisson(
  lambdaLink = "log",
  omegaLink  = "logit",
)

#y <- model_oiztpoisson$simulate(n = N, eta = eta, lower = -1)
observed_oiztpoisson <- test_inflated$oiztpoisson1

data_oiztpoisson <- data.frame(
  x = x1[observed_oiztpoisson > 0],
  y = observed_oiztpoisson[observed_oiztpoisson > 0]
)

fit_oiztpoisson <- estimatePopsize(
  formula = y ~ x,
  model   = model_oiztpoisson,
  data    = data_oiztpoisson,
  method = "IRLS",
  controlModel = controlModel(
    omegaFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

expect_silent(summary(fit_oiztpoisson))

expect_silent(print(summary(fit_oiztpoisson)))

# expect_equivalent(
#   coef(fit_oiztpoisson),
#   coefficients_oiztpoisson,
#   tolerance = .25
# )

expect_silent(
  population_oiztpoisson <- popSizeEst(fit_oiztpoisson)
)

# expect_equivalent(
#   population_oiztpoisson$pointEstimate,
#   N,
#   tolerance = .2
# )

# expect_true(
#   (population_oiztpoisson$confidenceInterval[1, 1] < N) &
#   (N < population_oiztpoisson$confidenceInterval[1, 2])
# )

# expect_true(
#   (population_oiztpoisson$confidenceInterval[2, 1] < N) &
#   (N < population_oiztpoisson$confidenceInterval[2, 2])
# )

expect_true(
  all(predict(
    fit_oiztpoisson,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_oiztpoisson,
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_oiztpoisson,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_oiztpoisson,
    type = "contr"
  )),
  fit_oiztpoisson$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_oiztpoisson,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_oiztpoisson,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_oiztpoisson,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_oiztpoisson,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_oiztpoisson,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_oiztpoisson,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)


# same with different link
model_oiztpoisson <- oiztpoisson(
  lambdaLink = "neglog",
  omegaLink  = "cloglog",
)

#y <- model_oiztpoisson$simulate(n = N, eta = eta, lower = -1)
observed_oiztpoisson <- test_inflated$oiztpoisson2

data_oiztpoisson <- data.frame(
  x = x1[observed_oiztpoisson > 0],
  y = observed_oiztpoisson[observed_oiztpoisson > 0]
)

fit_oiztpoisson <- estimatePopsize(
  formula = y ~ x,
  model   = model_oiztpoisson,
  data    = data_oiztpoisson,
  method = "IRLS",
  controlModel = controlModel(
    omegaFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

expect_silent(summary(fit_oiztpoisson))

expect_silent(print(summary(fit_oiztpoisson)))

# expect_equivalent(
#   coef(fit_oiztpoisson),
#   coefficients_oiztpoisson,
#   tolerance = .5
# )

expect_silent(
  population_oiztpoisson <- popSizeEst(fit_oiztpoisson)
)

# expect_equivalent(
#   population_oiztpoisson$pointEstimate,
#   N,
#   tolerance = .2
# )

expect_true(
  (population_oiztpoisson$confidenceInterval[1, 1] < N) &
  (N < population_oiztpoisson$confidenceInterval[1, 2])
)

# expect_true(
#   (population_oiztpoisson$confidenceInterval[2, 1] < N) &
#   (N < population_oiztpoisson$confidenceInterval[2, 2])
# )

expect_true(
  all(predict(
    fit_oiztpoisson,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_oiztpoisson,
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_oiztpoisson,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_oiztpoisson,
    type = "contr"
  )),
  fit_oiztpoisson$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_oiztpoisson,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_oiztpoisson,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_oiztpoisson,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[,3:4]> 0)
)

expect_true(
  all(predict(
    fit_oiztpoisson,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_oiztpoisson,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_oiztpoisson,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

source("helper_families.R")
if (!developer_tests_enabled()) exit_file("developer tests disabled")

inflated_test_data <- load_inflated_test_data()
N <- inflated_test_data$N
x1 <- inflated_test_data$x1
test_inflated <- inflated_test_data$test_inflated

# negbin
coefficients_oiztnegbin <- c(.4, .4, -.4, .2, -1.1, .37)
#eta <- cbind(coefficients_oiztnegbin[1] + coefficients_oiztnegbin[2] * x1, coefficients_oiztnegbin[3] + coefficients_oiztnegbin[4] * x1, coefficients_oiztnegbin[5] + coefficients_oiztnegbin[6] * x1)

model_oiztnegbin <- oiztnegbin(
  lambdaLink = "log",
  alphaLink  = "log",
  omegaLink  = "logit",
)

#y <- model_oiztnegbin$simulate(n = N, eta = eta, lower = -1)
observed_oiztnegbin <- test_inflated$oiztnegbin1

data_oiztnegbin <- data.frame(
  x = x1[observed_oiztnegbin > 0],
  y = observed_oiztnegbin[observed_oiztnegbin > 0]
)

fit_oiztnegbin <- estimatePopsize(
  formula = y ~ x,
  model   = model_oiztnegbin,
  data    = data_oiztnegbin,
  method = "IRLS",
  controlModel = controlModel(
    omegaFormula = ~ x,
    alphaFormula = ~ x
  ),
  controlMethod = controlMethod(silent = TRUE)
)

expect_silent(summary(fit_oiztnegbin, confint = TRUE, correlation = TRUE))

expect_silent(print(summary(fit_oiztnegbin, confint = TRUE, correlation = TRUE)))

expect_equivalent(
  coef(fit_oiztnegbin),
  coefficients_oiztnegbin,
  tolerance = .6
)

expect_silent(
  population_oiztnegbin <- popSizeEst(fit_oiztnegbin)
)

expect_equivalent(
  population_oiztnegbin$pointEstimate,
  N,
  tolerance = .2
)

expect_true(
  (population_oiztnegbin$confidenceInterval[1, 1] < N) &
  (N < population_oiztnegbin$confidenceInterval[1, 2])
)

expect_true(
  (population_oiztnegbin$confidenceInterval[2, 1] < N) &
  (N < population_oiztnegbin$confidenceInterval[2, 2])
)

expect_true(
  all(predict(
    fit_oiztnegbin,
    type = "response",
    se.fit = TRUE
  )[, 4:6] > 0)
)

expect_true(
  all(predict(
    fit_oiztnegbin,
    type = "link",
    se.fit = TRUE
  )[, 4:6]> 0)
)

expect_true(
  all(predict(
    fit_oiztnegbin,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_equivalent(
  sum(predict(
    fit_oiztnegbin,
    type = "contr"
  )),
  fit_oiztnegbin$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_oiztnegbin,
    type = "popSize",
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_oiztnegbin,
    newdata = data.frame(x = x1),
    type = "response",
    se.fit = TRUE
  )[, 4:6] > 0)
)

expect_true(
  all(predict(
    fit_oiztnegbin,
    newdata = data.frame(x = x1),
    type = "link",
    se.fit = TRUE
  )[, 4:6]> 0)
)

expect_true(
  all(predict(
    fit_oiztnegbin,
    newdata = data.frame(x = x1),
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_oiztnegbin,
    newdata = data.frame(x = x1),
    type = "contr"
  )) > N
)

expect_silent(
  predict(
    fit_oiztnegbin,
    newdata = data.frame(x = x1),
    type = "popSize",
    se.fit = TRUE
  )
)

#' @title Control parameters for regression
#' @author Piotr Chlebicki, Maciej Beręsewicz
#' \loadmathjax
#' 
#' @description \code{controlMethod} constructs a list with all necessary 
#' control parameters for regression fitting in 
#' \code{estimatePopsizeFit} and \code{estimatePopsize}.
#'
#' @param epsilon a tolerance level for fitting algorithms by default \code{1e-8}.
#' @param maxiter a maximum number of iterations.
#' @param verbose a numeric value indicating whether to trace steps of fitting algorithm for 
#' \code{IRLS} fitting method different values of verbose give the following information:
#' \itemize{
#'   \item 1 -- Returns information on the number of current 
#'   iteration and current log-likelihood.
#'   \item 2 -- Returns information on vector of regression parameters 
#'   at current iteration (and all of the above).
#'   \item 3 -- Returns information on reduction of log-likelihood 
#'   at current iteration (and all of the above).
#'   \item 4 -- Returns information on value of log-likelihood function gradient 
#'   at current iteration (and all of the above).
#'   \item 5 -- Returns information on convergence criterion and values that are 
#'   taken into account when considering convergence (and all of the above).
#' }
#' if \code{optim} method was chosen verbose will be passed to [stats::optim()] as trace.
#' @param printEveryN an integer value indicating how often to print information
#' specified in \code{verbose}, by default set to \code{1}.
#' @param coefStart,etaStart initial parameters for regression coefficients
#' or linear predictors if \code{NULL}. For \code{IRLS} fitting only \code{etaStart}
#' is needed so if \code{coefStart} is provided it will be converted to \code{etaStart},
#' for \code{optim} fitting \code{coefStart} is necessary and argument \code{etaStart}
#' will be ignored.
#' @param silent a logical value, indicating whether warnings in \code{IRLS} method should be suppressed.
#' @param optimPass an optional list of parameters passed to \code{stats::optim(..., control = optimPass)}
#' if FALSE then list of control parameters will be inferred from other parameters.
#' @param optimMethod a method of [stats::optim()] used  \code{"Nelder-Mead"} is the default .
#' @param stepsize only for \code{IRLS}, scaling of updates to \code{beta} vector 
#' lower value means slower convergence but more accuracy by default 1. 
#' In general if fitting algorithm fails lowering this value tends to 
#' be most effective at correcting it.
#' @param checkDiagWeights a logical value indicating whether to check if diagonal 
#' elements of working weights matrixes in \code{IRLS} are sufficiently positive 
#' so that these matrixes are positive defined. By default \code{TRUE}.
#' @param weightsEpsilon a small number to ensure positive definedness of weights matrixes. 
#' Only matters if \code{checkDiagWeights} is set to \code{TRUE}. 
#' By default \code{1e-8}.
#' @param momentumFactor an experimental parameter in \code{IRLS} only allowing for 
#' taking previous step into account at current step, i.e instead of 
#' updating regression parameters as:
#' \mjsdeqn{\boldsymbol{\beta}_{(a)} = 
#' \boldsymbol{\beta}_{(a-1)} + \text{stepsize} \cdot \text{step}_{(a)}}
#' the update will be made as:
#' \mjsdeqn{
#' \boldsymbol{\beta}_{(a)} = \boldsymbol{\beta}_{(a-1)} + \text{stepsize} 
#' \cdot (\text{step}_{(a)} + \text{momentum}\cdot\text{step}_{(a-1)})}
#' @param momentumActivation the value of log-likelihood reduction below 
#' which momentum will apply.
#' @param criterion a criterion used to determine convergence in \code{IRLS}, 
#' multiple values may be provided. By default \code{c("coef", "abstol")}.
#' @param saveIRLSlogs a logical value indicating if information specified in
#' \code{verbose} should be saved to output object, by default \code{FALSE}.
#'
#' @return List with selected parameters, it is also possible to call list directly.
#' @seealso [singleRcapture::estimatePopsize()] [singleRcapture::estimatePopsizeFit()] 
#' [singleRcapture::controlModel()] [singleRcapture::controlPopVar()]
#' @export
controlMethod <- function(epsilon             = 1e-8,
                          maxiter             = 1000,
                          verbose             = 0,
                          printEveryN         = 1L,
                          coefStart           = NULL,
                          etaStart            = NULL,
                          optimMethod         = "Nelder-Mead",
                          silent              = FALSE,
                          optimPass           = FALSE,
                          stepsize            = 1,
                          checkDiagWeights    = TRUE,
                          weightsEpsilon      = 1e-8,
                          momentumFactor      = 0,
                          saveIRLSlogs        = FALSE,
                          momentumActivation  = 5,
                          criterion           = c("coef", 
                                                  "abstol", 
                                                  "reltol")) {
  
  if (!missing(criterion) && all(c("abstol", "reltol") %in% criterion)) 
    stop("Choosing both absolute tolerance and relative tolerance for convergence criterion is not allowed.")
  
  if (!missing(criterion) && all(!(c("coef", "abstol", "reltol") %in% criterion))) 
    stop("At least one convergence criterion has to be chosen")
  
  if (!isTRUE(is.finite(epsilon)) || !isTRUE(0 < epsilon) || isTRUE(length(epsilon) > 1))
    stop("Argument epsilon has to be a positive numeric value (of length 1).")
  
  if (!isTRUE(is.finite(maxiter)) || !isTRUE(0 < maxiter) || isTRUE(length(maxiter) > 1))
    stop("Argument maxiter has to be a positive numeric value (of length 1).")
  
  if (!isTRUE(is.finite(verbose)) || isTRUE(length(verbose) > 1))
    stop("Argument verbose has to be a numeric value (of length 1).")
  
  if (!isTRUE(is.finite(printEveryN)) || !isFALSE(0 > printEveryN) || isTRUE(length(printEveryN) > 1)) {
    stop("Argument printEveryN has to be a nonnegative numeric value (of length 1).")
    # If it's not an integer
    printEveryN <- as.integer(printEveryN)
  }
  
  if (!is.null(etaStart) && !isTRUE(is.numeric(etaStart)))
    stop("Argument etaStart has to be either a numeric vector or NULL.")
  
  if (!is.null(coefStart) && !isTRUE(is.numeric(coefStart)))
    stop("Argument coefStart has to be either a numeric vector or NULL.")
  
  if (!isTRUE(is.logical(silent)) || isTRUE(length(silent) > 1))
    stop("Argument silent should be logical value (of length 1).")
  
  if (!isTRUE(is.logical(checkDiagWeights)) || isTRUE(length(checkDiagWeights) > 1))
    stop("Argument checkDiagWeights should be logical value (of length 1).")
  
  if (!isTRUE(is.logical(saveIRLSlogs)) || isTRUE(length(saveIRLSlogs) > 1))
    stop("Argument saveIRLSlogs should be logical value (of length 1).")
  
  if (!isTRUE(is.finite(stepsize)) || !isTRUE(0 < stepsize) || isTRUE(length(stepsize) > 1))
    stop("Argument stepsize has to be a positive numeric value (of length 1).")
  
  if (!isTRUE(is.finite(momentumActivation)) || !isTRUE(0 < momentumActivation) || 
       isTRUE(length(momentumActivation) > 1))
    stop("Argument momentumActivation has to be a positive numeric value (of length 1).")
  
  if (!isTRUE(is.finite(momentumFactor)) || !isFALSE(0 > momentumFactor) || 
       isTRUE(length(momentumFactor) > 1))
    stop("Argument momentumFactor has to be a non negative numeric value (of length 1).")
  
  list(
    epsilon             = epsilon,
    maxiter             = maxiter,
    verbose             = verbose,
    printEveryN         = printEveryN,
    etaStart            = etaStart,
    coefStart           = coefStart,
    optimMethod         = optimMethod,
    silent              = silent,
    optimPass           = optimPass,
    stepsize            = stepsize,
    checkDiagWeights    = checkDiagWeights,
    weightsEpsilon      = weightsEpsilon,
    momentumFactor      = momentumFactor,
    momentumActivation  = momentumActivation,
    saveIRLSlogs        = saveIRLSlogs,
    criterion           = if (missing(criterion)) c("coef", "abstol") else criterion
  )
}
#' @title Control parameters specific to some models
#' @author Piotr Chlebicki, Maciej Beręsewicz
#' 
#' @description \code{controlModel} constructs a list with all necessary 
#' control parameters in \code{estimatePopsize} that are either specific to 
#' selected model or do not fit anywhere else.
#' 
#' Specifying additional formulas should be done by using only right hand side of
#' the formula also for now all variables from additional formulas should also be
#' included in the "main" formula.
#' 
#' @param weightsAsCounts a boolean value indicating whether to treat \code{weights}
#' argument as number of occurrences for each row in the \code{data} and adjust
#' necessary methods and functionalities, like adjustments in bootstrap or
#' decreasing weights in \code{dfbeta} instead or deleting rows from data, 
#' to accommodate this form of model specification.
#' @param omegaFormula a formula for inflation parameter in one inflated zero 
#' truncated and zero truncated one inflated models.
#' @param alphaFormula a formula for dispersion parameter in negative binomial
#' based models.
#' @param piFormula a formula for probability parameter in pseudo hurdle zero 
#' truncated and zero truncated pseudo hurdle models.
#' 
#' @return A list with selected parameters, it is also possible to call list directly.
#' @seealso [singleRcapture::estimatePopsize()] [singleRcapture::controlMethod()] [singleRcapture::controlPopVar()] [singleRcapture::singleRmodels()]
#' @export
controlModel <- function(weightsAsCounts = FALSE,
                         omegaFormula = ~ 1,
                         alphaFormula = ~ 1,
                         piFormula = ~ 1) {
  list(
    weightsAsCounts = weightsAsCounts,
    omegaFormula    = omegaFormula,
    alphaFormula    = alphaFormula,
    piFormula       = piFormula
  )
}
#' @title  Control parameters for population size estimation
#' @author Piotr Chlebicki, Maciej Beręsewicz
#' \loadmathjax
#'
#' @description Creating control parameters for population size estimation and 
#' respective standard error and variance estimation.
#'
#' @param alpha a significance level, 0.05 used by default.
#' @param cores for bootstrap only, a number of processor cores to be used,
#' any number greater than 1 activates code designed with \code{doParallel}, 
#' \code{foreach} and \code{parallel} packages. Note that for now using parallel
#' computing makes tracing impossible so \code{traceBootstrapSize} and 
#' \code{bootstrapVisualTrace} parameters are ignored in this case.
#' @param bootType the bootstrap type to be used. Default is \code{"parametric"}, 
#' other possible values are: \code{"semiparametric"} and \code{"nonparametric"}.
#' @param B a number of bootstrap samples to be performed (default 500).
#' @param confType a type of confidence interval for bootstrap confidence interval, 
#' \code{"percentile"} by default. 
#' Other possibilities: \code{"studentized"} and \code{"basic"}.
#' @param keepbootStat a boolean value indicating whether to keep a vector of 
#' statistics produced by bootstrap.
#' @param traceBootstrapSize a boolean value indicating whether to print size of 
#' bootstrapped sample after truncation for semi- and fully parametric bootstraps.
#' @param bootstrapVisualTrace a boolean value indicating whether to plot bootstrap 
#' statistics in real time if \code{cores = 1} if \code{cores > 1} it instead
#' indicates whether to make progress bar.
#' @param fittingMethod a method used for fitting models from bootstrap samples.
#' @param bootstrapFitcontrol control parameters for each regression works exactly 
#' like \code{controlMethod} but for fitting models from bootstrap samples.
#' @param sd a character indicating how to compute standard deviation of population 
#' size estimator either as:
#' \mjsdeqn{\hat{\sigma}=\sqrt{\hat{\text{var}}(\hat{N})}}
#' for \code{sqrt} (which is slightly biased if \mjseqn{\hat{N}}
#' has a normal distribution) or for \code{normalMVUE} as the unbiased 
#' minimal variance estimator for normal distribution:
#' \mjsdeqn{\hat{\sigma}=\sqrt{\hat{\text{var}}(\hat{N})}
#' \frac{\Gamma\left(\frac{N_{obs}-1}{2}\right)}{\Gamma\left(\frac{N_{obs}}{2}\right)}
#' \sqrt{\frac{N_{obs}}{2}}}
#' where the ration involving gamma functions is computed by log gamma function.
#' @param covType a type of covariance matrix for regression parameters by default 
#' observed information matrix.
#'
#' @return A list with selected parameters, it is also possible to call list directly.
#' @seealso [singleRcapture::estimatePopsize()] [singleRcapture::controlModel()] [singleRcapture::controlMethod()]
#' @export
controlPopVar <- function(alpha = .05,
                          bootType = c("parametric",
                                       "semiparametric",
                                       "nonparametric"),
                          B = 500,
                          confType = c("percentile",
                                       "normal",
                                       "basic"),
                          keepbootStat = TRUE,
                          traceBootstrapSize = FALSE,
                          bootstrapVisualTrace = FALSE,
                          fittingMethod = c("optim", "IRLS"),
                          bootstrapFitcontrol = NULL,
                          sd = c("sqrtVar", "normalMVUE"),
                          covType = c("observedInform", "Fisher"),
                          cores = 1L) {
  
  if (missing(fittingMethod)) fittingMethod <- "IRLS"
  if (missing(bootType)) bootType <- "parametric"
  if (missing(confType)) confType <- "percentile"
  if (missing(covType))  covType <- "observedInform"
  if (missing(sd))       sd <- "sqrtVar"
  
  # Null's are not supposed to pass the tests
  if (!isTRUE(is.finite(alpha)) || isTRUE(0 > alpha || 1 < alpha) || isTRUE(length(alpha) > 1))
    stop("Argument alpha should be a numeric value between 0 and 1 (of length 1).")
  
  if (!isTRUE(sd %in% c("sqrtVar", "normalMVUE")) && !missing(sd))
    stop("Argument sd should be a character with value either sqrtVar or normalMVUE.")
  
  if ((!isTRUE(covType %in% c("observedInform", "Fisher")) && !missing(covType)) || isTRUE(length(covType) > 1))
    stop("Argument covType should be a character with value either observedInform or Fisher.")
  
  
  if (!missing(bootType) && !isTRUE(bootType %in% c("parametric",
                                                    "semiparametric",
                                                    "nonparametric")))
    stop("Argument bootType should be a character with value either in c(parametric, semiparametric, nonparametric).")
  
  if (!isTRUE(is.finite(B)) || isTRUE(B < 0) || isTRUE(length(B) > 1))
    stop("Argument B should be a numeric value greater than 0 (of length 1).")
  
  if (!isTRUE(all(c(is.logical(traceBootstrapSize), is.logical(bootstrapVisualTrace)))) || 
       isTRUE(length(bootstrapVisualTrace) > 1) || isTRUE(length(traceBootstrapSize) > 1))
    stop("Arguments traceBootstrapSize and bootstrapVisualTrace should be logical values (of length 1).")
    
    
  list(
    bootstrapVisualTrace = bootstrapVisualTrace,
    bootstrapFitcontrol  = bootstrapFitcontrol,
    traceBootstrapSize   = traceBootstrapSize,
    fittingMethod        = fittingMethod,
    keepbootStat         = keepbootStat,
    bootType             = bootType,
    confType             = confType,
    covType              = covType,
    cores                = cores,
    alpha                = alpha,
    B                    = B,
    sd                   = sd
  )
}

# These tests are only supposed to be run on developer's machine and
# package GitHub page not on CRAN (they take too long)

source("helper_families.R")
if (!developer_tests_enabled()) exit_file("developer tests disabled")

expect_error(
  estimatePopsize(
    formula = TOTAL_SUB ~ .,
    data = farmsubmission,
    model = "zotpoisson",
    method = "maxLik"
  )
)

expect_silent(
  xx <- estimatePopsize(
    formula = TOTAL_SUB ~ .,
    data = farmsubmission,
    model = "zotpoisson",
    controlMethod = controlMethod(
      epsilon = 1e-6, silent = TRUE
    )# testing silent
  )
)

expect_silent(
  summary(marginalFreq(xx), dropl5 = "group")
)

set.seed(123)
expect_silent(
  xx <- estimatePopsize(
    formula = TOTAL_SUB ~ .,
    data = farmsubmission,
    model = "zotpoisson",
    popVar = "bootstrap",
    controlMethod = controlMethod(epsilon = 1e-6, silent = TRUE),# testing silent
    controlPopVar = controlPopVar(
      B = 140,
      bootType = "parametric",
      bootstrapFitcontrol = controlMethod(
        silent = TRUE,
        epsilon = .Machine$double.eps
      ),
      cores = 2L,
      confType = "normal"
    )
  )
)

expect_silent(
  predict(
    xx,
    type = "mean",
    se.fit = TRUE
  )
)

expect_warning(
  xx <- estimatePopsize(
    formula = TOTAL_SUB ~ log_size,
    data = farmsubmission,
    model = "zotnegbin",
    controlModel  = controlModel(alphaFormula = ~ C_TYPE),
    controlMethod = controlMethod(
      silent = TRUE, stepsize = .7,
      verbose = 4, maxiter = 20
    )
  )
)

expect_silent(
  predict(
    xx,
    type = "mean",
    se.fit = TRUE
  )
)

expect_silent(
  predict(
    xx,
    type = "response",
    se.fit = TRUE
  )
)

expect_silent(
  predict(
    xx,
    type = "link",
    se.fit = TRUE
  )
)

set.seed(123)
expect_silent(
  xx <- estimatePopsize(
    formula = capture ~ gender,
    data = netherlandsimmigrant,
    model = "zotgeom",
    method = "optim",
    popVar = "bootstrap",
    controlMethod = controlMethod(epsilon = 1e-6, silent = TRUE),# testing silent
    controlPopVar = controlPopVar(
      B = 70,
      cores = 2L,
      bootstrapFitcontrol = controlMethod(),
      bootType = "nonparametric",
      confType = "basic"
    )
  )
)


expect_silent(
  dfbeta(xx)
)

expect_silent(
  dfbeta(xx, cores = 2L)
)

expect_silent(
  predict(
    xx,
    type = "mean",
    se.fit = TRUE
  )
)

expect_silent(
  estimatePopsize(
    formula = TOTAL_SUB ~ .,
    data = farmsubmission,
    model = "oiztgeom",
    popVar = "bootstrap",
    controlModel = controlModel(omegaFormula = ~ .),
    controlMethod = controlMethod(silent = TRUE),# testing silent
    controlPopVar = controlPopVar(
      B = 35,
      cores = 2L,
      bootstrapFitcontrol = controlMethod(),
      bootType = "semiparametric"
    )
  )
)

df <- farmsubmission[, c(1,4)]
df$ww <- 0
### this is dplyr::count but slower and without dependencies
df <- aggregate(ww ~ ., df, FUN = length)

expect_warning(
  xx <- estimatePopsize(
    formula = TOTAL_SUB ~ C_TYPE,
    data = df,
    model = ztpoisson,
    popVar = "bootstrap",
    weights = df$ww,
    controlMethod = controlMethod(
      verbose = 5,
      saveIRLSlogs = TRUE,
      criterion = "reltol"
    ),
    controlModel = controlModel(weightsAsCounts = TRUE),
    controlPopVar = controlPopVar(
      B = 70,
      cores = 2L,
      bootType = "semiparametric"
    )
  )
)

expect_true(
  !is.null(
    xx$fittingLog
  )
)

expect_silent(
  dfbeta(xx)
)

expect_silent(
  dfbeta(xx, cores = 2L)
)

expect_silent(
  estimatePopsize(
    formula = TOTAL_SUB ~ C_TYPE,
    data = df,
    model = ztoipoisson,
    popVar = "bootstrap",
    weights = df$ww,
    controlMethod = controlMethod(silent = TRUE),
    controlModel = controlModel(weightsAsCounts = TRUE),
    controlPopVar = controlPopVar(
      B = 70,
      cores = 2L,
      bootType = "parametric"
    )
  )
)

expect_silent(
  estimatePopsize(
    formula = TOTAL_SUB ~ C_TYPE,
    data = df,
    model = ztoigeom,
    popVar = "bootstrap",
    weights = df$ww,
    controlMethod = controlMethod(silent = TRUE),
    controlModel = controlModel(weightsAsCounts = TRUE),
    controlPopVar = controlPopVar(
      B = 70,
      cores = 2L,
      bootType = "nonparametric"
    )
  )
)

expect_silent(
  estimatePopsize(
    formula = TOTAL_SUB ~ C_TYPE,
    data = df,
    model = chao,
    popVar = "bootstrap",
    weights = df$ww,
    controlMethod = controlMethod(silent = TRUE),
    controlModel = controlModel(weightsAsCounts = TRUE),
    controlPopVar = controlPopVar(
      B = 70,
      bootType = "nonparametric"
    )
  )
)

expect_silent(
  estimatePopsize(
    formula = TOTAL_SUB ~ C_TYPE,
    data = df,
    model = zelterman,
    popVar = "bootstrap",
    weights = df$ww,
    controlMethod = controlMethod(silent = TRUE),
    controlModel = controlModel(weightsAsCounts = TRUE),
    controlPopVar = controlPopVar(
      B = 70,
      bootType = "semiparametric"
    )
  )
)

expect_silent(
  estimatePopsize(
    formula = TOTAL_SUB ~ C_TYPE,
    data = df,
    model = ztgeom,
    popVar = "bootstrap",
    weights = df$ww,
    controlMethod = controlMethod(silent = TRUE),
    controlModel = controlModel(weightsAsCounts = TRUE),
    controlPopVar = controlPopVar(
      B = 70,
      bootType = "parametric"
    )
  )
)

expect_error(
  controlPopVar(
    alpha = -.5
  )
)

expect_error(
  controlPopVar(
    alpha = c(.2, .5)
  )
)

expect_error(
  controlPopVar(
    sd = "abc"
  )
)

expect_error(
  controlPopVar(
    covType = "abc"
  )
)

expect_error(
  controlPopVar(
    bootType = "abc"
  )
)

expect_error(
  controlPopVar(
    B = -5
  )
)

expect_error(
  controlPopVar(
    traceBootstrapSize = -1
  )
)


expect_error(
  controlMethod(
    criterion = "abc"
  )
)

expect_error(
  controlMethod(
    criterion = c("abstol", "reltol")
  )
)

expect_error(
  controlMethod(
    epsilon = -1
  )
)

expect_error(
  controlMethod(
    maxiter = -1
  )
)

expect_error(
  controlMethod(
    verbose = c(2, 5)
  )
)

expect_silent(
  controlMethod(
    printEveryN = 2.5
  )
)

expect_error(
  controlMethod(
    etaStart = "abc"
  )
)

expect_error(
  controlMethod(
    coefStart = "abc"
  )
)

expect_error(
  controlMethod(
    silent = "abc"
  )
)

expect_error(
  controlMethod(
    checkDiagWeights = "abc"
  )
)

expect_error(
  controlMethod(
    saveIRLSlogs = "abc"
  )
)

expect_error(
  controlMethod(
    stepsize = -1
  )
)

expect_error(
  controlMethod(
    momentumFactor = -5
  )
)

expect_error(
  controlMethod(
    momentumActivation = -1
  )
)

# Testing on generated data for inflated and hurdle models
# These tests are only supposed to be run on developer's machine and 
# package GitHub page not on CRAN (they take too long)

if (isTRUE(tolower(Sys.getenv("TEST_SINGLERCAPTURE_MULTICORE_DEVELOPER")) == "true")) {
    
  N <- 500
  #x1 <- rbinom(n = N, size = 3, prob = .4)
  x1 <- rep(0:3, c(99, 228, 137, 36))
  ## generated data from distributions
  test_inflated <- read.csv("test_inflated.csv")
  #test_inflated <- read.csv("inst/tinytest/test_inflated.csv")
  
  # Ztoi ####
  
  # ztoipoisson
  beta <- c(.6, .2, -1.25, .1)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1)
  
  fn <- ztoipoisson(
    lambdaLink = "log",
    omegaLink  = "logit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$ztoipoisson1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      omegaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1), 
    beta, tol = .3
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate, 
    N, tol = .1
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  # test whether estimation of subpopulation sizes are 'good'
  XX <- stratifyPopsize(M1, strata = ~ x)
  order <- 1 + as.numeric(substr(XX$name, start = 4, stop = 4))
  
  expect_equivalent(
    XX$Estimated,
    as.numeric(table(x1))[order],
    tolerance = .15
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # same with different link
  fn <- ztoipoisson(
    lambdaLink = "neglog",
    omegaLink  = "cloglog",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$ztoipoisson2
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      omegaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .7
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  # expect_equivalent(
  #   pop$pointEstimate, N,
  #   tolerance = .25
  # )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # ztoigeom
  beta <- c(.6, .2, -1.25, .1)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1)
  
  fn <- ztoigeom(
    lambdaLink = "log",
    omegaLink  = "logit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$ztoigeom1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      omegaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .7
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .2
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # same for different link
  fn <- ztoigeom(
    lambdaLink = "neglog",
    omegaLink  = "cloglog",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$ztoigeom2
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      omegaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .2
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # ztoinegbin
  beta <- c(.6, .2, -1.25, .1, -1, -.3)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1, beta[5] + beta[6] * x1)
  
  fn <- ztoinegbin(
    lambdaLink = "log",
    alphaLink  = "log", 
    omegaLink  = "logit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$ztoinegbin1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      omegaFormula = ~ x,
      alphaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  # don't change this
  expect_true(
    sum(abs(coef(M1) - beta)) < 8
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .3
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 4:6] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[, 4:6]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 4:6] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[, 4:6]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # oizt ####
  
  # poisson
  beta <- c(.6, -.3, -.5, .1)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1)
  
  fn <- oiztpoisson(
    lambdaLink = "log",
    omegaLink  = "logit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$oiztpoisson1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      omegaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  # expect_equivalent(
  #   coef(M1),
  #   beta,
  #   tolerance = .25
  # )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  # expect_equivalent(
  #   pop$pointEstimate,
  #   N,
  #   tolerance = .2
  # )
  
  # expect_true(
  #   (pop$confidenceInterval[1, 1] < N) &
  #   (N < pop$confidenceInterval[1, 2])
  # )
  
  # expect_true(
  #   (pop$confidenceInterval[2, 1] < N) &
  #   (N < pop$confidenceInterval[2, 2])
  # )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  
  # same with different link
  fn <- oiztpoisson(
    lambdaLink = "neglog",
    omegaLink  = "cloglog",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$oiztpoisson2
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      omegaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  # expect_equivalent(
  #   coef(M1),
  #   beta,
  #   tolerance = .5
  # )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  # expect_equivalent(
  #   pop$pointEstimate,
  #   N,
  #   tolerance = .2
  # )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  # expect_true(
  #   (pop$confidenceInterval[2, 1] < N) &
  #   (N < pop$confidenceInterval[2, 2])
  # )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # geometric
  beta <- c(.6, .2, -1, .4)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1)
  
  fn <- oiztgeom(
    lambdaLink = "log",
    omegaLink  = "logit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$oiztgeom1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      omegaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .3
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .2
  )
  
  
  # expect_true(
  #   (pop$confidenceInterval[1, 1] < N) &
  #   (N < pop$confidenceInterval[1, 2])
  # )
  
  # expect_true(
  #   (pop$confidenceInterval[2, 1] < N) &
  #   (N < pop$confidenceInterval[2, 2])
  # )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # same for different link
  beta <- c(.6, -.3, -1, .4)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1)
  
  fn <- oiztgeom(
    lambdaLink = "neglog",
    omegaLink  = "cloglog",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$oiztgeom2
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      omegaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .7
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  # expect_equivalent(
  #   pop$pointEstimate,
  #   N,
  #   tolerance = .2
  # )
  
  # expect_true(
  #   (pop$confidenceInterval[1, 1] < N) &
  #   (N < pop$confidenceInterval[1, 2])
  # )
  
  # expect_true(
  #   (pop$confidenceInterval[2, 1] < N) &
  #   (N < pop$confidenceInterval[2, 2])
  # )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # negbin
  beta <- c(.4, .4, -.4, .2, -1.1, .37)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1, beta[5] + beta[6] * x1)
  
  fn <- oiztnegbin(
    lambdaLink = "log",
    alphaLink  = "log", 
    omegaLink  = "logit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$oiztnegbin1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      omegaFormula = ~ x,
      alphaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1, confint = TRUE, correlation = TRUE))
  
  expect_silent(print(summary(M1, confint = TRUE, correlation = TRUE)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .6
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .2
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 4:6] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[, 4:6]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 4:6] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[, 4:6]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # ztHurdle ####
  
  # poisson
  beta <- c(.6, -.3, -.5, .1)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1)
  
  fn <- ztHurdlepoisson(
    lambdaLink = "log",
    piLink  = "logit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$ztHurdlepoisson1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      piFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .7
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .5
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # same with different link
  fn <- ztHurdlepoisson(
    lambdaLink = "neglog",
    piLink  = "probit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$ztHurdlepoisson2
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      piFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .4
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .4
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # geometric
  beta <- c(.6, .2, -3, .4)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1)
  
  fn <- ztHurdlegeom(
    lambdaLink = "log",
    omegaLink  = "logit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$ztHurdlegeom1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      piFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .4
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .2
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # same for different link
  beta <- c(.6, -.3, -1, .4)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1)
  
  
  fn <- ztHurdlegeom(
    lambdaLink = "neglog", 
    piLink = "probit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$ztHurdlegeom2
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      piFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .65
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .25
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # negbin
  beta <- c(.4, .4, -.4, .2, -1.1, .37)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1, beta[5] + beta[6] * x1)
  
  fn <- ztHurdlenegbin(
    lambdaLink = "log",
    alphaLink  = "log", 
    piLink  = "logit",
    eimStep = 20
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$ztHurdlenegbin1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      piFormula = ~ x,
      alphaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1, confint = TRUE, correlation = TRUE))
  
  expect_silent(print(summary(M1, confint = TRUE, correlation = TRUE)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .5
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .3
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 4:6] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[, 4:6]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 4:6] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[, 4:6]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # Hurdlezt ####
  # poisson
  beta <- c(.6, -.3, -.5, .1)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1)
  
  fn <- Hurdleztpoisson(
    lambdaLink = "log",
    piLink  = "logit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$Hurdleztpoisson1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      piFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .6
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .25
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # same with different link
  fn <- Hurdleztpoisson(
    lambdaLink = "neglog",
    piLink  = "probit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$Hurdleztpoisson2
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      piFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE, stepsize = .3)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .6
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .3
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # geometric
  beta <- c(.6, .2, -3, .4)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1)
  
  fn <- Hurdleztgeom(
    lambdaLink = "log",
    omegaLink  = "logit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$Hurdleztgeom1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      piFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .2
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .2
  )
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  # same for different link
  beta <- c(.3, -.7, -.25, .3)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1)
  
  fn <- Hurdleztgeom(
    lambdaLink = "neglog", 
    piLink = "probit",
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$Hurdleztgeom2
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      piFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1))
  
  expect_silent(print(summary(M1)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .4
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .2
  )
  
  
  expect_true(
    (pop$confidenceInterval[1, 1] < N) &
    (N < pop$confidenceInterval[1, 2])
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 3:4] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[,3:4]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  # negbin
  beta <- c(.4, .4, -.4, .2, -1.1, .37)
  #eta <- cbind(beta[1] + beta[2] * x1, beta[3] + beta[4] * x1, beta[5] + beta[6] * x1)
  
  fn <- Hurdleztnegbin(
    lambdaLink = "log",
    alphaLink  = "log", 
    piLink  = "logit",
    eimStep = 30
  )
  
  #y <- fn$simulate(n = N, eta = eta, lower = -1)
  y <- test_inflated$Hurdleztnegbin1
  
  df <- data.frame(
    x = x1[y > 0],
    y = y[y > 0]
  )
  
  M1 <- estimatePopsize(
    formula = y ~ x,
    model   = fn,
    data    = df,
    method = "IRLS",
    controlModel = controlModel(
      piFormula = ~ x,
      alphaFormula = ~ x
    ),
    controlMethod = controlMethod(silent = TRUE)
  )
  
  expect_silent(summary(M1, confint = TRUE, correlation = TRUE))
  
  expect_silent(print(summary(M1, confint = TRUE, correlation = TRUE)))
  
  expect_equivalent(
    coef(M1),
    beta,
    tolerance = .7
  )
  
  expect_silent(
    pop <- popSizeEst(M1)
  )
  
  expect_equivalent(
    pop$pointEstimate,
    N,
    tolerance = .35
  )
  
  expect_true(
    (pop$confidenceInterval[2, 1] < N) &
    (N < pop$confidenceInterval[2, 2])
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "response",
      se.fit = TRUE
    )[, 4:6] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "link",
      se.fit = TRUE
    )[, 4:6]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_equivalent(
    sum(predict(
      M1, 
      type = "contr"
    )),
    M1$populationSize$pointEstimate
  )
  
  expect_silent(
    predict(
      M1, 
      type = "popSize",
      se.fit = TRUE
    )
  )
  
  expect_true(
    all(predict(
      M1,
      newdata = data.frame(x = x1),
      type = "response",
      se.fit = TRUE
    )[, 4:6] > 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "link",
      se.fit = TRUE
    )[, 4:6]> 0)
  )
  
  expect_true(
    all(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "mean",
      se.fit = TRUE
    )[, c(2, 4)]> 0)
  )
  
  expect_true(
    sum(predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "contr"
    )) > N
  )
  
  expect_silent(
    predict(
      M1, 
      newdata = data.frame(x = x1),
      type = "popSize",
      se.fit = TRUE
    )
  )
}

# cheking popsize estimation as in articles on which the package is based

expect_silent(
  opt <- estimatePopsize(
    formula = TOTAL_SUB ~ .,
    model = "ztpoisson",
    method = "optim",
    popVar = "analytic",
    data = farmsubmission,
    controlMethod = controlMethod(
      maxiter = 5000,
      optimMethod = "Nelder-Mead",
      silent = TRUE
    )
  )
)

expect_silent(
  dd <- estimatePopsize(
    formula = TOTAL_SUB ~ .,
    model = "zotnegbin",
    method = "optim",
    data = farmsubmission,
    controlMethod = controlMethod(
      maxiter = 5000,
      optimMethod = "Nelder-Mead",
      silent = TRUE
    )
  )
)

expect_silent(
  predict(
    dd,
    type = "mean"
  )
)

expect_silent(
  dd <- estimatePopsize(
    formula = TOTAL_SUB ~ .,
    model = "zotgeom",
    method = "optim",
    data = farmsubmission
  )
)

expect_silent(
  predict(
    dd,
    type = "mean"
  )
)

expect_silent(
  irls <- estimatePopsize(
    formula = TOTAL_SUB ~ .,
    model = "ztpoisson",
    data = farmsubmission,
    controlMethod = controlMethod(
      silent = TRUE
    )
  )
)

expect_equivalent(
  irls$populationSize$pointEstimate,
  18346, tolerance = .005
)

expect_equivalent(
  opt$populationSize$pointEstimate,
  18346, tolerance = .005
)

expect_equivalent(
  irls$populationSize$confidenceInterval[1, ],
  list(17932, 18760),
  tolerance = .005
)

expect_equivalent(
  opt$populationSize$confidenceInterval[1, ],
  list(17932, 18760),
  tolerance = .005
)

expect_silent(
  ch <- estimatePopsize(
    formula = TOTAL_SUB ~ .,
    model = chao(),
    data = farmsubmission,
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_silent(
  predict(
    ch,
    type = "mean"
  )
)

expect_equivalent(
  ch$populationSize$pointEstimate,
  21657,
  tolerance = .005
)

expect_equivalent(
  ch$populationSize$confidenceInterval[1, ],
  list(20883, 22430),
  tolerance = .005
)

oi_theory_data <- data.frame(
  y = c(rep(2, 95), rep(3, 32))
)

expect_silent(
  oi_theory <- estimatePopsize(
    formula = y ~ 1,
    model = "oichao",
    data = oi_theory_data,
    method = "IRLS",
    popVar = "analytic",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_equivalent(
  3 * exp(unname(coef(oi_theory))[1]),
  3 * 32 / 95,
  tolerance = .005
)

expect_equivalent(
  oi_theory$populationSize$pointEstimate,
  127 + (2 / 9) * 95 ^ 3 / 32 ^ 2,
  tolerance = .005
)

set.seed(2024)
oi_covariate_data <- data.frame(
  x = rnorm(500)
)
oi_covariate_data$lambda <- exp(0.2 + 0.35 * oi_covariate_data$x)
oi_covariate_data$y <- rpois(nrow(oi_covariate_data), oi_covariate_data$lambda)
oi_covariate_data <- subset(oi_covariate_data, y > 0)

expect_silent(
  oi_irls <- estimatePopsize(
    formula = y ~ x,
    model = "oichao",
    data = oi_covariate_data,
    method = "IRLS",
    popVar = "analytic",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_silent(
  oi_opt <- estimatePopsize(
    formula = y ~ x,
    model = oichao(),
    data = oi_covariate_data,
    method = "optim",
    popVar = "analytic",
    controlMethod = controlMethod(
      silent = TRUE,
      optimMethod = "BFGS",
      maxiter = 500
    )
  )
)

expect_true(
  max(abs(coef(oi_irls) - coef(oi_opt))) < 1e-3
)

expect_true(
  abs(oi_irls$populationSize$pointEstimate - oi_opt$populationSize$pointEstimate) < 0.25
)

expect_silent(
  predict(
    oi_irls,
    type = "response",
    se.fit = TRUE
  )
)

expect_silent(
  summary(oi_irls)
)

expect_silent(
  popSizeEst(oi_irls)
)

# on netherlandsimmigrant
# confint

expect_silent(
  zl <- estimatePopsize(
    formula = capture ~ . - reason - nation,
    data = netherlandsimmigrant,
    model = "zelterman",
    popVar = "analytic",
    method = "IRLS",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_silent(
  predict(
    zl,
    type = "mean"
  )
)

expect_silent(
  zl2 <- estimatePopsize(
    formula = capture ~ 1,
    data = netherlandsimmigrant,
    model = "zelterman",
    popVar = "analytic",
    method = "IRLS",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_silent(
  poi <- estimatePopsize(
    formula = capture ~ .,
    data = netherlandsimmigrant,
    model = "ztpoisson",
    popVar = "analytic",
    method = "IRLS",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_silent(
  poi2 <- estimatePopsize(
    formula = capture ~ 1,
    data = netherlandsimmigrant,
    model = "ztpoisson",
    popVar = "analytic",
    method = "IRLS",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_equivalent(
  zl$populationSize$confidenceInterval[1, ],
  list(8416, 12009),
  tolerance = .005
)

expect_equivalent(
  zl2$populationSize$confidenceInterval[1, ],
  list(8084, 10765),
  tolerance = .005
)

expect_equivalent(
  poi$populationSize$confidenceInterval[1, ],
  list(7185, 18198),
  tolerance = .005
)

expect_equivalent(
  poi2$populationSize$confidenceInterval[1, ],
  list(6363, 7797),
  tolerance = .005
)

# point
expect_equivalent(
  poi$populationSize$pointEstimate,
  12691,
  tolerance = .005
)

expect_equivalent(
  poi2$populationSize$pointEstimate,
  7080,
  tolerance = .005
)

expect_equivalent(
  zl$populationSize$pointEstimate,
  10213,
  tolerance = .005
)

expect_equivalent(
  zl2$populationSize$pointEstimate,
  9425,
  tolerance = .005
)

# other tests
expect_error(
  estimatePopsize(
    formula = Y ~ X^2, 
    data = data.frame(Y = rep(c(0, 1, 2, 3), 50),
                      X = rep(c(1, 2), 100)),
    model = "ztnegbin",
    method = "IRLS",
    popVar = "analytic"
  )
)

set.seed(123)
expect_silent(
  estimatePopsize(
    formula = TOTAL_SUB ~ .,
    data = farmsubmission,
    model = "zotpoisson",
    popVar = "bootstrap",
    controlMethod = controlMethod(epsilon = 1e-6, silent = TRUE),# testing silent
    controlPopVar = controlPopVar(B = 4, bootstrapFitcontrol = controlMethod(silent = TRUE, epsilon = .Machine$double.eps))
  )
)

expect_silent(
  estimatePopsize(
    formula = TOTAL_SUB ~ .,
    data = farmsubmission,
    model = "ztoigeom", 
    method = "optim",
    popVar = "bootstrap",
    controlMethod = controlMethod(epsilon = 1e-6, silent = TRUE),
    controlPopVar = controlPopVar(
      B = 6,
      bootType = "semiparametric"
    )
  )
)

expect_silent(
  estimatePopsize(
    formula = TOTAL_SUB ~ .,
    data = farmsubmission,
    model = "chao",
    popVar = "bootstrap",
    method = "IRLS",
    controlMethod = controlMethod(silent = TRUE),
    controlPopVar = controlPopVar(
      B = 6,
      bootType = "nonparametric"
    )
  )
)

expect_silent(
  estimatePopsize(
    formula = capture ~ gender,
    data = netherlandsimmigrant,
    model = "ztpoisson",
    popVar = "bootstrap",
    method = "IRLS",
    controlPopVar = controlPopVar(
      B = 4, 
      bootstrapFitcontrol = controlMethod(
        silent = TRUE, 
        epsilon = .Machine$double.eps)
      )
  )
)

expect_error(
  estimatePopsize(formula = cbind(TOTAL_SUB, log_size) ~ log_distance, 
                  data = farmsubmission, model = ztpoisson)
)

# testing for zero truncated models

eps <- if (capabilities("long.double")) sqrt(.Machine$double.eps) else 0.01

set.seed(123)

N <- 1000
#x1 <- rbinom(n = N, size = 3, prob = .4)
#x2 <- runif(n = N)
#x3 <- rnorm(n = N, mean = 3, sd = 4)

testTruncated <- read.csv("test_zerotruncated.csv")
#testTruncated <- read.csv("inst/tinytest/test_zerotruncated.csv")
x1 <- testTruncated$x1
x2 <- testTruncated$x2
x3 <- testTruncated$x3

## poisson ####
beta <- c(.6, .2, -1.25, .1)
#eta <- cbind(beta[1] + beta[2] * x1 + beta[3] * x2 + beta[4] * x3)

fn <- ztpoisson(
  lambdaLink = "neglog"
)

#y <- fn$simulate(n = N, eta = eta, lower = -1)
y <- testTruncated$ztpoisson1

df <- data.frame(
  x3 = x3[y > 0],
  x2 = x2[y > 0],
  x1 = x1[y > 0],
  y = y[y > 0]
)

M1 <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = fn,
  data    = df,
  method = "IRLS"
)

expect_equivalent(
  coef(M1),
  beta,
  tolerance = .25
)

expect_silent(
  pop <- popSizeEst(M1)
)

expect_equivalent(
  N,
  pop$pointEstimate,
  tol = .15
)

expect_true(
  all(N < pop$confidenceInterval$upperBound) &
  all(N > pop$confidenceInterval$lowerBound)
)

expect_silent(
  stra <- stratifyPopsize(M1, ~ x1)
)

expect_equivalent(
  as.numeric(table(df$x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Estimated,
  tolerance = .1
)

AA <- summary(marginalFreq(M1), dropl5 = "group", df = 1)

expect_true(
  all(AA$Test$`P(>X^2)` > .01)
)

expect_true(
  all(predict(
    M1, 
    type = "response",
    newdata = testTruncated[, paste0("x", 1:3)],
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    M1, 
    type = "link",
    newdata = testTruncated[, paste0("x", 1:3)],
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    M1, 
    type = "mean",
    newdata = testTruncated[, paste0("x", 1:3)],
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    M1, 
    type = "contr",
    newdata = testTruncated[, paste0("x", 1:3)]
  )) > N
)

expect_silent(
  predict(
    M1, 
    type = "popSize",
    newdata = testTruncated[, paste0("x", 1:3)],
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    M1, 
    type = "response",
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    M1, 
    type = "link",
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    M1, 
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    M1, 
    type = "contr"
  )) == M1$populationSize$pointEstimate
)

expect_silent(
  predict(
    M1, 
    type = "popSize",
    se.fit = TRUE
  )
)

# different link

fn <- ztpoisson(
  lambdaLink = "log"
)

#y <- fn$simulate(n = N, eta = eta, lower = -1)
y <- testTruncated$ztpoisson2

df <- data.frame(
  x3 = x3[y > 0],
  x2 = x2[y > 0],
  x1 = x1[y > 0],
  y = y[y > 0]
)

M1 <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = fn,
  data    = df,
  method = "IRLS"
)

expect_equivalent(
  coef(M1),
  beta,
  tolerance = .15
)

expect_silent(
  pop <- popSizeEst(M1)
)

expect_equivalent(
  N,
  pop$pointEstimate,
  tol = .1
)

expect_true(
  all(N < pop$confidenceInterval$upperBound) &
  all(N > pop$confidenceInterval$lowerBound)
)

expect_silent(
  stra <- stratifyPopsize(M1, ~ x1)
)

expect_equivalent(
  as.numeric(table(df$x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Estimated,
  tolerance = .1
)

AA <- summary(marginalFreq(M1), dropl5 = "group")

expect_true(
  all(AA$Test$`P(>X^2)` > .01)
)

expect_true(
  all(predict(
    M1, 
    type = "response",
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    M1, 
    type = "link",
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    M1, 
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    M1, 
    type = "contr"
  )) == M1$populationSize$pointEstimate
)

expect_silent(
  predict(
    M1, 
    type = "popSize",
    se.fit = TRUE
  )
)

## bootstrap

M1 <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = fn,
  data    = df,
  method = "IRLS",
  popVar = "bootstrap",
  controlPopVar = controlPopVar(
    B = 250,
    bootType = "parametric"
  )
)

expect_silent(
  plot(M1, "bootHist")
)

expect_silent(
  print(summary(M1, correlation = TRUE))
)

expect_silent(
  pop <- popSizeEst(M1)
)

expect_equivalent(
  N,
  pop$pointEstimate,
  tol = .1
)

expect_true(
  all(N < pop$confidenceInterval[2]) &
  all(N > pop$confidenceInterval[1])
)

expect_true(
  shapiro.test(pop$boot)$p.value > .05
)

M1 <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = fn,
  data    = df,
  method = "IRLS",
  popVar = "bootstrap",
  controlPopVar = controlPopVar(
    B = 250,
    bootType = "semiparametric"
  )
)

expect_silent(
  plot(M1, "bootHist")
)

expect_silent(
  print(summary(M1, correlation = TRUE))
)

expect_silent(
  pop <- popSizeEst(M1)
)

expect_equivalent(
  N,
  pop$pointEstimate,
  tol = .1
)

expect_true(
  all(N < pop$confidenceInterval[2]) &
  all(N > pop$confidenceInterval[1])
)

expect_true(
  shapiro.test(pop$boot)$p.value > .05
)

expect_true(
  all(predict(
    M1, 
    type = "response",
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    M1, 
    type = "link",
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    M1, 
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    M1, 
    type = "contr"
  )) == M1$populationSize$pointEstimate
)

expect_silent(
  predict(
    M1, 
    type = "popSize",
    se.fit = TRUE
  )
)

M1 <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = fn,
  data    = df,
  method = "IRLS",
  popVar = "bootstrap",
  controlPopVar = controlPopVar(
    B = 250,
    bootType = "nonparametric"
  )
)

expect_silent(
  plot(M1, "bootHist")
)

expect_silent(
  print(summary(M1, correlation = TRUE))
)

expect_silent(
  pop <- popSizeEst(M1)
)

expect_equivalent(
  N,
  pop$pointEstimate,
  tol = .1
)

## geometric ####
beta <- c(.6, .2, -1.25, .1)
#eta <- cbind(beta[1] + beta[2] * x1 + beta[3] * x2 + beta[4] * x3)

fn <- ztgeom(
  lambdaLink = "neglog"
)

#y <- fn$simulate(n = N, eta = eta, lower = -1)
y <- testTruncated$ztgeom1

df <- data.frame(
  x3 = x3[y > 0],
  x2 = x2[y > 0],
  x1 = x1[y > 0],
  y = y[y > 0]
)

M1 <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = fn,
  data    = df,
  method = "IRLS"
)

expect_equivalent(
  coef(M1),
  beta,
  tolerance = .25
)

expect_silent(
  pop <- popSizeEst(M1)
)

expect_equivalent(
  N,
  pop$pointEstimate,
  tol = .1
)

expect_true(
  all(N < pop$confidenceInterval$upperBound) &
  all(N > pop$confidenceInterval$lowerBound)
)

expect_silent(
  stra <- stratifyPopsize(M1, ~ x1)
)

expect_equivalent(
  as.numeric(table(df$x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Estimated,
  tolerance = .2
)

AA <- summary(marginalFreq(M1), dropl5 = "group")

expect_true(
  all(AA$Test$`P(>X^2)` > .03)
)

expect_true(
  all(predict(
    M1, 
    type = "response",
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    M1, 
    type = "link",
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    M1, 
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    M1, 
    type = "contr"
  )) == M1$populationSize$pointEstimate
)

expect_silent(
  predict(
    M1, 
    type = "popSize",
    se.fit = TRUE
  )
)

# different link

fn <- ztgeom(
  lambdaLink = "log"
)

#y <- fn$simulate(n = N, eta = eta, lower = -1)
y <- testTruncated$ztgeom2

df <- data.frame(
  x3 = x3[y > 0],
  x2 = x2[y > 0],
  x1 = x1[y > 0],
  y = y[y > 0]
)

M1 <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = fn,
  data    = df,
  method = "IRLS"
)

expect_equivalent(
  coef(M1),
  beta,
  tolerance = .2
)

expect_silent(
  pop <- popSizeEst(M1)
)

expect_equivalent(
  N,
  pop$pointEstimate,
  tol = .05
)

expect_true(
  all(N < pop$confidenceInterval$upperBound) &
  all(N > pop$confidenceInterval$lowerBound)
)

expect_silent(
  stra <- stratifyPopsize(M1, ~ x1)
)

expect_equivalent(
  as.numeric(table(df$x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Observed,
  tolerance = .1
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Estimated,
  tolerance = .1
)

AA <- summary(marginalFreq(M1), dropl5 = "group")

expect_true(
  all(AA$Test$`P(>X^2)` > .05)
)

expect_true(
  all(predict(
    M1, 
    type = "response",
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    M1, 
    type = "link",
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    M1, 
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    M1, 
    type = "contr"
  )) == M1$populationSize$pointEstimate
)

expect_silent(
  predict(
    M1, 
    type = "popSize",
    se.fit = TRUE
  )
)

## ztnegbin ####
beta <- c(.3, .2, -.3, .1,
          .2, .1, -.5, .3)
#eta <- cbind(beta[1] + beta[2] * x1 + beta[3] * x2 + beta[4] * x3,
#             beta[5] + beta[6] * x1 + beta[7] * x2 + beta[8] * x3)

fn <- ztnegbin(
  lambdaLink = "log",
  alphaLink = "log",
  eimStep = 10
)

#y <- fn$simulate(n = N, eta = eta, lower = -1)
y <- testTruncated$ztnegbin1

df <- data.frame(
  x3 = x3[y > 0],
  x2 = x2[y > 0],
  x1 = x1[y > 0],
  y = y[y > 0]
)

M1 <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = fn,
  data    = df,
  method = "IRLS",
  controlModel = controlModel(
    alphaFormula = ~ x1 + x2 + x3
  )
)

expect_silent(
  pop <- popSizeEst(M1)
)

expect_equivalent(
  pop$pointEstimate,
  N,
  tol = .3
)

expect_true(
  all(N < pop$confidenceInterval$upperBound) &
  all(N > pop$confidenceInterval$lowerBound)
)

expect_silent(
  stra <- stratifyPopsize(M1, ~ x1)
)

expect_equivalent(
  as.numeric(table(df$x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Observed
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Estimated,
  tolerance = .5
)

# different link
beta <- c(.1, .2, -.3, .1,
          .2, .1, -.1,   .3)

fn <- ztnegbin(
  lambdaLink = "log",
  alphaLink = "neglog",
  eimStep = 40
)

#y <- fn$simulate(n = N, eta = eta, lower = -1)
y <- testTruncated$ztnegbin2

df <- data.frame(
  x3 = x3[y > 0],
  x2 = x2[y > 0],
  x1 = x1[y > 0],
  y = y[y > 0]
)

M1 <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = fn,
  data    = df,
  method = "IRLS",
  controlModel = controlModel(
    alphaFormula = ~ x1 + x2 + x3
  )
)

expect_true(
  max(abs(coef(M1) - beta)) < .9
)

expect_silent(
  pop <- popSizeEst(M1)
)

expect_equivalent(
  N,
  pop$pointEstimate,
  tol = .1
)

expect_true(
  all(N < pop$confidenceInterval$upperBound) &
  all(N > pop$confidenceInterval$lowerBound)
)

expect_silent(
  stra <- stratifyPopsize(M1, ~ x1)
)

expect_equivalent(
  as.numeric(table(df$x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Estimated,
  tolerance = .1
)

AA <- summary(marginalFreq(M1), dropl5 = "no")

expect_true(
  all(AA$Test$`P(>X^2)` > .05)
)

expect_true(
  all(predict(
    M1, 
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    M1, 
    type = "link",
    se.fit = TRUE
  )[,3:4] > 0)
)

expect_true(
  all(predict(
    M1, 
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    M1, 
    type = "contr"
  )) == M1$populationSize$pointEstimate
)

expect_silent(
  predict(
    M1, 
    type = "popSize",
    se.fit = TRUE
  )
)

# different link
fn <- ztnegbin(
  lambdaLink = "neglog",
  alphaLink = "log",
  eimStep = 10
)

#y <- fn$simulate(n = N, eta = eta, lower = -1)
y <- testTruncated$ztnegbin3

df <- data.frame(
  x3 = x3[y > 0],
  x2 = x2[y > 0],
  x1 = x1[y > 0],
  y = y[y > 0]
)

M1 <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = fn,
  data    = df,
  method = "IRLS",
  controlModel = controlModel(
    alphaFormula = ~ x1 + x2 + x3
  )
)

expect_silent(
  pop <- popSizeEst(M1)
)

expect_equivalent(
  N,
  pop$pointEstimate,
  tol = .6
)

expect_true(
  (N < pop$confidenceInterval$upperBound[2]) &
  (N > pop$confidenceInterval$lowerBound[2])
)

expect_silent(
  stra <- stratifyPopsize(M1, ~ x1)
)

expect_equivalent(
  as.numeric(table(df$x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Estimated,
  tolerance = .6
)

expect_true(
  all(predict(
    M1, 
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    M1, 
    type = "link",
    se.fit = TRUE
  )[, 3:4]> 0)
)

expect_true(
  all(predict(
    M1, 
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)] >= 0)
)

expect_true(
  sum(predict(
    M1, 
    type = "contr"
  )) == M1$populationSize$pointEstimate
)

expect_silent(
  predict(
    M1, 
    type = "popSize",
    se.fit = TRUE
  )
)

# different link
beta <- c(.6, -.2, -1.25, -.1,
          .2, -.1, -.5,   .3)
#eta <- cbind(beta[1] + beta[2] * x1 + beta[3] * x2 + beta[4] * x3,
#             beta[5] + beta[6] * x1 + beta[7] * x2 + beta[8] * x3)

fn <- ztnegbin(
  lambdaLink = "log",
  alphaLink = "log",
  eimStep = 25
)

#y <- fn$simulate(n = N, eta = eta, lower = -1)
y <- testTruncated$ztnegbin4

df <- data.frame(
  x3 = x3[y > 0],
  x2 = x2[y > 0],
  x1 = x1[y > 0],
  y = y[y > 0]
)

M1 <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = fn,
  data    = df,
  method = "IRLS",
  controlModel = controlModel(
    alphaFormula = ~ x1 + x2 + x3
  )
)

expect_equivalent(
  beta,
  coef(M1),
  tolerance = .4
)

expect_silent(
  pop <- popSizeEst(M1)
)

expect_equivalent(
  N,
  pop$pointEstimate,
  tol = .15
)

expect_true(
  all(N < pop$confidenceInterval$upperBound) &
  all(N > pop$confidenceInterval$lowerBound)
)

expect_silent(
  stra <- stratifyPopsize(M1, ~ x1)
)

expect_equivalent(
  as.numeric(table(df$x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(stra$name, start = 5, stop = 5)) + 1],
  stra$Estimated,
  tolerance = .15
)

AA <- summary(marginalFreq(M1), dropl5 = "group")

expect_true(
  all(AA$Test$`P(>X^2)` > .05)
)

expect_true(
  all(predict(
    M1, 
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    M1, 
    type = "link",
    se.fit = TRUE
  )[, 3:4]> 0)
)

expect_true(
  all(predict(
    M1, 
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)] >= 0)
)

expect_true(
  sum(predict(
    M1, 
    type = "contr"
  )) == M1$populationSize$pointEstimate
)

expect_silent(
  predict(
    M1, 
    type = "popSize",
    se.fit = TRUE
  )
)

ratio_raw <- data.frame(
  observed = c(rep(1, 50), rep(2, 30), rep(3, 15), rep(4, 6), rep(5, 2))
)

ratio_tab <- aggregate(
  list(weight = rep(1, nrow(ratio_raw))),
  by = list(observed = ratio_raw$observed),
  FUN = sum
)

expect_error(
  ratioReg(
    observed ~ grp,
    data = data.frame(observed = c(1, 2, 2, 3), grp = c("a", "a", "b", "b")),
    confint = "none"
  ),
  pattern = "ratioFormula"
)

expect_silent(
  ratio_fit_raw <- ratioReg(
    observed ~ 1,
    data = ratio_raw,
    model = "M0",
    estimator = "SM",
    confint = "none"
  )
)

expect_silent(
  ratio_fit_tab <- ratioReg(
    observed ~ 1,
    data = ratio_tab,
    weights = ratio_tab$weight,
    model = "M0",
    estimator = "SM",
    confint = "none"
  )
)

expect_equal(
  coef(ratio_fit_raw),
  coef(ratio_fit_tab),
  tolerance = sqrt(.Machine$double.eps)
)

expect_equal(
  popSizeEst(ratio_fit_raw)$pointEstimate,
  popSizeEst(ratio_fit_tab)$pointEstimate,
  tolerance = sqrt(.Machine$double.eps)
)

gap_counts <- data.frame(
  observed = c(rep(1, 100), rep(2, 80), rep(4, 20), rep(5, 4))
)

gap_fit <- ratioReg(
  observed ~ 1,
  data = gap_counts,
  model = "M0",
  estimator = "HT",
  confint = "none"
)

expect_identical(
  gap_fit$ratioData$k,
  c(1L, 4L)
)

contig_counts <- data.frame(
  observed = c(rep(1, 100), rep(2, 80), rep(3, 48), rep(4, 19))
)

contig_fit <- ratioReg(
  observed ~ 1,
  data = contig_counts,
  model = "M0",
  estimator = "HT",
  confint = "none"
)

manual_ratio_data <- data.frame(
  k = 1:3,
  ratio = c(80 / 100, 48 / 80, 19 / 48)
)
manual_ratio_data$logRatio <- log(manual_ratio_data$ratio)
manual_ratio_data$weight <- 1 / c(
  1 / 100 + 1 / 80,
  1 / 80 + 1 / 48,
  1 / 48 + 1 / 19
)

manual_lm <- lm(
  logRatio ~ log(k + 1),
  data = manual_ratio_data,
  weights = weight
)

expect_equal(
  coef(contig_fit),
  coef(manual_lm),
  tolerance = 1e-10
)

manual_sse <- sum(manual_ratio_data$weight * residuals(manual_lm) ^ 2)
manual_sigma2 <- manual_sse / nrow(manual_ratio_data)
manual_loglik <- -0.5 * nrow(manual_ratio_data) *
  (log(2 * pi) + 1 + log(manual_sigma2)) +
  0.5 * sum(log(manual_ratio_data$weight))

expect_equal(
  as.numeric(logLik(contig_fit)),
  manual_loglik,
  tolerance = 1e-10
)

expect_equal(
  AIC(contig_fit),
  -2 * manual_loglik + 2 * (manual_lm$rank + 1),
  tolerance = 1e-10
)

expect_equal(
  BIC(contig_fit),
  -2 * manual_loglik + log(nrow(manual_ratio_data)) * (manual_lm$rank + 1),
  tolerance = 1e-10
)

manual_log_ratios <- coef(contig_fit)[1] + coef(contig_fit)[2] * log(1:4)
manual_terms <- exp(c(0, cumsum(manual_log_ratios)))
manual_p0 <- 1 / sum(manual_terms)
manual_ht <- sum(contig_fit$counts) / (1 - manual_p0)
manual_sm <- sum(contig_fit$counts) + contig_fit$counts["2"] /
  exp(sum(manual_log_ratios[1:2]))

expect_equal(
  unname(contig_fit$fittedProbabilities["0"]),
  manual_p0,
  tolerance = 1e-10
)

expect_equal(
  popSizeEst(contig_fit, estimator = "HT")$pointEstimate,
  manual_ht,
  tolerance = 1e-10
)

expect_equal(
  unname(popSizeEst(contig_fit, estimator = "SM")$pointEstimate),
  unname(manual_sm),
  tolerance = 1e-10
)

no_infl_counts <- data.frame(
  observed = c(rep(1, 30), rep(2, 27), rep(3, 20), rep(4, 12), rep(5, 7), rep(6, 4))
)

no_infl_fit <- ratioReg(
  observed ~ 1,
  data = no_infl_counts,
  model = "auto",
  estimator = "SM",
  confint = "none"
)

inflated_counts <- data.frame(
  observed = c(rep(1, 200), rep(2, 30), rep(3, 15), rep(4, 6), rep(5, 2))
)

inflated_fit <- ratioReg(
  observed ~ 1,
  data = inflated_counts,
  model = "auto",
  estimator = "SM",
  confint = "none"
)

expect_identical(
  no_infl_fit$selectedModel,
  "M0"
)

expect_identical(
  inflated_fit$selectedModel,
  "M1"
)

expect_silent(
  boot_fit <- ratioReg(
    observed ~ 1,
    data = ratio_raw,
    model = "auto",
    estimator = "SM",
    confint = "bootstrap",
    B = 20,
    seed = 123
  )
)

expect_silent(
  boot_fit_same <- ratioReg(
    observed ~ 1,
    data = ratio_raw,
    model = "auto",
    estimator = "SM",
    confint = "bootstrap",
    B = 20,
    seed = 123
  )
)

expect_equal(
  popSizeEst(boot_fit)$confidenceInterval,
  popSizeEst(boot_fit_same)$confidenceInterval,
  tolerance = sqrt(.Machine$double.eps)
)

expect_equal(
  confint(boot_fit),
  as.matrix(popSizeEst(boot_fit)$confidenceInterval),
  tolerance = sqrt(.Machine$double.eps)
)

expect_true(
  length(popSizeEst(boot_fit)$boot) == 20
)

expect_silent(summary(boot_fit))
expect_silent(print(boot_fit))

expect_silent({
  grDevices::pdf(tempfile(fileext = ".pdf"))
  plot(boot_fit, plotType = "ratio")
  grDevices::dev.off()
})

expect_silent({
  grDevices::pdf(tempfile(fileext = ".pdf"))
  plot(boot_fit, plotType = "bootHist")
  grDevices::dev.off()
})

expect_silent({
  grDevices::pdf(tempfile(fileext = ".pdf"))
  plot(boot_fit, plotType = "freqFit")
  grDevices::dev.off()
})

expect_error(
  estimatePopsize(
    capture ~ 1,
    data = netherlandsimmigrant,
    model = "ztpoisson",
    ratioReg = TRUE,
    controlMethod = controlMethod(silent = TRUE)
  ),
  pattern = "Use ratioReg"
)

# input validation -----------------------------------------------------------

expect_error(
  ratioReg(
    observed ~ 1,
    data = data.frame(observed = c(1, -1, 2)),
    confint = "none"
  ),
  pattern = "strictly positive"
)

expect_error(
  ratioReg(
    observed ~ 1,
    data = data.frame(observed = c(1, 2.5, 3)),
    confint = "none"
  ),
  pattern = "whole numbers"
)

expect_error(
  ratioReg(
    observed ~ 1,
    data = data.frame(observed = c(1, 2, Inf)),
    confint = "none"
  ),
  pattern = "finite"
)

expect_error(
  ratioReg(
    ~ 1,
    data = data.frame(observed = c(1, 2, 3)),
    confint = "none"
  ),
  pattern = "response variable"
)

expect_error(
  ratioReg(
    observed ~ 0,
    data = data.frame(observed = c(1, 2, 3)),
    confint = "none"
  ),
  pattern = "~ 1"
)

expect_error(
  ratioReg(
    observed ~ 1,
    data = ratio_raw,
    confint = "bootstrap",
    B = -1
  ),
  pattern = "non-negative"
)

expect_error(
  ratioReg(
    observed ~ 1,
    data = ratio_raw,
    confint = "bootstrap",
    B = 0
  ),
  pattern = "at least 1"
)

expect_error(
  ratioReg(
    observed ~ 1,
    data = data.frame(observed = c(rep(1, 10), rep(2, 5))),
    confint = "none",
    maxCount = 1
  ),
  pattern = "at least 2"
)

expect_error(
  ratioReg(
    observed ~ 1,
    data = ratio_raw,
    weights = c(1, -1, rep(1, nrow(ratio_raw) - 2)),
    confint = "none"
  ),
  pattern = "non-negative"
)

# maxCount truncates the support ---------------------------------------------

mc_counts <- data.frame(
  observed = c(rep(1, 100), rep(2, 80), rep(3, 48), rep(4, 19), rep(5, 5))
)
mc_full <- ratioReg(
  observed ~ 1,
  data = mc_counts,
  model = "M0",
  confint = "none"
)
mc_capped <- ratioReg(
  observed ~ 1,
  data = mc_counts,
  model = "M0",
  confint = "none",
  maxCount = 3
)

expect_identical(
  names(mc_full$counts),
  as.character(1:5)
)
expect_identical(
  names(mc_capped$counts),
  as.character(1:3)
)
expect_identical(mc_capped$ratioData$k, c(1L, 2L))

# subset works with both a bare expression and a precomputed logical ---------

set.seed(11)
subset_data <- data.frame(
  observed = sample(1:5, 300, replace = TRUE, prob = c(.4, .3, .15, .1, .05)),
  grp = sample(c("a", "b"), 300, replace = TRUE)
)
keep_a <- subset_data$grp == "a"

fit_bare <- ratioReg(
  observed ~ 1,
  data = subset_data,
  model = "M0",
  confint = "none",
  subset = grp == "a"
)
fit_pre <- ratioReg(
  observed ~ 1,
  data = subset_data,
  model = "M0",
  confint = "none",
  subset = keep_a
)

expect_equal(fit_bare$sizeObserved, sum(keep_a))
expect_equal(
  unname(coef(fit_bare)),
  unname(coef(fit_pre)),
  tolerance = sqrt(.Machine$double.eps)
)

# method variants -----------------------------------------------------------

method_fit <- ratioReg(
  observed ~ 1,
  data = ratio_raw,
  model = "auto",
  estimator = "SM",
  confint = "none"
)

expect_equal(
  nobs(method_fit),
  nrow(method_fit$ratioData)
)

expect_equal(
  unname(extractAIC(method_fit)),
  unname(AIC(method_fit)),
  tolerance = sqrt(.Machine$double.eps)
)

expect_equal(
  unname(coef(method_fit, model = "M0")),
  unname(coef(method_fit, model = "M0", component = "full")),
  tolerance = sqrt(.Machine$double.eps)
)

m1_base <- coef(method_fit, model = "M1", component = "base")
expect_false("oneInflation" %in% names(m1_base))

expect_equal(
  popSizeEst(method_fit)$pointEstimate,
  popSizeEst(method_fit, estimator = method_fit$primaryEstimator)$pointEstimate,
  tolerance = sqrt(.Machine$double.eps)
)

expect_true(
  popSizeEst(method_fit, estimator = "HT")$pointEstimate >=
    method_fit$sizeObserved
)
expect_true(
  popSizeEst(method_fit, estimator = "SM")$pointEstimate >=
    method_fit$sizeObserved
)

# fitted frequencies sum to the observed total ------------------------------

expect_equal(
  unname(sum(method_fit$fittedFrequencies)),
  method_fit$sizeObserved,
  tolerance = 1e-8
)

# fitted base probabilities form a proper distribution ----------------------

expect_equal(
  unname(sum(method_fit$fittedProbabilities)),
  1,
  tolerance = 1e-10
)
expect_true(all(method_fit$fittedProbabilities >= 0))

# custom ratioFormula -------------------------------------------------------

custom_fit <- ratioReg(
  observed ~ 1,
  data = ratio_raw,
  ratioFormula = ~ k,
  model = "M0",
  confint = "none"
)
expect_identical(
  names(coef(custom_fit, model = "M0")),
  c("(Intercept)", "k")
)

# bootstrap with model = "M0" exercises the f0-imputation branch ------------

set.seed(0)
boot_m0 <- ratioReg(
  observed ~ 1,
  data = ratio_raw,
  model = "M0",
  estimator = "HT",
  confint = "bootstrap",
  B = 15,
  seed = 17
)
expect_true(length(popSizeEst(boot_m0)$boot) == 15)
expect_true(is.finite(popSizeEst(boot_m0)$variance))

# These functions are only used internally in the package so there is no need for documenting them
#' @importFrom graphics points
noparBoot <- function(family, formulas, y, X, modelFrame,
                      beta, weights, trcount, numboot,
                      eta, trace, visT, controlBootstrapMethod = NULL,
                      method, N, offset, weightsFlag, ...) {
  strappedStatistic <- vector("numeric", length = numboot)
  n <- length(y)
  if (isTRUE(weightsFlag)) {
    n <- sum(weights)
  }
  famName <- family$family
  
  if (length(weights) == 1) {
    weights <- rep(1, n)
  }
  
  k <- 1
  
  if (visT) {
    plot(
      1, type = "n",
      xlab = "Bootstrap sample", 
      ylab = "Value of population size estimator",
      main = expression(paste("Plot of values of ", hat(N), " obtained from bootstrap samples")),
      sub = "Points will be added in real time",
      xlim = c(0, numboot + 1), ylim = c(0, 2 * N)
    )
  }
  
  #terms <- terms(modelFrame)
  
  while (k <= numboot) {
    if (isTRUE(weightsFlag)) {
      strap        <- sample(x = 1:length(y), 
                             size = n, 
                             prob = weights / n, 
                             replace = TRUE)
      
      ystrap       <- as.numeric(y[strap])
      offsetStrap  <- offset[strap, , drop = FALSE]
      Xstrap       <- modelFrame[strap, , drop = FALSE]
      
      weightsStrap <- rep(1, length(ystrap))
      
    } else {
      strap        <- sample.int(replace = TRUE, n = n)
      
      ystrap       <- as.numeric(y[strap])
      weightsStrap <- as.numeric(weights[strap])
      offsetStrap  <- offset[strap, , drop = FALSE]
      Xstrap       <- modelFrame[strap, , drop = FALSE]
    }
    
    if (!is.data.frame(Xstrap)) {
      Xstrap <- as.data.frame(Xstrap)
      colnames(Xstrap) <- colnames(modelFrame)
    }
    
    Xstrap <- singleRinternalGetXvlmMatrix(X = Xstrap, 
                                           formulas = formulas, 
                                           family$etaNames)
    theta <- NULL
    try(
      theta <- estimatePopsizeFit(
        y = ystrap,
        X = Xstrap,
        family = family,
        control = controlBootstrapMethod,
        method = method,
        priorWeights = weightsStrap,
        coefStart = jitter(beta),
        etaStart = matrix(Xstrap %*% jitter(beta), ncol = NCOL(offsetStrap)) + offsetStrap,
        offset = offsetStrap
      )$beta,
      silent = TRUE
    )
    k <- k + 1
    
    if (is.null(theta)) {
      if (isTRUE(trace)) cat("\n")
      k <- k - 1
    } else {
      theta <- matrix(Xstrap %*% theta, ncol = length(family$etaNames)) + offsetStrap
      
      if (isTRUE(trace)) {print(summary(theta))}
      est <- family$pointEst(pw = weightsStrap, eta = theta, y = ystrap)
      
      if (visT) graphics::points(k - 1, est, pch = 1)
      if (isTRUE(trace)) cat(" Estimated population size: ", est,"\n",sep = "")
      
      strappedStatistic[k - 1] <- est
    }
  }
  
  strappedStatistic
}
# multicore
#' @importFrom foreach %dopar%
#' @importFrom foreach foreach
#' @importFrom parallel makeCluster
#' @importFrom parallel stopCluster
#' @importFrom doParallel registerDoParallel
noparBootMultiCore <- function(family, formulas, y, X, modelFrame,
                                beta, weights, trcount, numboot,
                                eta, cores, controlBootstrapMethod = NULL,
                                method, N, offset, weightsFlag, visT, ...) {
  n <- length(y)
  if (isTRUE(weightsFlag)) {
    n <- sum(weights)
  }
  famName <- family$family
  
  if (length(weights) == 1) {
    weights <- rep(1, n)
  }
  
  cl <- parallel::makeCluster(cores)
  doParallel::registerDoParallel(cl)
  on.exit(parallel::stopCluster(cl))
  
  strappedStatistic <- foreach::`%dopar%`(
    obj = foreach::foreach(k = 1:numboot, .combine = c),
    ex = {
      theta <- NULL
      while (is.null(theta)) {
        if (isTRUE(weightsFlag)) {
          strap        <- sample(x = 1:length(y), 
                                 size = n, 
                                 prob = weights / n, 
                                 replace = TRUE)
          
          ystrap       <- as.numeric(y[strap])
          offsetStrap  <- offset[strap, , drop = FALSE]
          Xstrap       <- modelFrame[strap, , drop = FALSE]
          
          weightsStrap <- rep(1, length(ystrap))
          
        } else {
          strap        <- sample.int(replace = TRUE, n = n)
          
          ystrap       <- as.numeric(y[strap])
          weightsStrap <- as.numeric(weights[strap])
          offsetStrap  <- offset[strap, , drop = FALSE]
          Xstrap       <- modelFrame[strap, , drop = FALSE]
        }
        
        if (!is.data.frame(Xstrap)) {
          Xstrap <- as.data.frame(Xstrap)
          colnames(Xstrap) <- colnames(modelFrame)
        }
        
        Xstrap <- singleRinternalGetXvlmMatrix(X = Xstrap, 
                                               formulas = formulas, 
                                               family$etaNames)
        
        try(
          theta <- estimatePopsizeFit(
            y = ystrap,
            X = Xstrap,
            family = family,
            control = controlBootstrapMethod,
            method = method,
            priorWeights = weightsStrap,
            coefStart = jitter(beta),
            etaStart = matrix(Xstrap %*% jitter(beta), ncol = NCOL(offsetStrap)) + offsetStrap,
            offset = offsetStrap
          )$beta,
          silent = TRUE
        )
      }
      theta <- matrix(Xstrap %*% theta, ncol = length(family$etaNames)) + offsetStrap
      family$pointEst(pw = weightsStrap, eta = theta, y = ystrap)
    }
  )
  
  strappedStatistic
}

#' @rdname singleRmodels
#' @export
ztHurdlegeom <- function(lambdaLink = c("log", "neglog"), 
                         piLink = c("logit", "cloglog", "probit"), 
                         ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  if (missing(piLink))  piLink <- "logit"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink, piLink)
  
  lambdaLink <- switch(lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  piLink <- switch(piLink,
    "logit" = singleRinternallogitLink,
    "cloglog" = singleRinternalcloglogLink,
    "probit" = singleRinternalprobitLink
  )
  
  links[1:2] <- c(lambdaLink, piLink)
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    if (!deriv) {
      switch (type,
        "nontrunc" = (PI + (1 - PI) * (2 + lambda)) * lambda ^ 2 / 
          (lambda ^ 2 + lambda + 1),
        "trunc" = PI + (1 - PI) * (2 + lambda)
      )
    } else {
      switch (type,
        "nontrunc" = {
          matrix(c(
            -lambda * ((PI - 1) * lambda ^ 3 + (2 * PI - 2) * lambda ^ 2 + 
            (4 * PI - 5) * lambda + 2 * PI - 4) / (lambda ^ 2 + lambda + 1) ^ 2,
            -(1 + lambda) * lambda ^ 2 / (lambda ^ 2 + lambda + 1)
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
                piLink(eta[, 2], inverse = TRUE, deriv = 1)
          ), ncol = 2)
        },
        "trunc" = {
          matrix(c(
            1 - PI,
            1 - 2 - lambda
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
                piLink(eta[, 2], inverse = TRUE, deriv = 1)
          ), ncol = 2)
        }
      )
    }
  }
  
  variance <- function(eta, type = "nontrunc", ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    switch (type,
    "nontrunc" = (1 - 1 / (1 + lambda)) * (PI + (1 - PI) * lambda + 2 * lambda ^ 2),
    "trunc" = (PI + (1 - PI) * (lambda * (1 + lambda) + lambda ^ 2 - lambda * (1 + lambda) ^ (-2)) / 
      (1 - 1 / (1 + lambda) - lambda * (1 + lambda) ^ (-2)))
    ) - mu.eta(eta = eta, type = type) ^ 2
  }
  
  Wfun <- function(prior, y, eta, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    z <- PI
    
    ## expected for (1-I)Y
    YY <- mu.eta(eta) - z
    
    G00 <- (-1) * piLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 / ((1 - PI) * PI)
    
    G11 <- lambdaLink(eta[, 1], inverse = TRUE, deriv = 2) * 
      ((YY - (1 - z) * 2) / lambda - (YY - (1 - z)) / (lambda + 1)) -
      lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 *
      ((1 - z) / (lambda * (lambda + 1)) + 
      ((z - 1) * lambda + YY + (z - 1) * 2) / (lambda ^ 2 * (lambda + 1)) + 
      ((z - 1) * lambda + YY + (z - 1) * 2) / (lambda * (lambda + 1) ^ 2))
    
    matrix(
      -c(G11 * prior, # lambda
         rep(0, NROW(eta)) * prior, # mixed
         rep(0, NROW(eta)) * prior, # mixed
         G00 * prior  # pi
      ),
      dimnames = list(rownames(eta), c("lambda", "mixed", "mixed", "pi")),
      ncol = 4
    )
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    z <- as.numeric(y == 1)
    weight <- weight / prior
    
    G1 <- lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * 
      (1 - z) * ((y - 2) / lambda - (y - 1) / (lambda + 1))
    
    G0 <- piLink(eta[, 1], inverse = TRUE, deriv = 1) *  
      (z / PI - (1 - z) / (1 - PI))
    
    
    uMatrix <- matrix(c(G1, G0), ncol = 2)
    
    weight <- lapply(X = 1:nrow(weight), FUN = function (x) {
      matrix(as.numeric(weight[x, ]), ncol = 2)
    })
    
    pseudoResid <- sapply(X = 1:length(weight), FUN = function (x) {
      xx <- diag(weight[[x]])
      xx * uMatrix[x, ]
    })
    pseudoResid <- t(pseudoResid)
    dimnames(pseudoResid) <- dimnames(eta)
    pseudoResid
  }
  
  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, 
                           ...) {
    y <- as.numeric(y)
    if (is.null(weight)) {
      weight <- 1
    }
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X) / 2), rep(0, NROW(X) / 2))
    }
    z <- as.numeric(y == 1)
    
    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
      function(beta) {
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        PI     <- piLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        -sum(weight * (z * log(PI) + (1 - z) * log(1 - PI) +
        (1 - z) * ((y - 2) * log(lambda) - (y - 1) * log(1 + lambda))))
      },
      function(beta) {
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        PI     <- piLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        G1 <- weight * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * 
              (1 - z) * ((y - 2) / lambda - (y - 1) / (lambda + 1))
        
        G0 <- weight * piLink(eta[, 1], inverse = TRUE, deriv = 1) *  
              (z / PI - (1 - z) / (1 - PI))
        
        if (NbyK) {
          XX <- 1:(attr(X, "hwm")[1])
          return(cbind(as.data.frame(X[1:nrow(eta), XX]) * G1, as.data.frame(X[-(1:nrow(eta)), -XX]) * G0))
        }
        if (vectorDer) {
          return(cbind(G1, G0))
        }
        as.numeric(c(G1, G0) %*% X)
      },
      function (beta) {
        lambdaPredNumber <- attr(X, "hwm")[1]
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        PI     <- piLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        res <- matrix(nrow = length(beta), ncol = length(beta), 
                      dimnames = list(names(beta), names(beta)))
        
        # PI^2 derivative
        G00 <- piLink(eta[, 1], inverse = TRUE, deriv = 2) *  
               (z / PI - (1 - z) / (1 - PI)) -
               piLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 *
               (1 / ((1 - PI) * PI) + (PI - z) / ((1 - PI) ^ 2 * PI) + 
               (z - PI) / ((1 - PI) * PI ^ 2))
        
        # lambda^2 derivative
        G11 <- lambdaLink(eta[, 1], inverse = TRUE, deriv = 2) * 
          (1 - z) * ((y - 2) / lambda - (y - 1) / (lambda + 1)) -
          lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 *
          ((1 - z) / (lambda * (lambda + 1)) + 
          ((z - 1) * (lambda - y + 2)) / (lambda ^ 2 * (lambda + 1)) + 
          ((z - 1) * (lambda - y + 2)) / (lambda * (lambda + 1) ^ 2))
        
        
        res[-(1:lambdaPredNumber), -(1:lambdaPredNumber)] <- t(as.data.frame(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)] * G00 * weight)) %*% as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)])
        res[1:lambdaPredNumber, 1:lambdaPredNumber] <- t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber] * G11 * weight)) %*% X[1:(nrow(X) / 2), 1:lambdaPredNumber]
        res[1:lambdaPredNumber, -(1:lambdaPredNumber)] <- 0
        res[-(1:lambdaPredNumber), 1:lambdaPredNumber] <- 0
        
        res
      }
    )
  }
  
  validmu <- function(mu) {
    (sum(!is.finite(mu)) == 0) && all(0 < mu)
  }
  
  devResids <- function(y, eta, wt, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    idealLambda <- ifelse(y > 1, y - 2, 0)
    
    diff <- ifelse(
      y == 1, -log(PI),
      ifelse(y == 2, 0,
      (y - 2) * log(idealLambda) - (y - 1) * log(1 + idealLambda)) - 
      (log(1 - PI) + (y - 2) * log(lambda) - (y - 1) * log(1 + lambda))
    )
    
    diff[diff < 0] <- 0
    sign(y - mu.eta(eta = eta)) * sqrt(2 * wt * diff)
  }
  
  pointEst <- function (pw, eta, contr = FALSE, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    N <- pw * (lambda ^ 2 + lambda + 1) / lambda ^ 2
    if(!contr) {
      N <- sum(N)
    }
    N
  }
  
  popVar <- function (pw, eta, cov, Xvlm, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    # w.r to PI
    bigTheta1 <- rep(0, nrow(eta))
    # w.r to lambda
    bigTheta2 <- (pw * as.numeric(-(lambda + 2) / lambda ^ 3) *
    lambdaLink(eta[, 1], inverse = TRUE, deriv = 1))
    
    bigTheta <- t(c(bigTheta2, bigTheta1) %*% Xvlm)
    
    f1 <-  t(bigTheta) %*% as.matrix(cov) %*% bigTheta
    
    f2 <- sum(pw * (lambda ^ 2 + lambda + 1) / lambda ^ 2)
    
    f1 + f2
  }
  
  dFun <- function (x, eta, type = c("trunc", "nontrunc")) {
    if (missing(type)) type <- "trunc"
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    switch (type,
      "trunc" = {
        as.numeric(x == 1) * PI + as.numeric(x > 0) * 
        (1 - PI) * stats::dnbinom(x = x, mu = lambda, size = 1) / 
        (1 - 1 / (1 + lambda) - lambda * ((1 + lambda) ^ -2))
      },
      "nontrunc" = {
        as.numeric(x == 0) * (1 + lambda) / (lambda ^ 2 + lambda + 1) +
        as.numeric(x == 1) * PI * lambda ^ 2 / (lambda ^ 2 + lambda + 1) +
        as.numeric(x > 1) * (1 - PI) * (lambda ^ x / (1 + lambda) ^ (x + 1)) *
        (lambda + 1) ^ 2 / (lambda ^ 2 + lambda + 1)
      }
    )
  }
  
  simulate <- function(n, eta, lower = 0, upper = Inf) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    CDF <- function(x) {
      p <- lambda / (1 + lambda)
      const <- -p * (lambda * (p ^ x - 1) + p ^ x) / (1 + lambda)
      ifelse(x == Inf, 1, 
      ifelse(x < 0, 0, 
      ifelse(x < 1, (1 + lambda) / (1 + lambda + lambda ^ 2), 
      (1 + lambda) / (1 + lambda + lambda ^ 2) + PI * (lambda ^ 2
      ) / (1 + lambda + lambda ^ 2) +  (1 - PI) * ((1 + lambda) ^ 2
      ) * const / (lambda ^ 2 + lambda + 1))))
    }
    lb <- CDF(lower)
    ub <- CDF(upper)
    p_u <- stats::runif(n, lb, ub)
    sims <- rep(0, n)
    cond <- CDF(sims) <= p_u
    while (any(cond)) {
      sims[cond] <- sims[cond] + 1
      cond <- CDF(sims) <= p_u
    }
    sims
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12)),
        family$links[[2]](weighted.mean(observed == 1, priorWeights) * (.5 + .5 * (observed == 1)) + .01)
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights)),
        family$links[[2]](weighted.mean(observed == 1, priorWeights) + .01)
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
      if ("(Intercept):pi" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[2], rep(0, attr(Xvlm, "hwm")[2] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[2] / attr(Xvlm, "hwm")[2], attr(Xvlm, "hwm")[2]))
      }
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "ztHurdlegeom",
      etaNames = c("lambda", "pi"),
      simulate  = simulate,
      getStart  = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

#N <- 1000
###gender <- rbinom(N, 1, 0.2)
gender <- rep(0:1, c(812, 188))

etaLambda <- -1   + 0.5 * gender
etaAlpha  <- -1.7 + 1 * gender
etaOmega  <- -.2  + 0.8 * gender
etaPi     <- -.3  + 0.9 * gender

expect_silent(
  res <- data.frame(
    ztpoisson = simulate(
      ztpoisson(), 
      eta = cbind(etaLambda), 
      seed = 1
    ),
    ztgeom = simulate(
      ztgeom(), 
      eta = cbind(etaLambda), 
      seed = 1
    ),
    ztnegbin = simulate(
      ztnegbin(), 
      eta = cbind(etaLambda, etaAlpha), 
      seed = 1
    ),
    zotpoisson = simulate(
      zotpoisson(), 
      eta = cbind(etaLambda), 
      seed = 1
    ),
    zotgeom = simulate(
      zotgeom(), 
      eta = cbind(etaLambda), 
      seed = 1
    ),
    zotnegbin = simulate(
      zotnegbin(), 
      eta = cbind(etaLambda, etaAlpha), 
      seed = 1
    ),
    ztoipoisson = simulate(
      ztoipoisson(), 
      eta = cbind(etaLambda, etaOmega), 
      seed = 1
    ),
    ztoigeom = simulate(
      ztoigeom(), 
      eta = cbind(etaLambda, etaOmega), 
      seed = 1
    ),
    ztoinegbin = simulate(
      ztoinegbin(), 
      eta = cbind(etaLambda, etaAlpha, etaOmega), 
      seed = 1
    ),
    oiztpoisson = simulate(
      oiztpoisson(), 
      eta = cbind(etaLambda, etaOmega), 
      seed = 1
    ),
    oiztgeom = simulate(
      oiztgeom(), 
      eta = cbind(etaLambda, etaOmega), 
      seed = 1
    ),
    oiztnegbin = simulate(
      oiztnegbin(), 
      eta = cbind(etaLambda, etaAlpha, etaOmega), 
      seed = 1
    ),
    ztHurdlepoisson = simulate(
      ztHurdlepoisson(), 
      eta = cbind(etaLambda, etaPi), 
      seed = 1
    ),
    ztHurdlegeom = simulate(
      ztHurdlegeom(), 
      eta = cbind(etaLambda, etaPi), 
      seed = 1
    ),
    ztHurdlenegbin = simulate(
      ztHurdlenegbin(), 
      eta = cbind(etaLambda, etaAlpha, etaPi), 
      seed = 1
    ),
    Hurdleztpoisson = simulate(
      Hurdleztpoisson(), 
      eta = cbind(etaLambda, etaPi), 
      seed = 1
    ),
    Hurdleztgeom = simulate(
      Hurdleztgeom(), 
      eta = cbind(etaLambda, etaPi), 
      seed = 1
    ),
    Hurdleztnegbin = simulate(
      Hurdleztnegbin() , 
      eta = cbind(etaLambda, etaAlpha, etaPi), 
      seed = 1
    ),
    chao = simulate(
      chao(), 
      eta = cbind(etaLambda), 
      seed = 1
    ),
    oichao = simulate(
      oichao(),
      eta = cbind(etaLambda),
      truncated = TRUE,
      seed = 1
    ),
    zelterman = simulate(
      zelterman(), 
      eta = cbind(etaLambda), 
      seed = 1
    )
  )
)

expect_true(
  all(res$chao == res$zelterman)
)

expect_true(
  all(res$oichao %in% 2:3)
)

expect_true(
  all(res$zotpoisson == res$ztpoisson)
)

expect_true(
  all(res$zotgeom == res$ztgeom)
)

expect_true(
  all(res$zotnegbin == res$ztnegbin)
)

#' @import mathjaxr
NULL
#' \loadmathjax
#' @title Family functions in singleRcapture package
#' @author Piotr Chlebicki, Maciej Beręsewicz
#'
#' @description Package \code{singleRcapture} utilizes various family type
#' functions that specify variable parts of population size estimation,
#' regression, diagnostics and other necessary information that depends
#' on the model. These functions are used as \code{model} argument in
#' \code{estimatePopsize} function.
#' 
#' @param nSim,epsSim if working weights cannot be computed analytically these 
#' arguments specify maximum number of simulations allowed and
#' precision level for finding them numerically respectively.
#' @param lambdaLink a link for Poisson parameter, \code{"log"} 
#' by default except for rare-count Poisson models such as
#' \code{zelterman()} and \code{chao()} where only
#' \mjseqn{\ln\left(\frac{x}{2}\right)} is possible and \code{oichao()}
#' where only \mjseqn{\ln\left(\frac{x}{3}\right)} is possible.
#' @param alphaLink a link for dispersion parameter, \code{"log"} by default
#' @param omegaLink a link for inflation parameter, \code{"logit"} by default 
#' @param piLink a link for probability parameter,  \code{"logit"} by default
#' @param eimStep a non negative integer describing 
#' how many values should be used at each step of approximation
#' of information matrixes when no analytic solution is available 
#' (e.g. \code{"ztnegbin"}), default varies depending on a function.
#' Higher value usually means faster convergence but may potentially cause
#' issues with convergence.
#' @param ... Additional arguments, not used for now.
#' 
#' @details Most of these functions are based on some "base" distribution with
#' support \mjseqn{\mathbb{N}_{0}=\mathbb{N}\cup\lbrace 0\rbrace} that describe
#' distribution of \mjseqn{Y} before truncation. Currently they include:
#' \mjsdeqn{\mathbb{P}(Y=y|\lambda,\alpha)=\left\lbrace
#' \begin{array}{cc}
#' \frac{\lambda^{y}e^{-\lambda}}{y!}    & \text{Poisson distribution}  \cr
#' \frac{\Gamma(y+\alpha^{-1})}{\Gamma(\alpha^{-1})y!} 
#' \left(\frac{\alpha^{-1}}{\alpha^{-1}+\lambda}\right)^{\alpha^{-1}}
#' \left(\frac{\lambda}{\alpha^{-1}+\lambda}\right)^{y} & 
#' \text{negative binomial distribution} \cr
#' \frac{\lambda^{y}}{(1+\lambda)^{y+1}} & 
#' \text{geometric distribution}
#' \end{array}
#' \right.}
#' where \mjseqn{\lambda} is the Poisson parameter and 
#' \mjseqn{\alpha} is the dispersion parameter. Geometric distribution
#' is a special case of negative binomial distribution when 
#' \mjseqn{\alpha=1} it is included because negative binomial 
#' distribution is quite troublesome numerical regression in fitting.
#' It is important to know that PMF of negative binomial distribution 
#' approaches the PMF of Poisson distribution when 
#' \mjseqn{\alpha\rightarrow 0^{+}}.
#' 
#' 
#' **Note** in literature on single source capture recapture models 
#' the dispersion parameter which introduces greater variability 
#' in negative binomial distribution compared to Poisson distribution is 
#' generally interpreted as explaining the *unobserved* heterogeneity
#' i.e. presence of important unobserved independent variables.
#' All these methods for estimating population size are tied to Poisson 
#' processes hence we use \mjseqn{\lambda} as parameter symbol
#' instead of \mjseqn{\mu} to emphasize this connection.
#' Also will not be hard to see that **all** estimators derived from 
#' modifying the "base" distribution are unbiased if assumptions 
#' made by respective models are not violated.
#' 
#' 
#' The **zero-truncated** models corresponding to "base" distributions are
#' characterized by relation:
#' \mjsdeqn{\mathbb{P}(Y=y|Y>0)=\left\lbrace
#' \begin{array}{cc}
#' \frac{\mathbb{P}(Y=y)}{1-\mathbb{P}(Y=0)} & \text{when }y\neq 0 \cr
#' 0 & \text{when }y=0
#' \end{array}\right.}
#' which allows us to estimate parameter values using only observed part of
#' population. These models lead to the following estimates, respectively:
#' \mjsdeqn{
#' \begin{aligned}
#' \hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1}{1-\exp(-\lambda_{k})} &
#' \text{ For Poisson distribution} \cr
#' \hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1}{1-(1+\\alpha_{k}\lambda_{k})^{-\alpha_{k}^{-1}}} &
#' \text{ For negative binomial distribution} \cr
#' \hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1+\lambda_{k}}{\lambda_{k}} &
#' \text{ For geometric distribution}
#' \end{aligned}
#' }
#' 
#' One common way in which assumptions of zero-truncated models are violated is
#' presence of **one-inflation** the presence of which is somewhat similar in
#' single source capture-recapture models to zero-inflation in usual count data
#' analysis. There are two ways in which one-inflation may be understood,
#' they relate to whether \mjseqn{\mathbb{P}(Y=0)} is 
#' modified by inflation. The first approach is inflate 
#' (\mjseqn{\omega} parameter) zero-truncated distribution as:
#' \mjsdeqn{
#' \mathbb{P}_{new}(Y=y|Y>0) = \left\lbrace\begin{array}{cc}
#' \omega + (1 - \omega)\mathbb{P}_{old}(Y=1|Y>0)& \text{when: } y = 1 \cr
#' (1 - \omega) \mathbb{P}_{old}(Y=y|Y>0) & \text{when: } y \neq 1
#' \end{array}\right.}
#' which corresponds to:
#' \mjsdeqn{
#' \mathbb{P}_{new}(Y=y) = \left\lbrace\begin{array}{cc}
#' \mathbb{P}_{old}(Y=0) & \text{when: } y = 0 \cr
#' \omega(1 - \mathbb{P}(Y=0)) + (1 - \omega)\mathbb{P}_{old}(Y=1) & \text{when: } y = 1 \cr
#' (1 - \omega) \mathbb{P}_{old}(Y=y) & \text{when: } y > 1
#' \end{array}\right.
#' }
#' before zero-truncation. Models that utilize this
#' approach are commonly referred to as *one-inflated zero-truncated models* (`oizt*()` family).
#' Another way of accommodating one-inflation in SSCR is by putting inflation
#' parameter on base distribution as:
#' \mjsdeqn{
#' \mathbb{P}_{new}(Y=y) = \left\lbrace\begin{array}{cc}
#' \omega + (1 - \omega)\mathbb{P}_{old}(Y=1)& \text{when: } y = 1 \cr
#' (1 - \omega) \mathbb{P}_{old}(Y=y) & \text{when: } y \neq 1
#' \end{array}\right.
#' }
#' which then becomes:
#' \mjsdeqn{
#' \mathbb{P}_{new}(Y=y|Y>0) = \left\lbrace\begin{array}{cc}
#' \frac{\omega}{1 - (1-\omega)\mathbb{P}_{old}(Y=0)} + \frac{(1 - \omega)}{1 - (1-\omega)\mathbb{P}_{old}(Y=0)}\mathbb{P}_{old}(Y=1)& \text{when: } y = 1 \cr
#' \frac{(1 - \omega)}{1 - (1-\omega)\mathbb{P}_{old}(Y=0)}\mathbb{P}_{old}(Y=y) & \text{when: } y > 1
#' \end{array}\right.
#' }
#' after truncation and are commonly referred to as *zero-truncated one-inflated models* (`ztoi*()` family).
#' It was shown by Böhning in 2022 paper that these approaches are equivalent 
#' in terms of maximizing likelihoods if we do not put formula on 
#' \mjseqn{\omega}. They can however lead to different 
#' population size estimates.
#' 
#' For *one-inflated zero-truncated models* the formula for population size
#' estimate \mjseqn{\hat{N}} does not change since 
#' \mjseqn{\mathbb{P}(y=0)} remains the same but estimation of parameters 
#' changes all calculations.
#' 
#' For *zero-truncated one-inflated models* population size estimates are 
#' expressed, respectively by:
#' \mjsdeqn{
#' \begin{aligned}
#' \hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1}{1-(1-\omega_{k})\exp(-\lambda_{k})} &\text{ For base Poisson distribution} \cr
#' \hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1}{1-(1-\omega_{k})(1+\\alpha_{k}\lambda_{k})^{-\alpha_{k}^{-1}}} &\text{ For base negative binomial distribution} \cr
#' \hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1+\lambda_{k}}{\lambda_{k} + \omega_{k}} &\text{ For base geometric distribution}
#' \end{aligned}
#' }
#' 
#' **Zero-one-truncated** models ignore one counts instead of accommodating
#' one-inflation by utilizing the identity
#' \mjsdeqn{
#' \ell_{\text{ztoi}}=\boldsymbol{f}_{1}\ln{\frac{\boldsymbol{f}_{1}}{N_{obs}}}
#' +(N_{obs}-\boldsymbol{f}_{1})\ln{\left(1-\frac{\boldsymbol{f}_{1}}{N_{obs}}
#' \right)} + \ell_{\text{zot}}
#' }
#' where \mjseqn{\ell_{\text{zot}}} is the log likelihood 
#' of zero-one-truncated distribution characterized by probability mass function:
#' \mjsdeqn{\mathbb{P}(Y=y|Y>1)=\left\lbrace
#' \begin{array}{cc}
#' \frac{\mathbb{P}(Y=y)}{1-\mathbb{P}(Y=0)-\mathbb{P}(Y=1)} & \text{when }y > 1 \cr
#' 0 & \text{when }y\in\lbrace 0, 1\rbrace
#' \end{array}\right.}
#' where \mjseqn{\mathbb{P}(Y)} is the probability mass function of 
#' the "base" distribution. The identity above justifies use of zero-one-truncated,
#' unfortunately it was only proven for intercept only models, however
#' numerical simulations seem to indicate that even if the theorem cannot be
#' extended for (non trivial) regression population size estimation is still
#' possible. 
#' 
#' For *zero-one-truncated models* population size estimates are expressed by:
#' \mjsdeqn{
#' \begin{aligned}
#' \hat{N} &= \boldsymbol{f}_{1} + \sum_{k=1}^{N_{obs}}
#' \frac{1-\lambda_{k}\exp(-\lambda_{k})}{1-\exp(-\lambda_{k})-\lambda_{k}\exp(-\lambda_{k})} 
#' &\text{ For base Poisson distribution} \cr
#' \hat{N} &= \boldsymbol{f}_{1} + \sum_{k=1}^{N_{obs}}
#' \frac{1-\lambda_{k}(1+\alpha_{k}\lambda_{k})^{-1-\alpha_{k}^{-1}}}{
#' 1-(1+\alpha_{k}\lambda_{k})^{-\alpha_{k}^{-1}}-\lambda_{k}(1+\alpha_{k}\lambda_{k})^{-1-\alpha_{k}^{-1}}} 
#' &\text{ For base negative binomial distribution} \cr
#' \hat{N} &= \boldsymbol{f}_{1} + \sum_{k=1}^{N_{obs}}
#' \frac{\lambda_{k}^{2}+\lambda_{k}+1}{\lambda_{k}^{2}} 
#' &\text{ For base geometric distribution}
#' \end{aligned}
#' }
#' 
#' Pseudo hurdle models are experimental and not yet described in literature.
#' 
#' Lastly there are **chao** and **zelterman** models which are based on 
#' logistic regression on the dummy variable
#' \mjsdeqn{
#' Z = \left\lbrace\begin{array}{cc}
#' 0     & \text{if }Y = 1  \cr
#' 1     & \text{if }Y = 2
#' \end{array}\right.}
#' based on the equation:
#' \mjsdeqn{
#' \text{logit}(p_{k})=
#' \ln\left(\frac{\lambda_{k}}{2}\right)=
#' \boldsymbol{\beta}\mathbf{x}_{k}=\eta_{k}}
#' where \mjseqn{\lambda_{k}} is the Poisson parameter.
#' 
#' The *zelterman* estimator of population size is expressed as:
#' \mjsdeqn{\hat{N}=\sum_{k=1}^{N_{obs}}{1-\exp\left(-\lambda_{k}\right)}}
#' and *chao* estimator has the form:
#' \mjsdeqn{
#' \hat{N}=N_{obs}+\sum_{k=1}^{\boldsymbol{f}_{1}+\boldsymbol{f}_{2}}
#' \frac{1}{\lambda_{k}+ \frac{\lambda_{k}^{2}}{2}}
#' }
#' The *oichao* estimator is the modified Chao analogue robust to one-inflation
#' and is based on the dummy variable
#' \mjsdeqn{
#' Z = \left\lbrace\begin{array}{cc}
#' 0     & \text{if }Y = 2  \cr
#' 1     & \text{if }Y = 3
#' \end{array}\right.}
#' with equation:
#' \mjsdeqn{
#' \text{logit}(p_{k})=
#' \ln\left(\frac{\lambda_{k}}{3}\right)=
#' \boldsymbol{\beta}\mathbf{x}_{k}=\eta_{k}}
#' and population size estimator:
#' \mjsdeqn{
#' \hat{N}=N_{obs}+\sum_{k=1}^{\boldsymbol{f}_{2}+\boldsymbol{f}_{3}}
#' \frac{1}{\frac{\lambda_{k}^{2}}{2}+ \frac{\lambda_{k}^{3}}{6}}
#' }
#' 
#' @seealso [estimatePopsize()]
#'
#' @return A object of class \code{family} containing objects:
#' \itemize{
#' \item \code{makeMinusLogLike} -- A factory function for creating the following functions:
#' \mjseqn{\ell(\boldsymbol{\beta}), \frac{\partial\ell}{\partial\boldsymbol{\beta}},
#' \frac{\partial^{2}\ell}{\partial\boldsymbol{\beta}^{T}\partial\boldsymbol{\beta}}
#' } functions from the
#' \mjseqn{\boldsymbol{y}} vector and the
#' \mjseqn{\boldsymbol{X}_{vlm}} matrix
#' (or just \mjseqn{\boldsymbol{X}} if applied to model 
#' with single linear predictor)which has the \code{deriv} argument with possible 
#' values in \code{c(0, 1, 2)} that determine which derivative to return; 
#' the default value is \code{0}, which represents the minus log-likelihood.
#' \item \code{links} -- A list with link functions.
#' \item \code{mu.eta, variance} -- Functions of linear predictors that
#' return expected value and variance. The \code{type} argument with 2 possible 
#' values (\code{"trunc"} and \code{"nontrunc"}) that specifies whether
#' to return \mjseqn{\mathbb{E}(Y|Y>0), \text{var}(Y|Y>0)} or 
#' \mjseqn{\mathbb{E}(Y), \text{var}(Y)} respectively; the \code{deriv} argument 
#' with values in \code{c(0, 1, 2)} is used for indicating the derivative with 
#' respect to the linear predictors, which is used for providing 
#' standard errors in the \code{predict} method.
#' \item \code{family} -- A string that specifies name of the model.
#' \item \code{valideta, validmu} -- For now it only returns \code{TRUE}. 
#' In the near future, it will be used to check whether applied linear 
#' predictors are valid (i.e. are transformed into some elements of the 
#' parameter space subjected to the inverse link function).
#' \item \code{funcZ, Wfun} -- Functions that create pseudo residuals and
#' working weights used in IRLS algorithm.
#' \item \code{devResids} -- A function that given the linear predictors
#' prior weights vector and response vector returns deviance residuals.
#' Not all family functions have these functions implemented yet.
#' \item \code{pointEst, popVar} -- Functions that given prior weights
#' linear predictors and in the latter case also estimate of 
#' \mjseqn{\text{cov}(\hat{\boldsymbol{\beta}})} and \mjseqn{\boldsymbol{X_{vlm}}}
#' matrix return point estimate for population size and analytic estimation 
#' of its variance.There is a additional boolean parameter \code{contr} in the 
#' former function that if set to true returns contribution of each unit.
#' \item \code{etaNames} -- Names of linear predictors.
#' \item \code{densityFunction} -- A function that given linear predictors 
#' returns value of PMF at values \code{x}. Additional argument \code{type}
#' specifies whether to return \mjseqn{\mathbb{P}(Y|Y>0)} or
#' \mjseqn{\mathbb{P}(Y)}.
#' \item \code{simulate} -- A function that generates values of a dependent 
#' vector given linear predictors.
#' \item \code{getStart} -- An expression for generating starting points.
#' }
#' @name singleRmodels
NULL

#' \loadmathjax
#' @title Regression diagnostics in singleRcapture
#' @author Piotr Chlebicki, Maciej Beręsewicz
#' 
#'
#' @description List of some regression diagnostics implemented for
#' \code{singleRStaticCountData} class. Functions that either require no changes from 
#' \code{glm} class or are not relevant to context of \code{singleRcapture}
#' are omitted.
#' 
#' @param model,object an object of \code{singleRStaticCountData} class.
#' @param dfbeta if \code{dfbeta} was already obtained it is possible to pass 
#' them into function so that they need not be computed for the second time.
#' @param do.coef logical indicating if \code{dfbeta} computation for influence 
#' should be done. \code{FALSE} by default.
#' @param cores a number of processor cores to be used,
#' any number greater than 1 activates code designed with \code{doParallel}, 
#' \code{foreach} and \code{parallel} packages. Note that for now using parallel 
#' computing makes tracing impossible so \code{trace} parameter is ignored in this case.
#' @param type a type of residual to return.
#' @param trace a logical value specifying whether to tracking results when
#' \code{cores > 1} it will result in a progress bar being created.
#' @param maxitNew the maximal number of iterations for regressions with starting 
#' points \mjseqn{\hat{\boldsymbol{\beta}}} on data 
#' specified at call for \code{model} after the removal of k'th row. By default 1.
#' @param ... arguments passed to other methods. 
#' Notably \code{dfpopsize.singleRStaticCountData} calls 
#' \code{dfbeta.singleRStaticCountData} if no \code{dfbeta} argument was 
#' provided and \code{controlMethod} is called in \code{dfbeta} method.
#' 
#' @details 
#' 
#' \code{dfpopsize} and \code{dfbeta} are closely related. \code{dfbeta} 
#' fits a regression after removing a specific row from the data and returns the 
#' difference between regression coefficients estimated on full data set and
#' data set obtained after deletion of that row, and repeats procedure once
#' for every unit present in the data.\code{dfpopsize} does the same for 
#' population size estimation utilizing coefficients computed by \code{dfbeta}.
#' 
#' \code{cooks.distance} is implemented (for now) only for models with a single
#' linear predictor and works exactly like the method for \code{glm} class.
#' 
#' \code{sigma} computes the standard errors of predicted means. Returns a matrix
#' with two columns first for truncated mean and the other for the non-truncated mean.
#' 
#' \code{residuals.singleRStaticCountData} (can be abbreviated to \code{resid})
#'  works like \code{residuals.glm} with the exception that:
#' \itemize{
#' \item \code{"pearson"} -- returns non standardized residuals.
#' \item \code{"pearsonSTD"} -- is currently defined only for single predictors
#' models but will be extended to all models in a near future, but for families 
#' with more than one distribution parameter it will be a multivariate residual.
#' \item \code{"response"} -- returns both residuals computed with truncated
#' and non truncated fitted value.
#' \item \code{"working"} -- is possibly multivariate if more than one linear 
#' predictor is present.
#' \item \code{"deviance"} -- is not yet defined for all families in 
#' [singleRmodels()] e.g. negative binomial based methods.
#' \item \code{"all"} -- returns all available residual types.
#' }
#' 
#' 
#' \code{hatvalues.singleRStaticCountData} is method for \code{singleRStaticCountData} 
#' class for extracting diagonal elements of projection matrix. 
#' 
#' Since \code{singleRcapture} supports not only regular glm's but also vglm's the 
#' \code{hatvalues} returns a matrix with number of columns corresponding to number 
#' of linear predictors in a model, where kth column corresponds to elements of 
#' the diagonal of projection matrix associated with kth linear predictor. 
#' For glm's  
#' \mjsdeqn{\boldsymbol{W}^{\frac{1}{2}}\boldsymbol{X}
#' \left(\boldsymbol{X}^{T}\boldsymbol{W}\boldsymbol{X}\right)^{-1}
#' \boldsymbol{X}^{T}\boldsymbol{W}^{\frac{1}{2}}}
#' where: \mjseqn{\boldsymbol{W}=\mathbb{E}\left(\text{Diag}
#' \left(\frac{\partial^{2}\ell}{\partial\boldsymbol{\eta}^{T}
#' \partial\boldsymbol{\eta}}\right)\right)}
#' and \mjseqn{\boldsymbol{X}} is a model (lm) matrix. 
#' For vglm's present in the package it is instead :
#' \mjsdeqn{\boldsymbol{X}_{vlm}
#' \left(\boldsymbol{X}_{vlm}^{T}\boldsymbol{W}\boldsymbol{X}_{vlm}\right)^{-1}
#' \boldsymbol{X}_{vlm}^{T}\boldsymbol{W}}
#' where:
#' \mjsdeqn{
#' \boldsymbol{W} = \mathbb{E}\left(\begin{bmatrix}
#' \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{1}^{T}\partial\eta_{1}}\right) &
#' \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{1}^{T}\partial\eta_{2}}\right) &
#' \dotso & \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{1}^{T}\partial\eta_{p}}\right)\cr
#' \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{2}^{T}\partial\eta_{1}}\right) &
#' \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{2}^{T}\partial\eta_{2}}\right) &
#' \dotso & \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{2}^{T}\partial\eta_{p}}\right)\cr
#' \vdots & \vdots & \ddots & \vdots\cr
#' \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{p}^{T}\partial\eta_{1}}\right) &
#' \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{p}^{T}\partial\eta_{2}}\right) &
#' \dotso & \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{p}^{T}\partial\eta_{p}}\right)
#' \end{bmatrix}\right)}
#' is a block matrix constructed by taking the expected  value from diagonal 
#' matrixes corresponding to second derivatives with respect to each linear 
#' predictor (and mixed derivatives) and 
#' \mjseqn{\boldsymbol{X}_{vlm}} is a model (vlm) 
#' matrix constructed using specifications in \code{controlModel} and 
#' call to \code{estimatePopsize}.
#'
#' \code{influence} works like \code{glm} counterpart computing the most important
#' influence measures.
#'
#' @seealso [estimatePopsize()] [stats::hatvalues()] [controlMethod()] [stats::dfbeta()]
#' [stats::cooks.distance()]
#' @return \itemize{
#' \item For \code{hatvalues} -- A matrix with n rows and p columns where n is a 
#' number of observations in the data and p is number of regression parameters.
#' \item For \code{dfpopsize} -- A vector for which k'th element corresponds
#' to the difference between point estimate of population size estimation on 
#' full data set and point estimate of population size estimation after the 
#' removal of k'th unit from the data set.
#' \item For \code{dfbeta} -- A matrix with n rows and p observations where p 
#' is a number of units in data and p is the number of regression parameters. 
#' K'th row of this matrix corresponds to 
#' \mjseqn{\hat{\boldsymbol{\beta}}-\hat{\boldsymbol{\beta}}_{-k}}
#' where \mjseqn{\hat{\boldsymbol{\beta}}_{-k}} is a vector of estimates for 
#' regression parameters after the removal of k'th row from the data.
#' \item \code{cooks.distance} -- A matrix with a single columns with
#' values of cooks distance for every unit in \code{model.matrix}
#' \item \code{residuals.singleRStaticCountData} -- A \code{data.frame} 
#' with chosen residuals.
#' }
#' 
#' @examples
#' \donttest{
#' # For singleRStaticCountData class
#' # Get simple model
#' Model <- estimatePopsize(
#'   formula = capture ~ nation + age + gender, 
#'   data = netherlandsimmigrant, 
#'   model = ztpoisson, 
#'   method = "IRLS"
#' )
#' # Get dfbeta
#' dfb <- dfbeta(Model)
#' # The dfpopsize results are obtained via (It is also possible to not provide 
#' # dfbeta then they will be computed manually):
#' res <- dfpopsize(Model, dfbeta = dfb)
#' summary(res)
#' plot(res)
#' # see vaious types of residuals:
#' head(resid(Model, "all"))
#' }
#' @name regDiagSingleR
NULL

#' \loadmathjax
#' @title Summary statistics for model of singleRStaticCountData class
#' 
#' @description
#' A \code{summary} method for \code{singleRStaticCountData} class
#' 
#' 
#' @details Works 
#' analogically to \code{summary.glm} but includes population size estimation 
#' results. If any additional statistics, such as confidence intervals for
#' coefficients or coefficient correlation, are specified they will be printed.
#' 
#' @param object object of singleRStaticCountData class.
#' @param test type of test for significance of parameters \code{"t"} for t-test 
#' and \code{"z"} for normal approximation of students t distribution, by 
#' default \code{"z"} is used if there are more than 30 degrees of freedom
#' and \code{"t"} is used in other cases.
#' @param resType type of residuals to summarize any value that is allowed in
#' \code{residuals.singleRStaticCountData} except for \code{"all"} is allowed. By default 
#' pearson residuals are used.
#' @param correlation logical value indicating whether correlation matrix should
#' be computed from covariance matrix by default \code{FALSE}.
#' @param confint logical value indicating whether confidence intervals for
#' regression parameters should be constructed. By default \code{FALSE}.
#' @param cov covariance matrix corresponding to regression parameters. 
#' It is possible to give \code{cov} argument as a function of \code{object}.
#' If not specified it will be constructed using \code{vcov.singleRStaticCountData} method.
#' (i.e using Cramer-Rao lower bound)
#' @param popSizeEst a \code{popSizeEstResults} class object.
#' If not specified population size estimation results will be drawn from
#' \code{object}. If any post-hoc procedures, such as sandwich covariance matrix 
#' estimation or bias reduction, were taken it is possible to include them in 
#' population size estimation results by calling \code{redoPopEstimation}.
#' @param ... additional optional arguments passed to the following functions:
#' \itemize{
#' \item \code{vcov.singleRStaticCountData} -- if no \code{cov} argument was provided.
#' \item \code{cov} -- if \code{cov} parameter specified at call was a function.
#' \item \code{confint.singleRStaticCountData} -- if \code{confint} parameter was set to \code{TRUE} at function call.
#' In particular it is possible to set confidence level in \code{...}.
#' }
#' 
#' @return An object of \code{summarysingleRStaticCountData} class containing:
#' \itemize{
#' \item \code{call} -- A call which created \code{object}.
#' \item \code{coefficients} -- A dataframe with estimated regression coefficients
#' and their summary statistics such as standard error Wald test statistic and
#' p value for Wald test.
#' \item \code{residuals} -- A vector of residuals of type specified at call.
#' \item \code{aic} -- Akaike's information criterion.
#' \item \code{bic} -- Bayesian (Schwarz's) information criterion.
#' \item \code{iter} -- Number of iterations taken in fitting regression.
#' \item \code{logL} -- Logarithm of likelihood function evaluated at coefficients.
#' \item \code{deviance} -- Residual deviance.
#' \item \code{populationSize} -- Object with population size estimation results.
#' \item \code{dfResidual} -- Residual degrees of freedom.
#' \item \code{sizeObserved} -- Size of observed population.
#' \item \code{correlation} -- Correlation matrix if \code{correlation} parameter was set to \code{TRUE}
#' \item \code{test} -- Type of statistical test performed.
#' \item \code{model} -- Family class object specified in call for \code{object}.
#' \item \code{skew} -- If bootstrap sample was saved contains estimate of skewness.
#' }
#' @method summary singleRStaticCountData
#' @importFrom stats pt
#' @importFrom stats coef
#' @importFrom stats sd
#' @seealso [redoPopEstimation()] [stats::summary.glm()]
#' @exportS3Method 
summary.singleRStaticCountData <- function(object, 
                                           test = c("t", "z"), 
                                           resType = "pearson", 
                                           correlation = FALSE, 
                                           confint = FALSE, 
                                           cov, 
                                           popSizeEst, 
                                           ...) {
  if (resType == "all") {stop("Can't use 'resType = all' in summary.singleRStaticCountData method, if you wish to obtain all aviable types of residuals call residuals.singleRStaticCountData method directly.")}
  dfResidual <- object$dfResidual
  if (missing(test)) {if (dfResidual > 30) test <- "z" else test <- "t"}
  if (missing(cov)) {
    cov <- vcov(object, ...)
  } else if (is.function(cov)) {
    cov <- cov(object, ...)
  }
  
  pers <- residuals(object, type = resType)
  
  cf <- stats::coef(object)
  se <- sqrt(diag(cov))
  
  wValues <- cf / se
  
  pValues <- switch (test,
    "t" = 2 *    stats::pt(q = -abs(wValues), df = dfResidual),
    "z" = 2 * stats::pnorm(q = abs(wValues), lower.tail = FALSE)
  )
  
  crr <- if (isFALSE(correlation)) {NULL} else {cov / outer(se, se)}
  
  if(isTRUE(correlation)) {rownames(crr) <- colnames(crr) <- names(cf)}
  
  cnfint <- if(isTRUE(confint)) {confint(object, ...)} else {NULL}
  
  if (is.numeric(object$populationSize$boot)) {
    n <- length(object$populationSize$boot)
    m <- sum((object$populationSize$boot - mean(object$populationSize$boot)) ^ 3) / n
    s <- sd(object$populationSize$boot)
    skew <- m / (s ^ 3)
  } else {
    skew <- NULL
  }
  
  ob <- data.frame(cf, se, wValues, pValues)
  
  colnames(ob) <- switch(test,
    "t" = c("Estimate", "Std. Error", "t value", "P(>|t|)"),
    "z" = c("Estimate", "Std. Error", "z value", "P(>|z|)")
  )
  if (isTRUE(confint)) {
    ob[, 4] <- cnfint[, 1]
    ob[, 5] <- cnfint[, 2]
    ob[, 6] <- pValues
    colnames(ob)[4:6] <- c(colnames(cnfint), colnames(ob)[4])
  }
  
  cf <- ob
  structure(
    list(
      call = object$call,
      coefficients = cf,
      residuals = pers,
      aic = AIC(object, ...),
      bic = BIC(object, ...),
      iter = object$iter,
      logL = object$logL,
      deviance = object$deviance,
      populationSize = if (missing(popSizeEst)) object$populationSize else popSizeEst,
      dfResidual = dfResidual,
      sizeObserved = object$sizeObserved,
      correlation = crr,
      test = test,
      model = object$model,
      skew = skew
    ),
    class = "summarysingleRStaticCountData"
  )
}
#' @method print summarysingleRStaticCountData
#' @importFrom stats printCoefmat
#' @exportS3Method 
print.summarysingleRStaticCountData <- function(x, 
                                                signif.stars = getOption("show.signif.stars"), 
                                                digits = max(3L, getOption("digits") - 3L), 
                                                ...) {
  cat("\nCall:\n", paste(deparse(x$call), sep = "\n", collapse = "\n"), 
      "\n\n", sep = "")
  
  cat("Pearson Residuals:\n")
  print(summary(c(x$residuals[, 1])))
  
  cat("\nCoefficients:\n")
  
  cond <- matrix(sapply(
    x$model$etaNames, 
    FUN = function(k) {
      sapply(strsplit(rownames(x$coefficients), 
                      split = ":"), 
             FUN = function(x) x[[length(x)]] == k)
    }
  ), 
  ncol = length(x$model$etaNames), 
  dimnames = list(
    NULL,
    x$model$etaNames
    )
  )
  
  for (k in x$model$etaNames) {
    if (!all(cond[, k] == FALSE)) {
      cat("-----------------------\nFor linear predictors associated with:", k, "\n")
      # Base library has no str_length and no proper sub_string this is a 
      # slightly convoluted way of making do without them
      toPrint <- subset(x$coefficients, cond[,k])
      lengths <- sapply(rownames(toPrint), 
                        function(x) {length(strsplit(x, split = "")[[1]])})
      lK <- length(unlist(strsplit(k, split = "")))
      rownames(toPrint) <- sapply(
        1:nrow(toPrint), 
        function(x) {substr(
          x = rownames(toPrint)[x], 
          start = 1, 
          stop = lengths[[x]] - (1 + lK)
        )}
      )
      printCoefmat(
        toPrint, digits = digits, 
        signif.stars = signif.stars, 
        signif.legend = if (k == x$model$etaNames[length(x$model$etaNames)]) signif.stars else FALSE, 
        P.values = TRUE, has.Pvalue = TRUE, 
        na.print = "NA", 
        ...
      )
    } else {
      cat("-----------------------\nFor linear predictors associated with:", k, "\n")
      printCoefmat(
        subset(x$coefficients, rowSums(cond) == 0), 
        digits = digits, 
        signif.stars = signif.stars, 
        signif.legend = if (k == x$model$etaNames[length(x$model$etaNames)]) signif.stars else FALSE, 
        P.values = TRUE, has.Pvalue = TRUE, 
        na.print = "NA", 
        ...
      )
    }
  }
  
  cat("\n")
  
  if (!is.null(x$correlation)) {
    corr <- round(x$correlation, digits = 2)
    corr[!lower.tri(corr)] <- ""
    print(corr[-1, -dim(corr)[2]], quote = FALSE)
    cat("\n")
  }
  
  sd <- sqrt(x$populationSize$variance)
  
  if (x$populationSize$control$sd == "normalMVUE") {
    sd <- sd / (sqrt(2 / (x$sizeObserved - 1)) * 
    exp(lgamma(x$sizeObserved / 2) - lgamma((x$sizeObserved - 1) / 2)))
  }
  
  cat(
    "AIC: ", x$aic,
    "\nBIC: ", x$bic,
    "\nResidual deviance: ", x$deviance,
    "\n\nLog-likelihood: ", x$logL, " on ", x$dfResidual, " Degrees of freedom ",
    # optim does not allow for accessing information
    # on number of iterations performed only a number 
    # of calls for gradient and objective function
    if (isTRUE(x$call$method == "optim")) "\nNumber of calls to log-likelihood function: "
    else "\nNumber of iterations: " , x$iter[1], 
    "\n-----------------------",
    "\nPopulation size estimation results: ",
    "\nPoint estimate ", x$populationSize$pointEstimate, 
    "\nObserved proportion: ", round(100 * x$sizeObserved / x$populationSize$pointEstimate, 
                                     digits = 1), 
    "% (N obs = ", x$sizeObserved, ")",
    if (!is.null(x$skew)) "\nBoostrap sample skewness: ", 
    if (!is.null(x$skew)) x$skew, 
    if (!is.null(x$skew)) "\n0 skewness is expected for normally distributed variable\n---",
    if (isTRUE(x$call$popVar == "bootstrap")) "\nBootstrap Std. Error " else "\nStd. Error ", 
    sd, "\n", 
    (1 - x$populationSize$control$alpha) * 100, 
    "% CI for the population size:\n", 
    sep = ""
  )
  
  print(x$populationSize$confidenceInterval)
  
  cat((1 - x$populationSize$control$alpha) * 100, 
      "% CI for the share of observed population:\n", 
      sep = "")
  
  dd <- as.data.frame(x$populationSize$confidenceInterval)
  
  if (ncol(dd) == 1) {
    vctpop <- sort(x$populationSize$confidenceInterval, decreasing = TRUE)
    names(vctpop) <- rev(names(vctpop))
    print(100 * x$sizeObserved / vctpop)
  } else {
    print(data.frame(
      lowerBound = 100 * x$sizeObserved / dd[, 2], 
      upperBound = 100 * x$sizeObserved / dd[, 1],
      row.names = rownames(dd)
    ))
  }
  
  
  invisible(x)
}

#' @rdname singleRmodels
#' @importFrom lamW lambertW0
#' @importFrom stats weighted.mean
#' @export
ztpoisson <- function(lambdaLink = c("log", "neglog"),
                      ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink)
  
  lambdaLink <- switch (lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  links[1] <- c(lambdaLink)
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    if (!deriv) {
      switch (type,
        "nontrunc" = lambda,
        "trunc" = lambda / (1 - exp(-lambda))
      )
    } else {
      switch (type,
        "nontrunc" = cbind(lambdaLink(eta, inverse = TRUE, deriv = 1)),
        "trunc" = cbind(
          lambdaLink(eta, inverse = TRUE, deriv = 1) *
            (exp(lambda) * (exp(lambda) - lambda - 1)) / (exp(lambda) - 1) ^ 2
        )
      )
    }
  }

  variance <- function(eta, type = "nontrunc", ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    switch (type,
    "nontrunc" = lambda,
    "trunc" = mu.eta(eta = eta) * (1 + lambda - mu.eta(eta = eta))
    )
  }
  
  Wfun <- function(prior, eta, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    Ey <- mu.eta(eta)
    
    G11 <- (exp(2 * lambda) / (exp(lambda) - 1) ^ 2 -
    exp(lambda) / (exp(lambda) - 1) - Ey / lambda ^ 2) *
    lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2
    
    matrix(-G11 * prior, ncol = 1, 
           dimnames = list(rownames(eta), c("lambda")))
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    prior * ((y / lambda - 1 / (1 - exp(-lambda))) * 
    lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) / weight)
  }

  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, 
                           ...) {
    y <- as.numeric(y)
    if (is.null(weight)) {
      weight <- 1
    }
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X)))
    }
    
    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
      function(beta) {
        lambda <- lambdaLink((as.matrix(X) %*% beta + offset)[, 1], 
                             inverse = TRUE)
        
        -sum(weight * (y * log(lambda) - log(exp(lambda) - 1) - lgamma(y + 1)))
      },
      function(beta) {
        eta <- as.matrix(X) %*% beta + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        G1 <- (y / lambda - 1 / (1-exp(-lambda))) * weight * 
               lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
        
        if (NbyK) {
          return(as.data.frame(X) * G1)
        }
        if (vectorDer) {
          return(matrix(G1, ncol = 1))
        }
        t(as.matrix(X)) %*% G1
      },
      function(beta) {
        eta <- as.matrix(X) %*% beta + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        G1 <- (y / lambda - 1 / (1 - exp(-lambda))) * 
               lambdaLink(eta[, 1], inverse = TRUE, deriv = 2)
        
        G11 <- (exp(2*lambda) / (exp(lambda) - 1) ^ 2 -
                exp(lambda) / (exp(lambda) - 1) - y / lambda ^ 2) *
                lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2
        
        t(as.matrix(X) * weight * (G1 + G11)) %*% as.matrix(X)
      }
    )
  }

  validmu <- function(mu) {
    (sum(!is.finite(mu)) == 0) && all(0 < mu)
  }

  devResids <- function(y, eta, wt, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    idealLambda <- tryCatch(
      expr = {
        suppressWarnings(
          ifelse(y > 1, lamW::lambertW0(-y * exp(-y)) + y, 0)
        )
      },
      error = function (e) {
        warning("Deviance residuals could not have been computed and zero vector will be returned instead.", call. = FALSE)
        NULL
      }
    )
    if (is.null(idealLambda)) {
      return(rep(0, length(y)))
    }
    
    
    logL <- y * log(lambda) - lambda - log(1 - exp(-lambda))
    logSat <- y * ifelse(y > 1, log(idealLambda), 0) - idealLambda - ifelse(y > 1, log(1 - exp(-idealLambda)), 0)
    
    diff <- (y * log(lambda) - lambda - log(1 - exp(-lambda)) - 
    y * ifelse(y > 1, log(idealLambda), 0) - 
    idealLambda - ifelse(y > 1, log(1 - exp(-idealLambda)), 0)
    )
    
    if (any(diff > 0)) {
      warning(paste0(
        "Some of differences between log likelihood in saturated model",
        " and fitted model were positive which indicates either:\n",
        "(1): A very good model fit or\n",
        "(2): Incorrect computation of saturated model",
        "\nDouble check deviance before proceeding"
      ))
    }
    
    diff[diff > 0] <- -0
    
    ### Here we take pairwise minimum because in specific situations
    ### lambda and idealLambda are so close for some units that
    ### their respective likelihoods differ only by machine epsilon
    ### and rounding may cause warnings
    
    ### TLDR:: pmin must be here not because mathematical error, 
    ### rather because of a rounding error
    sign(y - mu.eta(eta = eta)) * sqrt(-2 * wt * pmin(0, logL - logSat))
  }

  pointEst <- function (pw, eta, contr = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    N <- pw / (1 - exp(-lambda))
    if(!contr) {
      N <- sum(N)
    }
    N
  }

  popVar <- function (pw, eta, cov, Xvlm, ...) {
    Xvlm <- as.data.frame(Xvlm)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)

    bigTheta <- -(lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * pw * 
                  exp(lambda) / (exp(lambda) - 1) ^ 2) %*% as.matrix(Xvlm)
    bigTheta <- as.vector(bigTheta)
    
    f1 <- t(bigTheta) %*% as.matrix(cov) %*% bigTheta

    f2 <- sum(pw * exp(-lambda) / ((1 - exp(-lambda)) ^ 2))

    f1 + f2
  }
  
  simulate <- function(n, eta, lower = 0, upper = Inf) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    lb <- stats::ppois(lower, lambda)
    ub <- stats::ppois(upper, lambda)
    p_u <- stats::runif(n, lb, ub)
    sims <- stats::qpois(p_u, lambda)
    sims
  }
  
  dFun <- function (x, eta, type = c("trunc", "nontrunc")) {
    if (missing(type)) type <- "trunc"
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    switch (type,
      "trunc" = {
        stats::dpois(x = x, lambda = lambda) / 
        (1 - stats::dpois(x = 0, lambda = lambda))
      },
      "nontrunc" = stats::dpois(x = x, lambda = lambda)
    )
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12))
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights))
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
    }
  )
  
  ratioFunc <- function(x) {x+1}
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "ztpoisson",
      etaNames  = c("lambda"),
      simulate  = simulate,
      getStart  = getStart,
      ratioFunc = ratioFunc,
      extraInfo = c(
        mean       = "lambda",
        variance   = "lambda",
        popSizeEst = "(1 + exp(-lambda)) ^ -1",
        meanTr     = "lambda / (1 - exp(-lambda))",
        varianceTr = 
          "(lambda / (1 - exp(-lambda))) * (1 + lambda - lambda / (1 - exp(-lambda)))"
      )
    ),
    class = c("singleRfamily", "family")
  )
}

source("helper_families.R")
zerotruncated_test_data <- load_zerotruncated_test_data()
eps <- zerotruncated_test_data$eps
N <- zerotruncated_test_data$N
test_zerotruncated <- zerotruncated_test_data$test_zerotruncated
x1 <- zerotruncated_test_data$x1
x2 <- zerotruncated_test_data$x2
x3 <- zerotruncated_test_data$x3

## poisson ####
coefficients_ztpoisson <- c(.6, .2, -1.25, .1)
#eta <- cbind(coefficients_ztpoisson[1] + coefficients_ztpoisson[2] * x1 + coefficients_ztpoisson[3] * x2 + coefficients_ztpoisson[4] * x3)

model_ztpoisson <- ztpoisson(
  lambdaLink = "neglog"
)

#y <- model_ztpoisson$simulate(n = N, eta = eta, lower = -1)
observed_ztpoisson <- test_zerotruncated$ztpoisson1

data_ztpoisson <- data.frame(
  x3 = x3[observed_ztpoisson > 0],
  x2 = x2[observed_ztpoisson > 0],
  x1 = x1[observed_ztpoisson > 0],
  y = observed_ztpoisson[observed_ztpoisson > 0]
)

fit_ztpoisson <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = model_ztpoisson,
  data    = data_ztpoisson,
  method = "IRLS"
)

expect_equivalent(
  coef(fit_ztpoisson),
  coefficients_ztpoisson,
  tolerance = .25
)

expect_silent(
  population_ztpoisson <- popSizeEst(fit_ztpoisson)
)

expect_equivalent(
  N,
  population_ztpoisson$pointEstimate,
  tol = .15
)

expect_true(
  all(N < population_ztpoisson$confidenceInterval$upperBound) &
  all(N > population_ztpoisson$confidenceInterval$lowerBound)
)

expect_silent(
  strata_ztpoisson <- stratifyPopsize(fit_ztpoisson, ~ x1)
)

expect_equivalent(
  as.numeric(table(data_ztpoisson$x1))[as.numeric(substr(strata_ztpoisson$name, start = 5, stop = 5)) + 1],
  strata_ztpoisson$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(strata_ztpoisson$name, start = 5, stop = 5)) + 1],
  strata_ztpoisson$Estimated,
  tolerance = .1
)

marginal_ztpoisson <- summary(marginalFreq(fit_ztpoisson), dropl5 = "group", df = 1)

expect_true(
  all(marginal_ztpoisson$Test$`P(>X^2)` > .01)
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "response",
    newdata = test_zerotruncated[, paste0("x", 1:3)],
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "link",
    newdata = test_zerotruncated[, paste0("x", 1:3)],
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "mean",
    newdata = test_zerotruncated[, paste0("x", 1:3)],
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztpoisson,
    type = "contr",
    newdata = test_zerotruncated[, paste0("x", 1:3)]
  )) > N
)

expect_silent(
  predict(
    fit_ztpoisson,
    type = "popSize",
    newdata = test_zerotruncated[, paste0("x", 1:3)],
    se.fit = TRUE
  )
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "response",
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "link",
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztpoisson,
    type = "contr"
  )) == fit_ztpoisson$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztpoisson,
    type = "popSize",
    se.fit = TRUE
  )
)

# different link

model_ztpoisson <- ztpoisson(
  lambdaLink = "log"
)

#y <- model_ztpoisson$simulate(n = N, eta = eta, lower = -1)
observed_ztpoisson <- test_zerotruncated$ztpoisson2

data_ztpoisson <- data.frame(
  x3 = x3[observed_ztpoisson > 0],
  x2 = x2[observed_ztpoisson > 0],
  x1 = x1[observed_ztpoisson > 0],
  y = observed_ztpoisson[observed_ztpoisson > 0]
)

fit_ztpoisson <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = model_ztpoisson,
  data    = data_ztpoisson,
  method = "IRLS"
)

expect_equivalent(
  coef(fit_ztpoisson),
  coefficients_ztpoisson,
  tolerance = .15
)

expect_silent(
  population_ztpoisson <- popSizeEst(fit_ztpoisson)
)

expect_equivalent(
  N,
  population_ztpoisson$pointEstimate,
  tol = .1
)

expect_true(
  all(N < population_ztpoisson$confidenceInterval$upperBound) &
  all(N > population_ztpoisson$confidenceInterval$lowerBound)
)

expect_silent(
  strata_ztpoisson <- stratifyPopsize(fit_ztpoisson, ~ x1)
)

expect_equivalent(
  as.numeric(table(data_ztpoisson$x1))[as.numeric(substr(strata_ztpoisson$name, start = 5, stop = 5)) + 1],
  strata_ztpoisson$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(strata_ztpoisson$name, start = 5, stop = 5)) + 1],
  strata_ztpoisson$Estimated,
  tolerance = .1
)

marginal_ztpoisson <- summary(marginalFreq(fit_ztpoisson), dropl5 = "group")

expect_true(
  all(marginal_ztpoisson$Test$`P(>X^2)` > .01)
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "response",
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "link",
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztpoisson,
    type = "contr"
  )) == fit_ztpoisson$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztpoisson,
    type = "popSize",
    se.fit = TRUE
  )
)

## bootstrap

fit_ztpoisson <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = model_ztpoisson,
  data    = data_ztpoisson,
  method = "IRLS",
  popVar = "bootstrap",
  controlPopVar = controlPopVar(
    B = 250,
    bootType = "parametric"
  )
)

expect_silent(
  plot(fit_ztpoisson, "bootHist")
)

expect_silent(
  print(summary(fit_ztpoisson, correlation = TRUE))
)

expect_silent(
  population_ztpoisson <- popSizeEst(fit_ztpoisson)
)

expect_equivalent(
  N,
  population_ztpoisson$pointEstimate,
  tol = .1
)

expect_true(
  all(N < population_ztpoisson$confidenceInterval[2]) &
  all(N > population_ztpoisson$confidenceInterval[1])
)

expect_true(
  shapiro.test(population_ztpoisson$boot)$p.value > .05
)

fit_ztpoisson <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = model_ztpoisson,
  data    = data_ztpoisson,
  method = "IRLS",
  popVar = "bootstrap",
  controlPopVar = controlPopVar(
    B = 250,
    bootType = "semiparametric"
  )
)

expect_silent(
  plot(fit_ztpoisson, "bootHist")
)

expect_silent(
  print(summary(fit_ztpoisson, correlation = TRUE))
)

expect_silent(
  population_ztpoisson <- popSizeEst(fit_ztpoisson)
)

expect_equivalent(
  N,
  population_ztpoisson$pointEstimate,
  tol = .1
)

expect_true(
  all(N < population_ztpoisson$confidenceInterval[2]) &
  all(N > population_ztpoisson$confidenceInterval[1])
)

expect_true(
  shapiro.test(population_ztpoisson$boot)$p.value > .05
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "response",
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "link",
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    fit_ztpoisson,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztpoisson,
    type = "contr"
  )) == fit_ztpoisson$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztpoisson,
    type = "popSize",
    se.fit = TRUE
  )
)

fit_ztpoisson <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = model_ztpoisson,
  data    = data_ztpoisson,
  method = "IRLS",
  popVar = "bootstrap",
  controlPopVar = controlPopVar(
    B = 250,
    bootType = "nonparametric"
  )
)

expect_silent(
  plot(fit_ztpoisson, "bootHist")
)

expect_silent(
  print(summary(fit_ztpoisson, correlation = TRUE))
)

expect_silent(
  population_ztpoisson <- popSizeEst(fit_ztpoisson)
)

expect_equivalent(
  N,
  population_ztpoisson$pointEstimate,
  tol = .1
)

#' \loadmathjax

#' @title Obtain covariance matrix estimation
#' 
#' @description A \code{vcov} method for \code{singleRStaticCountData} class.
#' 
#' @param object object of singleRStaticCountData class.
#' @param type type of estimate for covariance matrix for now either
#' expected (Fisher) information matrix or observed information matrix.
#' @param ... additional arguments for method functions
#' 
#' @details  Returns a estimated covariance matrix for model coefficients
#' calculated from analytic hessian or Fisher information matrix usually 
#' utilizing asymptotic effectiveness of maximum likelihood estimates.
#' Covariance type is taken from control parameter that have been provided
#' on call that created \code{object} if arguments \code{type} was not specified.
#' 
#' @method vcov singleRStaticCountData
#' @return A covariance matrix for fitted coefficients, rows and columns of which 
#' correspond to parameters returned by \code{coef} method.
#' @seealso [vcovHC.singleRStaticCountData()] [sandwich::sandwich()]
#' @exportS3Method
vcov.singleRStaticCountData <- function(object, 
                         type = c("Fisher", 
                                  "observedInform"), 
                         ...) {
  if (missing(type) ){type <- object$populationSize$control$covType}
  X <- model.matrix(object, "vlm")
  res <- switch(
    type,
    "observedInform" = solve(
      -object$model$makeMinusLogLike(y = object$y, X = X,
      weight = object$priorWeights, deriv = 2, 
      offset = object$offset)(object$coefficients),
      ...
    ),
    "Fisher" = {
      if (isTRUE(object$call$method == "IRLS")) {W <- object$weights} 
        else {
          W <- object$model$Wfun(
            prior = object$priorWeights, 
            eta   = object$linearPredictors, 
            y     = as.numeric(model.response(model.frame(object)))
          )
        };
      if (isTRUE(object$control$controlMethod$checkDiagWeights)) {
        W[, (1:length(family(object)$etaNames)) ^ 2] <- ifelse(
          W[, (1:length(family(object)$etaNames)) ^ 2] < 
          object$control$controlMethod$weightsEpsilon, 
          object$control$controlMethod$weightsEpsilon, 
          W[, (1:length(family(object)$etaNames)) ^ 2]
        )
      };
      solve(
      singleRinternalMultiplyWeight(X = X, W = W) %*% X,
      ...
    )}
  )
  dimnames(res) <- list(names(object$coefficients), names(object$coefficients))
  res
}

#' @title Confidence intervals for model parameters
#' 
#' @description A function that computes studentized confidence intervals
#' for model coefficients.
#' 
#' @param object object of singleRStaticCountData class.
#' @param parm names of parameters for which confidence intervals are to be 
#' computed, if missing all parameters will be considered.
#' @param level confidence level for intervals.
#' @param ... currently does nothing.
#' 
#' @method confint singleRStaticCountData
#' @return An object with named columns that include upper and 
#' lower limit of confidence intervals.
#' @exportS3Method
confint.singleRStaticCountData <- function(object,
                            parm, 
                            level = 0.95, 
                            ...) {
  std <- sqrt(diag(vcov.singleRStaticCountData(object)))
  if (missing(parm)) {
    coef <- object$coefficients
  } else {
    coef <- object$coefficients[parm]
    std <- std[parm]
  }
  sc <- qnorm(p = 1 - (1 - level) / 2)
  res <- data.frame(coef - sc * std, coef + sc * std)
  colnames(res) <- c(paste0(100 * (1 - level) / 2, "%"),
                     paste0(100 * (1 - (1 - level) / 2), "%"))
  res
}

# functions with documented elsewhere

#' @rdname regDiagSingleR
#' @method hatvalues singleRStaticCountData
#' @importFrom stats hatvalues
#' @exportS3Method 
hatvalues.singleRStaticCountData <- function(model, ...) {
  X <- model.frame.singleRStaticCountData(model, ...)
  X <- singleRinternalGetXvlmMatrix(
    X = X, 
    formulas = model$formula, 
    parNames = model$model$etaNames
  )
  if (isTRUE(model$call$method == "IRLS")) {
    W <- model$weights
  } else {
    W <- model$model$Wfun(
      prior = model$priorWeights, 
      eta   = model$linearPredictors,
      y     = if (is.null(model$y)) model.response(model.frame(model)) 
      else model$y
    )
  }
  
  mlt <- singleRinternalMultiplyWeight(X = X, W = W)
  hatvalues <- diag(X %*% solve(mlt %*% X) %*% mlt)
  hatvalues <- matrix(
    hatvalues, 
    ncol = length(model$model$etaNames), 
    dimnames = list(
      1:(length(hatvalues) / length(model$model$etaNames)), 
      model$model$etaNames
    )
  )
  hatvalues
}

#' @method residuals singleRStaticCountData
#' @importFrom stats residuals
#' @rdname regDiagSingleR
#' @exportS3Method
residuals.singleRStaticCountData <- function(object,
                                             type = c(
                                               "pearson",
                                               "pearsonSTD",
                                               "response",
                                               "working",
                                               "deviance",
                                               "all"
                                             ),
                                             ...) {
  type <- match.arg(type)
  res <- object$residuals
  wts <- object$priorWeights
  y <- if (is.null(object$y)) stats::model.response(model.frame(object)) 
    else object$y
  
  if (type == "pearsonSTD" && length(object$model$etaNames) > 1)
    stop(paste0("Standardized pearson residuals not yet",
                "implemented for models with multiple linear predictors"))
  
  rs <- switch(
    type,
    working = as.data.frame(
      object$model$funcZ(eta = object$linearPredictors, 
                         weight = object$weights, 
                         y = y, prior = wts), 
      col.names = paste0("working:", object$model$etaNames)
    ),
    response = res,
    pearson = data.frame(
      "pearson" = res$truncated / sqrt(
        object$model$variance(eta = object$linearPredictors, type = "trunc")
      )
    ),
    pearsonSTD = data.frame(
      "pearsonSTD" = res$truncated / sqrt(
        (1 - hatvalues(object)) * 
        object$model$variance(eta = object$linearPredictors, type = "trunc")
      )
    ),
    deviance = data.frame(
      "deviance" = object$model$devResids(y = y, 
                                          eta = object$linearPredictors, 
                                          wt = wts)
    ),
    all = {colnames(res) <- c("truncatedResponse", 
                              "nontruncatedResponse");
      data.frame(
        as.data.frame(object$model$funcZ(eta = object$linearPredictors, 
                                         weight = object$weights, 
                                         y = y, prior = wts),
                      col.names = paste0("working:", object$model$etaNames)),
        res,
        "pearson" = as.numeric(res$truncated / sqrt(object$model$variance(
          eta = object$linearPredictors, type = "trunc"
        ))),
        "pearsonSTD" = if (length(object$model$etaNames) == 1) as.numeric(
          res$truncated / sqrt(
            (1 - hatvalues(object)) * 
            object$model$variance(object$linearPredictors, type = "trunc")
          )
        ) else NA,
        "deviance" = as.numeric(object$model$devResids(
          y = y, eta = object$linearPredictors, wt = wts
        )),
        row.names = rownames(object$linearPredictors)
    )}
  )
  rs
}

#' @importFrom stats cooks.distance
#' @method cooks.distance singleRStaticCountData
#' @rdname regDiagSingleR
#' @exportS3Method 
cooks.distance.singleRStaticCountData <- function(model, ...) {
  if (length(model$model$etaNames) > 1) 
    stop("Cooks distance is only implemented for single parameter families.")
  
  res <- residuals(model, type = "pearsonSTD") ^ 2
  res <- res[, 1]
  
  ht <- hatvalues(model)
  res <- (res * (ht / (length(coef(model)))))
  rownames(res) <- rownames(ht)
  res
}

# Add in the future actual methods for this
#' @importFrom stats add1
#' @method add1 singleRStaticCountData
#' @exportS3Method 
add1.singleRStaticCountData <- function(object, scope, ...) {
  stop("The add1 method for singleRStaticCountData class doesn't work yet.")
}
#' @importFrom stats profile
#' @method profile singleRStaticCountData
#' @exportS3Method 
profile.singleRStaticCountData <- function(fitted, ...) {
  stop("The profile method for singleRStaticCountData class doesn't work yet.")
}
#' @importFrom stats drop1
#' @method drop1 singleRStaticCountData
#' @exportS3Method 
drop1.singleRStaticCountData <- function(object, scope, ...) {
  stop("The drop1 method for singleRStaticCountData class doesn't work yet.")
}
#' @importFrom stats anova
#' @method anova singleRStaticCountData
#' @exportS3Method 
anova.singleRStaticCountData <- function(object, ...) {
  stop("The custom anova method for singleRStaticCountData class is not yet implemented. If the goal is to compare models we recommend using `lmtest::lrtest` instead. ")
}

# ALL CODE IN THIS FILE WAS DEVELOPED BASED ON CODE IN sandwich PACKAGE

#' @importFrom sandwich estfun
#' @method estfun singleRStaticCountData
#' @rdname vcovHC.singleRStaticCountData
#' @exportS3Method
estfun.singleRStaticCountData <- function(x,...) {
  Y <- if (is.null(x$y)) stats::model.response(model.frame(x)) else x$y
  X <- stats::model.matrix(x, type = "vlm")
  
  beta <- stats::coef(x)
  wts <- x$priorWeights
  
  if (is.null(wts)) wts <- rep(1, length(Y))
  
  if (x$control$controlModel$weightsAsCounts) {
    nums <- rep(1:length(Y), wts)
    Y <- Y[nums]
    X <- X[rep(nums, length(x$model$etaNames)), , drop = FALSE]
    res <- x$model$makeMinusLogLike(
      y = Y, 
      X = X, 
      weight = 1, 
      NbyK = TRUE, 
      deriv = 1,
      offset = x$offset[rep(nums, length(x$model$etaNames)), , drop = FALSE]
    )(beta)
    
    colnames(res) <- names(beta)
  } else {
    res <- x$model$makeMinusLogLike(
      y = Y, 
      X = X, 
      weight = x$priorWeights, 
      NbyK = TRUE, 
      deriv = 1,
      offset = x$offset
    )(beta)
    
    colnames(res) <- names(beta)
    rownames(res) <- rownames(X)
  }
  
  res
}

#' @importFrom sandwich bread
#' @method bread singleRStaticCountData
#' @rdname vcovHC.singleRStaticCountData
#' @exportS3Method
bread.singleRStaticCountData <- function(x,...) {
  stats::vcov(x, ...) * nobs(x)
}

#' @title Heteroscedasticity-consistent covariance matrix estimation for singleRStaticCountData class
#' @author Piotr Chlebicki, Maciej Beręsewicz
#' 
#' @description S3 method for \code{vcovHC} to handle \code{singleRStaticCountData} class objects. 
#' Works exactly like \code{vcovHC.default} the only difference being that this method handles vector generalised linear models.
#' Updating the covariance matrix in variance/standard error estimation for population size estimator can be done via [singleRcapture::redoPopEstimation()]
#'
#' @param x a fitted \code{singleRStaticCountData} class object.
#' @param type a character string specifying the estimation type, same as in \code{sandwich::vcovHC.default}. HC3 is the default value.
#' @param omega a vector or a function depending on the arguments residuals (i.e. the derivative of log-likelihood with respect to each linear predictor), diaghat (the diagonal of the corresponding hat matrix) and df (the residual degrees of freedom), same as in \code{sandwich::vcovHC.default}.
#' @param sandwich logical. Should the sandwich estimator be computed? If set to FALSE only the meat matrix is returned. Same as in [sandwich::vcovHC()]
#' @param ... for \code{vcovHC} additional optional arguments passed to the following functions:
#' \itemize{
#'   \item \code{estfun} -- for empirical estimating functions.
#'   \item \code{hatvalues} -- for diagonal elements of projection matrix.
#'   \item \code{sandwich} -- only if \code{sandwich} argument in function call was set to \code{TRUE}.
#'   \item \code{vcov} -- when calling \code{bread} internally.
#' }
#'
#' @return Variance-covariance matrix estimation corrected for heteroscedasticity of regression errors
#' @seealso [sandwich::vcovHC()] [singleRcapture::redoPopEstimation()]
#' @examples 
#' set.seed(1)
#' N <- 10000
#' gender <- rbinom(N, 1, 0.2)
#' eta <- -1 + 0.5*gender
#' counts <- rpois(N, lambda = exp(eta))
#' df <- data.frame(gender, eta, counts)
#' df2 <- subset(df, counts > 0)
#' mod1 <-  estimatePopsize(
#'   formula = counts ~ 1 + gender, 
#'   data = df2, 
#'   model = "ztpoisson", 
#'   method = "optim", 
#'   popVar = "analytic"
#' )
#' require(sandwich)
#' HC <- sandwich::vcovHC(mod1, type = "HC4")
#' Fisher <- vcov(mod1, "Fisher") # variance covariance matrix obtained from 
#' #Fisher (expected) information matrix
#' HC
#' Fisher
#' # usual results
#' summary(mod1)
#' # updated results
#' summary(mod1, cov = HC,
#' popSizeEst = redoPopEstimation(mod1, cov = HC))
#' # estimating equations
#' mod1_sims <- sandwich::estfun(mod1)
#' head(mod1_sims)
#' # bread method
#' all(vcov(mod1, "Fisher") * nrow(df2) == sandwich::bread(mod1, type = "Fisher"))
#' @importFrom sandwich vcovHC
#' @method vcovHC singleRStaticCountData
#' @exportS3Method
vcovHC.singleRStaticCountData <- function(x, 
                           type = c("HC3", "const", "HC", 
                                    "HC0", "HC1", "HC2", 
                                    "HC4", "HC4m", "HC5"), 
                           omega = NULL, 
                           sandwich = TRUE, 
                           ...) {
  type <- match.arg(type)
  estfun <- estfun(x, ...)
  beta <- x$coefficients
  X <- model.matrix(x, "vlm")
  n <- nobs(x)
  k <- NCOL(X)
  
  df <- x$dfResidual
  hat <- as.vector(hatvalues(x, ...))
  Y <- if (is.null(x$y)) stats::model.response(model.frame(x)) else x$y
  
  if (x$control$controlModel$weightsAsCounts) {
    nums <- rep(1:length(Y), x$priorWeights)
    Y <- Y[nums]
    X <- X[rep(nums, length(x$model$etaNames)), , drop = FALSE]
    res <- as.vector(x$model$makeMinusLogLike(
      y = Y, 
      X = X, 
      weight = 1, 
      vectorDer = TRUE, 
      deriv = 1,
      offset = x$offset[rep(nums, length(x$model$etaNames)), , drop = FALSE]
    )(beta))
    
    hat <- (hat / x$priorWeights)[nums]
  } else {
    res <- as.vector(x$model$makeMinusLogLike(
      y = Y, 
      X = X, 
      weight = x$priorWeights, 
      vectorDer = TRUE, 
      deriv = 1,
      offset = x$offset
    )(beta))
  }

  if (is.null(omega)) {
    if (type == "HC") 
      type <- "HC0"
    switch(type, const = {
      omega <- function(residuals, diaghat, df) rep(1, length(residuals)) * sum(residuals^2)/df
    }, HC0 = {
      omega <- function(residuals, diaghat, df) residuals^2
    }, HC1 = {
      omega <- function(residuals, diaghat, df) residuals^2 * 
        length(residuals)/df
    }, HC2 = {
      omega <- function(residuals, diaghat, df) residuals^2/(1 - diaghat)
    }, HC3 = {
      omega <- function(residuals, diaghat, df) residuals^2/(1 - diaghat)^2
    }, HC4 = {
      omega <- function(residuals, diaghat, df) {
        n <- length(residuals)
        p <- as.integer(round(sum(diaghat), digits = 0))
        delta <- pmin(4, n * diaghat/p)
        residuals^2/(1 - diaghat)^delta
      }
    }, HC4m = {
      omega <- function(residuals, diaghat, df) {
        gamma <- c(1, 1.5)
        n <- length(residuals)
        p <- as.integer(round(sum(diaghat), digits = 0))
        delta <- pmin(gamma[1], n * diaghat/p) + pmin(gamma[2], 
                                                      n * diaghat/p)
        residuals^2/(1 - diaghat)^delta
      }
    }, HC5 = {
      omega <- function(residuals, diaghat, df) {
        k <- 0.7
        n <- length(residuals)
        p <- as.integer(round(sum(diaghat), digits = 0))
        delta <- pmin(n * diaghat/p, pmax(4, n * k * max(diaghat)/p))
        residuals^2 / sqrt((1 - diaghat)^delta)
      }
    })
  }
  if (is.function(omega)) 
    omega <- omega(res, hat, df)
  rval <- sqrt(omega) * X
  rval <- crossprod(rval)/n
  
  if (sandwich) 
    rval <- sandwich(x, meat. = rval, ...)
  rval
}

#' @rdname singleRmodels
#' @importFrom stats dbinom
#' @export
chao <- function(lambdaLink = "loghalf",
                 ...) {
  if (missing(lambdaLink)) lambdaLink <- "loghalf"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink)
  
  lambdaLink <- switch (lambdaLink,
    "loghalf" = singleRinternalloghalfLink
  )
  
  links[1] <- c(lambdaLink)
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    lambda <- lambdaLink(eta, inverse = TRUE)
    
    if (!deriv) {
      switch (type,
        "nontrunc" = lambda,
        "trunc" = (lambda / 2) / (1 + lambda / 2)
      )
    } else {
      switch (type,
        "nontrunc" = cbind(1 * lambdaLink(eta, inverse = TRUE, deriv = 1)),
        "trunc" = cbind(
          (2 / (lambda + 2) ^ 2) *
            lambdaLink(eta, inverse = TRUE, deriv = 1)
        )
      )
    }
  }
  
  variance <- function(eta, type = "nontrunc", ...) {
    lambda <- lambdaLink(eta, inverse = TRUE)
    switch (type,
    "nontrunc" = lambda,
    "trunc" = ((lambda / 2) / (1 + lambda / 2)) * (1 / (1 + lambda / 2))
    )
  }
  
  Wfun <- function(prior, eta, y, ...) {
    iddx <- y %in% 1:2
    lambda <- lambdaLink(eta, inverse = TRUE)[iddx]
    z <- (lambda / 2) / (1 + lambda / 2)
    res <- iddx
    res[iddx] <- -prior[iddx] * (
      -2 / (lambda * (2 + lambda) ^ 2) *
        lambdaLink(eta, inverse = TRUE, deriv = 1)[iddx] ^ 2)
    
    matrix(res, ncol = 1, 
           dimnames = list(rownames(eta), 
                           c("lambda")))
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    iddx <- y %in% 1:2
    z <- y - 1
    lambda <- lambdaLink(eta, inverse = TRUE)[iddx]
    
    res <- iddx
    res[iddx] <- prior[iddx] * (((z[iddx] - 1) * lambda + 2 * z[iddx]) / 
    (lambda * (2 + lambda))) * lambdaLink(eta, inverse = TRUE, deriv = 1)[iddx] / weight[iddx, ]
    res
  }

  minusLogLike <- function(y, X, 
                           weight = 1, 
                           NbyK = FALSE, 
                           vectorDer = FALSE, 
                           deriv = 0,
                           offset,
                           eta, 
                           ...) {
    y <- as.numeric(y)
    z <- y - 1
    if (is.null(weight)) {
      weight <- 1
    }
    
    iddx <- y %in% 1:2
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X)))
    }
    
    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
      function(beta, eta) {
        if (missing(eta)) {
          eta <- as.matrix(X) %*% beta + offset
        }
        lambda <- lambdaLink(eta, inverse = TRUE)[iddx]
        -sum(weight[iddx] * dbinom(size = 1, 
                                   prob = ((lambda / 2) / (1 + lambda / 2)),
                                   log  = TRUE,
                                   x    = z[iddx]))
      },
      function(beta) {
        eta <- as.matrix(X) %*% beta + offset
        lambda <- lambdaLink(eta, inverse = TRUE)[iddx]
        G0 <- iddx
        G0[iddx] <- ((z[iddx] - 1) * lambda + 2 * z[iddx]) / 
          (lambda * (2 + lambda)) * weight[iddx] * 
          lambdaLink(eta, inverse = TRUE, deriv = 1)[iddx]
        
        if (NbyK) {
          return(as.data.frame(X[iddx, , drop = FALSE]) * G0)
        }
        if (vectorDer) {
          return(matrix(G0, ncol = 1))
        }
        
        t(X) %*% G0
      },
      function(beta) {
        eta <- as.matrix(X) %*% beta + offset
        lambda <- lambdaLink(eta, inverse = TRUE)[iddx]
        
        G00 <- iddx
        G00[iddx] <- (-(4 * z[iddx] + 4 * z[iddx] * lambda + 
                          (z[iddx] - 1) * lambda ^ 2) / ((2 + lambda) ^ 2 * lambda ^ 2)) *
          lambdaLink(eta, inverse = TRUE, deriv = 1)[iddx] ^ 2 +
          ((z[iddx] - 1) * lambda + 2 * z[iddx]) / (lambda * (2 + lambda)) *
          lambdaLink(eta, inverse = TRUE, deriv = 2)[iddx]
        
        t(as.data.frame(X) * weight * G00) %*% as.matrix(X)
      }
    )
  }

  validmu <- function(mu) {
    (sum(!is.finite(mu)) == 0) && all(1 > mu)
  }

  devResids <- function(y, eta, wt, ...) {
    z <- y - 1
    lambda <- lambdaLink(eta, inverse = TRUE)
    diff <- (z * log(lambda / 2 / (1 + lambda / 2)) + (1 - z) * log(1 / (1 + lambda / 2)))
    diff[diff > 0] <- -0
    
    ((-1) ^ y) * sqrt((-2 * wt * diff) * (y %in% 1:2))
  }

  pointEst <- function (pw, eta, contr = FALSE, y, ...) {
    lambda <- lambdaLink(eta, inverse = TRUE)
    N <- ((1 + (y %in% 1:2) / (lambda + (lambda ^ 2) / 2)) * pw)
    if(!contr) {
      N <- sum(N)
    }
    N
  }

  popVar <- function (pw, eta, cov, Xvlm, y, ...) {
    iddx <- y %in% 1:2
    lambda <- lambdaLink(eta, inverse = TRUE)
    Xvlm <- as.matrix(Xvlm)
    prob <- lambda * exp(-lambda) + (lambda ^ 2) * exp(-lambda) / 2

    f1 <- t((-lambdaLink(eta[,1], inverse = TRUE, deriv = 1) * pw * 
      as.numeric((lambda+1)/(lambda^2/2+lambda)^2))[iddx] %*% Xvlm[iddx, , drop = FALSE])
    
    f1 <- t(f1) %*% as.matrix(cov) %*% f1

    f2 <- sum((pw * (1 - prob) * ((1 + exp(-lambda) / prob) ^ 2))[iddx])

    f1 + f2
  }
  
  simulate <- function(n, eta, lower = 0, upper = 2) {
    lambda <- lambdaLink(eta, inverse = TRUE)

    lb <- stats::ppois(lower, lambda)
    ub <- stats::ppois(upper, lambda)
    p_u <- stats::runif(n, lb, ub)
    sims <- stats::qpois(p_u, lambda)
    sims
  }

  dFun <- function (x, eta, type = "trunc") {
    if (missing(type)) type <- "trunc"
    lambda <- lambdaLink(eta, inverse = TRUE)
    
    switch (type,
      "trunc" = {
        stats::dpois(x = x, lambda = lambda) / 
        (1 - stats::dpois(x = 0, lambda = lambda))
      },
      "nontrunc" = stats::dpois(x = x, lambda = lambda)
    )
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      etaStart <- cbind(
        (priorWeights * (observed == 2) + .5) / (priorWeights + 1)
      ) + offset
      etaStart <- family$links[[1]](2 * (etaStart / (1 - etaStart)))
    } else if (method == "optim") {
      init <- mean(((priorWeights * (observed == 2) + .5) / (priorWeights + 1))[observed %in% 1:2])
      coefStart <- rep(family$links[[1]](2 * init / (1 - init)), NCOL(variables))
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "chao",
      etaNames  = c("lambda"),
      simulate  = simulate,
      simulateLower = 0,
      simulateUpper = 2,
      getStart  = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

# chao family tests

chao_family <- chao()
chao_eta <- matrix(chao_family$links[[1]](1.7))

expect_equal(
  as.numeric(chao_family$mu.eta(chao_eta, type = "trunc")),
  1.7 / (2 + 1.7),
  tolerance = 1e-10
)

expect_equal(
  as.numeric(chao_family$mu.eta(chao_eta, type = "nontrunc")),
  1.7,
  tolerance = 1e-10
)

expect_equal(
  as.numeric(chao_family$variance(chao_eta, type = "trunc")),
  (1.7 / (2 + 1.7)) * (2 / (2 + 1.7)),
  tolerance = 1e-10
)

expect_equal(
  as.numeric(chao_family$densityFunction(1, chao_eta, type = "trunc")),
  dpois(1, 1.7) / (1 - dpois(0, 1.7)),
  tolerance = 1e-10
)

expect_equal(
  as.numeric(chao_family$densityFunction(2, chao_eta, type = "nontrunc")),
  dpois(2, 1.7),
  tolerance = 1e-10
)

chao_dev_y2 <- as.numeric(chao_family$devResids(
  y = 2, eta = matrix(chao_family$links[[1]](100)), wt = 1
))
chao_dev_y1 <- as.numeric(chao_family$devResids(
  y = 1, eta = matrix(chao_family$links[[1]](0.01)), wt = 1
))
expect_true(is.finite(chao_dev_y2))
expect_true(is.finite(chao_dev_y1))
expect_true(chao_dev_y2 > 0)
expect_true(chao_dev_y1 < 0)
expect_true(abs(chao_dev_y2) < 1)
expect_true(abs(chao_dev_y1) < 1)

expect_equal(
  as.numeric(chao_family$devResids(y = 0, eta = chao_eta, wt = 1)),
  0,
  tolerance = 1e-10
)
expect_equal(
  as.numeric(chao_family$devResids(y = 3, eta = chao_eta, wt = 1)),
  0,
  tolerance = 1e-10
)

expect_silent(
  fit_chao <- estimatePopsize(
    formula = TOTAL_SUB ~ .,
    model = chao(),
    data = farmsubmission,
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_silent(
  predict(
    fit_chao,
    type = "mean"
  )
)

expect_equivalent(
  fit_chao$populationSize$pointEstimate,
  21657,
  tolerance = .005
)

expect_equivalent(
  fit_chao$populationSize$confidenceInterval[1, ],
  list(20883, 22430),
  tolerance = .005
)

expect_silent(
  population_chao <- popSizeEst(fit_chao)
)

expect_true(is.finite(population_chao$variance))
expect_true(population_chao$variance > 0)

#' \loadmathjax
#' @title Predict method for singleRStaticCountData class
#' 
#' @description
#' A method for \code{predict} function, works analogous to \code{predict.glm}
#' but gives the possibility to get standard errors of 
#' mean/distribution parameters and directly get pop size estimates for new data.
#' 
#' 
#' @param object an object of \code{singleRStaticCountData} class.
#' @param newdata an optional \code{data.frame} containing new data.
#' @param type the type of prediction required, possible values are:
#' \itemize{
#'   \item \code{"response"}-- For matrix containing estimated distributions
#'   parameters.
#'   \item \code{"link"}    -- For matrix of linear predictors.
#'   \item \code{"mean"}    -- For fitted values of both \mjseqn{Y} and
#'   \mjseqn{Y|Y>0}.
#'   \item \code{"contr"}   -- For inverse probability weights (here named for 
#'   observation contribution to population size estimate).
#'   \item \code{"popSize"} -- For population size estimation. Note
#'   this results in a call to \code{redoPopEstimation} and it is
#'   usually better to call this function directly.
#' } by default set to \code{"response"}.
#' @param se.fit a logical value indicating whether standard errors should be 
#' computed. Only matters for \code{type} in \code{"response", "mean", "link"}.
#' @param na.action does nothing yet.
#' @param weights optional vector of weights for \code{type} in \code{"contr", "popSize"}.
#' @param cov optional matrix or function or character specifying either
#' a covariance matrix or a function to compute that covariance matrix.
#' By default \code{vcov.singleRStaticCountData} can be set to e.g. \code{vcovHC}.
#' @param ... arguments passed to other functions, for now this only affects
#' \code{vcov.singleRStaticCountData} method and \code{cov} function.
#' 
#' @details Standard errors are computed with assumption of regression
#' coefficients being asymptotically normally distributed, if this assumption
#' holds then each of linear predictors i.e. each row of
#' \mjseqn{\boldsymbol{\eta}=\boldsymbol{X}_{vlm}\boldsymbol{\beta}}
#' is asymptotically normally distributed and their variances are expressed by
#' well known formula. The mean \mjseqn{\mu} and distribution parameters
#' are then differentiable functions of asymptotically normally distributed
#' variables and therefore their variances can be computed using (multivariate)
#' delta method.
#'
#' @return Depending on \code{type} argument if one of \code{"response", "link", "mean"}
#' a matrix with fitted values and possibly standard errors if \code{se.fit}
#' argument was set to \code{TRUE}, if \code{type} was set to \code{"contr"}
#' a vector with inverses of probabilities, finally for \code{"popSize"}
#' an object of class \code{popSizeEstResults} with its own methods containing
#' population size estimation results.
#' 
#' @method predict singleRStaticCountData
#' @seealso [redoPopEstimation()] [stats::summary.glm()] [estimatePopsize()]
#' @exportS3Method
predict.singleRStaticCountData <- function(object,
                                           newdata,
                                           type = c("response", "link", 
                                                    "mean", "popSize", 
                                                    "contr"),
                                           se.fit = FALSE,
                                           na.action = NULL,
                                           weights,
                                           cov,
                                           ...) {
  type <- match.arg(type)
  if (missing(weights)) {
    if (missing(newdata)) {
      weights <- object$priorWeights
    } else {
      weights <- rep(1, NROW(newdata))
    }
  }
  
  if (missing(cov)) {
    cov <- vcov
  }
  
  if (is.character(cov)) {
    cov <- get(cov, mode = "function", envir = parent.frame())
  }
  if (is.function(cov)) {
    cov <- cov(object, ...)
  }
  

  if (missing(newdata)) {
    Xvlm <- model.matrix(
      object, type = "vlm"
    )
    
    eta <- object$linearPredictors
    
    res <- switch (type,
      response = as.data.frame(
        lapply(1:length(family(object)$etaNames), FUN = function(x) {
          family(object)$links[[x]](
            eta[, x],
            inverse = TRUE
          )
      }), col.names = family(object)$etaNames),
      link     = eta,
      mean     = fitted(object, "all"),
      popSize  = popSizeEst(object, ...),
      contr    = family(object)$pointEst(
        pw     = weights,
        eta    = eta,
        contr  = TRUE,
        y      = model.response(model.frame(object))
      )
    )
  } else {
    mf <- model.frame(
      object, data = newdata
    )
    
    Xvlm <- model.matrix(
      object, type = "vlm", data = newdata
    )
    
    eta <- matrix(
      as.matrix(Xvlm) %*% stats::coef(object), 
      ncol = length(family(object)$etaNames),
      dimnames = list(
        rownames(Xvlm),
        family(object)$etaNames
      )
    )

    res <- switch (type,
      response = as.data.frame(
        lapply(1:length(family(object)$etaNames), FUN = function(x) {
        family(object)$links[[x]](
          eta[, x],
          inverse = TRUE
        )
      }), col.names = family(object)$etaNames),
      link = eta,
      mean = data.frame(
        "truncated"    = family(object)$mu.eta(eta = eta),
        "nontruncated" = family(object)$mu.eta(eta = eta, type = "nontrunc")
      ),
      popSize = redoPopEstimation(
        object = object, newdata = newdata,
        weights = if (missing(weights)) rep(1, NROW(mf)) else weights,
        cov = cov,
        ...
      ),
      contr = family(object)$pointEst(
        pw    = weights,
        eta   = eta,
        contr = TRUE,
        y     = model.response(mf) 
      )
    )
  }
  
  if (isTRUE(se.fit)) {
    #### beta is asymptotically normal and each eta is just a linear
    #### combination of beta so it is also asymptotically normal
    #### and its variance is computed by quadratic form
    if (type != "mean") {
      se <- matrix(
        sapply(1:NROW(Xvlm), function(x) {
          (t(Xvlm[x,]) %*% cov %*% Xvlm[x,]) ^ .5
        }),
        ncol = length(family(object)$etaNames),
        dimnames = list(
          rownames(eta),
          paste0("se:", family(object)$etaNames)
        )
      )
      
      if (type == "link") 
        res <- cbind(res, se)
      
      ## since beta is asymptotically normal we use delta metod for
      ## geting standard errors of invlink(eta)
      
      if (type == "response") {
        se <- matrix(
          sapply(
            1:length(family(object)$etaNames),
            function (x) {
              se[, x, drop = FALSE] *
                abs(family(object)$links[[x]](
                  eta[, x],
                  inverse = TRUE,
                  deriv = 1
                ))
            }
          ),
          ncol = length(family(object)$etaNames),
          dimnames = dimnames(se)
        )
        res <- cbind(res,se)
      }
    } else if (type == "mean") {
      ## covariance matrix foe each row of linear predictors
      auxVec <- c(0, cumsum(rep(
        NROW(eta), 
        length.out = length(family(object)$etaNames) - 1
      )))
      
      se <- lapply(
        1:NROW(eta),
        function (x) {
          Xvlm[x + auxVec, , drop = FALSE] %*% cov %*%
            t(Xvlm[x + auxVec, , drop = FALSE])
        }
      )
      
      derivMu <- list(
        family(object)$mu.eta(eta, type = "nontrunc", deriv = 1),
        family(object)$mu.eta(eta, type = "trunc", deriv = 1)
      )
      
      res <- data.frame(
        res,
        "se:truncated" = sapply(
          1:NROW(eta),
          function (x) {
            (derivMu[[2]][x, , drop = FALSE] %*% se[[x]] %*%
               t(derivMu[[2]][x, , drop = FALSE])) ^ .5
          }
        ),
        "se:nontruncated" = sapply(
          1:NROW(eta), 
          function (x) {
            (derivMu[[1]][x, , drop = FALSE] %*% se[[x]] %*%
               t(derivMu[[1]][x, , drop = FALSE])) ^ .5
          }
        )
      )
    }
  }

  res
}

#' @title Diagnostic plots for regression and population size estimation
#' @author Piotr Chlebicki
#'
#' @description Simple diagnostic plots for \code{singleRStaticCountData} class objects.
#'
#' @param x object of \code{singleRStaticCountData} class.
#' @param dfpop an optional pre-computed result of \code{dfpopsize()} to avoid recomputation when using \code{plotType = "dfpopContr"} or \code{plotType = "dfpopBox"}.
#' @param confIntStrata confidence interval type to use for strata plot.
#' Currently supported values are \code{"normal"} and \code{"logNormal"}.
#' @param plotType character parameter (default \code{"qq"}) specifying type of plot to be made.
#' The following list presents and briefly explains possible type of plots:
#' \itemize{
#'   \item \code{qq} -- The quantile-quantile plot for pearson residuals 
#'   (or standardized pearson residuals if these are available for the model) i.e. 
#'   empirical quantiles from residuals are plotted against theoretical quantiles 
#'   from standard distribution.
#'   \item \code{marginal} -- A plot made by \code{matplot} with fitted and 
#'   observed marginal frequencies with labels.
#'   \item \code{fitresid} -- Plot of fitted linear predictors against 
#'   (standardized) pearson residuals.
#'   \item \code{bootHist} -- Simple histogram for statistics obtained from 
#'   bootstrapping (if one was performed and the statistics were saved).
#'   \item \code{rootogram} -- Rootogram, for full explanation see: 
#'   Kleiber and Zeileis Visualizing Count Data Regressions Using Rootograms (2016), 
#'   in short it is a \code{barplot} where height is the square root of observed marginal 
#'   frequencies adjusted by difference between square root of observed and fitted marginal 
#'   frequencies connected by line representing fitted marginal frequencies. 
#'   The less of a difference there is between the 0 line and beginning of a bar 
#'   the more accurate fit was produced by the model.
#'   \item \code{dfpopContr} -- Plot of \code{dfpopsize} against unit contribution.
#'   On the plot is y = x line i.e. what deletion effect would be if removing the
#'   unit from the model didn't effect regression coefficients. The further away
#'   the observation is from this line the more influential it is.
#'   \item \code{dfpopBox} -- Boxplot of \code{dfpopsize} for getting the general 
#'   idea about the distribution of the "influence" of each unit on
#'   population size estimate.
#'   \item \code{scaleLoc} -- The scale - location plot i.e. square root of 
#'   absolute values of (standardized) pearson residuals against linear predictors
#'   for each column of linear predictors.
#'   \item \code{cooks} -- Plot of cooks distance for detecting influential observations.
#'   \item \code{hatplot} -- Plot of hat values for each linear predictor for detecting influential observations.
#'   \item \code{strata} -- Plot of confidence intervals and point estimates for strata provided in \code{...} argument
#' }
#' @param histKernels logical value indicating whether to add density lines
#' to histogram.
#' @param ... additional optional arguments passed to the following functions:
#' \itemize{
#'   \item For \code{plotType = "bootHist"}
#'   \itemize{
#'   \item \code{graphics::hist} -- with \code{x, main, xlab, ylab} parameters fixed.
#'   }
#'   \item For \code{plotType = "rootogram"} 
#'   \itemize{
#'   \item \code{graphics::barplot} -- with \code{height, offset, ylab, xlab, ylim} parameters fixed.
#'   \item \code{graphics::lines} -- with \code{x, y, pch, type, lwd, col} parameters fixed.
#'   }
#'   \item For \code{plotType = "dfpopContr"}
#'   \itemize{
#'   \item \code{dfpopsize} -- with \code{model, observedPop} parameters fixed.
#'   \item \code{plot.default} -- with \code{x, y, xlab, main} parameters fixed.
#'   }
#'   \item For \code{plotType = "dfpopBox"}
#'   \itemize{
#'   \item \code{dfpopsize} -- with \code{model, observedPop} parameters fixed.
#'   \item \code{graphics::boxplot} -- with \code{x, ylab, main} parameters fixed.
#'   }
#'   \item For \code{plotType = "scaleLoc"}
#'   \itemize{
#'   \item \code{plot.default} -- with \code{x, y, xlab, ylab, main, sub} parameters fixed.
#'   }
#'   \item For \code{plotType = "fitresid"}
#'   \itemize{
#'   \item \code{plot.default} -- with \code{x, y, xlab, ylab, main, sub} parameters fixed.
#'   }
#'   \item For \code{plotType = "cooks"}
#'   \itemize{
#'   \item \code{plot.default} -- with \code{x, xlab, ylab, main} parameters fixed.
#'   }
#'   \item For \code{plotType = "hatplot"} 
#'   \itemize{
#'   \item \code{hatvalues.singleRStaticCountData}
#'   \item \code{plot.default} -- with \code{x, xlab, ylab, main} parameters fixed.
#'   }
#'   \item For \code{plotType = "strata"}
#'   \itemize{
#'   \item \code{stratifyPopsize.singleRStaticCountData}
#'   }
#' }
#' 
#' @method plot singleRStaticCountData
#' @return No return value only the plot being made.
#' @importFrom stats ppoints qqline qqnorm density dlnorm
#' @importFrom graphics abline barplot hist lines matplot legend boxplot panel.smooth axis text arrows par
#' @seealso [estimatePopsize()] [dfpopsize()] [marginalFreq()] [stats::plot.lm()] [stats::cooks.distance()] [hatvalues.singleRStaticCountData()]
#' @export
plot.singleRStaticCountData <- function(x, 
                                        plotType = c("qq", "marginal", "fitresid",
                                                     "bootHist", "rootogram", "dfpopContr",
                                                     "dfpopBox", "scaleLoc", "cooks",
                                                     "hatplot", "strata"),
                                        confIntStrata = c("normal", "logNormal"),
                                        histKernels = TRUE,
                                        dfpop,
                                        ...) {
  
  if (missing(plotType) || is.null(plotType)) {
    plotType <- eval(formals()$plotType)[1]
  } else if (is.numeric(plotType) && length(plotType) == 1) {
    # Handle numeric index
    valid_plots <- eval(formals()$plotType)
    if (plotType > 0 && plotType <= length(valid_plots)) {
      plotType <- valid_plots[as.integer(plotType)]
    } else {
      stop("Numeric plotType must be between 1 and ", length(valid_plots))
    }
  } else if (is.character(plotType)) {
    # Use match.arg for character input validation against the options
    plotType <- match.arg(plotType)
  } else {
    stop("plotType must be a character string, or a numeric index")
  }
  
  ## sugested by Victoria Wimmer
  oldpar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(oldpar))
  
  if (isTRUE(plotType == "bootHist") && isTRUE(!is.numeric(x$populationSize$boot))) {
    stop("Trying to plot bootstrap results with no bootstrap performed")
  } 
  plotType <- match.arg(plotType)
  parNum <- length(x$model$etaNames)
  
  if (parNum == 1)
    type <- "pearsonSTD" 
  else 
    type <- "pearson"
  
  if (plotType == "fitresid") {
    res <- residuals(x, type = "response")[, 1]
  } else {
    res <- residuals(x, type = type)[, 1]
  }
  switch(plotType,
  qq = {
    stats::qqnorm(res);
    stats::qqline(res);
  },
  marginal = {
    M <- marginalFreq(x);
    FF <- M$table;
    FF[names(M$y)] <- M$y;
    FF[setdiff(names(M$table), names(M$y))] <- 0;
    graphics::matplot(
      y = cbind(M$table, FF),
      x = 0:max(x$y), type = "o",
      col = 1:2, pch = 21:22, lty = 2,
      main = "Plot of observed and fitted marginal frequencies",
      ylab = "Frequency",
      xlab = "Counts",
      ...);
    legend("topright",
           legend = c(x$model$family,
                      "Observed"),
           col = 1:2, pch = 21:22)
  },
  bootHist = {
    h <- graphics::hist(
      x$populationSize$boot,
      ylab = "Number of bootstrap samples",
      xlab = expression(hat(N)),
      main = "Bootstrap of population size estimates",
      ...
    );
    if (isTRUE(histKernels)) {
      xxx <- density(x$populationSize$boot, kernel = "epanechnikov");
      graphics::lines(
        x = xxx$x, 
        y = dlnorm(
          x       = xxx$x, 
          meanlog = log(mean(x$populationSize$boot) / 
                          sqrt(1 + var(x$populationSize$boot) / 
                                 mean(x$populationSize$boot) ^ 2)),
          sdlog   = sqrt(log(1 + var(x$populationSize$boot) / mean(x$populationSize$boot) ^ 2))
        ) * length(x$populationSize$boot) * diff(h$breaks)[1], 
        lty = 2, 
        col = 8
      )
      graphics::lines(
        x = xxx$x, 
        y = dnorm(
          x    = xxx$x, 
          mean = mean(x$populationSize$boot),
          sd   = sd(x$populationSize$boot)
        ) * length(x$populationSize$boot) * diff(h$breaks)[1], 
        lty = 3, 
        col = 9
      )
      graphics::lines(
        x = xxx$x, 
        y = xxx$y * length(x$populationSize$boot) * diff(h$breaks)[1], 
        lty = 4, 
        col = 10
      )
      graphics::legend(
        "topright",
        c("log-normal density", "normal density", "Epanechnikov kernel"),
        lty = 2:4,
        col = 8:10
      )
    }
  },
  rootogram = {
    M <- marginalFreq(x);
    FF <- M$table;
    FF[names(M$y)] <- M$y;
    FF[setdiff(names(M$table), names(M$y))] <- 0;
    FF <- FF[-1];
    bp <- graphics::barplot(
      sqrt(FF),
      offset = sqrt(M$table[-1]) - sqrt(FF),
      ylab = expression(sqrt("Frequency")),
      xlab = "captures",
      ylim = c(min(sqrt(M$table[-1]) - sqrt(FF)) - 1, max(sqrt(M$table[-1]) + 1)),
      ...);
    graphics::lines(bp, sqrt(M$table[-1]), 
          type = "o", 
          pch = 19, 
          lwd = 2, 
          col = 2);
    graphics::abline(h = 0, 
           lty = 2)
  },
  dfpopContr = {
    if (missing(dfpop)) dfpop <- dfpopsize(x, ...);
    contr <- x$model$pointEst(
      pw = x$priorWeights,
      eta = x$linearPredictors, 
      contr = TRUE
    );
    plot(x = dfpop, y = contr,
         main = paste0("Observation deletion effect on point estimate of",
                       "\npopulation size estimate vs observation contribution"),
         xlab = "Deletion effect", ylab = "Observation contribution", 
         ...);
    abline(a = 0, b = 1, col = "red")
  },
  dfpopBox = {
    if (missing(dfpop)) dfpop <- dfpopsize(x, ...);
    graphics::boxplot(
      dfpop, 
      ylab = "Deletion effect",
      main = paste0("Boxplot of observation deletion effect on",
                    "\npoint estimate of population size estimate"), 
      ...
    )
  },
  scaleLoc = {
    if (parNum == 1) {
      lp <- x$linearPredictors
      if (x$model$family == "zelterman") {
        lp <- lp[x$which$reg]
      }
      plot(y = sqrt(abs(res)), x = lp,
           xlab = "Linear predictors",
           ylab = expression(sqrt("Std. Pearson resid.")),
           main = "Scale-Location plot",
           ...);
      graphics::panel.smooth(y = sqrt(abs(res)), x = lp, iter = 0)
    } else {
      graphics::par(mfrow = c(parNum, 1));
      for (k in 1:parNum) {
        plot(y = sqrt(abs(res)), x = x$linearPredictors[, k],
             xlab = "Linear predictors",
             ylab = expression(sqrt("Pearson resid.")),
             main = "Scale-Location plot",
             sub = paste0("For linear predictors associated with: ", 
                          x$model$etaNames[k]),
             ...);
        graphics::panel.smooth(y = sqrt(abs(res)), 
                               x = x$linearPredictors[, k], 
                               iter = 0)
      }
    }
  },
  fitresid = {
    if (parNum == 1) {
      lp <- x$linearPredictors
      if (x$model$family == "zelterman") {
        lp <- lp[x$which$reg]
        res <- res[x$which$reg]
      }
      plot(y = res, x = lp,
           xlab = "Linear predictors",
           ylab = "Response residuals",
           main = "Residuals vs Fitted",
           ...);
      abline(lty = 2, col = "darkgrey", h = 0);
      graphics::panel.smooth(y = res, x = lp, iter = 0)
    } else {
      graphics::par(mfrow = c(parNum, 1));
      for (k in 1:parNum) {
        plot(y = res, x = x$linearPredictors[, k],
             xlab = "Linear predictors",
             ylab = "Response residuals",
             main = "Residuals vs Fitted",
             sub = paste0("For linear predictors associated with: ", 
                          x$model$etaNames[k]),
             ...);
        abline(lty = 2, col = "darkgrey", h = 0);
        graphics::panel.smooth(y = res, 
                               x = x$linearPredictors[, k], 
                               iter = 0)
      }
    }
  },
  cooks = {
    A <- cooks.distance.singleRStaticCountData(x);
    plot(A,
         main = "Cook's distance",
         ylab = "Cook's distance",
         xlab = "Observation index",
         ylim = c(0, max(A) * 1.1),
         ...)
  },
  hatplot = {
    A <- hatvalues.singleRStaticCountData(x, ...);
    graphics::par(mfrow = c(parNum, 1));
    for (k in 1:parNum) {
      plot(A[, k],
           xlab = "Observation index",
           ylab = "Hat values",
           main = paste0("For linear predictors associated with: ", 
                         x$model$etaNames[k]),
           ...)
    }
  },
  strata = {
    if (missing(confIntStrata)) confIntStrata <- "logNormal"
    result <- stratifyPopsize(x, ...)
    est <- result[, 3]
    obs <- result[, 2]
    nm <- result[, 1]
    if (confIntStrata == "logNormal") 
      cnf <- result[, 8:9] 
    else 
      cnf <- result[, 6:7]
    
    tilt <- 0
    plot(y = 1:NROW(result), x = est,
         xlim = range(cnf),
         xlab = "Sub population size estimate", ylab="",
         main = paste0(
           "Confidence intervals and point estimates for specified sub populations\n",
           "Observed population sizes are presented as navy coloured points"
         ),
         yaxt = "n", pch = 19,
         ...
    )
    points(y = 1:NROW(result), x = obs, col = "navy", pch = 19)
    axis(side = 2, at = 1:NROW(result), labels = FALSE)
    text(
      y = 1:NROW(result), 
      x = graphics::par("usr", no.readonly = TRUE)[3] - (range(cnf)[2] - range(cnf)[1]) / 20, 
      adj = 1,
      nm, 
      srt = tilt, 
      cex = .6,
      xpd = TRUE
    )
    arrows(cnf[ ,1], 1:NROW(result), 
           cnf[ ,2], 1:NROW(result), 
           length = 0.05, 
           angle  = 90, 
           code   = 3)
  })
  
  invisible()
}

#' @rdname singleRmodels
#' @importFrom lamW lambertW0
#' @export
ztHurdlepoisson <- function(lambdaLink = c("log", "neglog"), 
                            piLink = c("logit", "cloglog", "probit"), 
                            ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  if (missing(piLink))  piLink <- "logit"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink, piLink)
  
  lambdaLink <- switch(lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  piLink <- switch(piLink,
    "logit" = singleRinternallogitLink,
    "cloglog" = singleRinternalcloglogLink,
    "probit" = singleRinternalprobitLink
  )
  
  links[1:2] <- c(lambdaLink, piLink)
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    if (!deriv) {
      switch (type,
        "nontrunc" = PI - PI * lambda * exp(-lambda) + (lambda - lambda * exp(-lambda)) * (1 - PI),
        "trunc" = PI + (1 - PI) * (lambda - lambda * exp(-lambda)) / (1 - exp(-lambda) - lambda * exp(-lambda))
      )
    } else {
      switch (type,
        "nontrunc" = {
          matrix(c(
            -exp(-lambda) * ((PI - 1) * exp(lambda) - lambda + 1),
            1 - lambda
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
                piLink(eta[, 2], inverse = TRUE, deriv = 1)
          ), ncol = 2)
        },
        "trunc" = {
          matrix(c(
            (1 - PI) * (exp(2 * lambda) + (-lambda ^ 2 - 2) * exp(lambda) + 1) /
            (exp(lambda) - lambda - 1) ^ 2,
            -(lambda - lambda * exp(-lambda)) / 
            (1 - exp(-lambda) - lambda * exp(-lambda))
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
                piLink(eta[, 2], inverse = TRUE, deriv = 1)
          ), ncol = 2)
        }
      )
    }
  }
  
  variance <- function(eta, type = "nontrunc", ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    switch (type,
    "nontrunc" = PI * (1 - exp(-lambda)) + (1 - PI) * (lambda ^ 2 + lambda - lambda * exp(-lambda)),
    "trunc" = (PI + (1 - PI) * (lambda + lambda ^ 2 - lambda * exp(-lambda)) / (1 - exp(-lambda))) - (PI + (1 - PI) * lambda) ^ 2
    )
  }
  
  Wfun <- function(prior, eta, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    z <- PI
    ## expected for (1-z)Y
    YY <- mu.eta(eta) - PI
    
    G00 <- -piLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2 * 
    ((PI - z) / ((PI - 1) ^ 2 * PI) + (PI - z) / ((PI - 1) * PI ^ 2) -
    1 / ((PI - 1) * PI))
    
    G11 <- lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 * 
    (-YY * exp(2 * lambda) + ((1 - z) * lambda ^ 3 + (z - 1) * lambda ^ 2 + 2 * YY * lambda + 2 * YY) * 
    exp(lambda) + (-YY - z + 1) * lambda ^ 2 - YY * 2 * lambda - YY) / 
    (lambda ^ 2 * (exp(lambda) - lambda - 1) ^ 2)
    
    matrix(
      -c(G11 * prior, # lambda
         rep(0, NROW(eta)) * prior, # mixed
         rep(0, NROW(eta)) * prior, # mixed
         G00 * prior  # pi
      ),
      dimnames = list(rownames(eta), c("lambda", "mixed", "mixed", "pi")),
      ncol = 4
    )
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    z <- as.numeric(y == 1)
    weight <- weight / prior
    
    G1 <- lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
          (1 - z) * (-lambda / (-lambda - 1 + exp(lambda)) + y / lambda - 1)
    
    G0 <- piLink(eta[, 2], inverse = TRUE, deriv = 1) * 
          (z / PI - (1 - z) / (1 - PI))
    
    uMatrix <- matrix(c(G1, G0), ncol = 2)
    
    weight <- lapply(X = 1:nrow(weight), FUN = function (x) {
      matrix(as.numeric(weight[x, ]), ncol = 2)
    })
    
    pseudoResid <- sapply(X = 1:length(weight), FUN = function (x) {
      # in this case matrix is diagonal
      xx <- 1 / diag(weight[[x]])
      xx * uMatrix[x, ]
    })
    pseudoResid <- t(pseudoResid)
    dimnames(pseudoResid) <- dimnames(eta)
    pseudoResid
  }
  
  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, 
                           ...) {
    y <- as.numeric(y)
    if (is.null(weight)) {
      weight <- 1
    }
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X) / 2), rep(0, NROW(X) / 2))
    }
    
    z <- as.numeric(y == 1)
    
    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
      function(beta) {
        eta    <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        PI     <- piLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        logistic <- -sum(weight * (z * log(PI) + (1 - z) * log(1 - PI)))
        zot <- -sum(weight * (1 - z) * (y * log(lambda) - lambda - lgamma(y + 1) - log(1 - exp(-lambda) - lambda * exp(-lambda))))
        zot + logistic
      },
      function(beta) {
        eta    <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        PI     <- piLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        G1 <- lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * weight *
        (1 - z) * (-lambda / (-lambda - 1 + exp(lambda)) + y / lambda - 1)
        
        G0 <- piLink(eta[, 2], inverse = TRUE, deriv = 1) * 
              weight * (z / PI - (1 - z) / (1 - PI))
        
        if (NbyK) {
          XX <- 1:(attr(X, "hwm")[1])
          return(cbind(as.data.frame(X[1:nrow(eta), XX]) * G1, as.data.frame(X[-(1:nrow(eta)), -XX]) * G0))
        }
        if (vectorDer) {
          return(cbind(G1, G0))
        }
        as.numeric(c(G1, G0) %*% X)
      },
      function (beta) {
        lambdaPredNumber <- attr(X, "hwm")[1]
        eta    <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        PI     <- piLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        res <- matrix(nrow = length(beta), ncol = length(beta), 
                      dimnames = list(names(beta), names(beta)))
        
        G00 <- -piLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2 * 
        ((PI - z) / ((PI - 1) ^ 2 * PI) + (PI - z) / ((PI - 1) * PI ^ 2) -
        1 / ((PI - 1) * PI)) + piLink(eta[, 2], inverse = TRUE, deriv = 2) * 
        (z / PI - (1 - z) / (1 - PI))
        
        G11 <- lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 * 
        ((z - 1) * (y * exp(2 * lambda) + 
        (-lambda ^ 3 + lambda ^ 2 - 2 * y * lambda - 2 * y) * exp(lambda) +
        (y - 1) * lambda ^ 2 + 2 * y * lambda + y)) / 
        (lambda ^ 2 * (exp(lambda) - lambda - 1) ^ 2) +
        lambdaLink(eta[, 1], inverse = TRUE, deriv = 2) *
        (1 - z) * (-lambda / (-lambda - 1 + exp(lambda)) + y / lambda - 1)
        
        # PI^2 derivative
        res[-(1:lambdaPredNumber), -(1:lambdaPredNumber)] <- 
          t(as.data.frame(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)] * G00 * weight)) %*% 
          as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)])
        # lambda^2 derivative
        res[1:lambdaPredNumber, 1:lambdaPredNumber] <- 
          t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber] * G11 * weight)) %*% 
          X[1:(nrow(X) / 2), 1:lambdaPredNumber]
        
        #mixed
        res[1:lambdaPredNumber, -(1:lambdaPredNumber)] <- 0
        res[-(1:lambdaPredNumber), 1:lambdaPredNumber] <- 0
        
        res
      }
    )
  }
  
  validmu <- function(mu) {
    (sum(!is.finite(mu)) == 0) && all(0 < mu)
  }
  
  devResids <- function(y, eta, wt, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    # when pi = 0 distribution collapses to zotpoisson
    inverseFunction <- function(y) {stats::uniroot(
      f = function(x) {
        l <- exp(x)
        (l - l * exp(-l)) / (1 - exp(-l) - l * exp(-l)) - y
      }, 
      lower = -log(y), upper = y * 10, 
      tol = .Machine$double.eps
    )$root}
    
    yUnq <- unique(y)
    
    etaSat <- tryCatch(
      expr = {
        suppressWarnings(sapply(yUnq, 
          FUN = function(x) ifelse(x %in% c(1, 2), -Inf, inverseFunction(x))
        ))
      },
      error = function (e) {
        warning("Deviance residuals could not have been computed and zero vector will be returned instead.", call. = FALSE)
        NULL
      }
    )
    if (is.null(etaSat)) {
      return(rep(0, length(y)))
    }
    
    etaSat <- sapply(y, FUN = function(x) etaSat[yUnq == x])
    idealLambda <- exp(etaSat)
    diff <- ifelse(
      y == 1,
      -log(PI), ifelse(y == 2, 0,
      y * etaSat - idealLambda - log(1 - exp(-idealLambda) - idealLambda * exp(-idealLambda)) - lgamma(y + 1)) - 
      (log(1 - PI) + y * log(lambda) - lambda - log(1 - exp(-lambda) - lambda * exp(-lambda)) - lgamma(y + 1))
    )
    
    if (any(diff < 0)) {
      warning(paste0(
        "Some of differences between log likelihood in saturated model",
        " and fitted model were positive which indicates either:\n",
        "(1): A very good model fit or\n",
        "(2): Incorrect computation of saturated model",
        "\nDouble check deviance before proceeding"
      ))
      diff[diff < 0] <- 0
    }
    
    sign(y - mu.eta(eta = eta)) * sqrt(2 * wt * diff)
  }
  
  pointEst <- function (pw, eta, contr = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    N <- pw * (1 - lambda * exp(-lambda)) / (1 - exp(-lambda) - lambda * exp(-lambda))
    if(!contr) {
      N <- sum(N)
    }
    N
  }
  
  popVar <- function (pw, eta, cov, Xvlm, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    # w.r to PI
    bigTheta1 <- rep(0, nrow(eta))
    # w.r to lambda
    bigTheta2 <- -as.numeric(pw * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
      (exp(lambda) - 1) / (exp(lambda) - lambda - 1) ^ 2)
    
    bigTheta <- t(c(bigTheta2, bigTheta1) %*% Xvlm)
    
    f1 <-  t(bigTheta) %*% as.matrix(cov) %*% bigTheta
    
    f2 <- sum(pw * ((1 - lambda * exp(-lambda)) * exp(-lambda) / 
                    ((1 - exp(-lambda) - lambda * exp(-lambda)) ^ 2)))
    
    f1 + f2
  }
  
  dFun <- function (x, eta, type = c("trunc", "nontrunc"), ...) {
    if (missing(type)) type <- "trunc"
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    switch (type,
      "trunc" = {
        as.numeric(x == 1) * PI + as.numeric(x > 0) * 
        (1 - PI) * stats::dpois(x = x, lambda = lambda) / 
        (1 - exp(-lambda) - lambda * exp(-lambda))
      },
      "nontrunc" = {
        stats::dpois(x = x, lambda = lambda) / (1 - lambda * exp(-lambda)) *
        (as.numeric(x == 0) + as.numeric(x > 1) * (1 - PI)) + 
        as.numeric(x == 1) * PI * (1 - exp(-lambda) / (1 - lambda * exp(-lambda)))
      }
    )
  }
  
  simulate <- function(n, eta, lower = 0, upper = Inf) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    CDF <- function(x) {
      ifelse(x == Inf, 1, 
      ifelse(x < 0, 0, 
      ifelse(x < 1, exp(-lambda) / (1 - lambda * exp(-lambda)), 
      exp(-lambda) / (1 - lambda * exp(-lambda)) + PI * (1 - exp(-lambda) / 
      (1 - lambda * exp(-lambda))) + (1 - PI) * (stats::ppois(x, lambda) - 
      lambda * exp(-lambda) - exp(-lambda)) / (1 - lambda * exp(-lambda)))))
    }
    lb <- CDF(lower)
    ub <- CDF(upper)
    p_u <- stats::runif(n, lb, ub)
    sims <- rep(0, n)
    cond <- CDF(sims) <= p_u
    while (any(cond)) {
      sims[cond] <- sims[cond] + 1
      cond <- CDF(sims) <= p_u
    }
    sims
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12)),
        family$links[[2]](weighted.mean(observed == 1, priorWeights) * (.5 + .5 * (observed == 1)) + .01)
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights)),
        family$links[[2]](weighted.mean(observed == 1, priorWeights) + .01)
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
      if ("(Intercept):pi" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[2], rep(0, attr(Xvlm, "hwm")[2] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[2] / attr(Xvlm, "hwm")[2], attr(Xvlm, "hwm")[2]))
      }
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "ztHurdlepoisson",
      etaNames = c("lambda", "pi"),
      simulate  = simulate,
      getStart  = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

#' \loadmathjax

#' @title Estimate size of sub-populations
#' 
#' @description A function that estimates sizes of specific sub populations 
#' based on a capture-recapture model for the whole population.
#'
#' @param object an object on which the population size estimates should be based
#' in \code{singleRcapture} package this is a fitter \code{singleRStaticCountData} class object.
#' @param strata a specification of sub populations given by one of:
#' \itemize{
#' \item formula -- a formula to be applied to \code{model.frame} extracted from
#' the object.
#' \item Logical vector with number of entries equal to number of rows in the dataset.
#' \item A (named) list where each element is a logical vector, names of the list
#' will be used to specify names variable in returned object.
#' \item Vector of names of explanatory variables. For \code{singleRStaticCountData} method
#' for this function this specification of \code{strata} parameter will
#' result in every level of explanatory variable having its own sub population
#' for each variable specified.
#' \item If no value was provided the \code{singleRStaticCountData} method for this function 
#' will itself create sub populations based on levels of factor variables
#' in \code{model.frame}.
#' }
#' @param cov for \code{singleRStaticCountData} method an estimate of variance-covariance matrix
#' for estimate of regression parameters. It is possible to pass a function
#' such as for example \code{sandwich::vcovHC} which will be called as:
#' \code{foo(object, ...)} and a user may specify additional arguments of a 
#' function in \code{...} argument. If not provided an estimate for covariance
#' matrix will be set by calling appropriate \code{vcov} method.
#' @param alpha significance level for confidence intervals --
#' Either a single numeric value or a vector of length equal to number of 
#' sub populations specified in \code{strata}. 
#' If missing it is set to \code{.05} in \code{singleRStaticCountData} method.
#' @param ... a vector of arguments to be passed to other functions.
#' For \code{singleRStaticCountData} method for this functions arguments in \code{...} are 
#' passed to either \code{cov} if argument provided was a function or 
#' \code{vcov} if \code{cov} argument was missing at call.
#' 
#' @details In single source capture-recapture models the most frequently used
#' estimate for population size is Horvitz-Thompson type estimate:
#' 
#' \mjsdeqn{\hat{N} = \sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}(Y_{k}>0)} = 
#' \sum_{k=1}^{N_{obs}}\frac{1}{1-\mathbb{P}(Y_{k}=0)}}
#'
#' where \mjseqn{I_{k}=I_{Y_{k} > 0}} are 
#' indicator variables, with value 1 if kth unit was observed at least once 
#' and 0 otherwise and the inverse probabilistic weights weights for 
#' units observed in the data \mjseqn{\tfrac{1}{\mathbb{P}(Y_{k}>0)}}
#' are estimated using fitted linear predictors.
#' 
#' The estimates for different sub populations are made by changing the
#' \mjseqn{I_{k}=I_{Y_{k} > 0}} indicator variables to 
#' refer not to the population as a whole but to the sub populations that are 
#' being considered i.e. by changing values from 1 to 0 if kth unit is not a 
#' member of sub population that is being considered at the moment.
#' 
#' The estimation of variance for these estimates and estimation of variance for
#' estimate of population size for the whole population follow the same relation
#' as the one described above.
#' 
#' @seealso [vcov.singleRStaticCountData()] [estimatePopsize()]
#'
#' @return A \code{data.frame} object with row names being the names of specified 
#' sub populations either provided or inferred.
#' @export
stratifyPopsize <- function(object, strata, alpha, ...) {
  UseMethod("stratifyPopsize")
}



#' @method stratifyPopsize singleRStaticCountData
#' @rdname stratifyPopsize
#' @importFrom stats vcov
#' @importFrom stats contrasts
#' @exportS3Method
stratifyPopsize.singleRStaticCountData <- function(object, 
                                                   strata,
                                                   alpha, 
                                                   cov = NULL,
                                                   ...) {
  if (missing(strata)) {
    strata <- names(which(attr(object$terms, "dataClasses") == "factor"))
    strata <- strata[strata %in% attr(object$terms, "term.labels")]
    if (!length(strata)) {
      strata <- names(which(attr(object$terms, "dataClasses") == "character"))
      strata <- strata[strata %in% attr(object$terms, "term.labels")]
    }
    if (!length(strata)) {
      stop("No strata argument was provided and no factors or character columns are present in model.frame.")
    }
  }
  if (missing(alpha)) alpha <- .05
  
  # convert strata to list for all viable types of specifying the argument
  if (inherits(strata, "formula")) {
    mf <- model.frame(strata, model.frame(object))
    mmf <- model.matrix(
      strata, data = mf, 
      contrasts.arg = lapply(
        # this makes it so that all levels of factors are encoded
        subset(mf, select = sapply(mf, is.factor)),
        contrasts, contrasts = FALSE
      )
    )
    trm <- attr(mf, "terms")
    strata <- list()
    for (k in attr(trm, "term.labels")) {
      if (k %in% colnames(mf)) {
        if (is.integer(mf[, k]) | is.character(mf[, k])) {
          for (t in unique(mf[,k])) {
            strata[[paste0(k, "==", t)]] <- mf[,k] == t
          }
        } else if (is.factor(mf[, k])) {
          for (t in levels(mf[, k])) {
            strata[[paste0(k, "==", t)]] <- mf[,k] == t
          }
        }
      } else {
        tLevs <- colnames(mmf)[attr(mmf, "assign") == which(attr(trm, "term.labels") == k)]
        for (t in tLevs) {
          strata[[as.character(t)]] <- mmf[, t] == 1
        }
      }
    }
  } else if (is.list(strata)) {
    if (!all(sapply(strata, is.logical)))
      stop("Invalid way of specifying subpopulations in strata. If strata argument is a list then all elements of that list must be logical vectors.")
    
    if (length(strata[[1]]) != object$sizeObserved) 
      stop("Elements of strata object should have length equal to number of observed units.")
    
  } else if (is.logical(strata)) {
    if (length(strata) != object$sizeObserved) 
      stop("strata object should have length equal to number of observed units.")
    
    strata <- list(strata = strata)
  } else if (is.character(strata)) {
    modelFrame <- model.frame(object)
    out <- list()
    for (k in strata) {
      if (!(k %in% colnames(modelFrame))) 
        stop("Variable specified in strata is not present in model frame.")

      if (is.factor(modelFrame[, k])) {
        # this makes a difference on factors that have leves which are not represented in the data
        for (t in levels(modelFrame[, k])) {
          out[[paste0(as.character(k), "==", t)]] <- (modelFrame[, k] == t)
        }
      } else {
        for (t in unique(modelFrame[, k])) {
          out[[paste0(as.character(k), "==", t)]] <- (modelFrame[, k] == t)
        }
      }
    }
    strata <- out
  } else {
    errorMessage <- paste0(
      "Invalid way of specifying subpopulations in strata.\n", 
      "Please provide either:\n",
      "(1) - a list with logical vectors specifying different sub populations\n",
      "(2) - a single logical vector\n",
      "(3) - a formula\n",
      "(4) - a vector with names of variables by which strata will be created\n"
    )
    stop(errorMessage)
  }
  
  family <- family(object = object)
  priorWeights <- object$priorWeights
  eta <- object$linearPredictors
  Xvlm <- model.matrix(object, "vlm")
  y <- if (is.null(object$y)) model.response(model.frame(object)) else object$y
  flagWeighting <- object$control$controlModel$weightsAsCounts
  
  if (is.function(cov)) cov <- cov(object, ...)
  if (is.null(cov)) cov <- vcov(object, ...)
  
  obs <- vector(mode = "numeric", length = length(strata))
  est <- vector(mode = "numeric", length = length(strata))
  stdErr <- vector(mode = "numeric", length = length(strata))
  cnfStudent <- matrix(nrow = length(strata), ncol = 2)
  cnfChao <- matrix(nrow = length(strata), ncol = 2)
  sc <- qnorm(p = 1 - alpha / 2)
  if (length(sc) != length(strata)) sc <- rep(sc, length.out = length(strata))
  
  for (k in 1:length(strata)) {
    cond <- strata[[k]]
    
    if (isTRUE(object$control$controlModel$weightsAsCounts)) {
      obs[k] <- sum(priorWeights[cond])
    } else {
      obs[k] <- sum(cond)
    }
    
    if (obs[k] > 0) {
      est[k] <- family$pointEst(pw  = priorWeights[cond], 
                                eta = eta[cond, , drop = FALSE],
                                y   = y[cond])
      
      stdErr[k] <- family$popVar(
        pw = priorWeights[cond], 
        eta = eta[cond, , drop = FALSE], 
        cov = cov, 
        Xvlm = subset(Xvlm, subset = rep(cond, length(family$etaNames))),
        y   = y[cond]
      ) ^ .5
      
      cnfStudent[k, ] <- est[k] + c(-sc[k] * stdErr[k], sc[k] * stdErr[k])
      
      G <- exp(sc[k] * sqrt(log(1 + (stdErr[k]^2) / ((est[k] - obs[k]) ^ 2))))
      cnfChao[k, ] <- obs[k] + c((est[k] - obs[k]) / G, (est[k] - obs[k]) * G)
    } else {
      est[k] <- 0
      stdErr[k] <- 0
      cnfStudent[k, ] <- c(0, 0)
      cnfChao[k, ] <- c(0, 0)
    }
  }
  
  result <- data.frame(
    names(strata), obs, est, 
    100 * obs / est, stdErr, 
    cnfStudent[, 1], cnfStudent[, 2], 
    cnfChao[, 1], cnfChao[, 2], alpha
  )
  
  bounds <- c("LowerBound", "UpperBound")
  
  colnames(result) <- c(
    "name", "Observed", "Estimated", 
    "ObservedPercentage", "StdError", 
    paste0("normal", bounds), 
    paste0("logNormal", bounds), "confLevel"
  )
  
  result
}

# internal functions

# logit link
singleRinternallogitLink <- function(x,
                                     inverse = FALSE,
                                     deriv = 0) {
  deriv <- deriv + 1
  if (isFALSE(inverse)) {
    res <- switch(
      deriv,
      log(x / (1 - x)),
      1 / (x * (1 - x)),
      (2 * x - 1) / ((x * (1 - x)) ^ 2), 
      -2 * (3 * (x ^ 2) - 3 * x + 1) / ((x * (1 - x)) ^ 3)
    )
  } else {
    res <- switch(
      deriv,
      1 / (1 + exp(-x)),
      exp(x) / ((1 + exp(x)) ^ 2),
      -exp(x) * (exp(x) - 1) / ((1 + exp(x)) ^ 3),
      exp(x) * (exp(2 * x) - 4 * exp(x) + 1) / ((1 + exp(x)) ^ 4)
    )
  }
  
  res
}

# log link
singleRinternallogLink <- function(x,
                                  inverse = FALSE,
                                  deriv = 0) {
  deriv <- deriv + 1
  if (isFALSE(inverse)) {
    res <- switch(
      deriv,
      log(x),
      1 / x, 
      -1 / (x ^ 2), 
      2 / (x ^ 3)
    )
  } else {
    res <- exp(x)
  }
  
  res
}

# half log link for chao
singleRinternalloghalfLink <- function(x,
                                   inverse = FALSE,
                                   deriv = 0) {
  deriv <- deriv + 1
  if (isFALSE(inverse)) {
    res <- switch(
      deriv,
      log(x / 2),
      1 / x, 
      -1 / (x ^ 2),
      2 / (x ^ 3) 
    )
  } else {
    res <- 2 * exp(x)
  }
  
  res
}

# third log link for modified chao
singleRinternallogthirdLink <- function(x,
                                        inverse = FALSE,
                                        deriv = 0) {
  deriv <- deriv + 1
  if (isFALSE(inverse)) {
    res <- switch(
      deriv,
      log(x / 3),
      1 / x,
      -1 / (x ^ 2),
      2 / (x ^ 3)
    )
  } else {
    res <- 3 * exp(x)
  }
  
  res
}

# cloglog
singleRinternalcloglogLink <- function(x,
                                     inverse = FALSE,
                                     deriv = 0) {
  deriv <- deriv + 1
  if (isFALSE(inverse)) {
    res <- switch(
      deriv,
      log(-log(1 - x)),
      -1 / ((1 - x) * log(1 - x)),
      -(1 + log(1 - x)) / ((x - 1) ^ 2 * log(1 - x) ^ 2),
      (2*log(1 - x) ^ 2 + 3 * log(1 - x) + 2) / (log(1 - x) ^ 3 * (x - 1) ^ 3)
    )
  } else {
    res <- switch(
      deriv,
      1 - exp(-exp(x)),
      exp(x - exp(x)),
      (1 - exp(x)) * exp(x - exp(x)),
      (exp(2 * x) - 3 * exp(x) + 1) * exp(x - exp(x))
    )
  }
  
  res
}

# probit
#' @importFrom stats dnorm
#' @importFrom stats pnorm
singleRinternalprobitLink <- function(x,
                                      inverse = FALSE,
                                      deriv = 0) {
  deriv <- deriv + 1
  if (isFALSE(inverse)) {
    res <- switch(deriv,
      qnorm(x),
      1/dnorm(qnorm(x)),
      qnorm(x) / (dnorm(qnorm(x))) ^ 2,
      (1 + 2 * qnorm(x) ^ 2) / dnorm(qnorm(x)) ^ 3
    )
  } else {
    res <- switch(deriv,
      pnorm(x),
      dnorm(x),
      -(x * exp(-x ^ 2 / 2)) / (2 * pi) ^ .5,
      ((x ^ 2 - 1) * exp(-x ^ 2 / 2)) / (2 * pi) ^ .5
    )
  }
  
  res
}


# neglog link
singleRinternalneglogLink <- function(x,
                                     inverse = FALSE,
                                     deriv = 0) {
  if (isFALSE(inverse)) {
    deriv <- deriv + 1
    res <- switch(
      deriv,
      -log(x),
      -1 / x, 
      1 / x ^ 2, 
      -2 / x ^ 3 
    )
  } else {
    res <- exp(-x) * ((-1) ^ deriv)
  }
  
  res
}

#' @rdname singleRmodels
#' @export
ztgeom <- function(lambdaLink = c("log", "neglog"),
                   ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink)
  
  lambdaLink <- switch (lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  links[1] <- c(lambdaLink)
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    if (!deriv) {
      switch (
        type,
        "nontrunc" = lambda,
        "trunc" = 1 + lambda
      )
    } else {
      switch (type,
        "nontrunc" = cbind(lambdaLink(eta, inverse = TRUE, deriv = 1)),
        "trunc" = cbind(lambdaLink(eta, inverse = TRUE, deriv = 1))
      )
    }
  }
  
  variance <- function(eta, type = "nontrunc", ...) {
    mu <- mu.eta(eta)
    switch (type,
      nontrunc = mu ^ 2 - mu - 1 / mu + 2,
      trunc    = (mu + 1) / mu
    )
  }
  
  Wfun <- function(prior, eta, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    matrix(
      prior / (lambda * (lambda + 1)) * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2,
      ncol = 1, dimnames = list(rownames(eta), c("lambda"))
    )
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    prior * ((y - 1) / lambda - y / (1 + lambda)) * 
    lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) / weight
  }
  
  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, 
                           ...) {
    if (is.null(weight)) {
      weight <- 1
    }
    y <- as.numeric(y)
    X <- as.matrix(X)
    
    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
      function(beta) {
        eta <- as.matrix(X) %*% beta + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        -sum(weight * ((y - 1) * log(lambda) - y * log(1 + lambda)))
      },
      function(beta) {
        eta <- X %*% beta + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        G1 <- (y - 1) / lambda - y / (1 + lambda)
        G1 <- weight * G1 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
        if (NbyK) {
          return(as.data.frame(X) * G1)
        }
        if (vectorDer) {
          return(matrix(G1, ncol = 1))
        }
        t(G1) %*% X
      },
      function(beta) {
        eta <- X %*% beta + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        G1 <- (y - 1) / lambda - y / (1 + lambda)
        G1 <- G1 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 2)
        
        G11 <- y / (lambda + 1) ^ 2 - (y - 1) / lambda ^ 2
        G11 <- G11 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2
        
        t(as.data.frame(X) * (G1 + G11) * weight) %*% X
      },
    )
  }
  
  validmu <- function(mu) {
    (sum(!is.finite(mu)) == 0) && all(0 < mu)
  }
  
  devResids <- function (y, eta, wt, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    hm1y <- y - 1
    loghm1y <- ifelse(y > 1, log(hm1y), 0)
    
    diff <- (y - 1) * log(lambda) - y * log(1 + lambda) - 
            (y - 1) * loghm1y + y * log(y)
    
    diff[diff > 0] <- -0
    sign(y - 1 - lambda) * sqrt(-2 * wt * diff)
  }
  
  pointEst <- function (pw, eta, contr = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    N <- pw / (1 - 1 / (1 + lambda))
    if(!contr) {
      N <- sum(N)
    }
    N
  }
  
  popVar <- function (pw, eta, cov, Xvlm, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    pr <- 1 - 1 / (1 + lambda)
    
    bigTheta <- -(lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * pw / lambda ^ 2) %*% as.matrix(Xvlm)
    bigTheta <- as.vector(bigTheta)
    
    f1 <- t(bigTheta) %*% as.matrix(cov) %*% bigTheta
    f2 <- sum(pw * (1 - pr) / (pr ^ 2))
    
    f1 + f2
  }
  
simulate <- function(n, eta, lower = 0, upper = Inf) {
  lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    lb <- stats::pnbinom(lower, mu=lambda, size = 1)
    ub <- stats::pnbinom(upper, mu=lambda, size = 1)
    p_u <- stats::runif(n, lb, ub)
    sims <- stats::qnbinom(p_u, mu=lambda, size = 1)
    sims
  }
  
  dFun <- function (x, eta, type = c("trunc", "nontrunc")) {
    if (missing(type)) type <- "trunc"
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    switch (type,
      "trunc" = {
        stats::dgeom(x = x, prob = 1 / (1 + lambda)) / 
        (1 - stats::dgeom(x = 0, prob = 1 / (1 + lambda)))
      },
      "nontrunc" = stats::dgeom(x = x, prob = 1 / (1 + lambda))
    )
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12))
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights))
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "ztgeom",
      etaNames  = c("lambda"),
      simulate  = simulate,
      getStart  = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

source("helper_families.R")
zerotruncated_test_data <- load_zerotruncated_test_data()
eps <- zerotruncated_test_data$eps
N <- zerotruncated_test_data$N
test_zerotruncated <- zerotruncated_test_data$test_zerotruncated
x1 <- zerotruncated_test_data$x1
x2 <- zerotruncated_test_data$x2
x3 <- zerotruncated_test_data$x3

## geometric ####
coefficients_ztgeom <- c(.6, .2, -1.25, .1)
#eta <- cbind(coefficients_ztgeom[1] + coefficients_ztgeom[2] * x1 + coefficients_ztgeom[3] * x2 + coefficients_ztgeom[4] * x3)

model_ztgeom <- ztgeom(
  lambdaLink = "neglog"
)

#y <- model_ztgeom$simulate(n = N, eta = eta, lower = -1)
observed_ztgeom <- test_zerotruncated$ztgeom1

data_ztgeom <- data.frame(
  x3 = x3[observed_ztgeom > 0],
  x2 = x2[observed_ztgeom > 0],
  x1 = x1[observed_ztgeom > 0],
  y = observed_ztgeom[observed_ztgeom > 0]
)

fit_ztgeom <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = model_ztgeom,
  data    = data_ztgeom,
  method = "IRLS"
)

expect_equivalent(
  coef(fit_ztgeom),
  coefficients_ztgeom,
  tolerance = .25
)

expect_silent(
  population_ztgeom <- popSizeEst(fit_ztgeom)
)

expect_equivalent(
  N,
  population_ztgeom$pointEstimate,
  tol = .1
)

expect_true(
  all(N < population_ztgeom$confidenceInterval$upperBound) &
  all(N > population_ztgeom$confidenceInterval$lowerBound)
)

expect_silent(
  strata_ztgeom <- stratifyPopsize(fit_ztgeom, ~ x1)
)

expect_equivalent(
  as.numeric(table(data_ztgeom$x1))[as.numeric(substr(strata_ztgeom$name, start = 5, stop = 5)) + 1],
  strata_ztgeom$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(strata_ztgeom$name, start = 5, stop = 5)) + 1],
  strata_ztgeom$Estimated,
  tolerance = .2
)

marginal_ztgeom <- summary(marginalFreq(fit_ztgeom), dropl5 = "group")

expect_true(
  all(marginal_ztgeom$Test$`P(>X^2)` > .03)
)

expect_true(
  all(predict(
    fit_ztgeom,
    type = "response",
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    fit_ztgeom,
    type = "link",
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    fit_ztgeom,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztgeom,
    type = "contr"
  )) == fit_ztgeom$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztgeom,
    type = "popSize",
    se.fit = TRUE
  )
)

# different link

model_ztgeom <- ztgeom(
  lambdaLink = "log"
)

#y <- model_ztgeom$simulate(n = N, eta = eta, lower = -1)
observed_ztgeom <- test_zerotruncated$ztgeom2

data_ztgeom <- data.frame(
  x3 = x3[observed_ztgeom > 0],
  x2 = x2[observed_ztgeom > 0],
  x1 = x1[observed_ztgeom > 0],
  y = observed_ztgeom[observed_ztgeom > 0]
)

fit_ztgeom <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = model_ztgeom,
  data    = data_ztgeom,
  method = "IRLS"
)

expect_equivalent(
  coef(fit_ztgeom),
  coefficients_ztgeom,
  tolerance = .2
)

expect_silent(
  population_ztgeom <- popSizeEst(fit_ztgeom)
)

expect_equivalent(
  N,
  population_ztgeom$pointEstimate,
  tol = .05
)

expect_true(
  all(N < population_ztgeom$confidenceInterval$upperBound) &
  all(N > population_ztgeom$confidenceInterval$lowerBound)
)

expect_silent(
  strata_ztgeom <- stratifyPopsize(fit_ztgeom, ~ x1)
)

expect_equivalent(
  as.numeric(table(data_ztgeom$x1))[as.numeric(substr(strata_ztgeom$name, start = 5, stop = 5)) + 1],
  strata_ztgeom$Observed,
  tolerance = .1
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(strata_ztgeom$name, start = 5, stop = 5)) + 1],
  strata_ztgeom$Estimated,
  tolerance = .1
)

marginal_ztgeom <- summary(marginalFreq(fit_ztgeom), dropl5 = "group")

expect_true(
  all(marginal_ztgeom$Test$`P(>X^2)` > .05)
)

expect_true(
  all(predict(
    fit_ztgeom,
    type = "response",
    se.fit = TRUE
  )[, 2] > 0)
)

expect_true(
  all(predict(
    fit_ztgeom,
    type = "link",
    se.fit = TRUE
  )[,2]> 0)
)

expect_true(
  all(predict(
    fit_ztgeom,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztgeom,
    type = "contr"
  )) == fit_ztgeom$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztgeom,
    type = "popSize",
    se.fit = TRUE
  )
)

#' @rdname singleRmodels
#' @importFrom stats uniroot
#' @importFrom stats dnbinom
#' @export
oiztnegbin <- function(nSim = 1000, epsSim = 1e-8, eimStep = 6,
                       lambdaLink = c("log", "neglog"), 
                       alphaLink = c("log", "neglog"),
                       omegaLink = c("logit", "cloglog", "probit"), ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  if (missing(alphaLink))  alphaLink <- "log"
  if (missing(omegaLink))  omegaLink <- "logit"

  links <- list()
  attr(links, "linkNames") <- c(lambdaLink, alphaLink, omegaLink)
  
  lambdaLink <- switch(lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  alphaLink <- switch(alphaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  omegaLink <- switch(omegaLink,
    "logit" = singleRinternallogitLink,
    "cloglog" = singleRinternalcloglogLink,
    "probit" = singleRinternalprobitLink
  )
  
  links[1:3] <- c(lambdaLink, alphaLink, omegaLink)
  
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    omega  <-  omegaLink(eta[, 3], inverse = TRUE)
    
    if (!deriv) {
      switch (type,
        "nontrunc" = omega * (1 - (1 + alpha * lambda) ^ (-1 / alpha)) + (1 - omega) * lambda,
        "trunc" = omega + (1 - omega) * lambda / (1 - (1 + alpha * lambda) ^ (-1 / alpha))
      )
    } else {
      switch (
        type,
        "nontrunc" = {
          matrix(c(
            1 + omega * (alpha * lambda + 1) ^ (-1 / alpha - 1) - omega,
            -omega * (log(lambda * alpha + 1) / alpha ^ 2 -
            lambda / (alpha * (lambda * alpha + 1))) / 
            (lambda * alpha + 1) ^ (1 / alpha),
            1 - lambda - 1 / (alpha * lambda + 1) ^ (1 / alpha)
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
             alphaLink(eta[, 2], inverse = TRUE, deriv = 1),
             omegaLink(eta[, 3], inverse = TRUE, deriv = 1)
          ), ncol = 3)
        },
        "trunc" = {
          matrix(c(
            (1 - omega) * (alpha * lambda + 1) ^ (1 / alpha - 1) *
            ((alpha * lambda + 1) ^ (1 / alpha + 1) +
            (-alpha - 1) * lambda - 1) / 
            ((alpha * lambda + 1) ^ (1 / alpha) - 1) ^ 2,
            (1 - omega) * lambda * (lambda * alpha + 1) ^ (1 / alpha - 1) *
            ((lambda * alpha + 1) * log(lambda * alpha + 1) - lambda * alpha) /
            (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha) - 1) ^ 2),
            -lambda / (1 - (1 + alpha * lambda) ^ (-1 / alpha))
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
             alphaLink(eta[, 2], inverse = TRUE, deriv = 1),
             omegaLink(eta[, 3], inverse = TRUE, deriv = 1)
          ), ncol = 3)
        }
      )
    }
  }
  
  variance <- function(eta, type = "nontrunc", ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    omega  <-  omegaLink(eta[, 3], inverse = TRUE)
    switch (type,
      nontrunc = omega * (1 - (1 + alpha * lambda) ^ (-1 / alpha)) + 
      (1 - omega) * (lambda + (1 + alpha) * lambda ^ 2),
      trunc = omega + (1 - omega) * (lambda + (1 + alpha) * lambda ^ 2) / 
      (1 - (1 + alpha * lambda) ^ (-1 / alpha))
    ) - mu.eta(eta = eta, type = type) ^ 2
  }
  
  compdigamma <- function(y, alpha) {
    (-digamma(y + 1 / alpha) + digamma(1 / alpha)) / (alpha ^ 2)
  }
  
  
  comptrigamma <- function(y, alpha) {
    (2 * (digamma(y + 1 / alpha) - digamma(1 / alpha)) * alpha +
    trigamma(y + 1 / alpha) - trigamma(1 / alpha)) / (alpha ^ 4)
  }
  
  # Computing the expected value of di/trigamma functions on (y + 1/alpha)
  compExpectG1 <- function(eta) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    omega  <-  omegaLink(eta[, 3], inverse = TRUE)
    
    P0 <- (1 + alpha * lambda) ^ (-1 / alpha)
    #P0 <- stats::dnbinom(x = 0, size = 1 / alpha, mu = lambda)
    res <- rep(0, NROW(eta))
    k <- 2
    finished <- rep(FALSE, NROW(eta))
    while ((k < nSim) & !all(finished)) {
      prob <- apply(cbind(k:(k + eimStep)), MARGIN = 1, FUN = function(x) {
        (1 - omega) * stats::dnbinom(
          x = x, 
          size = 1 / alpha, 
          mu = lambda
        ) / (1 - P0)
      })
      trg <- apply(cbind(k:(k + eimStep)), MARGIN = 1, FUN = function(x) {
        comptrigamma(y = x, alpha = alpha)
      })
      prob[!(is.finite(prob))] <- 0
      trg[!(is.finite(trg))] <- 0
      toAdd <- trg * prob
      toAdd <- rowSums(toAdd)
      k <- k + eimStep + 1
      res <- res + toAdd
      finished <- abs(toAdd) < epsSim
    }
    res
  }
  
  Wfun <- function(prior, eta, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    omega  <-  omegaLink(eta[, 3], inverse = TRUE)
    
    z  <- omega + (1 - omega) * lambda * 
    ((1 + alpha * lambda) ^ (-1 / alpha - 1)) / 
    (1 - (1 + alpha * lambda) ^ (-1 / alpha))
    
    XXX <- mu.eta(eta, type = "trunc") - z
    
    Etrig <- compExpectG1(eta)
    
    # omega
    G00 <- (-z * ((alpha * lambda + 1) ^ (1 / alpha + 1) +
    (-alpha - 1) * lambda - 1) ^ 2) / (((alpha * lambda + 1) ^ (1 / alpha + 1) +
    (-alpha - 1) * lambda - 1) * omega + lambda) ^ 2 -
    (1 - z) / (1 - omega) ^ 2
    
    G00 <- G00 * omegaLink(eta[, 3], inverse = TRUE, deriv = 1) ^ 2
    
    # omega alpha
    G01 <- (-z) * lambda * ((alpha * lambda + 1) ^ (1 / alpha + 1) *
    log(alpha * lambda + 1) + (-alpha ^ 2 - alpha) * lambda * 
    (alpha * lambda + 1) ^ (1 / alpha) + alpha ^ 2 * lambda) /
    (alpha ^ 2 * (((alpha * lambda + 1) ^ (1 / alpha + 1) +
    (-alpha - 1) * lambda - 1) * omega + lambda) ^ 2)
    
    G01 <- G01 * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) *
                 omegaLink(eta[, 3], inverse = TRUE, deriv = 1)
    
    # omega lambda
    G02 <- z * ((lambda - 1) * (alpha * lambda + 1) ^ (1 / alpha) + 1) /
    (((alpha * lambda + 1) ^ (1 / alpha + 1) + (-alpha - 1) * lambda - 1) * omega + lambda) ^ 2
    
    G02 <- G02 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
                  omegaLink(eta[, 3], inverse = TRUE, deriv = 1)
    
    # alpha
    G11 <- (1 - z) * ((log(lambda * alpha + 1) / alpha ^ 2 - lambda / 
    (alpha * (lambda * alpha + 1))) ^ 2 / ((lambda * alpha + 1) ^ (1 / alpha) *
    (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha))) + (log(lambda * alpha + 1) / alpha ^ 2 -
    lambda / (alpha * (lambda * alpha + 1))) ^ 2 / ((lambda * alpha + 1) ^ (2 / alpha) *
    (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha)) ^ 2) + (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 +
    (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) + lambda ^ 2 / 
    (alpha * (lambda * alpha + 1) ^ 2)) / ((lambda * alpha + 1) ^ (1 / alpha) * 
    (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha))) - (2 * log(lambda * alpha + 1)) / alpha ^ 3 + 
    2 * lambda / (alpha ^ 2 * (lambda * alpha + 1))) - 
    (lambda ^ 2 * (-(1 - z) / alpha - XXX)) / 
    (lambda * alpha + 1) ^ 2 - XXX / alpha ^ 2 + Etrig +
    (1 - omega) * z * lambda * ((lambda * alpha + 1) ^ (1 / alpha + 1) *
    log(lambda * alpha + 1) * ((lambda * (1 / alpha + 1)) / (lambda * alpha + 1) -
    log(lambda * alpha + 1) / alpha ^ 2) + (lambda * alpha + 1) ^ (1 / alpha) * 
    (-lambda * alpha ^ 2 - lambda * alpha) * (lambda / (alpha * (lambda * alpha + 1)) - 
    log(lambda * alpha + 1) / alpha ^ 2) + (-2 * lambda * alpha - lambda) * 
    (lambda * alpha + 1) ^ (1 / alpha) + lambda * (lambda * alpha + 1) ^ (1 / alpha) +
    2 * lambda * alpha) / (alpha ^ 2 * (lambda * alpha + 1) * 
    ((lambda * alpha + 1) ^ (1 / alpha) * (omega * lambda * alpha + omega) - 
    omega * lambda * alpha + (1 - omega) * lambda - omega) * ((lambda * alpha + 1) ^ (1 / alpha) - 1)) -
    (1 - omega) * z * lambda * ((lambda * alpha + 1) ^ (1 / alpha + 1) * log(lambda * alpha + 1) +
    (lambda * alpha + 1) ^ (1 / alpha) * (-lambda * alpha ^ 2 - lambda * alpha) + lambda * alpha ^ 2) *
    ((lambda * alpha + 1) ^ (1 / alpha) * (omega * lambda * alpha + omega) *
    (lambda / (alpha * (lambda * alpha + 1)) - log(lambda * alpha + 1) / alpha ^ 2) + 
    omega * lambda * (lambda * alpha + 1) ^ (1 / alpha) - omega * lambda) / 
    (alpha ^ 2 * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) *
    (omega * lambda * alpha + omega) - omega * lambda * alpha + (1 - omega) * lambda - omega) ^ 2 *
    ((lambda * alpha + 1) ^ (1 / alpha) - 1)) + ((omega - 1) * z * lambda * 
    (lambda * alpha + 1) ^ (1 / alpha - 1) * (lambda / (alpha * (lambda * alpha + 1)) -
    log(lambda * alpha + 1) / alpha ^ 2) * ((lambda * alpha + 1) ^ (1 / alpha + 1) * 
    log(lambda * alpha + 1) + (lambda * alpha + 1) ^ (1 / alpha) *
    (-lambda * alpha ^ 2 - lambda * alpha) + lambda * alpha ^ 2)) /
    (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha) * (omega * lambda * alpha + omega) -
    omega * lambda * alpha + (1 - omega) * lambda - omega) * ((lambda * alpha + 1) ^ (1 / alpha) - 1) ^ 2) +
    (2 * (omega - 1) * z * lambda * ((lambda * alpha + 1) ^ (1 / alpha + 1) * 
    log(lambda * alpha + 1) + (lambda * alpha + 1) ^ (1 / alpha) * 
    (-lambda * alpha ^ 2 - lambda * alpha) + lambda * alpha ^ 2)) /
    (alpha ^ 3 * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) *
    (omega * lambda * alpha + omega) - omega * lambda * alpha + (1 - omega) * lambda - omega) *
    ((lambda * alpha + 1) ^ (1 / alpha) - 1)) + ((omega - 1) * z * lambda ^ 2 * 
    ((lambda * alpha + 1) ^ (1 / alpha + 1) * log(lambda * alpha + 1) + 
    (lambda * alpha + 1) ^ (1 / alpha) * (-lambda * alpha ^ 2 - lambda * alpha) + lambda * alpha ^ 2)) /
    (alpha ^ 2 * (lambda * alpha + 1) ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha) *
    (omega * lambda * alpha + omega) - omega * lambda * alpha + 
    (1 - omega) * lambda - omega) * ((lambda * alpha + 1) ^ (1 / alpha) - 1))

    G11 <- G11 * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2
    
    # alpha lambda
    G12 <- ((omega * (lambda * alpha + 1) ^ (1 + 1 / alpha) -
    lambda * (omega * alpha + omega - 1) - omega) * 
    (((lambda * alpha + 1) ^ (1 / alpha) - 1) * 
    ((omega - 1) * z * (lambda - 1) * (lambda * alpha + 1) ^ (1 / alpha) *
    (lambda * alpha - (lambda * alpha + 1) * log(lambda * alpha + 1)) -
    (omega - 1) * z * lambda * alpha ^ 2 * ((lambda - 1) * 
    (lambda * alpha + 1) ^ (1 / alpha) + 1)) - (omega - 1) * z * 
    (lambda * alpha + 1) ^ (1 / alpha) * ((lambda - 1) * (lambda * alpha + 1) ^ (1 / alpha) + 1) *
    (lambda * alpha - (lambda * alpha + 1) * log(lambda * alpha + 1))) -
    (omega - 1) * z * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) - 1) *
    ((lambda - 1) * (lambda * alpha + 1) ^ (1 / alpha) + 1) * 
    (omega * (lambda * alpha + 1) ^ (1 / alpha) * 
    (lambda * alpha * (alpha + 1) - (lambda * alpha + 1) * log(lambda * alpha + 1)) -
    omega * lambda * alpha ^ 2)) / (alpha ^ 2 * (lambda * alpha + 1) ^ 2 * 
    ((lambda * alpha + 1) ^ (1 / alpha) - 1) ^ 2 *
    (omega * (lambda * alpha + 1) ^ (1 + 1 / alpha) -
    lambda * (omega * alpha + omega - 1) - omega) ^ 2) +
    (1 - z) * (-((lambda * alpha + 1) ^ (-2 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1)))) /
    (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha)) ^ 2 - ((lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) / 
    (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha)) + 1 / (alpha * (lambda * alpha + 1))) + 
    (-(1 - z) / alpha - XXX) / (lambda * alpha + 1) - 
    (lambda * (-(1 - z) / alpha - XXX) * alpha) / (lambda * alpha + 1) ^ 2
    
    G12 <- G12 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
                  alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
    
    #lambda
    G22 <- (1 - z) * (-((-1 / alpha - 1) * alpha * (alpha * lambda + 1) ^ (-1 / alpha - 2)) / 
    (1 - 1 / (alpha * lambda + 1) ^ (1 / alpha)) + (alpha * lambda + 1) ^ (-2 / alpha - 2) /
    (1 - 1 / (alpha * lambda + 1) ^ (1 / alpha)) ^ 2) - XXX / lambda ^ 2 - 
    (alpha ^ 2 * (-XXX - (1 - z) / alpha)) / (alpha * lambda + 1) ^ 2 - 
    ((omega - 1) * z * ((alpha * lambda + 1) ^ (1 / alpha) *
    (((alpha ^ 2 - 1) * omega - alpha + 1) * lambda ^ 2 + 
    ((-6 * alpha ^ 2 - 5 * alpha - 1) * omega + 4 * alpha) * lambda +
    (-6 * alpha - 4) * omega + 2) + (alpha * lambda + 1) ^ (2 / alpha) *
    (((-2 * alpha ^ 2 - alpha) * omega + alpha) * lambda ^ 2 +
    ((6 * alpha ^ 2 + 4 * alpha) * omega - 2 * alpha) * lambda + (6 * alpha + 5) * omega - 1) +
    (alpha * lambda + 1) ^ (3 / alpha) * ((alpha ^ 2 + alpha) * omega * lambda ^ 2 +
    (-2 * alpha ^ 2 - alpha + 1) * omega * lambda + (-2 * alpha - 2) * omega) +
    ((2 * alpha ^ 2 + 2 * alpha) * omega - 2 * alpha) * lambda + (2 * alpha + 1) * omega - 1)) /
    ((alpha * lambda + 1) ^ 2 * ((alpha * lambda + 1) ^ (1 / alpha) * (alpha * omega * lambda + omega) +
    ((-alpha - 1) * omega + 1) * lambda - omega) ^ 2 * ((alpha * lambda + 1) ^ (1 / alpha) - 1) ^ 2)
    
    
    G22 <- G22 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2
    
    matrix(
      -c(G22 * prior, G12 * prior, G02 * prior,
         G12 * prior, G11 * prior, G01 * prior,
         G02 * prior, G01 * prior, G00 * prior),
      dimnames = list(
        rownames(eta), 
        c("lambda", "lambda:alpha", "lambda:omega",
          "lambda:alpha", "alpha", "alpha:omega",
          "lambda:omega", "alpha:omega", "omega")
      ),
      ncol = 9
    )
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    omega  <-  omegaLink(eta[, 3], inverse = TRUE)
    z <- as.numeric(y == 1)
    weight <- weight / prior
    
    dig <-  compdigamma(y = y, alpha = alpha)
    
    G0 <- z * ((alpha * lambda + 1) ^ (1 / alpha + 1) + (-alpha - 1) *
    lambda - 1) / (((alpha * lambda + 1) ^ (1 / alpha + 1) +
    (-alpha - 1) * lambda - 1) * omega + lambda) -
    (1 - z) / (1 - omega)
    
    G1 <- ((1 - omega) * z * lambda * ((lambda * alpha + 1) ^ (1 / alpha + 1) *
    log(lambda * alpha + 1) + (lambda * alpha + 1) ^ (1 / alpha) *
    (-lambda * alpha ^ 2 - lambda * alpha) + lambda * alpha ^ 2)) /
    (alpha ^ 2 * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) *
    (omega * lambda * alpha + omega) - omega * lambda * alpha + 
    (1 - omega) * lambda - omega) * ((lambda * alpha + 1) ^ (1 / alpha) - 1)) +
    
      (1 - z) * ((log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
    ((lambda * alpha + 1) ^ (1 / alpha) * (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha))) +
    log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - y)) /
    (lambda * alpha + 1) + y / alpha + dig)
    
    G2 <- (1 - z) * (-(alpha * lambda + 1) ^ (-1 / alpha - 1) / 
    (1 - 1 / (alpha * lambda + 1) ^ (1 / alpha)) + (alpha * (-y - 1 / alpha)) /
    (alpha * lambda + 1) + y / lambda) + 
    ((omega - 1) * z * ((lambda - 1) * (alpha * lambda + 1) ^ (1 / alpha) + 1)) /
    ((alpha * lambda + 1) * ((alpha * lambda + 1) ^ (1 / alpha) * 
    (alpha * omega * lambda + omega) + 
    ((-alpha - 1) * omega + 1) * lambda - omega) * 
    ((alpha * lambda + 1) ^ (1 / alpha) - 1))
    
    G0 <- G0 *  omegaLink(eta[, 3], inverse = TRUE, deriv = 1)
    G1 <- G1 *  alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
    G2 <- G2 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
    
    uMatrix <- cbind(G2, G1, G0)
    
    weight <- lapply(X = 1:nrow(weight), FUN = function (x) {
      matrix(as.numeric(weight[x, ]), ncol = 3)
    })

    pseudoResid <- sapply(X = 1:length(weight), FUN = function (x) {
      xx <- solve(weight[[x]])
      xx %*% uMatrix[x, ]
    })
    
    pseudoResid <- t(pseudoResid)
    dimnames(pseudoResid) <- dimnames(eta)
    pseudoResid
  }
  
  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, 
                           ...) {
    if (is.null(weight)) {
      weight <- 1
    }
    y <- as.numeric(y)
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X) / 3), rep(0, NROW(X) / 3), rep(0, NROW(X) / 3))
    }
    
    z <- as.numeric(y == 1)
    X <- as.matrix(X)
    
    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
            function(beta) {
              eta <- matrix(as.matrix(X) %*% beta, ncol = 3) + offset
              lambda <- lambdaLink(eta[, 1], inverse = TRUE)
              alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
              omega  <-  omegaLink(eta[, 3], inverse = TRUE)
              
              -sum(weight * (z * log(omega + (1 - omega) *
              lambda * (1 + alpha * lambda) ^ (-1 / alpha - 1) / 
              (1 - (1 + lambda * alpha) ^ (-1 / alpha))) + 
              (1 - z) * (log(1 - omega) + lgamma(y + 1 / alpha) - 
              lgamma(1 / alpha) - lgamma(y + 1) - 
              (y + 1 / alpha) * log(1 + lambda * alpha) + y * log(lambda * alpha) - 
              log(1 - (1 + lambda * alpha) ^ (-1 / alpha)))))
            },
            function(beta) {
              eta <- matrix(as.matrix(X) %*% beta, ncol = 3) + offset
              lambda <- lambdaLink(eta[, 1], inverse = TRUE)
              alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
              omega  <-  omegaLink(eta[, 3], inverse = TRUE)
              
              dig <-  compdigamma(y = y, alpha = alpha)
              
              G0 <- z * ((alpha * lambda+1) ^ (1 / alpha + 1) + (-alpha - 1) *
              lambda - 1) / (((alpha * lambda + 1) ^ (1 / alpha + 1) +
              (-alpha - 1) * lambda - 1) * omega + lambda) -
              (1 - z) / (1 - omega)
              
              G1 <- ((1 - omega) * z * lambda * ((lambda * alpha + 1) ^ (1 / alpha + 1) *
              log(lambda * alpha + 1) + (lambda * alpha + 1) ^ (1 / alpha) *
              (-lambda * alpha ^ 2 - lambda * alpha) + lambda * alpha ^ 2)) /
              (alpha ^ 2 * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) *
              (omega * lambda * alpha + omega) - omega * lambda * alpha + 
              (1 - omega) * lambda - omega) * ((lambda * alpha + 1) ^ (1 / alpha) - 1)) +
              (1 - z) * ((log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
              ((lambda * alpha + 1) ^ (1 / alpha) * (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha))) +
              log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - y)) /
              (lambda * alpha + 1) + y / alpha + dig)
              
              G2 <- (1 - z) * (-(alpha * lambda + 1) ^ (-1 / alpha - 1) / 
              (1 - 1 / (alpha * lambda + 1) ^ (1 / alpha)) + (alpha * (-y - 1 / alpha)) /
              (alpha * lambda + 1) + y / lambda) + 
              ((omega - 1) * z * ((lambda - 1) * (alpha * lambda + 1) ^ (1 / alpha) + 1)) /
              ((alpha * lambda + 1) * ((alpha * lambda + 1) ^ (1 / alpha) * 
              (alpha * omega * lambda + omega) + 
              ((-alpha - 1) * omega + 1) * lambda - omega) * 
              ((alpha * lambda + 1) ^ (1 / alpha) - 1))
              
              
              G0 <- G0 * weight *  omegaLink(eta[, 3], inverse = TRUE, deriv = 1)
              G1 <- G1 * weight *  alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
              G2 <- G2 * weight * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
              
              if (NbyK) {
                return(cbind(
                  G2 * as.data.frame(
                    X[1:nrow(eta), 
                      1:(attr(X, "hwm")[1])]
                  ), 
                  G1 * as.data.frame(
                    X[(nrow(eta) + 1):(2 * nrow(eta)), 
                      (attr(X, "hwm")[1] + 1):(attr(X, "hwm")[2])]
                  ), 
                  G0 * as.data.frame(
                    X[(2 * nrow(eta) + 1):(3 * nrow(eta)), 
                      (attr(X, "hwm")[2] + 1):(attr(X, "hwm")[3])]
                  )
                ))
              }
              if (vectorDer) {
                return(cbind(G2, G1, G0))
              }
              
              as.numeric(c(G2, G1, G0) %*% X)
            },
            function(beta) {
              predNumbers <- attr(X, "hwm")
              eta <- matrix(as.matrix(X) %*% beta, ncol = 3) + offset
              lambda <- lambdaLink(eta[, 1], inverse = TRUE)
              alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
              omega  <-  omegaLink(eta[, 3], inverse = TRUE)
              res <- matrix(nrow = length(beta), 
                            ncol = length(beta), 
                            dimnames = list(names(beta), names(beta)))
              
              trig <- comptrigamma(y = y, alpha = alpha)
              dig <-  compdigamma(y = y, alpha = alpha)
              
              # omega
              G0 <- z * ((alpha * lambda+1) ^ (1 / alpha + 1) + (-alpha - 1) *
              lambda - 1) / (((alpha * lambda + 1) ^ (1 / alpha + 1) +
              (-alpha - 1) * lambda - 1) * omega + lambda) -
              (1 - z) / (1 - omega)
              
              G00 <- (-z * ((alpha * lambda + 1) ^ (1 / alpha + 1) +
              (-alpha - 1) * lambda - 1) ^ 2) / (((alpha * lambda + 1) ^ (1 / alpha + 1) +
              (-alpha - 1) * lambda - 1) * omega + lambda) ^ 2 -
              (1 - z) / (1 - omega) ^ 2
              
              G00 <- G00 * omegaLink(eta[, 3], inverse = TRUE, deriv = 1) ^ 2 +
                      G0 * omegaLink(eta[, 3], inverse = TRUE, deriv = 2)
              
              # omega alpha
              G01 <- (-z) * lambda * ((alpha * lambda + 1) ^ (1 / alpha + 1) *
              log(alpha * lambda + 1) + (-alpha ^ 2 - alpha) * lambda * 
              (alpha * lambda + 1) ^ (1 / alpha) + alpha ^ 2 * lambda) /
              (alpha ^ 2 * (((alpha * lambda + 1) ^ (1 / alpha + 1) +
              (-alpha - 1) * lambda - 1) * omega + lambda) ^ 2)
              
              G01 <- G01 * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) *
                           omegaLink(eta[, 3], inverse = TRUE, deriv = 1)
              
              # omega lambda
              G02 <- z * ((lambda - 1) * (alpha * lambda + 1) ^ (1 / alpha) + 1) /
              (((alpha * lambda + 1) ^ (1 / alpha + 1) + (-alpha - 1) * lambda - 1) * omega + lambda) ^ 2
              
              G02 <- G02 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
                            omegaLink(eta[, 3], inverse = TRUE, deriv = 1)
              # alpha
              G1 <- ((1 - omega) * z * lambda * ((lambda * alpha + 1) ^ (1 / alpha + 1) *
              log(lambda * alpha + 1) + (lambda * alpha + 1) ^ (1 / alpha) *
              (-lambda * alpha ^ 2 - lambda * alpha) + lambda * alpha ^ 2)) /
              (alpha ^ 2 * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) *
              (omega * lambda * alpha + omega) - omega * lambda * alpha + 
              (1 - omega) * lambda - omega) * ((lambda * alpha + 1) ^ (1 / alpha) - 1)) +
              (1 - z) * ((log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
              ((lambda * alpha + 1) ^ (1 / alpha) * (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha))) +
              log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - y)) /
              (lambda * alpha + 1) + y / alpha + dig)
              
              G11 <- (1 - z) * ((log(lambda * alpha + 1) / alpha ^ 2 - lambda / 
              (alpha * (lambda * alpha + 1))) ^ 2 / ((lambda * alpha + 1) ^ (1 / alpha) *
              (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha))) + (log(lambda * alpha + 1) / alpha ^ 2 -
              lambda / (alpha * (lambda * alpha + 1))) ^ 2 / ((lambda * alpha + 1) ^ (2 / alpha) *
              (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha)) ^ 2) + (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 +
              (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) + lambda ^ 2 / 
              (alpha * (lambda * alpha + 1) ^ 2)) / ((lambda * alpha + 1) ^ (1 / alpha) * 
              (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha))) - (2 * log(lambda * alpha + 1)) / alpha ^ 3 + 
              (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) - (lambda ^ 2 * (-1 / alpha - y)) / 
              (lambda * alpha + 1) ^ 2 - y / alpha ^ 2 + trig) - 
              ((omega - 1) * z * lambda * ((lambda * alpha + 1) ^ (1 / alpha + 1) *
              log(lambda * alpha + 1) * ((lambda * (1 / alpha + 1)) / (lambda * alpha + 1) -
              log(lambda * alpha + 1) / alpha ^ 2) + (lambda * alpha + 1) ^ (1 / alpha) * 
              (-lambda * alpha ^ 2 - lambda * alpha) * (lambda / (alpha * (lambda * alpha + 1)) - 
              log(lambda * alpha + 1) / alpha ^ 2) + (-2 * lambda * alpha - lambda) * 
              (lambda * alpha + 1) ^ (1 / alpha) + lambda * (lambda * alpha + 1) ^ (1 / alpha) +
              2 * lambda * alpha)) / (alpha ^ 2 * (lambda * alpha + 1) * 
              ((lambda * alpha + 1) ^ (1 / alpha) * (omega * lambda * alpha + omega) - 
              omega * lambda * alpha + (1 - omega) * lambda - omega) * ((lambda * alpha + 1) ^ (1 / alpha) - 1)) +
              ((omega - 1) * z * lambda * ((lambda * alpha + 1) ^ (1 / alpha + 1) * log(lambda * alpha + 1) +
              (lambda * alpha + 1) ^ (1 / alpha) * (-lambda * alpha ^ 2 - lambda * alpha) + lambda * alpha ^ 2) *
              ((lambda * alpha + 1) ^ (1 / alpha) * (omega * lambda * alpha + omega) *
              (lambda / (alpha * (lambda * alpha + 1)) - log(lambda * alpha + 1) / alpha ^ 2) + 
              omega * lambda * (lambda * alpha + 1) ^ (1 / alpha) - omega * lambda)) / 
              (alpha ^ 2 * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) *
              (omega * lambda * alpha + omega) - omega * lambda * alpha + (1 - omega) * lambda - omega) ^ 2 *
              ((lambda * alpha + 1) ^ (1 / alpha) - 1)) + ((omega - 1) * z * lambda * 
              (lambda * alpha + 1) ^ (1 / alpha - 1) * (lambda / (alpha * (lambda * alpha + 1)) -
              log(lambda * alpha + 1) / alpha ^ 2) * ((lambda * alpha + 1) ^ (1 / alpha + 1) * 
              log(lambda * alpha + 1) + (lambda * alpha+ 1 ) ^ (1 / alpha) *
              (-lambda * alpha ^ 2 - lambda * alpha) + lambda * alpha ^ 2)) /
              (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha) * (omega * lambda * alpha + omega) -
              omega * lambda * alpha + (1 - omega) * lambda - omega) * ((lambda * alpha + 1) ^ (1 / alpha) - 1) ^ 2) +
              (2 * (omega - 1) * z * lambda * ((lambda * alpha + 1) ^ (1 / alpha + 1) * 
              log(lambda * alpha + 1) + (lambda * alpha + 1) ^ (1 / alpha) * 
              (-lambda * alpha ^ 2 - lambda * alpha) + lambda * alpha ^ 2)) /
              (alpha ^ 3 * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) *
              (omega * lambda * alpha + omega) - omega * lambda * alpha + (1 - omega) * lambda - omega) *
              ((lambda * alpha + 1) ^ (1 / alpha) - 1)) + ((omega - 1) * z * lambda ^ 2 * 
              ((lambda * alpha + 1) ^ (1 / alpha + 1) * log(lambda * alpha + 1) + 
              (lambda * alpha + 1) ^ (1 / alpha) * (-lambda * alpha ^ 2 - lambda * alpha) + lambda * alpha ^ 2)) /
              (alpha ^ 2 * (lambda * alpha + 1) ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha) *
              (omega * lambda * alpha + omega) - omega * lambda * alpha + 
              (1 - omega) * lambda - omega) * ((lambda * alpha + 1) ^ (1 / alpha) - 1))
              
              G11 <- G11 * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2 +
                      G1 * alphaLink(eta[, 2], inverse = TRUE, deriv = 2)
              
              # alpha lambda
              G12 <- ((omega * (lambda * alpha + 1) ^ (1 + 1 / alpha) -
              lambda * (omega * alpha + omega - 1) - omega) * 
              (((lambda * alpha + 1) ^ (1 / alpha) - 1) * 
              ((omega - 1) * z * (lambda - 1) * (lambda * alpha + 1) ^ (1 / alpha) *
              (lambda * alpha - (lambda * alpha + 1) * log(lambda * alpha + 1)) -
              (omega - 1) * z * lambda * alpha ^ 2 * ((lambda - 1) * 
              (lambda * alpha + 1) ^ (1 / alpha) + 1)) - (omega - 1) * z * 
              (lambda * alpha + 1) ^ (1 / alpha) * ((lambda - 1) * (lambda * alpha + 1) ^ (1 / alpha) + 1) *
              (lambda * alpha - (lambda * alpha + 1) * log(lambda * alpha + 1))) -
              (omega - 1) * z * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) - 1) *
              ((lambda - 1) * (lambda * alpha + 1) ^ (1 / alpha) + 1) * 
              (omega * (lambda * alpha + 1) ^ (1 / alpha) * 
              (lambda * alpha * (alpha + 1) - (lambda * alpha + 1) * log(lambda * alpha + 1)) -
              omega * lambda * alpha ^ 2)) / (alpha ^ 2 * (lambda * alpha + 1) ^ 2 * 
              ((lambda * alpha + 1) ^ (1 / alpha) - 1) ^ 2 *
              (omega * (lambda * alpha + 1) ^ (1 + 1 / alpha) -
              lambda * (omega * alpha + omega - 1) - omega) ^ 2) +
              (1 - z) * (-((lambda * alpha + 1) ^ (-2 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1)))) /
              (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha)) ^ 2 - ((lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) / 
              (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha)) + 1 / (alpha * (lambda * alpha + 1)) + 
              (-1 / alpha - y) / (lambda * alpha + 1) - 
              (lambda * (-1 / alpha - y) * alpha) / (lambda * alpha + 1) ^ 2)
              
              G12 <- G12 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
                            alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
              
              #lambda
              G2 <- (1 - z) * (-(alpha * lambda + 1) ^ (-1 / alpha - 1) / 
              (1 - 1 / (alpha * lambda + 1) ^ (1 / alpha)) + (alpha * (-y - 1 / alpha)) /
              (alpha * lambda + 1) + y / lambda) + 
              ((omega - 1) * z * ((lambda - 1) * (alpha * lambda + 1) ^ (1 / alpha) + 1)) /
              ((alpha * lambda + 1) * ((alpha * lambda + 1) ^ (1 / alpha) * 
              (alpha * omega * lambda + omega) + 
              ((-alpha - 1) * omega + 1) * lambda - omega) * 
              ((alpha * lambda + 1) ^ (1 / alpha) - 1))
              
              G22 <- (1 - z) * (-((-1 / alpha - 1) * alpha * (alpha * lambda + 1) ^ (-1 / alpha - 2)) / 
              (1 - 1 / (alpha * lambda + 1) ^ (1 / alpha)) + (alpha * lambda + 1) ^ (-2 / alpha - 2) /
              (1 - 1 / (alpha * lambda + 1) ^ (1 / alpha)) ^ 2 - (alpha ^ 2 * (-y - 1 / alpha)) /
              (alpha * lambda + 1) ^ 2 - y / lambda ^ 2) - ((omega - 1) * z * 
              ((alpha * lambda + 1) ^ (1 / alpha) * (((alpha ^ 2 - 1) * omega - alpha + 1) * lambda ^ 2 +
              ((-6 * alpha ^ 2 - 5 * alpha - 1) * omega + 4 * alpha) * lambda +
              (-6 * alpha - 4) * omega + 2) + (alpha * lambda + 1) ^ (2 / alpha) *
              (((-2 * alpha ^ 2 - alpha) * omega + alpha) * lambda ^ 2 + 
              ((6 * alpha ^ 2 + 4 * alpha) * omega - 2 * alpha) * lambda +
              (6 * alpha + 5) * omega - 1) + (alpha * lambda + 1) ^ (3 / alpha) *
              ((alpha ^ 2 + alpha) * omega * lambda ^ 2 + 
              (-2 * alpha ^ 2 - alpha + 1) * omega * lambda + (-2 * alpha - 2) * omega) +
              ((2 * alpha ^ 2 + 2 * alpha) * omega - 2 * alpha) * lambda +
              (2 * alpha + 1) * omega - 1)) / ((alpha * lambda + 1) ^ 2 *
              ((alpha * lambda + 1) ^ (1 / alpha) * (alpha * omega * lambda + omega) +
              ((-alpha - 1) * omega + 1) * lambda - omega) ^ 2 * 
              ((alpha * lambda + 1) ^ (1 / alpha) - 1) ^ 2)
               
              G22 <- G22 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 + 
                      G2 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 2)
              
              
              predNumbers <- cumsum(predNumbers)
              res[(predNumbers[2]+1):predNumbers[3], (predNumbers[2]+1):predNumbers[3]] <- #omega
                t(as.data.frame(X[(nrow(X) * 2 / 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]]) *
                  G00 * weight) %*% X[(nrow(X) * 2 / 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]]
              
              res[(predNumbers[1]+1):predNumbers[2], (predNumbers[2]+1):predNumbers[3]] <- #omega alpha
                t(as.data.frame(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1] + 1):predNumbers[2]]) * 
                  G01 * weight) %*% as.matrix(X[(nrow(X) * 2/ 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]])
              
              res[(predNumbers[2]+1):predNumbers[3], (predNumbers[1]+1):predNumbers[2]] <- #omega alpha
                t(t(as.data.frame(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1] + 1):predNumbers[2]]) * 
                    G01 * weight) %*% as.matrix(X[(nrow(X) * 2/ 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]]))
              
              res[1:predNumbers[1], (predNumbers[2]+1):predNumbers[3]] <- #omega lambda
                t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                  G02 * weight) %*% as.matrix(X[(nrow(X) * 2 / 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]])
              
              res[(predNumbers[2]+1):predNumbers[3], 1:predNumbers[1]] <- #omega lambda
                t(t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                    G02 * weight) %*% as.matrix(X[(nrow(X) * 2 / 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]]))
              
              res[(predNumbers[1]+1):predNumbers[2], (predNumbers[1]+1):predNumbers[2]] <- #alpha
                t(as.data.frame(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]]) *
                  G11 * weight) %*% X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]]
              
              res[1:predNumbers[1], 1:predNumbers[1]] <-  # lambda
                t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                  G22 * weight) %*% X[1:(nrow(X) / 3), 1:predNumbers[1]]
              
              res[1:predNumbers[1], (predNumbers[1]+1):predNumbers[2]] <- #alpha lambda
                t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                  G12 * weight) %*% as.matrix(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]])
              
              res[(predNumbers[1]+1):predNumbers[2], 1:predNumbers[1]] <- #alpha lambda
                t(t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                    G12 * weight) %*% as.matrix(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]]))
              
              res
            }
    )
  }
  
  validmu <- function(mu) {
    all(is.finite(mu)) && all(0 < mu)
  }
  
  devResids <- function (y, eta, wt, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    omega  <-  omegaLink(eta[, 3], inverse = TRUE)
    z <- (y == 1)
    mu <- mu.eta(eta = eta)
    
    prob_ge1 <- pmax(1 - (1 + lambda * alpha) ^ (-1 / alpha), .Machine$double.xmin)
    logLikFit <- (
      z * log(omega + (1 - omega) * lambda * (1 + alpha * lambda) ^ (-1 / alpha - 1) /
      prob_ge1) + (1 - z) * (log(1 - omega) + lgamma(y + 1 / alpha) -
      lgamma(1 / alpha) - lgamma(y + 1) - (y + 1 / alpha) * log(1 + lambda * alpha) + y * log(lambda * alpha) -
      log(prob_ge1))
    )
    
    yUnq <- unique(y)
    
    if (any(yUnq > 77)) {
      warning("Currently numerical deviance is unreliable for counts greater than 78.")
    }
    
    findL <- function(t) {
      yNow <- yUnq[t]
      stats::optim(
        #par = if(yNow < 26) c(0, .6, 0) else c(-.5, log(yNow), -20),
        par = c(0, log(yNow), -10),
        fn = function(x) {
          s <- x[1]
          l <- exp(x[2])
          a <- exp(x[3])
          
          prob <- 1 - (1+a*l)^(-1/a)
          prob <- 1 / prob
          sum(c((l*prob - yNow) * 4.5,# s der
                yNow/l+(-yNow*a-1)/(1+a*l)-(1+a*l)^(-1-1/a)*prob+s*(prob-prob^2*(l*(1+a*l)^(-1-1/a))),# lambda der
                (log(l*a+1)/a^2-l/(a*(l*a+1)))/((l*a+1)^(1/a)*(1-1/(l*a+1)^(1/a)))+(s*l*(log(l*a+1)/a^2-l/(a*(l*a+1))))/((l*a+1)^(1/a)*(1-1/(l*a+1)^(1/a))^2)+log(l*a+1)/a^2+(l*(-1/a-yNow))/(l*a+1)+yNow/a-digamma(yNow+1/a)/a^2+digamma(1/a)/a^2,#alpha der
                #this is experimental
                lgamma(yNow+1/a)-lgamma(1/a) - lgamma(yNow+1)-(yNow+1/a)*log(1+a*l)+yNow*log(l*a)-log(1-(1+a*l)^(-1/a))) ^ 2) ^ .5
        },
        method = "BFGS",
        control = list(maxit = 10000, abstol = .Machine$double.eps, reltol = .Machine$double.eps)
      )$par
    }
    
    suppressWarnings({
      logLikIdeal <- sapply(1:length(yUnq), FUN = function(x) {
        ifelse(yUnq[x] == 1, 0, {
          xx <- findL(x)
          lagrange <- xx[1]
          l <- exp(xx[2])
          a <- exp(xx[3])
          (lgamma(yUnq[x] + 1 / a) - lgamma(1 / a) -
              lgamma(yUnq[x] + 1) - (yUnq[x] + 1 / a) * log(1 + a * l) +
              yUnq[x] * log(l * a) - log(1 - (1 + a * l) ^ (-1 / a)))
        })
      })
    })
    
    logLikIdeal <- sapply(1:length(y), FUN = function(x) {
      logLikIdeal[yUnq == y[x]]
    })
    
    diff <- logLikIdeal - logLikFit
    
    if (any(logLikFit > 0)) {
      warning("Dispersion parameter values are on the boundary of parameter space. Deviance residuals will be assigned 0 on these observations.")
      diff[logLikFit > 0]   <- 0
    } else if (any(diff < 0)) {
      warning("Numerical deviance finder found worse saturated likelihood than fitted model. Expect NA's in deviance/deviance residuals.")
      diff[diff < 0] <- 0
    }
    
    sign(y - mu) * sqrt(2 * wt * diff)
  }
  
  pointEst <- function (pw, eta, contr = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    N <- pw / (1 - (1 + alpha * lambda) ^ (- 1 / alpha))
    if(!contr) {
      N <- sum(N)
    }
    N
  }
  
  popVar <- function (pw, eta, cov, Xvlm, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    pr <- 1 - (1 + alpha * lambda) ^ (- 1 / alpha)
    
    bigTheta0 <- pw * 0 # w.r to omega
    bigTheta1 <- pw  *  alphaLink(eta[, 2], inverse = TRUE, deriv = 1) *
    ((lambda * alpha + 1) ^ (1 / alpha - 1) * 
    ((lambda * alpha + 1) * log(lambda * alpha + 1) - lambda * alpha)) /
    (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha) - 1) ^ 2)# w.r to alpha
    bigTheta2 <- -pw * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * 
    ((alpha * lambda + 1) ^ (1 / alpha - 1) / ((alpha * lambda + 1) ^ (1 / alpha) - 1) ^ 2)# w.r to lambda
    
    bigTheta <- t(c(bigTheta2, bigTheta1, bigTheta0) %*% Xvlm)
    
    f1 <-  t(bigTheta) %*% as.matrix(cov) %*% bigTheta
    f2 <-  sum(pw * (1 - pr) / (pr ^ 2))
    
    f1 + f2
  }
  
  dFun <- function (x, eta, type = c("trunc", "nontrunc")) {
    if (missing(type)) type <- "trunc"
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    omega  <-  omegaLink(eta[, 3], inverse = TRUE)
    P0 <- (1 + alpha * lambda) ^ (-1 / alpha)
    
    switch (type,
      "trunc" = {
        (1 - omega) * stats::dnbinom(x = x, mu = lambda, size = 1 / alpha) / 
        (1 - P0) + omega * as.numeric(x == 1)
      },
      "nontrunc" = {
        stats::dnbinom(x = x, mu = lambda, size = 1 / alpha) * 
        (as.numeric(x == 0) + as.numeric(x > 0) * (1 - omega)) +
        omega * (1 - P0) * as.numeric(x == 1)
      }
    )
  }
  
  simulate <- function(n, eta, lower = 0, upper = Inf) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    omega  <-  omegaLink(eta[, 3], inverse = TRUE)
    CDF <- function(x) {
      ifelse(x == Inf, 1, 
      ifelse(x < 0, 0, 
      ifelse(x < 1, (1 + alpha * lambda) ^ (- 1 / alpha), 
      (1 + alpha * lambda) ^ (- 1 / alpha) + 
      omega * (1 - (1 + alpha * lambda) ^ (- 1 / alpha)) + 
      (1 - omega) * (stats::pnbinom(q = x, mu = lambda, size = 1 / alpha) -
                     (1 + alpha * lambda) ^ (- 1 / alpha)))))
    }
    lb <- CDF(lower)
    ub <- CDF(upper)
    p_u <- stats::runif(n, lb, ub)
    sims <- rep(0, n)
    cond <- CDF(sims) < p_u
    while (any(cond)) {
      sims[cond] <- sims[cond] + 1
      cond <- CDF(sims) < p_u
    }
    sims
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      init <- log(abs((observed / weighted.mean(observed, priorWeights) - 1) / observed) + .1)
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12)),
        family$links[[2]](ifelse(init < -.5, .1, init + .55)),
        family$links[[3]](weighted.mean(observed == 1, priorWeights) * (.5 + .5 * (observed == 1)) + .01)
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights)),
        family$links[[2]](abs((cov.wt(cbind(observed, observed), wt = priorWeights, method = "ML")$cov[1,1] / weighted.mean(observed, priorWeights) - 1) / weighted.mean(observed, priorWeights)) + .1),
        family$links[[3]](weighted.mean(observed == 1, priorWeights) + .01)
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
      if ("(Intercept):alpha" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[2], rep(0, attr(Xvlm, "hwm")[2] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[2] / attr(Xvlm, "hwm")[2], attr(Xvlm, "hwm")[2]))
      }
      if ("(Intercept):omega" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[3], rep(0, attr(Xvlm, "hwm")[3] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[3] / attr(Xvlm, "hwm")[3], attr(Xvlm, "hwm")[3]))
      }
    }
  )

  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "oiztnegbin",
      etaNames  = c("lambda", "alpha", "omega"),
      simulate  = simulate,
      getStart  = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

#' @rdname singleRmodels
#' @author Cyprian Jurkowski, Piotr Chlebicki, Maciej Beręsewicz
#' @importFrom stats dbinom
#' @export
oichao <- function(lambdaLink = "logthird",
                   ...) {
  if (missing(lambdaLink)) lambdaLink <- "logthird"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink)
  
  lambdaLink <- switch(
    lambdaLink,
    "logthird" = singleRinternallogthirdLink,
    stop("Currently only lambdaLink = 'logthird' is supported.")
  )
  
  links[1] <- c(lambdaLink)
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    lambda <- lambdaLink(eta, inverse = TRUE)
    
    if (!deriv) {
      switch(
        type,
        "nontrunc" = lambda,
        "trunc" = (lambda / 3) / (1 + lambda / 3)
      )
    } else {
      switch(
        type,
        "nontrunc" = cbind(lambdaLink(eta, inverse = TRUE, deriv = 1)),
        "trunc" = cbind(
          (3 / (lambda + 3) ^ 2) *
            lambdaLink(eta, inverse = TRUE, deriv = 1)
        )
      )
    }
  }
  
  variance <- function(eta, type = "nontrunc", ...) {
    lambda <- lambdaLink(eta, inverse = TRUE)
    switch(
      type,
      "nontrunc" = lambda,
      "trunc" = ((lambda / 3) / (1 + lambda / 3)) * (1 / (1 + lambda / 3))
    )
  }
  
  Wfun <- function(prior, eta, y, ...) {
    iddx <- y %in% 2:3
    lambda <- lambdaLink(eta, inverse = TRUE)[iddx]
    res <- iddx
    res[iddx] <- -prior[iddx] * (
      -3 / (lambda * (3 + lambda) ^ 2) *
        lambdaLink(eta, inverse = TRUE, deriv = 1)[iddx] ^ 2
    )
    
    matrix(
      res,
      ncol = 1,
      dimnames = list(rownames(eta), c("lambda"))
    )
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    iddx <- y %in% 2:3
    z <- y - 2
    lambda <- lambdaLink(eta, inverse = TRUE)[iddx]
    
    res <- iddx
    res[iddx] <- prior[iddx] * (((z[iddx] - 1) * lambda + 3 * z[iddx]) /
      (lambda * (3 + lambda))) *
      lambdaLink(eta, inverse = TRUE, deriv = 1)[iddx] /
      weight[iddx, ]
    res
  }
  
  minusLogLike <- function(y, X,
                           weight = 1,
                           NbyK = FALSE,
                           vectorDer = FALSE,
                           deriv = 0,
                           offset,
                           eta,
                           ...) {
    y <- as.numeric(y)
    z <- y - 2
    if (is.null(weight)) {
      weight <- 1
    }
    
    iddx <- y %in% 2:3
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X)))
    }
    
    if (!(deriv %in% c(0, 1, 2))) {
      stop("Only score function and derivatives up to 2 are supported.")
    }
    
    deriv <- deriv + 1
    
    switch(
      deriv,
      function(beta, eta) {
        if (missing(eta)) {
          eta <- as.matrix(X) %*% beta + offset
        }
        lambda <- lambdaLink(eta, inverse = TRUE)[iddx]
        -sum(weight[iddx] * dbinom(
          size = 1,
          prob = (lambda / 3) / (1 + lambda / 3),
          log = TRUE,
          x = z[iddx]
        ))
      },
      function(beta) {
        eta <- as.matrix(X) %*% beta + offset
        lambda <- lambdaLink(eta, inverse = TRUE)[iddx]
        G0 <- iddx
        G0[iddx] <- ((z[iddx] - 1) * lambda + 3 * z[iddx]) /
          (lambda * (3 + lambda)) * weight[iddx] *
          lambdaLink(eta, inverse = TRUE, deriv = 1)[iddx]
        
        if (NbyK) {
          return(as.data.frame(X[iddx, , drop = FALSE]) * G0)
        }
        if (vectorDer) {
          return(matrix(G0, ncol = 1))
        }
        
        t(X) %*% G0
      },
      function(beta) {
        eta <- as.matrix(X) %*% beta + offset
        lambda <- lambdaLink(eta, inverse = TRUE)[iddx]
        
        G00 <- iddx
        G00[iddx] <- (
          -(9 * z[iddx] + 6 * z[iddx] * lambda +
              (z[iddx] - 1) * lambda ^ 2) /
            ((3 + lambda) ^ 2 * lambda ^ 2)
        ) * lambdaLink(eta, inverse = TRUE, deriv = 1)[iddx] ^ 2 +
          ((z[iddx] - 1) * lambda + 3 * z[iddx]) /
          (lambda * (3 + lambda)) *
          lambdaLink(eta, inverse = TRUE, deriv = 2)[iddx]
        
        t(as.data.frame(X) * weight * G00) %*% as.matrix(X)
      }
    )
  }
  
  validmu <- function(mu) {
    (sum(!is.finite(mu)) == 0) && all(1 > mu)
  }
  
  devResids <- function(y, eta, wt, ...) {
    z <- y - 2
    lambda <- lambdaLink(eta, inverse = TRUE)
    diff <- z * log(lambda / 3 / (1 + lambda / 3)) +
      (1 - z) * log(1 / (1 + lambda / 3))
    diff[diff > 0] <- -0
    
    ((-1) ^ (y - 1)) * sqrt((-2 * wt * diff) * (y %in% 2:3))
  }
  
  pointEst <- function(pw, eta, contr = FALSE, y, ...) {
    lambda <- lambdaLink(eta, inverse = TRUE)
    N <- ((1 + (y %in% 2:3) / ((lambda ^ 2) / 2 + (lambda ^ 3) / 6)) * pw)
    if (!contr) {
      N <- sum(N)
    }
    N
  }
  
  popVar <- function(pw, eta, cov, Xvlm, y, ...) {
    iddx <- y %in% 2:3
    lambda <- lambdaLink(eta, inverse = TRUE)
    Xvlm <- as.matrix(Xvlm)
    prob <- (lambda ^ 2) * exp(-lambda) / 2 + (lambda ^ 3) * exp(-lambda) / 6
    
    f1 <- t((-lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * pw *
      as.numeric((lambda + lambda ^ 2 / 2) /
        (((lambda ^ 2) / 2 + (lambda ^ 3) / 6) ^ 2)))[iddx] %*%
      Xvlm[iddx, , drop = FALSE])
    
    f1 <- t(f1) %*% as.matrix(cov) %*% f1
    
    f2 <- sum((pw * (1 - prob) * ((1 + exp(-lambda) / prob) ^ 2))[iddx])
    
    f1 + f2
  }
  
  simulate <- function(n, eta, lower = 1, upper = 3) {
    lambda <- as.numeric(lambdaLink(eta, inverse = TRUE))
    lambda <- rep_len(lambda, n)

    if (is.infinite(upper)) {
      return(stats::rpois(n = n, lambda = lambda))
    }

    lower <- as.integer(floor(lower)) + 1L
    upper <- as.integer(floor(upper))
    lower <- max(lower, 0L)

    if (lower > upper) {
      stop("Simulation support must contain at least one integer value.")
    }

    support <- seq.int(lower, upper)
    probs <- vapply(
      X = support,
      FUN = function(x) stats::dpois(x = x, lambda = lambda),
      FUN.VALUE = numeric(length(lambda))
    )
    probs <- probs / rowSums(probs)

    if (length(support) == 1L) {
      return(rep.int(support, n))
    }

    as.numeric(apply(
      X = probs,
      MARGIN = 1L,
      FUN = function(prob) sample(x = support, size = 1L, prob = prob)
    ))
  }
  
  dFun <- function(x, eta, type = "trunc") {
    if (missing(type)) type <- "trunc"
    lambda <- lambdaLink(eta, inverse = TRUE)
    
    switch(
      type,
      "trunc" = {
        stats::dpois(x = x, lambda = lambda) /
          (1 - stats::dpois(x = 0, lambda = lambda))
      },
      "nontrunc" = stats::dpois(x = x, lambda = lambda)
    )
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      etaStart <- cbind(
        (priorWeights * (observed == 3) + .5) / (priorWeights + 1)
      ) + offset
      etaStart <- family$links[[1]](3 * (etaStart / (1 - etaStart)))
    } else if (method == "optim") {
      init <- mean(
        ((priorWeights * (observed == 3) + .5) / (priorWeights + 1))[observed %in% 2:3]
      )
      coefStart <- rep(family$links[[1]](3 * init / (1 - init)), NCOL(variables))
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links            = links,
      mu.eta           = mu.eta,
      valideta         = function(eta) {TRUE},
      variance         = variance,
      Wfun             = Wfun,
      funcZ            = funcZ,
      devResids        = devResids,
      validmu          = validmu,
      pointEst         = pointEst,
      popVar           = popVar,
      family           = "oichao",
      etaNames         = c("lambda"),
      simulate         = simulate,
      simulateLower    = 1,
      simulateUpper    = 3,
      getStart         = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

# working methods ---------------------------------------------------------

eps <- if (capabilities("long.double")) sqrt(.Machine$double.eps) else 0.01

# test simulate
set.seed(123)
expect_equivalent(
  {N <- 10000
   gender <- rbinom(N, 1, 0.2)
   eta <- -1 + 0.5*gender
   counts <- rnbinom(N, mu = exp(eta), size = 1)
   df <- data.frame(gender, eta, counts)
   df2 <- subset(df, counts > 0)
   mod1 <-  estimatePopsize(
     formula = counts ~ 1 + gender,
     data = df2,
     model = "ztnegbin",
     method = "optim"
   )
   mid1_sim <- simulate(mod1, 10)
   dim(mid1_sim)
  },
  c(2920, 10),
  tolerance = eps
)

expect_silent(
  Model <- estimatePopsize(
    formula = capture ~ nation + age + gender, 
    data = netherlandsimmigrant, 
    model = ztpoisson, 
    method = "IRLS",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_silent(
  Model1 <- estimatePopsize(
    formula = capture ~ 1, 
    data = netherlandsimmigrant, 
    model = ztpoisson, 
    method = "IRLS",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_silent(
  Model2 <- estimatePopsize(
    formula = capture ~ . - age - reason, 
    data = netherlandsimmigrant, 
    model = zelterman, 
    method = "IRLS",
    controlMethod = controlMethod(silent = TRUE)
  )
)

df <- netherlandsimmigrant[, c(1:3,5)]
df$ww <- 0
### this is dplyr::count but slower and without dependencies
df <- aggregate(ww ~ ., df, FUN = length)

expect_silent(
  Model6 <- estimatePopsize(
    formula = capture ~ nation + age + gender, 
    data = df, 
    model = ztpoisson, 
    method = "IRLS",
    weights = df$ww,
    controlMethod = controlMethod(silent = TRUE),
    controlModel = controlModel(weightsAsCounts = TRUE)
  )
)

expect_equal(
  nobs(Model6),
  nobs(Model),
  tolerance = eps
)

expect_equal(
  Model$populationSize$pointEstimate,
  Model6$populationSize$pointEstimate,
  tolerance = eps
)

expect_equal(
  Model$populationSize$confidenceInterval,
  Model6$populationSize$confidenceInterval,
  tolerance = eps
)

expect_equal(
  Model$populationSize$variance,
  Model6$populationSize$variance,
  tolerance = eps
)

expect_equal(
  Model$coefficients,
  Model6$coefficients,
  tolerance = eps
)

expect_equal(
  Model$logL,
  Model6$logL,
  tolerance = eps
)

# dfbetas and dfpopsize
# 4 takes too long
expect_silent(
  dfb <- dfbeta(Model)
)

expect_silent(
  dfb1 <- dfbeta(Model1)
)

expect_silent(
  dfb2 <- dfbeta(Model2)
)

expect_silent(
  dfb6 <- dfbeta(Model6)
)

expect_silent(
  hatvalues(Model)
)

expect_silent(
  hatvalues(Model2)
)

expect_silent(
  dfp <- dfpopsize(Model, dfbeta = dfb)
)

expect_silent(
  dfp6 <- dfpopsize(Model6, dfbeta = dfb6)
)

expect_equal(
  max(abs(dfp)),
  4236.412,
  tolerance = .05
)

expect_equal(
  abs(mean(abs(dfp))),
  19.19,
  tolerance = .05
)

expect_true(
  abs(max(abs(dfpopsize(Model1, dfbeta = dfb1))) - 88.349) < .2
)

expect_true(
  abs(mean(abs(dfpopsize(Model1, dfbeta = dfb1))) - 8.1945) < .1
)

expect_true(
  abs(max(abs(dfpopsize(Model2, dfbeta = dfb2))) - 3648.17) < .2
)

expect_equal(
  c(unique(dfp[netherlandsimmigrant$capture == 1 & netherlandsimmigrant$gender == "female" & netherlandsimmigrant$nation == "American and Australia" & netherlandsimmigrant$age == "<40yrs"])[1],
    unique(dfp[netherlandsimmigrant$capture == 2 & netherlandsimmigrant$gender == "female" & netherlandsimmigrant$nation == "American and Australia" & netherlandsimmigrant$age == "<40yrs"])[1],
    unique(dfp[netherlandsimmigrant$capture == 3 & netherlandsimmigrant$gender == "female" & netherlandsimmigrant$nation == "American and Australia" & netherlandsimmigrant$age == "<40yrs"])[1],
    unique(dfp[netherlandsimmigrant$capture == 1 & netherlandsimmigrant$gender == "male"   & netherlandsimmigrant$nation == "American and Australia" & netherlandsimmigrant$age == "<40yrs"])[1],
    unique(dfp[netherlandsimmigrant$capture == 2 & netherlandsimmigrant$gender == "male"   & netherlandsimmigrant$nation == "American and Australia" & netherlandsimmigrant$age == "<40yrs"])[1],
    unique(dfp[netherlandsimmigrant$capture == 3 & netherlandsimmigrant$gender == "male"   & netherlandsimmigrant$nation == "American and Australia" & netherlandsimmigrant$age == "<40yrs"])[1],
    unique(dfp[netherlandsimmigrant$capture == 4 & netherlandsimmigrant$gender == "male"   & netherlandsimmigrant$nation == "American and Australia" & netherlandsimmigrant$age == "<40yrs"])[1]),
  dfp6[1:7],
  tolerance = eps
)

# Extractors

expect_true(
  max(abs(
    c(AIC(Model), AIC(Model1), AIC(Model2)) -
      c(1712.901, 1805.904, 1131.723)
  )) < 1
)

expect_true(
  max(abs(
    c(BIC(Model), BIC(Model1), BIC(Model2)) -
      c(1757.213, 1811.443, 1170.3)
  )) < 1
)

expect_silent(
  c(extractAIC(Model), extractAIC(Model1), extractAIC(Model2))
)

expect_equal(AIC(Model), AIC(Model6),
             tolerance = eps)
expect_equal(BIC(Model), BIC(Model6),
             tolerance = eps)
expect_equal(extractAIC(Model), extractAIC(Model6),
             tolerance = eps)

# Sandwich

require(sandwich)

expect_equivalent(
  vcovHC(Model),
  vcovHC(Model6),
  tolerance = eps
)

expect_equivalent(
  vcov(Model, type = "observedInform"),
  vcov(Model, type = "Fisher"),
  tolerance = max(.0001, eps)
)

expect_equivalent(
  vcov(Model1, type = "observedInform"),
  vcov(Model1, type = "Fisher"),
  tolerance = max(.0001, eps)
)

expect_equivalent(
  vcov(Model2, type = "observedInform"),
  vcov(Model2, type = "Fisher"),
  tolerance = max(.00001, eps)
)

expect_equivalent(
  bread(Model, type = "observedInform"),
  bread(Model, type = "Fisher"),
  tolerance = max(.0001, eps)
)

expect_equivalent(
  bread(Model1, type = "observedInform"),
  bread(Model1, type = "Fisher"),
  tolerance = max(.0001, eps)
)

expect_equivalent(
  bread(Model2, type = "observedInform"),
  bread(Model2, type = "Fisher"),
  tolerance = max(.00001, eps)
)

expect_silent(
  sandwich(Model)
)

expect_silent(
  sandwich(Model1)
)

expect_silent(
  sandwich(Model2)
)

expect_silent(
  sandwich(Model6)
)

expect_silent(
  vcovHC(Model)
)

expect_silent(
  vcovHC(Model6, type = "HC")
)

expect_silent(
  vcovHC(Model1, type = "HC4m")
)

expect_silent(
  vcovHC(Model2, type = "HC5")
)

expect_silent(
  vcovHC(Model, type = "const")
)

expect_silent(
  vcovHC(Model6, type = "HC2")
)

expect_silent(
  vcovHC(Model1, type = "HC1")
)

# confint

expect_silent(
  confint(Model)
)

expect_silent(
  confint(Model, parm = 1:2, level = .99)
)

expect_silent(
  cooks.distance(Model)
)

expect_silent(
  cooks.distance(Model1)
)

expect_silent(
  cooks.distance(Model2)
)

expect_silent(
  model.frame(Model2)
)

expect_true(
  all(dim(model.matrix(Model)) == c(1880, 8))
)

temp <- Model
temp$X <- NULL

expect_identical(
  model.matrix(Model),
  model.matrix(temp)
)

expect_identical(
  model.matrix(Model, "vlm"),
  model.matrix(temp, "vlm")
)

temp$modelFrame <- NULL

expect_identical(
  model.matrix(Model),
  model.matrix(temp)
)

expect_identical(
  model.matrix(Model, "vlm"),
  model.matrix(temp, "vlm")
)

expect_silent(
  print(popSizeEst(Model))
)

expect_equal(
  c(popSizeEst(Model)$pointEstimate, popSizeEst(Model1)$pointEstimate, 
    popSizeEst(Model2)$pointEstimate),
  c(12690, 7080, 15816.14),
  tol = .05
)

expect_equal(
  c(popSizeEst(Model)$variance, popSizeEst(Model1)$variance, 
    popSizeEst(Model2)$variance),
  c(7885812, 133774.1, 9093464),
  tol = .05
)

expect_silent(
  plot(Model, "qq")
)

expect_silent(
  plot(Model, "marginal")
)

expect_silent(
  plot(Model, "fitresid")
)

expect_error(
  plot(Model, "bootHist")
)

expect_silent(
  plot(Model, "rootogram")
)

expect_silent(
  plot(Model, "dfpopContr", dfpop = dfp)
)

expect_silent(
  plot(Model, "dfpopBox", dfpop = dfp)
)

expect_silent(
  plot(Model, "scaleLoc")
)

expect_silent(
  plot(Model, "cooks")
)

expect_silent(
  plot(Model, "hatplot")
)

expect_silent(
  plot(Model, "qq")
)

expect_silent(
  plot(Model, "strata")
)

expect_silent(
  up <- redoPopEstimation(Model, cov = vcovHC(Model, "HC4m"))
)

expect_silent(
  summary(Model, cov = vcovHC(Model, "HC4m"), correlation = TRUE, 
          confint = TRUE, popSizeEst = up)
)

expect_equivalent(
  up$confidenceInterval[1, ],
  data.frame(6611.906, 18768.8),
  tolerance = .05
)

expect_silent(
  up <- summary(marginalFreq(Model6), df = 1, dropl5 = "group")
)

expect_silent(
  print(up)
)

expect_equivalent(
  predict(Model, type = "response", se.fit = TRUE),
  predict(Model6, type = "response", se.fit = TRUE, newdata = model.frame(Model)),
  tolerance = eps
)

expect_equivalent(
  predict(Model, type = "link", se.fit = TRUE),
  predict(Model6, type = "link", se.fit = TRUE, newdata = netherlandsimmigrant[,-4]),
  tolerance = eps
)

expect_equivalent(
  predict(Model, type = "mean", se.fit = TRUE),
  predict(Model6, type = "mean", se.fit = TRUE, newdata = netherlandsimmigrant[,-4]),
  tolerance = eps
)

pred_data <- data.frame(
  y = c(1, 2, 3, 1, 2),
  x = c(1, 2, 3, 4, 5)
)

expect_silent(
  pred_mod <- estimatePopsize(
    formula = y ~ x,
    data = pred_data,
    model = "ztpoisson",
    controlMethod = controlMethod(silent = TRUE)
  )
)

custom_cov <- diag(c(1.5, 2.5))
custom_pred <- predict(pred_mod, type = "link", se.fit = TRUE, cov = custom_cov)
custom_se_col <- paste0("se:", family(pred_mod)$etaNames)

expect_equal(
  as.numeric(custom_pred[, custom_se_col]),
  sqrt(diag(model.matrix(pred_mod, type = "vlm") %*% custom_cov %*%
              t(model.matrix(pred_mod, type = "vlm")))),
  tolerance = eps
)

expect_silent(
  residuals(Model, type = "all")
)

expect_equal(
  logLik(Model),
  logLik(Model6),
  tolerance = eps
)

offset_data <- data.frame(y = c(1, 2, 3, 1, 2))
offset_matrix <- matrix(log(c(1, 2, 1, 2, 1)), ncol = 1)

expect_silent(
  offset_mod <- estimatePopsize(
    formula = y ~ 1,
    data = offset_data,
    model = "ztpoisson",
    offset = offset_matrix,
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_equal(
  -logLik(offset_mod, type = "function")(coef(offset_mod)),
  as.numeric(logLik(offset_mod)),
  tolerance = eps
)

expect_silent(
  up <- redoPopEstimation(
    Model6, newdata = netherlandsimmigrant[,-4]
  )
)

expect_silent(
  up1 <- redoPopEstimation(
    Model, 
    newdata = model.frame(Model6), 
    weights = Model6$priorWeights,
    weightsAsCounts = TRUE
  )
)

expect_equal(
  up, up1, tolerance = .025
)

expect_equal(
  stratifyPopsize(Model),
  stratifyPopsize(Model6),
  tolerance = eps
)

expect_error(
  stratifyPopsize(Model, strata = 8L)
)

expect_silent(
  print(Model)
)

expect_silent(
  print(family(Model))
)

expect_equal(
  NCOL(fitted(Model, "all")),
  2L
)

expect_equivalent(
  as.numeric(table(simulate(Model6, seed = 123)[,1])),
  c(1619, 232, 28, 1),
  tolerance = eps
)

weight_data <- data.frame(
  y = c(1, 2, 3, 1, 2),
  x = c(1, 2, NA, 4, 5)
)

expect_error(
  estimatePopsize(
    formula = y ~ x,
    data = pred_data[-5, ],
    weights = c(1, 2),
    model = "ztpoisson",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_silent(
  weight_mod <- estimatePopsize(
    formula = y ~ x,
    data = weight_data,
    weights = c(1, 2, 3, 4, 5),
    model = "ztpoisson",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_silent(
  weight_mod_complete <- estimatePopsize(
    formula = y ~ x,
    data = na.omit(weight_data),
    weights = c(1, 2, 4, 5),
    model = "ztpoisson",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_equal(
  weight_mod$priorWeights,
  weight_mod_complete$priorWeights,
  tolerance = eps
)

expect_equal(
  weight_mod$sizeObserved,
  nrow(model.frame(weight_mod)),
  tolerance = eps
)

expect_equal(
  weight_mod$sizeObserved,
  weight_mod_complete$sizeObserved,
  tolerance = eps
)

expect_equal(
  weight_mod$populationSize$confidenceInterval,
  weight_mod_complete$populationSize$confidenceInterval,
  tolerance = eps
)

set.seed(321)
oichao_y <- sample(c(1, 2, 3, 4), 20, replace = TRUE)
oichao_X <- cbind(1, rnorm(20))
oichao_beta <- c(-0.3, 0.2)
oichao_family <- oichao()
oichao_fn <- oichao_family$makeMinusLogLike(
  y = oichao_y,
  X = oichao_X,
  weight = rep(1, 20),
  deriv = 0
)
oichao_grad <- oichao_family$makeMinusLogLike(
  y = oichao_y,
  X = oichao_X,
  weight = rep(1, 20),
  deriv = 1
)
oichao_hess <- oichao_family$makeMinusLogLike(
  y = oichao_y,
  X = oichao_X,
  weight = rep(1, 20),
  deriv = 2
)
eps_fd <- 1e-6
fd_grad <- sapply(1:2, function(j) {
  beta_plus <- oichao_beta
  beta_minus <- oichao_beta
  beta_plus[j] <- beta_plus[j] + eps_fd
  beta_minus[j] <- beta_minus[j] - eps_fd
  (oichao_fn(beta_plus) - oichao_fn(beta_minus)) / (2 * eps_fd)
})
fd_hess <- matrix(0, 2, 2)
for (i in 1:2) {
  for (j in 1:2) {
    ei <- rep(0, 2)
    ej <- rep(0, 2)
    ei[i] <- eps_fd
    ej[j] <- eps_fd
    fd_hess[i, j] <- (
      oichao_fn(oichao_beta + ei + ej) -
        oichao_fn(oichao_beta + ei - ej) -
        oichao_fn(oichao_beta - ei + ej) +
        oichao_fn(oichao_beta - ei - ej)
    ) / (4 * eps_fd ^ 2)
  }
}

expect_true(
  max(abs(as.numeric(-oichao_grad(oichao_beta)) - fd_grad)) < 1e-5
)

expect_true(
  max(abs(-oichao_hess(oichao_beta) - fd_hess)) < 1e-3
)

expect_silent(
  oichao_fit <- estimatePopsize(
    formula = y ~ 1,
    data = data.frame(y = c(rep(2, 20), rep(3, 10))),
    model = "oichao",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_true(
  all(as.matrix(simulate(oichao_fit, nsim = 4, seed = 1)) %in% 2:3)
)

# oichao family math sanity checks -------------------------------------------

oichao_fam <- oichao()
oichao_eta <- matrix(oichao_fam$links[[1]](1.7))

expect_equal(
  as.numeric(oichao_fam$mu.eta(oichao_eta, type = "trunc")),
  1.7 / (3 + 1.7),
  tolerance = 1e-10
)

expect_equal(
  as.numeric(oichao_fam$mu.eta(oichao_eta, type = "nontrunc")),
  1.7,
  tolerance = 1e-10
)

expect_equal(
  as.numeric(oichao_fam$variance(oichao_eta, type = "trunc")),
  (1.7 / (3 + 1.7)) * (3 / (3 + 1.7)),
  tolerance = 1e-10
)

# density: dFun returns truncated Poisson by default
expect_equal(
  as.numeric(oichao_fam$densityFunction(2, oichao_eta, type = "trunc")),
  dpois(2, 1.7) / (1 - dpois(0, 1.7)),
  tolerance = 1e-10
)

expect_equal(
  as.numeric(oichao_fam$densityFunction(3, oichao_eta, type = "nontrunc")),
  dpois(3, 1.7),
  tolerance = 1e-10
)

# deviance residual: sign tracks y, value is small near the appropriate boundary
oichao_dev_y3 <- as.numeric(oichao_fam$devResids(
  y = 3, eta = matrix(oichao_fam$links[[1]](100)), wt = 1
))
oichao_dev_y2 <- as.numeric(oichao_fam$devResids(
  y = 2, eta = matrix(oichao_fam$links[[1]](0.01)), wt = 1
))
expect_true(is.finite(oichao_dev_y3))
expect_true(is.finite(oichao_dev_y2))
expect_true(oichao_dev_y3 > 0)
expect_true(oichao_dev_y2 < 0)
# residual shrinks toward zero as lambda approaches the perfect-fit boundary
expect_true(abs(oichao_dev_y3) < 1)
expect_true(abs(oichao_dev_y2) < 1)
# devResids is zero for y outside {2,3} (multiplied by indicator)
expect_equal(
  as.numeric(oichao_fam$devResids(y = 1, eta = oichao_eta, wt = 1)),
  0,
  tolerance = 1e-10
)
expect_equal(
  as.numeric(oichao_fam$devResids(y = 4, eta = oichao_eta, wt = 1)),
  0,
  tolerance = 1e-10
)

# popVar is non-negative on the intercept-only example used elsewhere
oichao_pop_fit <- estimatePopsize(
  formula = y ~ 1,
  data = data.frame(y = c(rep(2, 95), rep(3, 32))),
  model = "oichao",
  popVar = "analytic",
  controlMethod = controlMethod(silent = TRUE)
)
expect_true(is.finite(oichao_pop_fit$populationSize$variance))
expect_true(oichao_pop_fit$populationSize$variance > 0)
expect_true(
  oichao_pop_fit$populationSize$confidenceInterval[1, "lowerBound"] <=
    oichao_pop_fit$populationSize$pointEstimate
)
expect_true(
  oichao_pop_fit$populationSize$confidenceInterval[1, "upperBound"] >=
    oichao_pop_fit$populationSize$pointEstimate
)

# unsupported link is rejected
expect_error(oichao(lambdaLink = "log"), pattern = "logthird")


# not working methods ---------------------------------------------------------

expect_error(
  add1(mod1),
  "The add1 method for singleRStaticCountData class doesn't work yet"
)

expect_error(
  profile(mod1),
  "The profile method for singleRStaticCountData class doesn't work yet"
)

expect_error(
  drop1(mod1),
  "The drop1 method for singleRStaticCountData class doesn't work yet"
)

expect_error(
  anova(mod1),
  "The custom anova method for singleRStaticCountData class is not yet implemented. If the goal is to compare models we recommend using `lmtest::lrtest` instead."
)

# oichao family tests

oichao_theory_data <- data.frame(
  y = c(rep(2, 95), rep(3, 32))
)

expect_silent(
  fit_oichao_theory <- estimatePopsize(
    formula = y ~ 1,
    model = "oichao",
    data = oichao_theory_data,
    method = "IRLS",
    popVar = "analytic",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_equivalent(
  3 * exp(unname(coef(fit_oichao_theory))[1]),
  3 * 32 / 95,
  tolerance = .005
)

expect_equivalent(
  fit_oichao_theory$populationSize$pointEstimate,
  127 + (2 / 9) * 95 ^ 3 / 32 ^ 2,
  tolerance = .005
)

set.seed(2024)
data_oichao_covariate <- data.frame(
  x = rnorm(500)
)
data_oichao_covariate$lambda <- exp(0.2 + 0.35 * data_oichao_covariate$x)
data_oichao_covariate$y <- rpois(
  nrow(data_oichao_covariate),
  data_oichao_covariate$lambda
)
data_oichao_covariate <- subset(data_oichao_covariate, y > 0)

expect_silent(
  fit_oichao_irls <- estimatePopsize(
    formula = y ~ x,
    model = "oichao",
    data = data_oichao_covariate,
    method = "IRLS",
    popVar = "analytic",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_silent(
  fit_oichao_optim <- estimatePopsize(
    formula = y ~ x,
    model = oichao(),
    data = data_oichao_covariate,
    method = "optim",
    popVar = "analytic",
    controlMethod = controlMethod(
      silent = TRUE,
      optimMethod = "BFGS",
      maxiter = 500
    )
  )
)

expect_true(
  max(abs(coef(fit_oichao_irls) - coef(fit_oichao_optim))) < 1e-3
)

expect_true(
  abs(
    fit_oichao_irls$populationSize$pointEstimate -
      fit_oichao_optim$populationSize$pointEstimate
  ) < 0.25
)

expect_silent(
  predict(
    fit_oichao_irls,
    type = "response",
    se.fit = TRUE
  )
)

expect_silent(summary(fit_oichao_irls))

expect_silent(popSizeEst(fit_oichao_irls))

set.seed(321)
observed_oichao <- sample(c(1, 2, 3, 4), 20, replace = TRUE)
design_oichao <- cbind(1, rnorm(20))
coefficients_oichao <- c(-0.3, 0.2)
family_oichao <- oichao()
objective_oichao <- family_oichao$makeMinusLogLike(
  y = observed_oichao,
  X = design_oichao,
  weight = rep(1, 20),
  deriv = 0
)
gradient_oichao <- family_oichao$makeMinusLogLike(
  y = observed_oichao,
  X = design_oichao,
  weight = rep(1, 20),
  deriv = 1
)
hessian_oichao <- family_oichao$makeMinusLogLike(
  y = observed_oichao,
  X = design_oichao,
  weight = rep(1, 20),
  deriv = 2
)
eps_fd_oichao <- 1e-6
finite_difference_gradient_oichao <- sapply(1:2, function(j) {
  coefficients_plus <- coefficients_oichao
  coefficients_minus <- coefficients_oichao
  coefficients_plus[j] <- coefficients_plus[j] + eps_fd_oichao
  coefficients_minus[j] <- coefficients_minus[j] - eps_fd_oichao
  (objective_oichao(coefficients_plus) - objective_oichao(coefficients_minus)) /
    (2 * eps_fd_oichao)
})
finite_difference_hessian_oichao <- matrix(0, 2, 2)
for (i in 1:2) {
  for (j in 1:2) {
    ei <- rep(0, 2)
    ej <- rep(0, 2)
    ei[i] <- eps_fd_oichao
    ej[j] <- eps_fd_oichao
    finite_difference_hessian_oichao[i, j] <- (
      objective_oichao(coefficients_oichao + ei + ej) -
        objective_oichao(coefficients_oichao + ei - ej) -
        objective_oichao(coefficients_oichao - ei + ej) +
        objective_oichao(coefficients_oichao - ei - ej)
    ) / (4 * eps_fd_oichao ^ 2)
  }
}

expect_true(
  max(abs(
    as.numeric(-gradient_oichao(coefficients_oichao)) -
      finite_difference_gradient_oichao
  )) < 1e-5
)

expect_true(
  max(abs(
    -hessian_oichao(coefficients_oichao) -
      finite_difference_hessian_oichao
  )) < 1e-3
)

expect_silent(
  fit_oichao <- estimatePopsize(
    formula = y ~ 1,
    data = data.frame(y = c(rep(2, 20), rep(3, 10))),
    model = "oichao",
    controlMethod = controlMethod(silent = TRUE)
  )
)

expect_true(
  all(as.matrix(simulate(fit_oichao, nsim = 4, seed = 1)) %in% 2:3)
)

eta_oichao <- matrix(family_oichao$links[[1]](1.7))

expect_equal(
  as.numeric(family_oichao$mu.eta(eta_oichao, type = "trunc")),
  1.7 / (3 + 1.7),
  tolerance = 1e-10
)

expect_equal(
  as.numeric(family_oichao$mu.eta(eta_oichao, type = "nontrunc")),
  1.7,
  tolerance = 1e-10
)

expect_equal(
  as.numeric(family_oichao$variance(eta_oichao, type = "trunc")),
  (1.7 / (3 + 1.7)) * (3 / (3 + 1.7)),
  tolerance = 1e-10
)

expect_equal(
  as.numeric(family_oichao$densityFunction(2, eta_oichao, type = "trunc")),
  dpois(2, 1.7) / (1 - dpois(0, 1.7)),
  tolerance = 1e-10
)

expect_equal(
  as.numeric(family_oichao$densityFunction(3, eta_oichao, type = "nontrunc")),
  dpois(3, 1.7),
  tolerance = 1e-10
)

oichao_dev_y3 <- as.numeric(family_oichao$devResids(
  y = 3, eta = matrix(family_oichao$links[[1]](100)), wt = 1
))
oichao_dev_y2 <- as.numeric(family_oichao$devResids(
  y = 2, eta = matrix(family_oichao$links[[1]](0.01)), wt = 1
))
expect_true(is.finite(oichao_dev_y3))
expect_true(is.finite(oichao_dev_y2))
expect_true(oichao_dev_y3 > 0)
expect_true(oichao_dev_y2 < 0)
expect_true(abs(oichao_dev_y3) < 1)
expect_true(abs(oichao_dev_y2) < 1)

expect_equal(
  as.numeric(family_oichao$devResids(y = 1, eta = eta_oichao, wt = 1)),
  0,
  tolerance = 1e-10
)
expect_equal(
  as.numeric(family_oichao$devResids(y = 4, eta = eta_oichao, wt = 1)),
  0,
  tolerance = 1e-10
)

expect_silent(
  population_oichao_theory <- popSizeEst(fit_oichao_theory)
)
expect_true(is.finite(population_oichao_theory$variance))
expect_true(population_oichao_theory$variance > 0)
expect_true(
  population_oichao_theory$confidenceInterval[1, "lowerBound"] <=
    population_oichao_theory$pointEstimate
)
expect_true(
  population_oichao_theory$confidenceInterval[1, "upperBound"] >=
    population_oichao_theory$pointEstimate
)

expect_error(oichao(lambdaLink = "log"), pattern = "logthird")

#' \loadmathjax
#' 
#' @title Regression fitting in single-source capture-recapture models
#' @author Piotr Chlebicki, Maciej Beresewicz
#'
#' @description \code{estimatePopsizeFit} does for \code{estimatePopsize} what
#' \code{glm.fit} does for \code{glm}. It is internally called in 
#' \code{estimatePopsize}. Since \code{estimatePopsize} does much more than
#' just regression fitting \code{estimatePopsizeFit} is much faster.
#'
#' @param y vector of dependent variables.
#' @param X model matrix, the vglm one.
#' @param family same as model in \code{estimatePopsize}.
#' @param control control parameters created in \code{controlModel}.
#' @param method method of estimation same as in \code{estimatePopsize}.
#' @param priorWeights vector of prior weights its the same argument as weights
#' in \code{estimatePopsize}.
#' @param etaStart,coefStart initial value of regression parameters or
#' linear predictors.
#' @param offset offset passed from by default passed from [estimatePopsize()].
#' @param ... arguments to pass to other methods.
#' 
#' @details If \code{method} argument was set to \code{"optim"} the \code{stats::optim}
#' function will be used to fit regression with analytically computed gradient and 
#' (minus) log likelihood functions as \code{gr} and \code{fn} arguments. 
#' Unfortunately \code{optim} does not allow for hessian to be specified.
#' More information about how to modify \code{optim} fitting is included in 
#' [controlMethod()].
#' 
#' 
#' If \code{method} argument was set to \code{"IRLS"} the iteratively reweighted 
#' least squares. The algorithm is well know in generalised linear models.
#' Thomas W. Yee later extended this algorithm to vector generalised linear models 
#' and in more general terms it can roughly be described as 
#' (this is Yee's description after changing some conventions):
#' 
#' 1. Initialize with:
#' \itemize{
#' \item \code{converged <- FALSE}
#' \item \code{iter <- 1}
#' \item \mjseqn{\boldsymbol{\beta}}\code{ <- start}
#' \item \mjseqn{\boldsymbol{W}}\code{ <- prior}
#' \item \mjseqn{\ell}\code{ <- }\mjseqn{\ell(\boldsymbol{\beta})}
#' }
#' 2. If \code{converged} or \code{iter > Maxiter} move to step 7.
#' 3. Store values from previous algorithm step:
#' \itemize{
#' \item \mjseqn{\boldsymbol{W}_{-}}\code{ <- }
#' \mjseqn{\boldsymbol{W}}
#' \item \mjseqn{\ell_{-}}\code{ <- }
#' \mjseqn{\ell}
#' \item \mjseqn{\boldsymbol{\beta}_{-}}\code{ <- }
#' \mjseqn{\boldsymbol{\beta}}
#' } and assign values at current step:
#' \itemize{
#' \item \mjseqn{\boldsymbol{\eta}}\code{ <- }
#' \mjseqn{\boldsymbol{X}_{vlm}\boldsymbol{\beta}}
#' \item \mjseqn{Z_{i}}\code{ <- }
#' \mjseqn{
#' \eta_{i}+\frac{\partial\ell_{i}}{\partial\eta_{i}}
#' \mathbb{E}\left(\frac{\partial^{2}\ell_{i}}{
#' \partial\eta_{i}^{T}\partial\eta_{i}}\right)^{-1}}
#' \item \mjseqn{\boldsymbol{W}_{ij}}\code{ <- }
#' \mjseqn{\mathbb{E}\left(\frac{\partial^{2}\ell}{
#' \partial\boldsymbol{\eta}_{j}^{T}\partial\boldsymbol{\eta}_{i}}\right)}
#' }
#' where \mjseqn{\ell_{i}} is the ith component of log likelihood 
#' function, \mjseqn{\eta_{i}} is the vector of linear predictors 
#' associated with ith row and \mjseqn{\mathbb{E}\left(\frac{\partial^{2}\ell_{i}}{
#' \partial\eta_{i}^{T}\partial\eta_{i}}\right)}
#' corresponds to weights associated with ith row and \mjseqn{\boldsymbol{W}} 
#' is a block matrix, made of diagonal matrixes 
#' \mjseqn{\mathbb{E}\left(\frac{\partial^{2}\ell}{
#' \partial\boldsymbol{\eta}_{j}^{T}\partial\boldsymbol{\eta}_{i}}\right)}
#' 4. Regress \mjseqn{\boldsymbol{Z}} on 
#' \mjseqn{\boldsymbol{X}_{vlm}} to obtain 
#' \mjseqn{\boldsymbol{\beta}} as:
#' \mjsdeqn{\boldsymbol{\beta}=
#' \left(\boldsymbol{X}_{vlm}^{T}\boldsymbol{W}\boldsymbol{X}_{vlm}\right)^{-1}
#' \boldsymbol{X}_{vlm}^{T}\boldsymbol{W}\boldsymbol{Z}}
#' 5. Assign:
#' \itemize{
#' \item\code{converged <- }\mjteqn{\ell(\boldsymbol{\beta})-\ell_{-} < \varepsilon\cdot\ell_{-}}{\ell(\boldsymbol{\beta})-\ell_{-} \lt \varepsilon\cdot\ell_{-}}{\ell(\boldsymbol{\beta})-\ell_{-} < \varepsilon\cdot\ell_{-}}
#' or
#' \mjteqn{||\boldsymbol{\beta}-\boldsymbol{\beta}_{-}||_{\infty} < \varepsilon}{||\boldsymbol{\beta}-\boldsymbol{\beta}_{-}||_{\infty} \lt \varepsilon}{||\boldsymbol{\beta}-\boldsymbol{\beta}_{-}||_{\infty} < \varepsilon}
#' \item\code{iter <- iter + 1}
#' }
#' where \mjseqn{\varepsilon} is the relative tolerance level, 
#' by default \code{1e-8}.
#' 6. Return to step 2.
#' 7. Return \mjseqn{\boldsymbol{\beta}, \boldsymbol{W}}, \code{iter}.
#' 
#' In this package we use different conventions for \mjseqn{\boldsymbol{X}_{vlm}}
#' matrix hence slight differences are present in algorithm description but 
#' results are identical.
#' 
#' @references Yee, T. W. (2015). Vector Generalized Linear and Additive Models: 
#' With an Implementation in R. New York, USA: Springer. ISBN 978-1-4939-2817-0.
#' 
#' @examples 
#' \donttest{
# Get data
#' summary(farmsubmission)
#' 
#' # construct vglm model matrix
#' X <- matrix(data = 0, nrow = 2 * NROW(farmsubmission), ncol = 7)
#' X[1:NROW(farmsubmission), 1:4] <- model.matrix(
#' ~ 1 + log_size + log_distance + C_TYPE, 
#' farmsubmission
#' )
#' 
#' 
#' X[-(1:NROW(farmsubmission)), 5:7] <- X[1:NROW(farmsubmission), c(1, 3, 4)]
#' 
#' # this attribute tells the function which elements of the design matrix 
#' # correspond to which linear predictor 
#' attr(X, "hwm") <- c(4, 3)
#' 
#' # get starting points
#' start <- glm.fit(
#' y = farmsubmission$TOTAL_SUB, 
#' x = X[1:NROW(farmsubmission), 1:4], 
#' family = poisson()
#' )$coefficients
#' 
#' res <- estimatePopsizeFit(
#' y = farmsubmission$TOTAL_SUB, 
#' X = X, 
#' method = "IRLS", 
#' priorWeights = 1, 
#' family = ztoigeom(), 
#' control = controlMethod(verbose = 5), 
#' coefStart = c(start, 0, 0, 0),
#' etaStart = matrix(X %*% c(start, 0, 0, 0), ncol = 2),
#' offset = cbind(rep(0, NROW(farmsubmission)), rep(0, NROW(farmsubmission)))
#' )
#' 
#' # extract results
#' 
#' # regression coefficient vector
#' res$beta
#' 
#' # check likelihood
#' ll <- ztoigeom()$makeMinusLogLike(y = farmsubmission$TOTAL_SUB, X = X)
#' 
#' -ll(res$beta)
#' 
#' # number of iterations
#' res$iter
#' 
#' # working weights
#' head(res$weights)
#' 
#' # Compare with optim call
#' 
#' res2 <- estimatePopsizeFit(
#'   y = farmsubmission$TOTAL_SUB, 
#'   X = X, 
#'   method = "optim", 
#'   priorWeights = 1, 
#'   family = ztoigeom(), 
#'   coefStart = c(start, 0, 0, 0),
#'   control = controlMethod(verbose = 1, silent = TRUE),
#'   offset = cbind(rep(0, NROW(farmsubmission)), rep(0, NROW(farmsubmission)))
#' )
#' # extract results
#' 
#' # regression coefficient vector
#' res2$beta
#' 
#' 
#' # check likelihood
#' -ll(res2$beta)
#' 
#' # number of calls to log lik function
#' # since optim does not return the number of
#' # iterations
#' res2$iter
#' 
#' # optim does not calculated working weights
#' head(res2$weights)
#' }
#' @return List with regression parameters, working weights 
#' (if IRLS fitting method) was chosen and number of iterations taken.
#' @seealso [stats::glm()] [estimatePopsize()] [controlMethod()] [stats::optim()] 
# #' @importFrom maxLik maxLik
#' @export
estimatePopsizeFit <- function(y, X,
                               family,
                               control,
                               method,
                               priorWeights,
                               coefStart,
                               etaStart,
                               offset,
                               ...) {
  hwm <- attr(X, "hwm")
  tbgname <- colnames(X)
  X <- as.matrix(X)
  logg <- NULL
  
  if (method == "IRLS") {
    
    FITT <- singleRcaptureinternalIRLSmultipar(
      dependent          = y,
      covariates         = X,
      eps                = control$epsilon,
      family             = family,
      maxiter            = control$maxiter,
      weights            = priorWeights,
      etaStart           = etaStart,
      silent             = control$silent,
      trace              = control$verbose,
      stepsize           = control$stepsize,
      hwm                = hwm,
      momentumFactor     = control$momentumFactor,
      momentumActivation = control$momentumActivation,
      check              = control$checkDiagWeights,
      epsWeights         = control$weightsEpsilon,
      crit               = control$criterion,
      saveLog            = control$saveIRLSlogs,
      printOften         = control$printEveryN,
      offset             = offset
    )
    
    iter <- FITT$iter
    weights <- FITT$weights
    beta <- FITT$coefficients
    logg <- FITT$logg
  } else if (method == "optim") {
    logLike <- family$makeMinusLogLike(
      y = y, X = X, 
      weight = priorWeights,
      offset = offset
    )
    
    grad <- family$makeMinusLogLike(
      y = y, X = X, 
      weight = priorWeights, 
      deriv  = 1,
      offset = offset
    )
    
    weights <- priorWeights
    methodopt <- control$optimMethod
    
    if (!isFALSE(control$optimPass)) {
      ctrl <- control$optimPass
    } else {
      if (methodopt == "L-BFGS-B") {
        ctrl <- list(
          factr = control$epsilon,
          maxit = control$maxiter,
          trace = control$verbose
        )
      } else {
        ctrl <- list(
          reltol = control$epsilon,
          maxit = control$maxiter,
          trace = control$verbose
        )
      }
    }
    
    giveError <- tryCatch(
      expr = {
        FITT <- stats::optim(
          fn = logLike,
          par = coefStart,
          gr = function(x) -grad(x),
          method = methodopt,
          # hessian = TRUE,
          control = ctrl
        )
        list(0)
      },
      error = function (e) {
        list(1, e)
      }
    )

    if (giveError[[1]] == 1) {
      stop(paste0(
        "stats::optim fitting returned with the following error: ",
        giveError[[2]]$message,
        " consider fitting with 'IRLS' or changing controlMethod argument."
      ))
    }
    
    if (FITT$convergence != 0 && !isTRUE(control$silent)) {
      warning("Convergence not obtained in ", control$maxiter, 
              " iterations of optim fitting algorithm", sep = "")
      
      switch (as.character(FITT$convergence),
        "1"  = warning("Convergence not obtained in ", control$maxiter, 
                       " iterations of optim fitting algorithm", sep = ""),
        "10" = warning("stats::optim returned with code 10 degeneracy of the Nelder-Mead simplex."),
        "51" =    stop("stats::optim returned with code 51 with message: ", FITT$message),
        "52" = warning("stats::optim returned with code 52 with message: ", FITT$message)
      )
    }
    
    iter <- FITT$counts
    beta <- FITT$par
  } else {
    stop("Method not implemented.")
  }
  if (is.null(FITT)) {
    stop("fitting error try another model
          (negative binomial models are highly volatile)")
  }
  
  beta <- as.vector(beta)

  list(
    beta = beta,
    weights = weights,
    iter = iter,
    logg = logg
  )
}

#' @rdname singleRmodels
#' @importFrom stats uniroot
#' @importFrom stats dnbinom
#' @export
ztHurdlenegbin <- function(nSim = 1000, epsSim = 1e-8, eimStep = 6,
                           lambdaLink = c("log", "neglog"), 
                           alphaLink = c("log", "neglog"),
                           piLink = c("logit", "cloglog", "probit"), 
                           ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  if (missing(alphaLink))  alphaLink  <- "log"
  if (missing(piLink))     piLink     <- "logit"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink, alphaLink, piLink)
  
  lambdaLink <- switch(lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  alphaLink <- switch(alphaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  piLink <- switch(piLink,
    "logit" = singleRinternallogitLink,
    "cloglog" = singleRinternalcloglogLink,
    "probit" = singleRinternalprobitLink
  )
  
  links[1:3] <- c(lambdaLink, alphaLink, piLink)
  
  nbLogP0 <- function(lambda, alpha) {
    -(1 / alpha) * log1p(lambda * alpha)
  }

  nbLogP1 <- function(lambda, alpha) {
    log(lambda) + (-1 / alpha - 1) * log1p(lambda * alpha)
  }

  nbP0 <- function(lambda, alpha) {
    exp(nbLogP0(lambda, alpha))
  }

  nbP1 <- function(lambda, alpha) {
    exp(nbLogP1(lambda, alpha))
  }

  probGe2 <- function(lambda, alpha) {
    logP0 <- nbLogP0(lambda, alpha)
    logP1 <- nbLogP1(lambda, alpha)
    logP01 <- pmax(logP0, logP1) + log1p(exp(-abs(logP0 - logP1)))
    pmax(-expm1(pmin(logP01, 0)), .Machine$double.xmin)
  }

  logProbGe2 <- function(lambda, alpha) {
    log(probGe2(lambda, alpha))
  }

  probGe2Scaled <- function(lambda, alpha) {
    (1 + alpha * lambda) ^ (1 / alpha + 1) * probGe2(lambda, alpha)
  }

  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    P1 <- nbP1(lambda, alpha)
    prob_ge2 <- probGe2(lambda, alpha)
    prob_ge2_scaled <- probGe2Scaled(lambda, alpha)
    
    if (!deriv) {
      switch (type,
        "nontrunc" = (1 - (1 + alpha * lambda) ^ (-1 / alpha)) * (PI + (1 - PI) * lambda),
        "trunc" = PI + (1 - PI) * 
        (lambda - P1) / prob_ge2
      )
    } else {
      switch (
        type,
        "nontrunc" = {
          matrix(c(
            (1 - PI) * (1 - 1 / (alpha * lambda + 1) ^ (1 / alpha)) +
            ((1 - PI) * lambda + PI) * (alpha * lambda + 1) ^ (-1 / alpha - 1),
            -((1 - PI) * lambda + PI) * 
            (log(lambda * alpha + 1) / alpha ^ 2 - 
            lambda / (alpha * (lambda * alpha + 1))) /
            (lambda * alpha + 1) ^ (1 / alpha),
            (1 - lambda) * (1 - 1 / (alpha * lambda + 1) ^ (1 / alpha))
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
              alphaLink(eta[,2], inverse = TRUE, deriv = 1),
                piLink(eta[, 3], inverse = TRUE, deriv = 1)
          ), ncol = 3)
        },
        "trunc" = {
	          matrix(c(
	            (1 - PI) * ((alpha * lambda + 1) ^ (2 / alpha) * (alpha ^ 2 * lambda ^ 2 + 2 * alpha * lambda + 1) +
	            (alpha * lambda + 1) ^ (1 / alpha) * ((-alpha ^ 2 - 2 * alpha - 1) * lambda ^ 2 - 2 * alpha * lambda - 2) + 1) /
	            prob_ge2_scaled ^ 2,
	            (1 - PI) * lambda ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha) *
	            (lambda * alpha ^ 2 + (lambda + 1) * alpha + 1) * log(lambda * alpha + 1) +
	            (lambda * alpha + 1) ^ (1 / alpha) * ((1 - 2 * lambda) * alpha ^ 2 - lambda * alpha) - alpha ^ 2) /
	            (alpha ^ 2 * prob_ge2_scaled ^ 2),
	            1 - (lambda - P1) / prob_ge2
	          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
             alphaLink(eta[, 2], inverse = TRUE, deriv = 1),
                piLink(eta[, 3], inverse = TRUE, deriv = 1)
          ), ncol = 3)
        }
      )
    }
  }
  
  variance <- function(eta, type = "nontrunc", ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    P0 <- nbP0(lambda, alpha)
    P1 <- nbP1(lambda, alpha)
    prob_ge2 <- probGe2(lambda, alpha)
    
    res <- switch (type,
      nontrunc = (1 - P0) * (PI + (1 - PI) * lambda + (1 + alpha) * lambda ^ 2),
      trunc = PI + (1 - PI) * (lambda * (1 + alpha * lambda) + lambda ^ 2 - 
      P1) / prob_ge2
    ) - mu.eta(eta = eta, type = type) ^ 2

    pmax(res, 0)
  }
  
  compdigamma <- function(y, alpha) {
    (-digamma(y + 1 / alpha) + digamma(1 / alpha)) / (alpha ^ 2)
  }
  
  
  comptrigamma <- function(y, alpha) {
    (2 * (digamma(y + 1 / alpha) - digamma(1 / alpha)) * alpha +
       trigamma(y + 1 / alpha) - trigamma(1 / alpha)) / (alpha ^ 4)
  }
  
  # Computing the expected value of di/trigamma functions on (y + 1/alpha)
  
  compExpectG1 <- function(eta) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-  piLink(eta[, 3], inverse = TRUE)
    
    prob_ge2 <- probGe2(lambda, alpha)
    res <- rep(0, NROW(eta))
    
    # 1 is the first possible y value for 0 truncated hurdle distribution
    # but here we compute the (1 - z) * psi function which takes 0 at y = 1
    k <- 2 
    finished <- rep(FALSE, NROW(eta))
    while ((k < nSim) & !all(finished)) {
      prob <- apply(cbind(k:(k + eimStep)), MARGIN = 1, FUN = function(x) {
	        (1 - PI) * stats::dnbinom(
	          x = x,
	          size = 1 / alpha,
	          mu = lambda
	        ) / prob_ge2
	      })
      trg <- apply(cbind(k:(k + eimStep)), MARGIN = 1, FUN = function(x) {
        comptrigamma(y = x, alpha = alpha)
      })
      prob[!(is.finite(prob))] <- 0
      trg[!(is.finite(trg))] <- 0
      toAdd <- trg * prob
      toAdd <- rowSums(toAdd)
      k <- k + eimStep + 1
      res <- res + toAdd
      finished <- abs(toAdd) < epsSim
    }
    res
  }
  
  Wfun <- function(prior, eta, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    prob_ge2 <- probGe2(lambda, alpha)
    prob_ge2_scaled <- probGe2Scaled(lambda, alpha)
    
    z  <- PI
    
    ## expected for (1-I)Y
    XXX <- mu.eta(eta, type = "trunc") - z
    
    Etrig <- compExpectG1(eta)
    
    # PI
    G00 <- (-z / PI ^ 2 - (1 - z) / (1 - PI) ^ 2)
    
    G00 <- G00 * piLink(eta[, 3], inverse = TRUE, deriv = 1) ^ 2
    
    G01 <- rep(0, NROW(eta))
    G02 <- rep(0, NROW(eta))
    
    # alpha
	    G11 <- Etrig + (1 - z) * (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
	    (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
	    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) ^ 2 /
	    prob_ge2 ^ 2 -
	    (1 - z) * (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) ^ 2 /
	    (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
	    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) /
    (lambda * alpha + 1)) ^ 2 - (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 + 
    (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) + lambda ^ 2 / 
    (alpha * (lambda * alpha + 1) ^ 2)) / (lambda * alpha + 1) ^ (1 / alpha) -
	    lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
	    (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 + (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) -
	    (lambda ^ 2 * (-1 / alpha - 1)) / (lambda * alpha + 1) ^ 2)) /
	    prob_ge2 -
	    (1 - z) * (2 * log(lambda * alpha + 1)) / alpha ^ 3 + (1 - z) * (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) -
	    (lambda ^ 2 * (-(1 - z) / alpha - XXX)) / (lambda * alpha + 1) ^ 2 - XXX / alpha ^ 2
    
    G11 <- G11 * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2
    
    # alpha lambda
    G12 <- (-1) * ((1 - z) * (alpha * lambda + 1) ^ (1 / alpha) * 
    ((alpha ^ 3 + alpha ^ 2) * lambda ^ 3 + (2 * alpha ^ 2 + 2 * alpha) * lambda ^ 2 + (alpha + 1) * lambda) * 
    log(alpha * lambda + 1) + (alpha * lambda + 1) ^ (1 / alpha) * 
    ((1 - z) * (alpha ^ 4 - alpha ^ 3 - alpha ^ 2) * lambda ^ 3 + 
    ((-2 * alpha ^ 4 - 2 * alpha ^ 3) * XXX + (1 - z) * 4 * alpha ^ 3 - 
    (1 - z) *  alpha ^ 2 - (1 - z) *  alpha) * lambda ^ 2 + 
    ((-4 * alpha ^ 3 - 2 * alpha ^ 2) * XXX + (1 - z) *  3 * alpha ^ 2) * lambda - 
	    2 * alpha ^ 2 * XXX) + (alpha * lambda + 1) ^ (2 / alpha) * (-(1 - z) * alpha ^ 4 * lambda ^ 3 +
	    (alpha ^ 4 * XXX - (1 - z) * 2 * alpha ^ 3) * lambda ^ 2 + (2 * alpha ^ 3 * XXX -
	    (1 - z) * alpha ^ 2) * lambda + alpha ^ 2 * XXX) + ((alpha ^ 4 + 2 * alpha ^ 3 + alpha ^ 2) * XXX -
    (1 - z) * 2 * alpha ^ 3 - (1 - z) * alpha ^ 2) * lambda ^ 2 +
	    ((2 * alpha ^ 3 + 2 * alpha ^ 2) * XXX - (1 - z) * 2 * alpha ^ 2) * lambda + alpha ^ 2 * XXX) /
	    (alpha ^ 2 * (alpha * lambda + 1) ^ 2 * prob_ge2_scaled ^ 2)
    
    G12 <- G12 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
                  alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
    
    #lambda
    G22 <- ((alpha * lambda + 1) ^ (2 / alpha) * ((1 - z) * alpha ^ 3 * lambda ^ 4 + 
    ((1 - z) * 2 * alpha ^ 2 - 2 * alpha ^ 3 * XXX) * lambda ^ 3 +
    ((1 - z) * alpha - 5 * alpha ^ 2 * XXX) * lambda ^ 2 - 4 * alpha * XXX * lambda - XXX) + 
    (alpha * lambda + 1) ^ (1 / alpha) * ((1 - z) * (alpha - alpha ^ 3) * lambda ^ 4 +
    ((4 * alpha ^ 3 + 4 * alpha ^ 2) * XXX - (1 - z) * 4 * alpha ^ 2 - (1 - z) * alpha + (1 - z)) * lambda ^ 3 +
    ((10 * alpha ^ 2 + 6 * alpha) * XXX - (1 - z) * 3 * alpha - (1 - z)) * lambda ^ 2 + 
    (8 * alpha + 2) * XXX * lambda + 2 * XXX) + 
	    ((-2 * alpha ^ 3 - 4 * alpha ^ 2 - 2 * alpha) * XXX + (1 - z) * 2 * alpha ^ 2 +
	    (1 - z) *  2 * alpha) * lambda ^ 3 + ((-5 * alpha ^ 2 - 6 * alpha - 1) * XXX +
	    (1 - z) * 2 * alpha + (1 - z)) * lambda ^ 2 + (-4 * alpha - 2) * XXX * lambda - XXX) /
	    (lambda ^ 2 * (alpha * lambda + 1) ^ 2 * prob_ge2_scaled ^ 2)
    
    G22 <- G22 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2
    
    matrix(
      -c(G22 * prior, G12 * prior, G02 * prior,
         G12 * prior, G11 * prior, G01 * prior,
         G02 * prior, G01 * prior, G00 * prior),
      dimnames = list(
        rownames(eta), 
        c("lambda", "lambda:alpha", "lambda:PI",
          "lambda:alpha", "alpha", "alpha:PI",
          "lambda:PI", "alpha:PI", "PI")
      ),
      ncol = 9
    )
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    z <- as.numeric(y == 1)
    weight <- weight / prior
    prob_ge2 <- probGe2(lambda, alpha)
    
    dig <- compdigamma(y = y, alpha = alpha)
    
    G0 <- z / PI - (1 - z) / (1 - PI)
    
	    G1 <- (1 - z) * (y / alpha + (digamma(1 / alpha) - digamma(1 / alpha + y)) / alpha ^ 2  +
	    log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - y)) / (lambda * alpha + 1) -
	    (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
	    (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
	    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) /
	    prob_ge2)
    
	    G2 <- (1 - z) * (y / lambda  + (alpha * (-y - 1 / alpha)) / (alpha * lambda + 1) -
	    ((1 / alpha + 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) /
	    prob_ge2)
    
    G0 <- G0 *  piLink(eta[, 3], inverse = TRUE, deriv = 1)
    G1 <- G1 *  alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
    G2 <- G2 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
    
    uMatrix <- cbind(G2, G1, G0)
    
    weight <- lapply(X = 1:nrow(weight), FUN = function (x) {
      matrix(as.numeric(weight[x, ]), ncol = 3)
    })
    
    pseudoResid <- sapply(X = 1:length(weight), FUN = function (x) {
      xx <- solve(weight[[x]])
      xx %*% uMatrix[x, ]
    })
    
    pseudoResid <- t(pseudoResid)
    dimnames(pseudoResid) <- dimnames(eta)
    pseudoResid
  }
  
  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, 
                           ...) {
    if (is.null(weight)) {
      weight <- 1
    }
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X) / 3), rep(0, NROW(X) / 3), rep(0, NROW(X) / 3))
    }
    
    y <- as.numeric(y)
    z <- as.numeric(y == 1)
    X <- as.matrix(X)

    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
            function(beta) {
              eta    <- matrix(as.matrix(X) %*% beta, ncol = 3) + offset
              lambda <- lambdaLink(eta[, 1], inverse = TRUE)
              alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
              PI     <-     piLink(eta[, 3], inverse = TRUE)
              
              -sum(weight * (z * log(PI) + (1 - z) * log(1 - PI) + (1 - z) *
              (lgamma(y + 1 / alpha) - lgamma(1 / alpha) -
              lgamma(y + 1) - (y + 1 / alpha) * log(1 + lambda * alpha) +
              y * log(lambda * alpha) - logProbGe2(lambda, alpha))))
            },
            function(beta) {
              eta    <- matrix(as.matrix(X) %*% beta, ncol = 3) + offset
	              lambda <- lambdaLink(eta[, 1], inverse = TRUE)
	              alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
	              PI     <-     piLink(eta[, 3], inverse = TRUE)
	              prob_ge2 <- probGe2(lambda, alpha)
              
	              dig <- compdigamma(y = y, alpha = alpha)
              
              G0 <- z / PI - (1 - z) / (1 - PI)
              
	              G1 <- (1 - z) * (y / alpha + (digamma(1 / alpha) - digamma(1 / alpha + y)) / alpha ^ 2  +
	              log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - y)) / (lambda * alpha + 1) -
	              (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
	              (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
	              (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) /
	              prob_ge2)
              
	              G2 <- (1 - z) * (y / lambda  + (alpha * (-y - 1 / alpha)) / (alpha * lambda + 1) -
	              ((1 / alpha + 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) /
	              prob_ge2)
              
              G0 <- G0 * weight *     piLink(eta[, 3], inverse = TRUE, deriv = 1)
              G1 <- G1 * weight *  alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
              G2 <- G2 * weight * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
              
              if (NbyK) {
                return(cbind(
                  G2 * as.data.frame(
                    X[1:nrow(eta), 
                      1:(attr(X, "hwm")[1])]
                  ), 
                  G1 * as.data.frame(
                    X[(nrow(eta) + 1):(2 * nrow(eta)), 
                      (attr(X, "hwm")[1] + 1):(attr(X, "hwm")[2])]
                  ), 
                  G0 * as.data.frame(
                    X[(2 * nrow(eta) + 1):(3 * nrow(eta)), 
                      (attr(X, "hwm")[2] + 1):(attr(X, "hwm")[3])]
                  )
                ))
              }
              if (vectorDer) {
                return(cbind(G2, G1, G0))
              }
              
              as.numeric(c(G2, G1, G0) %*% X)
            },
            function(beta) {
              predNumbers <- attr(X, "hwm")
              eta    <- matrix(as.matrix(X) %*% beta, ncol = 3) + offset
	              lambda <- lambdaLink(eta[, 1], inverse = TRUE)
	              alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
	              PI     <-     piLink(eta[, 3], inverse = TRUE)
	              prob_ge2 <- probGe2(lambda, alpha)
	              prob_ge2_scaled <- probGe2Scaled(lambda, alpha)
              
              res <- matrix(
                nrow = length(beta), 
                ncol = length(beta), 
                dimnames = list(names(beta), names(beta))
              )
              
              trig <- comptrigamma(y = y, alpha = alpha)
              dig  <-  compdigamma(y = y, alpha = alpha)
              
              # PI
              G0 <- z / PI - (1 - z) / (1 - PI)
              
              G00 <- (-z / PI ^ 2 - (1 - z) / (1 - PI) ^ 2)
              
              G00 <- G00 * piLink(eta[, 3], inverse = TRUE, deriv = 1) ^ 2 +
                      G0 * piLink(eta[, 3], inverse = TRUE, deriv = 2)
              
              # alpha
	              G1 <- (1 - z) * (y / alpha + (digamma(1 / alpha) - digamma(1 / alpha + y)) / alpha ^ 2 -
	              (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
	              (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
	              (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) /
	              prob_ge2 +
	              log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - y)) / (lambda * alpha + 1))
              
              G11 <- (1 - z) * ((2 * digamma(1 / alpha + y)) / alpha ^ 3 - 
              (2 * digamma(1 / alpha)) / alpha ^ 3 + trigamma(1 / alpha + y) / alpha ^ 4 - 
              trigamma(1 / alpha) / alpha ^ 4 + (-(log(lambda * alpha + 1) / alpha ^ 2 - 
	              lambda / (alpha * (lambda * alpha + 1))) /(lambda * alpha + 1) ^ (1 / alpha) -
	              lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) * (log(lambda * alpha + 1) / alpha ^ 2 +
	              (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) ^ 2 /
	              prob_ge2 ^ 2 -
              (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) ^ 2 / 
              (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) /
              (lambda * alpha + 1)) ^ 2 - (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 + 
              (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) + lambda ^ 2 / 
              (alpha * (lambda * alpha + 1) ^ 2)) / (lambda * alpha + 1) ^ (1 / alpha) -
	              lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) * (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 +
	              (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) - (lambda ^ 2 * (-1 / alpha - 1)) / (lambda * alpha + 1) ^ 2)) /
	              prob_ge2 -
              (2 * log(lambda * alpha + 1)) / alpha ^ 3 + (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) -
              (lambda ^ 2 * (-1 / alpha - y)) / (lambda * alpha + 1) ^ 2 - y / alpha ^ 2)
              
              G11 <- G11 * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2 +
                      G1 * alphaLink(eta[, 2], inverse = TRUE, deriv = 2)
              
              # alpha lambda
              G12 <- (z - 1) * ((alpha * lambda + 1) ^ (1 / alpha) * ((alpha ^ 3 + alpha ^ 2) * lambda ^ 3 +
              (2 * alpha ^ 2 + 2 * alpha) * lambda ^ 2 + (alpha + 1) * lambda) * 
              log(alpha * lambda + 1) + (alpha * lambda + 1) ^ (1 / alpha) * 
              ((alpha ^ 4 - alpha ^ 3 - alpha ^ 2) * lambda ^ 3 + 
              ((-2 * alpha ^ 4 - 2 * alpha ^ 3) * y + 4 * alpha ^ 3 - alpha ^ 2 - alpha) * lambda ^ 2 + 
              ((-4 * alpha ^ 3 - 2 * alpha ^ 2) * y + 3 * alpha ^ 2) * lambda - 2 * alpha ^ 2 * y) +
	              (alpha * lambda + 1) ^ (2 / alpha) * (-alpha ^ 4 * lambda ^ 3 +
	              (alpha ^ 4 * y - 2 * alpha ^ 3) * lambda ^ 2 + (2 * alpha ^ 3 * y - alpha ^ 2) * lambda + alpha ^ 2 * y) +
	              ((alpha ^ 4 + 2 * alpha ^ 3 + alpha ^ 2) * y - 2 * alpha ^ 3 - alpha ^ 2) * lambda ^ 2 +
	              ((2 * alpha ^ 3 + 2 * alpha ^ 2) * y - 2 * alpha ^ 2) * lambda + alpha ^ 2 * y) /
	              (alpha ^ 2 * (alpha * lambda + 1) ^ 2 * prob_ge2_scaled ^ 2)
              
              G12 <- G12 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
                            alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
              
              #lambda
	              G2 <- (1 - z) * (y / lambda + ((-1 / alpha-1) * alpha * lambda *
	              (alpha * lambda + 1) ^ (-1 / alpha - 2)) /
	              prob_ge2 +
	              (alpha * (-y - 1 / alpha)) / (alpha * lambda + 1))
              
              G22 <- (1 - z) * ((alpha * lambda + 1) ^ (2 / alpha) * 
              (alpha ^ 3 * lambda ^ 4 + (2 * alpha ^ 2 - 2 * alpha ^ 3 * y) * lambda ^ 3 +
              (alpha - 5 * alpha ^ 2 * y) * lambda ^ 2 - 4 * alpha * y * lambda - y) + 
              (alpha * lambda + 1) ^ (1 / alpha) * ((alpha - alpha ^ 3) * lambda ^ 4 +
              ((4 * alpha ^ 3 + 4 * alpha ^ 2) * y - 4 * alpha ^ 2 - alpha + 1) * lambda ^ 3 +
              ((10 * alpha ^ 2 + 6 * alpha) * y - 3 * alpha - 1) * lambda ^ 2 + 
              (8 * alpha + 2) * y * lambda + 2 * y) + 
	              ((-2 * alpha ^ 3 - 4 * alpha ^ 2 - 2 * alpha) * y + 2 * alpha ^ 2 + 2 * alpha) * lambda ^ 3 +
	              ((-5 * alpha ^ 2 - 6 * alpha - 1) * y + 2 * alpha + 1) * lambda ^ 2 +
	              (-4 * alpha - 2) * y * lambda - y) / (lambda ^ 2 * (alpha * lambda + 1) ^ 2 *
	              prob_ge2_scaled ^ 2)
              
              G22 <- G22 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 + 
                      G2 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 2)
              
              
              predNumbers <- cumsum(predNumbers)
              res[(predNumbers[2]+1):predNumbers[3], (predNumbers[2]+1):predNumbers[3]] <- #PI
                t(as.data.frame(X[(nrow(X) * 2 / 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]]) *
                    G00 * weight) %*% X[(nrow(X) * 2 / 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]]
              
              res[(predNumbers[1]+1):predNumbers[2], (predNumbers[2]+1):predNumbers[3]] <- 0
              
              res[(predNumbers[2]+1):predNumbers[3], (predNumbers[1]+1):predNumbers[2]] <- 0
              
              res[1:predNumbers[1], (predNumbers[2]+1):predNumbers[3]] <- 0
              
              res[(predNumbers[2]+1):predNumbers[3], 1:predNumbers[1]] <- 0
              
              res[(predNumbers[1]+1):predNumbers[2], (predNumbers[1]+1):predNumbers[2]] <- #alpha
                t(as.data.frame(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]]) *
                    G11 * weight) %*% X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]]
              
              res[1:predNumbers[1], 1:predNumbers[1]] <-  # lambda
                t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                    G22 * weight) %*% X[1:(nrow(X) / 3), 1:predNumbers[1]]
              
              res[1:predNumbers[1], (predNumbers[1]+1):predNumbers[2]] <- #alpha lambda
                t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                    G12 * weight) %*% as.matrix(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]])
              
              res[(predNumbers[1]+1):predNumbers[2], 1:predNumbers[1]] <- #alpha lambda
                t(t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                      G12 * weight) %*% as.matrix(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]]))
              
              res
            }
    )
  }
  
  validmu <- function(mu) {
    all(is.finite(mu)) && all(0 < mu)
  }
  
  devResids <- function (y, eta, wt, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    mu <- mu.eta(eta = eta)
    
    logLikFit <- (
      (y == 1) * log(PI) + (y>1) * log(1 - PI) + (y>1) *
      (lgamma(y + 1 / alpha) - lgamma(1 / alpha) - lgamma(y + 1) -
      (y + 1 / alpha) * log(1 + lambda * alpha) + y * log(lambda * alpha) -
      logProbGe2(lambda, alpha))
    )
    
    yUnq <- unique(y)
    
    if (any(yUnq > 77)) {
      warning("Currently numerical deviance is unreliable for counts greater than 78.")
    }
    
    findL <- function(yNow) {
      stats::optim(
        par = if(TRUE) c(0, log(yNow), -15 + 2 * (yNow %in% 3:5) - 2 * (yNow %in% 6:9)) else c(0, log(yNow), -6),
        fn = function(x) {
          s <- x[1]
          l <- exp(x[2])
          a <- exp(x[3])
          
          sum(c(((l-l*(a*l+1)^(-1/a-1))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)-yNow) ,# s der
                s*((-(a*l+1)^(-1/a-1)-(-1/a-1)*a*l*(a*l+1)^(-1/a-2)+1)/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+(a*(-yNow-1/a))/(a*l+1)+yNow/l,# lambda der
                s*(-((l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1))-(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)+log(l*a+1)/a^2+(l*(-1/a-yNow))/(l*a+1)+yNow/a-digamma(1/a+yNow)/a^2+digamma(1/a)/a^2) ^ 2)#alpha der
        },
        gr = function(x) {
          s <- x[1]
          l <- exp(x[2])
          a <- exp(x[3])
          d12.21 <- ((a*l+1)^(2/a)*(a^2*l^2+2*a*l+1)+(a*l+1)^(1/a)*((-a^2-2*a-1)*l^2-2*a*l-2)+1)/((a*l+1)^(1/a+1)+(-a-1)*l-1)^2
          d13.31 <- (l^2*((l*a+1)^(1/a)*(l*a^2+(l+1)*a+1)*log(l*a+1)+(l*a+1)^(1/a)*((1-2*l)*a^2-l*a)-a^2))/(a^2*((l*a+1)^(1/a+1)-l*a-l-1)^2)
          d22.22 <- s*((-2*(-1/a-1)*a*(a*l+1)^(-1/a-2)-(-1/a-2)*(-1/a-1)*a^2*l*(a*l+1)^(-1/a-3))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)*a*(a*l+1)^(-1/a-2)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2+((-1/a-2)*(-1/a-1)*a^2*l*(a*l+1)^(-1/a-3)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2+(2*(-1/a-1)*a*l*(a*l+1)^(-1/a-2)*(-(a*l+1)^(-1/a-1)-(-1/a-1)*a*l*(a*l+1)^(-1/a-2)+1))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2+(2*(-1/a-1)^2*a^2*l^2*(a*l+1)^(-2/a-4)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^3)+((-1/a-1)*a*(a*l+1)^(-1/a-2))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-2)*(-1/a-1)*a^2*l*(a*l+1)^(-1/a-3))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)^2*a^2*l^2*(a*l+1)^(-2/a-4))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2-(a^2*(-yNow-1/a))/(a*l+1)^2-yNow/l^2
          d23.32 <- (((l*a+1)^(2/a)*(l^5*s*a^5+(5*l^4*s-l^4)*a^4+((-l^5+2*l^4+9*l^3)*s-l^4-3*l^3)*a^3+((-3*l^4+6*l^3+7*l^2)*s-3*l^3-3*l^2)*a^2+((-3*l^3+6*l^2+2*l)*s-3*l^2-l)*a+(2*l-l^2)*s-l)+(l*a+1)^(1/a)*(-l^5*s*a^5+((-3*l^5-5*l^4)*s+l^4)*a^4+((-3*l^5-10*l^4-9*l^3)*s+2*l^4+3*l^3)*a^3+((-l^5-7*l^4-13*l^3-7*l^2)*s+l^4+5*l^3+3*l^2)*a^2+((-2*l^4-5*l^3-8*l^2-2*l)*s+2*l^3+4*l^2+l)*a+(-l^3-l^2-2*l)*s+l^2+l))*log(l*a+1)+(l*a+1)^(2/a)*((3*l^3*yNow-l^5*s-2*l^4)*a^5+((3*l^3+9*l^2)*yNow+(2*l^5-8*l^4+2*l^3)*s-8*l^3)*a^4+((6*l^2+9*l)*yNow+(l^5+2*l^4-13*l^3+4*l^2)*s+l^4-10*l^2)*a^3+((3*l+3)*yNow+(2*l^4-2*l^3-6*l^2+2*l)*s+2*l^3-4*l)*a^2+((l^3-2*l^2)*s+l^2)*a)+(l*a+1)^(3/a)*((l^4-l^3*yNow)*a^5+(3*l^3-3*l^2*yNow)*a^4+(3*l^2-3*l*yNow)*a^3+(l-yNow)*a^2)+(l*a+1)^(1/a)*((-3*l^3*yNow+l^5*s+l^4)*a^5+((-6*l^3-9*l^2)*yNow+(3*l^5+8*l^4-4*l^3)*s+7*l^3)*a^4+((-3*l^3-12*l^2-9*l)*yNow+(3*l^5+7*l^4+13*l^3-8*l^2)*s-2*l^4+3*l^3+11*l^2)*a^3+((-3*l^2-6*l-3)*yNow+(l^5+4*l^4+6*l^3+6*l^2-4*l)*s-l^4-3*l^3+3*l^2+5*l)*a^2+((l^4+l^3+2*l^2)*s-l^3-l^2)*a)+l^3*yNow*a^5+((3*l^3+3*l^2)*yNow+2*l^3*s-2*l^3)*a^4+((3*l^3+6*l^2+3*l)*yNow+4*l^2*s-3*l^3-4*l^2)*a^3+((l^3+3*l^2+3*l+1)*yNow+2*l*s-l^3-3*l^2-2*l)*a^2)/(a^2*(l*a+1)^2*((l*a+1)^(1/a+1)-l*a-l-1)^3)
          d33.33 <- s*((2*(l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))^2)/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^3-((l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))^2/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))^2-(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))+l^2/(a*(l*a+1)^2))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-1))/(l*a+1)^2)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))^2)/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)+(2*l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-1))/(l*a+1)^2))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1))+(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))^2/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(-(log(l*a+1)/a^2-l/(a*(l*a+1)))^2/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))^2-(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))+l^2/(a*(l*a+1)^2))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-1))/(l*a+1)^2))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-yNow))/(l*a+1)^2-yNow/a^2+(2*digamma(1/a+yNow))/a^3-(2*digamma(1/a))/a^3+trigamma(1/a+yNow)/a^4-trigamma(1/a)/a^4
          
          f2 <- 2*c(((l-l*(a*l+1)^(-1/a-1))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)-yNow) ,# s der
                    s*((-(a*l+1)^(-1/a-1)-(-1/a-1)*a*l*(a*l+1)^(-1/a-2)+1)/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+(a*(-yNow-1/a))/(a*l+1)+yNow/l,# lambda der
                    s*(-((l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1))-(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)+log(l*a+1)/a^2+(l*(-1/a-yNow))/(l*a+1)+yNow/a-digamma(1/a+yNow)/a^2+digamma(1/a)/a^2)
          
          c(sum(f2 * c(0, d12.21, d13.31)),
            sum(f2 * c(d12.21, d22.22, d23.32)),
            sum(f2 * c(d13.31, d23.32, d33.33)))
        },
        method = "CG",
        control = list(maxit = 50000, abstol = .Machine$double.eps)
      )
    }
    fff <- function(yNow) {
      if (yNow < 3) return(0)
      
      xx <- findL(yNow)
      a <- exp(xx$par[3])
      l <- exp(xx$par[2])
      
      lgamma(yNow+1/a)-lgamma(1/a)-lgamma(yNow+1)-(yNow+1/a)*log(1+l*a)+yNow*log(l*a)-logProbGe2(l, a)
    }
    
    suppressWarnings(logLikIdeal <- sapply(yUnq, FUN = fff))
    names(logLikIdeal) <- yUnq
    
    logLikIdeal <- sapply(y, FUN = function(x) {
      logLikIdeal[names(logLikIdeal) == x]
    })
    diff <- logLikIdeal - logLikFit
    
    if (any(logLikFit > 0)) {
      warning("Dispersion parameter values are on the boundary of parameter space. Deviance residuals will be assigned 0 on these observations.")
      diff[logLikFit > 0]   <- 0
    } else if (any(diff < 0)) {
      warning("Numerical deviance finder found worse saturated likelihood than fitted model. Expect NA's in deviance/deviance residuals.")
    }
    
    diff[diff < 0] <- 0
    
    sign(y - mu) * sqrt(2 * wt * diff)
  }
  
  pointEst <- function (pw, eta, contr = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    P1 <- nbP1(lambda, alpha)
    N <- pw * (1 - P1) / probGe2(lambda, alpha)
    if(!contr) {
      N <- sum(N)
    }
    N
  }
  
  popVar <- function (pw, eta, cov, Xvlm, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    P0 <- nbP0(lambda, alpha)
    P1 <- nbP1(lambda, alpha)
    prob_ge2 <- probGe2(lambda, alpha)
    prob_ge2_scaled <- probGe2Scaled(lambda, alpha)
    
    bigTheta0 <- pw * 0 # w.r to PI
    bigTheta1 <- pw  *  alphaLink(eta[, 2], inverse = TRUE, deriv = 1) *
      ((lambda * alpha + 1) ^ (1 / alpha) * (lambda ^ 2 * alpha ^ 2 + 2 * lambda * alpha + 1) * 
      log(lambda * alpha + 1) + (lambda * alpha + 1) ^ (1 / alpha) * 
      (-lambda ^ 2 * alpha ^ 2 - lambda * alpha) - lambda ^ 2 * alpha ^ 2) / 
      (alpha ^ 2 * prob_ge2_scaled ^ 2)# w.r to alpha
    bigTheta2 <- -pw * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * 
      ((alpha * lambda + 1) ^ (1 / alpha + 1) - 1) /
      prob_ge2_scaled ^ 2# w.r to lambda
    
    bigTheta <- t(c(bigTheta2, bigTheta1, bigTheta0) %*% Xvlm)
    
    f1 <-  t(bigTheta) %*% as.matrix(cov) %*% bigTheta
    f2 <-  sum(pw * (1 - P1) * P0 / prob_ge2 ^ 2)
    
    f1 + f2
  }
  
  dFun <- function (x, eta, type = c("trunc", "nontrunc")) {
    if (missing(type)) type <- "trunc"
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-  piLink(eta[, 3], inverse = TRUE)
    prob_ge2 <- probGe2(lambda, alpha)
    
    switch (type,
      "trunc" = {
	        as.numeric(x == 1) * PI + as.numeric(x > 0) *
	        (1 - PI) * stats::dnbinom(x = x, mu = lambda, size = 1 / alpha) /
	        prob_ge2
      },
      "nontrunc" = {
        stats::dnbinom(x = x, mu = lambda, size = 1 / alpha) / 
        (1 - lambda * (1 + alpha * lambda) ^ (- 1 / alpha - 1)) *
        (as.numeric(x == 0) + as.numeric(x > 1) * (1 - PI)) + 
        as.numeric(x == 1) * PI * (1 - (1 + alpha * lambda) ^ (-1 / alpha) / 
        (1 - lambda * (1 + alpha * lambda) ^ (- 1 / alpha - 1)))
      }
    )
  }
  
  simulate <- function(n, eta, lower = 0, upper = Inf) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <- alphaLink(eta[, 2], inverse = TRUE)
    PI     <- piLink(eta[, 3], inverse = TRUE)
    P0 <- nbP0(lambda, alpha)
    P1 <- nbP1(lambda, alpha)
    CDF <- function(x) {
      ifelse(x == Inf, 1, 
      ifelse(x < 0, 0, 
      ifelse(x < 1, P0 / (1 - P1), 
      P0 / (1 - P1) + PI * (1 - P0 / (1 - P1)) + 
      (1 - PI) * (stats::pnbinom(x, mu = lambda, size = 1 / alpha) - P1 - P0) / (1 - P1))))
    }
    lb <- CDF(lower)
    ub <- CDF(upper)
    p_u <- stats::runif(n, lb, ub)
    sims <- rep(0, n)
    cond <- CDF(sims) < p_u
    while (any(cond)) {
      sims[cond] <- sims[cond] + 1
      cond <- CDF(sims) < p_u
    }
    sims
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      init <- log(abs((observed / weighted.mean(observed, priorWeights) - 1) / observed) + .1)
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12)),
        family$links[[2]](ifelse(init < -.5, .1, init + .55)),
        family$links[[3]](weighted.mean(observed == 1, priorWeights) * (.5 + .5 * (observed == 1)) + .01)
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights)),
        family$links[[2]](abs((cov.wt(cbind(observed, observed), wt = priorWeights, method = "ML")$cov[1,1] / weighted.mean(observed, priorWeights) - 1) / weighted.mean(observed, priorWeights)) + .1),
        family$links[[3]](weighted.mean(observed == 1, priorWeights) + .01)
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
      if ("(Intercept):alpha" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[2], rep(0, attr(Xvlm, "hwm")[2] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[2] / attr(Xvlm, "hwm")[2], attr(Xvlm, "hwm")[2]))
      }
      if ("(Intercept):pi" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[3], rep(0, attr(Xvlm, "hwm")[3] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[3] / attr(Xvlm, "hwm")[3], attr(Xvlm, "hwm")[3]))
      }
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "ztHurdlenegbin",
      etaNames  = c("lambda", "alpha", "pi"),
      simulate  = simulate,
      getStart  = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

#' \loadmathjax
#' @title Updating population size estimation results
#'
#' @description A function that applies all post-hoc procedures that were taken
#' (such as heteroscedastic consistent covariance matrix estimation or bias
#' reduction) to population size estimation and standard error estimation.
#' 
#' @param object object for which update of population size estimation results will be done.
#' @param newdata optional \code{data.frame} with new data for pop size estimation.
#' @param cov an updated covariance matrix estimate.
#' @param coef optional vector of coefficients of regression on which to base 
#' population size estimation. If missing it is set to \code{coef(object)}.
#' @param weights optional vector of weights to use in population size estimation. 
#' @param control similar to \code{controlPopVar} in [estimatePopsize()].
#' If missing set to controls provided on call to \code{object}.
#' @param popVar similar to \code{popVar} in [estimatePopsize()].
#' If missing set to \code{"analytic"}.
#' @param offset offset argument for new data
#' @param weightsAsCounts for \code{singleRStaticCountData} method used to specify
#' whether weights should be treated as number of occurrences for rows in data
#' @param ... additional optional arguments, currently not used in \code{singleRStaticCountData} class method.
#'
#' @details
#' Any non specified arguments will be inferred from the \code{object}
#' 
#'
#' @return An object of class \code{popSizeEstResults} containing updated 
#' population size estimation results.
#' 
#' @examples
#' # Create simple model
#' Model <- estimatePopsize(
#'   formula = capture ~ nation + gender, 
#'   data = netherlandsimmigrant, 
#'   model = ztpoisson, 
#'   method = "IRLS"
#' )
#' # Apply heteroscedasticity consistent covariance matrix estimation
#' require(sandwich)
#' cov <- vcovHC(Model, type = "HC3")
#' summary(Model, cov = cov,
#' popSizeEst = redoPopEstimation(Model, cov = cov))
#' # Compare to results with usual covariance matrix estimation
#' summary(Model)
#' 
#' ## get confidence interval with larger significance level
#' redoPopEstimation(Model, control = controlPopVar(alpha = .000001))
#' @export
redoPopEstimation <- function(object, newdata, ...) {
  UseMethod("redoPopEstimation")
}

#' @method redoPopEstimation singleRStaticCountData
#' @rdname redoPopEstimation
#' @exportS3Method
redoPopEstimation.singleRStaticCountData <- function(object, 
                                                     newdata, 
                                                     cov, 
                                                     weights,
                                                     coef,
                                                     control,
                                                     popVar,
                                                     offset,
                                                     weightsAsCounts,
                                                     ...) {
  if (missing(cov)) {
    cov <- vcov
  }
  
  if (is.character(cov)) {
    cov <- get(cov, mode = "function", envir = parent.frame())
  }
  if (is.function(cov)) {
    cov <- cov(object, ...)
  }
  
  if (missing(newdata)) {
    Xvlm <- model.matrix(object, "vlm")
    
    pw <- if (missing(weights))
      object$priorWeights
    else weights
    
    offset <- if (missing(offset))
      object$offset
    else offset
    
    etaNew <- if (missing(coef))
      object$linearPredictors
    else matrix(as.matrix(Xvlm) %*% coef, ncol = length(family(object)$etaNames))
    
    if (missing(control))
      control <- object$populationSize$control
    
    if (is.null(control$bootstrapFitcontrol)) {
      control$bootstrapFitcontrol <- object$control$controlMethod
    }
    
    Y <- object$y
    X <- model.matrix(object)
    MM <- model.frame(object, ...)
    nn <- nobs(object)
    
  } else {
    if (missing(control))
      control <- object$populationSize$control
    
    if (is.null(control$bootstrapFitcontrol)) {
      control$bootstrapFitcontrol <- object$control$controlMethod
    }
    
    MM <- model.frame(object, data = newdata, ...)
    X <- model.matrix(object, data = newdata)
    Xvlm <- model.matrix(object, type = "vlm", data = newdata)
    
    Y <- model.response(MM)
    
    nn <- length(Y)
    
    pw <- if (missing(weights))
      rep(1, nn)
    else weights
    
    offset <- if (missing(offset))
      matrix(0, nrow = length(pw), ncol = length(family(object)$etaNames))
    else offset
    
    coef <- if (missing(coef)) stats::coef(object)
    
    etaNew <- matrix(
      as.matrix(Xvlm) %*% coef, 
      ncol = length(family(object)$etaNames),
      dimnames = list(
        rownames(Xvlm),
        family(object)$etaNames
      )
    ) + offset
  }
  
  singleRcaptureinternalpopulationEstimate(
    y = Y,
    formulas = object$formula,
    X = X,
    grad = object$model$makeMinusLogLike(
      y = Y,
      X = Xvlm,
      weight = pw, 
      deriv = 1
    ),
    hessian = object$model$makeMinusLogLike(
      y = Y,
      X = Xvlm, 
      weight = pw, 
      deriv = 2
    ),
    popVar = if (missing(popVar)) 
      "analytic" 
    else popVar,
    weights = pw,
    eta = etaNew,
    family = family(object),
    beta = if (missing(coef))
      stats::coef(object)
    else coef,
    control = if (missing(control))
      object$populationSize$control
    else control,
    Xvlm = Xvlm,
    W = if (isTRUE(object$call$method == "IRLS") & missing(newdata)) 
      object$weights 
    else 
      family(object)$Wfun(prior = pw, eta = etaNew, y = Y),
    sizeObserved = nn,
    modelFrame = MM,
    cov = cov,
    offset = offset,
    weightsFlag = if (missing(weightsAsCounts)) object$control$controlModel$weightsAsCounts else weightsAsCounts
  )
}

#' @import mathjaxr
NULL
#' @title Single-source capture-recapture models
#' @author Piotr Chlebicki, Maciej Beręsewicz
#'
#' \loadmathjax
#' @description \code{estimatePopsize} first fits appropriate (v)glm model and 
#' then estimates full (observed and unobserved) population size.
#' In this types of models it is assumed that the response vector 
#' (i.e. the dependent variable) corresponds to the number of times a given unit 
#' was observed in the source.
#' Population size is then usually estimated by Horvitz-Thompson type estimator:
#' 
#' \mjsdeqn{\hat{N} = \sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}(Y_{k}>0)} = 
#' \sum_{k=1}^{N_{obs}}\frac{1}{1-\mathbb{P}(Y_{k}=0)}}
#'
#' where \mjseqn{I_{k}=I_{Y_{k} > 0}} are indicator 
#' variables, with value 1 if kth unit was observed at least once and 0 otherwise.
#'
#' @param data a data frame or object coercible to data.frame class containing 
#' data for the regression and population size estimation.
#' @param formula a formula for the model to be fitted, only applied to the "main" 
#' linear predictor. Only single response models are available.
#' @param ratioReg deprecated placeholder for ratio regression. Use
#'   [ratioReg()] instead.
#' @param model a model for regression and population estimate full description in [singleRmodels()]. 
#' @param weights an optional object of prior weights used in fitting the model. 
#' Can be used to specify number of occurrences of rows in data see [controlModel()]
#' @param subset a logical vector indicating which observations should be used 
#' in regression and population size estimation. It will be evaluated on \code{data} argument provided on call.
#' @param naAction Not yet implemented.
#' @param method a method for fitting values currently supported: iteratively 
#' reweighted least squares (\code{IRLS}) and maximum likelihood (\code{optim}).
#' @param popVar a method of constructing confidence interval and estimating
#' the standard error either analytic or bootstrap. 
#' Bootstrap confidence interval type may be specified in 
#' \code{controlPopVar.} There is also the third possible value of \code{noEst} 
#' which skips the population size estimate all together.
#' @param controlMethod a list indicating parameters to use in fitting the model 
#' may be constructed with \code{singleRcapture::controlMethod} function. 
#' More information included in [controlMethod()].
#' @param controlModel a list indicating additional formulas for regression 
#' (like formula for inflation parameter/dispersion parameter) may be 
#' constructed with \code{singleRcapture::controlModel} function. 
#' More information will eventually be included in [controlModel()].
#' @param controlPopVar a list indicating parameters to use in estimating variance 
#' of population size estimation may be constructed with 
#' \code{singleRcapture::controlPopVar} function. 
#' More information included in [controlPopVar()].
#' @param modelFrame,x,y logical values indicating whether to return model matrix, 
#' dependent vector and model matrix as a part of output.
#' @param contrasts not yet implemented.
#' @param offset a matrix of offset values with the number of columns matching the
#' number of distribution parameters providing offset values to each of 
#' linear predictors.
#' @param ... additional optional arguments passed to other methods eg. 
#' \code{estimatePopsizeFit}.
#' 
#' @details The generalized linear model is characterized by equation
#' \mjsdeqn{\boldsymbol{\eta}=\boldsymbol{X}\boldsymbol{\beta}}
#' where \mjseqn{\boldsymbol{X}} is the (lm) model matrix. 
#' The vector generalized linear model is similarly characterized by equations
#' \mjsdeqn{\boldsymbol{\eta}_{k}=\boldsymbol{X}_{k}\boldsymbol{\beta}_{k}}
#' where \mjseqn{\boldsymbol{X}_{k}} is a (lm) model 
#' matrix constructed from appropriate formula 
#' (specified in \code{controlModel} parameter).
#' 
#' The \mjseqn{\boldsymbol{\eta}} is then a vector constructed as:
#' 
#' \mjsdeqn{\boldsymbol{\eta}=\begin{pmatrix}
#' \boldsymbol{\eta}_{1} \cr 
#' \boldsymbol{\eta}_{2} \cr
#' \dotso \cr
#' \boldsymbol{\eta}_{p}
#' \end{pmatrix}^{T}}
#' 
#' and in cases of models in our package the (vlm) model matrix 
#' is constructed as a block matrix:
#' 
#' \mjsdeqn{\boldsymbol{X}_{vlm}=
#' \begin{pmatrix}
#' \boldsymbol{X}_{1} & \boldsymbol{0} &\dotso &\boldsymbol{0}\cr
#' \boldsymbol{0} & \boldsymbol{X}_{2} &\dotso &\boldsymbol{0}\cr
#' \vdots & \vdots & \ddots & \vdots\cr
#' \boldsymbol{0} & \boldsymbol{0} &\dotso &\boldsymbol{X}_{p}
#' \end{pmatrix}}
#' 
#' this differs from convention in \code{VGAM} package (if we only consider our 
#' special cases of vglm models) but this is just a convention and does not 
#' affect the model, this convention is taken because it makes fitting with 
#' IRLS (explanation of algorithm in [estimatePopsizeFit()]) algorithm easier.
#' (If \code{constraints} matrixes in \code{vglm} match the ones we implicitly
#' use the \code{vglm} model matrix differs with respect to order of 
#' \code{kronecker} multiplication of \code{X} and \code{constraints}.)
#' In this package we use observed likelihood to fit regression models.
#' 
#' As mentioned above usually the population size estimation is done via:
#' \mjsdeqn{\hat{N} = \sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}(Y_{k}>0)} = 
#' \sum_{k=1}^{N_{obs}}\frac{1}{1-\mathbb{P}(Y_{k}=0)}}
#'
#' where \mjseqn{I_{k}=I_{Y_{k} > 0}} are indicator variables, 
#' with value 1 if kth unit was observed at least once and 0 otherwise.
#' The \mjseqn{\mathbb{P}(Y_{k}>0)} are estimated by maximum likelihood.
#' 
#' The following assumptions are usually present when using 
#' the method of estimation described above:
#' 1. The specified regression model is correct. This entails linear
#' relationship between independent variables and dependent ones and
#' dependent variable being generated by appropriate distribution.
#' 2. No unobserved heterogeneity. If this assumption is broken there
#' are some possible (admittedly imperfect) workarounds see details in 
#' [singleRmodels()].
#' 3. The population size is constant in relevant time frame.
#' 4. Depending on confidence interval construction (asymptotic) normality
#' of \mjseqn{\hat{N}} statistic is assumed.
#' 
#' There are two ways of estimating variance of estimate \mjseqn{\hat{N}},
#' the first being \code{"analytic"} usually done by application of 
#' law of total variance to \mjseqn{\hat{N}}:
#'
#' \mjsdeqn{\text{var}(\hat{N})=\mathbb{E}\left(\text{var}
#' \left(\hat{N}|I_{1},\dots,I_{n}\right)\right)+
#' \text{var}\left(\mathbb{E}(\hat{N}|I_{1},\dots,I_{n})\right)}
#' 
#' and then by \mjseqn{\delta} method to 
#' \mjseqn{\hat{N}|I_{1},\dots I_{N}}:
#' 
#' \mjsdeqn{\mathbb{E}\left(\text{var}
#' \left(\hat{N}|I_{1},\dots,I_{n}\right)\right)=
#' \left.\left(\frac{\partial(N|I_1,\dots,I_N)}{\partial\boldsymbol{\beta}}\right)^{T}
#' \text{cov}\left(\boldsymbol{\beta}\right)
#' \left(\frac{\partial(N|I_1,\dots,I_N)}{\partial\boldsymbol{\beta}}\right)
#' \right|_{\boldsymbol{\beta}=\hat{\boldsymbol{\beta}}}}
#' 
#' and the \mjseqn{\text{var}\left(\mathbb{E}(\hat{N}|I_{1},\dots,I_{n})\right)}
#' term may be derived analytically (if we assume independence of
#' observations) since \mjseqn{\hat{N}|I_{1},\dots,I_{n}}
#' is just a constant. 
#' 
#' In general this gives us:
#' \mjsdeqn{
#' \begin{aligned}
#' \text{var}\left(\mathbb{E}(\hat{N}|I_{1},\dots,I_{n})\right)&=
#' \text{var}\left(\sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}(Y_{k}>0)}\right)\cr
#' &=\sum_{k=1}^{N}\text{var}\left(\frac{I_{k}}{\mathbb{P}(Y_{k}>0)}\right)\cr
#' &=\sum_{k=1}^{N}\frac{1}{\mathbb{P}(Y_{k}>0)^{2}}\text{var}(I_{k})\cr
#' &=\sum_{k=1}^{N}\frac{1}{\mathbb{P}(Y_{k}>0)^{2}}\mathbb{P}(Y_{k}>0)(1-\mathbb{P}(Y_{k}>0))\cr
#' &=\sum_{k=1}^{N}\frac{1}{\mathbb{P}(Y_{k}>0)}(1-\mathbb{P}(Y_{k}>0))\cr
#' &\approx\sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}(Y_{k}>0)^{2}}(1-\mathbb{P}(Y_{k}>0))\cr
#' &=\sum_{k=1}^{N_{obs}}\frac{1-\mathbb{P}(Y_{k}>0)}{\mathbb{P}(Y_{k}>0)^{2}}
#' \end{aligned}
#' }
#' 
#' Where the approximation on 6th line appears because in 5th line we sum over 
#' all units, that includes unobserved units, since \mjseqn{I_{k}} are 
#' independent and \mjseqn{I_{k}\sim b(\mathbb{P}(Y_{k}>0))} the 6th line
#' is an unbiased estimator of the 5th line.
#' 
#' The other method for estimating variance is \code{"bootstrap"}, but since
#' \mjseqn{N_{obs}=\sum_{k=1}^{N}I_{k}} is also a random variable bootstrap 
#' will not be as simple as just drawing \mjseqn{N_{obs}} units from data 
#' with replacement and just computing \mjseqn{\hat{N}}.
#' 
#' Method described above is referred to in literature as \code{"nonparametric"}
#' bootstrap (see [controlPopVar()]), due to ignoring variability in observed
#' sample size it is likely to underestimate variance.
#' 
#' A more sophisticated bootstrap procedure may be described as follows:
#' 1. Compute the probability distribution as: 
#' \mjsdeqn{
#' \frac{\hat{\boldsymbol{f}}_{0}}{\hat{N}}, 
#' \frac{\boldsymbol{f}_{1}}{\hat{N}}, 
#' \dots,
#' \frac{\boldsymbol{f}_{\max{y}}}{\hat{N}}}
#' where \mjseqn{\boldsymbol{f}_{n}} denotes observed 
#' marginal frequency of units being observed exactly n times.
#' 2. Draw \mjseqn{\hat{N}} units from the distribution above 
#' (if \mjseqn{\hat{N}} is not an integer than draw 
#' \mjseqn{\lfloor\hat{N}\rfloor + b(\hat{N}-\lfloor\hat{N}\rfloor)}),
#' where \mjseqn{\lfloor\cdot\rfloor} is the floor function.
#' 3. Truncated units with \mjseqn{y=0}.
#' 4. If there are covariates draw them from original data with replacement from 
#' uniform distribution. For example if unit drawn to new data has 
#' \mjseqn{y=2} choose one of covariate vectors from original data that 
#' was associated with unit for which was observed 2 times.
#' 5. Regress \mjseqn{\boldsymbol{y}_{new}} on \mjseqn{\boldsymbol{X}_{vlm new}} 
#' and obtain \mjseqn{\hat{\boldsymbol{\beta}}_{new}}, with starting point 
#' \mjseqn{\hat{\boldsymbol{\beta}}} to make it slightly faster, use them 
#' to compute \mjseqn{\hat{N}_{new}}.
#' 6. Repeat 2-5 until there are at least \code{B} statistics are obtained.
#' 7. Compute confidence intervals based on \code{alpha} and \code{confType} 
#' specified in [controlPopVar()].
#' 
#' To do step 1 in procedure above it is convenient to first draw binary vector of length
#' \mjseqn{\lfloor\hat{N}\rfloor + b(\hat{N}-\lfloor\hat{N}\rfloor)} with probability 
#' \mjseqn{1-\frac{\hat{\boldsymbol{f}}_{0}}{\hat{N}}}, sum elements in that vector to
#' determine the sample size and then draw sample of this size uniformly from the data.
#' 
#' This procedure is known in literature as \code{"semiparametric"} bootstrap
#' it is necessary to assume that the have a correct estimate 
#' \mjseqn{\hat{N}} in order to use this type of bootstrap.
#' 
#' Lastly there is \code{"parametric"} bootstrap where we assume that the 
#' probabilistic model used to obtain \mjseqn{\hat{N}} is correct the 
#' bootstrap procedure may then be described as:
#' 
#' 1. Draw \mjseqn{\lfloor\hat{N}\rfloor + b(\hat{N}-\lfloor\hat{N}\rfloor)} 
#' covariate information vectors with replacement from data according to 
#' probability distribution that is proportional to: \mjseqn{N_{k}},
#' where \mjseqn{N_{k}} is the contribution of kth unit i.e. 
#' \mjseqn{\dfrac{1}{\mathbb{P}(Y_{k}>0)}}.
#' 2. Determine \mjseqn{\boldsymbol{\eta}} matrix using estimate 
#' \mjseqn{\hat{\boldsymbol{\beta}}}.
#' 3. Generate \mjseqn{\boldsymbol{y}} (dependent variable) 
#' vector using \mjseqn{\boldsymbol{\eta}} and 
#' probability mass function associated with chosen model.
#' 4. Truncated units with \mjseqn{y=0} and construct 
#' \mjseqn{\boldsymbol{y}_{new}} and \mjseqn{\boldsymbol{X}_{vlm new}}.
#' 5. Regress \mjseqn{\boldsymbol{y}_{new}} on \mjseqn{\boldsymbol{X}_{vlm new}} 
#' and obtain \mjseqn{\hat{\boldsymbol{\beta}}_{new}}
#' use them to compute \mjseqn{\hat{N}_{new}}.
#' 6. Repeat 1-5 until there are at least \code{B} statistics are obtained.
#' 7. Compute confidence intervals based on \code{alpha} and \code{confType}
#' specified in [controlPopVar()]
#' 
#' It is also worth noting that in the \code{"analytic"} method \code{estimatePopsize}
#' only uses "standard" covariance matrix estimation. It is possible that improper
#' covariance matrix estimate is the only part of estimation that has its assumptions
#' violated. In such cases post-hoc procedures are implemented in this package
#' to address this issue.
#' 
#' Lastly confidence intervals for \mjseqn{\hat{N}} are computed 
#' (in analytic case) either by assuming that it follows a normal distribution 
#' or that variable \mjseqn{\ln(N-\hat{N})}
#' follows a normal distribution. 
#' 
#' These estimates may be found using either \code{summary.singleRStaticCountData}
#' method or \code{popSizeEst.singleRStaticCountData} function. They're labelled as 
#' \code{normal} and \code{logNormal} respectively.
#'
#' @references General single source capture recapture literature:
#' 
#' Zelterman, Daniel (1988). ‘Robust estimation in truncated discrete distributions
#' with application to capture-recapture experiments’. In: Journal of statistical
#' planning and inference 18.2, pp. 225–237.
#' 
#' Heijden, Peter GM van der et al. (2003). ‘Point and interval estimation of the
#' population size using the truncated Poisson regression model’. 
#' In: Statistical Modelling 3.4, pp. 305–322. doi: 10.1191/1471082X03st057oa.
#' 
#' Cruyff, Maarten J. L. F. and Peter G. M. van der Heijden (2008). ‘Point and 
#' Interval Estimation of the Population Size Using a Zero-Truncated Negative 
#' Binomial Regression Model’. In: Biometrical Journal 50.6, pp. 1035–1050. 
#' doi: 10.1002/bimj.200810455
#' 
#' Böhning, Dankmar and Peter G. M. van der Heijden (2009). ‘A covariate adjustment 
#' for zero-truncated approaches to estimating the size of hidden and 
#' elusive populations’. In: The Annals of Applied Statistics 3.2, pp. 595–610. 
#' doi: 10.1214/08-AOAS214.
#' 
#' Böhning, Dankmar, Alberto Vidal-Diez et al. (2013). ‘A Generalization of 
#' Chao’s Estimator for Covariate Information’. In: Biometrics 69.4, pp. 1033– 
#' 1042. doi: 10.1111/biom.12082
#' 
#' Böhning, Dankmar and Peter G. M. van der Heijden (2019). ‘The identity of the 
#' zero-truncated, one-inflated likelihood and the zero-one-truncated likelihood 
#' for general count densities with an application to drink-driving in Britain’. 
#' In: The Annals of Applied Statistics 13.2, pp. 1198–1211. 
#' doi: 10.1214/18-AOAS1232.
#' 
#' Navaratna WC, Del Rio Vilas VJ, Böhning D. Extending Zelterman's approach for 
#' robust estimation of population size to zero-truncated clustered Data. 
#' Biom J. 2008 Aug;50(4):584-96. doi: 10.1002/bimj.200710441.
#' 
#' Böhning D. On the equivalence of one-inflated zero-truncated and zero-truncated 
#' one-inflated count data likelihoods. Biom J. 2022 Aug 15. doi: 10.1002/bimj.202100343.
#' 
#' Böhning, D., Friedl, H. Population size estimation based upon zero-truncated, 
#' one-inflated and sparse count data. Stat Methods Appl 30, 1197–1217 (2021). 
#' doi: 10.1007/s10260-021-00556-8
#' 
#' Bootstrap:
#' 
#' 
#' Zwane, PGM EN and Van der Heijden, Implementing the parametric bootstrap in capture-recapture 
#' models with continuous covariates 2003 Statistics & probability letters 65.2 pp 121-125
#' 
#' Norris, James L and Pollock, Kenneth H Including model uncertainty in estimating variances 
#' in multiple capture studies 1996 in Environmental and Ecological Statistics 3.3 pp 235-244
#' 
#' Vector generalized linear models: 
#' 
#' 
#' Yee, T. W. (2015). Vector Generalized Linear and Additive Models: 
#' With an Implementation in R. New York, USA: Springer. ISBN 978-1-4939-2817-0.
#'
#' @returns Returns an object of class \code{c("singleRStaticCountData", "singleR", "glm", "lm")}
#' with type \code{list} containing:\cr
#' \itemize{
#'  \item{\code{y} -- Vector of dependent variable if specified at function call.}
#'  \item{\code{X} -- Model matrix if specified at function call.}
#'  \item{\code{formula} -- A list with formula provided on call and additional formulas specified in \code{controlModel}.}
#'  \item{\code{call} -- Call matching original input.}
#'  \item{\code{coefficients} -- A vector of fitted coefficients of regression.}
#'  \item{\code{control} -- A list of control parameters for \code{controlMethod} and \code{controlModel}, \code{controlPopVar} is included in populationSize.}
#'  \item{\code{model} -- Model which estimation of population size and regression was built, object of class family.}
#'  \item{\code{deviance} -- Deviance for the model.}
#'  \item{\code{priorWeights} -- Prior weight provided on call.}
#'  \item{\code{weights} -- If \code{IRLS} method of estimation was chosen weights returned by \code{IRLS}, otherwise same as \code{priorWeights}.}
#'  \item{\code{residuals} -- Vector of raw residuals.}
#'  \item{\code{logL} -- Logarithm likelihood obtained at final iteration.}
#'  \item{\code{iter} -- Numbers of iterations performed in fitting or if \code{stats::optim} was used number of call to loglikelihood function.}
#'  \item{\code{dfResiduals} -- Residual degrees of freedom.}
#'  \item{\code{dfNull} -- Null degrees of freedom.}
#'  \item{\code{fittValues} -- Data frame of fitted values for both mu (the expected value) and lambda (Poisson parameter).}
#'  \item{\code{populationSize} -- A list containing information of population size estimate.}
#'  \item{\code{modelFrame} -- Model frame if specified at call.}
#'  \item{\code{linearPredictors} -- Vector of fitted linear predictors.}
#'  \item{\code{sizeObserved} -- Number of observations in original model frame.}
#'  \item{\code{terms} -- terms attribute of model frame used.}
#'  \item{\code{contrasts} -- contrasts specified in function call.}
#'  \item{\code{naAction} -- naAction used.}
#'  \item{\code{which} -- list indicating which observations were used in regression/population size estimation.}
#'  \item{\code{fittingLog} -- log of fitting information for \code{"IRLS"} fitting if specified in \code{controlMethod}.}
#' }
#' 
#' @seealso 
#' [stats::glm()] -- For more information on generalized linear models.
#' 
#' [stats::optim()] -- For more information on \code{optim} function used in 
#' \code{optim} method of fitting regression.
#' 
#' [controlMethod()] -- For control parameters related to regression.
#' 
#' [controlPopVar()] -- For control parameters related to population size estimation.
#' 
#' [controlModel()] -- For control parameters related to model specification.
#' 
#' [estimatePopsizeFit()] -- For more information on fitting procedure in
#' \code{estimate_popsize}.
#' 
#' [popSizeEst()] [redoPopEstimation()] -- For extracting population size 
#' estimation results are applying post-hoc procedures.
#' 
#' [summary.singleRStaticCountData()] -- For summarizing important information about the
#' model and population size estimation results.
#' 
#' [marginalFreq()] -- For information on marginal frequencies and comparison
#' between observed and fitted quantities.
#' 
#' [VGAM::vglm()] -- For more information on vector generalized linear models.
#' 
#' [singleRmodels()] -- For description of various models.

#' @examples
#' \donttest{
#' # Model from 2003 publication 
#' # Point and interval estimation of the
#' # population size using the truncated Poisson regression mode
#' # Heijden, Peter GM van der et al. (2003)
#' model <- estimatePopsize(
#'    formula = capture ~ gender + age + nation, 
#'    data = netherlandsimmigrant, 
#'    model = ztpoisson
#' )
#' summary(model)
#' # Graphical presentation of model fit
#' plot(model, "rootogram")
#' # Statistical test
#' # see documentation for summary.singleRmargin
#' summary(marginalFreq(model), df = 1, "group")
#' 
#' # We currently support 2 methods of numerical fitting
#' # (generalized) IRLS algorithm and via stats::optim
#' # the latter one is faster when fitting negative binomial models 
#' # (and only then) due to IRLS having to numerically compute
#' # (expected) information matrixes, optim is also less reliable when
#' # using alphaFormula argument in controlModel
#' modelNegBin <- estimatePopsize(
#'     formula = TOTAL_SUB ~ ., 
#'     data = farmsubmission, 
#'     model = ztnegbin, 
#'     method = "optim"
#' )
#' summary(modelNegBin)
#' summary(marginalFreq(modelNegBin))
#' 
#' # More advanced call that specifies additional formula and shows
#' # in depth information about fitting procedure
#' pseudoHurdleModel <- estimatePopsize(
#'     formula = capture ~ nation + age,
#'     data = netherlandsimmigrant,
#'     model = Hurdleztgeom,
#'     method = "IRLS",
#'     controlMethod = controlMethod(verbose = 5),
#'     controlModel = controlModel(piFormula = ~ gender)
#' )
#' summary(pseudoHurdleModel)
#' # Assessing model fit
#' plot(pseudoHurdleModel, "rootogram")
#' summary(marginalFreq(pseudoHurdleModel), "group", df = 1)
#' 
#' 
#' # A advanced input with additional information for fitting procedure and
#' # additional formula specification and different link for inflation parameter.
#' Model <- estimatePopsize(
#'  formula = TOTAL_SUB ~ ., 
#'  data = farmsubmission, 
#'  model = oiztgeom(omegaLink = "cloglog"), 
#'  method = "IRLS", 
#'  controlMethod = controlMethod(
#'    stepsize = .85, 
#'    momentumFactor = 1.2,
#'    epsilon = 1e-10, 
#'    silent = TRUE
#'  ),
#'  controlModel = controlModel(omegaFormula = ~ C_TYPE + log_size)
#' )
#' summary(marginalFreq(Model), df = 18 - length(Model$coefficients))
#' summary(Model)
#' }
#' @export
estimatePopsize <- function(formula, 
                            ...) {
  # So that other packages such as singleRcaptureExtra
  # can implement custom methods
  UseMethod("estimatePopsize")
}
#' @rdname estimatePopsize
#' @importFrom stats model.frame model.matrix model.response lm predict
#' @export
estimatePopsize.default <- function(formula,
                                    data,
                                    model = c(
                                      "ztpoisson", "ztnegbin", "ztgeom",
                                      "zotpoisson", "ztoipoisson", "oiztpoisson",
                                      "ztHurdlepoisson", "Hurdleztpoisson", "zotnegbin",
                                      "ztoinegbin", "oiztnegbin", "ztHurdlenegbin",
                                      "Hurdleztnegbin", "zotgeom", "ztoigeom",
                                      "oiztgeom", "ztHurdlegeom", "Hurdleztgeom",
                                      "zelterman", "chao", "oichao"
                                    ),
                                    weights  = NULL,
                                    subset   = NULL,
                                    naAction = NULL,
                                    method   = c("optim", 
                                                 "IRLS"),
                                    popVar   = c("analytic",
                                                 "bootstrap",
                                                 "noEst"),
                                    controlMethod = NULL,
                                    controlModel  = NULL,
                                    controlPopVar = NULL,
                                    modelFrame    = TRUE,
                                    x             = FALSE,
                                    y             = TRUE,
                                    contrasts     = NULL,
                                    ratioReg      = FALSE,
                                    offset,
                                    ...) {
  if (missing(method)) method <- "IRLS"
  if (missing(popVar)) popVar <- "analytic"
  
  subset <- parse(text = deparse(substitute(subset)))
  
  if (!is.data.frame(data)) {
    data <- data.frame(data)
  }
  fullDataRows <- rownames(data)
  
  if (!is.logical(subset)) {subset <- eval(subset, data)}
  if (is.null(subset)) {subset <- TRUE}
  # subset is often in conflict with some common packages hence the explicit call
  data <- base::subset(data, subset = subset)
  
  family <- model
  if (is.character(family)) {
    family <- get(family, mode = "function", envir = parent.frame())
  }
  if (is.function(family)) {
    family <- family()
  }
  
  returnElements <- list(y, x, modelFrame)
  
  if (!ratioReg) {
    m1 <- controlPopVar
    m1 <- m1[sapply(m1, is.null) == FALSE]
    m2 <- controlPopVar(
      fittingMethod = match.arg(method), 
      bootstrapFitcontrol = controlMethod(
        epsilon = 1e-3, 
        maxiter = 20, 
        optimMethod = "Nelder-Mead", 
        silent = TRUE
      )
    )
    m2 <- m2[names(m2) %in% names(m1) == FALSE]
    controlPopVar <- append(m1, m2)
    
    m1 <- controlMethod
    m2 <- controlMethod(
      optimMethod = if (grepl(x = family$family, pattern = "negbin") || 
                        grepl(x = family$family, pattern = "^ztoi")) 
        "Nelder-Mead" else "BFGS",
      maxiter = if (isTRUE(method == "IRLS")) 100
                else 1000
    )
    m2 <- m2[names(m2) %in% names(m1) == FALSE]
    controlMethod <- append(m1, m2)
    
    m1 <- controlModel
    m2 <- controlModel()
    m2 <- m2[names(m2) %in% names(m1) == FALSE]
    controlModel <- append(m1, m2)
    
    formulas <- list(formula)
    
    if ("alpha" %in% family$etaNames) {
      formulas <- append(x = formulas, controlModel$alphaFormula)
    }
    if ("omega" %in% family$etaNames) {
      formulas <- append(x = formulas, controlModel$omegaFormula)
    }
    if ("pi" %in% family$etaNames) {
      formulas <- append(x = formulas, controlModel$piFormula)
    }
  
    combinedFromula <- singleRinternalMergeFormulas(formulas)
    
    modelFrame <- stats::model.frame(combinedFromula, 
                                     data,  
                                     ...)
    variables  <- stats::model.matrix(combinedFromula, 
                                      modelFrame, 
                                      contrasts = contrasts, 
                                      ...)
    
    terms     <- attr(modelFrame, "terms")
    contrasts <- attr(variables, "contrasts")
    observed  <- model.response(modelFrame)
    
    if (NCOL(observed) > 1) 
      stop("Single source capture-recapture models support only single dependent variable.")
  
    if (!is.null(weights)) {
      weights <- as.numeric(weights)
      modelFrameRows <- rownames(modelFrame)
      dataRows <- rownames(data)
      
      if (length(weights) == 1L) {
        priorWeights <- rep(weights, nrow(modelFrame))
      } else if (length(weights) == nrow(modelFrame)) {
        priorWeights <- weights
      } else if (length(weights) == nrow(data)) {
        alignedRows <- match(modelFrameRows, dataRows)
        if (anyNA(alignedRows)) {
          stop("Unable to align argument weights with rows used in fitting.")
        }
        priorWeights <- weights[alignedRows]
      } else if (length(weights) == length(fullDataRows)) {
        alignedRows <- match(modelFrameRows, fullDataRows)
        if (anyNA(alignedRows)) {
          stop("Unable to align argument weights with rows used in fitting.")
        }
        priorWeights <- weights[alignedRows]
      } else {
        stop(
          "Argument weights must have length 1, the number of rows used in fitting, ",
          "the number of rows in data after subsetting, or the number of rows in the original data."
        )
      }
    } else {
      priorWeights <- rep(1, nrow(modelFrame))
    }
    weights <- 1
    
    if (controlModel$weightsAsCounts) {
      sizeObserved <- sum(priorWeights)
    } else {
      sizeObserved <- nrow(modelFrame)
    }
    
    if (!all(observed > 0)) {
      stop("Error in function estimatePopsize, data contains zero-counts.")
    }
    
    Xvlm <- singleRinternalGetXvlmMatrix(
      X = modelFrame,
      formulas = formulas, 
      parNames = family$etaNames
    )
    
    if (missing(offset)) {
      offset <- matrix(0, nrow = NROW(modelFrame), ncol = length(family$etaNames))
    } else if (!is.matrix(offset)) {
      offset <- matrix(offset, nrow = NROW(modelFrame), ncol = length(family$etaNames))
    }
    
    colnames(offset) <- family$etaNames
    
    if (!is.null(controlMethod$coefStart) || !is.null(controlMethod$etaStart)) {
      if (method == "IRLS") {
        etaStart  <- controlMethod$etaStart
        coefStart <- controlMethod$coefStart
        if (is.null(etaStart)) {
          etaStart <- matrix(Xvlm %*% coefStart, ncol = length(family$etaNames))
        }
      } else if (method == "optim") {
        etaStart  <- controlMethod$etaStart
        coefStart <- controlMethod$coefStart
        if (is.null(coefStart)) {
          eval(family$getStart)
        }
      }
    }
    else {
      eval(family$getStart)
    }
    
    
    FITT <- estimatePopsizeFit(
      y            = observed,
      X            = Xvlm,
      family       = family,
      control      = controlMethod,
      method       = method,
      priorWeights = priorWeights,
      coefStart    = coefStart,
      etaStart     = etaStart,
      offset       = offset
    )
    
    coefficients        <- FITT$beta
    names(coefficients) <- colnames(Xvlm)
    iter                <- FITT$iter
    dfReduced           <- sizeObserved * length(family$etaNames) - length(coefficients)
    IRLSlog             <- FITT$logg
    
    
    logLike <- family$makeMinusLogLike(
      y      = observed, 
      X      = Xvlm,
      weight = priorWeights,
      offset = offset
    )
    
    grad <- family$makeMinusLogLike(
      y      = observed, 
      X      = Xvlm, 
      weight = priorWeights, 
      deriv  = 1,
      offset = offset
    )
    
    hessian <- family$makeMinusLogLike(
      y      = observed,
      X      = Xvlm,
      weight = priorWeights,
      deriv  = 2,
      offset = offset
    )
  
    eta           <- matrix(as.matrix(Xvlm) %*% coefficients, 
                            ncol = length(family$etaNames)) + offset
    
    colnames(eta) <- family$etaNames
    rownames(eta) <- rownames(variables)
    weights       <- FITT$weights
    
    if (family$family == "zelterman") {
      eta <- matrix(as.matrix(variables) %*% coefficients, ncol = 1) + offset
      colnames(eta) <- family$etaNames
      rownames(eta) <- rownames(variables)
    }
  
    fitt <- data.frame(
      family$mu.eta(eta = eta),
      family$mu.eta(eta = eta, type = "nontrunc")
    )
    colnames(fitt) <- c("truncated", "nontruncated")
    
    eig <- eigen(hessian(coefficients), symmetric = TRUE, only.values = TRUE)
    if (!all(sign(eig$values) == -1)) {
      warningMessage <- paste0(
        "The (analytically computed) hessian of the score function ",
        "is not negative define.\nNOTE: Second derivative test failing does not 
        necessarily mean that the maximum of score function that was found 
        numericaly is invalid since R^k is not a bounded space.\n",
        "Additionally in one inflated and hurdle models second ",
        "derivative test often fails even on valid arguments."
      )
      if (!isTRUE(controlMethod$silent)) {
        warning(warningMessage)
      }
      if (controlPopVar$covType == "observedInform") {
        if (!isTRUE(controlMethod$silent)) 
          warning(paste0(
            "Switching from observed information matrix to Fisher information",
            " matrix because hessian of log-likelihood is not negative define."
          ))
        controlPopVar$covType <- "Fisher"
      }
    }
    
    nullDeviance <- as.numeric(NULL)
    LOG          <- suppressWarnings(-logLike(coefficients))
    resRes       <- priorWeights * (observed - fitt)
    
    if (family$family %in% c("zelterman", "chao")) {
      resRes <- resRes - priorWeights
    } else if (family$family == "oichao") {
      resRes <- resRes - 2 * priorWeights
    }
  
    deviance <- sum(family$devResids(y   = observed, 
                                     wt  = priorWeights,
                                     eta = eta + offset) ^ 2)
    
    if (popVar == "noEst") {
      Pop <- NULL
    } else {
      POP <- singleRcaptureinternalpopulationEstimate(
        y = observed,
        formulas = formulas,
        X = variables,
        grad = grad,
        hessian = hessian,
        popVar = popVar,
        weights = priorWeights,
        eta = eta,
        family = family,
        beta = coefficients,
        control = controlPopVar,
        Xvlm = Xvlm,
        W = if (method == "IRLS")
              weights 
            else 
              family$Wfun(prior = priorWeights, eta = eta),
        sizeObserved = sizeObserved,
        modelFrame = modelFrame,
        cov = NULL,
        offset = offset,
        weightsFlag = controlModel$weightsAsCounts
      )
    }
    
    structure(
      list(
        y                = if(isTRUE(returnElements[[1]])) as.numeric(observed) else NULL, # droping names
        X                = if(isTRUE(returnElements[[2]])) variables else NULL,
        modelFrame       = if (isTRUE(returnElements[[3]])) modelFrame else NULL,
        formula          = formulas,
        call             = match.call(),
        coefficients     = coefficients,
        control          = list(controlModel  = controlModel,
                                controlMethod = controlMethod),
        nullDeviance     = nullDeviance,
        model            = family,
        deviance         = deviance,
        priorWeights     = priorWeights,
        weights          = weights,
        residuals        = resRes,
        logL             = LOG,
        iter             = iter,
        dfResidual       = dfReduced,
        dfNull           = length(observed) - 1,
        fittValues       = fitt,
        populationSize   = POP,
        linearPredictors = eta,
        offset           = offset,
        sizeObserved     = sizeObserved,
        terms            = terms,
        contrasts        = contrasts,
        naAction         = naAction,
        fittingLog       = if (is.null(IRLSlog)) "IRLS logs were not saved." else IRLSlog
      ),
      class = c("singleRStaticCountData", "glm", "lm")
    )
  } else {
    stop(
      "Argument ratioReg is not supported in estimatePopsize(). ",
      "Use ratioReg() as a standalone function instead."
    )
  }
}

#' @title Generating data in singleRcapture
#' 
#' @description
#' An S3 method for \code{stats::simulate} to handle \code{singleRStaticCountData} and 
#' \code{singleRfamily} classes.
#'
#' @param object an object representing a fitted model.
#' @param nsim a numeric scalar specifying:
#' \itemize{
#'    \item number of response vectors to simulate in \code{simulate.singleRStaticCountData}, defaults to \code{1L}.
#'    \item number of units to draw in \code{simulate.singleRfamily}, defaults to \code{NROW(eta)}.
#' }
#' @param seed an object specifying if and how the random number generator should be initialized (‘seeded’).
#' @param truncated logical value indicating whether to sample from truncated or
#' full distribution.
#' @param eta a matrix of linear predictors
#' @param ... additional optional arguments.
#' @return a \code{data.frame} with \code{n} rows and \code{nsim} columns.
#' @seealso [stats::simulate()] [singleRcapture::estimatePopsize()]
#' @examples
#' N <- 10000
#' ###gender <- rbinom(N, 1, 0.2)
#' gender <- rep(0:1, c(8042, 1958))
#' eta <- -1 + 0.5*gender
#' counts <- simulate(ztpoisson(), eta = cbind(eta), seed = 1)
#' df <- data.frame(gender, eta, counts)
#' df2 <- subset(df, counts > 0)
#' ### check coverage with summary
#' mod1 <-  estimatePopsize(
#'   formula       = counts ~ 1 + gender, 
#'   data          = df2, 
#'   model         = ztpoisson, 
#'   controlMethod = list(silent = TRUE)
#' )
#' mod1_sims <- simulate(mod1, nsim=10, seed = 1)
#' colMeans(mod1_sims)
#' mean(df2$counts)
#' @author Maciej Beręsewicz, Piotr Chlebicki
#' @importFrom stats simulate
#' @method simulate singleRStaticCountData
#' @exportS3Method
#' @name simulate
#' @export
simulate.singleRStaticCountData <- function(object, nsim = 1, seed = NULL, ...) {
  n <- nobs(object)
  eta <- object$linearPredictors
  
  if (isTRUE(object$control$controlModel$weightsAsCounts)) {
    eta <- matrix(
      unlist(sapply(1:NROW(eta), function(x) {
        rep(eta[x,], object$priorWeights[x])
      })),
      ncol = NCOL(eta),
      byrow = TRUE,
      dimnames = list(
        1:n, colnames(eta)
      )
    )
  }
  
  val <- simulate(
    object    = family(object), 
    seed      = seed, 
    nsim      = n * nsim, 
    eta       = eta, 
    truncated = TRUE
  )

  dim(val) <- c(n, nsim)
  val <- as.data.frame(val)
  names(val) <- paste0("sim_", seq_len(nsim))
  val
}

#' @rdname simulate
#' @importFrom stats simulate
#' @method simulate singleRfamily
#' @exportS3Method
#' @export
simulate.singleRfamily <- function(object, 
                                   nsim,
                                   seed = NULL, 
                                   eta, 
                                   truncated = FALSE, 
                                   ...) {
  if (missing(nsim)) nsim <- NROW(eta)
  lower <- if (truncated) {
    if (is.null(object$simulateLower)) 0 else object$simulateLower
  } else {
    -1
  }
  upper <- if (is.null(object$simulateUpper)) Inf else object$simulateUpper
  if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
    runif(1)
  if (is.null(seed))
    RNGstate <- get(".Random.seed", envir = .GlobalEnv)
  else {
    R.seed <- get(".Random.seed", envir = .GlobalEnv)
    set.seed(seed)
    RNGstate <- structure(seed, kind = as.list(RNGkind()))
    on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
  }
  object$simulate(nsim, eta, lower = lower, upper = upper)
}

#' @rdname singleRmodels
#' @importFrom lamW lambertW0
#' @export
Hurdleztpoisson <- function(lambdaLink = c("log", "neglog"), 
                            piLink = c("logit", "cloglog", "probit"), 
                            ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  if (missing(piLink))  piLink <- "logit"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink, piLink)
  
  lambdaLink <- switch(lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  piLink <- switch(piLink,
    "logit" = singleRinternallogitLink,
    "cloglog" = singleRinternalcloglogLink,
    "probit" = singleRinternalprobitLink
  )
  
  links[1:2] <- c(lambdaLink, piLink)
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    if (!deriv) {
      switch (type,
        "nontrunc" = PI + (1 - PI) * exp(-lambda) * (lambda * exp(lambda) - lambda) / (1 - lambda * exp(-lambda)),
        "trunc" = PI * (1 - lambda * exp(-lambda)) / (1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda)) + 
        (1 - PI) * lambda * (exp(lambda) - 1) * exp(-lambda) / (1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda))
      )
    } else {
      switch (type,
        "nontrunc" = {
          matrix(c(
            (1 - PI) * exp(lambda) * (exp(lambda) - lambda ^ 2 + lambda - 1) /
            (exp(lambda) - lambda) ^ 2,
            (1 - lambda) * exp(lambda) / (exp(lambda) - lambda)
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
                piLink(eta[, 2], inverse = TRUE, deriv = 1)
          ), ncol = 2)
        },
        "trunc" = {
          matrix(c(
            (exp(2 * lambda) + (-lambda ^ 2 + PI * lambda - 2) * exp(lambda) + 1) *
            (1 - PI) / (exp(lambda) - lambda + PI - 1) ^ 2,
            -(exp(lambda) - lambda) * ((lambda - 1) * exp(lambda) + 1) / 
            (PI + exp(lambda) - lambda - 1) ^ 2
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
                piLink(eta[, 2], inverse = TRUE, deriv = 1)
          ), ncol = 2)
        }
      )
    }
  }
  
  variance <- function(eta, type = "nontrunc", ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    switch(type,
    "nontrunc" = PI + (1 - PI) * exp(-lambda) * lambda * 
      (exp(lambda) * (1 + lambda) - 1) / (1 - lambda * exp(-lambda)),
    "trunc" = PI * (1 - lambda * exp(-lambda)) / 
      (1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda)) + 
      (1 - PI) * exp(-lambda) * lambda * (exp(lambda) * (1 + lambda) - 1) / 
      (1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda))
    ) - (mu.eta(eta = eta, type = type) ^ 2)
  }
  
  Wfun <- function(prior, eta, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    z <- PI * (1 - lambda * exp(-lambda)) / (1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda))
    
    ## expected for (1-z)Y
    YY <- mu.eta(eta) - z
    
    # PI^2 derivative
    G00 <- exp(-2 * lambda) / (-exp(-lambda) * (1 - PI) - lambda * exp(-lambda) + 1) ^ 2 - 
      z / PI ^ 2 - (1 - z) / (1 - PI) ^ 2
    
    G00 <- G00 * piLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2
    
    # mixed
    G01 <- (exp(lambda) - 1) / (exp(lambda) - lambda + PI - 1) ^ 2
    
    G01 <- G01 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
      piLink(eta[, 2], inverse = TRUE, deriv = 1)
    
    # Beta^2 derivative
    G11 <- (lambda * exp(-lambda) + (1 - PI) * exp(-lambda) - exp(-lambda)) ^ 2 / 
      (-lambda * exp(-lambda) - (1 - PI) * exp(-lambda) + 1) ^ 2 - 
      (z * (lambda * exp(-lambda) - exp(-lambda)) ^ 2) / 
      (1 - lambda * exp(-lambda)) ^ 2 + 
      (lambda * exp(-lambda) + (1 - PI) * exp(-lambda) - 2 * exp(-lambda)) /
      (-lambda * exp(-lambda) - (1 - PI) * exp(-lambda) + 1) + 
      (z * (2 * exp(-lambda) - lambda * exp(-lambda))) /
      (1 - lambda * exp(-lambda)) - YY / lambda ^ 2
    
    G11 <- G11 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2
    
    matrix(
      -c(G11, # lambda
         G01, # mixed
         G01, # mixed
         G00  # pi
      ),
      dimnames = list(rownames(eta), c("lambda", "mixed", "mixed", "pi")),
      ncol = 4
    )
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    weight <- weight / prior
    
    z <- ifelse(y == 1, y, 0)
    
    # lambda derivative
    G1 <- -(lambda * exp(-lambda) + (1 - PI) * exp(-lambda) - exp(-lambda)) /
      (-lambda * exp(-lambda) - (1 - PI) * exp(-lambda) + 1) +
      (z * (lambda * exp(-lambda) - exp(-lambda))) /
      (1 - lambda * exp(-lambda)) + (1 - z) * (y / lambda - 1)
    G1 <- G1 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
    
    # PI derivative
    G0 <- z / PI - (1 - z) / (1 - PI) - 
      exp(-lambda) / (-exp(-lambda) * (1-PI) - lambda * exp(-lambda) + 1)
    G0 <- G0 * piLink(eta[, 2], inverse = TRUE, deriv = 1)
    
    uMatrix <- matrix(c(G1, G0), ncol = 2)
    
    weight <- lapply(X = 1:nrow(weight), FUN = function (x) {
      matrix(as.numeric(weight[x, ]), ncol = 2)
    })
    
    pseudoResid <- sapply(X = 1:length(weight), FUN = function (x) {
      xx <- solve(weight[[x]])
      xx %*% uMatrix[x, ]
    })
    pseudoResid <- t(pseudoResid)
    dimnames(pseudoResid) <- dimnames(eta)
    pseudoResid
  }
  
  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, 
                           ...) {
    y <- as.numeric(y)
    if (is.null(weight)) {
      weight <- 1
    }
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X) / 2), rep(0, NROW(X) / 2))
    }
    z <- as.numeric(y == 1)
    
    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
      function(beta) {
        eta    <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        PI     <- piLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)

        -sum((z * (log(1 - lambda * exp(-lambda)) + log(PI)) +
        (1 - z) * (log(1 - PI) + y * log(lambda) - lambda - lgamma(y + 1)) -
        log(1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda))) * weight)
      },
      function(beta) {
        eta    <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        PI     <- piLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        # lambda derivative
        G1 <- -(lambda * exp(-lambda) + (1 - PI) * exp(-lambda) - exp(-lambda)) /
          (-lambda * exp(-lambda) - (1 - PI) * exp(-lambda) + 1) +
          (z * (lambda * exp(-lambda) - exp(-lambda))) /
          (1 - lambda * exp(-lambda)) + (1 - z) * (y / lambda - 1)
        G1 <- G1 * weight * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
        
        # PI derivative
        G0 <- z / PI - (1 - z) / (1 - PI) - 
          exp(-lambda) / (-exp(-lambda) * (1-PI) - lambda * exp(-lambda) + 1)
        G0 <- G0 * weight * piLink(eta[, 2], inverse = TRUE, deriv = 1)
        
        if (NbyK) {
          XX <- 1:(attr(X, "hwm")[1])
          return(cbind(as.data.frame(X[1:nrow(eta), XX]) * G1, as.data.frame(X[-(1:nrow(eta)), -XX]) * G0))
        }
        if (vectorDer) {
          return(cbind(G1, G0))
        }
        as.numeric(c(G1, G0) %*% X)
      },
      function (beta) {
        lambdaPredNumber <- attr(X, "hwm")[1]
        eta    <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        PI     <- piLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        res <- matrix(nrow = length(beta), ncol = length(beta), 
                      dimnames = list(names(beta), names(beta)))
        
        G1 <- -(lambda * exp(-lambda) + (1 - PI) * exp(-lambda) - exp(-lambda)) /
          (-lambda * exp(-lambda) - (1 - PI) * exp(-lambda) + 1) +
          (z * (lambda * exp(-lambda) - exp(-lambda))) /
          (1 - lambda * exp(-lambda)) + (1 - z) * (y / lambda - 1)
        
        G0 <- z / PI - (1 - z) / (1 - PI) - 
          exp(-lambda) / (-exp(-lambda) * (1-PI) - lambda * exp(-lambda) + 1)
        
        # PI^2 derivative
        G00 <- exp(-2 * lambda) / (-exp(-lambda) * (1 - PI) - lambda * exp(-lambda) + 1) ^ 2 - 
          z / PI ^ 2 - (1 - z) / (1 - PI) ^ 2
        
        G00 <- G00 * piLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2 +
               G0  * piLink(eta[, 2], inverse = TRUE, deriv = 2)
        
        # mixed
        
        G01 <- (exp(lambda) - 1) / (exp(lambda) - lambda + PI - 1) ^ 2
        G01 <- G01 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
          piLink(eta[, 2], inverse = TRUE, deriv = 1)
        
        # lambda^2 derivative
        G11 <- (lambda * exp(-lambda) + (1 - PI) * exp(-lambda) - exp(-lambda)) ^ 2 / 
          (-lambda * exp(-lambda) - (1 - PI) * exp(-lambda) + 1) ^ 2 - 
          (z * (lambda * exp(-lambda) - exp(-lambda)) ^ 2) / 
          (1 - lambda * exp(-lambda)) ^ 2 + 
          (lambda * exp(-lambda) + (1 - PI) * exp(-lambda) - 2 * exp(-lambda)) /
          (-lambda * exp(-lambda) - (1 - PI) * exp(-lambda) + 1) + 
          (z * (2 * exp(-lambda) - lambda * exp(-lambda))) /
          (1 - lambda * exp(-lambda)) - (y * (1 - z)) / lambda ^ 2
        G11 <- G11 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 +
               G1  * lambdaLink(eta[, 1], inverse = TRUE, deriv = 2)
        
        res[-(1:lambdaPredNumber), -(1:lambdaPredNumber)] <- 
          t(as.data.frame(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)] * G00 * weight)) %*% 
          as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)])
        
        res[1:lambdaPredNumber, 1:lambdaPredNumber] <- 
          t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber] * G11 * weight)) %*% 
          X[1:(nrow(X) / 2), 1:lambdaPredNumber]
        
        res[1:lambdaPredNumber, -(1:lambdaPredNumber)] <- 
          t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber] * G01 * weight)) %*% 
          as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)])
        
        res[-(1:lambdaPredNumber), 1:lambdaPredNumber] <- 
          t(t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber] * G01 * weight)) %*% 
              as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)]))
        
        res
      }
    )
  }
  
  validmu <- function(mu) {
    all(0 < mu & is.finite(mu))
  }
  
  devResids <- function(y, eta, wt, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    # when pi = 0 distribution collapses to zotpoisson
    inverseFunction <- function(y) {stats::uniroot(
      f = function(x) {
        lambda <- exp(x)
        (lambda - lambda * exp(-lambda)) / (1 - exp(-lambda) - lambda * exp(-lambda)) - y
      }, 
      lower = -log(y), upper = y * 10, 
      tol = .Machine$double.eps
    )$root}
    
    yUnq <- unique(y)
    idealLambda <- tryCatch(
      expr = {
        suppressWarnings(sapply(yUnq, 
          FUN = function(x) ifelse(x %in% c(1, 2), -Inf, inverseFunction(x))
        ))
      },
      error = function (e) {
        warning("Deviance residuals could not have been computed and zero vector will be returned instead.", call. = FALSE)
        NULL
      }
    )
    if (is.null(idealLambda)) {
      return(rep(0, length(y)))
    }
    
    idealLambda <- sapply(y, FUN = function(x) idealLambda[yUnq == x])
    idealLambda <- exp(idealLambda)
    
    diff <- ifelse(
      y == 1,
      -(log(PI) + log(1 - lambda * exp(-lambda)) - log(1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda))),
      ifelse(y == 2, 0,
      y * log(idealLambda) - idealLambda - log(1 - exp(-idealLambda) - idealLambda * exp(-idealLambda)) - lgamma(y + 1)) - 
      (log(1 - PI) + y * log(lambda) - lambda - lgamma(y + 1) - log(1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda)))
    )
    
    if (any(diff < 0)) {
      warning(paste0(
        "Some of differences between log likelihood in saturated model",
        " and fitted model were positive which indicates either:\n",
        "(1): A very good model fit or\n",
        "(2): Incorrect computation of saturated model",
        "\nDouble check deviance before proceeding"
      ))
      diff[diff < 0] <- 0
    }
    
    sign(y - mu.eta(eta = eta)) * sqrt(2 * wt * diff)
  }
  
  pointEst <- function (pw, eta, contr = FALSE, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    N <- pw * (1 - lambda * exp(-lambda)) / 
      (1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda))
    
    if(!contr) {
      N <- sum(N)
    }
    N
  }
  
  popVar <- function (pw, eta, cov, Xvlm, ...) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    prob <- 1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda)
    
    # w.r to PI
    bigTheta1 <- -pw * piLink(eta[, 2], inverse = TRUE, deriv = 1) * 
      ((exp(lambda) - lambda) / (PI + exp(lambda) - lambda - 1) ^ 2)
    
    # w.r to lambda
    bigTheta2 <- pw * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * 
      (PI - 1) * (exp(lambda) - 1) / (exp(lambda) - lambda + PI - 1) ^ 2
    
    bigTheta <- t(c(bigTheta2, bigTheta1) %*% Xvlm)
    
    f1 <-  t(bigTheta) %*% as.matrix(cov) %*% bigTheta
    
    f2 <- sum(pw * (1 - lambda * exp(-lambda)) * (1 - PI) * exp(-lambda) / (prob ^ 2))
    
    f1 + f2
  }
  
  dFun <- function (x, eta, type = c("trunc", "nontrunc")) {
    if (missing(type)) type <- "trunc"
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    switch (type,
      "trunc" = {
        ifelse(x == 1, PI * (1 - lambda * exp(-lambda)) / 
          (1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda)),
          (1 - PI) * (lambda ^ x) * exp(-lambda) / 
          ((1 - (1 - PI) * exp(-lambda) - lambda * exp(-lambda)) * factorial(x))
        )
      },
      "nontrunc" = {
        ifelse(x == 1, PI, (1 - PI) *
        stats::dpois(x, lambda) / (1 - lambda * exp(-lambda)))
      }
    )
  }
  
  simulate <- function(n, eta, lower = 0, upper = Inf) {
    PI     <- piLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    CDF <- function(x) {
      ifelse(x == Inf, 1, 
      ifelse(x < 0, 0, 
      ifelse(x < 1, (1 - PI) * exp(-lambda) / (1 - lambda * exp(-lambda)), 
      PI + (1 - PI) * (stats::ppois(x, lambda) - lambda * exp(-lambda)
      ) / (1 - lambda * exp(-lambda)))))
    }
    lb <- CDF(lower)
    ub <- CDF(upper)
    p_u <- stats::runif(n, lb, ub)
    sims <- rep(0, n)
    cond <- CDF(sims) <= p_u
    while (any(cond)) {
      sims[cond] <- sims[cond] + 1
      cond <- CDF(sims) <= p_u
    }
    sims
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12)),
        family$links[[2]](weighted.mean(observed == 1, priorWeights) * (.5 + .5 * (observed == 1)) + .01)
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights)),
        family$links[[2]](weighted.mean(observed == 1, priorWeights) + .01)
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
      if ("(Intercept):pi" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[2], rep(0, attr(Xvlm, "hwm")[2] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[2] / attr(Xvlm, "hwm")[2], attr(Xvlm, "hwm")[2]))
      }
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "Hurdleztpoisson",
      etaNames = c("lambda", "pi"),
      simulate  = simulate,
      getStart  = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

#' @rdname regDiagSingleR
#' @export
dfpopsize <- function(model, ...) {
  UseMethod("dfpopsize")
}

#' @method dfpopsize singleRStaticCountData
#' @rdname regDiagSingleR
#' @exportS3Method 
dfpopsize.singleRStaticCountData <- function(model, dfbeta = NULL, ...) {
  dfb <- if (is.null(dfbeta)) dfbeta(model, ...) else dfbeta
  
  X <- model.frame(model, ...)
  
  N <- model$populationSize$pointEstimate
  range <- 1:NROW(dfb)
  
  res <- vector("numeric", length = NROW(dfb))
  pw <- model$priorWeights
  y <- model.response(model.frame(model))
  
  for (k in range) {
    cf <- model$coefficients - dfb[k, ]
    
    if (isTRUE(model$control$controlModel$weightsAsCounts == FALSE)) {
      res[k] <- model$model$pointEst(
        eta = matrix(
          singleRinternalGetXvlmMatrix(
            X = X[rownames(X) != rownames(X)[k], , drop = FALSE], 
            formulas = model$formula, 
            parNames = model$model$etaNames
          ) %*% cf, 
          ncol = length(model$model$etaNames)
        ),
        y = y[-k],
        pw = pw[-k]
      )
    } else {
      # If weights are zero nothing breaks
      kk <- rep(0, length(pw))
      kk[k] <- 1
      res[k] <- model$model$pointEst(
        eta = matrix(
          model.matrix(model, "vlm") %*% cf, 
          ncol = length(model$model$etaNames)
        ),
        y = y,
        pw = pw - kk
      )
    }
  }
  
  N - res
}

#' @method dfbeta singleRStaticCountData
#' @importFrom stats dfbeta
#' @importFrom foreach %dopar%
#' @importFrom foreach foreach
#' @importFrom parallel makeCluster
#' @importFrom parallel stopCluster
#' @importFrom doParallel registerDoParallel
#' @rdname regDiagSingleR
#' @exportS3Method 
dfbeta.singleRStaticCountData <- function(model,
                           maxitNew = 1,
                           trace = FALSE,
                           cores = 1,
                           ...) {
  # The usual formula from glms doesn't work
  X <- model.frame.singleRStaticCountData(model, ...)
  y <- if (is.null(model$y)) stats::model.response(X) else model$y
  X <- X
  y <- y
  cf <- coef(model)
  pw <- model$priorWeights
  offset <- model$offset
  eta <- model$linearPredictors
  
  if (cores > 1) {
    cl <- parallel::makeCluster(cores)
    doParallel::registerDoParallel(cl)
    on.exit(parallel::stopCluster(cl))
    
    if (isFALSE(model$control$controlModel$weightsAsCounts)) {
      res <- foreach::`%dopar%`(
        obj = foreach::foreach(k = 1:NROW(X), .combine = rbind),
        ex = {
          c(cf - estimatePopsizeFit(
            control = controlMethod(
              silent = TRUE,
              maxiter = maxitNew + 1,
              ...
            ),
            y = y[-k],
            X = singleRinternalGetXvlmMatrix(
              X        = X[rownames(X) != rownames(X)[k], , drop = FALSE],
              formulas = model$formula,
              parNames = model$model$etaNames
            ),
            coefStart = cf,
            etaStart  = eta[-k, , drop = FALSE] + offset[-k, , drop = FALSE],
            family = model$model,
            priorWeights = pw[-k],
            method = if (is.null(model$call$method)) "IRLS" else model$call$method,
            offset = offset[-k, , drop = FALSE]
          )$beta)
        }
      )
    } else {
      res <- foreach::`%dopar%`(
        obj = foreach::foreach(k = 1:NROW(X), .combine = rbind),
        ex = {
          if (isFALSE(pw[k] - 1 > 0)) {
            c(cf - estimatePopsizeFit(
              control = controlMethod(
                silent = TRUE,
                maxiter = maxitNew + 1,
                ...
              ),
              y = y[-k],
              X = singleRinternalGetXvlmMatrix(
                X        = X[-k, , drop = FALSE],
                formulas = model$formula,
                parNames = model$model$etaNames
              ),
              coefStart = cf,
              etaStart  = eta[-k, , drop = FALSE] + offset[-k, , drop = FALSE],
              family = model$model,
              priorWeights = pw[-k],
              method = if (is.null(model$call$method)) "IRLS" else model$call$method,
              offset = offset[-k, , drop = FALSE]
            )$beta)
          } else {
            kk <- rep(0, length(pw))
            kk[k] <- 1
            
            c(cf - estimatePopsizeFit(
              control = controlMethod(
                silent = TRUE,
                maxiter = maxitNew + 1,
                ...
              ),
              y = y,
              X = model.matrix(model, "vlm"),
              coefStart = cf,
              etaStart  = eta + offset,
              family = model$model,
              priorWeights = pw - kk,
              method = if (is.null(model$call$method)) "IRLS" else model$call$method,
              offset = offset
            )$beta)
          }
        }
      )
    }
  } else {
    res <- matrix(nrow = nrow(X), ncol = length(cf))
    
    for (k in 1:nrow(X)) {
      if (isTRUE(trace)) {
        cat("-----\nRemoving observation number: ", k, "\n", sep = "")
      }
      if (isFALSE(model$control$controlModel$weightsAsCounts)) {
        res[k, ] <- cf - estimatePopsizeFit(
          control = controlMethod(
            silent = TRUE,
            maxiter = maxitNew + 1,
            ...
          ),
          y = y[-k],
          X = singleRinternalGetXvlmMatrix(
            X        = X[rownames(X) != rownames(X)[k], , drop = FALSE],
            formulas = model$formula, 
            parNames = model$model$etaNames
          ),
          coefStart = cf,
          etaStart  = eta[-k, , drop = FALSE] + offset[-k, , drop = FALSE],
          family = model$model,
          priorWeights = pw[-k],
          method = if (is.null(model$call$method)) "IRLS" else model$call$method,
          offset = offset[-k, , drop = FALSE]
        )$beta
      } else {
        if (isFALSE(pw[k] - 1 > 0)) {
          res[k, ] <- cf - estimatePopsizeFit(
            control = controlMethod(
              silent = TRUE,
              maxiter = maxitNew + 1,
              ...
            ),
            y = y[-k],
            X = singleRinternalGetXvlmMatrix(
              X        = X[rownames(X) != rownames(X)[k], , drop = FALSE],
              formulas = model$formula, 
              parNames = model$model$etaNames
            ),
            coefStart = cf,
            etaStart  = eta[-k, , drop = FALSE] + offset[-k, , drop = FALSE],
            family = model$model,
            priorWeights = pw[-k],
            method = if (is.null(model$call$method)) "IRLS" else model$call$method,
            offset = offset[-k, , drop = FALSE]
          )$beta
        } else {
          kk <- rep(0, length(pw))
          kk[k] <- 1
          
          res[k, ] <- cf - estimatePopsizeFit(
            control = controlMethod(
              silent = TRUE,
              maxiter = maxitNew + 1,
              ...
            ),
            y = y,
            X = model.matrix(model, "vlm"),
            coefStart = cf,
            etaStart  = eta + offset,
            family = model$model,
            priorWeights = pw - kk,
            method = if (is.null(model$call$method)) "IRLS" else model$call$method,
            offset = offset
          )$beta
        }
      }
    }
  }
  
  colnames(res) <- names(cf)
  res
}

#' @rdname singleRmodels
#' @importFrom stats uniroot
#' @importFrom stats dnbinom
#' @export
Hurdleztnegbin <- function(nSim = 1000, epsSim = 1e-8, eimStep = 6, 
                           lambdaLink = c("log", "neglog"), 
                           alphaLink = c("log", "neglog"),
                           piLink = c("logit", "cloglog", "probit"), 
                           ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  if (missing(alphaLink))  alphaLink  <- "log"
  if (missing(piLink))     piLink     <- "logit"

  links <- list()
  attr(links, "linkNames") <- c(lambdaLink, alphaLink, piLink)
  
  lambdaLink <- switch(lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  alphaLink <- switch(alphaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  piLink <- switch(piLink,
    "logit" = singleRinternallogitLink,
    "cloglog" = singleRinternalcloglogLink,
    "probit" = singleRinternalprobitLink
  )
  
  links[1:3] <- c(lambdaLink, alphaLink, piLink)
  
  nbLogP0 <- function(lambda, alpha) {
    -(1 / alpha) * log1p(lambda * alpha)
  }

  nbLogP1 <- function(lambda, alpha) {
    log(lambda) + (-1 / alpha - 1) * log1p(lambda * alpha)
  }

  nbP0 <- function(lambda, alpha) {
    exp(nbLogP0(lambda, alpha))
  }

  nbP1 <- function(lambda, alpha) {
    exp(nbLogP1(lambda, alpha))
  }

  probNotOne <- function(lambda, alpha) {
    pmax(-expm1(pmin(nbLogP1(lambda, alpha), 0)), .Machine$double.xmin)
  }

  logProbNotOne <- function(lambda, alpha) {
    log(probNotOne(lambda, alpha))
  }

  probGe2 <- function(lambda, alpha) {
    logP0 <- nbLogP0(lambda, alpha)
    logP1 <- nbLogP1(lambda, alpha)
    logP01 <- pmax(logP0, logP1) + log1p(exp(-abs(logP0 - logP1)))
    pmax(-expm1(pmin(logP01, 0)), .Machine$double.xmin)
  }

  logProbGe2 <- function(lambda, alpha) {
    log(probGe2(lambda, alpha))
  }

  probTrunc <- function(lambda, alpha, PI) {
    logP0 <- log1p(-PI) + nbLogP0(lambda, alpha)
    logP1 <- nbLogP1(lambda, alpha)
    logPdrop <- pmax(logP0, logP1) + log1p(exp(-abs(logP0 - logP1)))
    pmax(-expm1(pmin(logPdrop, 0)), .Machine$double.xmin)
  }

  logProbTrunc <- function(lambda, alpha, PI) {
    log(probTrunc(lambda, alpha, PI))
  }

  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    P1 <- nbP1(lambda, alpha)
    prob_not_one <- probNotOne(lambda, alpha)
    prob_trunc <- probTrunc(lambda, alpha, PI)
    
    if (!deriv) {
      switch (type,
        "nontrunc" = PI + (1 - PI) * 
        (lambda - P1) / prob_not_one,
        "trunc" = (PI - P1 +
        (1 - PI) * lambda) / prob_trunc
      )
    } else {
      switch (type,
        "nontrunc" = {
          matrix(c(
            (1 - PI) * (alpha * lambda + 1) ^ (1 / alpha) * 
            ((alpha * lambda + 1) ^ (1 / alpha) * (alpha ^ 2 * lambda ^ 2 + 2 * alpha * lambda + 1) +
            (-alpha - 1) * lambda ^ 2 + lambda - 1) /
            ((alpha * lambda + 1) ^ (1 / alpha + 1) - lambda) ^ 2,
            (1 - PI) * (lambda - 1) * lambda * (lambda * alpha + 1) ^ (1 / alpha) *
            ((lambda * alpha + 1) * log(lambda * alpha + 1) - lambda * alpha ^ 2 - lambda * alpha) /
            (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha + 1) - lambda) ^ 2),
            1 - (lambda - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1)) /
            (1 - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1))
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
             alphaLink(eta[, 2], inverse = TRUE, deriv = 1),
                piLink(eta[, 3], inverse = TRUE, deriv = 1)
          ), ncol = 3)
        },
        "trunc" = {
          matrix(c(
            (1 - PI) * ((alpha * lambda + 1) ^ (1 / alpha) * 
            (((PI - 1) * alpha ^ 2 + (PI - 2) * alpha - 1) * lambda ^ 2 +
            ((PI - 2) * alpha + PI) * lambda - 2) + (alpha * lambda + 1) ^ (2 / alpha) *
            (alpha ^ 2 * lambda ^ 2 + 2 * alpha * lambda + 1) + 1) /
            ((alpha * lambda + 1) ^ (1 / alpha + 1) + ((PI - 1) * alpha - 1) * lambda + PI - 1) ^ 2,
            (PI - 1) * ((lambda * alpha + 1) ^ (1 / alpha) *
            (((PI - 1) * lambda ^ 3 - PI * lambda ^ 2) * alpha ^ 2 +
            (-lambda ^ 3 + (2 * PI - 1) * lambda ^ 2 - 2 * PI * lambda) * alpha -
            lambda ^ 2 + PI * lambda - PI) * log(lambda * alpha + 1) +
            (lambda * alpha + 1) ^ (1 / alpha) * 
            (((2 - PI) * lambda ^ 3 + (PI - 1) * lambda ^ 2) * alpha ^ 2 +
            (lambda ^ 3 - PI * lambda ^ 2 + PI * lambda) * alpha) + lambda ^ 2 * alpha ^ 2) /
            (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha + 1) + 
            (PI - 1) * lambda * alpha - lambda + PI - 1) ^ 2),
            -(alpha * lambda + 1) * ((alpha * lambda + 1) ^ (2 / alpha) *
            (alpha * lambda ^ 2 + (1 - alpha) * lambda - 1) + 
            (alpha * lambda + 1) ^ (1 / alpha) * (-lambda ^ 2 + (alpha + 1) * lambda + 1) - lambda) /
            ((alpha * lambda + 1) * PI + (alpha * lambda + 1) ^ (1 / alpha + 1) + (-alpha - 1) * lambda - 1) ^ 2
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
             alphaLink(eta[, 2], inverse = TRUE, deriv = 1),
                piLink(eta[, 3], inverse = TRUE, deriv = 1)
          ), ncol = 3)
        }
      )
    }
  }
  
  variance <- function(eta, type = "nontrunc", ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    P0 <- nbP0(lambda, alpha)
    P1 <- nbP1(lambda, alpha)
    prob_not_one <- probNotOne(lambda, alpha)
    normTerm <- probTrunc(lambda, alpha, PI)
    
    res <- switch (type,
      nontrunc = PI + (1 - PI) * (lambda * (1 + lambda + alpha * lambda) - P1) / prob_not_one,
      trunc    = (PI - P1 + (1 - PI) * lambda * (1 + lambda + alpha * lambda)) / normTerm
    ) - mu.eta(eta = eta, type = type) ^ 2

    pmax(res, 0)
  }
  
  compdigamma <- function(y, alpha) {
    (-digamma(y + 1 / alpha) + digamma(1 / alpha)) / (alpha ^ 2)
  }
  
  
  comptrigamma <- function(y, alpha) {
    (2 * (digamma(y + 1 / alpha) - digamma(1 / alpha)) * alpha +
    trigamma(y + 1 / alpha) - trigamma(1 / alpha)) / alpha ^ 4
  }
  
  # Computing the expected value of di/trigamma functions on (y + 1/alpha)
  compExpectG1 <- function(eta) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-  piLink(eta[, 3], inverse = TRUE)
    
    normTerm <- probTrunc(lambda, alpha, PI)
    
    res <- rep(0, NROW(eta))
    # 1 is the first possible y value for 0 truncated hurdle distribution
    # but here we compute the (1 - z) * psi function which takes 0 at y = 1
    k <- 2 
    finished <- rep(FALSE, NROW(eta))
    while ((k < nSim) & !all(finished)) {
      prob <- apply(cbind(k:(k + eimStep)), MARGIN = 1, FUN = function(x) {
	        (1 - PI) * stats::dnbinom(
	          x = x,
	          size = 1 / alpha,
	          mu = lambda
	        ) / normTerm
	      })
      trg <- apply(cbind(k:(k + eimStep)), MARGIN = 1, FUN = function(x) {
        comptrigamma(y = x, alpha = alpha)
      })
      prob[!(is.finite(prob))] <- 0
      trg[!(is.finite(trg))] <- 0
      toAdd <- trg * prob
      toAdd <- rowSums(toAdd)
      k <- k + eimStep + 1
      res <- res + toAdd
      finished <- abs(toAdd) < epsSim
    }
    res
  }
  
  Wfun <- function(prior, eta, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    
    z <- PI * (1 - lambda * (1 + alpha * lambda) ^ (-1 / alpha - 1))
    z <- z / (1 - (1 - PI) * (1 + alpha * lambda) ^ (-1 / alpha) - lambda * (1 + alpha * lambda) ^ (- 1 / alpha - 1))
    
    ## expected for (1-z)Y
    XXX <- mu.eta(eta, type = "trunc") - z

    Etrig  <- compExpectG1(eta)

    # PI derivative
    G00 <- 1 / ((alpha * lambda + 1) ^ (2 / alpha) * 
    (-(1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) -
    lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1) ^ 2) -
    z / PI ^ 2 - (1 - z) / (1 - PI) ^ 2
    
    G00 <- G00 * piLink(eta[, 3], inverse = TRUE, deriv = 1) ^ 2
    
    # PI alpha derivative
    G01 <- -((lambda * alpha + 1) ^ (1 / alpha) * (1 + alpha * lambda) ^ 2 *
    log(lambda * alpha + 1) + (lambda * alpha + 1) ^ (1 / alpha) *
    (-lambda ^ 2 * alpha ^ 2 - lambda * alpha) - lambda ^ 2 * alpha ^ 2) /
    (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha + 1) + 
    (PI - 1) * lambda * alpha - lambda + PI - 1) ^ 2)
    
    G01 <- G01 * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) *
                    piLink(eta[, 3], inverse = TRUE, deriv = 1)
    
    # PI lambda derivative
    G02 <- ((alpha * lambda + 1) ^ (1 / alpha + 1) - 1) /
    ((alpha * lambda + 1) ^ (1 / alpha + 1) + 
    ((PI - 1) * alpha - 1) * lambda + PI - 1) ^ 2
    
    G02 <- G02 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
                     piLink(eta[, 3], inverse = TRUE, deriv = 1)
    
    # alpha derivative
    G11 <- (-(1 - PI) * (log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
    (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 + lambda * (-1 / alpha - 1) / (lambda * alpha + 1))) ^ 2 /
    (1 - (1 - PI) / (lambda * alpha + 1) ^ (1 / alpha) -
    lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) ^ 2 -
    (-(1 - PI) * (log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) ^ 2 /
    (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 + lambda * (-1 / alpha - 1) / (lambda * alpha + 1)) ^ 2 -
    (1 - PI) * (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 + 2 * lambda / (alpha ^ 2 * (lambda * alpha + 1)) +
    lambda ^ 2 / (alpha * (lambda * alpha + 1)^2)) / (lambda * alpha + 1) ^ (1 / alpha) -
    lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) * 
    (-2 * log(lambda * alpha + 1) / alpha ^ 3 + 
    2 * lambda / (alpha ^ 2 * (lambda * alpha + 1)) - 
    lambda ^ 2 * (-1 / alpha - 1) / (lambda * alpha + 1) ^ 2)) /
    (1 - (1 - PI) / (lambda * alpha + 1) ^ (1 / alpha) -
    lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) -
    z * lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) * 
    (log(lambda * alpha + 1) / alpha ^ 2 + 
    (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) ^ 2 /
    (1 - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) -
    z * lambda ^ 2 * (lambda * alpha + 1) ^ (-2 / alpha - 2) *
    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) ^ 2 /
    (1 - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) ^ 2 +
    (1 - z) * (-2 * log(lambda * alpha + 1) / alpha ^ 3 + 
    2 * lambda / (alpha ^ 2 * (lambda * alpha + 1)) + lambda ^ 2 / (alpha * (lambda * alpha + 1) ^ 2)) -
    2 * XXX / ((1 / alpha + lambda) * alpha ^ 3) + 
    XXX / ((1 / alpha + lambda) ^ 2 * alpha ^ 4) + Etrig - 
    z * lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) * 
    (-2 * log(lambda * alpha + 1) / alpha ^ 3 + 2 * lambda / (alpha ^ 2 * (lambda * alpha + 1)) -
    lambda ^ 2 * (-1 / alpha - 1) / (lambda * alpha + 1) ^ 2) / 
    (1 - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1))
    
    G11 <- G11 * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2
    
    # alpha lambda derivative
    G12 <- (1 - z) * lambda / (lambda * alpha + 1) ^ 2 - 
    XXX / ((1 / alpha + lambda) ^ 2 * alpha ^ 2) -
    ((-(1 - PI) * (lambda * alpha + 1) ^ (-1 / alpha - 1) +
    (lambda * alpha + 1) ^ (-1 / alpha - 1) - 
    lambda * (1 / alpha + 1) * alpha * (lambda * alpha + 1) ^ (-1 / alpha- 2)) *
    (((PI - 1) * (log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1)))) /
    (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)))) / 
    ((PI - 1) / (lambda * alpha + 1) ^ (1 / alpha) - 
    lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) + 1) ^ 2 +
    (-(1 - PI) * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 + 
    (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) +
    (lambda * alpha + 1) ^ (-1 / alpha - 1) * 
    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) /
    (lambda * alpha + 1)) - lambda * (1 / alpha + 1) * alpha * 
    (lambda * alpha + 1) ^ (-1 / alpha - 2) * 
    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 2)) / (lambda * alpha + 1)) -
    lambda * (1 / alpha + 1) * (lambda * alpha + 1) ^ (-1 / alpha - 2) + 
    (lambda * (lambda * alpha + 1) ^ (-1 / alpha -2)) / alpha) / 
    ((PI - 1) / (lambda * alpha + 1) ^ (1 / alpha) -
    lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) + 1) +
    z * (-(lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) -
    lambda * (-1 / alpha - 1) * alpha * (lambda * alpha + 1) ^ (-1 / alpha - 2) *
    (log(lambda * alpha + 1) / alpha ^ 2 + lambda * (-1 / alpha - 2) / (lambda * alpha + 1)) -
    (lambda * (lambda * alpha + 1) ^ (-1 / alpha - 2)) / alpha - 
    lambda * (-1 / alpha - 1) * (lambda * alpha + 1) ^ (-1 / alpha - 2)) /
    (1 - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) +
    z * lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (-(lambda * alpha + 1) ^ (-1 / alpha - 1) - lambda * (-1 / alpha - 1) *
    alpha * (lambda * alpha + 1) ^ (-1 / alpha - 2)) *
    (log(lambda * alpha + 1) / alpha ^ 2 + 
    (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) /
    (1 - lambda * (lambda * alpha + 1) ^( -1 / alpha - 1)) ^ 2
    
    G12 <- G12 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
                  alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
    
    #lambda derivative
    G22 <- (1 - z) * alpha / (alpha * lambda + 1) ^ 2 + XXX / (lambda + 1 / alpha) ^ 2 - XXX / lambda ^ 2 - 
    ((1 - PI) * (-1 / alpha - 1) * alpha * (alpha * lambda + 1) ^ (-1 / alpha - 2) -
    2 * (-1 / alpha - 1) * alpha * (alpha * lambda + 1) ^ (-1 / alpha - 2) -
    (-1 / alpha - 2) * (-1 / alpha - 1) * alpha ^ 2 * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 3)) /
    (1 - (1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1)) +
    ((1 - PI) * (alpha * lambda + 1) ^ (-1 / alpha - 1) - (alpha * lambda + 1) ^ (-1 / alpha - 1) -
    (-1 / alpha - 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) ^ 2 /
    (-(1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) -
    lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1) ^ 2 +
    z * (-2 * (-1 / alpha - 1) * alpha * (alpha * lambda + 1) ^ (-1 / alpha - 2) -
    (-1 / alpha - 2) * (-1 / alpha - 1) * alpha ^ 2 * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 3)) /
    (1 - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1)) - 
    z * (-(alpha * lambda + 1) ^ (-1 / alpha - 1) - 
    (-1 / alpha - 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) ^ 2 /
    (1 - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1)) ^ 2
    
    G22 <- G22 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2
    
    matrix(
      -c(G22 * prior, G12 * prior, G02 * prior,
         G12 * prior, G11 * prior, G01 * prior,
         G02 * prior, G01 * prior, G00 * prior),
      dimnames = list(
        rownames(eta), 
        c("lambda",       "lambda:alpha", "lambda:PI",
          "lambda:alpha", "alpha",        "alpha:PI",
          "lambda:PI",    "alpha:PI",     "PI")
      ),
      ncol = 9
    )
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    z <- as.numeric(y == 1)
    weight <- weight / prior
    
    dig <- compdigamma(y = y, alpha = alpha)
    
    G0 <- z / PI - (1 - z) / (1 - PI) - 1 /
    ((alpha * lambda + 1) ^ (1 / alpha) * 
    (-(1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) -
    lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1))
    
    G1 <- (1 - z) * (log(lambda * alpha + 1) / alpha ^ 2 - 
    lambda / (alpha * (lambda * alpha + 1)) + y / ((1 / alpha + lambda) * alpha ^ 2) + dig) - 
    z * lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) /
    (1 - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) - 
    (-(1 - PI) * (log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
    (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 + lambda * (-1 / alpha - 1) / (lambda * alpha + 1))) /
    (1 - (1 - PI) / (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1))
    
    G2 <- (1 - z) * (-1 / (alpha * lambda + 1) - y / (lambda + 1 / alpha) + y / lambda) -
    ((1 - PI) * (alpha * lambda + 1) ^ (-1 / alpha - 1) - (alpha * lambda + 1) ^ (-1 / alpha - 1) - 
    (-1 / alpha - 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) /
    (-(1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1) +
    z * (-(alpha * lambda + 1) ^ (-1 / alpha - 1) - 
    (-1 / alpha - 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) /
    (1 - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1))
    
    
    G0 <- G0 *     piLink(eta[, 3], inverse = TRUE, deriv = 1)
    G1 <- G1 *  alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
    G2 <- G2 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
    
    uMatrix <- cbind(G2, G1, G0)
    
    weight <- lapply(X = 1:nrow(weight), FUN = function (x) {
      matrix(as.numeric(weight[x, ]), ncol = 3)
    })
    
    pseudoResid <- sapply(X = 1:length(weight), FUN = function (x) {
      xx <- solve(weight[[x]])
      xx %*% uMatrix[x, ]
    })
    
    pseudoResid <- t(pseudoResid)
    dimnames(pseudoResid) <- dimnames(eta)
    pseudoResid
  }
  
  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, 
                           ...) {
    if (is.null(weight)) {
      weight <- 1
    }
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X) / 3), rep(0, NROW(X) / 3), rep(0, NROW(X) / 3))
    }
    
    
    y <- as.numeric(y)
    z <- as.numeric(y == 1)
    X <- as.matrix(X)
    
    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
	            function(beta) {
	              eta    <- matrix(as.matrix(X) %*% beta, ncol = 3) + offset
	              lambda <- lambdaLink(eta[, 1], inverse = TRUE)
	              alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
	              PI     <-     piLink(eta[, 3], inverse = TRUE)

	              -sum(weight * (
	              z * (log(PI) + logProbNotOne(lambda, alpha)) +
	              (1 - z) * (log(1 - PI) + lgamma(y + 1 / alpha) - lgamma(y + 1) -
	              lgamma(1 / alpha) - log(1 + alpha * lambda) / alpha +
	              y * log(lambda) - y * log(lambda + 1 / alpha)) -
	              logProbTrunc(lambda, alpha, PI)))
	            },
            function(beta) {
              eta    <- matrix(as.matrix(X) %*% beta, ncol = 3) + offset
              lambda <- lambdaLink(eta[, 1], inverse = TRUE)
              alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
              PI     <-     piLink(eta[, 3], inverse = TRUE)
              
              dig <- compdigamma(y = y, alpha = alpha)
              
              G0 <- z / PI - (1 - z) / (1 - PI) - 1 /
              ((alpha * lambda + 1) ^ (1 / alpha) * 
              (-(1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) -
              lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1))
              
              G1 <- (1 - z) * (log(lambda * alpha + 1) / alpha ^ 2 - 
              lambda / (alpha * (lambda * alpha + 1)) +
              y / ((1 / alpha + lambda) * alpha ^ 2) + dig) - 
              z * lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) /
              (1 - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) - 
              (-(1 - PI) * (log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
              (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 + lambda * (-1 / alpha - 1) / (lambda * alpha + 1))) /
              (1 - (1 - PI) / (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1))
              
              G2 <- (1 - z) * (-1 / (alpha * lambda + 1) - y / (lambda + 1 / alpha) + y / lambda) -
              ((1 - PI) * (alpha * lambda + 1) ^ (-1 / alpha - 1) - (alpha * lambda + 1) ^ (-1 / alpha - 1) - 
              (-1 / alpha - 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) /
              (-(1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1) +
              z * (-(alpha * lambda + 1) ^ (-1 / alpha - 1) - 
              (-1 / alpha - 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) /
              (1 - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1))
              
              G0 <- G0 * weight *     piLink(eta[, 3], inverse = TRUE, deriv = 1)
              G1 <- G1 * weight *  alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
              G2 <- G2 * weight * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
              
              if (NbyK) {
                return(cbind(
                  G2 * as.data.frame(
                    X[1:nrow(eta), 
                      1:(attr(X, "hwm")[1])]
                  ), 
                  G1 * as.data.frame(
                    X[(nrow(eta) + 1):(2 * nrow(eta)), 
                      (attr(X, "hwm")[1] + 1):(attr(X, "hwm")[2])]
                  ), 
                  G0 * as.data.frame(
                    X[(2 * nrow(eta) + 1):(3 * nrow(eta)), 
                      (attr(X, "hwm")[2] + 1):(attr(X, "hwm")[3])]
                  )
                ))
              }
              if (vectorDer) {
                return(cbind(G2, G1, G0))
              }
              
              as.numeric(c(G2, G1, G0) %*% X)
            },
            function(beta) {
              predNumbers <- attr(X, "hwm")
              eta    <- matrix(as.matrix(X) %*% beta, ncol = 3) + offset
              lambda <- lambdaLink(eta[, 1], inverse = TRUE)
              alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
              PI     <-     piLink(eta[, 3], inverse = TRUE)
              
              res <- matrix(
                nrow = length(beta), 
                ncol = length(beta), 
                dimnames = list(names(beta), names(beta))
              )
              
              trig <- comptrigamma(y = y, alpha = alpha)
              dig  <-  compdigamma(y = y, alpha = alpha)
              
              # PI dervative
              G0 <- z / PI - (1 - z) / (1 - PI) - 1 /
              ((alpha * lambda + 1) ^ (1 / alpha) * 
              (-(1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) -
              lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1))
              
              G00 <- 1 / ((alpha * lambda + 1) ^ (2 / alpha) * 
              (-(1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) -
              lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1) ^ 2) -
              z / PI ^ 2 - (1 - z) / (1 - PI) ^ 2
              
              G00 <- G00 * piLink(eta[, 3], inverse = TRUE, deriv = 1) ^ 2 +
                      G0 * piLink(eta[, 3], inverse = TRUE, deriv = 2)
              
              # PI alpha dervative
              G01 <- -((lambda * alpha + 1) ^ (1 / alpha) * (1 + alpha * lambda) ^ 2 *
              log(lambda * alpha + 1) + (lambda * alpha + 1) ^ (1 / alpha) *
              (-lambda ^ 2 * alpha ^ 2 - lambda * alpha) - lambda ^ 2 * alpha ^ 2) /
              (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha + 1) + 
                            (PI - 1) * lambda * alpha - lambda + PI - 1) ^ 2)
              
              G01 <- G01 * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) *
                              piLink(eta[, 3], inverse = TRUE, deriv = 1)
              
              # PI lambda dervative
              G02 <- ((alpha * lambda + 1) ^ (1 / alpha + 1) - 1) /
              ((alpha * lambda + 1) ^ (1 / alpha + 1) + 
              ((PI - 1) * alpha - 1) * lambda + PI - 1) ^ 2
              
              G02 <- G02 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
                               piLink(eta[, 3], inverse = TRUE, deriv = 1)
              
              # alpha dervative
              G1 <- (1 - z) * (log(lambda * alpha + 1) / alpha ^ 2 - 
              lambda / (alpha * (lambda * alpha + 1)) +
              y / ((1 / alpha + lambda) * alpha ^ 2) + dig) - 
              z * lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) /
              (1 - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) - 
              (-(1 - PI) * (log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
              (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 + lambda * (-1 / alpha - 1) / (lambda * alpha + 1))) /
              (1 - (1 - PI) / (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1))
              
              G11 <- (-(1 - PI) * (log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
              (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 + lambda * (-1 / alpha - 1) / (lambda * alpha + 1))) ^ 2 /
              (1 - (1 - PI) / (lambda * alpha + 1) ^ (1 / alpha) -
              lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) ^ 2 -
              (-(1 - PI) * (log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) ^ 2 /
              (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 + lambda * (-1 / alpha - 1) / (lambda * alpha + 1)) ^ 2 -
              (1 - PI) * (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 + 2 * lambda / (alpha ^ 2 * (lambda * alpha + 1)) +
              lambda ^ 2 / (alpha * (lambda * alpha + 1)^2)) / (lambda * alpha + 1) ^ (1 / alpha) -
              lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) * 
              (-2 * log(lambda * alpha + 1) / alpha ^ 3 + 
              2 * lambda / (alpha ^ 2 * (lambda * alpha + 1)) - 
              lambda ^ 2 * (-1 / alpha - 1) / (lambda * alpha + 1) ^ 2)) /
              (1 - (1 - PI) / (lambda * alpha + 1) ^ (1 / alpha) -
              lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) -
              z * lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) * 
              (log(lambda * alpha + 1) / alpha ^ 2 + 
              (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) ^ 2 /
              (1 - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) -
              z * lambda ^ 2 * (lambda * alpha + 1) ^ (-2 / alpha - 2) *
              (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) ^ 2 /
              (1 - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) ^ 2 +
              (1 - z) * (-2 * log(lambda * alpha + 1) / alpha ^ 3 + 
              2 * lambda / (alpha ^ 2 * (lambda * alpha + 1)) + lambda ^ 2 / (alpha * (lambda * alpha + 1) ^ 2) -
              2 * y / ((1 / alpha + lambda) * alpha ^ 3) + 
              y / ((1 / alpha + lambda) ^ 2 * alpha ^ 4) + trig) - 
              z * lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) * 
              (-2 * log(lambda * alpha + 1) / alpha ^ 3 + 2 * lambda / (alpha ^ 2 * (lambda * alpha + 1)) -
              lambda ^ 2 * (-1 / alpha - 1) / (lambda * alpha + 1) ^ 2) / 
              (1 - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1))
              
              G11 <- G11 * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2 +
                      G1 * alphaLink(eta[, 2], inverse = TRUE, deriv = 2)
              
              # alpha lambda dervative
              G12 <- (1 - z) * (lambda / (lambda * alpha + 1) ^ 2 - 
              y / ((1 / alpha + lambda) ^ 2 * alpha ^ 2)) -
              ((-(1 - PI) * (lambda * alpha + 1) ^ (-1 / alpha - 1) +
              (lambda * alpha + 1) ^ (-1 / alpha - 1) - 
              lambda * (1 / alpha + 1) * alpha * (lambda * alpha + 1) ^ (-1 / alpha- 2)) *
              (((PI - 1) * (log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1)))) /
              (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)))) / 
              ((PI - 1) / (lambda * alpha + 1) ^ (1 / alpha) - 
              lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) + 1) ^ 2 +
              (-(1 - PI) * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 + 
              (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) +
              (lambda * alpha + 1) ^ (-1 / alpha - 1) * 
              (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) /
              (lambda * alpha + 1)) - lambda * (1 / alpha + 1) * alpha * 
              (lambda * alpha + 1) ^ (-1 / alpha - 2) * 
              (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 2)) / (lambda * alpha + 1)) -
              lambda * (1 / alpha + 1) * (lambda * alpha + 1) ^ (-1 / alpha - 2) + 
              (lambda * (lambda * alpha + 1) ^ (-1 / alpha -2)) / alpha) / 
              ((PI - 1) / (lambda * alpha + 1) ^ (1 / alpha) -
              lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) + 1) +
              z * (-(lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) -
              lambda * (-1 / alpha - 1) * alpha * (lambda * alpha + 1) ^ (-1 / alpha - 2) *
              (log(lambda * alpha + 1) / alpha ^ 2 + lambda * (-1 / alpha - 2) / (lambda * alpha + 1)) -
              (lambda * (lambda * alpha + 1) ^ (-1 / alpha - 2)) / alpha - 
              lambda * (-1 / alpha - 1) * (lambda * alpha + 1) ^ (-1 / alpha - 2)) /
              (1 - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) +
              z * lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
              (-(lambda * alpha + 1) ^ (-1 / alpha - 1) - lambda * (-1 / alpha - 1) *
              alpha * (lambda * alpha + 1) ^ (-1 / alpha - 2)) *
              (log(lambda * alpha + 1) / alpha ^ 2 + 
              (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1)) /
              (1 - lambda * (lambda * alpha + 1) ^( -1 / alpha - 1)) ^ 2
              
              G12 <- G12 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
                            alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
              
              #lambda dervative
              G2 <- (1 - z) * (-1 / (alpha * lambda + 1) - y / (lambda + 1 / alpha) + y / lambda) -
              ((1 - PI) * (alpha * lambda + 1) ^ (-1 / alpha - 1) - (alpha * lambda + 1) ^ (-1 / alpha - 1) - 
              (-1 / alpha - 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) /
              (-(1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1) +
              z * (-(alpha * lambda + 1) ^ (-1 / alpha - 1) - 
              (-1 / alpha - 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) /
              (1 - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1))
              
              G22 <- (1 - z) * (alpha / (alpha * lambda + 1) ^ 2 + y / (lambda + 1 / alpha) ^ 2 - y / lambda ^ 2) - 
              ((1 - PI) * (-1 / alpha - 1) * alpha * (alpha * lambda + 1) ^ (-1 / alpha - 2) -
              2 * (-1 / alpha - 1) * alpha * (alpha * lambda + 1) ^ (-1 / alpha - 2) -
              (-1 / alpha - 2) * (-1 / alpha - 1) * alpha ^ 2 * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 3)) /
              (1 - (1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1)) +
              ((1 - PI) * (alpha * lambda + 1) ^ (-1 / alpha - 1) - (alpha * lambda + 1) ^ (-1 / alpha - 1) -
              (-1 / alpha - 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) ^ 2 /
              (-(1 - PI) / (alpha * lambda + 1) ^ (1 / alpha) -
              lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1) ^ 2 +
              z * (-2 * (-1 / alpha - 1) * alpha * (alpha * lambda + 1) ^ (-1 / alpha - 2) -
              (-1 / alpha - 2) * (-1 / alpha - 1) * alpha ^ 2 * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 3)) /
              (1 - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1)) - 
              z * (-(alpha * lambda + 1) ^ (-1 / alpha - 1) - 
              (-1 / alpha - 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) ^ 2 /
              (1 - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1)) ^ 2
              
              G22 <- G22 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 + 
                      G2 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 2)
              
              
              predNumbers <- cumsum(predNumbers)
              
              res[(predNumbers[2]+1):predNumbers[3], (predNumbers[2]+1):predNumbers[3]] <- #PI
                t(as.data.frame(X[(nrow(X) * 2 / 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]]) *
                    G00 * weight) %*% X[(nrow(X) * 2 / 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]]
              
              res[(predNumbers[1]+1):predNumbers[2], (predNumbers[2]+1):predNumbers[3]] <- #PI alpha
                t(as.data.frame(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1] + 1):predNumbers[2]]) * 
                    G01 * weight) %*% as.matrix(X[(nrow(X) * 2/ 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]])
              
              res[(predNumbers[2]+1):predNumbers[3], (predNumbers[1]+1):predNumbers[2]] <- #PI alpha
                t(t(as.data.frame(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1] + 1):predNumbers[2]]) * 
                    G01 * weight) %*% as.matrix(X[(nrow(X) * 2/ 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]]))
              
              res[1:predNumbers[1], (predNumbers[2]+1):predNumbers[3]] <- #PI lambda
                t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                    G02 * weight) %*% as.matrix(X[(nrow(X) * 2 / 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]])
              
              res[(predNumbers[2]+1):predNumbers[3], 1:predNumbers[1]] <- #PI lambda
                t(t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                    G02 * weight) %*% as.matrix(X[(nrow(X) * 2 / 3 + 1):nrow(X), (predNumbers[2]+1):predNumbers[3]]))
              
              res[(predNumbers[1]+1):predNumbers[2], (predNumbers[1]+1):predNumbers[2]] <- #alpha
                t(as.data.frame(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]]) *
                    G11 * weight) %*% X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]]
              
              res[1:predNumbers[1], 1:predNumbers[1]] <-  # lambda
                t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                    G22 * weight) %*% X[1:(nrow(X) / 3), 1:predNumbers[1]]
              
              res[1:predNumbers[1], (predNumbers[1]+1):predNumbers[2]] <- #alpha lambda
                t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                    G12 * weight) %*% as.matrix(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]])
              
              res[(predNumbers[1]+1):predNumbers[2], 1:predNumbers[1]] <- #alpha lambda
                t(t(as.data.frame(X[1:(nrow(X) / 3), 1:predNumbers[1]]) * 
                    G12 * weight) %*% as.matrix(X[(nrow(X) / 3 + 1):(nrow(X) * 2 / 3), (predNumbers[1]+1):predNumbers[2]]))
              
              res
            }
    )
  }
  
  validmu <- function(mu) {
    all(is.finite(mu)) && all(0 < mu)
  }
  
  devResids <- function (y, eta, wt, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    mu <- mu.eta(eta = eta)
    
    logLikFit <- (
      (y==1) * (log(PI) + logProbNotOne(lambda, alpha)) +
      (y!=1) * (log(1 - PI) + lgamma(y + 1 / alpha) - lgamma(y + 1) -
      lgamma(1 / alpha) - log(1 + alpha * lambda) / alpha +
      y * log(lambda) - y * log(lambda + 1 / alpha)) -
      logProbTrunc(lambda, alpha, PI)
    )
    
    yUnq <- unique(y)
    
    if (any(yUnq > 77)) {
      warning("Currently numerical deviance is unreliable for counts greater than 78.")
    }
    
    
    findL <- function(yNow) {
      stats::optim(
        par = if(TRUE) c(0, log(yNow), -15 + 2 * (yNow %in% 3:5) - 2 * (yNow %in% 6:9)) else c(0, log(yNow), -6),
        fn = function(x) {
          s <- x[1]
          l <- exp(x[2])
          a <- exp(x[3])
          
          sum(c(((l-l*(a*l+1)^(-1/a-1))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)-yNow) ,# s derivative
                s*((-(a*l+1)^(-1/a-1)-(-1/a-1)*a*l*(a*l+1)^(-1/a-2)+1)/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+(a*(-yNow-1/a))/(a*l+1)+yNow/l,# lambda derivative
                s*(-((l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1))-(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)+log(l*a+1)/a^2+(l*(-1/a-yNow))/(l*a+1)+yNow/a-digamma(1/a+yNow)/a^2+digamma(1/a)/a^2) ^ 2)#alpha derivative
        },
        gr = function(x) {
          s <- x[1]
          l <- exp(x[2])
          a <- exp(x[3])
          d12.21 <- ((a*l+1)^(2/a)*(a^2*l^2+2*a*l+1)+(a*l+1)^(1/a)*((-a^2-2*a-1)*l^2-2*a*l-2)+1)/((a*l+1)^(1/a+1)+(-a-1)*l-1)^2
          d13.31 <- (l^2*((l*a+1)^(1/a)*(l*a^2+(l+1)*a+1)*log(l*a+1)+(l*a+1)^(1/a)*((1-2*l)*a^2-l*a)-a^2))/(a^2*((l*a+1)^(1/a+1)-l*a-l-1)^2)
          d22.22 <- s*((-2*(-1/a-1)*a*(a*l+1)^(-1/a-2)-(-1/a-2)*(-1/a-1)*a^2*l*(a*l+1)^(-1/a-3))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)*a*(a*l+1)^(-1/a-2)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2+((-1/a-2)*(-1/a-1)*a^2*l*(a*l+1)^(-1/a-3)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2+(2*(-1/a-1)*a*l*(a*l+1)^(-1/a-2)*(-(a*l+1)^(-1/a-1)-(-1/a-1)*a*l*(a*l+1)^(-1/a-2)+1))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2+(2*(-1/a-1)^2*a^2*l^2*(a*l+1)^(-2/a-4)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^3)+((-1/a-1)*a*(a*l+1)^(-1/a-2))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-2)*(-1/a-1)*a^2*l*(a*l+1)^(-1/a-3))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)^2*a^2*l^2*(a*l+1)^(-2/a-4))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2-(a^2*(-yNow-1/a))/(a*l+1)^2-yNow/l^2
          d23.32 <- (((l*a+1)^(2/a)*(l^5*s*a^5+(5*l^4*s-l^4)*a^4+((-l^5+2*l^4+9*l^3)*s-l^4-3*l^3)*a^3+((-3*l^4+6*l^3+7*l^2)*s-3*l^3-3*l^2)*a^2+((-3*l^3+6*l^2+2*l)*s-3*l^2-l)*a+(2*l-l^2)*s-l)+(l*a+1)^(1/a)*(-l^5*s*a^5+((-3*l^5-5*l^4)*s+l^4)*a^4+((-3*l^5-10*l^4-9*l^3)*s+2*l^4+3*l^3)*a^3+((-l^5-7*l^4-13*l^3-7*l^2)*s+l^4+5*l^3+3*l^2)*a^2+((-2*l^4-5*l^3-8*l^2-2*l)*s+2*l^3+4*l^2+l)*a+(-l^3-l^2-2*l)*s+l^2+l))*log(l*a+1)+(l*a+1)^(2/a)*((3*l^3*yNow-l^5*s-2*l^4)*a^5+((3*l^3+9*l^2)*yNow+(2*l^5-8*l^4+2*l^3)*s-8*l^3)*a^4+((6*l^2+9*l)*yNow+(l^5+2*l^4-13*l^3+4*l^2)*s+l^4-10*l^2)*a^3+((3*l+3)*yNow+(2*l^4-2*l^3-6*l^2+2*l)*s+2*l^3-4*l)*a^2+((l^3-2*l^2)*s+l^2)*a)+(l*a+1)^(3/a)*((l^4-l^3*yNow)*a^5+(3*l^3-3*l^2*yNow)*a^4+(3*l^2-3*l*yNow)*a^3+(l-yNow)*a^2)+(l*a+1)^(1/a)*((-3*l^3*yNow+l^5*s+l^4)*a^5+((-6*l^3-9*l^2)*yNow+(3*l^5+8*l^4-4*l^3)*s+7*l^3)*a^4+((-3*l^3-12*l^2-9*l)*yNow+(3*l^5+7*l^4+13*l^3-8*l^2)*s-2*l^4+3*l^3+11*l^2)*a^3+((-3*l^2-6*l-3)*yNow+(l^5+4*l^4+6*l^3+6*l^2-4*l)*s-l^4-3*l^3+3*l^2+5*l)*a^2+((l^4+l^3+2*l^2)*s-l^3-l^2)*a)+l^3*yNow*a^5+((3*l^3+3*l^2)*yNow+2*l^3*s-2*l^3)*a^4+((3*l^3+6*l^2+3*l)*yNow+4*l^2*s-3*l^3-4*l^2)*a^3+((l^3+3*l^2+3*l+1)*yNow+2*l*s-l^3-3*l^2-2*l)*a^2)/(a^2*(l*a+1)^2*((l*a+1)^(1/a+1)-l*a-l-1)^3)
          d33.33 <- s*((2*(l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))^2)/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^3-((l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))^2/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))^2-(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))+l^2/(a*(l*a+1)^2))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-1))/(l*a+1)^2)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))^2)/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)+(2*l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-1))/(l*a+1)^2))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1))+(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))^2/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(-(log(l*a+1)/a^2-l/(a*(l*a+1)))^2/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))^2-(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))+l^2/(a*(l*a+1)^2))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-1))/(l*a+1)^2))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-yNow))/(l*a+1)^2-yNow/a^2+(2*digamma(1/a+yNow))/a^3-(2*digamma(1/a))/a^3+trigamma(1/a+yNow)/a^4-trigamma(1/a)/a^4
          
          f2 <- 2*c(((l-l*(a*l+1)^(-1/a-1))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)-yNow) ,# s derivative
                    s*((-(a*l+1)^(-1/a-1)-(-1/a-1)*a*l*(a*l+1)^(-1/a-2)+1)/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+(a*(-yNow-1/a))/(a*l+1)+yNow/l,# lambda derivative
                    s*(-((l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1))-(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)+log(l*a+1)/a^2+(l*(-1/a-yNow))/(l*a+1)+yNow/a-digamma(1/a+yNow)/a^2+digamma(1/a)/a^2)
          
          c(sum(f2 * c(0, d12.21, d13.31)),
            sum(f2 * c(d12.21, d22.22, d23.32)),
            sum(f2 * c(d13.31, d23.32, d33.33)))
        },
        method = "CG",
        control = list(maxit = 50000, abstol = .Machine$double.eps)
      )
    }
    fff <- function(yNow) {
      if (yNow < 3) return(0)
      
      xx <- findL(yNow)
      a <- exp(xx$par[3])
      l <- exp(xx$par[2])
      
      lgamma(yNow+1/a)-lgamma(1/a)-lgamma(yNow+1)-(yNow+1/a)*log(1+l*a)+yNow*log(l*a)-logProbGe2(l, a)
    }
    
    suppressWarnings(logLikIdeal <- sapply(yUnq, FUN = fff))
    names(logLikIdeal) <- yUnq
    
    logLikIdeal <- sapply(y, FUN = function(x) {
      logLikIdeal[names(logLikIdeal) == x]
    })
    diff <- logLikIdeal - logLikFit
    
    if (any(logLikFit > 0)) {
      warning("Dispersion parameter values are on the boundary of parameter space. Deviance residuals will be assigned 0 on these observations.")
      diff[logLikFit > 0]   <- 0
    } else if (any(diff < 0)) {
      warning("Numerical deviance finder found worse saturated likelihood than fitted model. Expect NA's in deviance/deviance residuals.")
      diff[diff < 0] <- 0
    }
    
    #diff <- ifelse(abs(diff) < 1e-1 & diff > 0, 0, diff)
    
    sign(y - mu) * sqrt(2 * wt * diff)
  }
  
  pointEst <- function (pw, eta, contr = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    P1 <- nbP1(lambda, alpha)
    
    N <- pw * (1 - P1) / probTrunc(lambda, alpha, PI)
    
    if(!contr) {
      N <- sum(N)
    }
    N
  }
  
  popVar <- function (pw, eta, cov, Xvlm, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    pr <- (1 - PI) * nbP0(lambda, alpha)
    pr <- 1 - pr / probNotOne(lambda, alpha)
    
    bigTheta0 <- -pw * piLink(eta[, 3], inverse = TRUE, deriv = 1) * 
      ((alpha * lambda + 1) * ((alpha * lambda + 1) ^ (1 / alpha + 1) - lambda)) /
      ((alpha * lambda + 1) * PI + (alpha * lambda + 1) ^ (1 / alpha + 1) + (-alpha - 1) * lambda -1 ) ^ 2
    
    bigTheta1 <- pw * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) *
      (1 - PI) * ((lambda * alpha + 1) ^ (1 / alpha) *
      (lambda ^ 2 * alpha ^ 2 + 2 * lambda * alpha + 1) * log(lambda * alpha + 1) +
      (lambda * alpha + 1) ^ (1 / alpha) * (-lambda ^ 2 * alpha ^ 2 - lambda * alpha) -lambda ^ 2 * alpha ^ 2) /
      (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha + 1) + (PI - 1) * lambda * alpha - lambda + PI - 1) ^ 2)
    
    bigTheta2 <- pw * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * 
      (PI - 1) * ((alpha * lambda + 1) ^ (1 / alpha + 1) - 1) /
      ((alpha * lambda + 1) ^ (1 / alpha + 1) + ((PI - 1) * alpha - 1) * lambda + PI - 1) ^ 2
    
    bigTheta <- t(c(bigTheta2, bigTheta1, bigTheta0) %*% Xvlm)
    
    f1 <-  t(bigTheta) %*% as.matrix(cov) %*% bigTheta
    f2 <-  sum(pw * (1 - pr) / (pr ^ 2))
    
    f1 + f2
  }
  
  dFun <- function (x, eta, type = "trunc") {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-  piLink(eta[, 3], inverse = TRUE)
    P1 <- nbP1(lambda, alpha)
    normTerm <- probTrunc(lambda, alpha, PI)
    prob_not_one <- probNotOne(lambda, alpha)
    
    switch (type,
      "trunc" = {
        ifelse(x == 1, PI,
        stats::dnbinom(x = x, mu = lambda, size = 1 / alpha) * (1 - PI) / prob_not_one)
      },
      "nontrunc" = {
        ifelse(x == 1, PI * (1 - P1) / normTerm,
        stats::dnbinom(x = x, mu = lambda, size = 1 / alpha) * (1 - PI) / normTerm)
      }
    )
  }
  
  simulate <- function(n, eta, lower = 0, upper = Inf) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    PI     <-     piLink(eta[, 3], inverse = TRUE)
    
    P0 <- nbP0(lambda, alpha)
    P1 <- nbP1(lambda, alpha)
    prob_not_one <- probNotOne(lambda, alpha)
    
    CDF <- function(x) {
      ifelse(x == Inf, 1, 
      ifelse(x < 0, 0, 
      ifelse(x < 1, (1 - PI) * P0 / prob_not_one,
      PI + ((1 - PI) / prob_not_one) *
      (stats::pnbinom(x, mu = lambda, size = 1 / alpha) - P1)
      )))
    }
    lb <- CDF(lower)
    ub <- CDF(upper)
    p_u <- stats::runif(n, lb, ub)
    sims <- rep(0, n)
    cond <- CDF(sims) < p_u
    while (any(cond)) {
      sims[cond] <- sims[cond] + 1
      cond <- CDF(sims) < p_u
    }
    sims
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      init <- log(abs((observed / weighted.mean(observed, priorWeights) - 1) / observed) + .1)
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12)),
        family$links[[2]](ifelse(init < -.5, .1, init + .55)),
        family$links[[3]](weighted.mean(observed == 1, priorWeights) * (.5 + .5 * (observed == 1)) + .01)
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights)),
        family$links[[2]](abs((cov.wt(cbind(observed, observed), wt = priorWeights, method = "ML")$cov[1,1] / weighted.mean(observed, priorWeights) - 1) / weighted.mean(observed, priorWeights)) + .1),
        family$links[[3]](weighted.mean(observed == 1, priorWeights) + .01)
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
      if ("(Intercept):alpha" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[2], rep(0, attr(Xvlm, "hwm")[2] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[2] / attr(Xvlm, "hwm")[2], attr(Xvlm, "hwm")[2]))
      }
      if ("(Intercept):pi" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[3], rep(0, attr(Xvlm, "hwm")[3] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[3] / attr(Xvlm, "hwm")[3], attr(Xvlm, "hwm")[3]))
      }
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "Hurdleztnegbin",
      etaNames  = c("lambda", "alpha", "pi"),
      simulate  = simulate,
      getStart  = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

# These functions are only used internally in the package so there is no need for documenting them
#' @importFrom stats runif
#' @importFrom stats var
#' @importFrom stats quantile
#' @importFrom stats qnorm
singleRcaptureinternalpopulationEstimate <- function(y, X, grad,
                                                     beta, weights,
                                                     hessian, family,
                                                     eta, popVar,
                                                     control,
                                                     Xvlm, W, formulas,
                                                     sizeObserved,
                                                     modelFrame, 
                                                     cov, offset,
                                                     weightsFlag) {
  hwm <- attr(Xvlm, "hwm")
  siglevel <- control$alpha
  numboot <- control$B
  sc <- qnorm(p = 1 - siglevel / 2)
  
  if (popVar == "analytic") {
    strappedStatistic <- "No bootstrap performed"
    N <- family$pointEst(pw = weights, eta = eta, y = y)
    if (is.null(cov)) {
      
      cov <- switch(control$covType,
        "observedInform" = {
          tryCatch(
            expr = {suppressWarnings(solve(-hessian(beta)))},
            error = function (e) "e"
          )
        },
        "Fisher" = solve(singleRinternalMultiplyWeight(X = Xvlm, W = W) %*% Xvlm)
      )
      
      if (isTRUE(cov == "e")) {
        warning(
          "The analytically computed hessian of log-likelihood is numerically singular.",
          "Fisher information matrix will be used in population size estimation instead.",
          sep = " "
        )
        control$covType <- "Fisher"
        cov <- solve(singleRinternalMultiplyWeight(X = Xvlm, W = W) %*% Xvlm)
      }
    }
    
    variation <- as.numeric(family$popVar(
      eta = eta, 
      pw = weights,
      cov = cov,
      Xvlm = if (family$family == "zelterman") X else Xvlm,
      y = y
    ))
    
    if (!is.finite(variation))
      stop("Computed variance is infinite/NaN/NULL")
    
    sd <- sqrt(variation)
    if (control$sd == "normalMVUE") {
      sd <- sd / (sqrt(2 / (sizeObserved - 1)) * 
      exp(lgamma(sizeObserved / 2) - lgamma((sizeObserved - 1) / 2)))
    }
    
    G <- exp(sc * sqrt(log(1 + variation / ((N - sizeObserved) ^ 2))))
    confidenceInterval <- data.frame(t(data.frame(
      "normal" = c(lowerBound = max(N - sc * sd, sizeObserved), 
                   upperBound = N + sc * sd),
      "logNormal" = c(lowerBound = sizeObserved + max((N - sizeObserved) / G, 0), 
                      upperBound = sizeObserved + (N - sizeObserved) * G)
    )))
  } else if (grepl("bootstrap", popVar, fixed = TRUE)) {
    
    N <- family$pointEst(pw = weights, eta = eta, y = y)
    
    if (control$cores > 1) {
      funBoot <- switch(
        control$bootType,
        "parametric" = parBootMultiCore,
        "semiparametric" = semparBootMultiCore,
        "nonparametric" = noparBootMultiCore
      )
      
      strappedStatistic <- funBoot(
        family = family, formulas = formulas,
        y = y, X = X, hwm = hwm,
        beta = beta, weights = weights, numboot = numboot,
        eta = eta, cores = control$cores,
        method = control$fittingMethod,
        controlBootstrapMethod = control$bootstrapFitcontrol,
        N = N, Xvlm = Xvlm, modelFrame = modelFrame,
        offset = offset, weightsFlag = weightsFlag,
        visT = control$bootstrapVisualTrace
      )
    } else {
      funBoot <- switch(
        control$bootType,
        "parametric" = parBoot,
        "semiparametric" = semparBoot,
        "nonparametric" = noparBoot
      )
      
      strappedStatistic <- funBoot(
        family = family, formulas = formulas,
        y = y, X = X, hwm = hwm,
        beta = beta, weights = weights, numboot = numboot,
        eta = eta, trace = control$traceBootstrapSize,
        visT = control$bootstrapVisualTrace,
        method = control$fittingMethod,
        controlBootstrapMethod = control$bootstrapFitcontrol,
        N = N, Xvlm = Xvlm, modelFrame = modelFrame,
        offset = offset, weightsFlag = weightsFlag
      )
    }

    if (N < stats::quantile(strappedStatistic, .05)) {
      warning("bootstrap statistics unusually high, try higher maxiter/lower epsilon for fitting bootstrap samples (bootstrapFitcontrol)\n")
    } else if (N > stats::quantile(strappedStatistic, .95)) {
      warning("bootstrap statistics unusually low, try higher maxiter/lower epsilon for fitting bootstrap samples (bootstrapFitcontrol)\n")
    }
    if (max(strappedStatistic) > N ^ 1.5) {
      warning("Outlier(s) in statistics from bootstrap sampling detected, consider higher maxiter/lower epsilon for fitting bootstrap samples (bootstrapFitcontrol)\n")
    }
    
    variation <- stats::var(strappedStatistic)
    
    
    if (!is.finite(variation))
      stop("Computed variance is infinite/NaN/NULL")
    
    sd <- sqrt(variation)
    if (control$sd == "normalMVUE") {
      sd <- sd / (sqrt(2 / (sizeObserved - 1)) * exp(lgamma(sizeObserved / 2) - lgamma((sizeObserved- 1) / 2)))
    }
    
    if (control$confType == "percentile") {
      confidenceInterval <- stats::quantile(strappedStatistic,
                                            c(siglevel / 2,
                                              1 - siglevel / 2))
      names(confidenceInterval) <- c("lowerBound", "upperBound")
    } else if (control$confType == "normal") {
      G <- exp(sc * sqrt(log(1 + variation / ((N - sizeObserved) ^ 2))))
      
      confidenceInterval <- data.frame(t(data.frame(
        "normal" = c(lowerBound = max(N - sc * sd, 
                                           sizeObserved), 
                          upperBound = N + sc * sd),
        "logNormal" = c(lowerBound = max(sizeObserved + (N - sizeObserved) / G, 
                                            sizeObserved), 
                           upperBound = sizeObserved + (N - sizeObserved) * G)
      )))
    } else if (control$confType == "basic") {
      confidenceInterval <- 2 * N - stats::quantile(strappedStatistic,
                                                    c(1 - siglevel / 2, 
                                                      siglevel / 2))
      names(confidenceInterval) <- c("lowerBound", "upperBound")
    }
  }
  
  structure(
    list(
      pointEstimate = N,
      variance = variation,
      confidenceInterval = confidenceInterval,
      boot = if (isTRUE(control$keepbootStat)) strappedStatistic else NULL,
      control = control
    ),
    class = "popSizeEstResults"
  )
}
# multiparameter IRLS
singleRcaptureinternalIRLSmultipar <- function(dependent,
                                               family,
                                               formulas,
                                               covariates,
                                               etaStart,
                                               weights,
                                               maxiter = 10000,
                                               eps = .Machine$double.eps,
                                               silent = FALSE,
                                               trace = 0,
                                               stepsize = 1,
                                               momentumFactor,
                                               momentumActivation,
                                               check,
                                               epsWeights,
                                               crit,
                                               printOften,
                                               saveLog,
                                               offset,
                                               ...) {
  dg <- 8
  converged <- FALSE
  
  # Lowering stepsize to about .3 usually helps a great deal in IRLS fitting
  if (!silent && family$family == "zotnegbin" && stepsize == 1) {
    cat("Zero one truncated negative binomial distribution is prone to taking alpha parameter to infinity.",
        "\nConsider lowering stepsize control parameter if fitting fails.\n")
  }
  
  mu.eta    <- family$mu.eta
  validmu   <- family$validmu
  variance  <- family$variance
  famName   <- family$family
  Zfun      <- family$funcZ
  Wfun      <- family$Wfun
  prior     <- as.numeric(weights)
  dependent <- as.numeric(dependent)
  
  iter <- 1
  step <- NULL
  betaPrev <- NULL
  beta <- rep(0, NCOL(covariates))
  
  logLike <- family$makeMinusLogLike(
    y      = dependent, 
    X      = covariates, 
    weight = prior, 
    offset = offset
  )
  
  grad    <- family$makeMinusLogLike(
    y      = dependent, 
    X      = covariates, 
    weight = prior, 
    deriv  = 1, 
    offset = offset
  )
  
  logg <- NULL
  if (isTRUE(saveLog)) {
    logg <- data.frame()
  }
  
  addToLog <- expression(
    if (isTRUE(saveLog)) {
      if (trace == 1) {logg[NROW(logg) + 1, c(1, 2, 3)] <- c(iter, halfstepsizing, L)}
      if (trace == 2) {logg[NROW(logg) + 1, c(1:3, 4:(3+length(beta)))] <- c(iter, halfstepsizing, L, beta)}
      if (trace  > 2) {logg[NROW(logg) + 1, c(1:3, 4:(3+2*length(beta)))] <- c(iter, halfstepsizing, L, beta, grad(beta))}
    }
  )
  
  traceGreaterThanFourMessegeExpr <- expression(
    if (trace > 4) {
      cat(sep = " ", "\nAlgorithm will terminate if one of following conditions will be met:\n")
      if ("abstol" %in% crit) {
        cat("The increase to minus log-likelihood will be below chosen value of epsilon", eps, "\n")
      }
      if ("reltol" %in% crit) {
        cat("The relative increase to minus log-likelihood will be below chosen value of epsilon", eps, 
            "value at current step", format((L - LPrev) / LPrev, scientific = FALSE, digits = dg), "\n")
      }
      if ("coef" %in% crit) {
        cat("Maximum change to the vector of regression parameters will be below the chosen value of epsilon.\nAt current step the highest change was:", 
            format(max(abs(beta - betaPrev)), scientific = FALSE, digits = dg))
      }
    }
  )
  
  convergence <- expression(
    any(
      if ("abstol" %in% crit) {
        L - LPrev < eps
      } else FALSE,
      if ("reltol" %in% crit) {
        L - LPrev < eps * LPrev
      } else FALSE,
      if ("coef" %in% crit) {
        max(abs(beta - betaPrev)) < eps
      } else FALSE
    )
  )
  
  parNum <- length(family$etaNames)
  W <- prior
  LPrev <- -Inf
  L   <- -Inf
  eta <- etaStart
  
  while (!converged & (iter < maxiter)) {
    halfstepsizing <- FALSE
    mu <- mu.eta(eta = eta, ...)
    if (!validmu(mu)) {
      mu <- mu.eta(eta = eta, ...)
      stop(paste0(
        "Fit error infinite values reached consider another model, mu is too close to zero/infinity.\n"
      ))
    }

    if (any(!is.finite(
      sapply(1:length(family$etaNames), FUN = function(x) {
        family$links[[x]](eta[, x], inverse = TRUE)
      })
    )))
      stop(paste0(
        "Infinite values of distribution ",
        "parameters obtained for some IRLS iteration.\n"
      ))
    
    WPrev <- W
    tryCatch(
      expr = {
        W <- Wfun(prior = prior, eta = eta, y = dependent)
      },
      error = function (e) {
        stop(
          "Working weight matrixes at iteration: ",
          iter,
          " could not have been computed."
        )
      }
    )
    
    if (any(!is.finite(W))) {
      if (!silent) {
        warning(paste0(
          "NA's or NaN's or infinite values in weights matrixes ",
          "detected IRLS may not work propperly.\n",
          "Replacing these values by weightsEpsilon control parameter.\n"
        ))
      }
      W[!is.finite(W)] <- epsWeights
    }
    if (check) {
      W[, (1:parNum) ^ 2] <- ifelse(
        W[, (1:parNum) ^ 2] < epsWeights, 
        epsWeights, 
        W[, (1:parNum) ^ 2]
      )
    }
    
    z <- NULL
    try(
      expr = {z <- eta + Zfun(eta = eta, weight = W, y = dependent, prior = prior) - offset}
    )
    
    XbyW     <- singleRinternalMultiplyWeight(X = covariates, W = W)
    A        <- XbyW %*% covariates
    B        <- XbyW %*% as.numeric(z)
    # The code above does basically:
    # A <- t(Xvlm) %*% WW %*% (Xvlm)
    # B <- t(Xvlm) %*% WW %*% (as.numeric(z))
    # but does not allocate memmory to the full weight matrix
    
    if (any(!is.finite(A)) || any(!is.finite(B))) {
      stop("IRLS step could not have been computed at iteration ", iter, call. = FALSE)
    }
    
    betaPrev <- beta
    stepPrev <- step
    step     <- solve(A, B)
    
    if (!is.null(betaPrev)) {
      step <- step - betaPrev
      beta <- betaPrev + stepsize * 
        (step + if ((is.null(stepPrev) | !momentumFactor)) 0 else {
          if (L-LPrev < momentumActivation) momentumFactor * stepPrev else 0
        })
    } else {
      beta <- step
    }
    
    eta <- covariates %*% beta
    eta <- matrix(eta, ncol = parNum) + offset
    
    LPrev <- L
    L     <- -logLike(beta)
    
    if ((iter - 1) %% printOften == 0) {
      if (trace > 0) {cat(sep = "", "Iteration number ", iter, 
                          " log-likelihood: ", 
                          format(L, scientific = FALSE, 
                                 digits = dg))}
      if (trace > 1) {cat(sep = " ", "\nParameter vector: ", 
                          format(beta, scientific = FALSE, 
                                 digits = dg))}
      if (trace > 2) {cat(sep = " ", "\nlog-likelihood reduction: ", 
                          format(L - LPrev, scientific = FALSE, 
                                 digits = dg))}
      if (trace > 3) {cat(sep = " ", "\nValue of gradient at current step:\n", 
                          format(grad(beta), scientific = FALSE, 
                                 digits = dg))}
      eval(traceGreaterThanFourMessegeExpr)
      eval(addToLog)
    }

    if (isTRUE(L < LPrev) || is.infinite(L) || is.nan(L)) {
      halfstepsizing <- TRUE
      h <- step <- stepsize * (betaPrev - beta)
      if ((trace > 0) && (iter - 1) %% printOften == 0) {
        cat("\nTaking a modified step....\n")
      }
      repeat {
        h    <- step <- h / 2
        beta <- betaPrev - h
        L    <- -logLike(beta)
        if (isTRUE(L > LPrev) && is.finite(L)) {
          break
        }

        if (isTRUE(max(abs(h)) < .Machine$double.eps / 10^6)) {
          halfstepsizing <- FALSE
          if (isTRUE(L < LPrev)) {
            if (!silent) {
              warning("IRLS half-stepping terminated because the step is too small.")
            }
            L    <- LPrev
            beta <- betaPrev
          } else {
            if (!silent) {
              warning("IRLS half-stepping terminated because change to score function was too small.")
            }
          }
          break
        }
      }
      if ((iter - 1) %% printOften == 0) {
        if (trace > 0) {cat(sep = "", "Iteration number ", iter, 
                            " log-likelihood: ", 
                            format(L, scientific = FALSE, 
                                   digits = dg))}
        if (trace > 1) {cat(sep = " ", "\nParameter vector: ", 
                            format(beta, scientific = FALSE, 
                                   digits = dg))}
        if (trace > 2) {cat(sep = " ", "\nlog-likelihood reduction: ", 
                            format(L - LPrev, scientific = FALSE, 
                                   digits = dg))}
        if (trace > 3) {cat(sep = " ", "\nValue of gradient at current step:\n", 
                            format(grad(beta), scientific = FALSE, 
                                   digits = dg))}
        eval(traceGreaterThanFourMessegeExpr)
        eval(addToLog)
      }
    }
    
    eta <- covariates %*% beta
    eta <- matrix(eta, ncol = parNum) + offset
    
    if (trace > 0 && (iter - 1) %% printOften == 0) {cat(sep = "", "\n----\n")}
    converged <- eval(convergence)

    if (!converged && (iter + 1 <= maxiter)) {
      iter <- iter + 1
    } else if ((L < LPrev) & converged) {
      beta <- betaPrev
      L <- LPrev
      W <- WPrev
    }
    
    if (converged && halfstepsizing && !silent) {
      warning("Convergence at halfstepsize")
    }
    
  }
  
  if(iter == maxiter && !converged && !silent) {
    warning("Fitting algorithm (IRLS) has not converged")
  }
  
  if (trace > 4) {
    hhh <- family$makeMinusLogLike(y = dependent, X = covariates, weight = prior, deriv = 2, offset = offset)(beta)
    cat("Value of analytically computed hessian at fitted regression coefficients:\n")
    print(hhh)
    cat("The matrix above has the following eigen values:\n", 
        eigen(hhh, only.values = TRUE)$values, "\n", sep = " ")
    
    if (isTRUE(saveLog)) attr(logg, "hessian") <- hhh
  }
  
  if (isTRUE(saveLog)) {
    if (trace == 1) colnames(logg) <- c("iterationNumber", "halfStep", 
                                        "Log-likelihood")
    if (trace == 2) colnames(logg) <- c("iterationNumber", "halfStep", 
                                        "Log-likelihood", colnames(covariates))
    if (trace  > 2) colnames(logg) <- c("iterationNumber", "halfStep", 
                                        "Log-likelihood", colnames(covariates), 
                                        paste0("gradient -- ", colnames(covariates)))
  }
  
  mu <- mu.eta(eta = eta, ...)
  if (!validmu(mu)) {
    stop("Fit error infinite values reached consider another model, mu is too close to zero/infinity")
  }

  list(coefficients = beta, iter = iter, weights = W, logg = logg)
}
#' @importFrom stats terms
singleRinternalGetXvlmMatrix <- function(X, formulas, parNames, contrasts = NULL) {
  nPar <- length(parNames)
  Xses <- list()
  
  for (k in 1:nPar) {
    # Contrasts could be added in this place
    if (length(attr(terms(formulas[[k]], data = X), "term.labels")) != 0) {
      Xses[[k]] <- tryCatch(
        expr = {model.matrix(
          terms(formulas[[k]], data = X),
          data = X
        )},
        error = function (e) {
          ff <- formulas[[k]]
          ff[[2]] <- NULL
          model.matrix(
            terms(ff, data = X),
            data = X
          )
        }
      )
    } else {
      Xses[[k]] <- model.matrix(
        ~ 1,
        X
      )
      if (attr(terms(formulas[[k]], data = X), "intercept") == 0)
        warning(paste0(
          "One of formulas for the model has no variable ",
          "and no intercept and will be coerced to ~ 1."
        ))
    }
    if (k != 1) {
      colnames(Xses[[k]]) <- paste0(colnames(Xses[[k]]), ":", parNames[k])
    }
  }
  hwm <- sapply(Xses, ncol)
  
  Xvlm <- matrix(0, nrow = nPar * nrow(X), ncol = sum(hwm))
  colnames(Xvlm) <- unlist(sapply(X = Xses, FUN = colnames))
  row <- 0
  col <- 0
  for (k in Xses) {
    Xvlm[(row + 1):(row + nrow(k)), (col + 1):(col + ncol(k))] <- as.matrix(k)
    row <- row + nrow(k)
    col <- col + ncol(k)
  }
  attr(Xvlm, "hwm") <- hwm
  Xvlm
}
#' @importFrom stats reformulate
singleRinternalMergeFormulas <- function(ff) {
  # This code was inspired by: https://stevencarlislewalker.wordpress.com/2012/08/06/merging-combining-adding-together-two-formula-objects-in-r/
  env <- environment(ff[[1]])
  for (k in 1:length(ff)) { 
    # eliminate left hand side
    if (length(ff[[k]]) == 3) {
      resp <- ff[[k]][[2]]
      ff[[k]][[2]] <- NULL
    }
  }
  dotCheck <- sapply(ff, FUN = function(x) {x == ~.})
  if (any(dotCheck)) {
    out <- reformulate(".", resp)
  } else {
    for (k in 1:length(ff)) { 
      # extract right hand side and create string vector
      ff[[k]] <- strsplit(deparse(ff[[k]][[2]]), " \\+ ")[[1]]
    }
    ff <- unlist(ff)
    out <- reformulate(ff, resp)
  }
  
  environment(out) <- env
  out
}
# This is almost certainly an overkill but it supports arbitrary number of linear predictors
singleRinternalMultiplyWeight <- function (X, W, ...) {
  hwm <- attr(X, "hwm")
  thick <- sqrt(ncol(W))
  XbyW <- matrix(0, nrow = ncol(X), ncol = nrow(X))
  
  wch <- c(0, cumsum(hwm))
  whichVector <- t(matrix(1:(thick ^ 2), ncol = thick))
  
  index1 <- 1:(nrow(X) / thick)
  for (i in 1:thick) {
    index <- 1:(nrow(X) / thick)
    for (j in 1:thick) {
      XbyW[(wch[i]+1):wch[i+1], index] <- t(X[index1, (wch[i]+1):wch[i+1]] * W[, whichVector[i, j]])
      index <- index + nrow(X) / thick
    }
    index1 <- index1 + nrow(X) / thick
  }
  XbyW
}

#' @method print singleRStaticCountData
#' @importFrom stats AIC BIC
#' @exportS3Method 
print.singleRStaticCountData <- function(x, ...) {
  cat("Call: ")
  print(x$call)
  cat("\nCoefficients:\n")
  print(coef(x))
  cat("\nDegrees of Freedom:", x$df.null, 
      "Total (i.e NULL);", x$dfResidual, 
      "Residual")
  
  cat("\nAIC: ", signif(AIC(x)), 
      "\nBIC: ", signif(BIC(x)), 
      "\nResidual deviance: ", signif(x$deviance))
  cat("\n-----------------------------------\n",
      "Population size estimation results:\n",
      sep = "")
  print(x$populationSize)

  invisible(x)
}

#' @method print singleRfamily
#' @exportS3Method 
print.singleRfamily <- function(x, hideFormulas = c(FALSE, TRUE), ...) {
  cat("Family of distributions:", x$family,
      "\nNames of parameters:", x$etaNames, 
      "\nLinks:", attr(x$links, "linkNames"), 
      sep = " ")
  if (!is.null(x$extraInfo)) {
    
    nn <- c("mean", "variance", "popSizeEst", "meanTr", "varianceTr")
    if (any(xx <- is.na(x$extraInfo[nn])))
      warning(cat("The: ", nn[is.na(xx)], sep = " ",
                  "slots are empty for family: ", x$family))
    
    if (!hideFormulas[1])
      cat("\n--\nFormula for mean in this distribution: ", x$extraInfo["mean"],
          "\nFormula for variance in this distribution: ", x$extraInfo["variance"],
          "\nFormula for population size estimation in this distribution: sum(", 
          x$extraInfo["popSizeEst"], ")",
          sep = "")
    
    if (!hideFormulas[2])
      cat("\n\nFormula for mean in truncated distribution: ", x$extraInfo["meanTr"],
          "\nFormula for variance in truncated distribution: ", x$extraInfo["varianceTr"],
          sep = "")
  }
  invisible(x)
}

#' @rdname singleRmodels
#' @export
zotpoisson <- function(lambdaLink = c("log", "neglog"),
                       ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink)
  
  lambdaLink <- switch (lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  links[1] <- c(lambdaLink)
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    lambda <- lambdaLink(eta, inverse = TRUE)
    
    if (!deriv) {
      switch (type,
        "nontrunc" = lambda,
        "trunc" = (lambda - lambda * exp(-lambda)) / (1 - exp(-lambda) - lambda * exp(-lambda))
      )
    } else {
      switch (type,
        "nontrunc" = cbind(lambdaLink(eta, inverse = TRUE, deriv = 1)),
        "trunc" = cbind(
          lambdaLink(eta, inverse = TRUE, deriv = 1) *
            (exp(2 * lambda) + (-lambda ^ 2 - 2) * exp(lambda) + 1) /
            (exp(lambda) - lambda - 1) ^ 2
        )
      )
    }
  }

  variance <- function(eta, type = "nontrunc", ...) {
    lambda <- lambdaLink(eta, inverse = TRUE)
    switch (type,
    "nontrunc" = lambda,
    "trunc" = (lambda - lambda * exp(-lambda)) / (1 - exp(-lambda) - lambda * exp(-lambda))
    )
  }
  
  Wfun <- function(prior, eta, y, ...) {
    iddx <- y > 1
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)[iddx]
    Ey <- mu.eta(eta)[iddx]
    
    G11 <- iddx
    
    G11[iddx] <- (-((lambda - Ey) * exp(lambda) + exp(lambda) + Ey - 1) /
    (lambda * (exp(lambda) - lambda - 1)) + 
    ((lambda - Ey) * exp(lambda) + (Ey - 1) * lambda + Ey) /
    (lambda ^ 2 * (exp(lambda) - lambda - 1)) +
    ((exp(lambda) - 1) * ((lambda - Ey) * exp(lambda) + (Ey - 1) * lambda + Ey)) /
    (lambda * (exp(lambda) - lambda - 1) ^ 2)) *
    lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)[iddx] ^ 2
    
    matrix(-prior * G11, ncol = 1, dimnames = list(rownames(eta), c("lambda")))
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    iddx <- y > 1
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)[iddx]
    
    res <- iddx
    res[iddx] <- (-lambda/(-lambda -1 + exp(lambda)) + y[iddx] / lambda - 1) *
    lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)[iddx] * prior[iddx] / weight[iddx, ]
    res
  }

  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, ...) {
    y <- as.numeric(y)
    iddx <- y > 1
    if (is.null(weight)) {
      weight <- 1
    }
    
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X)))
    }
    
    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
      function(beta) {
        eta <- as.matrix(X) %*% beta + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        -sum(weight * iddx * (y * log(lambda) - lambda - lgamma(y + 1) -
        log(1 - exp(-lambda) - lambda * exp(-lambda))))
      },
      function(beta) {
        eta <- as.matrix(X) %*% beta + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)[iddx]
        G1 <- iddx
        
        G1[iddx] <- (-lambda/(-lambda -1 + exp(lambda)) + y[iddx] / lambda - 1) * 
          weight[iddx] * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)[iddx]
        if (NbyK) {
          return(as.data.frame(X) * G1)
        }
        if (vectorDer) {
          return(matrix(G1, ncol = 1))
        }
        
        t(as.matrix(X)) %*% G1
      },
      function(beta) {
        eta <- as.matrix(X) %*% beta + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)[iddx]
        
        G1 <- iddx
        
        G1[iddx] <- (-lambda/(-lambda - 1 + exp(lambda)) + y[iddx] / lambda - 1) * weight[iddx] *
               lambdaLink(eta[, 1], inverse = TRUE, deriv = 2)[iddx]
        
        G11 <- iddx
        G11[iddx] <- (-((lambda - y[iddx]) * exp(lambda)+exp(lambda) + y[iddx] - 1) /
        (lambda * (exp(lambda) - lambda - 1)) + 
         ((lambda - y[iddx]) * exp(lambda) + (y[iddx] - 1) * lambda + y[iddx]) /
        (lambda ^ 2 * (exp(lambda) - lambda - 1)) +
        ((exp(lambda) - 1) * ((lambda - y[iddx]) * exp(lambda) + (y[iddx] - 1) * lambda + y[iddx])) /
        (lambda * (exp(lambda) - lambda - 1) ^ 2)) * weight[iddx] *
        lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)[iddx] ^ 2
              
        
        t(X) %*% as.matrix(t(t(as.data.frame(X) * (G1 + G11))))
      }
    )
  }

  validmu <- function(mu) {
    (sum(!is.finite(mu)) == 0) && all(0 < mu)
  }

  devResids <- function(y, eta, wt, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    iddx <- y > 1
    
    inverseFunction <- function(y) {stats::uniroot(
      f = function(x) {mu.eta(x) - y}, 
      lower = -log(y), upper = y * 10, 
      tol = .Machine$double.eps
    )$root}
    
    # only compute predictors in saturated model for unique values of y
    # this is faster because stats::uniroot in slow whereas lamW::lambertW0 is really fast
    yUnq <- unique(y[iddx])
    lambdaSat <- sapply(yUnq, FUN = function(x) ifelse(x == 2, -Inf, inverseFunction(x)))
    
    idealLambda <- tryCatch(
      expr = {
        suppressWarnings(sapply(yUnq, 
          FUN = function(x) ifelse(x == 2, -Inf, inverseFunction(x))
        ))
      },
      error = function (e) {
        warning("Deviance residuals could not have been computed and zero vector will be returned instead.", call. = FALSE)
        NULL
      }
    )
    if (is.null(idealLambda)) {
      return(rep(0, length(y)))
    }
    
    lambdaSat <- lambdaLink(sapply(y[iddx], FUN = function(x) lambdaSat[yUnq == x]), inverse = TRUE)
    
    diff <- iddx
    lambda <- lambda[iddx]
    diff[iddx] <- y[iddx] * log(lambda) - lambda - log(1 - exp(-lambda) - lambda * exp(-lambda)) -
    ifelse(y[iddx] == 2, lgamma(y[iddx] + 1), y[iddx] * log(lambdaSat) - lambdaSat - 
    log(1 - exp(-lambdaSat) - lambdaSat * exp(-lambdaSat)))
    
    if (any(diff > 0)) {
      warning(paste0(
        "Some of differences between log likelihood in saturated model",
        " and fitted model were positive which indicates either:\n",
        "(1): A very good model fit or\n",
        "(2): Incorrect computation of saturated model",
        "\nDouble check deviance before proceeding"
      ))
      diff[diff > 0] <- -0
    }

    sign(y - mu.eta(eta = eta)) * sqrt(-2 * wt * diff)
  }

  pointEst <- function (pw, eta, contr = FALSE, y, ...) {
    iddx <- y > 1
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    N <- pw * (iddx * (1 - lambda * exp(-lambda)) / 
    (1 - exp(-lambda) - lambda * exp(-lambda)) + (1 - iddx) * 1)
    
    if(!contr) {
      N <- sum(N)
    }
    
    N
  }

  popVar <- function (pw, eta, cov, Xvlm, y, ...) {
    iddx <- y > 1
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)[iddx]
    Xvlm <- as.data.frame(Xvlm)
    
    f1 <- -t(Xvlm[iddx,, drop = FALSE]) %*% 
      (as.numeric(pw[iddx] * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)[iddx] * 
                                     (exp(lambda) - 1) / (exp(lambda) - lambda - 1) ^ 2))
    
    f1 <- t(f1) %*% as.matrix(cov) %*% f1
    
    f2 <- sum((pw[iddx] * (1 - lambda * exp(-lambda)) ^ 2 * 
    (exp(-lambda) + lambda * exp(-lambda)) / (1 - exp(-lambda) - lambda * exp(-lambda)) ^ 2))
    
    f1 + f2
  }
  
  dFun <- function (x, eta, type = c("trunc", "nontrunc")) {
    if (missing(type)) type <- "trunc"
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    switch (type,
      "trunc" = {
        stats::dpois(x = x, lambda = lambda) / 
        (1 - stats::dpois(x = 0, lambda = lambda) - 
        stats::dpois(x = 1, lambda = lambda))
      },
      "nontrunc" = stats::dpois(x = x, lambda = lambda)
    )
  }

  simulate <- function(n, eta, lower = 1, upper = Inf) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    lb <- stats::ppois(lower, lambda)
    ub <- stats::ppois(upper, lambda)
    p_u <- stats::runif(n, lb, ub)
    sims <- stats::qpois(p_u, lambda)
    sims
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12))
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights))
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "zotpoisson",
      etaNames  = c("lambda"),
      simulate  = simulate,
      getStart  = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

#' @rdname singleRmodels
#' @importFrom lamW lambertW0
#' @export
ztoipoisson <- function(lambdaLink = c("log", "neglog"), 
                        omegaLink = c("logit", "cloglog", "probit"), 
                        ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  if (missing(omegaLink))  omegaLink <- "logit"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink, omegaLink)
  
  lambdaLink <- switch(lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  omegaLink <- switch(omegaLink,
    "logit" = singleRinternallogitLink,
    "cloglog" = singleRinternalcloglogLink,
    "probit" = singleRinternalprobitLink
  )
  
  links[1:2] <- c(lambdaLink, omegaLink)
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    omega  <-  omegaLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    if (!deriv) {
      switch (type,
        "nontrunc" = omega + lambda * (1 - omega),
        "trunc" = exp(lambda) * (omega + lambda - omega * lambda) / (exp(lambda) - 1 + omega)
      )
    } else {
      switch (type,
        "nontrunc" = {
          matrix(c(
            1 - omega, 
            1 - lambda
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
            omegaLink(eta[, 2], inverse = TRUE, deriv = 1)
          ), ncol = 2)
        },
        "trunc" = {
          matrix(c(
            (1 - omega) * exp(lambda) * (exp(lambda) + (omega - 1) * lambda - 1) /
            (exp(lambda) + omega - 1) ^ 2, 
            -exp(lambda) * ((lambda - 1) * exp(lambda) + 1) /
            (omega + exp(lambda) - 1) ^ 2
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
             omegaLink(eta[, 2], inverse = TRUE, deriv = 1)
          ), ncol = 2)
        }
      )
    }
  }
  
  variance <- function(eta, type = "nontrunc", ...) {
    omega  <-  omegaLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    switch (type,
    "nontrunc" = omega + (1 - omega) * lambda * (lambda + 1),
    "trunc" = ((exp(lambda) * omega + lambda * (1 - omega)) / (exp(lambda) - 1 + omega) +
    (1 - omega) * lambda * ((1 + lambda) * exp(lambda) - 1) / (exp(lambda) - 1 + omega))
    ) - mu.eta(type = type, eta = eta) ^ 2
  }
  
  Wfun <- function(prior, y, eta, ...) {
    omega  <-  omegaLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    # expected for I's
    z <- (exp(lambda) * omega + lambda * (1 - omega)) / (exp(lambda) - 1 + omega)
    Ey <- mu.eta(eta)
    ## z here is the prob of 1 and XXX is expected for (1-I(y=1))*y
    XXX <- Ey - z
    
    # omega^2 derivative
    G00 <- (-(z * (exp(lambda) - lambda) ^ 2) /
    (exp(lambda) * omega + lambda * (1 - omega)) ^2 +
    1 / (omega + exp(lambda) - 1) ^ 2 - (1 - z) / (1 - omega) ^ 2)
    
    # mixed derivative
    G01 <- (z * (exp(lambda) - 1)) / (omega * exp(lambda) + (1 - omega) * lambda) -
    (z * (exp(lambda) - lambda) * (omega * exp(lambda) - omega + 1)) /
    (omega * exp(lambda) + (1 - omega) * lambda) ^ 2 + 
    exp(lambda) / (exp(lambda) + omega - 1) ^ 2
    
    # Beta^2 derivative
    G11 <- (exp(2 * lambda) / (exp(lambda) + omega - 1) ^ 2 +
    (omega * z * exp(lambda)) / (omega * exp(lambda) + (1 - omega) * lambda) -
    (z * (omega * exp(lambda) - omega + 1) ^ 2) / 
    (omega * exp(lambda) + (1 - omega) * lambda) ^ 2 -
    exp(lambda) / (exp(lambda) + omega - 1) - XXX / lambda ^ 2)
    
    G00 <- prior * G00 * omegaLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2
    
    G01 <- prior * G01 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) *
           omegaLink(eta[, 2], inverse = TRUE, deriv = 1)
    
    G11 <- prior * G11 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2
    
    matrix(
      -c(G11, # lambda
         G01, # mixed
         G01, # mixed
         G00  # omega
      ),
      dimnames = list(rownames(eta), c("lambda", "mixed", "mixed", "omega")),
      ncol = 4
    )
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    omega  <-  omegaLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    z <- ifelse(y == 1, y, 0)
    weight <- weight / prior
    
    G1 <- (z  * (exp(lambda) * omega + 1 - omega) / 
    (exp(lambda) * omega + lambda - omega * lambda) + 
    (1 - z) * y / lambda - exp(lambda) / (exp(lambda) - 1 + omega))
    G1 <- G1 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
    
    G0 <- ((z * (exp(lambda) - lambda)) / (exp(lambda) * omega + lambda * (1 - omega)) - 
            1 / (omega + exp(lambda) - 1) - (1 - z) / (1 - omega))
    
    G0 <- G0 * omegaLink(eta[, 2], inverse = TRUE, deriv = 1)
    
    uMatrix <- matrix(c(G1, G0), ncol = 2)
    
    weight <- lapply(X = 1:nrow(weight), FUN = function (x) {
      matrix(as.numeric(weight[x, ]), ncol = 2)
    })
    
    pseudoResid <- sapply(X = 1:length(weight), FUN = function (x) {
      xx <- chol2inv(chol(weight[[x]]))
      xx %*% uMatrix[x, ]
    })
    pseudoResid <- t(pseudoResid)
    dimnames(pseudoResid) <- dimnames(eta)
    pseudoResid
  }
  
  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, 
                           ...) {
    y <- as.numeric(y)
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X) / 2), rep(0, NROW(X) / 2))
    }
    
    if (is.null(weight)) {
      weight <- 1
    }
    z <- as.numeric(y == 1)
    
    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
      function(beta) {
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        omega  <-  omegaLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        -sum(weight * (z * (log(exp(lambda) * omega + (1 - omega) * lambda)) +
        (1 - z) * (log(1 - omega) + y * log(lambda) - lgamma(y + 1)) - 
        log(exp(lambda) - 1 + omega)))
      },
      function(beta) {
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        omega  <-  omegaLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        G1 <- (z  * (exp(lambda) * omega + 1 - omega) / 
        (exp(lambda) * omega + lambda - omega * lambda) + 
        (1 - z) * y / lambda - exp(lambda) / (exp(lambda) - 1 + omega))
        
        G1 <- G1 * weight * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
        
        G0 <- ((z * (exp(lambda) - lambda)) / (exp(lambda) * omega + lambda * (1 - omega)) - 
               1 / (omega + exp(lambda) - 1) - (1 - z) / (1 - omega))
        
        G0 <- G0 * weight * omegaLink(eta[, 2], inverse = TRUE, deriv = 1)
        
        if (NbyK) {
          XX <- 1:(attr(X, "hwm")[1])
          return(cbind(as.data.frame(X[1:nrow(eta), XX]) * G1, as.data.frame(X[-(1:nrow(eta)), -XX]) * G0))
        }
        if (vectorDer) {
          return(cbind(G1, G0))
        }
        
        as.numeric(c(G1, G0) %*% X)
      },
      function (beta) {
        lambdaPredNumber <- attr(X, "hwm")[1]
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        omega  <-  omegaLink(eta[, 2], inverse = TRUE)
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        
        res <- matrix(nrow = length(beta), ncol = length(beta), 
                      dimnames = list(names(beta), names(beta)))
        
        temp <- exp(lambda) * omega + lambda * (1 - omega)
        temp1 <- exp(lambda) * omega + 1 - omega
        temp2 <- exp(lambda) + 1 - omega
        
        G1 <- (z  * (exp(lambda) * omega + 1 - omega) / 
        (exp(lambda) * omega + lambda - omega * lambda) + 
        (1 - z) * y / lambda - exp(lambda) / (exp(lambda) - 1 + omega))
        
        G0 <- ((z * (exp(lambda) - lambda)) / (exp(lambda) * omega + lambda * (1 - omega)) - 
                1 / (omega + exp(lambda) - 1) - (1 - z) / (1 - omega))
        
        # omega^2 derivative
        G00 <- (-(z * (exp(lambda) - lambda) ^ 2) /
        (exp(lambda) * omega + lambda * (1 - omega)) ^2 +
        1 / (omega + exp(lambda) - 1) ^ 2 - (1 - z) / (1 - omega) ^ 2)
        
        # mixed derivative
        G01 <- (z * (exp(lambda) - 1)) / (omega * exp(lambda) + (1 - omega) * lambda) -
        (z * (exp(lambda) - lambda) * (omega * exp(lambda) - omega + 1)) /
        (omega * exp(lambda) + (1 - omega) * lambda) ^ 2 + 
        exp(lambda) / (exp(lambda) + omega - 1) ^ 2
        
        # Beta^2 derivative
        G11 <- (exp(2 * lambda) / (exp(lambda) + omega - 1) ^ 2 +
        (omega * z * exp(lambda)) / (omega * exp(lambda) + (1 - omega) * lambda) -
        (z * (omega * exp(lambda) - omega + 1) ^ 2) / 
        (omega * exp(lambda) + (1 - omega) * lambda) ^ 2 -
        exp(lambda) / (exp(lambda) + omega - 1) - (y * (1 - z)) / lambda ^ 2)
        
        res[-(1:lambdaPredNumber), -(1:lambdaPredNumber)] <- (
          t(as.data.frame(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)] * 
           (G00 * (omegaLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2) + 
            G0 * omegaLink(eta[, 2], inverse = TRUE, deriv = 2)) * weight)) %*% 
            as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)])
        )
        res[1:lambdaPredNumber, 1:lambdaPredNumber] <- (
          t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber] *
            (G11 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 +
             G1 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 2)) * weight)) %*% 
            X[1:(nrow(X) / 2), 1:lambdaPredNumber]
        )
        res[1:lambdaPredNumber, -(1:lambdaPredNumber)] <- t(
          t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber]) * G01 * 
            omegaLink(eta[, 2], inverse = TRUE, deriv = 1) * 
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * weight) %*% 
            as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)])
        )
        res[-(1:lambdaPredNumber), 1:lambdaPredNumber] <- (
          t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber]) * G01 * 
            omegaLink(eta[, 2], inverse = TRUE, deriv = 1) * 
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * weight) %*% 
            as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)])
        )
        
        res
      }
    )
  }
  
  validmu <- function(mu) {
    (sum(!is.finite(mu)) == 0) && all(0 < mu)
  }
  
  devResids <- function(y, eta, wt, ...) {
    omega  <-  omegaLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    mu1 <- mu.eta(eta = eta)
    idealOmega <- ifelse(y == 1, 1, 0)
    
    idealLambda <- tryCatch(
      expr = {
        suppressWarnings(
          ifelse(y > 1, lamW::lambertW0(-y * exp(-y)) + y, 0)
        )
      },
      error = function (e) {
        warning("Deviance residuals could not have been computed and zero vector will be returned instead.", call. = FALSE)
        NULL
      }
    )
    if (is.null(idealLambda)) {
      return(rep(0, length(y)))
    }
    
    diff <- ifelse(
      y == 1,
      -(log(exp(lambda) * omega + lambda * (1 - omega)) - log(exp(lambda) - 1 + omega)),
      y * log(idealLambda) - log(exp(idealLambda) - 1) - log(1 - omega) - y * log(lambda) + log(exp(lambda) - 1 + omega)
    )
    
    if (any(diff < 0)) {
      warning(paste0(
        "Some of differences between log likelihood in saturated model",
        " and fitted model were positive which indicates either:\n",
        "(1): A very good model fit or\n",
        "(2): Incorrect computation of saturated model",
        "\nDouble check deviance before proceeding"
      ))
    }
    
    ## see comments in ztpoisson for explanation of pmax
    sign(y - mu1) * sqrt(2 * wt * pmax(diff, 0))
  }
  
  pointEst <- function (pw, eta, contr = FALSE, ...) {
    omega  <-  omegaLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    N <- pw / (1 - (1 - omega) * exp(-lambda))
    if(!contr) {
      N <- sum(N)
    }
    N
  }
  
  popVar <- function (pw, eta, cov, Xvlm, ...) {
    omega  <-  omegaLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    bigTheta1 <- -pw * omegaLink(eta[, 2], inverse = TRUE,deriv = 1) * 
      exp(lambda)/(omega + exp(lambda) - 1) ^ 2# w.r to omega
    bigTheta2 <- pw * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * 
      ((omega - 1) * exp(lambda)) / (exp(lambda) + omega - 1) ^ 2 # w.r to lambda
    
    bigTheta <- t(c(bigTheta2, bigTheta1) %*% Xvlm)
    
    f1 <-  t(bigTheta) %*% as.matrix(cov) %*% bigTheta
    
    f2 <- sum(pw * (1 - omega) * exp(-lambda) / ((1 - (1 - omega) * exp(-lambda)) ^ 2))
    
    f1 + f2
  }
  
  dFun <- function (x, eta, type = c("trunc", "nontrunc")) {
    if (missing(type)) type <- "trunc"
    omega  <-  omegaLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    
    switch (type,
      "trunc" = {
        (omega * as.numeric(x == 1) +
        (1 - omega) * stats::dpois(x = x, lambda = lambda)) / 
        (1 - (1 - omega) * stats::dpois(x = 0, lambda = lambda))
      },
      "nontrunc" = {
        (1 - omega) * stats::dpois(x = x, lambda = lambda) +
        omega * as.numeric(x == 1)
      }
    )
  }
  
  simulate <- function(n, eta, lower = 0, upper = Inf) {
    omega  <-  omegaLink(eta[, 2], inverse = TRUE)
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    CDF <- function(x) {
      ifelse(x == Inf, 1, 
      ifelse(x < 0, 0, 
      ifelse(x < 1, (1 - omega) * exp(-lambda), 
      omega +  (1 - omega) * stats::ppois(x, lambda))))
    }
    lb <- CDF(lower)
    ub <- CDF(upper)
    p_u <- stats::runif(n, lb, ub)
    sims <- rep(0, n)
    cond <- CDF(sims) <= p_u
    while (any(cond)) {
      sims[cond] <- sims[cond] + 1
      cond <- CDF(sims) <= p_u
    }
    sims
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12)),
        (sizeObserved * priorWeights * (observed == 1) + .5) / (sizeObserved * priorWeights * sum(observed == 1) + 1)
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights)),
        family$links[[2]](weighted.mean(observed == 1, priorWeights) + .01)
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
      if ("(Intercept):omega" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[2], rep(0, attr(Xvlm, "hwm")[2] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[2] / attr(Xvlm, "hwm")[2], attr(Xvlm, "hwm")[2]))
      }
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "ztoipoisson",
      etaNames  = c("lambda", "omega"),
      simulate  = simulate,
      getStart  = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

#' @title Extract population size estimation results
#' 
#' @description An extractor function with \code{singleRStaticCountData} method for extracting
#' important information regarding pop size estimate.
#'
#' @param object object with population size estimates.
#' @param estimator for \code{singleRRatioReg} objects, which population-size
#'   estimator to return. The default \code{"primary"} returns the estimator
#'   selected when calling [ratioReg()].
#' @param ... additional optional arguments, currently not used in \code{singleRStaticCountData} class method. 
#'
#' @return An object of class \code{popSizeEstResults} containing population size estimation results.
#' @export
popSizeEst <- function(object, ...) {
  UseMethod("popSizeEst")
}

#' @method family singleRStaticCountData
#' @importFrom stats family
#' @exportS3Method
family.singleRStaticCountData <- function(object, ...) {
  object$model
}
#' @method extractAIC singleRStaticCountData
#' @importFrom stats extractAIC
#' @exportS3Method 
extractAIC.singleRStaticCountData <- function(fit, scale, k = 2, ...) {
  -2 * fit$logL + k * length(fit$coefficients)
}
# CODE MODIFIED FROM stats:::logLik.glm
#' @method logLik singleRStaticCountData
#' @importFrom stats logLik
#' @exportS3Method 
logLik.singleRStaticCountData <- function(object, 
                                          type = c("value",
                                                   "function"), 
                                          deriv = 0:2,
                                          ...) {
  if (missing(type))
    type <- "value"
  
  if (missing(deriv))
    deriv <- 0
  
  if (type == "value") {
    val <- object$logL
    attr(val, "nobs") <- nobs(object)
    attr(val, "df") <- length(stats::coef(object))
    class(val) <- "logLik"
    val
  } else {
    object$model$makeMinusLogLike(
      y = if (is.null(object$y)) as.numeric(model.response(model.frame(object))) else object$y,
      X = model.matrix(object, type = "vlm"),
      weight = object$priorWeights,
      deriv = deriv,
      offset = object$offset
    )
  }
}
# CODE MODIFIED FROM stats:::model.frame.glm
#' @method model.frame singleRStaticCountData
#' @importFrom stats glm
#' @importFrom stats model.frame
#' @importFrom stats update
#' @exportS3Method 
model.frame.singleRStaticCountData <- function(formula, ...) {
  dots <- list(...)
  dotargs <- dots[match(c("data", "na.action", "subset"), names(dots), 0L)]
  
  if (length(dotargs) || is.null(formula$modelFrame)) {
    combinedFromula <- singleRinternalMergeFormulas(formula$formula)
    if (!is.null(dotargs$data)) {
      jj <- all.vars(combinedFromula)[attr(terms(combinedFromula), "response")]
      if (!(jj %in% colnames(dotargs$data))) {
        combinedFromula <- update(combinedFromula, NULL ~ .)
      }
    }
    stats::model.frame(
      combinedFromula,
      data = if (is.null(dotargs$data)) eval(formula$call$data) else dotargs$data
    )
  }
  else formula$modelFrame
}

#' @method model.matrix singleRStaticCountData
#' @importFrom stats model.matrix
#' @exportS3Method 
model.matrix.singleRStaticCountData <- function(object, 
                                                type = c("lm", "vlm"), 
                                                ...) {
  if (missing(type)) type <- "lm"
  
  switch (type,
    lm = {
      X <- model.frame(object);
      X <- model.matrix(object$terms, X)
    },
    vlm = {
      X <- model.frame(object, ...);
      singleRinternalGetXvlmMatrix(
        X = X, 
        formulas = object$formula, 
        parNames = object$model$etaNames
      );
    }
  )
}

#' @method popSizeEst singleRStaticCountData
#' @rdname popSizeEst
#' @exportS3Method
popSizeEst.singleRStaticCountData <- function(object, ...) {
  object$populationSize
}

#' @method print popSizeEstResults
#' @exportS3Method 
print.popSizeEstResults <- function(x, ...) {
  cat("Point estimate: ", x$pointEstimate, 
      "\nVariance: ", x$variance, "\n", 
      (1 - x$control$alpha) * 100, "% confidence intervals:\n", 
      sep = "")
  print(x$confidenceInterval)
  
  invisible(x)
}

#' @importFrom stats fitted
#' @method fitted singleRStaticCountData
#' @exportS3Method 
fitted.singleRStaticCountData <- function(object,
                                          type = c("truncated", 
                                                   "nontruncated",
                                                   "all"),
                                          ...) {
  if (missing(type)) type <- "truncated"
  switch (type,
          truncated    = object$fittValues$truncated,
          nontruncated = object$fittValues$nontruncated,
          all          = object$fittValues
  )
}

#' @importFrom stats nobs
#' @method nobs singleRStaticCountData
#' @exportS3Method 
nobs.singleRStaticCountData <- function(object, ...) {
  object$sizeObserved
}

#' @importFrom stats df.residual
#' @method df.residual singleRStaticCountData
#' @exportS3Method 
df.residual.singleRStaticCountData <- function(object, ...) {
  object$dfResidual
}

#' @importFrom stats sigma
#' @method sigma singleRStaticCountData
#' @rdname regDiagSingleR
#' @exportS3Method 
sigma.singleRStaticCountData <- function(object, ...) {
  predict(object, type = "mean", se = TRUE)[c(3, 4)]
}

#' @importFrom stats influence
#' @method influence singleRStaticCountData
#' @rdname regDiagSingleR
#' @exportS3Method 
influence.singleRStaticCountData <- function(model, do.coef = FALSE, ...) {
  res <- list()
  hat <- hatvalues(model)
  if (NCOL(hat) > 1) {
    for (k in 1L:NCOL(hat)) {
      res[[paste0("hat:", colnames(hat)[k])]] <- hat[, k]
    }
  } else {
    res[["hat"]] <- hat[, 1]
  }
  
  if (isTRUE(do.coef)) {
    dfb <- dfbeta(model, ...)
    res[["coefficients"]] <- dfb
  }
  
  sigma <- sigma(model)
  res[["sigma:truncated"]]    <- sigma[, 1]
  res[["sigma:nontruncated"]] <- sigma[, 2]
  
  res[["dev.res"]] <- residuals(model, type = "deviance")[, 1]
  
  res[["pear.res"]] <- residuals(model, type = "pearson")[, 1]
  
  res
}

#' @importFrom stats rstudent
#' @method rstudent singleRStaticCountData
#' @rdname regDiagSingleR
#' @exportS3Method 
rstudent.singleRStaticCountData <- function(model, ...) {
  res <- residuals(model, type = "pearson")[, 1]
  hat <- hatvalues(model)[, 1]
  
  res <- res / sqrt((1 - hat))
  
  res[is.infinite(res)] <- NaN
  res
}

#' @importFrom stats rstandard
#' @method rstandard singleRStaticCountData
#' @rdname regDiagSingleR
#' @exportS3Method 
rstandard.singleRStaticCountData <- function(model,
                                             type = c("deviance", "pearson"), 
                                             ...) {
  type <- match.arg(type)
  res <- switch (type,
    pearson  = residuals(model, type = "pearsonSTD")[, 1],
    deviance = residuals(model, type = "deviance")[, 1] / 
      (sqrt(1 - hatvalues(model)[, 1]) * sigma(model)[, 1]),
  )
  
  res[is.infinite(res)] <- NaN
  res
}

#' @rdname singleRmodels
#' @export
zotnegbin <- function(nSim = 1000, epsSim = 1e-8, eimStep = 6, 
                      lambdaLink = c("log", "neglog"), 
                      alphaLink = c("log", "neglog"),
                      ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  if (missing(alphaLink))  alphaLink <- "log"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink, alphaLink)
  
  lambdaLink <- switch(lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  alphaLink <- switch(alphaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  links[1:2] <- c(lambdaLink, alphaLink)
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    
    if (!deriv) {
      switch (
        type,
        "nontrunc" = lambda,
        "trunc" = (lambda - lambda * ((1 + alpha * lambda) ^ (-1 - 1 / alpha))) / 
        (1 - (1 + alpha * lambda) ^ (-1 / alpha) - lambda * ((1 + alpha * lambda) ^ (-1 - 1 / alpha)))
      )
    } else {
      switch (
        type,
        "nontrunc" = {
          matrix(c(1, 0) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
             alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
          ), ncol = 2)
        },
        "trunc" = {
          matrix(c(
            ((alpha * lambda + 1) ^ (2 / alpha) * (alpha ^ 2 * lambda ^ 2 + 2 * alpha * lambda + 1) +
            (alpha * lambda + 1) ^ (1 / alpha) * ((-alpha ^ 2 - 2 * alpha - 1) * lambda ^ 2 - 2 * alpha * lambda - 2) + 1) /
            ((alpha * lambda + 1) ^ (1 / alpha + 1) + (-alpha - 1) * lambda - 1) ^ 2,
            lambda ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha) * 
            (lambda * alpha ^ 2 + (lambda + 1) * alpha + 1) * log(lambda * alpha + 1) +
            (lambda * alpha + 1) ^ (1 / alpha) * ((1 - 2 * lambda) * alpha ^ 2 - lambda * alpha) - alpha ^ 2) /
            (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha + 1) - lambda * alpha - lambda - 1) ^ 2)
          ) * c(
            lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
             alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
          ), ncol = 2)
        }
      )
    }
  }

  variance <- function(eta, type = "nontrunc", ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    switch (type,
    "nontrunc" = lambda * (1 + alpha * lambda),
    "trunc" = (lambda * (1 + alpha * lambda) + lambda ^ 2 - 
    lambda * ((1 + alpha * lambda) ^ (-1 - 1 / alpha))) / 
    (1 - (1 + alpha * lambda) ^ (-1 / alpha) - 
    lambda * ((1 + alpha * lambda) ^ (-1 - 1 / alpha))) - 
    mu.eta(eta, type = "trunc") ^ 2
    )
  }
  
  compExpect <- function(eta) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    
    P0 <- (1 + alpha * lambda) ^ (-1 / alpha)
    P1 <- lambda * (1 + alpha * lambda) ^ (- 1 / alpha - 1)
    
    res <- rep(0, NROW(eta))
    k <- 2 
    
    finished <- rep(FALSE, NROW(eta))
    while ((k < nSim) & !all(finished)) {
      prob <- apply(cbind(k:(k + eimStep)), MARGIN = 1, FUN = function(x) {
        stats::dnbinom(
          x = x, 
          size = 1 / alpha, 
          mu = lambda
        ) / (1 - P0 - P1)
      })
      trg <- apply(cbind(k:(k + eimStep)), MARGIN = 1, FUN = function(x) {
        (2 * (digamma(x + 1 / alpha) - digamma(1 / alpha)) * alpha +
           trigamma(x + 1 / alpha) - trigamma(1 / alpha)) / alpha ^ 4
      })
      prob[!(is.finite(prob))] <- 0
      trg[!(is.finite(trg))] <- 0
      toAdd <- trg * prob
      toAdd <- rowSums(toAdd)
      k <- k + eimStep + 1
      res <- res + toAdd
      finished <- abs(toAdd) < epsSim
    }
    res
  }
  
  Wfun <- function(prior, eta, y, ...) {
    iddx <- y > 1
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)[iddx]
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)[iddx]
    Ey <- mu.eta(eta = eta)[iddx]
    
    prob <- (1 - (1 + lambda * alpha) ^ (-1 / alpha) - 
    lambda * (1 + lambda * alpha) ^ (-1 - 1 / alpha))[iddx]
    
    Etrig <- compExpect(eta[iddx, , drop = FALSE])
    
    # 2nd alpha derivative
    G00 <- iddx
    G00[iddx] <- Etrig + (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
    (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) ^ 2 / 
    (-1 / (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) + 1) ^ 2 -
    (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) ^ 2 / 
    (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) /
    (lambda * alpha + 1)) ^ 2 - (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 + 
    (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) + lambda ^ 2 / 
    (alpha * (lambda * alpha + 1) ^ 2)) / (lambda * alpha + 1) ^ (1 / alpha) -
    lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) * 
    (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 + (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) -
    (lambda ^ 2 * (-1 / alpha - 1)) / (lambda * alpha + 1) ^ 2)) / 
    (-1 / (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) + 1) -
    (2 * log(lambda * alpha + 1)) / alpha ^ 3 + (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) -
    (lambda ^ 2 * (-1 / alpha - Ey)) / (lambda * alpha + 1) ^ 2 - Ey / alpha ^ 2
    
    G00[iddx] <- G00[iddx] * alphaLink(eta[, 2], inverse = TRUE, deriv = 1)[iddx] ^ 2
    
    # mixed derivative
    G01 <- iddx
    G01[iddx] <- -((alpha * lambda + 1) ^ (1 / alpha) * ((alpha ^ 3 + alpha ^ 2) * lambda ^ 3 +
    (2 * alpha ^ 2 + 2 * alpha) * lambda ^ 2 + (alpha + 1) * lambda) * 
    log(alpha * lambda + 1) + (alpha * lambda + 1) ^ (1 / alpha) * 
    ((alpha ^ 4 - alpha ^ 3 - alpha ^ 2) * lambda ^ 3 + 
    ((-2 * alpha ^ 4 - 2 * alpha ^ 3) * Ey + 4 * alpha ^ 3 - alpha ^ 2 - alpha) * lambda ^ 2 + 
    ((-4 * alpha ^ 3 - 2 * alpha ^ 2) * Ey + 3 * alpha ^ 2) * lambda - 2 * alpha ^ 2 * Ey) +
    (alpha * lambda + 1) ^ (2 / alpha) * (-alpha ^ 4 * lambda ^ 3 + 
    (alpha ^ 4 * Ey - 2 * alpha ^ 3) * lambda ^ 2 + 
    (2 * alpha ^ 3 * Ey - alpha ^ 2) * lambda + alpha ^ 2 * Ey) +
    ((alpha ^ 4 + 2 * alpha ^ 3 + alpha ^ 2) * Ey - 2 * alpha ^ 3 - alpha ^ 2) * lambda ^ 2 +
    ((2 * alpha ^ 3 + 2 * alpha ^ 2) * Ey - 2 * alpha ^ 2) * lambda + alpha ^ 2 * Ey) /
    (alpha ^ 2 * (alpha * lambda + 1) ^ 2 * ((alpha * lambda + 1) ^ (1 / alpha + 1) + (-alpha - 1) * lambda - 1) ^ 2)
    
    G01[iddx] <- G01[iddx] * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)[iddx] *
      alphaLink(eta[, 2], inverse = TRUE, deriv = 1)[iddx]
    
    # second lambda derivative
    G11 <- iddx
    G11[iddx] <- ((alpha * lambda + 1) ^ (2 / alpha) * 
    (alpha ^ 3 * lambda ^ 4 + (2 * alpha ^ 2 - 2 * alpha ^ 3 * Ey) * lambda ^ 3 +
    (alpha - 5 * alpha ^ 2 * Ey) * lambda ^ 2 - 4 * alpha * Ey * lambda - Ey) + 
    (alpha * lambda + 1) ^ (1 / alpha) * ((alpha - alpha ^ 3) * lambda ^ 4 +
    ((4 * alpha ^ 3 + 4 * alpha ^ 2) * Ey - 4 * alpha ^ 2 - alpha + 1) * lambda ^ 3 +
    ((10 * alpha ^ 2 + 6 * alpha) * Ey - 3 * alpha - 1) * lambda ^ 2 + 
    (8 * alpha + 2) * Ey * lambda + 2 * Ey) + 
    ((-2 * alpha ^ 3 - 4 * alpha ^ 2 - 2 * alpha) * Ey + 2 * alpha ^ 2 + 2 * alpha) * lambda ^ 3 +
    ((-5 * alpha ^ 2 - 6 * alpha - 1) * Ey + 2 * alpha + 1) * lambda ^ 2 + 
    (-4 * alpha - 2) * Ey * lambda - Ey) / 
    (lambda ^ 2 * (alpha * lambda + 1) ^ 2 * 
    ((alpha * lambda + 1) ^ (1 / alpha + 1) + (-alpha - 1) * lambda - 1) ^ 2)
    
    G11[iddx] <- G11[iddx] * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)[iddx] ^ 2
    
    matrix(
      -c(G11 * prior, # lambda
        G01 * prior,  # mixed 
        G01 * prior,  # mixed 
        G00 * prior  # alpha 
      ),
      dimnames = list(rownames(eta), 
                      c("lambda", "mixed", "mixed", "alpha")),
      ncol = 4
    )
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    iddx <- y > 1
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)[iddx]
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)[iddx]
    weight[iddx, ] <- weight[iddx, ] / prior[iddx]
    
    G0 <- iddx
    G0[iddx] <- y[iddx] / alpha + (digamma(1 / alpha) - digamma(1 / alpha + y[iddx])) / alpha ^ 2 +
    ((log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
    (lambda * alpha + 1) ^ (1 / alpha) + lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
    (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) /
    (1 - 1 / (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1)) +
    log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - y[iddx])) / (lambda * alpha + 1)
    
    G0[iddx] <- G0[iddx] * alphaLink(eta[, 2], inverse = TRUE, deriv = 1)[iddx]
    
    G1 <- iddx
    G1[iddx] <- y[iddx] / lambda - ((1 / alpha + 1) * alpha * lambda * 
    (alpha * lambda + 1) ^ (-1 / alpha - 2)) / 
    (-1 / (alpha * lambda + 1) ^ (1 / alpha) - 
    lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1) -
    (alpha * (y[iddx] + 1 / alpha)) / (alpha * lambda + 1)
    
    G1[iddx] <- G1[iddx] * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)[iddx]
    
    weight <- lapply(X = 1:nrow(weight), FUN = function (x) {
      matrix(as.numeric(weight[x, ]), ncol = 2)
    })
    
    uMatrix <- matrix(c(G1, G0), ncol = 2)
    
    pseudoResid <- sapply(X = 1:length(weight), FUN = function (x) {
      if (iddx[x]) {
        xx <- solve(weight[[x]])
        xx %*% uMatrix[x, ]
      } else {
        uMatrix[x, ]
      }
    })
    pseudoResid <- t(pseudoResid)
    dimnames(pseudoResid) <- dimnames(eta)
    pseudoResid
  }

  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, 
                           ...) {
    if (is.null(weight)) {
      weight <- 1
    }
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X) / 2), rep(0, NROW(X) / 2))
    }
    y <- as.numeric(y)
    iddx <- y > 1
    X <- as.matrix(X)

    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
      function(beta) {
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        alpha  <-  alphaLink(eta[, 2], inverse = TRUE)

        -sum(weight * iddx * (lgamma(y + 1 / alpha) - lgamma(1 / alpha) -
        lgamma(y + 1) - (y + 1 / alpha) * log(1 + lambda * alpha) +
        y * log(lambda * alpha) - log(1 - (1 + lambda * alpha) ^ (-1 / alpha) - 
        lambda * (1 + lambda * alpha) ^ (-1 - 1 / alpha)))
        )
      },
      function(beta) {
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)[iddx]
        alpha  <-  alphaLink(eta[, 2], inverse = TRUE)[iddx]
        
        G0 <- iddx
        G0[iddx] <- y[iddx] / alpha + (digamma(1 / alpha) - digamma(1 / alpha + y[iddx])) / alpha ^ 2  +
        log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - y[iddx])) / (lambda * alpha + 1) -
        (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
        (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
        (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) /
        (-(lambda * alpha + 1) ^ (-1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) + 1)
        
        G0[iddx] <- G0[iddx] * alphaLink(eta[, 2], inverse = TRUE, deriv = 1)[iddx] * weight[iddx]
        
        G1 <- iddx
        G1[iddx] <- y[iddx] / lambda  + (alpha * (-y[iddx] - 1 / alpha)) / (alpha * lambda + 1) -
        ((1 / alpha + 1) * alpha * lambda * (alpha * lambda + 1) ^ (-1 / alpha - 2)) / 
        (-(alpha * lambda + 1) ^ (-1 / alpha) - lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1)
        
        G1[iddx] <- G1[iddx] * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)[iddx] * weight[iddx]
        
        if (NbyK) {
          XX <- 1:(attr(X, "hwm")[1])
          return(cbind(as.data.frame(X[1:nrow(eta), XX]) * G1, 
                       as.data.frame(X[-(1:nrow(eta)), -XX]) * G0))
        }
        if (vectorDer) {
          return(cbind(G1, G0))
        }
        
        as.numeric(c(G1, G0) %*% X)
      },
      function(beta) {
        lambdaPredNumber <- attr(X, "hwm")[1]
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)[iddx]
        alpha  <-  alphaLink(eta[, 2], inverse = TRUE)[iddx]
        
        res <- matrix(nrow = length(beta), ncol = length(beta), 
                      dimnames = list(names(beta), names(beta)))
        
        G0 <- iddx
        G0[iddx] <- y[iddx] / alpha + (digamma(1 / alpha) - digamma(1 / alpha + y[iddx])) / alpha ^ 2 -
        (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
        (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
        (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) /
        (-1 / (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) + 1) +
        log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - y[iddx])) / (lambda * alpha + 1)
        
        G1 <- iddx
        G1[iddx] <- y[iddx] / lambda + ((-1 / alpha-1) * alpha * lambda * 
        (alpha * lambda + 1) ^ (-1 / alpha - 2)) / 
        (-1 / (alpha * lambda + 1) ^ (1 / alpha) - 
        lambda * (alpha * lambda + 1) ^ (-1 / alpha - 1) + 1) + 
        (alpha * (-y[iddx] - 1 / alpha)) / (alpha * lambda + 1)
        
        # 2nd alpha derivative
        
        G00 <- iddx
        G00[iddx] <- (2 * digamma(1 / alpha + y[iddx])) / alpha ^ 3 - (2 * digamma(1 / alpha)) / alpha ^ 3 +
        trigamma(1 / alpha + y[iddx]) / alpha ^ 4 - trigamma(1 / alpha) / alpha ^ 4 + 
        (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) /
        (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
        (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) / (lambda * alpha + 1))) ^ 2 / 
        (-1 / (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) + 1) ^ 2 -
        (-(log(lambda * alpha + 1) / alpha ^ 2 - lambda / (alpha * (lambda * alpha + 1))) ^ 2 / 
        (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) *
        (log(lambda * alpha + 1) / alpha ^ 2 + (lambda * (-1 / alpha - 1)) /
        (lambda * alpha + 1)) ^ 2 - (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 + 
        (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) + lambda ^ 2 / 
        (alpha * (lambda * alpha + 1) ^ 2)) / (lambda * alpha + 1) ^ (1 / alpha) -
        lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) * 
        (-(2 * log(lambda * alpha + 1)) / alpha ^ 3 + (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) -
        (lambda ^ 2 * (-1 / alpha - 1)) / (lambda * alpha + 1) ^ 2)) / 
        (-1 / (lambda * alpha + 1) ^ (1 / alpha) - lambda * (lambda * alpha + 1) ^ (-1 / alpha - 1) + 1) -
        (2 * log(lambda * alpha + 1)) / alpha ^ 3 + (2 * lambda) / (alpha ^ 2 * (lambda * alpha + 1)) -
        (lambda ^ 2 * (-1 / alpha - y[iddx])) / (lambda * alpha + 1) ^ 2 - y[iddx] / alpha ^ 2
          
        G00[iddx] <- G00[iddx] * alphaLink(eta[, 2], inverse = TRUE, deriv = 1)[iddx] ^ 2 +
                     G0[iddx]  * alphaLink(eta[, 2], inverse = TRUE, deriv = 2)[iddx]
        
        # mixed derivative
        G01 <- iddx
        G01[iddx] <- -((alpha * lambda + 1) ^ (1 / alpha) * ((alpha ^ 3 + alpha ^ 2) * lambda ^ 3 +
        (2 * alpha ^ 2 + 2 * alpha) * lambda ^ 2 + (alpha + 1) * lambda) * 
        log(alpha * lambda + 1) + (alpha * lambda + 1) ^ (1 / alpha) * 
        ((alpha ^ 4 - alpha ^ 3 - alpha ^ 2) * lambda ^ 3 + 
        ((-2 * alpha ^ 4 - 2 * alpha ^ 3) * y[iddx] + 4 * alpha ^ 3 - alpha ^ 2 - alpha) * lambda ^ 2 + 
        ((-4 * alpha ^ 3 - 2 * alpha ^ 2) * y[iddx] + 3 * alpha ^ 2) * lambda - 2 * alpha ^ 2 * y[iddx]) +
        (alpha * lambda + 1) ^ (2 / alpha) * (-alpha ^ 4 * lambda ^ 3 + 
        (alpha ^ 4 * y[iddx] - 2 * alpha ^ 3) * lambda ^ 2 + (2 * alpha ^ 3 * y[iddx] - alpha ^ 2) * lambda + alpha ^ 2 * y[iddx]) +
        ((alpha ^ 4 + 2 * alpha ^ 3 + alpha ^ 2) * y[iddx] - 2 * alpha ^ 3 - alpha ^ 2) * lambda ^ 2 +
        ((2 * alpha ^ 3 + 2 * alpha ^ 2) * y[iddx] - 2 * alpha ^ 2) * lambda + alpha ^ 2 * y[iddx]) /
        (alpha ^ 2 * (alpha * lambda + 1) ^ 2 * ((alpha * lambda + 1) ^ (1 / alpha + 1) + (-alpha - 1) * lambda - 1) ^ 2)
        
        
        G01[iddx] <- G01[iddx] * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)[iddx] *
                                  alphaLink(eta[, 2], inverse = TRUE, deriv = 1)[iddx]
        
        # second lambda derivative
        G11 <- iddx
        G11[iddx] <- ((alpha * lambda + 1) ^ (2 / alpha) * 
        (alpha ^ 3 * lambda ^ 4 + (2 * alpha ^ 2 - 2 * alpha ^ 3 * y[iddx]) * lambda ^ 3 +
        (alpha - 5 * alpha ^ 2 * y[iddx]) * lambda ^ 2 - 4 * alpha * y[iddx] * lambda - y[iddx]) + 
        (alpha * lambda + 1) ^ (1 / alpha) * ((alpha - alpha ^ 3) * lambda ^ 4 +
        ((4 * alpha ^ 3 + 4 * alpha ^ 2) * y[iddx] - 4 * alpha ^ 2 - alpha + 1) * lambda ^ 3 +
        ((10 * alpha ^ 2 + 6 * alpha) * y[iddx] - 3 * alpha - 1) * lambda ^ 2 + 
        (8 * alpha + 2) * y[iddx] * lambda + 2 * y[iddx]) + 
        ((-2 * alpha ^ 3 - 4 * alpha ^ 2 - 2 * alpha) * y[iddx] + 2 * alpha ^ 2 + 2 * alpha) * lambda ^ 3 +
        ((-5 * alpha ^ 2 - 6 * alpha - 1) * y[iddx] + 2 * alpha + 1) * lambda ^ 2 + 
        (-4 * alpha - 2) * y[iddx] * lambda - y[iddx]) / 
        (lambda ^ 2 * (alpha * lambda + 1) ^ 2 * 
        ((alpha * lambda + 1) ^ (1 / alpha + 1) + (-alpha - 1) * lambda - 1) ^ 2)
        
        G11[iddx] <- G11[iddx] * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)[iddx] ^ 2 +
                     G1[iddx]  * lambdaLink(eta[, 1], inverse = TRUE, deriv = 2)[iddx]
        
        
        res[-(1:lambdaPredNumber), -(1:lambdaPredNumber)] <- t(as.data.frame(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)]) * weight * G00) %*% X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)]
        res[1:lambdaPredNumber, 1:lambdaPredNumber] <- t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber]) * weight * G11) %*% X[1:(nrow(X) / 2), 1:lambdaPredNumber]
        res[1:lambdaPredNumber, -(1:lambdaPredNumber)] <- t(t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber]) * weight * G01) %*% as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)]))
        res[-(1:lambdaPredNumber), 1:lambdaPredNumber] <- t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber]) * weight * G01) %*% as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)])
        
        res
      }
    )
  }

  validmu <- function(mu) {
    (sum(!is.finite(mu)) == 0) && all(0 < mu)
  }

  devResids <- function (y, eta, wt, ...) {
    iddx <- y > 1
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    mu <- mu.eta(eta = eta)

    prob_ge2 <- pmax(1 - (1 + lambda * alpha) ^ (-1 / alpha) - lambda * (1 + lambda * alpha) ^ (-1 - 1 / alpha), .Machine$double.xmin)
    logLikFit <- (
      lgamma(y + 1 / alpha) - lgamma(1 / alpha) - lgamma(y + 1) - (y + 1 / alpha) * log(1 + lambda * alpha) +
      y * log(lambda * alpha) - log(prob_ge2)
    )[iddx]

    yUnq <- unique(y[iddx])

    if (any(yUnq > 77)) {
      warning("Currently numerical deviance is unreliable for counts greater than 78.")
    }
    
    findL <- function(yNow) {
      stats::optim(
        par = if(TRUE) c(0, log(yNow), -15 + 2 * (yNow %in% 3:5) - 2 * (yNow %in% 6:9)) else c(0, log(yNow), -6),
        fn = function(x) {
          s <- x[1]
          l <- exp(x[2])
          a <- exp(x[3])
          
          sum(c(((l-l*(a*l+1)^(-1/a-1))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)-yNow) ,# s der
                s*((-(a*l+1)^(-1/a-1)-(-1/a-1)*a*l*(a*l+1)^(-1/a-2)+1)/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+(a*(-yNow-1/a))/(a*l+1)+yNow/l,# lambda der
                s*(-((l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1))-(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)+log(l*a+1)/a^2+(l*(-1/a-yNow))/(l*a+1)+yNow/a-digamma(1/a+yNow)/a^2+digamma(1/a)/a^2) ^ 2)#alpha der
        },
        gr = function(x) {
          s <- x[1]
          l <- exp(x[2])
          a <- exp(x[3])
          d12.21 <- ((a*l+1)^(2/a)*(a^2*l^2+2*a*l+1)+(a*l+1)^(1/a)*((-a^2-2*a-1)*l^2-2*a*l-2)+1)/((a*l+1)^(1/a+1)+(-a-1)*l-1)^2
          d13.31 <- (l^2*((l*a+1)^(1/a)*(l*a^2+(l+1)*a+1)*log(l*a+1)+(l*a+1)^(1/a)*((1-2*l)*a^2-l*a)-a^2))/(a^2*((l*a+1)^(1/a+1)-l*a-l-1)^2)
          d22.22 <- s*((-2*(-1/a-1)*a*(a*l+1)^(-1/a-2)-(-1/a-2)*(-1/a-1)*a^2*l*(a*l+1)^(-1/a-3))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)*a*(a*l+1)^(-1/a-2)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2+((-1/a-2)*(-1/a-1)*a^2*l*(a*l+1)^(-1/a-3)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2+(2*(-1/a-1)*a*l*(a*l+1)^(-1/a-2)*(-(a*l+1)^(-1/a-1)-(-1/a-1)*a*l*(a*l+1)^(-1/a-2)+1))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2+(2*(-1/a-1)^2*a^2*l^2*(a*l+1)^(-2/a-4)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^3)+((-1/a-1)*a*(a*l+1)^(-1/a-2))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-2)*(-1/a-1)*a^2*l*(a*l+1)^(-1/a-3))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)^2*a^2*l^2*(a*l+1)^(-2/a-4))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2-(a^2*(-yNow-1/a))/(a*l+1)^2-yNow/l^2
          d23.32 <- (((l*a+1)^(2/a)*(l^5*s*a^5+(5*l^4*s-l^4)*a^4+((-l^5+2*l^4+9*l^3)*s-l^4-3*l^3)*a^3+((-3*l^4+6*l^3+7*l^2)*s-3*l^3-3*l^2)*a^2+((-3*l^3+6*l^2+2*l)*s-3*l^2-l)*a+(2*l-l^2)*s-l)+(l*a+1)^(1/a)*(-l^5*s*a^5+((-3*l^5-5*l^4)*s+l^4)*a^4+((-3*l^5-10*l^4-9*l^3)*s+2*l^4+3*l^3)*a^3+((-l^5-7*l^4-13*l^3-7*l^2)*s+l^4+5*l^3+3*l^2)*a^2+((-2*l^4-5*l^3-8*l^2-2*l)*s+2*l^3+4*l^2+l)*a+(-l^3-l^2-2*l)*s+l^2+l))*log(l*a+1)+(l*a+1)^(2/a)*((3*l^3*yNow-l^5*s-2*l^4)*a^5+((3*l^3+9*l^2)*yNow+(2*l^5-8*l^4+2*l^3)*s-8*l^3)*a^4+((6*l^2+9*l)*yNow+(l^5+2*l^4-13*l^3+4*l^2)*s+l^4-10*l^2)*a^3+((3*l+3)*yNow+(2*l^4-2*l^3-6*l^2+2*l)*s+2*l^3-4*l)*a^2+((l^3-2*l^2)*s+l^2)*a)+(l*a+1)^(3/a)*((l^4-l^3*yNow)*a^5+(3*l^3-3*l^2*yNow)*a^4+(3*l^2-3*l*yNow)*a^3+(l-yNow)*a^2)+(l*a+1)^(1/a)*((-3*l^3*yNow+l^5*s+l^4)*a^5+((-6*l^3-9*l^2)*yNow+(3*l^5+8*l^4-4*l^3)*s+7*l^3)*a^4+((-3*l^3-12*l^2-9*l)*yNow+(3*l^5+7*l^4+13*l^3-8*l^2)*s-2*l^4+3*l^3+11*l^2)*a^3+((-3*l^2-6*l-3)*yNow+(l^5+4*l^4+6*l^3+6*l^2-4*l)*s-l^4-3*l^3+3*l^2+5*l)*a^2+((l^4+l^3+2*l^2)*s-l^3-l^2)*a)+l^3*yNow*a^5+((3*l^3+3*l^2)*yNow+2*l^3*s-2*l^3)*a^4+((3*l^3+6*l^2+3*l)*yNow+4*l^2*s-3*l^3-4*l^2)*a^3+((l^3+3*l^2+3*l+1)*yNow+2*l*s-l^3-3*l^2-2*l)*a^2)/(a^2*(l*a+1)^2*((l*a+1)^(1/a+1)-l*a-l-1)^3)
          d33.33 <- s*((2*(l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))^2)/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^3-((l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))^2/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))^2-(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))+l^2/(a*(l*a+1)^2))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-1))/(l*a+1)^2)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))^2)/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)+(2*l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-1))/(l*a+1)^2))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1))+(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))^2/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(-(log(l*a+1)/a^2-l/(a*(l*a+1)))^2/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))^2-(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))+l^2/(a*(l*a+1)^2))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-1))/(l*a+1)^2))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)-(2*log(l*a+1))/a^3+(2*l)/(a^2*(l*a+1))-(l^2*(-1/a-yNow))/(l*a+1)^2-yNow/a^2+(2*digamma(1/a+yNow))/a^3-(2*digamma(1/a))/a^3+trigamma(1/a+yNow)/a^4-trigamma(1/a)/a^4
          
          f2 <- 2*c(((l-l*(a*l+1)^(-1/a-1))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)-yNow) ,# s der
                    s*((-(a*l+1)^(-1/a-1)-(-1/a-1)*a*l*(a*l+1)^(-1/a-2)+1)/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2)*(l-l*(a*l+1)^(-1/a-1)))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)^2)+((-1/a-1)*a*l*(a*l+1)^(-1/a-2))/(-1/(a*l+1)^(1/a)-l*(a*l+1)^(-1/a-1)+1)+(a*(-yNow-1/a))/(a*l+1)+yNow/l,# lambda der
                    s*(-((l-l*(l*a+1)^(-1/a-1))*(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1))))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)^2-(l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1))-(-(log(l*a+1)/a^2-l/(a*(l*a+1)))/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)*(log(l*a+1)/a^2+(l*(-1/a-1))/(l*a+1)))/(-1/(l*a+1)^(1/a)-l*(l*a+1)^(-1/a-1)+1)+log(l*a+1)/a^2+(l*(-1/a-yNow))/(l*a+1)+yNow/a-digamma(1/a+yNow)/a^2+digamma(1/a)/a^2)
          
          c(sum(f2 * c(0, d12.21, d13.31)),
            sum(f2 * c(d12.21, d22.22, d23.32)),
            sum(f2 * c(d13.31, d23.32, d33.33)))
        },
        method = "CG",
        control = list(maxit = 50000, abstol = .Machine$double.eps)
      )
    }
    fff <- function(yNow) {
      if (yNow == 2) return(0)
      
      xx <- findL(yNow)
      a <- exp(xx$par[3])
      l <- exp(xx$par[2])
      
      lgamma(yNow+1/a)-lgamma(1/a)-lgamma(yNow+1)-(yNow+1/a)*log(1+l*a)+yNow*log(l*a)-log(1-(1+l*a)^(-1/a)-l*(1+l*a)^(-1-1/a))
    }
    
    suppressWarnings(logLikIdeal <- sapply(yUnq, FUN = fff))
    names(logLikIdeal) <- yUnq
    
    logLikIdeal <- sapply(y[iddx], FUN = function(x) {
      logLikIdeal[names(logLikIdeal) == x]
    })
    
    diff <- iddx
    diff[iddx] <- logLikIdeal - logLikFit

    if (any(logLikFit > 0)) {
      warning("Dispersion parameter values are on the boundary of parameter space. Deviance residuals will be assigned 0 on these observations.")
      diff[iddx][logLikFit > 0]   <- 0
    } else if (any(diff < 0)) {
      warning("Numerical deviance finder found worse saturated likelihood than fitted model. Expect NA's in deviance/deviance residuals.")
    }

    diff[diff < 0] <- 0

    sign(y - mu) * sqrt(2 * wt * diff)
  }

  pointEst <- function (pw, eta, contr = FALSE, y, ...) {
    iddx <- y > 1
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    
    N <- pw * (iddx * (1 - lambda * ((1 + lambda * alpha) ^ (-1 - 1 / alpha))) / 
    (1 - (1 + lambda * alpha) ^ (- 1 / alpha) - lambda * 
    ((1 + lambda * alpha) ^ (-1 - 1 / alpha))) + (1 - iddx))
    
    if(!contr) {
      N <- sum(N)
    }
    N
  }

  popVar <- function (pw, eta, cov, Xvlm, y, ...) {
    iddx <- y > 1
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    
    prob <- 1 - (1 + lambda * alpha) ^ (-1 / alpha) - 
      lambda * (1 + lambda * alpha) ^ (-1 - 1 / alpha)
    
    bigTheta1 <- pw * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) * 
    as.numeric(((lambda*alpha+1)^(1/alpha)*(lambda^2*alpha^2+2*lambda*alpha+1)*log(lambda*alpha+1)+(lambda*alpha+1)^(1/alpha)*(-lambda^2*alpha^2-lambda*alpha)-lambda^2*alpha^2)/(alpha^2*((lambda*alpha+1)^(1/alpha+1)-lambda*alpha-lambda-1)^2))
    
    bigTheta2 <- pw * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * 
    as.numeric(-((alpha*lambda+1)^(1/alpha+1)-1)/((alpha*lambda+1)^(1/alpha+1)+(-alpha-1)*lambda-1)^2)
    
    bigTheta <- t(c(bigTheta2, bigTheta1)[rep(iddx, 2)] %*% Xvlm[rep(iddx, 2), , drop = FALSE])
    
    f1 <-  t(bigTheta) %*% as.matrix(cov) %*% bigTheta
    f2 <-  sum((pw * ((1 - prob) / (prob ^ 2)) * 
    (1 - lambda * ((1 + lambda * alpha) ^ (-1 - 1 / alpha))) ^ 2)[iddx])
    
    f1 + f2
  }
  
  dFun <- function (x, eta, type = c("trunc", "nontrunc")) {
    if (missing(type)) type <- "trunc"
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    
    switch (type,
      "trunc" = {
        stats::dnbinom(x = x, mu = lambda, size = 1 / alpha) / 
        (1 - stats::dnbinom(x = 0, mu = lambda, size = 1 / alpha) - 
        stats::dnbinom(x = 1, mu = lambda, size = 1 / alpha))
      },
      "nontrunc" = stats::dnbinom(x = x, mu = lambda, size = 1 / alpha)
    )
  }

  simulate <- function(n, eta, lower = 0, upper = Inf) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    
    lb <- stats::pnbinom(lower, mu = lambda, size = 1 / alpha)
    ub <- stats::pnbinom(upper, mu = lambda, size = 1 / alpha)
    p_u <- stats::runif(n, lb, ub)
    
    sims <- stats::qnbinom(p_u, mu = lambda, size = 1 / alpha)
    sims
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      init <- log(abs((observed / weighted.mean(observed, priorWeights) - 1) / observed) + .1)
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12)),
        family$links[[2]](ifelse(init < -.5, .1, init + .55))
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights)),
        family$links[[2]](abs((cov.wt(cbind(observed, observed), wt = priorWeights, method = "ML")$cov[1,1] / weighted.mean(observed, priorWeights) - 1) / weighted.mean(observed, priorWeights)) + .1)
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
      if ("(Intercept):alpha" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[2], rep(0, attr(Xvlm, "hwm")[2] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[2] / attr(Xvlm, "hwm")[2], attr(Xvlm, "hwm")[2]))
      }
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "zotnegbin",
      etaNames  = c("lambda", "alpha"),
      simulate  = simulate,
      getStart  = getStart
    ),
    class = c("singleRfamily", "family")
  )
}

#' @title The British farm submissions data
#' @description Data on British animal farms submissions to AHVLA. British farms 
#' are able to submit samples to AHVLA if cause of death for an animal
#' cannot be determined and private veterinary surgeon decides to submit them, 
#' unless there is notifiable disease suspected then such a submission is not 
#' required. 
#' 
#' This data set contains information about such farms.
#' All submissions from farms are included in this data frame
#' not only carcasses but also blood samples etc.
#' @docType data
#' @format Data frame with 12,036 rows and 4 columns.
#' \describe{
#'   \item{\code{TOTAL_SUB}}{Number of submissions of animal material.}
#'   \item{\code{log_size}}{Numerical value equal to logarithm of size of farm.}
#'   \item{\code{log_distance}}{Numerical value equal to logarithm of distance to nearest AHVLA center.}
#'   \item{\code{C_TYPE}}{Factor describing type of activity on farm that animals are 
#'   used for. Either \code{Dairy} or \code{Beef}}
#' }
#' @name farmsubmission
#' @references This data set and its description was provided in publication:
#' B&ouml;hning, D., Vidal Diez, A., Lerdsuwansri, R., Viwatwongkasem, C., and Arnold, M. (2013). "A generalization of Chao's estimator for covariate information". *Biometrics*, 69(4), 1033-1042. doi:10.1111/biom.12082
#' @usage data("farmsubmission")
"farmsubmission"

#' @title Immigration data in the Netherlands
#' @description  This data set contains information about immigrants in four
#' cities (Amsterdam, Rotterdam, The Hague and Utrecht) in Netherlands that have
#' been staying in the country illegally in 1995 and have appeared in police
#' records that year.
#' @docType data
#' @format Data frame with 1,880 rows and 5 columns.
#' \describe{
#'   \item{\code{capture}}{Number of times a person has been captured by police.}
#'   \item{\code{gender}}{Factor describing gender of the apprehended person.}
#'   \item{\code{age}}{Factor describing age of apprehended person. 
#'   Either below or above 40 years old.}
#'   \item{\code{reason}}{Factor describing reason for being apprehended by 
#'   police either illegal stay in Netherlands or other reasons.}
#'   \item{\code{nation}}{Factor with nation of origin of the captured person. 
#'   There are 6 levels of this variable: \code{"American and Australia",
#'   "Asia", "North Africa", "Rest of Africa", "Surinam", "Turkey"}.}
#' }
#' @name netherlandsimmigrant
#' @references This data set and its description was provided in publication:
#' van Der Heijden, P. G., Bustami, R., Cruyff, M. J., Engbersen, G., and Van Houwelingen, H. C. (2003). Point and interval estimation of the population size using the truncated Poisson regression model. *Statistical Modelling*, 3(4), 305-322. doi:10.1191/1471082X03st057oa
#' @usage data("netherlandsimmigrant")
"netherlandsimmigrant"

#' @title The British farm carcass submissions data
#' @description Data on British animal farms submissions to AHVLA. British farms 
#' are able to submit samples to AHVLA if cause of death for an animal
#' cannot be determined and private veterinary surgeon decides to submit them, 
#' unless there is notifiable disease suspected then such a submission is not 
#' required. 
#' 
#' This data set contains information about such farms.
#' Only submissions that are included in this data frame
#' are submissions of carcasses i.e. submissions of blood samples etc. 
#' are excluded.
#' @docType data
#' @format Data frame with 1,858 rows and 4 columns.
#' \describe{
#'   \item{\code{TOTAL_SUB}}{Number of submissions of animal carcasses.}
#'   \item{\code{log_size}}{Numerical value equal to logarithm of size of farm.}
#'   \item{\code{log_distance}}{Numerical value equal to logarithm of distance to nearest AHVLA center.}
#'   \item{\code{C_TYPE}}{Factor describing type of activity on farm that animals are 
#'   used for. Either \code{Dairy} or \code{Beef}}
#' }
#' @name carcassubmission
#' @references This data set and its description was provided in publication:
#' B&ouml;hning, D., Vidal Diez, A., Lerdsuwansri, R., Viwatwongkasem, C., and Arnold, M. (2013). "A generalization of Chao's estimator for covariate information". *Biometrics*, 69(4), 1033-1042. doi:10.1111/biom.12082
#' @usage data("carcassubmission")
"carcassubmission"

#' @rdname singleRmodels
#' @importFrom stats uniroot
#' @importFrom stats dnbinom
#' @importFrom stats optim
#' @export
ztnegbin <- function(nSim = 1000, epsSim = 1e-8, eimStep = 6,
                     lambdaLink = c("log", "neglog"), 
                     alphaLink = c("log", "neglog"),
                     ...) {
  if (missing(lambdaLink)) lambdaLink <- "log"
  if (missing(alphaLink))  alphaLink <- "log"
  
  links <- list()
  attr(links, "linkNames") <- c(lambdaLink, alphaLink)
  
  lambdaLink <- switch(lambdaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  alphaLink <- switch(alphaLink,
    "log"    = singleRinternallogLink,
    "neglog" = singleRinternalneglogLink
  )
  
  links[1:2] <- c(lambdaLink, alphaLink)
  
  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    
    if (!deriv) {
      switch (type,
          "nontrunc" = lambda,
          "trunc" = lambda / (1 - (1 + alpha * lambda) ^ (-1 / alpha))
      )
    } else {
      switch (type,
          "nontrunc" = {
            matrix(c(1, 0) * c(
              lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
              alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
            ), ncol = 2)
          },
          "trunc" = {
            matrix(c(
              (alpha * lambda + 1) ^ (1 / alpha - 1) *
              ((alpha * lambda + 1) ^ (1 / alpha + 1) +
              (-alpha - 1) * lambda - 1) / 
              ((alpha * lambda + 1) ^ (1 / alpha) - 1) ^ 2, 
              lambda * (lambda * alpha + 1) ^ (1 / alpha - 1) *
              ((lambda * alpha + 1) * log(lambda * alpha + 1) - lambda * alpha) /
              (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha) - 1) ^ 2)
            ) * c(
              lambdaLink(eta[, 1], inverse = TRUE, deriv = 1),
               alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
            ), ncol = 2)
          }
      )
    }
  }

  variance <- function(eta, type = "nontrunc", ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    P0 <- (1 + alpha * lambda) ^ (-1 / alpha)
    switch (type,
    nontrunc = lambda * (1 + alpha * lambda),
    trunc = (lambda + alpha * (lambda ^ 2) - alpha * (lambda ^ 2) * P0) / ((1 - P0) ^ 2)
    )
  }
  
  compdigamma <- function(y, alpha) {
    (-digamma(y + 1 / alpha) + digamma(1 / alpha)) / (alpha ^ 2)
  }
  
  
  comptrigamma <- function(y, alpha) {
    (2 * (digamma(y + 1 / alpha) - digamma(1 / alpha)) * alpha +
     trigamma(y + 1 / alpha) - trigamma(1 / alpha)) / (alpha ^ 4)
  }
  
  # Computing the expected value of di/trigamma functions on (y + 1/alpha)
  
  compExpect <- function(eta) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    
    P0 <- (1 + alpha * lambda) ^ (-1 / alpha)
    res <- rep(0, NROW(eta))
    k <- 1
    finished <- rep(FALSE, NROW(eta))
    while ((k < nSim) & !all(finished)) {
      prob <- apply(cbind(k:(k + eimStep)), MARGIN = 1, FUN = function(x) {
        stats::dnbinom(
          x = x, 
          size = 1 / alpha, 
          mu = lambda
        ) / (1 - P0)
      })
      trg <- apply(cbind(k:(k + eimStep)), MARGIN = 1, FUN = function(x) {
        comptrigamma(y = x, alpha = alpha)
      })
      prob[!(is.finite(prob))] <- 0
      trg[!(is.finite(trg))] <- 0
      toAdd <- trg * prob
      toAdd <- rowSums(toAdd)
      k <- k + eimStep + 1
      res <- res + toAdd
      finished <- abs(toAdd) < epsSim
    }
    res
  }
  
  Wfun <- function(prior, eta, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    Ey <- mu.eta(eta = eta)
    Etrig <- compExpect(eta)
    
    G00 <- prior * (Etrig + ((lambda * alpha + 1) ^ (1 / alpha) * 
    (lambda ^ 2 * alpha ^ 2 + 2 * lambda * alpha + 1) * log(lambda * alpha + 1) ^ 2 +
    ((lambda * alpha + 1) ^ (1 / alpha) * (2 * lambda ^ 2 * alpha ^ 3 + 
    (4 * lambda - 2 * lambda ^ 2) * alpha ^ 2 + (2 - 2 * lambda) * alpha) + 
    (lambda * alpha + 1) ^ (2 / alpha) * (-2 * lambda ^ 2 * alpha ^ 3 - 4 * lambda * alpha ^ 2 - 2 * alpha)) * 
    log(lambda * alpha + 1) + (lambda * alpha + 1) ^ (2 / alpha) * 
    ((3 * lambda ^ 2 - 2 * Ey * lambda) * alpha ^ 3 +
    (2 * lambda - Ey) * alpha ^ 2) + (lambda * alpha + 1) ^ (1 / alpha) *
    ((4 * Ey * lambda - 3 * lambda ^ 2) * alpha ^ 3 +
    (lambda ^ 2 - 2 * lambda + 2 * Ey) * alpha ^ 2) -
    2 * Ey * lambda * alpha ^ 3 - Ey * alpha ^ 2) /
    (alpha ^ 4 * (lambda * alpha + 1) ^ 2 * 
    ((lambda * alpha + 1) ^ (1 / alpha) - 1) ^ 2)) * 
    alphaLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2
    
    # mixed derivative
    G01 <- -((alpha * lambda + 1) ^ (1 / alpha + 1) *
    log(alpha * lambda + 1) + (alpha * lambda + 1) ^ (1 / alpha) *
    ((alpha ^ 2 - alpha) * lambda - 2 * alpha ^ 2 * Ey) +
    (alpha * lambda + 1) ^ (2 / alpha) * (alpha ^ 2 * Ey - alpha ^ 2 * lambda) + alpha ^ 2 * Ey) /
    (alpha ^ 2 * (alpha * lambda + 1) ^ 2 * ((alpha * lambda + 1) ^ (1 / alpha) - 1) ^ 2)
    
    G01 <- G01 * prior * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) *
           lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
    
    # second lambda derivative
    
    G11 <- prior * ((alpha * lambda + 1) ^ (2 / alpha) * (alpha * lambda ^ 2 - 2 * alpha * Ey * lambda - Ey) +
    (alpha * lambda + 1) ^ (1 / alpha) * ((1 - alpha) * lambda ^ 2 + 4 * alpha * Ey * lambda + 2 * Ey) - 2 * alpha * Ey * lambda - Ey) /
    (lambda ^ 2 * (alpha * lambda + 1) ^ 2 * ((alpha * lambda + 1) ^ (1 / alpha) - 1) ^ 2) *
    lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2
    
    matrix(
      -c(G11, # lambda
         G01, # mixed
         G01, # mixed
         G00  # alpha
      ),
      dimnames = list(rownames(eta), c("lambda", "mixed", "mixed", "alpha")),
      ncol = 4
    )
  }
  
  funcZ <- function(eta, weight, y, prior, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    dig <- compdigamma(y = y, alpha = alpha)
    weight <- weight / prior

    weight <- lapply(X = 1:nrow(weight), FUN = function (x) {
      matrix(as.numeric(weight[x, ]), ncol = 2)
    })
    
    # alpha derivative
    dig <- compdigamma(y = y, alpha = alpha)
    
    G0 <- (dig + ((lambda * alpha + 1) ^ (1 / alpha + 1) * log(lambda * alpha + 1) +
    (y - lambda) * alpha * (lambda * alpha + 1) ^ (1 / alpha) - y * alpha) /
    (alpha ^ 2 * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) - 1))) *
    alphaLink(eta[, 2], inverse = TRUE, deriv = 1)
    
    # lambda derivative
    G1 <- -((lambda - y) * (alpha * lambda + 1) ^ (1 / alpha) + y) /
    (lambda * (alpha * lambda + 1) * ((alpha * lambda + 1) ^ (1 / alpha) - 1)) *
    lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
    
    uMatrix <- matrix(c(G1, G0), ncol = 2)
    
    pseudoResid <- sapply(X = 1:length(weight), FUN = function (x) {
      xx <- solve(weight[[x]])
      xx %*% uMatrix[x, ]
    })
    
    pseudoResid <- t(pseudoResid)
    dimnames(pseudoResid) <- dimnames(eta)
    pseudoResid
  }

  minusLogLike <- function(y, X, 
                           weight    = 1, 
                           NbyK      = FALSE, 
                           vectorDer = FALSE, 
                           deriv     = 0,
                           offset, ...) {
    if (is.null(weight)) {
      weight <- 1
    }
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X) / 2), rep(0, NROW(X) / 2))
    }
    y <- as.numeric(y)
    X <- as.matrix(X)
    
    if (!(deriv %in% c(0, 1, 2))) 
      stop("Only score function and derivatives up to 2 are supported.")
    
    # to make it conform to how switch in R works, i.e. indexing begins with 1
    deriv <- deriv + 1
    
    switch (deriv,
      function(beta) {
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
        
        -sum(weight * (lgamma(y + 1 / alpha) - lgamma(1 / alpha) - lgamma(y + 1) - 
        (y + 1 / alpha) * log(1 + lambda * alpha) + y * log(lambda * alpha) - 
        log(1 - (1 + lambda * alpha) ^ (-1 / alpha))))
      },
      function(beta) {
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
        
        # alpha derivative
        dig <- compdigamma(y = y, alpha = alpha)
        
        G0 <- (dig + ((lambda * alpha + 1) ^ (1 / alpha + 1) * log(lambda * alpha + 1) +
        (y - lambda) * alpha * (lambda * alpha + 1) ^ (1 / alpha) - y * alpha) /
        (alpha ^ 2 * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) - 1))) *
        alphaLink(eta[, 2], inverse = TRUE, deriv = 1) * weight
        
        # lambda derivative
        G1 <- -((lambda - y) * (alpha * lambda + 1) ^ (1 / alpha) + y) /
        (lambda * (alpha * lambda + 1) * ((alpha * lambda + 1) ^ (1 / alpha) - 1)) *
        lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * weight
        
        if (NbyK) {
          XX <- 1:(attr(X, "hwm")[1])
          return(cbind(as.data.frame(X[1:nrow(eta), XX]) * G1, as.data.frame(X[-(1:nrow(eta)), -XX]) * G0))
        }
        if (vectorDer) {
          return(cbind(G1, G0))
        }
        
        as.numeric(c(G1, G0) %*% X)
      },
      function(beta) {
        lambdaPredNumber <- attr(X, "hwm")[1]
        eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
        lambda <- lambdaLink(eta[, 1], inverse = TRUE)
        alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
        res <- matrix(nrow = length(beta), 
                      ncol = length(beta), 
                      dimnames = list(names(beta), names(beta)))
        
        trig <- comptrigamma(y = y, alpha = alpha)
        dig <-  compdigamma(y = y, alpha = alpha)
        
        # 2nd alpha derivative
        G00 <- ((trig + ((lambda * alpha + 1) ^ (1 / alpha) * 
        (lambda ^ 2 * alpha ^ 2 + 2 * lambda * alpha + 1) * log(lambda * alpha + 1) ^ 2 +
        ((lambda * alpha + 1) ^ (1 / alpha) * (2 * lambda ^ 2 * alpha ^ 3 + 
        (4 * lambda - 2 * lambda ^ 2) * alpha ^ 2 + (2 - 2 * lambda) * alpha) + 
        (lambda * alpha + 1) ^ (2 / alpha) * (-2 * lambda ^ 2 * alpha ^ 3 - 4 * lambda * alpha ^ 2 - 2 * alpha)) * 
        log(lambda * alpha + 1) + (lambda * alpha + 1) ^ (2 / alpha) * 
        ((3 * lambda ^ 2 - 2 * y * lambda) * alpha ^ 3 +
        (2 * lambda - y) * alpha ^ 2) + (lambda * alpha + 1) ^ (1 / alpha) *
        ((4 * y * lambda - 3 * lambda ^ 2) * alpha ^ 3 +
        (lambda ^ 2 - 2 * lambda + 2 * y) * alpha ^ 2) -
        2 * y * lambda * alpha ^ 3 - y * alpha ^ 2) /
        (alpha ^ 4 * (lambda * alpha + 1) ^ 2 * 
        ((lambda * alpha + 1) ^ (1 / alpha) - 1) ^ 2)) * 
        alphaLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2 +
        (dig + ((lambda * alpha + 1) ^ (1 / alpha + 1) * log(lambda * alpha + 1) +
        (y - lambda) * alpha * (lambda * alpha + 1) ^ (1 / alpha) - y * alpha) /
        (alpha ^ 2 * (lambda * alpha + 1) * ((lambda * alpha + 1) ^ (1 / alpha) - 1))) *
        alphaLink(eta[, 2], inverse = TRUE, deriv = 2))
        
        # mixed derivative
        G01 <- -((alpha * lambda + 1) ^ (1 / alpha + 1) *
        log(alpha * lambda + 1) + (alpha * lambda + 1) ^ (1 / alpha) *
        ((alpha ^ 2 - alpha) * lambda - 2 * alpha ^ 2 * y) +
        (alpha * lambda + 1) ^ (2 / alpha) * (alpha ^ 2 * y - alpha ^ 2 * lambda) + alpha ^ 2 * y) /
        (alpha ^ 2 * (alpha * lambda + 1) ^ 2 * ((alpha * lambda + 1) ^ (1 / alpha) - 1) ^ 2)
        
        G01 <- G01 * alphaLink(eta[, 2], inverse = TRUE, deriv = 1) *
               lambdaLink(eta[, 1], inverse = TRUE, deriv = 1)
        
        # second lambda derivative
        
        G11 <- (((alpha * lambda + 1) ^ (2 / alpha) * (alpha * lambda ^ 2 - 2 * alpha * y * lambda - y) +
        (alpha * lambda + 1) ^ (1 / alpha) * ((1 - alpha) * lambda ^ 2 + 4 * alpha * y * lambda + 2 * y) - 2 * alpha * y * lambda - y) /
        (lambda ^ 2 * (alpha * lambda + 1) ^ 2 * ((alpha * lambda + 1) ^ (1 / alpha) - 1) ^ 2) *
        lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2 -
        ((lambda - y) * (alpha * lambda + 1) ^ (1 / alpha) + y) /
        (lambda * (alpha * lambda + 1) * ((alpha * lambda + 1) ^ (1 / alpha) - 1)) *
        lambdaLink(eta[, 1], inverse = TRUE, deriv = 2)) * weight
        
        res[-(1:lambdaPredNumber), -(1:lambdaPredNumber)] <- 
          t(as.data.frame(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)]) *
            G00 * weight) %*% X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)]
        
        res[1:lambdaPredNumber, 1:lambdaPredNumber] <- 
          t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber]) * 
            G11 * weight) %*% X[1:(nrow(X) / 2), 1:lambdaPredNumber]
        
        res[1:lambdaPredNumber, -(1:lambdaPredNumber)] <- 
          t(t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber]) * 
            G01 * weight) %*% as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)]))
        
        res[-(1:lambdaPredNumber), 1:lambdaPredNumber] <- 
          t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber]) * 
            G01 * weight) %*% as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)])
        
        res
      }
    )
  }

  validmu <- function(mu) {
    all(is.finite(mu)) && all(0 < mu)
  }

  devResids <- function (y, eta, wt, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    mu <- mu.eta(eta = eta)
    
    prob_ge1 <- pmax(1 - (1 + alpha * lambda) ^ (-1 / alpha), .Machine$double.xmin)
    logLikFit <- (
      lgamma(y + 1 / alpha) - lgamma(1 / alpha) - lgamma(y + 1) -
      (y + 1 / alpha) * log(1 + alpha * lambda) +
      y * log(lambda * alpha) - log(prob_ge1)
    )
    
    yUnq <- unique(y)
    
    if (any(yUnq > 77)) {
      warning("Currently numerical deviance is unreliable for counts greater than 78.")
    }
    
    findL <- function(t) {
      yNow <- yUnq[t]
      stats::optim(
        par = c(0, log(yNow), -10),
        fn = function(x) {
          s <- x[1]
          l <- exp(x[2])
          a <- exp(x[3])
          
          prob <- 1 - (1+a*l)^(-1/a)
          prob <- 1 / prob
          sum(c((l*prob - yNow) * 4.5,# s der
            yNow/l+(-yNow*a-1)/(1+a*l)-(1+a*l)^(-1-1/a)*prob+s*(prob-prob^2*(l*(1+a*l)^(-1-1/a))),# lambda der
            (log(l*a+1)/a^2-l/(a*(l*a+1)))/((l*a+1)^(1/a)*(1-1/(l*a+1)^(1/a)))+(s*l*(log(l*a+1)/a^2-l/(a*(l*a+1))))/((l*a+1)^(1/a)*(1-1/(l*a+1)^(1/a))^2)+log(l*a+1)/a^2+(l*(-1/a-yNow))/(l*a+1)+yNow/a-digamma(yNow+1/a)/a^2+digamma(1/a)/a^2,#alpha der
            lgamma(yNow+1/a)-lgamma(1/a) - lgamma(yNow+1)-(yNow+1/a)*log(1+a*l)+yNow*log(l*a)-log(1-(1+a*l)^(-1/a))) ^ 2) ^ .5
        },
        method = "BFGS",
        control = list(maxit = 10000, abstol = .Machine$double.eps, reltol = .Machine$double.eps)
      )$par
    }
    
    suppressWarnings({
      logLikIdeal <- sapply(1:length(yUnq), FUN = function(x) {
        ifelse(yUnq[x] == 1, 0, {
          xx <- findL(x)
          lagrange <- xx[1]
          l <- exp(xx[2])
          a <- exp(xx[3])
          (lgamma(yUnq[x] + 1 / a) - lgamma(1 / a) -
              lgamma(yUnq[x] + 1) - (yUnq[x] + 1 / a) * log(1 + a * l) +
              yUnq[x] * log(l * a) - log(1 - (1 + a * l) ^ (-1 / a)))
        })
      })
    })
    
    logLikIdeal <- sapply(1:length(y), FUN = function(x) {
      logLikIdeal[yUnq == y[x]]
    })

    diff <- logLikIdeal - logLikFit
    
    if (any(logLikFit > 0)) {
      warning("Dispersion parameter values are on the boundary of parameter space. Deviance residuals will be assigned 0 on these observations.")
      diff[logLikFit > 0]   <- 0
    } else if (any(diff < 0)) {
      warning("Numerical deviance finder found worse saturated likelihood than fitted model. Expect NA's in deviance/deviance residuals.")
      diff[diff < 0] <- 0
    }
    
    sign(y - mu) * sqrt(2 * wt * diff)
  }

  pointEst <- function (pw, eta, contr = FALSE, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    N <- pw / (1 - (1 + alpha * lambda) ^ (- 1 / alpha))
    if(!contr) {
      N <- sum(N)
    }
    N
  }

  popVar <- function (pw, eta, cov, Xvlm, ...) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    pr <- 1 - (1 + alpha * lambda) ^ (- 1 / alpha)

    # w.r to alpha
    bigTheta1 <- pw  *  alphaLink(eta[, 2], inverse = TRUE, deriv = 1) *
    ((lambda * alpha + 1) ^ (1 / alpha - 1) * ((lambda * alpha + 1) * log(lambda * alpha + 1) - lambda * alpha)) /
    (alpha ^ 2 * ((lambda * alpha + 1) ^ (1 / alpha) - 1) ^ 2)
    
    # w.r to lambda
    bigTheta2 <- -pw * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * 
    ((alpha * lambda + 1) ^ (1 / alpha - 1) / ((alpha * lambda + 1) ^ (1 / alpha) - 1) ^ 2)

    bigTheta <- t(c(bigTheta2, bigTheta1) %*% Xvlm)

    f1 <-  t(bigTheta) %*% as.matrix(cov) %*% bigTheta
    f2 <-  sum(pw * (1 - pr) / (pr ^ 2))

    f1 + f2
  }
  
  dFun <- function (x, eta, type = c("trunc", "nontrunc")) {
    if (missing(type)) type <- "trunc"
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)

    switch (type,
      "trunc" = {
        stats::dnbinom(x = x, mu = lambda, size = 1 / alpha) / 
        (1 - (1 + alpha * lambda) ^ (-1 / alpha))
      },
      "nontrunc" = stats::dnbinom(x = x, mu = lambda, size = 1 / alpha)
    )
  }

  simulate <- function(n, eta, lower = 0, upper = Inf) {
    lambda <- lambdaLink(eta[, 1], inverse = TRUE)
    alpha  <-  alphaLink(eta[, 2], inverse = TRUE)
    lb <- stats::pnbinom(lower, mu = lambda, size = 1 / alpha)
    ub <- stats::pnbinom(upper, mu = lambda, size = 1 / alpha)
    p_u <- stats::runif(n, lb, ub)
    sims <- stats::qnbinom(p_u, mu = lambda, size = 1 / alpha)
    sims
  }
  
  getStart <- expression(
    if (method == "IRLS") {
      init <- log(abs((observed / weighted.mean(observed, priorWeights) - 1) / observed) + .1)
      etaStart <- cbind(
        pmin(family$links[[1]](observed), family$links[[1]](12)),
        family$links[[2]](ifelse(init < -.5, .1, init + .55))
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed, priorWeights)),
        family$links[[2]](abs((cov.wt(cbind(observed, observed), wt = priorWeights, method = "ML")$cov[1,1] / weighted.mean(observed, priorWeights) - 1) / weighted.mean(observed, priorWeights)) + .1)
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
      if ("(Intercept):alpha" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[2], rep(0, attr(Xvlm, "hwm")[2] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[2] / attr(Xvlm, "hwm")[2], attr(Xvlm, "hwm")[2]))
      }
    }
  )
  
  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "ztnegbin",
      etaNames  = c("lambda", "alpha"),
      simulate  = simulate,
      getStart  = getStart,
      extraInfo = c(
        mean       = "lambda",
        variance   = "lambda * (1 + alpha * lambda)",
        popSizeEst = "1 / (1 - (1 + alpha * lambda) ^ (- 1 / alpha))",
        meanTr     = "lambda / (1 - (1 + alpha * lambda) ^ (-1 / alpha))",
        varianceTr =
          "(lambda + alpha * (lambda ^ 2) - alpha * (lambda ^ 2) * (1 + alpha * lambda) ^ (-1 / alpha)) / ((1 - (1 + alpha * lambda) ^ (-1 / alpha)) ^ 2)"
      )
    ),
    class = c("singleRfamily", "family")
  )
}

source("helper_families.R")
zerotruncated_test_data <- load_zerotruncated_test_data()
eps <- zerotruncated_test_data$eps
N <- zerotruncated_test_data$N
test_zerotruncated <- zerotruncated_test_data$test_zerotruncated
x1 <- zerotruncated_test_data$x1
x2 <- zerotruncated_test_data$x2
x3 <- zerotruncated_test_data$x3

## ztnegbin ####
coefficients_ztnegbin <- c(.3, .2, -.3, .1,
          .2, .1, -.5, .3)
#eta <- cbind(coefficients_ztnegbin[1] + coefficients_ztnegbin[2] * x1 + coefficients_ztnegbin[3] * x2 + coefficients_ztnegbin[4] * x3,
#             coefficients_ztnegbin[5] + coefficients_ztnegbin[6] * x1 + coefficients_ztnegbin[7] * x2 + coefficients_ztnegbin[8] * x3)

model_ztnegbin <- ztnegbin(
  lambdaLink = "log",
  alphaLink = "log",
  eimStep = 10
)

#y <- model_ztnegbin$simulate(n = N, eta = eta, lower = -1)
observed_ztnegbin <- test_zerotruncated$ztnegbin1

data_ztnegbin <- data.frame(
  x3 = x3[observed_ztnegbin > 0],
  x2 = x2[observed_ztnegbin > 0],
  x1 = x1[observed_ztnegbin > 0],
  y = observed_ztnegbin[observed_ztnegbin > 0]
)

fit_ztnegbin <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = model_ztnegbin,
  data    = data_ztnegbin,
  method = "IRLS",
  controlModel = controlModel(
    alphaFormula = ~ x1 + x2 + x3
  )
)

expect_silent(
  population_ztnegbin <- popSizeEst(fit_ztnegbin)
)

expect_equivalent(
  population_ztnegbin$pointEstimate,
  N,
  tol = .3
)

expect_true(
  all(N < population_ztnegbin$confidenceInterval$upperBound) &
  all(N > population_ztnegbin$confidenceInterval$lowerBound)
)

expect_silent(
  strata_ztnegbin <- stratifyPopsize(fit_ztnegbin, ~ x1)
)

expect_equivalent(
  as.numeric(table(data_ztnegbin$x1))[as.numeric(substr(strata_ztnegbin$name, start = 5, stop = 5)) + 1],
  strata_ztnegbin$Observed
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(strata_ztnegbin$name, start = 5, stop = 5)) + 1],
  strata_ztnegbin$Estimated,
  tolerance = .5
)

# different link
coefficients_ztnegbin <- c(.1, .2, -.3, .1,
          .2, .1, -.1,   .3)

model_ztnegbin <- ztnegbin(
  lambdaLink = "log",
  alphaLink = "neglog",
  eimStep = 40
)

#y <- model_ztnegbin$simulate(n = N, eta = eta, lower = -1)
observed_ztnegbin <- test_zerotruncated$ztnegbin2

data_ztnegbin <- data.frame(
  x3 = x3[observed_ztnegbin > 0],
  x2 = x2[observed_ztnegbin > 0],
  x1 = x1[observed_ztnegbin > 0],
  y = observed_ztnegbin[observed_ztnegbin > 0]
)

fit_ztnegbin <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = model_ztnegbin,
  data    = data_ztnegbin,
  method = "IRLS",
  controlModel = controlModel(
    alphaFormula = ~ x1 + x2 + x3
  )
)

expect_true(
  max(abs(coef(fit_ztnegbin) - coefficients_ztnegbin)) < .9
)

expect_silent(
  population_ztnegbin <- popSizeEst(fit_ztnegbin)
)

expect_equivalent(
  N,
  population_ztnegbin$pointEstimate,
  tol = .1
)

expect_true(
  all(N < population_ztnegbin$confidenceInterval$upperBound) &
  all(N > population_ztnegbin$confidenceInterval$lowerBound)
)

expect_silent(
  strata_ztnegbin <- stratifyPopsize(fit_ztnegbin, ~ x1)
)

expect_equivalent(
  as.numeric(table(data_ztnegbin$x1))[as.numeric(substr(strata_ztnegbin$name, start = 5, stop = 5)) + 1],
  strata_ztnegbin$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(strata_ztnegbin$name, start = 5, stop = 5)) + 1],
  strata_ztnegbin$Estimated,
  tolerance = .1
)

marginal_ztnegbin <- summary(marginalFreq(fit_ztnegbin), dropl5 = "no")

expect_true(
  all(marginal_ztnegbin$Test$`P(>X^2)` > .05)
)

expect_true(
  all(predict(
    fit_ztnegbin,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztnegbin,
    type = "link",
    se.fit = TRUE
  )[,3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztnegbin,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)]> 0)
)

expect_true(
  sum(predict(
    fit_ztnegbin,
    type = "contr"
  )) == fit_ztnegbin$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztnegbin,
    type = "popSize",
    se.fit = TRUE
  )
)

# different link
model_ztnegbin <- ztnegbin(
  lambdaLink = "neglog",
  alphaLink = "log",
  eimStep = 10
)

#y <- model_ztnegbin$simulate(n = N, eta = eta, lower = -1)
observed_ztnegbin <- test_zerotruncated$ztnegbin3

data_ztnegbin <- data.frame(
  x3 = x3[observed_ztnegbin > 0],
  x2 = x2[observed_ztnegbin > 0],
  x1 = x1[observed_ztnegbin > 0],
  y = observed_ztnegbin[observed_ztnegbin > 0]
)

fit_ztnegbin <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = model_ztnegbin,
  data    = data_ztnegbin,
  method = "IRLS",
  controlModel = controlModel(
    alphaFormula = ~ x1 + x2 + x3
  )
)

expect_silent(
  population_ztnegbin <- popSizeEst(fit_ztnegbin)
)

expect_equivalent(
  N,
  population_ztnegbin$pointEstimate,
  tol = .6
)

expect_true(
  (N < population_ztnegbin$confidenceInterval$upperBound[2]) &
  (N > population_ztnegbin$confidenceInterval$lowerBound[2])
)

expect_silent(
  strata_ztnegbin <- stratifyPopsize(fit_ztnegbin, ~ x1)
)

expect_equivalent(
  as.numeric(table(data_ztnegbin$x1))[as.numeric(substr(strata_ztnegbin$name, start = 5, stop = 5)) + 1],
  strata_ztnegbin$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(strata_ztnegbin$name, start = 5, stop = 5)) + 1],
  strata_ztnegbin$Estimated,
  tolerance = .6
)

expect_true(
  all(predict(
    fit_ztnegbin,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztnegbin,
    type = "link",
    se.fit = TRUE
  )[, 3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztnegbin,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)] >= 0)
)

expect_true(
  sum(predict(
    fit_ztnegbin,
    type = "contr"
  )) == fit_ztnegbin$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztnegbin,
    type = "popSize",
    se.fit = TRUE
  )
)

# different link
coefficients_ztnegbin <- c(.6, -.2, -1.25, -.1,
          .2, -.1, -.5,   .3)
#eta <- cbind(coefficients_ztnegbin[1] + coefficients_ztnegbin[2] * x1 + coefficients_ztnegbin[3] * x2 + coefficients_ztnegbin[4] * x3,
#             coefficients_ztnegbin[5] + coefficients_ztnegbin[6] * x1 + coefficients_ztnegbin[7] * x2 + coefficients_ztnegbin[8] * x3)

model_ztnegbin <- ztnegbin(
  lambdaLink = "log",
  alphaLink = "log",
  eimStep = 25
)

#y <- model_ztnegbin$simulate(n = N, eta = eta, lower = -1)
observed_ztnegbin <- test_zerotruncated$ztnegbin4

data_ztnegbin <- data.frame(
  x3 = x3[observed_ztnegbin > 0],
  x2 = x2[observed_ztnegbin > 0],
  x1 = x1[observed_ztnegbin > 0],
  y = observed_ztnegbin[observed_ztnegbin > 0]
)

fit_ztnegbin <- estimatePopsize(
  formula = y ~ x1 + x2 + x3,
  model   = model_ztnegbin,
  data    = data_ztnegbin,
  method = "IRLS",
  controlModel = controlModel(
    alphaFormula = ~ x1 + x2 + x3
  )
)

expect_equivalent(
  coefficients_ztnegbin,
  coef(fit_ztnegbin),
  tolerance = .4
)

expect_silent(
  population_ztnegbin <- popSizeEst(fit_ztnegbin)
)

expect_equivalent(
  N,
  population_ztnegbin$pointEstimate,
  tol = .15
)

expect_true(
  all(N < population_ztnegbin$confidenceInterval$upperBound) &
  all(N > population_ztnegbin$confidenceInterval$lowerBound)
)

expect_silent(
  strata_ztnegbin <- stratifyPopsize(fit_ztnegbin, ~ x1)
)

expect_equivalent(
  as.numeric(table(data_ztnegbin$x1))[as.numeric(substr(strata_ztnegbin$name, start = 5, stop = 5)) + 1],
  strata_ztnegbin$Observed,
  tolerance = eps
)

expect_equivalent(
  as.numeric(table(x1))[as.numeric(substr(strata_ztnegbin$name, start = 5, stop = 5)) + 1],
  strata_ztnegbin$Estimated,
  tolerance = .15
)

marginal_ztnegbin <- summary(marginalFreq(fit_ztnegbin), dropl5 = "group")

expect_true(
  all(marginal_ztnegbin$Test$`P(>X^2)` > .05)
)

expect_true(
  all(predict(
    fit_ztnegbin,
    type = "response",
    se.fit = TRUE
  )[, 3:4] > 0)
)

expect_true(
  all(predict(
    fit_ztnegbin,
    type = "link",
    se.fit = TRUE
  )[, 3:4]> 0)
)

expect_true(
  all(predict(
    fit_ztnegbin,
    type = "mean",
    se.fit = TRUE
  )[, c(2, 4)] >= 0)
)

expect_true(
  sum(predict(
    fit_ztnegbin,
    type = "contr"
  )) == fit_ztnegbin$populationSize$pointEstimate
)

expect_silent(
  predict(
    fit_ztnegbin,
    type = "popSize",
    se.fit = TRUE
  )
)

#' @importFrom graphics points
#' @importFrom stats rbinom
parBoot <- function(family, formulas, y, X, beta, weights,
                    trcount, numboot, eta, trace, visT,
                    controlBootstrapMethod = NULL, method,
                    modelFrame, offset, weightsFlag, ...) {
  strappedStatistic <- vector("numeric", length = numboot)
  n <- length(y)
  if (isTRUE(weightsFlag)) {
    n <- sum(weights)
  }
  if (length(weights) == 1) {
    weights <- rep(1, length(y))
  }
  
  if (family$family %in% c("chao", "zelterman", "oichao")) {
    message(paste(
      "Probability model will be taken as given by poisson distribution",
      "since rare-count chao-type models are based on mixture of poisson",
      "distribution semi-parametric bootstrap may be a better choice.", 
      sep = "\n"
    ))
  }
  
  contr <- family$pointEst(pw = weights,
                           eta = eta,
                           contr = TRUE,
                           y = y)
  
  N <- sum(contr)
  
  prob <- contr / N
  
  if (visT) {
    plot(
      1, type = "n",
      xlab = "Bootstrap sample", 
      ylab = "Value of population size estimator",
      main = expression(paste("Plot of values of ", hat(N), " obtained from bootstrap samples")),
      sub = "Points will be added in real time",
      xlim = c(0, numboot + 1), ylim = c(0, 2.5 * N)
    )
  }
  
  k <- 1
  while (k <= numboot) {
    nn <- floor(N) + stats::rbinom(n = 1, size = 1, prob = N - floor(N))
    if (isTRUE(weightsFlag)) {
      strap <- sample.int(replace = TRUE, n = length(y), size = nn, prob = prob)
      
      weightsStrap <- as.numeric(weights[strap])
      offsetStrap  <- offset[strap, , drop = FALSE]
      Xstrap       <- modelFrame[strap, , drop = FALSE]
      
      colnames(Xstrap) <- colnames(modelFrame)
      
      Xstrap <- singleRinternalGetXvlmMatrix(X = Xstrap, 
                                             formulas = formulas, 
                                             family$etaNames)
      
      ystrap <- family$simulate(
        n = nn,
        eta = matrix(Xstrap %*% beta, ncol = length(family$etaNames)),
        lower = -1, upper = Inf
      )
      offsetStrap  <- offsetStrap[ystrap > 0, , drop = FALSE]
      
      strap <- rep(ystrap > 0, length(family$etaNames))
      hwm <- attr(Xstrap, "hwm")
      
      Xstrap <- Xstrap[strap, , drop = FALSE]
      ystrap <- ystrap[ystrap > 0]
      
      weightsStrap <- rep(1, length(ystrap))
    } else {
      strap <- sample.int(replace = TRUE, n = length(y), size = nn, prob = prob)
      
      weightsStrap <- as.numeric(weights[strap])
      offsetStrap  <- offset[strap, , drop = FALSE]
      Xstrap       <- modelFrame[strap, , drop = FALSE]
      
      colnames(Xstrap) <- colnames(modelFrame)
      
      Xstrap <- singleRinternalGetXvlmMatrix(X = Xstrap, 
                                             formulas = formulas, 
                                             family$etaNames)
      
      ystrap <- family$simulate(
        n = nn,
        eta = matrix(Xstrap %*% beta, ncol = length(family$etaNames)),
        lower = -1, upper = Inf
      )
      
      weightsStrap <- weightsStrap[ystrap > 0]
      offsetStrap  <- offsetStrap[ystrap > 0, , drop = FALSE]
      
      strap <- rep(ystrap > 0, length(family$etaNames))
      hwm <- attr(Xstrap, "hwm")
      
      Xstrap <- Xstrap[strap, , drop = FALSE]
      ystrap <- ystrap[ystrap > 0]
    }
    
    if (isTRUE(trace)) cat("Iteration number:", k, 
                           "sample size:", length(ystrap), sep = " ")
    
    theta <- NULL
    attr(Xstrap, "hwm") <- hwm
    
    try(
      theta <- estimatePopsizeFit(
        y = ystrap,
        X = Xstrap,
        family = family,
        control = controlBootstrapMethod,
        method = method,
        priorWeights = weightsStrap,
        coefStart = jitter(beta),
        etaStart = matrix(Xstrap %*% jitter(beta), ncol = NCOL(offsetStrap)) + offsetStrap,
        offset = offsetStrap
      )$beta,
      silent = TRUE
    )
    
    k <- k + 1
    
    if (is.null(theta)) {
      if (isTRUE(trace)) cat("\n")
      k <- k - 1
    } else {
      theta <- matrix(Xstrap %*% theta, ncol = length(family$etaNames)) + offsetStrap
      
      est <- family$pointEst(pw = weightsStrap, eta = theta, y = ystrap)
      if (visT) graphics::points(k - 1, est, pch = 1)
      if (isTRUE(trace)) cat(" Estimated population size: ", est,"\n",sep = "")
      #if (visT) graphics::points(k - 1, est, pch = 1)
      
      strappedStatistic[k - 1] <- est
    }
    
  }
  
  strappedStatistic
}
# multicore
#' @importFrom foreach %dopar%
#' @importFrom foreach foreach
#' @importFrom parallel makeCluster
#' @importFrom parallel stopCluster
#' @importFrom doParallel registerDoParallel
#' @importFrom stats rbinom
parBootMultiCore <- function(family, formulas, y, X, modelFrame,
                             beta, weights, trcount, numboot,
                             eta, cores, controlBootstrapMethod = NULL,
                             method, N, offset, weightsFlag, visT, ...) {
  n <- length(y)
  if (isTRUE(weightsFlag)) {
    n <- sum(weights)
  }
  
  if (length(weights) == 1) {
    weights <- rep(1, length(y))
  }
  
  if (family$family %in% c("chao", "zelterman", "oichao")) {
    message(paste("Probability model will be taken as given by poisson distribution",
                  "since rare-count chao-type models are based on mixture of poisson",
                  "distribution semi-parametric bootstrap may be a better choice.", 
                  sep = "\n"))
  }
  
  
  contr <- family$pointEst(pw = weights,
                           eta = eta,
                           contr = TRUE,
                           y = y)
  N <- sum(contr)
  
  prob <- contr / N
  
  cl <- parallel::makeCluster(cores)
  doParallel::registerDoParallel(cl)
  on.exit(parallel::stopCluster(cl))
  
  strappedStatistic <- foreach::`%dopar%`(
    obj = foreach::foreach(k = 1:numboot, .combine = c),
    ex = {
      theta <- NULL
      while (is.null(theta)) {
        nn <- floor(N) + stats::rbinom(n = 1, size = 1, prob = N - floor(N))
        if (isTRUE(weightsFlag)) {
          strap <- sample.int(replace = TRUE, n = length(y), size = nn, prob = prob)
          
          weightsStrap <- as.numeric(weights[strap])
          offsetStrap  <- offset[strap, , drop = FALSE]
          Xstrap       <- modelFrame[strap, , drop = FALSE]
          
          colnames(Xstrap) <- colnames(modelFrame)
          
          Xstrap <- singleRinternalGetXvlmMatrix(X = Xstrap, 
                                                 formulas = formulas, 
                                                 family$etaNames)
          
          ystrap <- family$simulate(
            n = nn,
            eta = matrix(Xstrap %*% beta, ncol = length(family$etaNames)),
            lower = -1, upper = Inf
          )
          offsetStrap  <- offsetStrap[ystrap > 0, , drop = FALSE]
          
          strap <- rep(ystrap > 0, length(family$etaNames))
          hwm <- attr(Xstrap, "hwm")
          
          Xstrap <- Xstrap[strap, , drop = FALSE]
          ystrap <- ystrap[ystrap > 0]
          
          weightsStrap <- rep(1, length(ystrap))
        } else {
          strap <- sample.int(replace = TRUE, n = length(y), size = nn, prob = prob)
          
          weightsStrap <- as.numeric(weights[strap])
          offsetStrap  <- offset[strap, , drop = FALSE]
          Xstrap       <- modelFrame[strap, , drop = FALSE]
          
          colnames(Xstrap) <- colnames(modelFrame)
          
          Xstrap <- singleRinternalGetXvlmMatrix(X = Xstrap, 
                                                 formulas = formulas, 
                                                 family$etaNames)
          
          ystrap <- family$simulate(
            n = nn,
            eta = matrix(Xstrap %*% beta, ncol = length(family$etaNames)),
            lower = -1, upper = Inf
          )
          
          weightsStrap <- weightsStrap[ystrap > 0]
          offsetStrap  <- offsetStrap[ystrap > 0, , drop = FALSE]
          
          strap <- rep(ystrap > 0, length(family$etaNames))
          hwm <- attr(Xstrap, "hwm")
          
          Xstrap <- Xstrap[strap, , drop = FALSE]
          ystrap <- ystrap[ystrap > 0]
        }
        
        attr(Xstrap, "hwm") <- hwm
        
        try(
          theta <- estimatePopsizeFit(
            y = ystrap,
            X = Xstrap,
            family = family,
            control = controlBootstrapMethod,
            method = method,
            priorWeights = weightsStrap,
            coefStart = jitter(beta),
            etaStart = matrix(Xstrap %*% jitter(beta), ncol = NCOL(offsetStrap)) + offsetStrap,
            offset = offsetStrap
          )$beta,
          silent = TRUE
        )
      }
      family$pointEst(
        pw = weightsStrap, 
        eta = matrix(Xstrap %*% theta, 
                     ncol = length(family$etaNames)) + offsetStrap, 
        y = ystrap
      )
    }
  )
  
  strappedStatistic
}

#' @title Observed and fitted marginal frequencies
#' @author Piotr Chlebicki
#' \loadmathjax
#' @description A function that given a fitted \code{singleR} class object 
#' computed marginal frequencies by as sum of probability density functions
#' for each unit in data at each point i.e. kth element of marginal frequency
#' table is given by \mjseqn{\sum_{j=1}^{N_{obs}}\mathbb{P}(Y_{j}=k|\eta_{j})}. 
#' For k=0 only (if specified at call) they are computed as 
#' \mjseqn{\hat{N}-N_{obs}} because 
#' \mjseqn{\boldsymbol{f}_{0}} is assumed to 
#' the unobserved part of the studied population.
#' 
#' These frequencies are useful in diagnostics for count data regression, such
#' as assessment of fit.
#'
#' @param object object of \code{singleR} class.
#' @param includeones logical value indicating whether to include the estimated number of zero counts.
#' @param includezeros logical value indicating whether to include one counts in the zero-one truncated models.
#' @param onecount a numeric value indicating number of one counts if null \code{trcount} from object will be assumed to be a number one counts.
#' @param range optional argument specifying range of selected Y values.
#' @param ... currently does nothing.
#'
#' @return A list with observed name of the fitted model family degrees of freedom and observed and fitted marginal frequencies.
#' @seealso [estimatePopsize()] -- where example of usage is provided
#' @export
marginalFreq <- function(object,
                         includeones = TRUE,
                         includezeros = TRUE,
                         onecount = NULL,
                         range,
                         ...) {
  y <- if (is.null(object$y)) stats::model.response(model.frame(object)) else object$y
  if (missing(range)) {range <- (min(y):max(y))}
  y <- table(y)[names(table(y)) %in% as.character(range)]
  y <- y[!is.na(y)]
  
  trcount <- object$trcount
  if(!is.null(onecount)) {
    trcount <- onecount
  }
  
  # PMF for truncated distributions:
  probFun <- object$model$densityFunction
  
  res <- sapply(
    1:nrow(object$linearPredictors), 
    FUN = function(x) {object$model$densityFunction(
      x = range, 
      eta = object$linearPredictors[x, , drop = FALSE]
    )}
  )
  res <- rowSums(res)
  names(res) <- as.character(range)
  
  if(isTRUE(includeones) & grepl(object$model$family, pattern = "^zot")) {
    res <- c(trcount, res)
    names(res)[1] <- "1"
    y <- c(trcount, y)
    names(y)[1] <- "1"
    if (!is.null(onecount)) {res[1] <- onecount; y[1] <- onecount}
  }
  
  if(isTRUE(includezeros)) {
    res <- c(object$populationSize$pointEstimate - length(object$y), res)
    names(res)[1] <- "0"
  }
  res <- structure(list(
    table = res, y = y, 
    df = length(y) - length(object$coefficients) - 1,
    name = object$model$family
  ),
  class = c("singleRmargin"))
  res
}

#' @title Statistical tests of goodness of fit
#'
#' @description Performs two statistical test on observed and fitted
#' marginal frequencies. For G test the test statistic is computed as:
#' \loadmathjax
#' \mjsdeqn{G = 2\sum_{k}O_{k}\ln{\left(\frac{O_{k}}{E_{k}}\right)}}
#' and for \mjseqn{\chi^{2}} the test statistic is computed as:
#' \mjsdeqn{\chi^{2} = \sum_{k}\frac{\left(O_{k}-E_{k}\right)^{2}}{E_{k}}}
#' where \mjseqn{O_{k},E_{k}} denoted observed and fitted 
#' frequencies respectively. Both of these statistics converge to 
#' \mjseqn{\chi^2} distribution asymptotically with the same 
#' degrees of freedom.
#' 
#' The convergence of \mjseqn{G, \chi^2} statistics to 
#' \mjseqn{\chi^2} distribution may be violated if expected counts 
#' in cells are too low, say < 5, so it is customary to either censor or 
#' omit these cells.
#' 
#' @param object object of singleRmargin class.
#' @param df degrees of freedom if not provided the function will try and manually
#' but it is not always possible.
#' @param dropl5 a character indicating treatment of cells with frequencies < 5 
#' either grouping them, dropping or leaving them as is. Defaults to \code{drop}.
#' @param ... currently does nothing.
#'
#' @method summary singleRmargin
#' @return A chi squared test and G test for comparison between fitted and 
#' observed marginal frequencies.
#' @examples 
#' # Create a simple model
#' Model <- estimatePopsize(
#'   formula = capture ~ ., 
#'   data = netherlandsimmigrant, 
#'   model = ztpoisson, 
#'   method = "IRLS"
#' )
#' plot(Model, "rootogram")
#' # We see a considerable lack of fit
#' summary(marginalFreq(Model), df = 1, dropl5 = "group")
#' @exportS3Method
summary.singleRmargin <- function(object, df,
                                  dropl5 = c("drop", 
                                             "group", 
                                             "no"), 
                                  ...) {
  if (missing(dropl5)) {
    dropl5 <- "drop"
  }
  
  if (!is.character(dropl5) | length(dropl5) > 1) {
    warning("The argument dropl5 should be a 1 length character vector")
    dropl5 <- dropl5[1]
  }
  
  y <- object$y
  
  if (missing(dropl5) | !is.character(dropl5)) {dropl5 <- "no"}
  if (grepl("zot", object$name) & (1 %in% names(y))) {y <- y[-1]}
  
  if ((missing(df)) && (object$df < 1)) {
    warning("Degrees of freedom may be inacurate.")
    df <- 1
  } else if (missing(df)) {
    df <- object$df
  }
  
  A <- object$table[names(y)]
  
  switch (dropl5,
    "group" = {
      l <- (A < 5)
      if(!all(l == FALSE)) {
        y <- c(y[!l], sum(y[l]))
        A <- c(A[!l], sum(A[l]))
      }
    },
    "drop" = {
      l <- (A < 5)
      y <- y[!l]
      A <- A[!l]
    }
  )
  
  X2 <- sum(((A - y) ^ 2) / A)
  G <- 2 * sum(y * log(y / A))
  
  pval <- stats::pchisq(q = c(X2, G), df = df, lower.tail = FALSE)
  
  vect <- data.frame(
    round(c(X2, G), digits = 2),
    rep(df, 2), signif(pval, digits = 2)
  )
  rownames(vect) <- c("Chi-squared test", "G-test")
  colnames(vect) <- c("Test statistics", "df", "P(>X^2)")
  
  structure(
    list(Test = vect,
         l5 = switch(dropl5,
           "drop"  = "dropped",
           "group" = "grouped",
           "no"    = "preserved"
         ),
         y = y),
    class = "summarysingleRmargin"
  )
}


#' @method print summarysingleRmargin
#' @exportS3Method 
print.summarysingleRmargin <- function(x, ...) {
  cat("Test for Goodness of fit of a regression model:\n",
      "\n", sep = "")
  
  print(x$Test)
  
  cat("\n--------------------------------------------------------------",
      "\nCells with fitted frequencies of < 5 have been", x$l5, 
      "\nNames of cells used in calculating test(s) statistic:", names(x$y), 
      "\n", sep = " ")
  invisible()
}

#' @title Estimate Population Size Using Ratio Regression
#' @author Cyprian Jurkowski, Piotr Chlebicki, Maciej Beręsewicz
#'
#' @description
#' Fits the ratio-regression approach for single-source capture-recapture
#' count data as a standalone estimator. The regression is performed on
#' neighbouring marginal-frequency ratios rather than on unit-level counts,
#' which is why this method is separate from the `singleRmodels()` family
#' framework used by [estimatePopsize()].
#'
#' @details
#' Let \eqn{f_k} denote the observed frequency of units captured exactly
#' \eqn{k} times. Ratio regression models the neighbouring ratios
#' \eqn{r_k = f_{k + 1} / f_k}. The default model uses
#' \eqn{\log(r_k) = \beta_0 + \beta_1 \log(k + 1)} and weighted least squares
#' with weights \eqn{1 / (1 / f_k + 1 / f_{k + 1})}.
#'
#' Two published models are supported:
#'
#' - `M0`: ratio regression without one-inflation term
#' - `M1`: ratio regression with an additional one-inflation indicator
#'   `I(k == 1)`
#'
#' When `model = "auto"`, both models are fitted and compared by `criterion`.
#' Population size can then be extracted either through the Horvitz-Thompson
#' (`HT`) or semi-parametric (`SM`) estimator, with `estimator` selecting which
#' one is treated as primary by [popSizeEst()], [summary()], and [plot()].
#'
#' Bootstrap intervals are percentile intervals. For fixed `model = "M0"` or
#' `model = "M1"` the bootstrap uses zero-count imputation based on the
#' primary estimator. For `model = "auto"` it uses the single-bootstrap
#' reselection scheme where model selection is repeated inside each bootstrap
#' sample.
#'
#' @param formula a formula identifying the observed count column. It must use
#'   `~ 1`; for example `observed ~ 1`.
#' @param data a data frame or an object coercible to `data.frame`.
#' @param ratioFormula a one-sided formula used for the ratio regression on the
#'   derived ratio-level data. The default is `~ log(k + 1)`.
#' @param model which ratio-regression model to use: `M0`, `M1`, or `auto` to
#'   select between them by `criterion`.
#' @param criterion information criterion used when `model = "auto"`.
#' @param estimator which population-size estimator should be treated as the
#'   primary one.
#' @param weights optional prior weights interpreted as multiplicities of
#'   observed units.
#' @param subset optional logical expression selecting rows from `data`.
#' @param confint either `"bootstrap"` for percentile bootstrap intervals or
#'   `"none"` to skip interval estimation.
#' @param B number of bootstrap replicates when `confint = "bootstrap"`.
#' @param seed optional random seed used for bootstrap sampling.
#' @param maxCount optional upper support bound. Counts larger than `maxCount`
#'   are discarded before fitting.
#' @param ... additional arguments passed to [stats::model.frame()].
#'
#' @return
#' An object of class `singleRRatioReg` containing the fitted ratio-regression
#' models, model-selection criteria, derived ratio-level data, observed
#' frequencies, and population-size estimates.
#'
#' @examples
#' toy_counts <- data.frame(
#'   observed = c(rep(1, 50), rep(2, 30), rep(3, 15), rep(4, 6), rep(5, 2))
#' )
#'
#' toy_fit <- ratioReg(
#'   observed ~ 1,
#'   data = toy_counts,
#'   model = "auto",
#'   estimator = "SM",
#'   confint = "none"
#' )
#'
#' summary(toy_fit)
#' popSizeEst(toy_fit)
#' plot(toy_fit)
#'
#' toy_tab <- aggregate(list(weight = rep(1, nrow(toy_counts))),
#'                      by = list(observed = toy_counts$observed),
#'                      FUN = sum)
#' ratioReg(
#'   observed ~ 1,
#'   data = toy_tab,
#'   weights = toy_tab$weight,
#'   model = "M0",
#'   confint = "none"
#' )
#'
#' @seealso [estimatePopsize()] [popSizeEst()]
#' @export
ratioReg <- function(formula,
                     data,
                     ratioFormula = ~ log(k + 1),
                     model = c("auto", "M0", "M1"),
                     criterion = c("AIC", "BIC"),
                     estimator = c("SM", "HT"),
                     weights = NULL,
                     subset,
                     confint = c("bootstrap", "none"),
                     B = 1000,
                     seed = NULL,
                     maxCount = NULL,
                     ...) {
  model <- match.arg(model)
  criterion <- match.arg(criterion)
  estimator <- match.arg(estimator)
  confint <- match.arg(confint)

  if (!is.numeric(B) || length(B) != 1L || !is.finite(B) || B < 0) {
    stop("Argument B must be a single non-negative number.")
  }
  B <- as.integer(B)
  if (identical(confint, "bootstrap") && B < 1L) {
    stop("Argument B must be at least 1 when confint = 'bootstrap'.")
  }

  if (!is.null(seed) && (!is.numeric(seed) || length(seed) != 1L || !is.finite(seed))) {
    stop("Argument seed must be NULL or a single finite number.")
  }

  if (!is.data.frame(data)) {
    data <- data.frame(data)
  }

  singleRratioRegInternalCheckPrimaryFormula(formula)
  ratioFormula <- singleRratioRegInternalCheckRatioFormula(ratioFormula)

  fullDataRows <- rownames(data)
  subsetExpr <- if (missing(subset)) NULL else substitute(subset)
  modelFrameArgs <- list(
    formula = formula,
    data = data,
    ...
  )
  if (!is.null(subsetExpr)) {
    modelFrameArgs$subset <- subsetExpr
  }
  modelFrame <- do.call(
    stats::model.frame,
    modelFrameArgs,
    envir = parent.frame()
  )
  observed <- stats::model.response(modelFrame)

  if (NCOL(observed) > 1L) {
    stop("ratioReg() supports only a single response variable.")
  }

  observed <- as.numeric(observed)
  singleRratioRegInternalCheckObserved(observed)

  priorWeights <- singleRratioRegInternalAlignWeights(
    weights = weights,
    modelFrame = modelFrame,
    data = data,
    fullDataRows = fullDataRows
  )

  counts <- singleRratioRegInternalBuildCounts(
    observed = observed,
    weights = priorWeights,
    maxCount = maxCount
  )

  singleRratioRegInternalFitCounts(
    counts = counts,
    formula = formula,
    ratioFormula = ratioFormula,
    call = match.call(),
    requestedModel = model,
    criterion = criterion,
    primaryEstimator = estimator,
    confintType = confint,
    B = B,
    seed = seed
  )
}

singleRratioRegInternalCheckPrimaryFormula <- function(formula) {
  tt <- stats::terms(formula)
  if (attr(tt, "response") == 0L) {
    stop("Argument formula must include a response variable.")
  }

  if (length(attr(tt, "term.labels")) > 0L || attr(tt, "intercept") == 0L) {
    stop(
      "Argument formula must use '~ 1'. ",
      "Use ratioFormula for the ratio regression terms."
    )
  }

  invisible(formula)
}

singleRratioRegInternalCheckRatioFormula <- function(ratioFormula) {
  tt <- stats::terms(ratioFormula)
  if (attr(tt, "response") != 0L) {
    stop("Argument ratioFormula must be a one-sided formula.")
  }
  ratioFormula
}

singleRratioRegInternalCheckObserved <- function(observed) {
  if (any(!is.finite(observed))) {
    stop("Observed counts must be finite.")
  }
  if (any(observed <= 0)) {
    stop("Observed counts must be strictly positive.")
  }
  if (any(abs(observed - round(observed)) > sqrt(.Machine$double.eps))) {
    stop("Observed counts must be whole numbers.")
  }
  invisible(observed)
}

singleRratioRegInternalAlignWeights <- function(weights,
                                                modelFrame,
                                                data,
                                                fullDataRows) {
  if (is.null(weights)) {
    return(rep(1, nrow(modelFrame)))
  }

  weights <- as.numeric(weights)
  modelFrameRows <- rownames(modelFrame)
  dataRows <- rownames(data)

  if (length(weights) == 1L) {
    priorWeights <- rep(weights, nrow(modelFrame))
  } else if (length(weights) == nrow(modelFrame)) {
    priorWeights <- weights
  } else if (length(weights) == nrow(data)) {
    alignedRows <- match(modelFrameRows, dataRows)
    if (anyNA(alignedRows)) {
      stop("Unable to align argument weights with rows used in ratioReg().")
    }
    priorWeights <- weights[alignedRows]
  } else if (length(weights) == length(fullDataRows)) {
    alignedRows <- match(modelFrameRows, fullDataRows)
    if (anyNA(alignedRows)) {
      stop("Unable to align argument weights with rows used in ratioReg().")
    }
    priorWeights <- weights[alignedRows]
  } else {
    stop(
      "Argument weights must have length 1, the number of rows used in fitting, ",
      "the number of rows in data, or the number of rows in the original data."
    )
  }

  if (any(!is.finite(priorWeights)) || any(priorWeights < 0)) {
    stop("Argument weights must be finite and non-negative.")
  }

  priorWeights
}

singleRratioRegInternalBuildCounts <- function(observed, weights, maxCount = NULL) {
  counts <- stats::aggregate(
    x = list(weight = weights),
    by = list(observed = observed),
    FUN = sum
  )
  counts <- counts[order(counts$observed), , drop = FALSE]

  if (!is.null(maxCount)) {
    if (length(maxCount) != 1L || !is.finite(maxCount)) {
      stop("Argument maxCount must be a single finite number.")
    }
    maxCount <- as.integer(maxCount)
    if (maxCount < 2L) {
      stop("Argument maxCount must be at least 2.")
    }
    counts <- subset(counts, observed <= maxCount)
  }

  if (!nrow(counts)) {
    stop("No positive counts remained after applying maxCount.")
  }

  supportMax <- max(counts$observed)
  fullCounts <- rep(0, supportMax)
  names(fullCounts) <- as.character(seq_len(supportMax))
  fullCounts[as.character(counts$observed)] <- counts$weight
  if (sum(fullCounts) <= 0) {
    stop("Weighted observed frequencies must sum to a positive value.")
  }
  fullCounts
}

singleRratioRegInternalBuildRatioData <- function(counts) {
  if (length(counts) < 2L) {
    stop("ratioReg() requires support up to at least count 2.")
  }

  k <- seq_len(length(counts) - 1L)
  keep <- counts[k] > 0 & counts[k + 1L] > 0
  ratioData <- data.frame(
    k = k[keep],
    freq_k = as.numeric(counts[k[keep]]),
    freq_k1 = as.numeric(counts[k[keep] + 1L]),
    stringsAsFactors = FALSE
  )

  ratioData$ratio <- ratioData$freq_k1 / ratioData$freq_k
  ratioData$logRatio <- log(ratioData$ratio)
  ratioData$weight <- 1 / (1 / ratioData$freq_k + 1 / ratioData$freq_k1)
  ratioData$oneInflation <- as.numeric(ratioData$k == 1)
  ratioData
}

singleRratioRegInternalModelFormula <- function(ratioFormula,
                                                response = "logRatio",
                                                addOneInflation = FALSE) {
  tt <- stats::terms(ratioFormula)
  labels <- attr(tt, "term.labels")
  if (isTRUE(addOneInflation)) {
    labels <- c(labels, "oneInflation")
  }
  stats::reformulate(
    termlabels = labels,
    response = response,
    intercept = attr(tt, "intercept")
  )
}

singleRratioRegInternalFitModel <- function(ratioData, ratioFormula, modelName) {
  baseFormula <- singleRratioRegInternalModelFormula(ratioFormula)
  fitFormula <- singleRratioRegInternalModelFormula(
    ratioFormula = ratioFormula,
    addOneInflation = identical(modelName, "M1")
  )

  if (identical(modelName, "M1") && !any(ratioData$oneInflation == 1)) {
    return(list(
      model = modelName,
      available = FALSE,
      message = "Model M1 requires an observed ratio for k = 1."
    ))
  }

  design <- stats::model.matrix(fitFormula, data = ratioData)
  if (nrow(ratioData) < ncol(design)) {
    return(list(
      model = modelName,
      available = FALSE,
      message = "Not enough usable ratios to identify the requested ratio-regression model."
    ))
  }

  if (qr(design)$rank < ncol(design)) {
    return(list(
      model = modelName,
      available = FALSE,
      message = "The ratio regression design matrix is rank deficient."
    ))
  }

  fit <- stats::lm(
    formula = fitFormula,
    data = ratioData,
    weights = weight
  )

  coefficients <- stats::coef(fit)
  if (anyNA(coefficients)) {
    return(list(
      model = modelName,
      available = FALSE,
      message = "The requested model produced aliased coefficients."
    ))
  }

  nRatio <- nrow(ratioData)
  sse <- sum(ratioData$weight * stats::residuals(fit) ^ 2)
  sigma2 <- max(sse / nRatio, .Machine$double.eps)
  logL <- -0.5 * nRatio * (log(2 * pi) + 1 + log(sigma2)) +
    0.5 * sum(log(ratioData$weight))
  nPar <- fit$rank + 1L

  list(
    model = modelName,
    available = TRUE,
    fit = fit,
    formula = fitFormula,
    baseFormula = baseFormula,
    coefficients = coefficients,
    baseCoefficients = if (identical(modelName, "M1")) {
      coefficients[names(coefficients) != "oneInflation"]
    } else {
      coefficients
    },
    rank = fit$rank,
    sigma2 = sigma2,
    logL = logL,
    AIC = -2 * logL + 2 * nPar,
    BIC = -2 * logL + log(nRatio) * nPar,
    message = NULL
  )
}

singleRratioRegInternalPredictLinear <- function(fitInfo,
                                                 newdata,
                                                 component = c("full", "base")) {
  component <- match.arg(component)
  if (!isTRUE(fitInfo$available)) {
    stop("Cannot predict from an unavailable ratio-regression fit.")
  }

  if (identical(component, "full")) {
    stats::predict(fitInfo$fit, newdata = newdata)
  } else {
    X <- stats::model.matrix(
      stats::delete.response(stats::terms(fitInfo$baseFormula)),
      data = newdata
    )
    drop(X %*% fitInfo$baseCoefficients)
  }
}

singleRratioRegInternalLogSumExp <- function(x) {
  xmax <- max(x)
  xmax + log(sum(exp(x - xmax)))
}

singleRratioRegInternalBaseProbabilities <- function(fitInfo, supportMax) {
  ratioGrid <- data.frame(
    k = 0:(supportMax - 1L),
    oneInflation = as.numeric(0:(supportMax - 1L) == 1L)
  )
  logRatios <- singleRratioRegInternalPredictLinear(
    fitInfo = fitInfo,
    newdata = ratioGrid,
    component = "base"
  )
  logTerms <- c(0, cumsum(logRatios))
  logDenom <- singleRratioRegInternalLogSumExp(logTerms)
  probabilities <- exp(logTerms - logDenom)
  names(probabilities) <- as.character(0:supportMax)
  list(
    probabilities = probabilities,
    baseRatios = exp(logRatios)
  )
}

singleRratioRegInternalPlotX <- function(baseFormula, newdata) {
  X <- stats::model.matrix(
    stats::delete.response(stats::terms(baseFormula)),
    data = newdata
  )
  keep <- colnames(X) != "(Intercept)"
  if (sum(keep) == 1L) {
    x <- as.numeric(X[, keep])
    attr(x, "label") <- colnames(X)[keep]
    return(x)
  }

  x <- newdata$k
  attr(x, "label") <- "k"
  x
}

singleRratioRegInternalPopulationEstimates <- function(fitInfo, counts) {
  baseInfo <- singleRratioRegInternalBaseProbabilities(
    fitInfo = fitInfo,
    supportMax = length(counts)
  )
  probs <- baseInfo$probabilities
  p0 <- probs[1]
  p1 <- if (length(probs) >= 2L) probs[2] else 0
  f1 <- counts[1]
  f2 <- if (length(counts) >= 2L) counts[2] else 0
  sizeObserved <- sum(counts)

  hiddenHT <- if (identical(fitInfo$model, "M1")) {
    (sizeObserved - f1) * p0 / max(1 - p0 - p1, .Machine$double.eps)
  } else {
    sizeObserved * p0 / max(1 - p0, .Machine$double.eps)
  }

  hiddenSM <- f2 / max(prod(baseInfo$baseRatios[1:2]), .Machine$double.eps)

  list(
    HT = sizeObserved + hiddenHT,
    SM = sizeObserved + hiddenSM,
    baseProbabilities = probs,
    truncatedProbabilities = probs[-1] / max(1 - p0, .Machine$double.eps),
    baseRatios = baseInfo$baseRatios
  )
}

singleRratioRegInternalMakePopResult <- function(pointEstimate,
                                                 boot = NULL,
                                                 alpha = 0.05) {
  if (is.numeric(boot) && length(boot) > 1L) {
    quant <- stats::quantile(
      x = boot,
      probs = c(alpha / 2, 1 - alpha / 2),
      names = FALSE
    )
    variance <- stats::var(boot)
    confidenceInterval <- data.frame(
      lowerBound = quant[1],
      upperBound = quant[2]
    )
  } else {
    variance <- NA_real_
    confidenceInterval <- data.frame(
      lowerBound = NA_real_,
      upperBound = NA_real_
    )
    boot <- NULL
  }

  structure(
    list(
      pointEstimate = unname(pointEstimate),
      variance = variance,
      confidenceInterval = confidenceInterval,
      boot = boot,
      control = list(alpha = alpha)
    ),
    class = "popSizeEstResults"
  )
}

singleRratioRegInternalDrawSize <- function(x) {
  base <- floor(x)
  base + stats::rbinom(1, 1, x - base)
}

singleRratioRegInternalBootstrapSample <- function(object) {
  if (identical(object$requestedModel, "auto")) {
    probs <- object$counts / sum(object$counts)
    size <- singleRratioRegInternalDrawSize(sum(object$counts))
    sampled <- as.numeric(stats::rmultinom(1, size = size, prob = probs))
    names(sampled) <- names(object$counts)
    return(sampled)
  }

  pointEstimate <- object$populationSizeAll[[object$primaryEstimator]]$pointEstimate
  f0hat <- max(pointEstimate - sum(object$counts), 0)
  probs <- c(f0hat, object$counts)
  probs <- probs / sum(probs)
  size <- singleRratioRegInternalDrawSize(pointEstimate)
  sampled <- as.numeric(stats::rmultinom(1, size = size, prob = probs))
  counts <- sampled[-1]
  names(counts) <- names(object$counts)
  counts
}

singleRratioRegInternalBootstrap <- function(object, B, seed = NULL) {
  if (!is.null(seed)) {
    set.seed(seed)
  }

  bootValues <- matrix(
    NA_real_,
    nrow = 0,
    ncol = 2,
    dimnames = list(NULL, c("HT", "SM"))
  )
  attempts <- 0L
  maxAttempts <- max(B * 10L, B + 25L)

  while (nrow(bootValues) < B && attempts < maxAttempts) {
    attempts <- attempts + 1L
    sampledCounts <- singleRratioRegInternalBootstrapSample(object)
    bootFit <- tryCatch(
      singleRratioRegInternalFitCounts(
        counts = sampledCounts,
        formula = object$formula,
        ratioFormula = object$ratioFormula,
        call = object$call,
        requestedModel = object$requestedModel,
        criterion = object$criterion,
        primaryEstimator = object$primaryEstimator,
        confintType = "none",
        B = 0L,
        seed = NULL
      ),
      error = function(e) e
    )

    if (inherits(bootFit, "error")) {
      next
    }

    values <- c(
      HT = bootFit$populationSizeAll$HT$pointEstimate,
      SM = bootFit$populationSizeAll$SM$pointEstimate
    )

    if (all(is.finite(values))) {
      bootValues <- rbind(bootValues, values)
    }
  }

  if (nrow(bootValues) < B) {
    warning(
      "Only ", nrow(bootValues), " successful bootstrap replicates were obtained ",
      "out of the requested ", B, "."
    )
  }

  bootValues
}

singleRratioRegInternalCriteriaTable <- function(fits, selectedModel) {
  data.frame(
    model = c("M0", "M1"),
    available = c(fits$M0$available, fits$M1$available),
    logLik = c(
      if (fits$M0$available) fits$M0$logL else NA_real_,
      if (fits$M1$available) fits$M1$logL else NA_real_
    ),
    AIC = c(
      if (fits$M0$available) fits$M0$AIC else NA_real_,
      if (fits$M1$available) fits$M1$AIC else NA_real_
    ),
    BIC = c(
      if (fits$M0$available) fits$M0$BIC else NA_real_,
      if (fits$M1$available) fits$M1$BIC else NA_real_
    ),
    selected = c(selectedModel == "M0", selectedModel == "M1"),
    row.names = c("M0", "M1")
  )
}

singleRratioRegInternalFitCounts <- function(counts,
                                             formula,
                                             ratioFormula,
                                             call,
                                             requestedModel,
                                             criterion,
                                             primaryEstimator,
                                             confintType,
                                             B,
                                             seed = NULL) {
  ratioData <- singleRratioRegInternalBuildRatioData(counts)
  baseFormula <- singleRratioRegInternalModelFormula(ratioFormula)
  ratioData$xPlot <- singleRratioRegInternalPlotX(baseFormula, ratioData)

  fits <- list(
    M0 = singleRratioRegInternalFitModel(ratioData, ratioFormula, "M0"),
    M1 = singleRratioRegInternalFitModel(ratioData, ratioFormula, "M1")
  )

  selectedModel <- switch(
    requestedModel,
    M0 = {
      if (!fits$M0$available) {
        stop(fits$M0$message)
      }
      "M0"
    },
    M1 = {
      if (!fits$M1$available) {
        stop(fits$M1$message)
      }
      "M1"
    },
    auto = {
      available <- c(M0 = fits$M0$available, M1 = fits$M1$available)
      if (!any(available)) {
        stop("Neither M0 nor M1 could be fitted to the available ratios.")
      }
      if (sum(available) == 1L) {
        names(which(available))[1]
      } else {
        stats <- c(M0 = fits$M0[[criterion]], M1 = fits$M1[[criterion]])
        names(which.min(stats))[1]
      }
    }
  )

  selectedFit <- fits[[selectedModel]]
  popEstimates <- singleRratioRegInternalPopulationEstimates(selectedFit, counts)
  truncatedProbabilities <- popEstimates$truncatedProbabilities
  fittedFrequencies <- sum(counts) * truncatedProbabilities
  names(fittedFrequencies) <- names(counts)

  plotGrid <- data.frame(
    k = seq_len(length(counts) - 1L),
    oneInflation = as.numeric(seq_len(length(counts) - 1L) == 1L)
  )
  plotX <- singleRratioRegInternalPlotX(baseFormula, plotGrid)
  plotXLabel <- attr(plotX, "label")

  for (k in names(fits)) {
    if (!fits[[k]]$available) {
      next
    }

    fits[[k]]$plotGrid <- plotGrid
    fits[[k]]$plotX <- plotX
    fits[[k]]$plotPredictions <- list(
      full = singleRratioRegInternalPredictLinear(
        fitInfo = fits[[k]],
        newdata = plotGrid,
        component = "full"
      ),
      base = singleRratioRegInternalPredictLinear(
        fitInfo = fits[[k]],
        newdata = plotGrid,
        component = "base"
      )
    )
  }

  object <- structure(
    list(
      call = call,
      formula = formula,
      ratioFormula = ratioFormula,
      requestedModel = requestedModel,
      selectedModel = selectedModel,
      criterion = criterion,
      primaryEstimator = primaryEstimator,
      counts = counts,
      sizeObserved = sum(counts),
      supportMax = length(counts),
      ratioData = ratioData,
      fits = fits,
      criteria = singleRratioRegInternalCriteriaTable(fits, selectedModel),
      fittedProbabilities = popEstimates$baseProbabilities,
      fittedFrequencies = fittedFrequencies,
      plotInfo = list(
        xLabel = plotXLabel,
        xObserved = ratioData$xPlot,
        yObserved = ratioData$logRatio
      ),
      populationSize = NULL,
      populationSizeAll = list(
        HT = singleRratioRegInternalMakePopResult(pointEstimate = popEstimates$HT),
        SM = singleRratioRegInternalMakePopResult(pointEstimate = popEstimates$SM)
      ),
      bootstrap = NULL
    ),
    class = "singleRRatioReg"
  )

  bootstrap <- NULL
  if (identical(confintType, "bootstrap")) {
    bootstrap <- singleRratioRegInternalBootstrap(
      object = object,
      B = as.integer(B),
      seed = seed
    )
  }

  object$bootstrap <- bootstrap
  object$populationSizeAll <- list(
    HT = singleRratioRegInternalMakePopResult(
      pointEstimate = popEstimates$HT,
      boot = if (is.numeric(bootstrap)) bootstrap[, "HT"] else NULL
    ),
    SM = singleRratioRegInternalMakePopResult(
      pointEstimate = popEstimates$SM,
      boot = if (is.numeric(bootstrap)) bootstrap[, "SM"] else NULL
    )
  )
  object$populationSize <- object$populationSizeAll[[primaryEstimator]]
  object
}

#' @method coef singleRRatioReg
#' @exportS3Method
coef.singleRRatioReg <- function(object,
                                 model = c("selected", "M0", "M1"),
                                 component = c("full", "base"),
                                 ...) {
  model <- match.arg(model)
  component <- match.arg(component)
  if (identical(model, "selected")) {
    model <- object$selectedModel
  }

  fitInfo <- object$fits[[model]]
  if (!isTRUE(fitInfo$available)) {
    stop("Requested model is not available in this ratioReg() fit.")
  }

  switch(
    component,
    full = fitInfo$coefficients,
    base = fitInfo$baseCoefficients
  )
}

#' @method logLik singleRRatioReg
#' @importFrom stats logLik
#' @exportS3Method
logLik.singleRRatioReg <- function(object,
                                   model = c("selected", "M0", "M1"),
                                   ...) {
  model <- match.arg(model)
  if (identical(model, "selected")) {
    model <- object$selectedModel
  }

  fitInfo <- object$fits[[model]]
  if (!isTRUE(fitInfo$available)) {
    stop("Requested model is not available in this ratioReg() fit.")
  }

  val <- fitInfo$logL
  attr(val, "nobs") <- nrow(object$ratioData)
  attr(val, "df") <- fitInfo$rank + 1L
  class(val) <- "logLik"
  val
}

#' @method extractAIC singleRRatioReg
#' @importFrom stats extractAIC
#' @exportS3Method
extractAIC.singleRRatioReg <- function(fit,
                                       scale,
                                       k = 2,
                                       model = c("selected", "M0", "M1"),
                                       ...) {
  model <- match.arg(model)
  if (identical(model, "selected")) {
    model <- fit$selectedModel
  }

  fitInfo <- fit$fits[[model]]
  if (!isTRUE(fitInfo$available)) {
    stop("Requested model is not available in this ratioReg() fit.")
  }

  -2 * fitInfo$logL + k * (fitInfo$rank + 1L)
}

#' @method nobs singleRRatioReg
#' @importFrom stats nobs
#' @exportS3Method
nobs.singleRRatioReg <- function(object, ...) {
  nrow(object$ratioData)
}

#' @method popSizeEst singleRRatioReg
#' @rdname popSizeEst
#' @exportS3Method
popSizeEst.singleRRatioReg <- function(object,
                                       estimator = c("primary", "HT", "SM"),
                                       ...) {
  estimator <- match.arg(estimator)
  if (identical(estimator, "primary")) {
    estimator <- object$primaryEstimator
  }
  object$populationSizeAll[[estimator]]
}

#' @method confint singleRRatioReg
#' @importFrom stats confint
#' @exportS3Method
confint.singleRRatioReg <- function(object,
                                    parm,
                                    level = 0.95,
                                    estimator = c("primary", "HT", "SM"),
                                    ...) {
  if (!missing(level) && (length(level) != 1L || !is.finite(level))) {
    stop("Argument level must be a single finite number.")
  }

  estimator <- match.arg(estimator)
  if (identical(estimator, "primary")) {
    estimator <- object$primaryEstimator
  }

  boot <- object$populationSizeAll[[estimator]]$boot
  if (is.numeric(boot) && length(boot) > 1L) {
    alpha <- 1 - level
    return(
      matrix(
        stats::quantile(
          x = boot,
          probs = c(alpha / 2, 1 - alpha / 2),
          names = FALSE
        ),
        nrow = 1,
        dimnames = list(NULL, c("lowerBound", "upperBound"))
      )
    )
  }

  as.matrix(object$populationSizeAll[[estimator]]$confidenceInterval)
}

#' @method print singleRRatioReg
#' @exportS3Method
print.singleRRatioReg <- function(x, ...) {
  cat("Call: ")
  print(x$call)
  cat(
    "\nSelected ratio regression model:", x$selectedModel,
    "\nSelection criterion:", x$criterion,
    "\nPrimary population estimator:", x$primaryEstimator,
    "\nObserved population size:", signif(x$sizeObserved),
    "\n"
  )
  cat("\nCoefficients:\n")
  print(stats::coef(x))
  cat("\nPopulation size estimation results:\n")
  print(popSizeEst(x))

  alternative <- setdiff(c("HT", "SM"), x$primaryEstimator)
  cat("\nAlternative estimator point estimate:", alternative, "=",
      signif(popSizeEst(x, estimator = alternative)$pointEstimate), "\n")

  invisible(x)
}

#' @title Summary Statistics for Ratio Regression Fits
#'
#' @description
#' Summarizes a fitted `singleRRatioReg` object, including model-selection
#' criteria and population-size estimates.
#'
#' @param object object of class `singleRRatioReg`.
#' @param ... currently not used.
#'
#' @return An object of class `summarysingleRRatioReg`.
#' @method summary singleRRatioReg
#' @exportS3Method
summary.singleRRatioReg <- function(object, ...) {
  estimates <- data.frame(
    estimator = c("HT", "SM"),
    pointEstimate = c(
      object$populationSizeAll$HT$pointEstimate,
      object$populationSizeAll$SM$pointEstimate
    ),
    variance = c(
      object$populationSizeAll$HT$variance,
      object$populationSizeAll$SM$variance
    ),
    lowerBound = c(
      object$populationSizeAll$HT$confidenceInterval$lowerBound,
      object$populationSizeAll$SM$confidenceInterval$lowerBound
    ),
    upperBound = c(
      object$populationSizeAll$HT$confidenceInterval$upperBound,
      object$populationSizeAll$SM$confidenceInterval$upperBound
    )
  )

  structure(
    list(
      call = object$call,
      selectedModel = object$selectedModel,
      criterion = object$criterion,
      primaryEstimator = object$primaryEstimator,
      sizeObserved = object$sizeObserved,
      coefficients = lapply(
        object$fits[c("M0", "M1")],
        function(x) if (isTRUE(x$available)) x$coefficients else NULL
      ),
      criteria = object$criteria,
      estimates = estimates
    ),
    class = "summarysingleRRatioReg"
  )
}

#' @method print summarysingleRRatioReg
#' @exportS3Method
print.summarysingleRRatioReg <- function(x, digits = max(3L, getOption("digits") - 3L), ...) {
  cat("\nCall:\n")
  print(x$call)
  cat(
    "\nSelected ratio regression model:", x$selectedModel,
    "\nSelection criterion:", x$criterion,
    "\nPrimary population estimator:", x$primaryEstimator,
    "\nObserved population size:", signif(x$sizeObserved, digits),
    "\n\nModel selection table:\n",
    sep = " "
  )
  print(x$criteria, digits = digits)

  for (nm in names(x$coefficients)) {
    cat("\nCoefficients for", nm, ":\n")
    if (is.null(x$coefficients[[nm]])) {
      cat("Model not available.\n")
    } else {
      print(round(x$coefficients[[nm]], digits = digits))
    }
  }

  cat("\nPopulation size estimates:\n")
  toPrint <- x$estimates
  isNum <- vapply(toPrint, is.numeric, logical(1))
  toPrint[isNum] <- lapply(toPrint[isNum], round, digits = digits)
  print(toPrint, row.names = FALSE)
  invisible(x)
}

#' @title Diagnostic Plots for Ratio Regression Fits
#'
#' @description
#' Plot methods for `singleRRatioReg` objects.
#'
#' @param x object of class `singleRRatioReg`.
#' @param plotType one of `"ratio"`, `"bootHist"`, or `"freqFit"`.
#' @param ... additional arguments passed to the underlying graphics calls.
#'
#' @return No return value, called for side effects.
#' @method plot singleRRatioReg
#' @export
plot.singleRRatioReg <- function(x,
                                 plotType = c("ratio", "bootHist", "freqFit"),
                                 ...) {
  plotType <- match.arg(plotType)

  switch(
    plotType,
    ratio = {
      graphics::plot(
        x = x$plotInfo$xObserved,
        y = x$plotInfo$yObserved,
        xlab = x$plotInfo$xLabel,
        ylab = "log(f[k+1] / f[k])",
        main = "Ratio regression plot",
        pch = 19,
        ...
      )

      if (isTRUE(x$fits$M0$available)) {
        graphics::lines(
          x = x$fits$M0$plotX,
          y = x$fits$M0$plotPredictions$full,
          col = 2,
          lwd = if (x$selectedModel == "M0") 2.5 else 1.5,
          lty = 2
        )
      }

      if (isTRUE(x$fits$M1$available)) {
        graphics::lines(
          x = x$fits$M1$plotX,
          y = x$fits$M1$plotPredictions$full,
          col = 4,
          lwd = if (x$selectedModel == "M1") 2.5 else 1.5,
          lty = 1
        )
      }

      graphics::legend(
        "topright",
        legend = c("Observed", "M0", "M1"),
        col = c(1, 2, 4),
        pch = c(19, NA, NA),
        lty = c(NA, 2, 1),
        bty = "n"
      )
    },
    bootHist = {
      boot <- popSizeEst(x)$boot
      if (!is.numeric(boot)) {
        stop("No bootstrap results were stored for the primary estimator.")
      }

      graphics::hist(
        boot,
        main = "Bootstrap population size estimates",
        xlab = expression(hat(N)),
        ...
      )
    },
    freqFit = {
      xx <- seq_len(length(x$counts))
      graphics::matplot(
        x = xx,
        y = cbind(observed = x$counts, fitted = x$fittedFrequencies),
        type = "o",
        pch = c(19, 17),
        lty = c(1, 2),
        col = c(1, 2),
        xlab = "Counts",
        ylab = "Frequency",
        main = "Observed and fitted frequencies",
        ...
      )
      graphics::legend(
        "topright",
        legend = c("Observed", "Fitted"),
        col = c(1, 2),
        pch = c(19, 17),
        lty = c(1, 2),
        bty = "n"
      )
    }
  )

  invisible(x)
}

if ( requireNamespace("tinytest", quietly=TRUE) ){
  xx <- isTRUE(tolower(Sys.getenv("TEST_SINGLERCAPTURE_MULTICORE_DEVELOPER", unset = "FALSE")) == "true")
  tinytest::test_package("singleRcapture", ncpu = if (xx) 2 else NULL)
}

---
title: "singleRcapture: An R Package for Single-Source Capture-Recapture Models"
author: "Piotr Chlebicki, Maciej Beręsewicz"
date: "`r Sys.Date()`"
output: 
    html_vignette:
        df_print: kable
        toc: true
        number_sections: true
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{singleRcapture: An R Package for Single-Source Capture-Recapture Models}
  %\VignetteEngine{knitr::rmarkdown_notangle}
bibliography: bibliography.bib
link-citations: true
editor_options: 
  chunk_output_type: console
header-includes:
  - \usepackage{amsmath}
  - \usepackage{amsthm}
  - \usepackage{amssymb}
  - \usepackage{calc}
  - \usepackage{ragged2e}
  - \usepackage[ruled]{algorithm2e}
  - \DeclareMathAlphabet{\mathmybb}{U}{bbold}{m}{n}
  - \newcommand{\1}{\mathcal{I}}
  - \newcommand{\bZero}{\boldsymbol{0}}
  - \newcommand{\bx}{\boldsymbol{x}}
  - \newcommand{\bX}{\boldsymbol{X}}
  - \newcommand{\bbeta}{\boldsymbol{\beta}}
  - \newcommand{\boeta}{\boldsymbol{\eta}}
  - \newcommand{\bW}{\boldsymbol{W}}
  - \newcommand{\bo}{\boldsymbol{o}}  
---

```{=html}
<style>
pre {
  overflow-x: auto;
}
pre code {
  word-wrap: normal;
  white-space: pre;
}
</style>
```

```{r plot-parameters, include=FALSE}
par(mar = c(2.5, 8.5, 4.1, 2.5), cex.main = .7, cex.lab = .6)
```

# Introduction {#sec-introduction}

Population size estimation is a methodological approach employed across
multiple scientific disciplines, which serves as the basis for research,
policy formulation, and decision-making processes [cf.
@bohning2018capture]. In the field of statistics, particularly official
statistics, precise population estimates are essential in order to
develop robust economic models, optimize resource allocation, and inform
evidence-based policy [cf. @baffour2013modern]. Social scientists
utilize advanced population estimation techniques to investigate
*hard-to-reach* populations, such as homeless individuals or illicit
drug users in an effort to overcome the inherent limitations of
conventional census methodologies. These techniques are crucial for
obtaining accurate data on populations that are typically
under-represented or difficult to access using traditional sampling
methods [cf. @vincent2022estimating]. In ecology and epidemiology,
researchers focus on estimating the size of individual species or
disease-affected populations within defined geographical regions as part
of conservation efforts, ecosystem management, and public health
interventions.

Population size estimation can be approached using various
methodologies, each with distinct advantages and limitations.
Traditional approaches include full enumeration (e.g. censuses) and
comprehensive sample surveys, which, while providing detailed data, are
often resource-intensive and may result in delayed estimates,
particularly for human populations. Alternative methods leverage
existing data sources, such as administrative registers or carefully
designed small-scale studies in wildlife research, or census coverage
surveys [cf. @wolter1986some; @zhang2019note]. Information from these
sources is often extracted by applying statistical methods, known as
*capture-recapture* or *multiple system estimation*, which rely on data
from multiple enumerations of the same population [cf.
@dunne2024system]. This approach can be implemented using either a
single source with repeated observations, two sources, or multiple
sources.

In this paper we focus on methods that involve a single data source with
multiple enumerations of the same units [cf. @ztpoisson]. In human
population studies, such data can be derived from police records, health
system databases, or border control logs; in the case of non-human
populations, data of this kind can come from veterinary records or
specialized field data. These methods are often applied to estimate
hard-to-reach or hidden populations, where standard sampling methods may
be inappropriate because of prohibitive costs or problems with
identifying population members.

While methods for two or more sources are implemented in various
open-source software packages, for instance **CARE-4** [@yang2006care4]
(in **GAUSS**), **Rcapture** [@baillargeon2007rcapture], **marked**
[@laake2013marked] or **VGAM** [@yee2015vgam] (all in **R**),
single-source capture-recapture (SSCR) methods are either not available
at all or are only partially implemented in existing **R** packages or
other software. Therefore, the paper attempts to bridge this gap by
introducing the **singleRcapture** package, which implement
*state-of-the-art* SSCR methods and offer a user friendly API resembling
existing **R** functions (e.g., `glm`). In the next subsection we
describe existing **R** packages and other software that could be used
for estimating population size using SSCR methods.

## Software for capture-recapture with a single source {#sec-software}

The majority of SSCR methods assume zero-truncated distributions or
their extensions (e.g., inclusion of one-inflation). The
**Distributions.jl** [@Distributionsjl] (in **Julia**), **StatsModels**
[@seabold2010statsmodels] (in **Python**), **countreg** [@countreg],
**VGAM** [@VGAM-main] or **distributions3** [@distributions3] (in **R**)
implement some of those truncated distributions (e.g.
`distributions3::ZTPoisson` or `countreg::zerotrunc`) and the most
general distributions, such as Generally Altered, Inflated, Truncated
and Deflated, can be found in the **VGAM** package (e.g.
`VGAM::gaitdpoisson` for the Poisson distribution, see @gaitdcount for a
detailed description). However, the estimation of parameters of a given
truncated (and possibly inflated) distribution is just the first step
(as in the case of log-linear models in capture-recapture with two
sources) and, to the best of our knowledge, there is no open-source
software that can be used to estimate population size using on SSCR
methods and includes variance estimators or diagnostics.

Therefore, the purpose of the **singleRcapture**, an **R** package, is
to bridge this gap by providing scientists and other practitioners with
a tool for estimating population size with SSCR methods. We have
implemented state-of-the-art methods, as recently described by
@bohning2018capture or @bohning2024one and provided their extensions
(e.g., inclusion of covariates, different treatment of one-inflation),
which will be covered in detail in Section 2. The package implements
variance estimation based on various methods, can be used to create
custom models and diagnostic plots (e.g. rootograms) with parameters
estimated using a modified iteratively reweighted least squares (IRLS)
algorithm we have implemented for estimation stability. To the best of
our knowledge, no existing open-source package allows the estimation of
population size by selecting an appropriate SSCR model, conducting the
estimation, and providing informative diagnostics of the results.

The remaining part of the paper is structured as follows. Section 2
contains a brief description of the theoretical background and
information about fitting methods and available methods of variance
estimation. Section 3 introduces the main functions of the package.
Section 4 presents a case study and contains an assessment of its
results, diagnostics and estimates of specific sub-populations. Section
5 describes classes and `S3methods` implemented in the package. The
paper ends with conclusions and an appendix showing how to use the
`estimatePopsizeFit` function which is aimed to mimic the `glm.fit` or
similar functions. In replication materials we show how to implement a
custom model as this option could be of interest to users wishing to
apply any new bootstrap methods not implemented in the package.

# Theoretical background {#sec-theory}

## How to estimate population size with a single source?

Let $Y_{k}$ represent the number of times the $k$-th unit was observed
in a single source (e.g. register). Clearly, we only observe $k:Y_{k}>0$
and do not know how many units have been missed (i.e. $Y_{k}=0$), so the
population size, denoted by $N$, needs to be estimated. In general, we
assume that the conditional distribution of $Y_{k}$ given a vector of
covariates $\bx_{k}$ follows a version of the zero-truncated count data
distribution (and its extensions). When we know the parameters of the
distribution we can estimate the population size using a
Horvitz-Thompson type estimator given by:

$$
\begin{equation}
\hat{N}=
\sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}[Y_{k}>0|\bX_{k}]}=
\sum_{k=1}^{N_{obs}}\frac{1}{\mathbb{P}[Y_{k}>0|\bX_{k}]},
\end{equation}
$$

where $I_{k}:=\1_{\mathbb{N}}(Y_{k})$, $N_{obs}$ is the number of
observed units and $\1$ is the indicator function, while the maximum
likelihood estimate of $N$ is obtained after substituting regression
parameters $\bbeta$ for $\mathbb{P}[Y_{k}>0|\bx_{k}]$ in the above
equation.

The basic SSCR assumes independence between counts, which is a rather
naive assumption, since the first capture may significantly influence
the behavior of a given unit or limit the possibility of subsequent
captures (e.g. due to incarceration).

To solve these issues, @godwin2017estimation and @ztoi-oizt-poisson
introduced one-inflated distributions, which explicitly model the
probability of singletons by giving additional mass $\omega$ to
singletons denoted as $\1_{\{1\}}(y)$ [cf. @bohning2024one]. In other
words they considered a new random variable $Y^{\ast}$ that corresponds
to the data collection process which exhibits one inflation:

$$
\begin{equation*}
  \mathbb{P}\left[Y^{\ast}=y|Y^{\ast}>0\right] =
  \omega\1_{\{1\}}(y)+(1-\omega)\mathbb{P}[Y=y|Y>0].
\end{equation*}
$$

Analytic variance estimation is then performed by computing two parts of
the decomposition according to the law of total variance given by:

$$
\begin{equation}\label{eq-law_of_total_variance_decomposition}
  \text{var}[\hat{N}] = \mathbb{E}\left[\text{var}
  \left[\hat{N}|I_{1},\dots,I_{n}\right]\right] + 
  \text{var}\left[\mathbb{E}[\hat{N}|I_{1},\dots,I_{n}]\right],
\end{equation}
$$

where the first part can be estimated using the multivariate $\delta$
method given by:

$$
\begin{equation*}
  \mathbb{E}\left[\text{var} \left[\hat{N}|I_{1},\dots,I_{n}\right]\right] =
  \left.\left(\frac{\partial(N|I_1,\dots,I_N)}{\partial\bbeta}\right)^\top
  \text{cov}\left[\hat{\bbeta}\right]
  \left(\frac{\partial(N|I_1,\dots,I_N)}{\partial\bbeta}\right)
  \right|_{\bbeta=\hat{\bbeta}},
\end{equation*}
$$

while the second part of the decomposition above, assuming independence
of $I_{k}$'s and after some omitted simplifications, is optimally
estimated by:

$$
\begin{equation*}
  \text{var}\left[\mathbb{E}(\hat{N}|I_{1},\dots,I_{n})\right] =
  \text{var}\left[\sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}(Y_{k}>0)}\right]
  \approx\sum_{k=1}^{N_{obs}}\frac{1-\mathbb{P}(Y_{k}>0)}{\mathbb{P}(Y_{k}>0)^{2}},
\end{equation*}
$$

which serves as the basis for interval estimation. Confidence intervals
are usually constructed under the assumption of (asymptotic) normality
of $\hat{N}$ or asymptotic normality of $\ln(\hat{N}-N)$ (or log
normality of $\hat{N}$). The latter is an attempt to address a common
criticism of student type confidence intervals in SSCR, namely a
possibly skewed distribution of $\hat{N}$, and results in the $1-\alpha$
confidence interval given by:

$$
\begin{equation*}
  \left(N_{obs}+\frac{\hat{N}-N_{obs}}{\xi},N_{obs} +
  \left(\hat{N}-N_{obs}\right)\xi\right),
\end{equation*}
$$

where:

$$
\begin{equation*}
  \xi = \exp\left(z\left(1-\frac{\alpha}{2}\right)
  \sqrt{\ln\left(1+\frac{\widehat{\text{Var}}(\hat{N})}{\left(\hat{N}-N_{obs}\right)^{2}}\right)}\right),
\end{equation*}
$$

and where $z$ is the quantile function of the standard normal
distribution. The estimator $\hat{N}$ is best interpreted as being an
estimator of the total number of \underline{observable} units in the
population, since we have no means of estimating the number of units in
the population for which the probability of being included in the data
is $0$ [@ztpoisson].

## Available models

The full list of models implemented in **singleRcapture** along with
corresponding expressions for probability density functions and point
estimates can be found in the collective help file for all family
functions:

```{r, eval=FALSE}
?ztpoisson
```

For the sake of simplicity, we only list the family functions together
with brief descriptions. For more detailed information, please consult
the relevant literature.

The current list of these family functions includes:

-   Generalized Chao's [@chao1987estimating] and Zelterman's
    [@zelterman1988robust] estimators via logistic regression on
    variable $Z$ defined as $Z=1$ if $Y=2$ and $Z=0$ if $Y=1$ with
    $Z\sim b(p)$ where $b(\cdot)$ is the Bernoulli distribution and $p$
    can be modeled for each unit $k$ by
    $\text{logit}(p_k)=\ln(\lambda_k/2)$ with Poisson parameter
    $\lambda_k=\bx_{k}\bbeta$ (for a covariate extension see
    @chao-generalization and @zelterman):

$$
\hat{N}_{\text{Chao}} = N_{obs}+
\sum_{k=1}^{\boldsymbol{f}_{1}+\boldsymbol{f}_{2}}
\left(2\exp\left(\bx_{k}\hat{\bbeta}\right)+
2\exp\left(2\bx_{k}\hat{\bbeta}\right)\right)^{-1},
\tag{Chao's estimator}
$$

$$
\hat{N}_{\text{Zelt}}=\sum_{k=1}^{N_{obs}}
\left(1-\exp\left(-2\exp\left(\bx_{k}\hat{\bbeta}\right)\right)\right)^{-1}.
\tag{Zelterman's estimator}
$$

where $\boldsymbol{f}_{1}$ and $\boldsymbol{f}_{2}$ denotes number of
units observed once and twice.

-   Zero-truncated (`zt`$^\ast$) and zero-one-truncated (`zot`$^\ast$)
    Poisson [@zotmodels], geometric, NB type II (NB2) regression, where
    the non-truncated distribution is parameterized as:

$$
\mathbb{P}[Y=y|\lambda,\alpha] = \frac{\Gamma\left(y+\alpha^{-1}\right)}{\Gamma\left(\alpha^{-1}\right)y!}
\left(\frac{\alpha^{-1}}{\alpha^{-1}+\lambda}\right)^{\alpha^{-1}}
\left(\frac{\lambda}{\lambda + \alpha^{-1}}\right)^{y}.
$$

-   Zero-truncated one-inflated (`ztoi`$^\ast$) modifications, where the
    count data variable $Y^{\ast}$ is defined as:

$$
\mathbb{P}\left[Y^{\ast}=y\right] = \omega \1_{\{1\}}(y)+(1-\omega)\mathbb{P}[Y=y],
$$

$$
\mathbb{P}\left[Y^{\ast}=y|Y^{\ast}>0\right] = 
\omega\frac{\1_{\{1\}}(y)}{1-(1-\omega)\mathbb{P}[Y=0]}+
(1-\omega)\frac{\mathbb{P}[Y=y]}{1-(1-\omega)\mathbb{P}[Y=0]}.
$$

-   One-inflated zero-truncated (`oizt`$^\ast$) modifications, where the
    count data variable $Y^{\ast}$ is defined such that its distribution
    statisfies:

$$
\mathbb{P}\left[Y^{\ast}=y\right]=
\begin{cases}
\mathbb{P}[Y=0] & y=0, \\
\omega\left(1-\mathbb{P}[Y=0]\right)+(1-\omega)\mathbb{P}[Y=1] & y=1, \\
(1-\omega)\mathbb{P}[Y=y] & y>1,
\end{cases}
$$

$$
\mathbb{P}\left[Y^{\ast}=y|Y^{\ast}>0\right]=\omega\1_{\{1\}}(y)+(1-\omega)\mathbb{P}[Y=y|Y>0].
$$

Note that `ztoi`$^\ast$ and `oizt`$^\ast$ distributions are equivalent,
in the sense that the maximum value of the likelihood function is equal
for both of those distributions given any data, as shown by
[@bohning2023equivalence] but population size estimators are different.

In addition, we propose two new approaches to model singletons in a
similar way as in hurdle models. In particular, we have proposed the
following:

-   The zero-truncated hurdle model (`ztHurdle`\*) for Poisson,
    geometric and NB2 is defined as:

$$
\mathbb{P}[Y^{\ast}=y]=\begin{cases}
\frac{\mathbb{P}[Y=0]}{1-\mathbb{P}[Y=1]} & y=0, \\
\pi(1-\mathbb{P}[Y=1]) & y=1, \\
(1-\pi) \frac{\mathbb{P}[Y=y]}{1-\mathbb{P}[Y=1]} & y>1,
\end{cases}
$$

$$
\mathbb{P}[Y^{\ast}=y|Y^{\ast}>0]=\pi\mathbf{1}_{\{1\}}(y)+
(1-\pi)\mathbf{1}_{\mathbb{N}\setminus\{1\}}(y)\frac{\mathbb{P}[Y=y]}{1-\mathbb{P}[Y=0]-\mathbb{P}[Y=1]}.
$$

where $\pi$ denotes the conditional probability of observing singletons.

-   The hurdle zero-truncated model (`Hurdlezt`\*) for Poisson,
    geometric and NB2 is defined as:

$$
\mathbb{P}[Y^{\ast}=y]=\begin{cases}
\pi & y=1, \\
(1-\pi) \frac{\mathbb{P}[Y=y]}{1-\mathbb{P}[Y=1]} & y\neq1,
\end{cases}
$$

$$
\mathbb{P}[Y^{\ast}=y|Y^{\ast}>0]=\begin{cases}
\pi\frac{1-\mathbb{P}[Y=1]}{1-\mathbb{P}[Y=0]-\mathbb{P}[Y=1]} & y=1,\\
(1-\pi)\frac{\mathbb{P}[Y=y]}{1-\mathbb{P}[Y=0]-\mathbb{P}[Y=1]} & y>1,
\end{cases}
$$

where $\pi$ denotes the unconditional probability of observing
singletons.

-   A modified Chao estimator robust to one-inflation (`oichao`), which
    discards singletons and runs logistic regression on the indicator
    $Z=1$ if $Y=3$ and $Z=0$ if $Y=2$ with
    $\text{logit}(p_k)=\ln(\lambda_k/3)$, so that
    $p_k=\lambda_k/(3+\lambda_k)$ matches
    $\mathbb{P}[Y=3\mid Y\in\{2,3\}]$ for $Y\sim\mathrm{Poisson}(\lambda_k)$.
    The associated population-size estimator is

$$
\hat{N}_{\text{oiChao}} = N_{obs}+
\sum_{k=1}^{\boldsymbol{f}_{2}+\boldsymbol{f}_{3}}
\left(\tfrac{9}{2}\exp\left(2\bx_{k}\hat{\bbeta}\right)+
\tfrac{9}{2}\exp\left(3\bx_{k}\hat{\bbeta}\right)\right)^{-1},
\tag{oiChao estimator}
$$

where $\boldsymbol{f}_{2}$ and $\boldsymbol{f}_{3}$ denote the number of
units observed twice and three times. The inflation factor
$(\lambda_k^{2}/2+\lambda_k^{3}/6)^{-1}$ equals
$\mathbb{P}[Y=0]/\mathbb{P}[Y\in\{2,3\}]$ under $\mathrm{Poisson}(\lambda_k)$,
so the estimator is unaffected by extra mass at $Y=1$.

The approaches presented above differ in their assumptions,
computational complexity, or in the way they treat heterogeneity of
captures and singletons. For instance, the dispersion parameter $\alpha$
in the NB2 type models is often interpreted as measuring the `severity`
of unobserved heterogeneity in the underlying Poisson process
[@ztnegbin]. When using any truncated NB model, the hope is that given
the class of models considered, the consistency is not lost despite the
lack of information.

While not discussed in the literature, the interpretation of
heterogeneous $\alpha$ across the population (specified in
`controlModel`) would be that the unobserved heterogeneity affects the
accuracy of the prediction for the dependent variable $Y$ more severely
than others. The geometric model (NB with $\alpha=1$) is singled out in
the package and often considered in the literature because of inherent
computational issues with NB models, which are exacerbated by the fact
that data used for SSCR are usually of rather low quality. Data sparsity
is a particularly common problem in SSCR and a big challenge for all
numerical methods for fitting the (zero-truncated) NB model.

The extra mass $\omega$ in one-inflated models is an important extension
to the researcher's toolbox for SSCR models, since the inflation at
$y=1$ is likely to occur in many types of applications. For example,
when estimating the number of active people who committed criminal acts
in a given time period, the fact of being captured for the first time
following an arrest is associated with the risk of no longer being able
to be captured a second time. One constraint present in modelling via
inflated models is that attempts to include both the possibility of one
inflation and one deflation lead to both numerical and inferential
problems since the parameter space (of $(\omega, \lambda)$ or
$(\omega, \lambda, \alpha)$) is then given by
$\{(\omega, \lambda, \alpha) | \forall x\in \mathbb{N}: p(x|\omega, \lambda, \alpha)\geq0\}$
for the probability mass function $p$. The boundary of this set is then
a $1$ or $2-$dimensional manifold, transforming this parameter space
into $\mathbb{R}^{3}$ would require using `link` functions that depend
on more than one parameter.

Hurdle models represent another approach to modelling one-inflation.
They can also model deflation as well as inflation and deflation
simultaneously, so they are more flexible and, in the case of hurdle
zero-truncated models, appear to be more numerically stable.

Although the question of how to interpret regression parameters tends to
be somewhat overlooked in SSCR studies, we should point out that the
interpretation of the $\omega$ inflation parameter (in `ztoi`$^\ast$ or
`oizt`$^\ast$) is more convenient than the interpretation of the $\pi$
probability parameter (in hurdle models). Additionally, the
interpretation of the $\lambda$ parameter in (one) inflated models
conforms to the following intuition: given that unit $k$ comes from the
non-inflated part of the population, it follows a Poisson distribution
(respectively geometric or negative binomial) with the $\lambda$
parameter (or $\lambda,\alpha$); no such interpretation exists for
hurdle models. Interestingly, estimates from hurdle zero-truncated and
zero-truncated one-inflated models tend to be quite close to one
another, although more rigorous studies are required to confirm this
observation.

## Fitting method

As previously noted, the **singleRcapture** package can be used to model
the (linear) dependence of all parameters on covariates. A modified IRLS
algorithm is employed for this purpose as presented in Algorithm 1; full
details are available in @VGAM-main. In order to apply the algorithm, a
modified model matrix $\bX_{\text{vlm}}$ is created when the
`estimatePopsize` function is called. In the context of the models
implemented in **singleRcapture**, this matrix can be written as:

$$
\begin{equation}\label{X_vlm-definition}
  \bX_{\text{vlm}}=
  \begin{pmatrix}
    \bX_{1} & \boldsymbol{0} &\dotso &\boldsymbol{0}\\
    \boldsymbol{0} & \bX_{2} &\dotso &\boldsymbol{0}\\
    \vdots & \vdots & \ddots & \vdots\\
    \boldsymbol{0} & \boldsymbol{0} &\dotso &\bX_{p}
  \end{pmatrix}
\end{equation}
$$

where each $\bX_{i}$ corresponds to a model matrix associated with a
user specified formula.

### Algorithm 1: The modified IRLS algorithm used in the **singleRcapture** package

1.  Initialize with `iter` ← 1, $\boeta$ ← `start`, $\bW$ ← I, $\ell$ ←
    $\ell(\bbeta)$

2.  Store values from the previous step:\
    $\ell_{-}$ ← $\ell$, $\bW_{-}$ ← $\bW$, $\bbeta_{-}$ ← $\bbeta$ (the
    last assignment is omitted during the first iteration), and assign
    values in the current iteration:

    $\displaystyle\boeta$ ← $\bX_{\text{vlm}}\bbeta+\bo$

    $\bW_{(k)}$ ←
    $\mathbb{E}\left[-\frac{\partial^{2}\ell}{\partial\boeta_{(k)}^\top\partial\boeta_{(k)}}\right]$

    $\boldsymbol{Z}_{(k)}$ ←
    $\boeta_{(k)}+\frac{\partial\ell}{\partial\boeta_{(k)}}\bW_{(k)}^{-1}-\bo_{(k)}$

    where $\bo$ denotes offset.

3.  Assign the current coefficient value:\
    $\bbeta$ ←
    $\left(\bX_{\text{vlm}}\bW\bX_{\text{vlm}}\right)^{-1}\bX_{\text{vlm}}\bW\boldsymbol{Z}$

4.  If $\ell(\bbeta)<\ell(\bbeta_{-})$ try selecting the smallest value
    $h$ such that for $\bbeta_{h}$ ←
    $2^{-h}\left(\bbeta+\bbeta_{-}\right)$ the inequality
    $\ell(\bbeta_{h})>\ell(\bbeta_{-})$ holds if this is successful
    $\bbeta$ ← $\bbeta_{h}$, else stop the algorithm.

5.  If convergence is achieved or `iter` is higher than `maxiter`, stop
    the algorithm, else `iter` ← 1 + `iter` and return to step 2.

In the case of multi-parameter families, we get a matrix of linear
predictors $\boeta$ instead of a vector, with the number of columns
matching the number of parameters in the distribution. "Weights" (matrix
$\bW$) are then modified to be information matrices
$\displaystyle\mathbb{E}\left[-\frac{\partial^{2}\ell}{\partial\boeta_{(k)}^\top\partial\boeta_{(k)}}\right]$,
where $\ell$ is the log-likelihood function and $\boeta_{(k)}$ is the
$k$-th row of $\boeta$, while in the typical IRLS they are scalars
$\displaystyle\mathbb{E}\left[-\frac{\partial^{2}\ell}{\partial\eta_{k}^{2}}\right]$,
which is often just
$\displaystyle-\frac{\partial^{2}\ell}{\partial\eta^{2}}$.

## Bootstrap variance estimators {#sec-boostrap}

We have implemented three types of bootstrap algorithms: parametric
(adapted from theory in @zwane, @norrpoll for multiple source setting
with covariates), semi-parametric (see e.g. @BoehningFriedl2021) and
nonparametric. The nonparametric version is the usual bootstrap
algorithm; which will typically underestimate the variance of $\hat{N}$.
In this section, the focus is on the first two approaches.

The idea of semi-parametric bootstrap is to modify the usual bootstrap
to include the additional uncertainty resulting from the fact that the
sample size is a random variable. This type of bootstrap is performed in
steps listed in Algorithm 2.

### Algorithm 2: Semi-parametric bootstrap

1.  Draw a sample of size
    $N_{obs}'\sim\text{Binomial}\left(N', \frac{N_{obs}}{N'}\right)$,
    where
    $N'=\lfloor\hat{N}\rfloor+\text{Bernoulli}\left(\lfloor\hat{N}\rfloor-\hat{N}\right)$
2.  Draw $N_{obs}'$ units from the data uniformly without replacement
3.  Obtain a new population size estimate $N_b$ using bootstrap data
4.  Repeat 1-3 steps $B$ times

In other words, we first draw a sample size and then a sample
conditional on the sample size. Note that when using the semi-parametric
bootstrap one implicitly assumes that the population size estimate
$\hat{N}$ is accurate. The last implemented bootstrap type is the
parametric algorithm, which first draws a finite population of size
$\approx\hat{N}$ from the superpopulation model and then samples from
this population according to the selected model, as described in
Algorithm 3.

### Algorithm 3: Parametric bootstrap

1.  Draw the number of covariates equal to
    $\lfloor\hat{N}\rfloor+\text{Bernoulli}\left(\lfloor\hat{N}\rfloor-\hat{N}\right)$
    proportional to the estimated contribution
    $(\mathbb{P}\left[Y_{k}>0|\bx_{k}\right])^{-1}$ with replacement
2.  Using the fitted model and regression coefficients $\hat{\bbeta}$
    draw for each covariate the $Y$ value from the corresponding
    probability measure on $\mathbb{N}\cup\{0\}$
3.  Truncate units with the drawn $Y$ value equal to $0$
4.  Obtain a population size estimate $N_b$ based on the truncated data
5.  Repeat 1-4 steps $B$ times

Note that in order for this type of algorithm to result in consistent
standard error estimates, it is imperative that the estimated model for
the entire superpopulation probability space is consistent, which may be
much less realistic than in the case of the semi-parametric bootstrap.
The parametric bootstrap algorithm is the default option in
**singleRcapture**.

# The main function {#sec-main}

## The `estimatePopsize` function

The **singleRcapture** package is built around the `estimatePopsize`
function. The main design objective was to make using `estimatePopsize`
as similar as possible to the standard `stats::glm` function or packages
for fitting zero-truncated regression models, such as **countreg** (e.g.
`countreg::zerotrunc` function). The `estimatePopsize` function is used
to first fit an appropriate (vector) generalized linear model and to
estimate the population size along with its variance. It is assumed that
the response vector (i.e. the dependent variable) corresponds to the
number of times a given unit was observed in the source. The most
important arguments are given in Table below; the obligatory ones are
`formula, data, model`.

| Argument | Description |
|--------------------|----------------------------------------------------|
| `formula` | The main formula (i.e., for the Poisson $\lambda$ parameter); |
| `data` | A `data.frame` (or `data.frame` coercible) object; |
| `model` | Either a function, a string, or a family class object specifying which model should be used; possible values are listed in the documentation. The supplied argument should have the form `model = "ztpoisson"`, `model = ztpoisson`, or if a link function should be specified, then `model = ztpoisson(lambdaLink = "log")` can be used; |
| `method` | A numerical method used to fit regression `IRLS` or `optim`; |
| `popVar` | A method for estimating variance of $\hat{N}$ and creating confidence intervals (either bootstrap, analytic, or skipping the estimation entirely); |
| `controlMethod`, `controlModel`, or `controlPopVar` | Control parameters for numerical fitting, specifying additional formulas (inflation, dispersion) and population size estimation, respectively; |
| `offset` | A matrix of offset values with the number of columns matching the number of distribution parameters, providing offset values to each of the linear predictors; |
| `...` | Additional optional arguments passed to other methods, e.g., `estimatePopsizeFit`; |

An important step in using `estimatePopsize` is specifying the `model`
parameter, which indicates the type of model that will be used for
estimating the *unobserved* part of the population. For instance, to fit
Chao's or Zelterman's model one should select `chao` or `zelterman` and,
assuming that one-inflation is present, one can select one of the
zero-truncated one-inflated (`ztoi`$^\ast$) or one-inflated
zero-truncated (`oizt`$^\ast$) models, such as `oiztpoisson` for Poisson
or `ztoinegbin` for NB2.

If it is assumed that heterogeneity is observed for NB2 models, one can
specify the formula in the `controlModel` argument with the
`controlModel` function and the `alphaFormula` argument. This enables
the user to provide a formula for the dispersion parameter in the NB2
models. If heterogeneity is assumed for `ztoi`$^\ast$ or `oizt`$^\ast$,
one can specify the `omegaFormula` argument, which corresponds to the
$\omega$ parameter in these models. Finally, if covariates are assumed
to be available for the hurdle models (`ztHurdle`$^\ast$ or
`Hurdlezt`$^\ast$), then `piFormula` can be specified, as it provides a
formula for the probability parameter in these models.

## Controlling variance estimation with `controlPopVar`

The `estimatePopsize` function makes it possible to specify the variance
estimation method via `popVar` (e.g. analytic or variance bootstrap) and
control the estimation process by specifying `controlPopVar`. In the
control function `controlPopVar` the user can specify the `bootType`
argument, which has three possible values:
`"parametric", "semiparametric"` and `"nonparametric"`. Additional
arguments accepted by the `contorlPopVar` function, which are relevant
to bootstrap, include:

-   `alpha`, `B` -- the significance level and the number of bootstrap
    samples to be performed, respectively, with $0.05$ and $500$ being
    the default options.
-   `cores` -- the number of process cores to be used in bootstrap (1 by
    default); parallel computing is enabled by **doParallel**
    [@doParallel], **foreach** [@foreach] and **parallel** packages
    [@parallel].
-   `keepbootStat` -- a logical value indicating whether to keep a
    vector of statistics produced by the bootstrap.
-   `traceBootstrapSize`, `bootstrapVisualTrace` -- logical values
    indicating whether sample and population size should be tracked
    (`FALSE` by default); these work only when `cores` = 1.
-   `fittingMethod`, `bootstrapFitcontrol` -- the fitting method (by
    default the same as the one used in the original call) and control
    parameters (`controlMethod`) for model fitting in the bootstrap.

In addition, the user can specify the type of confidence interval by
means of `confType` and the type of covariance matrix by using `covType`
for the analytical variance estimator (observed or the Fisher
information matrix).

In the next sections we present a case study involving the use of a
simple zero-truncated Poisson regression and a more advanced model:
zero-truncated one-inflated geometric regression with the `cloglog` link
function. First, we present the example dataset, then we describe how to
estimate the population size and assess the quality and diagnostics
measures. Finally, we show how to estimate the population size in
user-specified sub-populations.

# Data analysis example {#sec-study}

The package can be installed in the standard manner using:

```{r, installation, eval=FALSE}
install.packages("singleRcapture")
```

Then, we need to load the package using the following code:

```{r, loading-package, warning=FALSE}
library(singleRcapture)
```

## Dataset

We use a dataset from @ztpoisson, which contains information about
immigrants in four Dutch cities (Amsterdam, Rotterdam, The Hague and
Utrecht), who were staying in the country without a legal permit in 1995
and appeared in police records for that year. This dataset is included
in the package called `netherlandsimmigrant`:

```{r, loading_singleRcapture}
data(netherlandsimmigrant)
head(netherlandsimmigrant)
```

The number of times each individual appeared in the records is included
in the `capture` variable. The available covariates include
`gender, age, reason, nation`; the last two represent the reason for
being captured and the region of the world a given person comes from:

```{r, netherlandsimmigrant_summary}
summary(netherlandsimmigrant)
```

One notable characteristic of this dataset is that it contains a
disproportionately large number of individuals who were observed only
once (i.e. `r sum(netherlandsimmigrant$capture==1)`).

```{r, table_netherlands}
table(netherlandsimmigrant$capture)
```

The basic syntax of `estimatePopsize` is very similar to that of `glm`,
the same can be said about the output of the summary method except for
additional results of population size estimates (denoted as
`Population size estimation results`).

```{r, basic_model_netherlands}
basicModel <- estimatePopsize(
  formula = capture ~ gender + age + nation,
  model   = ztpoisson(),
  data    = netherlandsimmigrant,
  controlMethod = controlMethod(silent = TRUE)
)
summary(basicModel)
```

The output regarding the population size contains the point estimate,
the observed proportion (based on the input dataset), the standard error
and two confidence intervals: one relating to the point estimate, the
second -- to the observed proportion.

According to this simple model, the population size is about 12,500,
with about 15% of units observed in the register. The 95% CI under
normality indicates that the true population size is likely to be
between 7,000-18,000, with about 10-26% of the target population
observed in the register.

Since there is a reasonable suspicion that the act of observing a unit
in the dataset may lead to undesirable consequences for the person
concerned (in this case, a possible deportation, detention or something
similar). For these reasons, the user may consider one-inflated models,
such as zero-truncated one-inflated geometric model (specified by
`ztoigeom` family) and those presented below.

```{r, model_inflated_netherlands, warning=FALSE, cache=T}
set.seed(123456)
modelInflated <- estimatePopsize(
    formula = capture ~ nation,
    model   = ztoigeom(omegaLink = "cloglog"),
    data    = netherlandsimmigrant,
    controlModel = controlModel(
        omegaFormula = ~ gender + age
    ),
    popVar = "bootstrap",
    controlPopVar = controlPopVar(bootType = "semiparametric",
                                  B = 50)
)
summary(modelInflated)
```

According to this approach, the population size is about 7,000, which is
about 5,000 less than in the case of the naive Poisson approach. A
comparison of AIC and BIC suggests that the one-inflation model fits the
data better with BIC for `ztoigeom` `r round(BIC(modelInflated))` and
`r round(BIC(basicModel))` for `ztpoisson`.

We can access population size estimates using the following code, which
returns a list with numerical results.

```{r, extract_pop_size}
popSizeEst(basicModel)    # alternative: basicModel$populationSize
popSizeEst(modelInflated) # alternative: modelInflated$populationSize
```

The decision whether to use a zero-truncated Poisson or zero-truncated
one-inflated geometric model should be based on the assessment of the
model and the assumptions regarding the data generation process. One
possible method of selection is based on the likelihood ratio test,
which can be computed quickly and conveniently with the **lmtest**
(@lmtest) interface:

```{r, message=FALSE}
library(lmtest)
```

```{r}
lrtest(basicModel, modelInflated, 
       name = function(x) {
    if (family(x)$family == "ztpoisson")
        "Basic model"
    else "Inflated model"
})
```

However, the above is not a standard method of model selection in SSCR.
The next sections are dedicated to a detailed description of how to
assess the results using standard statistical tests and diagnostics.

## Testing marginal frequencies

A popular method of testing the model fit in single source
capture-recapture studies consists in comparing the fitted marginal
frequencies
$\displaystyle\sum_{j=1}^{N_{obs}}\hat{\mathbb{P}}\left[Y_{j}=k|\bx_{j}, Y_{j} > 0\right]$
with the observed marginal frequencies
$\displaystyle\sum_{j=1}^{N}\1_{\{k\}}(Y_{j})=\sum_{j=1}^{N_{obs}}\1_{\{k\}}(Y_{j})$
for $k\geq1$. If the fitted model bears sufficient resemblance to the
real data collection process, these quantities should be quite close and
both $G$ and $\chi^{2}$ tests can be used to test the statistical
significance of the discrepancy with the following **singleRcapture**
syntax for the Poisson model (rather poor fit):

```{r, marginal_freq_basic_model}
margFreq <- marginalFreq(basicModel)
summary(margFreq, df = 1, dropl5 = "group")
```

and for the one-inflated model (better fit):

```{r, marginal_freq_inflated_model}
margFreq_inf <- marginalFreq(modelInflated)
summary(margFreq_inf, df = 1, dropl5 = "group")
```

where the `dropl5` argument is used to indicate how to handle cells with
less than $5$ fitted observations. Note, however, that currently there
is no continuity correction.

## Diagnostics

The `singleRStaticCountData` class has a `plot` method implementing
several types of quick demonstrative plots, such as the rootogram
[@rootogram], for comparing fitted and marginal frequencies, which can
be generated with the following syntax:

```{r, rootogram, fig.dim=c(5,5), fig.pos="ht", fig.show='hold', fig.cap="Rootograms for ztpoisson (left) and ztoigeom (right) models"}
plot(   basicModel, plotType = "rootogram", main = "ZT Poisson model")
plot(modelInflated, plotType = "rootogram", main = "ZTOI Geometric model")
```

The above plots suggest that the `ztoigeom` model fits the data better.
Another important issue in population size estimation is to conduct
model diagnostics in order to verify whether influential observations
are present in the data. For this purpose the leave-one-out (LOO)
diagnostic implemented in the `dfbeta` from the **stats** package has
been adapted as shown below (multiplied by a factor of a hundred for
better readability):

```{r, dfbeta_basic_model}
dfb <- dfbeta(basicModel)
round(t(apply(dfb, 2, quantile)*100), 4)
```

```{r, df_beta_inflated_model}
dfi <- dfbeta(modelInflated)
round(t(apply(dfi, 2, quantile)*100), 4)
```

The result of the `dfbeta` can be further used in the `dfpopsize`
function, which can be used to quantify LOO on the population size. Note
the warning when the bootstap variance estimation is applied.

```{r, dfpopsize_basic_model}
dfb_pop <- dfpopsize(basicModel, dfbeta = dfb)
dfi_pop <- dfpopsize(modelInflated, dfbeta = dfi)
summary(dfb_pop)
summary(dfi_pop)
```

Figure 2 shows a comparison of the effect of deleting an observation on
the population size estimate and inverse probability weights, which
refer to the contribution of a given observation to the population size
estimate:

```{r dfpopsize_plot, fig.dim=c(5,5), fig.pos="ht", fig.show='hold', fig.cap="Results for ztpoisson (left) and ztoigeom (right) model"}
plot(basicModel, plotType = "dfpopContr", 
     dfpop = dfb_pop, xlim = c(-4500, 150))
plot(modelInflated, plotType = "dfpopContr", 
     dfpop = dfi_pop, xlim = c(-4500, 150))
```

These plots show how the population size changes if a given observation
is removed. For instance, if we remove observation
`r which.min(dfb_pop)`, then the population size will increase by about
`r abs(round(min(dfb_pop)))` for the `ztpoisson` model. In the case of
`ztoigeom`, the largest change is equal to `r abs(round(min(dfi_pop)))`
for observation `r which.min(dfi_pop)`.

The full list of plot types along with the list of optional arguments
that can be passed from the call to the `plot` method down to base **R**
and **graphics** functions can be found in the help file of the `plot`
method.

```{r, eval=FALSE}
?plot.singleRStaticCountData
```

## The `stratifyPopsize` function

Researchers may be interested not only in the total population size but
also in the size of specific sub-populations (e.g. males, females,
particular age groups). For this reason we have created the
`stratifyPopsize` function, which estimates the size by strata defined
by the coefficients in the model (the default option). The following
output presents results based on the `ztpoisson` and `ztoigeom` models.

```{r, strata}
popSizestrata <- stratifyPopsize(basicModel)
cols <- c("name", "Observed", "Estimated", "logNormalLowerBound", 
          "logNormalUpperBound")
popSizestrata_report <- popSizestrata[, cols]
cols_custom <- c("Name", "Obs", "Estimated", "LowerBound", "UpperBound")
names(popSizestrata_report) <- cols_custom
popSizestrata_report
```

```{r, strata-inflated}
popSizestrata_inflated <- stratifyPopsize(modelInflated)
popSizestrata_inflated_report <- popSizestrata_inflated[, cols]
names(popSizestrata_inflated_report) <- cols_custom
popSizestrata_inflated_report
```

The `stratifyPopsize` function prepared to handle objects of the
`singleRStaticCountData` class, accepts three optional parameters
`strata, alpha, cov`, which are used for specifying sub-populations,
significance levels and the covariance matrix to be used for computing
standard errors. An example of the full call is presented below.

```{r, custom-strata, message=F, warning=F}
library(sandwich)
popSizestrataCustom <- stratifyPopsize(
  object  = basicModel,
  strata = ~ gender + age, 
  alpha   = rep(c(0.1, 0.05), each=2), 
  cov     = vcovHC(basicModel, type = "HC4")
)

popSizestrataCustom_report <- popSizestrataCustom[, c(cols, "confLevel")]
names(popSizestrataCustom_report) <- c(cols_custom, "alpha")
popSizestrataCustom_report
```

We have provided integration with the **sandwich** [@sandwich] package
to correct the variance-covariance matrix in the $\delta$ method. In the
code we have used the `vcovHC` method for `singleRStaticCountData` class
from the **sandwich** package, different significance levels for
confidence intervals in each stratum and a formula to specify that we
want estimates for both males and females to be grouped by `nation` and
`age`. The `strata` parameter can be specified either as:

-   a formula with the empty left hand side, as shown in the example
    above (e.g. `~ gender * age`),
-   a logical vector with the number of entries equal to the number of
    rows in the dataset, in which case only one stratum will be created
    (e.g. `netherlandsimmigrant$gender == "male"`),
-   a vector of names of explanatory variables, which will result in
    every level of the explanatory variable having its own
    sub-population for each variable specified (e.g.
    `c("gender", "age")`),
-   not supplied at all, in which case strata will correspond to levels
    of each factor in the data without any interactions (string vectors
    will be converted to factors for the convenience of the user),
-   a (named) list where each element is a logical vector; names of the
    list will be used to specify variable names in the returned object,
    for example:

```{r, eval=FALSE}
list(
  "Stratum 1" = netherlandsimmigrant$gender == "male"   & 
    netherlandsimmigrant$nation == "Suriname", 
  "Stratum 2" = netherlandsimmigrant$gender == "female" & 
    netherlandsimmigrant$nation == "North Africa"
)
```

One can also specify `plotType = "strata"` in the `plot` function, which
results in a plot with point and CI estimates of the population size.

```{r, strata_plot, fig.dim=c(6,6), fig.show='hold', fig.pos = "ht", fig.cap="Population size by covariates for ztpoisson (left) and ztoigeom (right) model"}
plot(basicModel, plotType = "strata")
plot(modelInflated, plotType = "strata")
```

Only the `logNormal` type of confidence interval is used for plotting
since the studentized confidence intervals often result in negative
lower bounds.

# Classes and `S3Methods` {#sec-methods}

We have created a number of classes. The main ones are:
`singleRStaticCountData`, `singleRfamily`, and supplementary are:
`popSizeEstResults`, `summarysingleRmargin` and \newline
`summarysingleRStaticCountData`, which make it possible to extract
relevant information regarding the population size.

For instance, the `popSizeEst` function can be used to extract
information about the estimated size of the population as given below:

```{r, popsize_extract}
(popEst <- popSizeEst(basicModel))
```

and the resulting object `popEst` of the `popSizeEstResults` class
contains the following fields:

-   `pointEstimate`, `variance` -- numerics containing the point
    estimate and variance of this estimate.
-   `confidenceInterval` -- a `data.frame` with confidence intervals.
-   `boot` -- If the bootstrap was performed a numeric vector containing
    the $\hat{N}$ values from the bootstrap, a character vector with
    value `"No bootstrap performed"` otherwise.
-   `control` -- a `controlPopVar` object with controls used to obtain
    the object.

The only explicitly defined method for `popSizeEstResults`,
`summarysingleRmargin` and `summarysingleRStaticCountData` classes is
the `print` method, but the former one also accepts **R** primitives
like `coef`:

```{r, extract-coefs}
coef(summary(basicModel))
```

analogously to `glm` from **stats**. The `singleRfamily` inherits the
`family` class from **stats** and has explicitly defined `print` and
`simulate` methods. Example usage is presented below

| Function | Description |
|-------------------|-----------------------------------------------------|
| `fitted` | It works almost exactly like `glm` counterparts but returns more information, namely on fitted values for the truncated and non-truncated probability distribution; |
| `logLik` | Compared to `glm` method, it has the possibility of returning not just the value of the fitted log-likelihood but also the entire function (argument `type = "function"`) along with two first derivatives (argument `deriv = 0:2`); |
| `model.matrix` | It has the possibility of returning the $X_{\text{vlm}}$ matrix defined previously; |
| `simulate` | It calls the `simulate` method for the chosen model and fitted $\boeta$; |
| `predict` | It has the possibility of returning either fitted distribution parameters for each unit (`type = "response"`), or just linear predictors (`type = "link"`), or means of the fitted distributions of $Y$ and $Y|Y>0$ (`type = "mean"`) or the inverse probability weights (`type = "contr"`). It is possible to set the `se.fit` argument to `TRUE` in order to obtain standard errors for each of those by using the $\delta$ method. Also, it is possible to use a custom covariance matrix for standard error computation (argument `cov`); |
| `redoPopEstimation` | A function that applies all post-hoc procedures that were performed (such as heteroscedastic consistent covariance matrix estimation via **countreg**) to estimate the population size and standard errors; |
| `residuals` | Used for obtaining residuals of several types, we refer interested readers to the manual `?singleRcapture:::residuals.singleRStaticCountData`; |
| `stratifyPopsize, summary` | Compared to the `glm` class, summary has the possibility of adding confidence intervals to the coefficient matrix (argument `confint = TRUE`) and using a custom covariance matrix (argument `cov = someMatrix`); |
| `plot` | It has been discussed above; |
| `popSizeEst` | An extractor showcased above; |
| `cooks.distance` | It works only for single predictor models; |
| `dfbeta, dfpopsize` | Multi-threading in `dfbeta` is available and `dfpopsize` calls `dfbeta` if no `dfbeta` object was provided in the call; |
| `bread, estfun, vcovHC` | For (almost) full **sandwich** compatibility; |
| `AIC, BIC, extractAIC, family, confint, df.residual, model.frame, hatvalues, nobs, print, sigma, influence, rstudent, rstandard` | These work exactly like `glm` counterparts. |

```{r, simulate-method}
set.seed(1234567890)
N <- 10000
gender <- rbinom(N, 1, 0.2)
eta <- -1 + 0.5*gender
counts <- simulate(ztpoisson(), eta = cbind(eta), seed = 1)
summary(data.frame(gender, eta, counts))
```

The full list of explicitly defined methods for `singleRStaticCountData`
methods is presented in Table above.

# Concluding remarks

In this paper we have introduced the **singleRcapture** package for
single source capture-recapture models. The package implement
state-of-the-art methods for estimating population size based on a
single data set with multiple counts. The package implements different
methods to account for heterogeneity in capture probabilities, modelled
using covariates, as well as behavioural change, modelled using
one-inflation. We have built the package to facilitate the
implementation of new models using `family` objects; their application
is exemplified in the Section 7. An example of implementing a custom
family described in Section 8 is presented in replication materials.

Furthermore, since many **R** users are familiar with **countreg** or
**VGAM** packages, we have implemented a lightweight extension called
**singleRcaptureExtra**, available through Github
(\url{https://github.com/ncn-foreigners/singleRcaptureExtra}), which can
be used to integrate **singleRcapture** with these packages.

In future work we plan to implement Bayesian estimation using **Stan**
(e.g. via the **brms** package; @carpenter2017stan, @brms) and for
one-inflation models we can use the recent approach proposed by
@tuoto2022bayesian and implement our own families using the **brms**
package.

# Acknowledgements {#Acknowledgements}

The authors' work has been financed by the National Science Centre in
Poland, OPUS 20, grant no. 2020/39/B/HS4/00941.

The authors would like to thank Peter van der Heijden, Maarten Cruyff,
Dankmar Böhning, Łukasz Chrostowski and Layna Dennett for useful
comments that have helped to improve the functionality of the package.
In addition, we would like to thank Marcin Szymkowiak and Tymon
Świtalski for their valuable comments that have considerably improved
the paper.

# Detailed information {#sec-details}

## The `estimatePopsizeFit` function {#sec-estimatePopsizeFit short-title="The estimatePopsizeFit function"}

In this section we provide a step-by-step description of how to prepare
data in order to use the `estimatePopsizeFit` function, which may be
useful to some users, e.g. those wishing to make modifications to the
$\hat{N}$ estimate or to the bootstrap. In order to show how to apply
the function we will fit a one-inflated zero-truncated geometric model on the data
from @chao-generalization with covariate dependency:

\begin{align*}
  \log(\lambda) &=
  \beta_{1, 1} + \beta_{1, 2} \text{log\_distance} + \beta_{1, 3} \text{C\_TYPE} +
  \beta_{1, 4} \text{log\_size}, \\
  \text{logit}(\omega) &=
  \beta_{2, 1} + \beta_{2, 2} \text{log\_distance} + \beta_{2, 3} \text{C\_TYPE}.
\end{align*}

This would be equivalent to the following `esimatePopsize` call:

```{r, example-code-farm, eval=FALSE}
estimatePopsize(
  TOTAL_SUB ~ .,
  data = farmsubmission,
  model = oiztgeom(),
  controlModel(
    omegaFormula = ~ 1 + log_size + C_TYPE
  )
)
```

1.  Create a data matrix $\bX_{\text{vlm}}$

```{r, estimatePopsizeFit-1}
X <- matrix(data = 0, nrow = 2 * NROW(farmsubmission), ncol = 7)
```

2.  Fill the first $n$ rows with `model.matrix` according to the
    specified formula and specify the attribute `attr(X, "hwm")` that
    informs the function which elements of the design matrix correspond
    to which linear predictor (covariates for counts and covariates for
    one-inflation)

```{r, estimatePopsizeFit-2}
X[1:NROW(farmsubmission), 1:4] <- model.matrix(
  ~ 1 + log_size + log_distance + C_TYPE, 
  farmsubmission
)
X[-(1:NROW(farmsubmission)), 5:7] <- model.matrix(
  ~ 1 + log_distance + C_TYPE, 
  farmsubmission
)
attr(X, "hwm") <- c(4, 3)
```

3.  Obtain starting $\bbeta$ parameters using the `glm.fit` function.

```{r, estimatePopsizeFit-3}
start <- glm.fit(
  y = farmsubmission$TOTAL_SUB, 
  x = X[1:NROW(farmsubmission), 1:4], 
  family = poisson()
)$coefficients
start
```

4.  Use the `estimatePopsizeFit` function to fit the model assuming a
    zero-truncated one-inflated geometric distribution as specified in
    the `family` argument.

```{r, estimatePopsizeFit-4}
res <- estimatePopsizeFit(
  y            = farmsubmission$TOTAL_SUB, 
  X            = X, 
  method       = "IRLS", 
  priorWeights = 1, 
  family       = oiztgeom(), 
  control      = controlMethod(silent = TRUE), 
  coefStart    = c(start, 0, 0, 0),
  etaStart     = matrix(X %*% c(start, 0, 0, 0), ncol = 2),
  offset       = cbind(rep(0, NROW(farmsubmission)), 
                       rep(0, NROW(farmsubmission)))
)
```

5.  Compare our results with those obtained by applying the
    `stats::optim` function.

```{r, estimatePopsizeFit-5}
ll <- oiztgeom()$makeMinusLogLike(y = farmsubmission$TOTAL_SUB, X = X)
```

```{r, estimatePopsizeFit-6}
res2 <- estimatePopsizeFit(
  y = farmsubmission$TOTAL_SUB, 
  X = X, 
  method = "optim", 
  priorWeights = 1, 
  family = oiztgeom(), 
  coefStart = c(start, 0, 0, 0),
  control = controlMethod(silent = TRUE, maxiter = 10000),
  offset = cbind(rep(0, NROW(farmsubmission)), rep(0, NROW(farmsubmission)))
)
```

```{r, estimatePopsizeFit-7}
data.frame(IRLS  = round(c(res$beta, -ll(res$beta), res$iter), 4),
           optim = round(c(res2$beta, -ll(res2$beta), res2$iter[1]), 4))
```

The default `maxiter` parameter for `"optim"` fitting is $1000$, but we
needed to increase it since the `optim` does not converge in $1000$
steps and "gets stuck" at a plateau, which results in a lower
log-likelihood value compared to the standard `"IRLS"`.

The above situation is rather typical. While we did not conduct any
formal numerical analyses, it seems that when one attempts to model more
than one parameter of the distribution as covariate dependent `optim`
algorithms, both `"Nelder-Mead"` and `"L-BFGS-B"` seem to be ill-suited
for the task despite being provided with the analytically computed
gradient. This is one of the reasons why `"IRLS"` is the default fitting
method.

## Structure of a family function {#sec-family}

In this section we provide details regarding the `family` object for the
**singleRcapture** package. This object contains additional parameters
in comparison to the standard `family` object from the `stats` package.

| Function | Description |
|------------------|------------------------------------------------------|
| `makeMinusLogLike` | A factory function for creating the following functions: <br> $\ell(\bbeta)$, $\frac{\partial\ell}{\partial\bbeta}$, $\frac{\partial^{2}\ell}{\partial\bbeta^\top\partial\bbeta}$ <br> from the $\boldsymbol{y}$ vector and the $\bX_{vlm}$ matrix. The `deriv` argument has possible values in `c(0, 1, 2)` that determine which derivative to return; the default value is `0`, which represents the minus log-likelihood. |
| `links` | A list with link functions; |
| `mu.eta, variance` | Functions of linear predictors that return the expected value and variance. The `type` argument with 2 possible values (`"trunc"` and `"nontrunc"`) specifies whether to return $\mathbb{E}[Y|Y>0], \text{var}[Y|Y>0]$ or $\mathbb{E}[Y], \text{var}[Y]$ respectively. The `deriv` argument with values in `c(0, 1, 2)` is used for indicating the derivative with respect to the linear predictors, which is used for providing standard errors in the `predict` method. |
| `family` | A string that specifies the model name; |
| `valideta, validmu` | For now, it only returns `TRUE`. In the near future, it will be used to check whether applied linear predictors are valid (i.e., are transformed into some elements of the parameter space subjected to the inverse link function). |
| `funcZ, Wfun` | Functions that create pseudo residuals and working weights used in the IRLS algorithm; |
| `devResids` | A function that returns deviance residuals given a vector of prior weights of linear predictors and the response vector; |
| `pointEst, popVar` | Functions that return the point estimate for the population size and analytic estimation of its variance given prior weights of linear predictors and, in the latter case, also estimates of $\text{cov}(\hat{\bbeta})$ and $\bX_{\text{vlm}}$ matrix. There is an additional boolean parameter `contr` in the former function, which, if set to `TRUE`, returns the contribution of each unit. |
| `etaNames` | Names of linear predictors; |
| `densityFunction` | A function that returns the value of PMF at values of `x` given linear predictors. The `type` argument specifies whether to return $\mathbb{P}[Y|Y>0]$ or $\mathbb{P}[Y]$; |
| `simulate` | A function that generates values of a dependent vector given linear predictors; |
| `getStart` | An expression for generating starting points; |

# Implementing a custom **singleRcapture** family function {#sec-custom short-title="Implementing custom singleRcapture family function"}

Suppose we want to implement a very specific zero truncated family
function in the **singleRcapture**, which corresponds to the following
"untruncated" distribution:

$$
  \mathbb{P}[Y=y|\lambda, \pi] = \begin{cases}
    1 - \frac{1}{2}\lambda - \frac{1}{2}\pi & \text{when: } y=0\\
    \frac{1}{2}\pi & \text{when: } y=1\\
    \frac{1}{2}\lambda & \text{when: } y=2,
  \end{cases}
$$

with $\lambda, \pi\in\left(0, 1\right)$ being dependent on covariates.

Below we provide a possible way of implementing the above model, with
`lambda, pi` meaning $\frac{1}{2}\lambda,\frac{1}{2}\pi$. We provide a
simple example that shows that the proposed approach works as expected.

```{r custom-family-example}
myFamilyFunction <- function(lambdaLink = c("logit", "cloglog", "probit"),
                             piLink     = c("logit", "cloglog", "probit"),
                             ...) {
  if (missing(lambdaLink)) lambdaLink <- "logit"
  if (missing(piLink))         piLink <- "logit"

  links <- list()
  attr(links, "linkNames") <- c(lambdaLink, piLink)

  lambdaLink <- switch(lambdaLink,
                       "logit"   = singleRcapture:::singleRinternallogitLink,
                       "cloglog" = singleRcapture:::singleRinternalcloglogLink,
                       "probit"  = singleRcapture:::singleRinternalprobitLink
  )

  piLink <- switch(piLink,
                   "logit"   = singleRcapture:::singleRinternallogitLink,
                   "cloglog" = singleRcapture:::singleRinternalcloglogLink,
                   "probit"  = singleRcapture:::singleRinternalprobitLink
  )

  links[1:2] <- c(lambdaLink, piLink)

  mu.eta <- function(eta, type = "trunc", deriv = FALSE, ...) {
    pi     <-     piLink(eta[, 2], inverse = TRUE) / 2
    lambda <- lambdaLink(eta[, 1], inverse = TRUE) / 2

    if (!deriv) {
      switch (type,
              "nontrunc" = pi + 2 * lambda,
              "trunc" = 1 + lambda / (pi + lambda)
      )
    } else {
      # Only necessary if one wishes to use standard errors in predict method
      switch (type,
              "nontrunc" = {
                matrix(c(2, 1) * c(
                  lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) / 2,
                  piLink(eta[, 2], inverse = TRUE, deriv = 1) / 2
                ), ncol = 2)
              },
              "trunc" = {
                matrix(c(
                  pi / (pi + lambda) ^ 2,
                  -lambda / (pi + lambda) ^ 2
                ) * c(
                  lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) / 2,
                  piLink(eta[, 2], inverse = TRUE, deriv = 1) / 2
                ), ncol = 2)
              }
      )
    }
  }

  variance <- function(eta, type = "nontrunc", ...) {
    pi     <-     piLink(eta[, 2], inverse = TRUE) / 2
    lambda <- lambdaLink(eta[, 1], inverse = TRUE) / 2

    switch (type,
            "nontrunc" = pi * (1 - pi) + 4 * lambda * (1 - lambda - pi),
            "trunc" = lambda * (1 - lambda) / (pi + lambda)
    )
  }

  Wfun <- function(prior, y, eta, ...) {
    pi     <-     piLink(eta[, 2], inverse = TRUE) / 2
    lambda <- lambdaLink(eta[, 1], inverse = TRUE) / 2

    G01 <- ((lambda + pi) ^ (-2)) * piLink(eta[, 2], inverse = TRUE, deriv = 1) *
      lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) * prior / 4

    G00 <- ((lambda + pi) ^ (-2)) - (pi ^ (-2)) - lambda / ((lambda + pi) * (pi ^ 2))
    G00 <- G00 * prior * (piLink(eta[, 2], inverse = TRUE, deriv = 1) ^ 2) / 4

    G11 <- ((lambda + pi) ^ (-2)) - (((lambda + pi) * lambda) ^ -1)
    G11 <- G11 * prior * (lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) ^ 2) / 4

    matrix(
      -c(G11, # lambda
         G01, # mixed
         G01, # mixed
         G00  # pi
      ),
      dimnames = list(rownames(eta), c("lambda", "mixed", "mixed", "pi")),
      ncol = 4
    )
  }

  funcZ <- function(eta, weight, y, prior, ...) {
    pi     <-     piLink(eta[, 2], inverse = TRUE) / 2
    lambda <- lambdaLink(eta[, 1], inverse = TRUE) / 2

    weight <- weight / prior

    G0 <- (2 - y) / pi     - ((lambda + pi) ^ -1)
    G1 <- (y - 1) / lambda - ((lambda + pi) ^ -1)

    G1 <- G1 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) / 2
    G0 <- G0 *     piLink(eta[, 2], inverse = TRUE, deriv = 1) / 2

    uMatrix <- matrix(c(G1, G0), ncol = 2)

    weight <- lapply(X = 1:nrow(weight), FUN = function (x) {
      matrix(as.numeric(weight[x, ]), ncol = 2)
    })

    pseudoResid <- sapply(X = 1:length(weight), FUN = function (x) {
      #xx <- chol2inv(chol(weight[[x]])) # less computationally demanding
      xx <- solve(weight[[x]]) # more stable
      xx %*% uMatrix[x, ]
    })
    pseudoResid <- t(pseudoResid)
    dimnames(pseudoResid) <- dimnames(eta)
    pseudoResid
  }

  minusLogLike <- function(y, X, offset,
                           weight    = 1,
                           NbyK      = FALSE,
                           vectorDer = FALSE,
                           deriv     = 0,
                           ...) {
    y <- as.numeric(y)
    if (is.null(weight)) {
      weight <- 1
    }
    if (missing(offset)) {
      offset <- cbind(rep(0, NROW(X) / 2), rep(0, NROW(X) / 2))
    }

    if (!(deriv %in% c(0, 1, 2)))
      stop("Only score function and derivatives up to 2 are supported.")
    deriv <- deriv + 1

    switch (deriv,
            function(beta) {
              eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
              pi     <-     piLink(eta[, 2], inverse = TRUE) / 2
              lambda <- lambdaLink(eta[, 1], inverse = TRUE) / 2
              -sum(weight * ((2 - y) * log(pi) + (y - 1) * log(lambda) - log(pi + lambda)))
            },
            function(beta) {
              eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
              pi     <-     piLink(eta[, 2], inverse = TRUE) / 2
              lambda <- lambdaLink(eta[, 1], inverse = TRUE) / 2

              G0 <- (2 - y) / pi     - ((lambda + pi) ^ -1)
              G1 <- (y - 1) / lambda - ((lambda + pi) ^ -1)

              G1 <- G1 * weight * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) / 2
              G0 <- G0 * weight *     piLink(eta[, 2], inverse = TRUE, deriv = 1) / 2

              if (NbyK) {
                XX <- 1:(attr(X, "hwm")[1])
                return(cbind(as.data.frame(X[1:nrow(eta), XX]) * G1,
                             as.data.frame(X[-(1:nrow(eta)), -XX]) * G0))
              }
              if (vectorDer) {
                return(cbind(G1, G0))
              }

              as.numeric(c(G1, G0) %*% X)
            },
            function (beta) {
              lambdaPredNumber <- attr(X, "hwm")[1]
              eta <- matrix(as.matrix(X) %*% beta, ncol = 2) + offset
              pi     <-     piLink(eta[, 2], inverse = TRUE) / 2
              lambda <- lambdaLink(eta[, 1], inverse = TRUE) / 2

              res <- matrix(nrow = length(beta), ncol = length(beta),
                            dimnames = list(names(beta), names(beta)))

              # pi^2 derivative
              dpi <- (2 - y) / pi - (lambda + pi) ^ -1
              G00 <- ((lambda + pi) ^ (-2)) - (2 - y) / (pi ^ 2)

              G00 <- t(as.data.frame(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)] *
                                       (G00 * ((piLink(eta[, 2], inverse = TRUE, deriv = 1) / 2) ^ 2) +
                                          dpi * piLink(eta[, 2], inverse = TRUE, deriv = 2) / 2) * weight)) %*%
                as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)])
              # mixed derivative
              G01 <- (lambda + pi) ^ (-2)

              G01 <- t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber]) *
                         G01 * (lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) / 2) *
                         (piLink(eta[, 2], inverse = TRUE, deriv = 1) / 2) * weight) %*%
                as.matrix(X[-(1:(nrow(X) / 2)), -(1:lambdaPredNumber)])
              # lambda^2 derivative
              G11 <- ((lambda + pi) ^ (-2)) - (y - 1) / (lambda ^ 2)
              dlambda <- (y - 1) / lambda - ((lambda + pi) ^ -1)

              G11 <- t(as.data.frame(X[1:(nrow(X) / 2), 1:lambdaPredNumber] *
                                       (G11 * ((lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) / 2) ^ 2) +
                                          dlambda * lambdaLink(eta[, 1], inverse = TRUE, deriv = 2) / 2) * weight)) %*%
                X[1:(nrow(X) / 2), 1:lambdaPredNumber]

              res[-(1:lambdaPredNumber), -(1:lambdaPredNumber)] <- G00
              res[1:lambdaPredNumber, 1:lambdaPredNumber] <- G11
              res[1:lambdaPredNumber, -(1:lambdaPredNumber)] <- t(G01)
              res[-(1:lambdaPredNumber), 1:lambdaPredNumber] <- G01

              res
            }
    )
  }

  validmu <- function(mu) {
    (sum(!is.finite(mu)) == 0) && all(0 < mu) && all(2 > mu)
  }

  # this is optional
  devResids <- function(y, eta, wt, ...) {
    0
  }

  pointEst <- function (pw, eta, contr = FALSE, ...) {
    pi     <-     piLink(eta[, 2], inverse = TRUE) / 2
    lambda <- lambdaLink(eta[, 1], inverse = TRUE) / 2
    N <- pw / (lambda + pi)
    if(!contr) {
      N <- sum(N)
    }
    N
  }

  popVar <- function (pw, eta, cov, Xvlm, ...) {
    pi     <-     piLink(eta[, 2], inverse = TRUE) / 2
    lambda <- lambdaLink(eta[, 1], inverse = TRUE) / 2

    bigTheta1 <- -pw / (pi + lambda) ^ 2 # w.r to pi
    bigTheta1 <- bigTheta1 * piLink(eta[, 2], inverse = TRUE, deriv = 1) / 2
    bigTheta2 <- -pw / (pi + lambda) ^ 2 # w.r to lambda
    bigTheta2 <- bigTheta2 * lambdaLink(eta[, 1], inverse = TRUE, deriv = 1) / 2 # w.r to lambda

    bigTheta <- t(c(bigTheta2, bigTheta1) %*% Xvlm)

    f1 <- t(bigTheta) %*% as.matrix(cov) %*% bigTheta

    f2 <- sum(pw * (1 - pi - lambda) / ((pi + lambda) ^ 2))

    f1 + f2
  }

  dFun <- function (x, eta, type = c("trunc", "nontrunc")) {
    if (missing(type)) type <- "trunc"
    pi     <-     piLink(eta[, 2], inverse = TRUE) / 2
    lambda <- lambdaLink(eta[, 1], inverse = TRUE) / 2

    switch (type,
            "trunc" = {
              (pi * as.numeric(x == 1) + lambda * as.numeric(x == 2)) / (pi + lambda)
            },
            "nontrunc" = {
              (1 - pi - lambda) * as.numeric(x == 0) +
                pi * as.numeric(x == 1) + lambda * as.numeric(x == 2)
            }
    )
  }

  simulate <- function(n, eta, lower = 0, upper = Inf) {
    pi     <-     piLink(eta[, 2], inverse = TRUE) / 2
    lambda <- lambdaLink(eta[, 1], inverse = TRUE) / 2
    CDF <- function(x) {
      ifelse(x == Inf, 1,
             ifelse(x < 0, 0,
                    ifelse(x < 1, 1 - pi - lambda,
                           ifelse(x < 2, 1 - lambda, 1))))
    }
    lb <- CDF(lower)
    ub <- CDF(upper)
    p_u <- stats::runif(n, lb, ub)
    sims <- rep(0, n)
    cond <- CDF(sims) <= p_u
    while (any(cond)) {
      sims[cond] <- sims[cond] + 1
      cond <- CDF(sims) <= p_u
    }
    sims
  }

  getStart <- expression(
    if (method == "IRLS") {
      etaStart <- cbind(
        family$links[[1]](mean(observed == 2) * (1 + 0 * (observed == 2))), # lambda
        family$links[[2]](mean(observed == 1) * (1 + 0 * (observed == 1)))  # pi
      ) + offset
    } else if (method == "optim") {
      init <- c(
        family$links[[1]](weighted.mean(observed == 2, priorWeights) * 1 + .0001),
        family$links[[2]](weighted.mean(observed == 1, priorWeights) * 1 + .0001)
      )
      if (attr(terms, "intercept")) {
        coefStart <- c(init[1], rep(0, attr(Xvlm, "hwm")[1] - 1))
      } else {
        coefStart <- rep(init[1] / attr(Xvlm, "hwm")[1], attr(Xvlm, "hwm")[1])
      }
      if ("(Intercept):pi" %in% colnames(Xvlm)) {
        coefStart <- c(coefStart, init[2], rep(0, attr(Xvlm, "hwm")[2] - 1))
      } else {
        coefStart <- c(coefStart, rep(init[2] / attr(Xvlm, "hwm")[2], attr(Xvlm, "hwm")[2]))
      }
    }
  )

  structure(
    list(
      makeMinusLogLike = minusLogLike,
      densityFunction  = dFun,
      links     = links,
      mu.eta    = mu.eta,
      valideta  = function (eta) {TRUE},
      variance  = variance,
      Wfun      = Wfun,
      funcZ     = funcZ,
      devResids = devResids,
      validmu   = validmu,
      pointEst  = pointEst,
      popVar    = popVar,
      family    = "myFamilyFunction",
      etaNames  = c("lambda", "pi"),
      simulate  = simulate,
      getStart  = getStart,
      extraInfo = c(
        mean       = "pi / 2 + lambda",
        variance   = paste0("(pi / 2) * (1 - pi / 2) + 2 * lambda * (1 - lambda / 2 - pi / 2)"),
        popSizeEst = "(1 - (pi + lambda) / 2) ^ -1",
        meanTr     = "1 + lambda / (pi + lambda)",
        varianceTr = paste0("lambda * (1 - lambda / 2) / (pi + lambda)")
      )
    ),
    class = c("singleRfamily", "family")
  )
}
```

A quick tests shows us that this implementation in fact works:

```{r}
set.seed(123)
Y <- simulate(
  myFamilyFunction(lambdaLink = "logit", piLink = "logit"),
  nsim = 1000, eta = matrix(0, nrow = 1000, ncol = 2),
  truncated = FALSE
)
mm <- estimatePopsize(
  formula = Y ~ 1,
  data = data.frame(Y = Y[Y > 0]),
  model = myFamilyFunction(lambdaLink = "logit",
                           piLink = "logit"),
  # the usual observed information matrix
  # is ill-suited for this distribution
  controlPopVar = controlPopVar(covType = "Fisher")
)
summary(mm)
```

where the link functions, such as
`singleRcapture:::singleRinternalcloglogLink`, are just internal
functions in `singleRcapture` that compute link functions, their
inverses and derivatives of both links and inverse links up to the third
order:

```{r single-link}
singleRcapture:::singleRinternalcloglogLink
```

One could, of course, include the code for computing them manually.

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/strataEstimation.R
\name{stratifyPopsize}
\alias{stratifyPopsize}
\alias{stratifyPopsize.singleRStaticCountData}
\title{Estimate size of sub-populations}
\usage{
stratifyPopsize(object, strata, alpha, ...)

\method{stratifyPopsize}{singleRStaticCountData}(object, strata, alpha, cov = NULL, ...)
}
\arguments{
\item{object}{an object on which the population size estimates should be based
in \code{singleRcapture} package this is a fitter \code{singleRStaticCountData} class object.}

\item{strata}{a specification of sub populations given by one of:
\itemize{
\item formula -- a formula to be applied to \code{model.frame} extracted from
the object.
\item Logical vector with number of entries equal to number of rows in the dataset.
\item A (named) list where each element is a logical vector, names of the list
will be used to specify names variable in returned object.
\item Vector of names of explanatory variables. For \code{singleRStaticCountData} method
for this function this specification of \code{strata} parameter will
result in every level of explanatory variable having its own sub population
for each variable specified.
\item If no value was provided the \code{singleRStaticCountData} method for this function
will itself create sub populations based on levels of factor variables
in \code{model.frame}.
}}

\item{alpha}{significance level for confidence intervals --
Either a single numeric value or a vector of length equal to number of
sub populations specified in \code{strata}.
If missing it is set to \code{.05} in \code{singleRStaticCountData} method.}

\item{...}{a vector of arguments to be passed to other functions.
For \code{singleRStaticCountData} method for this functions arguments in \code{...} are
passed to either \code{cov} if argument provided was a function or
\code{vcov} if \code{cov} argument was missing at call.}

\item{cov}{for \code{singleRStaticCountData} method an estimate of variance-covariance matrix
for estimate of regression parameters. It is possible to pass a function
such as for example \code{sandwich::vcovHC} which will be called as:
\code{foo(object, ...)} and a user may specify additional arguments of a
function in \code{...} argument. If not provided an estimate for covariance
matrix will be set by calling appropriate \code{vcov} method.}
}
\value{
A \code{data.frame} object with row names being the names of specified
sub populations either provided or inferred.
}
\description{
A function that estimates sizes of specific sub populations
based on a capture-recapture model for the whole population.
}
\details{
\loadmathjax

In single source capture-recapture models the most frequently used
estimate for population size is Horvitz-Thompson type estimate:

\mjsdeqn{\hat{N} = \sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}(Y_{k}>0)} = 
\sum_{k=1}^{N_{obs}}\frac{1}{1-\mathbb{P}(Y_{k}=0)}}

where \mjseqn{I_{k}=I_{Y_{k} > 0}} are
indicator variables, with value 1 if kth unit was observed at least once
and 0 otherwise and the inverse probabilistic weights weights for
units observed in the data \mjseqn{\tfrac{1}{\mathbb{P}(Y_{k}>0)}}
are estimated using fitted linear predictors.

The estimates for different sub populations are made by changing the
\mjseqn{I_{k}=I_{Y_{k} > 0}} indicator variables to
refer not to the population as a whole but to the sub populations that are
being considered i.e. by changing values from 1 to 0 if kth unit is not a
member of sub population that is being considered at the moment.

The estimation of variance for these estimates and estimation of variance for
estimate of population size for the whole population follow the same relation
as the one described above.
}
\seealso{
\code{\link[=vcov.singleRStaticCountData]{vcov.singleRStaticCountData()}} \code{\link[=estimatePopsize]{estimatePopsize()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ratioReg.R
\name{ratioReg}
\alias{ratioReg}
\title{Estimate Population Size Using Ratio Regression}
\usage{
ratioReg(
  formula,
  data,
  ratioFormula = ~log(k + 1),
  model = c("auto", "M0", "M1"),
  criterion = c("AIC", "BIC"),
  estimator = c("SM", "HT"),
  weights = NULL,
  subset,
  confint = c("bootstrap", "none"),
  B = 1000,
  seed = NULL,
  maxCount = NULL,
  ...
)
}
\arguments{
\item{formula}{a formula identifying the observed count column. It must use
\code{~ 1}; for example \code{observed ~ 1}.}

\item{data}{a data frame or an object coercible to \code{data.frame}.}

\item{ratioFormula}{a one-sided formula used for the ratio regression on the
derived ratio-level data. The default is \code{~ log(k + 1)}.}

\item{model}{which ratio-regression model to use: \code{M0}, \code{M1}, or \code{auto} to
select between them by \code{criterion}.}

\item{criterion}{information criterion used when \code{model = "auto"}.}

\item{estimator}{which population-size estimator should be treated as the
primary one.}

\item{weights}{optional prior weights interpreted as multiplicities of
observed units.}

\item{subset}{optional logical expression selecting rows from \code{data}.}

\item{confint}{either \code{"bootstrap"} for percentile bootstrap intervals or
\code{"none"} to skip interval estimation.}

\item{B}{number of bootstrap replicates when \code{confint = "bootstrap"}.}

\item{seed}{optional random seed used for bootstrap sampling.}

\item{maxCount}{optional upper support bound. Counts larger than \code{maxCount}
are discarded before fitting.}

\item{...}{additional arguments passed to \code{\link[stats:model.frame]{stats::model.frame()}}.}
}
\value{
An object of class \code{singleRRatioReg} containing the fitted ratio-regression
models, model-selection criteria, derived ratio-level data, observed
frequencies, and population-size estimates.
}
\description{
Fits the ratio-regression approach for single-source capture-recapture
count data as a standalone estimator. The regression is performed on
neighbouring marginal-frequency ratios rather than on unit-level counts,
which is why this method is separate from the \code{singleRmodels()} family
framework used by \code{\link[=estimatePopsize]{estimatePopsize()}}.
}
\details{
Let \eqn{f_k} denote the observed frequency of units captured exactly
\eqn{k} times. Ratio regression models the neighbouring ratios
\eqn{r_k = f_{k + 1} / f_k}. The default model uses
\eqn{\log(r_k) = \beta_0 + \beta_1 \log(k + 1)} and weighted least squares
with weights \eqn{1 / (1 / f_k + 1 / f_{k + 1})}.

Two published models are supported:
\itemize{
\item \code{M0}: ratio regression without one-inflation term
\item \code{M1}: ratio regression with an additional one-inflation indicator
\code{I(k == 1)}
}

When \code{model = "auto"}, both models are fitted and compared by \code{criterion}.
Population size can then be extracted either through the Horvitz-Thompson
(\code{HT}) or semi-parametric (\code{SM}) estimator, with \code{estimator} selecting which
one is treated as primary by \code{\link[=popSizeEst]{popSizeEst()}}, \code{\link[=summary]{summary()}}, and \code{\link[=plot]{plot()}}.

Bootstrap intervals are percentile intervals. For fixed \code{model = "M0"} or
\code{model = "M1"} the bootstrap uses zero-count imputation based on the
primary estimator. For \code{model = "auto"} it uses the single-bootstrap
reselection scheme where model selection is repeated inside each bootstrap
sample.
}
\examples{
toy_counts <- data.frame(
  observed = c(rep(1, 50), rep(2, 30), rep(3, 15), rep(4, 6), rep(5, 2))
)

toy_fit <- ratioReg(
  observed ~ 1,
  data = toy_counts,
  model = "auto",
  estimator = "SM",
  confint = "none"
)

summary(toy_fit)
popSizeEst(toy_fit)
plot(toy_fit)

toy_tab <- aggregate(list(weight = rep(1, nrow(toy_counts))),
                     by = list(observed = toy_counts$observed),
                     FUN = sum)
ratioReg(
  observed ~ 1,
  data = toy_tab,
  weights = toy_tab$weight,
  model = "M0",
  confint = "none"
)

}
\seealso{
\code{\link[=estimatePopsize]{estimatePopsize()}} \code{\link[=popSizeEst]{popSizeEst()}}
}
\author{
Cyprian Jurkowski, Piotr Chlebicki, Maciej Beręsewicz
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/miscDiagnostics.R
\name{vcov.singleRStaticCountData}
\alias{vcov.singleRStaticCountData}
\title{Obtain covariance matrix estimation}
\usage{
\method{vcov}{singleRStaticCountData}(object, type = c("Fisher", "observedInform"), ...)
}
\arguments{
\item{object}{object of singleRStaticCountData class.}

\item{type}{type of estimate for covariance matrix for now either
expected (Fisher) information matrix or observed information matrix.}

\item{...}{additional arguments for method functions}
}
\value{
A covariance matrix for fitted coefficients, rows and columns of which
correspond to parameters returned by \code{coef} method.
}
\description{
A \code{vcov} method for \code{singleRStaticCountData} class.
}
\details{
\loadmathjax

Returns a estimated covariance matrix for model coefficients
calculated from analytic hessian or Fisher information matrix usually
utilizing asymptotic effectiveness of maximum likelihood estimates.
Covariance type is taken from control parameter that have been provided
on call that created \code{object} if arguments \code{type} was not specified.
}
\seealso{
\code{\link[=vcovHC.singleRStaticCountData]{vcovHC.singleRStaticCountData()}} \code{\link[sandwich:sandwich]{sandwich::sandwich()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ratioReg.R, R/smallMethods.R
\name{popSizeEst.singleRRatioReg}
\alias{popSizeEst.singleRRatioReg}
\alias{popSizeEst}
\alias{popSizeEst.singleRStaticCountData}
\title{Extract population size estimation results}
\usage{
\method{popSizeEst}{singleRRatioReg}(object, estimator = c("primary", "HT", "SM"), ...)

popSizeEst(object, ...)

\method{popSizeEst}{singleRStaticCountData}(object, ...)
}
\arguments{
\item{object}{object with population size estimates.}

\item{estimator}{for \code{singleRRatioReg} objects, which population-size
estimator to return. The default \code{"primary"} returns the estimator
selected when calling \code{\link[=ratioReg]{ratioReg()}}.}

\item{...}{additional optional arguments, currently not used in \code{singleRStaticCountData} class method.}
}
\value{
An object of class \code{popSizeEstResults} containing population size estimation results.
}
\description{
An extractor function with \code{singleRStaticCountData} method for extracting
important information regarding pop size estimate.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{plot.singleRStaticCountData}
\alias{plot.singleRStaticCountData}
\title{Diagnostic plots for regression and population size estimation}
\usage{
\method{plot}{singleRStaticCountData}(
  x,
  plotType = c("qq", "marginal", "fitresid", "bootHist", "rootogram", "dfpopContr",
    "dfpopBox", "scaleLoc", "cooks", "hatplot", "strata"),
  confIntStrata = c("normal", "logNormal"),
  histKernels = TRUE,
  dfpop,
  ...
)
}
\arguments{
\item{x}{object of \code{singleRStaticCountData} class.}

\item{plotType}{character parameter (default \code{"qq"}) specifying type of plot to be made.
The following list presents and briefly explains possible type of plots:
\itemize{
\item \code{qq} -- The quantile-quantile plot for pearson residuals
(or standardized pearson residuals if these are available for the model) i.e.
empirical quantiles from residuals are plotted against theoretical quantiles
from standard distribution.
\item \code{marginal} -- A plot made by \code{matplot} with fitted and
observed marginal frequencies with labels.
\item \code{fitresid} -- Plot of fitted linear predictors against
(standardized) pearson residuals.
\item \code{bootHist} -- Simple histogram for statistics obtained from
bootstrapping (if one was performed and the statistics were saved).
\item \code{rootogram} -- Rootogram, for full explanation see:
Kleiber and Zeileis Visualizing Count Data Regressions Using Rootograms (2016),
in short it is a \code{barplot} where height is the square root of observed marginal
frequencies adjusted by difference between square root of observed and fitted marginal
frequencies connected by line representing fitted marginal frequencies.
The less of a difference there is between the 0 line and beginning of a bar
the more accurate fit was produced by the model.
\item \code{dfpopContr} -- Plot of \code{dfpopsize} against unit contribution.
On the plot is y = x line i.e. what deletion effect would be if removing the
unit from the model didn't effect regression coefficients. The further away
the observation is from this line the more influential it is.
\item \code{dfpopBox} -- Boxplot of \code{dfpopsize} for getting the general
idea about the distribution of the "influence" of each unit on
population size estimate.
\item \code{scaleLoc} -- The scale - location plot i.e. square root of
absolute values of (standardized) pearson residuals against linear predictors
for each column of linear predictors.
\item \code{cooks} -- Plot of cooks distance for detecting influential observations.
\item \code{hatplot} -- Plot of hat values for each linear predictor for detecting influential observations.
\item \code{strata} -- Plot of confidence intervals and point estimates for strata provided in \code{...} argument
}}

\item{confIntStrata}{confidence interval type to use for strata plot.
Currently supported values are \code{"normal"} and \code{"logNormal"}.}

\item{histKernels}{logical value indicating whether to add density lines
to histogram.}

\item{dfpop}{an optional pre-computed result of \code{dfpopsize()} to avoid recomputation when using \code{plotType = "dfpopContr"} or \code{plotType = "dfpopBox"}.}

\item{...}{additional optional arguments passed to the following functions:
\itemize{
\item For \code{plotType = "bootHist"}
\itemize{
\item \code{graphics::hist} -- with \code{x, main, xlab, ylab} parameters fixed.
}
\item For \code{plotType = "rootogram"}
\itemize{
\item \code{graphics::barplot} -- with \code{height, offset, ylab, xlab, ylim} parameters fixed.
\item \code{graphics::lines} -- with \code{x, y, pch, type, lwd, col} parameters fixed.
}
\item For \code{plotType = "dfpopContr"}
\itemize{
\item \code{dfpopsize} -- with \code{model, observedPop} parameters fixed.
\item \code{plot.default} -- with \code{x, y, xlab, main} parameters fixed.
}
\item For \code{plotType = "dfpopBox"}
\itemize{
\item \code{dfpopsize} -- with \code{model, observedPop} parameters fixed.
\item \code{graphics::boxplot} -- with \code{x, ylab, main} parameters fixed.
}
\item For \code{plotType = "scaleLoc"}
\itemize{
\item \code{plot.default} -- with \code{x, y, xlab, ylab, main, sub} parameters fixed.
}
\item For \code{plotType = "fitresid"}
\itemize{
\item \code{plot.default} -- with \code{x, y, xlab, ylab, main, sub} parameters fixed.
}
\item For \code{plotType = "cooks"}
\itemize{
\item \code{plot.default} -- with \code{x, xlab, ylab, main} parameters fixed.
}
\item For \code{plotType = "hatplot"}
\itemize{
\item \code{hatvalues.singleRStaticCountData}
\item \code{plot.default} -- with \code{x, xlab, ylab, main} parameters fixed.
}
\item For \code{plotType = "strata"}
\itemize{
\item \code{stratifyPopsize.singleRStaticCountData}
}
}}
}
\value{
No return value only the plot being made.
}
\description{
Simple diagnostic plots for \code{singleRStaticCountData} class objects.
}
\seealso{
\code{\link[=estimatePopsize]{estimatePopsize()}} \code{\link[=dfpopsize]{dfpopsize()}} \code{\link[=marginalFreq]{marginalFreq()}} \code{\link[stats:plot.lm]{stats::plot.lm()}} \code{\link[stats:influence.measures]{stats::cooks.distance()}} \code{\link[=hatvalues.singleRStaticCountData]{hatvalues.singleRStaticCountData()}}
}
\author{
Piotr Chlebicki
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sandwichMethods.R
\name{estfun.singleRStaticCountData}
\alias{estfun.singleRStaticCountData}
\alias{bread.singleRStaticCountData}
\alias{vcovHC.singleRStaticCountData}
\title{Heteroscedasticity-consistent covariance matrix estimation for singleRStaticCountData class}
\usage{
\method{estfun}{singleRStaticCountData}(x, ...)

\method{bread}{singleRStaticCountData}(x, ...)

\method{vcovHC}{singleRStaticCountData}(
  x,
  type = c("HC3", "const", "HC", "HC0", "HC1", "HC2", "HC4", "HC4m", "HC5"),
  omega = NULL,
  sandwich = TRUE,
  ...
)
}
\arguments{
\item{x}{a fitted \code{singleRStaticCountData} class object.}

\item{...}{for \code{vcovHC} additional optional arguments passed to the following functions:
\itemize{
\item \code{estfun} -- for empirical estimating functions.
\item \code{hatvalues} -- for diagonal elements of projection matrix.
\item \code{sandwich} -- only if \code{sandwich} argument in function call was set to \code{TRUE}.
\item \code{vcov} -- when calling \code{bread} internally.
}}

\item{type}{a character string specifying the estimation type, same as in \code{sandwich::vcovHC.default}. HC3 is the default value.}

\item{omega}{a vector or a function depending on the arguments residuals (i.e. the derivative of log-likelihood with respect to each linear predictor), diaghat (the diagonal of the corresponding hat matrix) and df (the residual degrees of freedom), same as in \code{sandwich::vcovHC.default}.}

\item{sandwich}{logical. Should the sandwich estimator be computed? If set to FALSE only the meat matrix is returned. Same as in \code{\link[sandwich:vcovHC]{sandwich::vcovHC()}}}
}
\value{
Variance-covariance matrix estimation corrected for heteroscedasticity of regression errors
}
\description{
S3 method for \code{vcovHC} to handle \code{singleRStaticCountData} class objects.
Works exactly like \code{vcovHC.default} the only difference being that this method handles vector generalised linear models.
Updating the covariance matrix in variance/standard error estimation for population size estimator can be done via \code{\link[=redoPopEstimation]{redoPopEstimation()}}
}
\examples{
set.seed(1)
N <- 10000
gender <- rbinom(N, 1, 0.2)
eta <- -1 + 0.5*gender
counts <- rpois(N, lambda = exp(eta))
df <- data.frame(gender, eta, counts)
df2 <- subset(df, counts > 0)
mod1 <-  estimatePopsize(
  formula = counts ~ 1 + gender, 
  data = df2, 
  model = "ztpoisson", 
  method = "optim", 
  popVar = "analytic"
)
require(sandwich)
HC <- sandwich::vcovHC(mod1, type = "HC4")
Fisher <- vcov(mod1, "Fisher") # variance covariance matrix obtained from 
#Fisher (expected) information matrix
HC
Fisher
# usual results
summary(mod1)
# updated results
summary(mod1, cov = HC,
popSizeEst = redoPopEstimation(mod1, cov = HC))
# estimating equations
mod1_sims <- sandwich::estfun(mod1)
head(mod1_sims)
# bread method
all(vcov(mod1, "Fisher") * nrow(df2) == sandwich::bread(mod1, type = "Fisher"))
}
\seealso{
\code{\link[sandwich:vcovHC]{sandwich::vcovHC()}} \code{\link[=redoPopEstimation]{redoPopEstimation()}}
}
\author{
Piotr Chlebicki, Maciej Beręsewicz
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{controlMethod}
\alias{controlMethod}
\title{Control parameters for regression}
\usage{
controlMethod(
  epsilon = 1e-08,
  maxiter = 1000,
  verbose = 0,
  printEveryN = 1L,
  coefStart = NULL,
  etaStart = NULL,
  optimMethod = "Nelder-Mead",
  silent = FALSE,
  optimPass = FALSE,
  stepsize = 1,
  checkDiagWeights = TRUE,
  weightsEpsilon = 1e-08,
  momentumFactor = 0,
  saveIRLSlogs = FALSE,
  momentumActivation = 5,
  criterion = c("coef", "abstol", "reltol")
)
}
\arguments{
\item{epsilon}{a tolerance level for fitting algorithms by default \code{1e-8}.}

\item{maxiter}{a maximum number of iterations.}

\item{verbose}{a numeric value indicating whether to trace steps of fitting algorithm for
\code{IRLS} fitting method different values of verbose give the following information:
\itemize{
\item 1 -- Returns information on the number of current
iteration and current log-likelihood.
\item 2 -- Returns information on vector of regression parameters
at current iteration (and all of the above).
\item 3 -- Returns information on reduction of log-likelihood
at current iteration (and all of the above).
\item 4 -- Returns information on value of log-likelihood function gradient
at current iteration (and all of the above).
\item 5 -- Returns information on convergence criterion and values that are
taken into account when considering convergence (and all of the above).
}
if \code{optim} method was chosen verbose will be passed to \code{\link[stats:optim]{stats::optim()}} as trace.}

\item{printEveryN}{an integer value indicating how often to print information
specified in \code{verbose}, by default set to \code{1}.}

\item{coefStart, etaStart}{initial parameters for regression coefficients
or linear predictors if \code{NULL}. For \code{IRLS} fitting only \code{etaStart}
is needed so if \code{coefStart} is provided it will be converted to \code{etaStart},
for \code{optim} fitting \code{coefStart} is necessary and argument \code{etaStart}
will be ignored.}

\item{optimMethod}{a method of \code{\link[stats:optim]{stats::optim()}} used  \code{"Nelder-Mead"} is the default .}

\item{silent}{a logical value, indicating whether warnings in \code{IRLS} method should be suppressed.}

\item{optimPass}{an optional list of parameters passed to \code{stats::optim(..., control = optimPass)}
if FALSE then list of control parameters will be inferred from other parameters.}

\item{stepsize}{only for \code{IRLS}, scaling of updates to \code{beta} vector
lower value means slower convergence but more accuracy by default 1.
In general if fitting algorithm fails lowering this value tends to
be most effective at correcting it.}

\item{checkDiagWeights}{a logical value indicating whether to check if diagonal
elements of working weights matrixes in \code{IRLS} are sufficiently positive
so that these matrixes are positive defined. By default \code{TRUE}.}

\item{weightsEpsilon}{a small number to ensure positive definedness of weights matrixes.
Only matters if \code{checkDiagWeights} is set to \code{TRUE}.
By default \code{1e-8}.}

\item{momentumFactor}{an experimental parameter in \code{IRLS} only allowing for
taking previous step into account at current step, i.e instead of
updating regression parameters as:
\mjsdeqn{\boldsymbol{\beta}_{(a)} = 
\boldsymbol{\beta}_{(a-1)} + \text{stepsize} \cdot \text{step}_{(a)}}
the update will be made as:
\mjsdeqn{
\boldsymbol{\beta}_{(a)} = \boldsymbol{\beta}_{(a-1)} + \text{stepsize} 
\cdot (\text{step}_{(a)} + \text{momentum}\cdot\text{step}_{(a-1)})}}

\item{saveIRLSlogs}{a logical value indicating if information specified in
\code{verbose} should be saved to output object, by default \code{FALSE}.}

\item{momentumActivation}{the value of log-likelihood reduction below
which momentum will apply.}

\item{criterion}{a criterion used to determine convergence in \code{IRLS},
multiple values may be provided. By default \code{c("coef", "abstol")}.}
}
\value{
List with selected parameters, it is also possible to call list directly.
}
\description{
\code{controlMethod} constructs a list with all necessary
control parameters for regression fitting in
\code{estimatePopsizeFit} and \code{estimatePopsize}.
}
\seealso{
\code{\link[=estimatePopsize]{estimatePopsize()}} \code{\link[=estimatePopsizeFit]{estimatePopsizeFit()}}
\code{\link[=controlModel]{controlModel()}} \code{\link[=controlPopVar]{controlPopVar()}}
}
\author{
Piotr Chlebicki, Maciej Beręsewicz
\loadmathjax
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ratioReg.R
\name{summary.singleRRatioReg}
\alias{summary.singleRRatioReg}
\title{Summary Statistics for Ratio Regression Fits}
\usage{
\method{summary}{singleRRatioReg}(object, ...)
}
\arguments{
\item{object}{object of class \code{singleRRatioReg}.}

\item{...}{currently not used.}
}
\value{
An object of class \code{summarysingleRRatioReg}.
}
\description{
Summarizes a fitted \code{singleRRatioReg} object, including model-selection
criteria and population-size estimates.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal.R
\name{summary.singleRmargin}
\alias{summary.singleRmargin}
\title{Statistical tests of goodness of fit}
\usage{
\method{summary}{singleRmargin}(object, df, dropl5 = c("drop", "group", "no"), ...)
}
\arguments{
\item{object}{object of singleRmargin class.}

\item{df}{degrees of freedom if not provided the function will try and manually
but it is not always possible.}

\item{dropl5}{a character indicating treatment of cells with frequencies < 5
either grouping them, dropping or leaving them as is. Defaults to \code{drop}.}

\item{...}{currently does nothing.}
}
\value{
A chi squared test and G test for comparison between fitted and
observed marginal frequencies.
}
\description{
Performs two statistical test on observed and fitted
marginal frequencies. For G test the test statistic is computed as:
\loadmathjax
\mjsdeqn{G = 2\sum_{k}O_{k}\ln{\left(\frac{O_{k}}{E_{k}}\right)}}
and for \mjseqn{\chi^{2}} the test statistic is computed as:
\mjsdeqn{\chi^{2} = \sum_{k}\frac{\left(O_{k}-E_{k}\right)^{2}}{E_{k}}}
where \mjseqn{O_{k},E_{k}} denoted observed and fitted
frequencies respectively. Both of these statistics converge to
\mjseqn{\chi^2} distribution asymptotically with the same
degrees of freedom.

The convergence of \mjseqn{G, \chi^2} statistics to
\mjseqn{\chi^2} distribution may be violated if expected counts
in cells are too low, say < 5, so it is customary to either censor or
omit these cells.
}
\examples{
# Create a simple model
Model <- estimatePopsize(
  formula = capture ~ ., 
  data = netherlandsimmigrant, 
  model = ztpoisson, 
  method = "IRLS"
)
plot(Model, "rootogram")
# We see a considerable lack of fit
summary(marginalFreq(Model), df = 1, dropl5 = "group")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimatePopsize.R
\name{estimatePopsize}
\alias{estimatePopsize}
\alias{estimatePopsize.default}
\title{Single-source capture-recapture models}
\usage{
estimatePopsize(formula, ...)

\method{estimatePopsize}{default}(
  formula,
  data,
  model = c("ztpoisson", "ztnegbin", "ztgeom", "zotpoisson", "ztoipoisson",
    "oiztpoisson", "ztHurdlepoisson", "Hurdleztpoisson", "zotnegbin", "ztoinegbin",
    "oiztnegbin", "ztHurdlenegbin", "Hurdleztnegbin", "zotgeom", "ztoigeom", "oiztgeom",
    "ztHurdlegeom", "Hurdleztgeom", "zelterman", "chao", "oichao"),
  weights = NULL,
  subset = NULL,
  naAction = NULL,
  method = c("optim", "IRLS"),
  popVar = c("analytic", "bootstrap", "noEst"),
  controlMethod = NULL,
  controlModel = NULL,
  controlPopVar = NULL,
  modelFrame = TRUE,
  x = FALSE,
  y = TRUE,
  contrasts = NULL,
  ratioReg = FALSE,
  offset,
  ...
)
}
\arguments{
\item{formula}{a formula for the model to be fitted, only applied to the "main"
linear predictor. Only single response models are available.}

\item{...}{additional optional arguments passed to other methods eg.
\code{estimatePopsizeFit}.}

\item{data}{a data frame or object coercible to data.frame class containing
data for the regression and population size estimation.}

\item{model}{a model for regression and population estimate full description in \code{\link[=singleRmodels]{singleRmodels()}}.}

\item{weights}{an optional object of prior weights used in fitting the model.
Can be used to specify number of occurrences of rows in data see \code{\link[=controlModel]{controlModel()}}}

\item{subset}{a logical vector indicating which observations should be used
in regression and population size estimation. It will be evaluated on \code{data} argument provided on call.}

\item{naAction}{Not yet implemented.}

\item{method}{a method for fitting values currently supported: iteratively
reweighted least squares (\code{IRLS}) and maximum likelihood (\code{optim}).}

\item{popVar}{a method of constructing confidence interval and estimating
the standard error either analytic or bootstrap.
Bootstrap confidence interval type may be specified in
\code{controlPopVar.} There is also the third possible value of \code{noEst}
which skips the population size estimate all together.}

\item{controlMethod}{a list indicating parameters to use in fitting the model
may be constructed with \code{singleRcapture::controlMethod} function.
More information included in \code{\link[=controlMethod]{controlMethod()}}.}

\item{controlModel}{a list indicating additional formulas for regression
(like formula for inflation parameter/dispersion parameter) may be
constructed with \code{singleRcapture::controlModel} function.
More information will eventually be included in \code{\link[=controlModel]{controlModel()}}.}

\item{controlPopVar}{a list indicating parameters to use in estimating variance
of population size estimation may be constructed with
\code{singleRcapture::controlPopVar} function.
More information included in \code{\link[=controlPopVar]{controlPopVar()}}.}

\item{modelFrame, x, y}{logical values indicating whether to return model matrix,
dependent vector and model matrix as a part of output.}

\item{contrasts}{not yet implemented.}

\item{ratioReg}{deprecated placeholder for ratio regression. Use
\code{\link[=ratioReg]{ratioReg()}} instead.}

\item{offset}{a matrix of offset values with the number of columns matching the
number of distribution parameters providing offset values to each of
linear predictors.}
}
\value{
Returns an object of class \code{c("singleRStaticCountData", "singleR", "glm", "lm")}
with type \code{list} containing:\cr
\itemize{
\item{\code{y} -- Vector of dependent variable if specified at function call.}
\item{\code{X} -- Model matrix if specified at function call.}
\item{\code{formula} -- A list with formula provided on call and additional formulas specified in \code{controlModel}.}
\item{\code{call} -- Call matching original input.}
\item{\code{coefficients} -- A vector of fitted coefficients of regression.}
\item{\code{control} -- A list of control parameters for \code{controlMethod} and \code{controlModel}, \code{controlPopVar} is included in populationSize.}
\item{\code{model} -- Model which estimation of population size and regression was built, object of class family.}
\item{\code{deviance} -- Deviance for the model.}
\item{\code{priorWeights} -- Prior weight provided on call.}
\item{\code{weights} -- If \code{IRLS} method of estimation was chosen weights returned by \code{IRLS}, otherwise same as \code{priorWeights}.}
\item{\code{residuals} -- Vector of raw residuals.}
\item{\code{logL} -- Logarithm likelihood obtained at final iteration.}
\item{\code{iter} -- Numbers of iterations performed in fitting or if \code{stats::optim} was used number of call to loglikelihood function.}
\item{\code{dfResiduals} -- Residual degrees of freedom.}
\item{\code{dfNull} -- Null degrees of freedom.}
\item{\code{fittValues} -- Data frame of fitted values for both mu (the expected value) and lambda (Poisson parameter).}
\item{\code{populationSize} -- A list containing information of population size estimate.}
\item{\code{modelFrame} -- Model frame if specified at call.}
\item{\code{linearPredictors} -- Vector of fitted linear predictors.}
\item{\code{sizeObserved} -- Number of observations in original model frame.}
\item{\code{terms} -- terms attribute of model frame used.}
\item{\code{contrasts} -- contrasts specified in function call.}
\item{\code{naAction} -- naAction used.}
\item{\code{which} -- list indicating which observations were used in regression/population size estimation.}
\item{\code{fittingLog} -- log of fitting information for \code{"IRLS"} fitting if specified in \code{controlMethod}.}
}
}
\description{
\code{estimatePopsize} first fits appropriate (v)glm model and
then estimates full (observed and unobserved) population size.
In this types of models it is assumed that the response vector
(i.e. the dependent variable) corresponds to the number of times a given unit
was observed in the source.
Population size is then usually estimated by Horvitz-Thompson type estimator:

\mjsdeqn{\hat{N} = \sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}(Y_{k}>0)} = 
\sum_{k=1}^{N_{obs}}\frac{1}{1-\mathbb{P}(Y_{k}=0)}}

where \mjseqn{I_{k}=I_{Y_{k} > 0}} are indicator
variables, with value 1 if kth unit was observed at least once and 0 otherwise.
}
\details{
The generalized linear model is characterized by equation
\mjsdeqn{\boldsymbol{\eta}=\boldsymbol{X}\boldsymbol{\beta}}
where \mjseqn{\boldsymbol{X}} is the (lm) model matrix.
The vector generalized linear model is similarly characterized by equations
\mjsdeqn{\boldsymbol{\eta}_{k}=\boldsymbol{X}_{k}\boldsymbol{\beta}_{k}}
where \mjseqn{\boldsymbol{X}_{k}} is a (lm) model
matrix constructed from appropriate formula
(specified in \code{controlModel} parameter).

The \mjseqn{\boldsymbol{\eta}} is then a vector constructed as:

\mjsdeqn{\boldsymbol{\eta}=\begin{pmatrix}
\boldsymbol{\eta}_{1} \cr 
\boldsymbol{\eta}_{2} \cr
\dotso \cr
\boldsymbol{\eta}_{p}
\end{pmatrix}^{T}}

and in cases of models in our package the (vlm) model matrix
is constructed as a block matrix:

\mjsdeqn{\boldsymbol{X}_{vlm}=
\begin{pmatrix}
\boldsymbol{X}_{1} & \boldsymbol{0} &\dotso &\boldsymbol{0}\cr
\boldsymbol{0} & \boldsymbol{X}_{2} &\dotso &\boldsymbol{0}\cr
\vdots & \vdots & \ddots & \vdots\cr
\boldsymbol{0} & \boldsymbol{0} &\dotso &\boldsymbol{X}_{p}
\end{pmatrix}}

this differs from convention in \code{VGAM} package (if we only consider our
special cases of vglm models) but this is just a convention and does not
affect the model, this convention is taken because it makes fitting with
IRLS (explanation of algorithm in \code{\link[=estimatePopsizeFit]{estimatePopsizeFit()}}) algorithm easier.
(If \code{constraints} matrixes in \code{vglm} match the ones we implicitly
use the \code{vglm} model matrix differs with respect to order of
\code{kronecker} multiplication of \code{X} and \code{constraints}.)
In this package we use observed likelihood to fit regression models.

As mentioned above usually the population size estimation is done via:
\mjsdeqn{\hat{N} = \sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}(Y_{k}>0)} = 
\sum_{k=1}^{N_{obs}}\frac{1}{1-\mathbb{P}(Y_{k}=0)}}

where \mjseqn{I_{k}=I_{Y_{k} > 0}} are indicator variables,
with value 1 if kth unit was observed at least once and 0 otherwise.
The \mjseqn{\mathbb{P}(Y_{k}>0)} are estimated by maximum likelihood.

The following assumptions are usually present when using
the method of estimation described above:
\enumerate{
\item The specified regression model is correct. This entails linear
relationship between independent variables and dependent ones and
dependent variable being generated by appropriate distribution.
\item No unobserved heterogeneity. If this assumption is broken there
are some possible (admittedly imperfect) workarounds see details in
\code{\link[=singleRmodels]{singleRmodels()}}.
\item The population size is constant in relevant time frame.
\item Depending on confidence interval construction (asymptotic) normality
of \mjseqn{\hat{N}} statistic is assumed.
}

There are two ways of estimating variance of estimate \mjseqn{\hat{N}},
the first being \code{"analytic"} usually done by application of
law of total variance to \mjseqn{\hat{N}}:

\mjsdeqn{\text{var}(\hat{N})=\mathbb{E}\left(\text{var}
\left(\hat{N}|I_{1},\dots,I_{n}\right)\right)+
\text{var}\left(\mathbb{E}(\hat{N}|I_{1},\dots,I_{n})\right)}

and then by \mjseqn{\delta} method to
\mjseqn{\hat{N}|I_{1},\dots I_{N}}:

\mjsdeqn{\mathbb{E}\left(\text{var}
\left(\hat{N}|I_{1},\dots,I_{n}\right)\right)=
\left.\left(\frac{\partial(N|I_1,\dots,I_N)}{\partial\boldsymbol{\beta}}\right)^{T}
\text{cov}\left(\boldsymbol{\beta}\right)
\left(\frac{\partial(N|I_1,\dots,I_N)}{\partial\boldsymbol{\beta}}\right)
\right|_{\boldsymbol{\beta}=\hat{\boldsymbol{\beta}}}}

and the \mjseqn{\text{var}\left(\mathbb{E}(\hat{N}|I_{1},\dots,I_{n})\right)}
term may be derived analytically (if we assume independence of
observations) since \mjseqn{\hat{N}|I_{1},\dots,I_{n}}
is just a constant.

In general this gives us:
\mjsdeqn{
\begin{aligned}
\text{var}\left(\mathbb{E}(\hat{N}|I_{1},\dots,I_{n})\right)&=
\text{var}\left(\sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}(Y_{k}>0)}\right)\cr
&=\sum_{k=1}^{N}\text{var}\left(\frac{I_{k}}{\mathbb{P}(Y_{k}>0)}\right)\cr
&=\sum_{k=1}^{N}\frac{1}{\mathbb{P}(Y_{k}>0)^{2}}\text{var}(I_{k})\cr
&=\sum_{k=1}^{N}\frac{1}{\mathbb{P}(Y_{k}>0)^{2}}\mathbb{P}(Y_{k}>0)(1-\mathbb{P}(Y_{k}>0))\cr
&=\sum_{k=1}^{N}\frac{1}{\mathbb{P}(Y_{k}>0)}(1-\mathbb{P}(Y_{k}>0))\cr
&\approx\sum_{k=1}^{N}\frac{I_{k}}{\mathbb{P}(Y_{k}>0)^{2}}(1-\mathbb{P}(Y_{k}>0))\cr
&=\sum_{k=1}^{N_{obs}}\frac{1-\mathbb{P}(Y_{k}>0)}{\mathbb{P}(Y_{k}>0)^{2}}
\end{aligned}
}

Where the approximation on 6th line appears because in 5th line we sum over
all units, that includes unobserved units, since \mjseqn{I_{k}} are
independent and \mjseqn{I_{k}\sim b(\mathbb{P}(Y_{k}>0))} the 6th line
is an unbiased estimator of the 5th line.

The other method for estimating variance is \code{"bootstrap"}, but since
\mjseqn{N_{obs}=\sum_{k=1}^{N}I_{k}} is also a random variable bootstrap
will not be as simple as just drawing \mjseqn{N_{obs}} units from data
with replacement and just computing \mjseqn{\hat{N}}.

Method described above is referred to in literature as \code{"nonparametric"}
bootstrap (see \code{\link[=controlPopVar]{controlPopVar()}}), due to ignoring variability in observed
sample size it is likely to underestimate variance.

A more sophisticated bootstrap procedure may be described as follows:
\enumerate{
\item Compute the probability distribution as:
\mjsdeqn{
\frac{\hat{\boldsymbol{f}}_{0}}{\hat{N}}, 
\frac{\boldsymbol{f}_{1}}{\hat{N}}, 
\dots,
\frac{\boldsymbol{f}_{\max{y}}}{\hat{N}}}
where \mjseqn{\boldsymbol{f}_{n}} denotes observed
marginal frequency of units being observed exactly n times.
\item Draw \mjseqn{\hat{N}} units from the distribution above
(if \mjseqn{\hat{N}} is not an integer than draw
\mjseqn{\lfloor\hat{N}\rfloor + b(\hat{N}-\lfloor\hat{N}\rfloor)}),
where \mjseqn{\lfloor\cdot\rfloor} is the floor function.
\item Truncated units with \mjseqn{y=0}.
\item If there are covariates draw them from original data with replacement from
uniform distribution. For example if unit drawn to new data has
\mjseqn{y=2} choose one of covariate vectors from original data that
was associated with unit for which was observed 2 times.
\item Regress \mjseqn{\boldsymbol{y}_{new}} on \mjseqn{\boldsymbol{X}_{vlm new}}
and obtain \mjseqn{\hat{\boldsymbol{\beta}}_{new}}, with starting point
\mjseqn{\hat{\boldsymbol{\beta}}} to make it slightly faster, use them
to compute \mjseqn{\hat{N}_{new}}.
\item Repeat 2-5 until there are at least \code{B} statistics are obtained.
\item Compute confidence intervals based on \code{alpha} and \code{confType}
specified in \code{\link[=controlPopVar]{controlPopVar()}}.
}

To do step 1 in procedure above it is convenient to first draw binary vector of length
\mjseqn{\lfloor\hat{N}\rfloor + b(\hat{N}-\lfloor\hat{N}\rfloor)} with probability
\mjseqn{1-\frac{\hat{\boldsymbol{f}}_{0}}{\hat{N}}}, sum elements in that vector to
determine the sample size and then draw sample of this size uniformly from the data.

This procedure is known in literature as \code{"semiparametric"} bootstrap
it is necessary to assume that the have a correct estimate
\mjseqn{\hat{N}} in order to use this type of bootstrap.

Lastly there is \code{"parametric"} bootstrap where we assume that the
probabilistic model used to obtain \mjseqn{\hat{N}} is correct the
bootstrap procedure may then be described as:
\enumerate{
\item Draw \mjseqn{\lfloor\hat{N}\rfloor + b(\hat{N}-\lfloor\hat{N}\rfloor)}
covariate information vectors with replacement from data according to
probability distribution that is proportional to: \mjseqn{N_{k}},
where \mjseqn{N_{k}} is the contribution of kth unit i.e.
\mjseqn{\dfrac{1}{\mathbb{P}(Y_{k}>0)}}.
\item Determine \mjseqn{\boldsymbol{\eta}} matrix using estimate
\mjseqn{\hat{\boldsymbol{\beta}}}.
\item Generate \mjseqn{\boldsymbol{y}} (dependent variable)
vector using \mjseqn{\boldsymbol{\eta}} and
probability mass function associated with chosen model.
\item Truncated units with \mjseqn{y=0} and construct
\mjseqn{\boldsymbol{y}_{new}} and \mjseqn{\boldsymbol{X}_{vlm new}}.
\item Regress \mjseqn{\boldsymbol{y}_{new}} on \mjseqn{\boldsymbol{X}_{vlm new}}
and obtain \mjseqn{\hat{\boldsymbol{\beta}}_{new}}
use them to compute \mjseqn{\hat{N}_{new}}.
\item Repeat 1-5 until there are at least \code{B} statistics are obtained.
\item Compute confidence intervals based on \code{alpha} and \code{confType}
specified in \code{\link[=controlPopVar]{controlPopVar()}}
}

It is also worth noting that in the \code{"analytic"} method \code{estimatePopsize}
only uses "standard" covariance matrix estimation. It is possible that improper
covariance matrix estimate is the only part of estimation that has its assumptions
violated. In such cases post-hoc procedures are implemented in this package
to address this issue.

Lastly confidence intervals for \mjseqn{\hat{N}} are computed
(in analytic case) either by assuming that it follows a normal distribution
or that variable \mjseqn{\ln(N-\hat{N})}
follows a normal distribution.

These estimates may be found using either \code{summary.singleRStaticCountData}
method or \code{popSizeEst.singleRStaticCountData} function. They're labelled as
\code{normal} and \code{logNormal} respectively.
}
\examples{
\donttest{
# Model from 2003 publication 
# Point and interval estimation of the
# population size using the truncated Poisson regression mode
# Heijden, Peter GM van der et al. (2003)
model <- estimatePopsize(
   formula = capture ~ gender + age + nation, 
   data = netherlandsimmigrant, 
   model = ztpoisson
)
summary(model)
# Graphical presentation of model fit
plot(model, "rootogram")
# Statistical test
# see documentation for summary.singleRmargin
summary(marginalFreq(model), df = 1, "group")

# We currently support 2 methods of numerical fitting
# (generalized) IRLS algorithm and via stats::optim
# the latter one is faster when fitting negative binomial models 
# (and only then) due to IRLS having to numerically compute
# (expected) information matrixes, optim is also less reliable when
# using alphaFormula argument in controlModel
modelNegBin <- estimatePopsize(
    formula = TOTAL_SUB ~ ., 
    data = farmsubmission, 
    model = ztnegbin, 
    method = "optim"
)
summary(modelNegBin)
summary(marginalFreq(modelNegBin))

# More advanced call that specifies additional formula and shows
# in depth information about fitting procedure
pseudoHurdleModel <- estimatePopsize(
    formula = capture ~ nation + age,
    data = netherlandsimmigrant,
    model = Hurdleztgeom,
    method = "IRLS",
    controlMethod = controlMethod(verbose = 5),
    controlModel = controlModel(piFormula = ~ gender)
)
summary(pseudoHurdleModel)
# Assessing model fit
plot(pseudoHurdleModel, "rootogram")
summary(marginalFreq(pseudoHurdleModel), "group", df = 1)


# A advanced input with additional information for fitting procedure and
# additional formula specification and different link for inflation parameter.
Model <- estimatePopsize(
 formula = TOTAL_SUB ~ ., 
 data = farmsubmission, 
 model = oiztgeom(omegaLink = "cloglog"), 
 method = "IRLS", 
 controlMethod = controlMethod(
   stepsize = .85, 
   momentumFactor = 1.2,
   epsilon = 1e-10, 
   silent = TRUE
 ),
 controlModel = controlModel(omegaFormula = ~ C_TYPE + log_size)
)
summary(marginalFreq(Model), df = 18 - length(Model$coefficients))
summary(Model)
}
}
\references{
General single source capture recapture literature:

Zelterman, Daniel (1988). ‘Robust estimation in truncated discrete distributions
with application to capture-recapture experiments’. In: Journal of statistical
planning and inference 18.2, pp. 225–237.

Heijden, Peter GM van der et al. (2003). ‘Point and interval estimation of the
population size using the truncated Poisson regression model’.
In: Statistical Modelling 3.4, pp. 305–322. doi: 10.1191/1471082X03st057oa.

Cruyff, Maarten J. L. F. and Peter G. M. van der Heijden (2008). ‘Point and
Interval Estimation of the Population Size Using a Zero-Truncated Negative
Binomial Regression Model’. In: Biometrical Journal 50.6, pp. 1035–1050.
doi: 10.1002/bimj.200810455

Böhning, Dankmar and Peter G. M. van der Heijden (2009). ‘A covariate adjustment
for zero-truncated approaches to estimating the size of hidden and
elusive populations’. In: The Annals of Applied Statistics 3.2, pp. 595–610.
doi: 10.1214/08-AOAS214.

Böhning, Dankmar, Alberto Vidal-Diez et al. (2013). ‘A Generalization of
Chao’s Estimator for Covariate Information’. In: Biometrics 69.4, pp. 1033–
1042. doi: 10.1111/biom.12082

Böhning, Dankmar and Peter G. M. van der Heijden (2019). ‘The identity of the
zero-truncated, one-inflated likelihood and the zero-one-truncated likelihood
for general count densities with an application to drink-driving in Britain’.
In: The Annals of Applied Statistics 13.2, pp. 1198–1211.
doi: 10.1214/18-AOAS1232.

Navaratna WC, Del Rio Vilas VJ, Böhning D. Extending Zelterman's approach for
robust estimation of population size to zero-truncated clustered Data.
Biom J. 2008 Aug;50(4):584-96. doi: 10.1002/bimj.200710441.

Böhning D. On the equivalence of one-inflated zero-truncated and zero-truncated
one-inflated count data likelihoods. Biom J. 2022 Aug 15. doi: 10.1002/bimj.202100343.

Böhning, D., Friedl, H. Population size estimation based upon zero-truncated,
one-inflated and sparse count data. Stat Methods Appl 30, 1197–1217 (2021).
doi: 10.1007/s10260-021-00556-8

Bootstrap:

Zwane, PGM EN and Van der Heijden, Implementing the parametric bootstrap in capture-recapture
models with continuous covariates 2003 Statistics & probability letters 65.2 pp 121-125

Norris, James L and Pollock, Kenneth H Including model uncertainty in estimating variances
in multiple capture studies 1996 in Environmental and Ecological Statistics 3.3 pp 235-244

Vector generalized linear models:

Yee, T. W. (2015). Vector Generalized Linear and Additive Models:
With an Implementation in R. New York, USA: Springer. ISBN 978-1-4939-2817-0.
}
\seealso{
\code{\link[stats:glm]{stats::glm()}} -- For more information on generalized linear models.

\code{\link[stats:optim]{stats::optim()}} -- For more information on \code{optim} function used in
\code{optim} method of fitting regression.

\code{\link[=controlMethod]{controlMethod()}} -- For control parameters related to regression.

\code{\link[=controlPopVar]{controlPopVar()}} -- For control parameters related to population size estimation.

\code{\link[=controlModel]{controlModel()}} -- For control parameters related to model specification.

\code{\link[=estimatePopsizeFit]{estimatePopsizeFit()}} -- For more information on fitting procedure in
\code{estimate_popsize}.

\code{\link[=popSizeEst]{popSizeEst()}} \code{\link[=redoPopEstimation]{redoPopEstimation()}} -- For extracting population size
estimation results are applying post-hoc procedures.

\code{\link[=summary.singleRStaticCountData]{summary.singleRStaticCountData()}} -- For summarizing important information about the
model and population size estimation results.

\code{\link[=marginalFreq]{marginalFreq()}} -- For information on marginal frequencies and comparison
between observed and fitted quantities.

\code{\link[VGAM:vglm]{VGAM::vglm()}} -- For more information on vector generalized linear models.

\code{\link[=singleRmodels]{singleRmodels()}} -- For description of various models.
}
\author{
Piotr Chlebicki, Maciej Beręsewicz

\loadmathjax
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate.R
\name{simulate}
\alias{simulate}
\alias{simulate.singleRStaticCountData}
\alias{simulate.singleRfamily}
\title{Generating data in singleRcapture}
\usage{
\method{simulate}{singleRStaticCountData}(object, nsim = 1, seed = NULL, ...)

\method{simulate}{singleRfamily}(object, nsim, seed = NULL, eta, truncated = FALSE, ...)
}
\arguments{
\item{object}{an object representing a fitted model.}

\item{nsim}{a numeric scalar specifying:
\itemize{
\item number of response vectors to simulate in \code{simulate.singleRStaticCountData}, defaults to \code{1L}.
\item number of units to draw in \code{simulate.singleRfamily}, defaults to \code{NROW(eta)}.
}}

\item{seed}{an object specifying if and how the random number generator should be initialized (‘seeded’).}

\item{...}{additional optional arguments.}

\item{eta}{a matrix of linear predictors}

\item{truncated}{logical value indicating whether to sample from truncated or
full distribution.}
}
\value{
a \code{data.frame} with \code{n} rows and \code{nsim} columns.
}
\description{
An S3 method for \code{stats::simulate} to handle \code{singleRStaticCountData} and
\code{singleRfamily} classes.
}
\examples{
N <- 10000
###gender <- rbinom(N, 1, 0.2)
gender <- rep(0:1, c(8042, 1958))
eta <- -1 + 0.5*gender
counts <- simulate(ztpoisson(), eta = cbind(eta), seed = 1)
df <- data.frame(gender, eta, counts)
df2 <- subset(df, counts > 0)
### check coverage with summary
mod1 <-  estimatePopsize(
  formula       = counts ~ 1 + gender, 
  data          = df2, 
  model         = ztpoisson, 
  controlMethod = list(silent = TRUE)
)
mod1_sims <- simulate(mod1, nsim=10, seed = 1)
colMeans(mod1_sims)
mean(df2$counts)
}
\seealso{
\code{\link[stats:simulate]{stats::simulate()}} \code{\link[=estimatePopsize]{estimatePopsize()}}
}
\author{
Maciej Beręsewicz, Piotr Chlebicki
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.singleRStaticCountData}
\alias{summary.singleRStaticCountData}
\title{Summary statistics for model of singleRStaticCountData class}
\usage{
\method{summary}{singleRStaticCountData}(
  object,
  test = c("t", "z"),
  resType = "pearson",
  correlation = FALSE,
  confint = FALSE,
  cov,
  popSizeEst,
  ...
)
}
\arguments{
\item{object}{object of singleRStaticCountData class.}

\item{test}{type of test for significance of parameters \code{"t"} for t-test
and \code{"z"} for normal approximation of students t distribution, by
default \code{"z"} is used if there are more than 30 degrees of freedom
and \code{"t"} is used in other cases.}

\item{resType}{type of residuals to summarize any value that is allowed in
\code{residuals.singleRStaticCountData} except for \code{"all"} is allowed. By default
pearson residuals are used.}

\item{correlation}{logical value indicating whether correlation matrix should
be computed from covariance matrix by default \code{FALSE}.}

\item{confint}{logical value indicating whether confidence intervals for
regression parameters should be constructed. By default \code{FALSE}.}

\item{cov}{covariance matrix corresponding to regression parameters.
It is possible to give \code{cov} argument as a function of \code{object}.
If not specified it will be constructed using \code{vcov.singleRStaticCountData} method.
(i.e using Cramer-Rao lower bound)}

\item{popSizeEst}{a \code{popSizeEstResults} class object.
If not specified population size estimation results will be drawn from
\code{object}. If any post-hoc procedures, such as sandwich covariance matrix
estimation or bias reduction, were taken it is possible to include them in
population size estimation results by calling \code{redoPopEstimation}.}

\item{...}{additional optional arguments passed to the following functions:
\itemize{
\item \code{vcov.singleRStaticCountData} -- if no \code{cov} argument was provided.
\item \code{cov} -- if \code{cov} parameter specified at call was a function.
\item \code{confint.singleRStaticCountData} -- if \code{confint} parameter was set to \code{TRUE} at function call.
In particular it is possible to set confidence level in \code{...}.
}}
}
\value{
An object of \code{summarysingleRStaticCountData} class containing:
\itemize{
\item \code{call} -- A call which created \code{object}.
\item \code{coefficients} -- A dataframe with estimated regression coefficients
and their summary statistics such as standard error Wald test statistic and
p value for Wald test.
\item \code{residuals} -- A vector of residuals of type specified at call.
\item \code{aic} -- Akaike's information criterion.
\item \code{bic} -- Bayesian (Schwarz's) information criterion.
\item \code{iter} -- Number of iterations taken in fitting regression.
\item \code{logL} -- Logarithm of likelihood function evaluated at coefficients.
\item \code{deviance} -- Residual deviance.
\item \code{populationSize} -- Object with population size estimation results.
\item \code{dfResidual} -- Residual degrees of freedom.
\item \code{sizeObserved} -- Size of observed population.
\item \code{correlation} -- Correlation matrix if \code{correlation} parameter was set to \code{TRUE}
\item \code{test} -- Type of statistical test performed.
\item \code{model} -- Family class object specified in call for \code{object}.
\item \code{skew} -- If bootstrap sample was saved contains estimate of skewness.
}
}
\description{
A \code{summary} method for \code{singleRStaticCountData} class
}
\details{
\loadmathjax

Works
analogically to \code{summary.glm} but includes population size estimation
results. If any additional statistics, such as confidence intervals for
coefficients or coefficient correlation, are specified they will be printed.
}
\seealso{
\code{\link[=redoPopEstimation]{redoPopEstimation()}} \code{\link[stats:summary.glm]{stats::summary.glm()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{netherlandsimmigrant}
\alias{netherlandsimmigrant}
\title{Immigration data in the Netherlands}
\format{
Data frame with 1,880 rows and 5 columns.
\describe{
\item{\code{capture}}{Number of times a person has been captured by police.}
\item{\code{gender}}{Factor describing gender of the apprehended person.}
\item{\code{age}}{Factor describing age of apprehended person.
Either below or above 40 years old.}
\item{\code{reason}}{Factor describing reason for being apprehended by
police either illegal stay in Netherlands or other reasons.}
\item{\code{nation}}{Factor with nation of origin of the captured person.
There are 6 levels of this variable: \code{"American and Australia",
  "Asia", "North Africa", "Rest of Africa", "Surinam", "Turkey"}.}
}
}
\usage{
data("netherlandsimmigrant")
}
\description{
This data set contains information about immigrants in four
cities (Amsterdam, Rotterdam, The Hague and Utrecht) in Netherlands that have
been staying in the country illegally in 1995 and have appeared in police
records that year.
}
\references{
This data set and its description was provided in publication:
van Der Heijden, P. G., Bustami, R., Cruyff, M. J., Engbersen, G., and Van Houwelingen, H. C. (2003). Point and interval estimation of the population size using the truncated Poisson regression model. \emph{Statistical Modelling}, 3(4), 305-322. doi:10.1191/1471082X03st057oa
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.singleRStaticCountData}
\alias{predict.singleRStaticCountData}
\title{Predict method for singleRStaticCountData class}
\usage{
\method{predict}{singleRStaticCountData}(
  object,
  newdata,
  type = c("response", "link", "mean", "popSize", "contr"),
  se.fit = FALSE,
  na.action = NULL,
  weights,
  cov,
  ...
)
}
\arguments{
\item{object}{an object of \code{singleRStaticCountData} class.}

\item{newdata}{an optional \code{data.frame} containing new data.}

\item{type}{the type of prediction required, possible values are:
\itemize{
\item \code{"response"}-- For matrix containing estimated distributions
parameters.
\item \code{"link"}    -- For matrix of linear predictors.
\item \code{"mean"}    -- For fitted values of both \mjseqn{Y} and
\mjseqn{Y|Y>0}.
\item \code{"contr"}   -- For inverse probability weights (here named for
observation contribution to population size estimate).
\item \code{"popSize"} -- For population size estimation. Note
this results in a call to \code{redoPopEstimation} and it is
usually better to call this function directly.
} by default set to \code{"response"}.}

\item{se.fit}{a logical value indicating whether standard errors should be
computed. Only matters for \code{type} in \code{"response", "mean", "link"}.}

\item{na.action}{does nothing yet.}

\item{weights}{optional vector of weights for \code{type} in \code{"contr", "popSize"}.}

\item{cov}{optional matrix or function or character specifying either
a covariance matrix or a function to compute that covariance matrix.
By default \code{vcov.singleRStaticCountData} can be set to e.g. \code{vcovHC}.}

\item{...}{arguments passed to other functions, for now this only affects
\code{vcov.singleRStaticCountData} method and \code{cov} function.}
}
\value{
Depending on \code{type} argument if one of \code{"response", "link", "mean"}
a matrix with fitted values and possibly standard errors if \code{se.fit}
argument was set to \code{TRUE}, if \code{type} was set to \code{"contr"}
a vector with inverses of probabilities, finally for \code{"popSize"}
an object of class \code{popSizeEstResults} with its own methods containing
population size estimation results.
}
\description{
A method for \code{predict} function, works analogous to \code{predict.glm}
but gives the possibility to get standard errors of
mean/distribution parameters and directly get pop size estimates for new data.
}
\details{
\loadmathjax

Standard errors are computed with assumption of regression
coefficients being asymptotically normally distributed, if this assumption
holds then each of linear predictors i.e. each row of
\mjseqn{\boldsymbol{\eta}=\boldsymbol{X}_{vlm}\boldsymbol{\beta}}
is asymptotically normally distributed and their variances are expressed by
well known formula. The mean \mjseqn{\mu} and distribution parameters
are then differentiable functions of asymptotically normally distributed
variables and therefore their variances can be computed using (multivariate)
delta method.
}
\seealso{
\code{\link[=redoPopEstimation]{redoPopEstimation()}} \code{\link[stats:summary.glm]{stats::summary.glm()}} \code{\link[=estimatePopsize]{estimatePopsize()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{carcassubmission}
\alias{carcassubmission}
\title{The British farm carcass submissions data}
\format{
Data frame with 1,858 rows and 4 columns.
\describe{
\item{\code{TOTAL_SUB}}{Number of submissions of animal carcasses.}
\item{\code{log_size}}{Numerical value equal to logarithm of size of farm.}
\item{\code{log_distance}}{Numerical value equal to logarithm of distance to nearest AHVLA center.}
\item{\code{C_TYPE}}{Factor describing type of activity on farm that animals are
used for. Either \code{Dairy} or \code{Beef}}
}
}
\usage{
data("carcassubmission")
}
\description{
Data on British animal farms submissions to AHVLA. British farms
are able to submit samples to AHVLA if cause of death for an animal
cannot be determined and private veterinary surgeon decides to submit them,
unless there is notifiable disease suspected then such a submission is not
required.

This data set contains information about such farms.
Only submissions that are included in this data frame
are submissions of carcasses i.e. submissions of blood samples etc.
are excluded.
}
\references{
This data set and its description was provided in publication:
Böhning, D., Vidal Diez, A., Lerdsuwansri, R., Viwatwongkasem, C., and Arnold, M. (2013). "A generalization of Chao's estimator for covariate information". \emph{Biometrics}, 69(4), 1033-1042. doi:10.1111/biom.12082
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{controlModel}
\alias{controlModel}
\title{Control parameters specific to some models}
\usage{
controlModel(
  weightsAsCounts = FALSE,
  omegaFormula = ~1,
  alphaFormula = ~1,
  piFormula = ~1
)
}
\arguments{
\item{weightsAsCounts}{a boolean value indicating whether to treat \code{weights}
argument as number of occurrences for each row in the \code{data} and adjust
necessary methods and functionalities, like adjustments in bootstrap or
decreasing weights in \code{dfbeta} instead or deleting rows from data,
to accommodate this form of model specification.}

\item{omegaFormula}{a formula for inflation parameter in one inflated zero
truncated and zero truncated one inflated models.}

\item{alphaFormula}{a formula for dispersion parameter in negative binomial
based models.}

\item{piFormula}{a formula for probability parameter in pseudo hurdle zero
truncated and zero truncated pseudo hurdle models.}
}
\value{
A list with selected parameters, it is also possible to call list directly.
}
\description{
\code{controlModel} constructs a list with all necessary
control parameters in \code{estimatePopsize} that are either specific to
selected model or do not fit anywhere else.

Specifying additional formulas should be done by using only right hand side of
the formula also for now all variables from additional formulas should also be
included in the "main" formula.
}
\seealso{
\code{\link[=estimatePopsize]{estimatePopsize()}} \code{\link[=controlMethod]{controlMethod()}} \code{\link[=controlPopVar]{controlPopVar()}} \code{\link[=singleRmodels]{singleRmodels()}}
}
\author{
Piotr Chlebicki, Maciej Beręsewicz
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/miscDiagnostics.R
\name{confint.singleRStaticCountData}
\alias{confint.singleRStaticCountData}
\title{Confidence intervals for model parameters}
\usage{
\method{confint}{singleRStaticCountData}(object, parm, level = 0.95, ...)
}
\arguments{
\item{object}{object of singleRStaticCountData class.}

\item{parm}{names of parameters for which confidence intervals are to be
computed, if missing all parameters will be considered.}

\item{level}{confidence level for intervals.}

\item{...}{currently does nothing.}
}
\value{
An object with named columns that include upper and
lower limit of confidence intervals.
}
\description{
A function that computes studentized confidence intervals
for model coefficients.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/documentationFiles.R, R/leave-one-out.R,
%   R/miscDiagnostics.R, R/smallMethods.R
\name{regDiagSingleR}
\alias{regDiagSingleR}
\alias{dfpopsize}
\alias{dfpopsize.singleRStaticCountData}
\alias{dfbeta.singleRStaticCountData}
\alias{hatvalues.singleRStaticCountData}
\alias{residuals.singleRStaticCountData}
\alias{cooks.distance.singleRStaticCountData}
\alias{sigma.singleRStaticCountData}
\alias{influence.singleRStaticCountData}
\alias{rstudent.singleRStaticCountData}
\alias{rstandard.singleRStaticCountData}
\title{Regression diagnostics in singleRcapture}
\usage{
dfpopsize(model, ...)

\method{dfpopsize}{singleRStaticCountData}(model, dfbeta = NULL, ...)

\method{dfbeta}{singleRStaticCountData}(model, maxitNew = 1, trace = FALSE, cores = 1, ...)

\method{hatvalues}{singleRStaticCountData}(model, ...)

\method{residuals}{singleRStaticCountData}(
  object,
  type = c("pearson", "pearsonSTD", "response", "working", "deviance", "all"),
  ...
)

\method{cooks.distance}{singleRStaticCountData}(model, ...)

\method{sigma}{singleRStaticCountData}(object, ...)

\method{influence}{singleRStaticCountData}(model, do.coef = FALSE, ...)

\method{rstudent}{singleRStaticCountData}(model, ...)

\method{rstandard}{singleRStaticCountData}(model, type = c("deviance", "pearson"), ...)
}
\arguments{
\item{model, object}{an object of \code{singleRStaticCountData} class.}

\item{...}{arguments passed to other methods.
Notably \code{dfpopsize.singleRStaticCountData} calls
\code{dfbeta.singleRStaticCountData} if no \code{dfbeta} argument was
provided and \code{controlMethod} is called in \code{dfbeta} method.}

\item{dfbeta}{if \code{dfbeta} was already obtained it is possible to pass
them into function so that they need not be computed for the second time.}

\item{maxitNew}{the maximal number of iterations for regressions with starting
points \mjseqn{\hat{\boldsymbol{\beta}}} on data
specified at call for \code{model} after the removal of k'th row. By default 1.}

\item{trace}{a logical value specifying whether to tracking results when
\code{cores > 1} it will result in a progress bar being created.}

\item{cores}{a number of processor cores to be used,
any number greater than 1 activates code designed with \code{doParallel},
\code{foreach} and \code{parallel} packages. Note that for now using parallel
computing makes tracing impossible so \code{trace} parameter is ignored in this case.}

\item{type}{a type of residual to return.}

\item{do.coef}{logical indicating if \code{dfbeta} computation for influence
should be done. \code{FALSE} by default.}
}
\value{
\itemize{
\item For \code{hatvalues} -- A matrix with n rows and p columns where n is a
number of observations in the data and p is number of regression parameters.
\item For \code{dfpopsize} -- A vector for which k'th element corresponds
to the difference between point estimate of population size estimation on
full data set and point estimate of population size estimation after the
removal of k'th unit from the data set.
\item For \code{dfbeta} -- A matrix with n rows and p observations where p
is a number of units in data and p is the number of regression parameters.
K'th row of this matrix corresponds to
\mjseqn{\hat{\boldsymbol{\beta}}-\hat{\boldsymbol{\beta}}_{-k}}
where \mjseqn{\hat{\boldsymbol{\beta}}_{-k}} is a vector of estimates for
regression parameters after the removal of k'th row from the data.
\item \code{cooks.distance} -- A matrix with a single columns with
values of cooks distance for every unit in \code{model.matrix}
\item \code{residuals.singleRStaticCountData} -- A \code{data.frame}
with chosen residuals.
}
}
\description{
List of some regression diagnostics implemented for
\code{singleRStaticCountData} class. Functions that either require no changes from
\code{glm} class or are not relevant to context of \code{singleRcapture}
are omitted.
}
\details{
\loadmathjax

\code{dfpopsize} and \code{dfbeta} are closely related. \code{dfbeta}
fits a regression after removing a specific row from the data and returns the
difference between regression coefficients estimated on full data set and
data set obtained after deletion of that row, and repeats procedure once
for every unit present in the data.\code{dfpopsize} does the same for
population size estimation utilizing coefficients computed by \code{dfbeta}.

\code{cooks.distance} is implemented (for now) only for models with a single
linear predictor and works exactly like the method for \code{glm} class.

\code{sigma} computes the standard errors of predicted means. Returns a matrix
with two columns first for truncated mean and the other for the non-truncated mean.

\code{residuals.singleRStaticCountData} (can be abbreviated to \code{resid})
works like \code{residuals.glm} with the exception that:
\itemize{
\item \code{"pearson"} -- returns non standardized residuals.
\item \code{"pearsonSTD"} -- is currently defined only for single predictors
models but will be extended to all models in a near future, but for families
with more than one distribution parameter it will be a multivariate residual.
\item \code{"response"} -- returns both residuals computed with truncated
and non truncated fitted value.
\item \code{"working"} -- is possibly multivariate if more than one linear
predictor is present.
\item \code{"deviance"} -- is not yet defined for all families in
\code{\link[=singleRmodels]{singleRmodels()}} e.g. negative binomial based methods.
\item \code{"all"} -- returns all available residual types.
}

\code{hatvalues.singleRStaticCountData} is method for \code{singleRStaticCountData}
class for extracting diagonal elements of projection matrix.

Since \code{singleRcapture} supports not only regular glm's but also vglm's the
\code{hatvalues} returns a matrix with number of columns corresponding to number
of linear predictors in a model, where kth column corresponds to elements of
the diagonal of projection matrix associated with kth linear predictor.
For glm's
\mjsdeqn{\boldsymbol{W}^{\frac{1}{2}}\boldsymbol{X}
\left(\boldsymbol{X}^{T}\boldsymbol{W}\boldsymbol{X}\right)^{-1}
\boldsymbol{X}^{T}\boldsymbol{W}^{\frac{1}{2}}}
where: \mjseqn{\boldsymbol{W}=\mathbb{E}\left(\text{Diag}
\left(\frac{\partial^{2}\ell}{\partial\boldsymbol{\eta}^{T}
\partial\boldsymbol{\eta}}\right)\right)}
and \mjseqn{\boldsymbol{X}} is a model (lm) matrix.
For vglm's present in the package it is instead :
\mjsdeqn{\boldsymbol{X}_{vlm}
\left(\boldsymbol{X}_{vlm}^{T}\boldsymbol{W}\boldsymbol{X}_{vlm}\right)^{-1}
\boldsymbol{X}_{vlm}^{T}\boldsymbol{W}}
where:
\mjsdeqn{
\boldsymbol{W} = \mathbb{E}\left(\begin{bmatrix}
\text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{1}^{T}\partial\eta_{1}}\right) &
\text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{1}^{T}\partial\eta_{2}}\right) &
\dotso & \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{1}^{T}\partial\eta_{p}}\right)\cr
\text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{2}^{T}\partial\eta_{1}}\right) &
\text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{2}^{T}\partial\eta_{2}}\right) &
\dotso & \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{2}^{T}\partial\eta_{p}}\right)\cr
\vdots & \vdots & \ddots & \vdots\cr
\text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{p}^{T}\partial\eta_{1}}\right) &
\text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{p}^{T}\partial\eta_{2}}\right) &
\dotso & \text{Diag}\left(\frac{\partial^{2}\ell}{\partial\eta_{p}^{T}\partial\eta_{p}}\right)
\end{bmatrix}\right)}
is a block matrix constructed by taking the expected  value from diagonal
matrixes corresponding to second derivatives with respect to each linear
predictor (and mixed derivatives) and
\mjseqn{\boldsymbol{X}_{vlm}} is a model (vlm)
matrix constructed using specifications in \code{controlModel} and
call to \code{estimatePopsize}.

\code{influence} works like \code{glm} counterpart computing the most important
influence measures.
}
\examples{
\donttest{
# For singleRStaticCountData class
# Get simple model
Model <- estimatePopsize(
  formula = capture ~ nation + age + gender, 
  data = netherlandsimmigrant, 
  model = ztpoisson, 
  method = "IRLS"
)
# Get dfbeta
dfb <- dfbeta(Model)
# The dfpopsize results are obtained via (It is also possible to not provide 
# dfbeta then they will be computed manually):
res <- dfpopsize(Model, dfbeta = dfb)
summary(res)
plot(res)
# see vaious types of residuals:
head(resid(Model, "all"))
}
}
\seealso{
\code{\link[=estimatePopsize]{estimatePopsize()}} \code{\link[stats:influence.measures]{stats::hatvalues()}} \code{\link[=controlMethod]{controlMethod()}} \code{\link[stats:influence.measures]{stats::dfbeta()}}
\code{\link[stats:influence.measures]{stats::cooks.distance()}}
}
\author{
Piotr Chlebicki, Maciej Beręsewicz
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimatePopsizeFit.R
\name{estimatePopsizeFit}
\alias{estimatePopsizeFit}
\title{Regression fitting in single-source capture-recapture models}
\usage{
estimatePopsizeFit(
  y,
  X,
  family,
  control,
  method,
  priorWeights,
  coefStart,
  etaStart,
  offset,
  ...
)
}
\arguments{
\item{y}{vector of dependent variables.}

\item{X}{model matrix, the vglm one.}

\item{family}{same as model in \code{estimatePopsize}.}

\item{control}{control parameters created in \code{controlModel}.}

\item{method}{method of estimation same as in \code{estimatePopsize}.}

\item{priorWeights}{vector of prior weights its the same argument as weights
in \code{estimatePopsize}.}

\item{etaStart, coefStart}{initial value of regression parameters or
linear predictors.}

\item{offset}{offset passed from by default passed from \code{\link[=estimatePopsize]{estimatePopsize()}}.}

\item{...}{arguments to pass to other methods.}
}
\value{
List with regression parameters, working weights
(if IRLS fitting method) was chosen and number of iterations taken.
}
\description{
\code{estimatePopsizeFit} does for \code{estimatePopsize} what
\code{glm.fit} does for \code{glm}. It is internally called in
\code{estimatePopsize}. Since \code{estimatePopsize} does much more than
just regression fitting \code{estimatePopsizeFit} is much faster.
}
\details{
\loadmathjax

If \code{method} argument was set to \code{"optim"} the \code{stats::optim}
function will be used to fit regression with analytically computed gradient and
(minus) log likelihood functions as \code{gr} and \code{fn} arguments.
Unfortunately \code{optim} does not allow for hessian to be specified.
More information about how to modify \code{optim} fitting is included in
\code{\link[=controlMethod]{controlMethod()}}.

If \code{method} argument was set to \code{"IRLS"} the iteratively reweighted
least squares. The algorithm is well know in generalised linear models.
Thomas W. Yee later extended this algorithm to vector generalised linear models
and in more general terms it can roughly be described as
(this is Yee's description after changing some conventions):
\enumerate{
\item Initialize with:
\itemize{
\item \code{converged <- FALSE}
\item \code{iter <- 1}
\item \mjseqn{\boldsymbol{\beta}}\code{ <- start}
\item \mjseqn{\boldsymbol{W}}\code{ <- prior}
\item \mjseqn{\ell}\code{ <- }\mjseqn{\ell(\boldsymbol{\beta})}
}
\item If \code{converged} or \code{iter > Maxiter} move to step 7.
\item Store values from previous algorithm step:
\itemize{
\item \mjseqn{\boldsymbol{W}_{-}}\code{ <- }
\mjseqn{\boldsymbol{W}}
\item \mjseqn{\ell_{-}}\code{ <- }
\mjseqn{\ell}
\item \mjseqn{\boldsymbol{\beta}_{-}}\code{ <- }
\mjseqn{\boldsymbol{\beta}}
} and assign values at current step:
\itemize{
\item \mjseqn{\boldsymbol{\eta}}\code{ <- }
\mjseqn{\boldsymbol{X}_{vlm}\boldsymbol{\beta}}
\item \mjseqn{Z_{i}}\code{ <- }
\mjseqn{
\eta_{i}+\frac{\partial\ell_{i}}{\partial\eta_{i}}
\mathbb{E}\left(\frac{\partial^{2}\ell_{i}}{
\partial\eta_{i}^{T}\partial\eta_{i}}\right)^{-1}}
\item \mjseqn{\boldsymbol{W}_{ij}}\code{ <- }
\mjseqn{\mathbb{E}\left(\frac{\partial^{2}\ell}{
\partial\boldsymbol{\eta}_{j}^{T}\partial\boldsymbol{\eta}_{i}}\right)}
}
where \mjseqn{\ell_{i}} is the ith component of log likelihood
function, \mjseqn{\eta_{i}} is the vector of linear predictors
associated with ith row and \mjseqn{\mathbb{E}\left(\frac{\partial^{2}\ell_{i}}{
\partial\eta_{i}^{T}\partial\eta_{i}}\right)}
corresponds to weights associated with ith row and \mjseqn{\boldsymbol{W}}
is a block matrix, made of diagonal matrixes
\mjseqn{\mathbb{E}\left(\frac{\partial^{2}\ell}{
\partial\boldsymbol{\eta}_{j}^{T}\partial\boldsymbol{\eta}_{i}}\right)}
\item Regress \mjseqn{\boldsymbol{Z}} on
\mjseqn{\boldsymbol{X}_{vlm}} to obtain
\mjseqn{\boldsymbol{\beta}} as:
\mjsdeqn{\boldsymbol{\beta}=
\left(\boldsymbol{X}_{vlm}^{T}\boldsymbol{W}\boldsymbol{X}_{vlm}\right)^{-1}
\boldsymbol{X}_{vlm}^{T}\boldsymbol{W}\boldsymbol{Z}}
\item Assign:
\itemize{
\item\code{converged <- }\mjteqn{\ell(\boldsymbol{\beta})-\ell_{-} < \varepsilon\cdot\ell_{-}}{\ell(\boldsymbol{\beta})-\ell_{-} \lt \varepsilon\cdot\ell_{-}}{\ell(\boldsymbol{\beta})-\ell_{-} < \varepsilon\cdot\ell_{-}}
or
\mjteqn{||\boldsymbol{\beta}-\boldsymbol{\beta}_{-}||_{\infty} < \varepsilon}{||\boldsymbol{\beta}-\boldsymbol{\beta}_{-}||_{\infty} \lt \varepsilon}{||\boldsymbol{\beta}-\boldsymbol{\beta}_{-}||_{\infty} < \varepsilon}
\item\code{iter <- iter + 1}
}
where \mjseqn{\varepsilon} is the relative tolerance level,
by default \code{1e-8}.
\item Return to step 2.
\item Return \mjseqn{\boldsymbol{\beta}, \boldsymbol{W}}, \code{iter}.
}

In this package we use different conventions for \mjseqn{\boldsymbol{X}_{vlm}}
matrix hence slight differences are present in algorithm description but
results are identical.
}
\examples{
\donttest{
summary(farmsubmission)

# construct vglm model matrix
X <- matrix(data = 0, nrow = 2 * NROW(farmsubmission), ncol = 7)
X[1:NROW(farmsubmission), 1:4] <- model.matrix(
~ 1 + log_size + log_distance + C_TYPE, 
farmsubmission
)


X[-(1:NROW(farmsubmission)), 5:7] <- X[1:NROW(farmsubmission), c(1, 3, 4)]

# this attribute tells the function which elements of the design matrix 
# correspond to which linear predictor 
attr(X, "hwm") <- c(4, 3)

# get starting points
start <- glm.fit(
y = farmsubmission$TOTAL_SUB, 
x = X[1:NROW(farmsubmission), 1:4], 
family = poisson()
)$coefficients

res <- estimatePopsizeFit(
y = farmsubmission$TOTAL_SUB, 
X = X, 
method = "IRLS", 
priorWeights = 1, 
family = ztoigeom(), 
control = controlMethod(verbose = 5), 
coefStart = c(start, 0, 0, 0),
etaStart = matrix(X \%*\% c(start, 0, 0, 0), ncol = 2),
offset = cbind(rep(0, NROW(farmsubmission)), rep(0, NROW(farmsubmission)))
)

# extract results

# regression coefficient vector
res$beta

# check likelihood
ll <- ztoigeom()$makeMinusLogLike(y = farmsubmission$TOTAL_SUB, X = X)

-ll(res$beta)

# number of iterations
res$iter

# working weights
head(res$weights)

# Compare with optim call

res2 <- estimatePopsizeFit(
  y = farmsubmission$TOTAL_SUB, 
  X = X, 
  method = "optim", 
  priorWeights = 1, 
  family = ztoigeom(), 
  coefStart = c(start, 0, 0, 0),
  control = controlMethod(verbose = 1, silent = TRUE),
  offset = cbind(rep(0, NROW(farmsubmission)), rep(0, NROW(farmsubmission)))
)
# extract results

# regression coefficient vector
res2$beta


# check likelihood
-ll(res2$beta)

# number of calls to log lik function
# since optim does not return the number of
# iterations
res2$iter

# optim does not calculated working weights
head(res2$weights)
}
}
\references{
Yee, T. W. (2015). Vector Generalized Linear and Additive Models:
With an Implementation in R. New York, USA: Springer. ISBN 978-1-4939-2817-0.
}
\seealso{
\code{\link[stats:glm]{stats::glm()}} \code{\link[=estimatePopsize]{estimatePopsize()}} \code{\link[=controlMethod]{controlMethod()}} \code{\link[stats:optim]{stats::optim()}}
}
\author{
Piotr Chlebicki, Maciej Beresewicz
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chao.R, R/documentationFiles.R,
%   R/hurdleztgeom.R, R/hurdleztnegbin.R, R/hurdleztpoisson.R, R/oichao.R,
%   R/oiztgeom.R, R/oiztnegbin.R, R/oiztpoisson.R, R/zelterman.R, R/zotgeom.R,
%   R/zotnegbin.R, R/zotpoisson.R, R/ztHurdlegeom.R, R/ztHurdlenegbin.R,
%   R/ztHurdlepoisson.R, R/ztgeom.R, R/ztnegbin.R, R/ztoigeom.R, R/ztoinegbin.R,
%   R/ztoipoisson.R, R/ztpoisson.R
\name{chao}
\alias{chao}
\alias{singleRmodels}
\alias{Hurdleztgeom}
\alias{Hurdleztnegbin}
\alias{Hurdleztpoisson}
\alias{oichao}
\alias{oiztgeom}
\alias{oiztnegbin}
\alias{oiztpoisson}
\alias{zelterman}
\alias{zotgeom}
\alias{zotnegbin}
\alias{zotpoisson}
\alias{ztHurdlegeom}
\alias{ztHurdlenegbin}
\alias{ztHurdlepoisson}
\alias{ztgeom}
\alias{ztnegbin}
\alias{ztoigeom}
\alias{ztoinegbin}
\alias{ztoipoisson}
\alias{ztpoisson}
\title{Family functions in singleRcapture package}
\usage{
chao(lambdaLink = "loghalf", ...)

Hurdleztgeom(
  lambdaLink = c("log", "neglog"),
  piLink = c("logit", "cloglog", "probit"),
  ...
)

Hurdleztnegbin(
  nSim = 1000,
  epsSim = 1e-08,
  eimStep = 6,
  lambdaLink = c("log", "neglog"),
  alphaLink = c("log", "neglog"),
  piLink = c("logit", "cloglog", "probit"),
  ...
)

Hurdleztpoisson(
  lambdaLink = c("log", "neglog"),
  piLink = c("logit", "cloglog", "probit"),
  ...
)

oichao(lambdaLink = "logthird", ...)

oiztgeom(
  lambdaLink = c("log", "neglog"),
  omegaLink = c("logit", "cloglog", "probit"),
  ...
)

oiztnegbin(
  nSim = 1000,
  epsSim = 1e-08,
  eimStep = 6,
  lambdaLink = c("log", "neglog"),
  alphaLink = c("log", "neglog"),
  omegaLink = c("logit", "cloglog", "probit"),
  ...
)

oiztpoisson(
  lambdaLink = c("log", "neglog"),
  omegaLink = c("logit", "cloglog", "probit"),
  ...
)

zelterman(lambdaLink = "loghalf", ...)

zotgeom(lambdaLink = c("log", "neglog"), ...)

zotnegbin(
  nSim = 1000,
  epsSim = 1e-08,
  eimStep = 6,
  lambdaLink = c("log", "neglog"),
  alphaLink = c("log", "neglog"),
  ...
)

zotpoisson(lambdaLink = c("log", "neglog"), ...)

ztHurdlegeom(
  lambdaLink = c("log", "neglog"),
  piLink = c("logit", "cloglog", "probit"),
  ...
)

ztHurdlenegbin(
  nSim = 1000,
  epsSim = 1e-08,
  eimStep = 6,
  lambdaLink = c("log", "neglog"),
  alphaLink = c("log", "neglog"),
  piLink = c("logit", "cloglog", "probit"),
  ...
)

ztHurdlepoisson(
  lambdaLink = c("log", "neglog"),
  piLink = c("logit", "cloglog", "probit"),
  ...
)

ztgeom(lambdaLink = c("log", "neglog"), ...)

ztnegbin(
  nSim = 1000,
  epsSim = 1e-08,
  eimStep = 6,
  lambdaLink = c("log", "neglog"),
  alphaLink = c("log", "neglog"),
  ...
)

ztoigeom(
  lambdaLink = c("log", "neglog"),
  omegaLink = c("logit", "cloglog", "probit"),
  ...
)

ztoinegbin(
  nSim = 1000,
  epsSim = 1e-08,
  eimStep = 6,
  lambdaLink = c("log", "neglog"),
  alphaLink = c("log", "neglog"),
  omegaLink = c("logit", "cloglog", "probit"),
  ...
)

ztoipoisson(
  lambdaLink = c("log", "neglog"),
  omegaLink = c("logit", "cloglog", "probit"),
  ...
)

ztpoisson(lambdaLink = c("log", "neglog"), ...)
}
\arguments{
\item{lambdaLink}{a link for Poisson parameter, \code{"log"}
by default except for rare-count Poisson models such as
\code{zelterman()} and \code{chao()} where only
\mjseqn{\ln\left(\frac{x}{2}\right)} is possible and \code{oichao()}
where only \mjseqn{\ln\left(\frac{x}{3}\right)} is possible.}

\item{...}{Additional arguments, not used for now.}

\item{piLink}{a link for probability parameter,  \code{"logit"} by default}

\item{nSim, epsSim}{if working weights cannot be computed analytically these
arguments specify maximum number of simulations allowed and
precision level for finding them numerically respectively.}

\item{eimStep}{a non negative integer describing
how many values should be used at each step of approximation
of information matrixes when no analytic solution is available
(e.g. \code{"ztnegbin"}), default varies depending on a function.
Higher value usually means faster convergence but may potentially cause
issues with convergence.}

\item{alphaLink}{a link for dispersion parameter, \code{"log"} by default}

\item{omegaLink}{a link for inflation parameter, \code{"logit"} by default}
}
\value{
A object of class \code{family} containing objects:
\itemize{
\item \code{makeMinusLogLike} -- A factory function for creating the following functions:
\mjseqn{\ell(\boldsymbol{\beta}), \frac{\partial\ell}{\partial\boldsymbol{\beta}},
\frac{\partial^{2}\ell}{\partial\boldsymbol{\beta}^{T}\partial\boldsymbol{\beta}}
} functions from the
\mjseqn{\boldsymbol{y}} vector and the
\mjseqn{\boldsymbol{X}_{vlm}} matrix
(or just \mjseqn{\boldsymbol{X}} if applied to model
with single linear predictor)which has the \code{deriv} argument with possible
values in \code{c(0, 1, 2)} that determine which derivative to return;
the default value is \code{0}, which represents the minus log-likelihood.
\item \code{links} -- A list with link functions.
\item \code{mu.eta, variance} -- Functions of linear predictors that
return expected value and variance. The \code{type} argument with 2 possible
values (\code{"trunc"} and \code{"nontrunc"}) that specifies whether
to return \mjseqn{\mathbb{E}(Y|Y>0), \text{var}(Y|Y>0)} or
\mjseqn{\mathbb{E}(Y), \text{var}(Y)} respectively; the \code{deriv} argument
with values in \code{c(0, 1, 2)} is used for indicating the derivative with
respect to the linear predictors, which is used for providing
standard errors in the \code{predict} method.
\item \code{family} -- A string that specifies name of the model.
\item \code{valideta, validmu} -- For now it only returns \code{TRUE}.
In the near future, it will be used to check whether applied linear
predictors are valid (i.e. are transformed into some elements of the
parameter space subjected to the inverse link function).
\item \code{funcZ, Wfun} -- Functions that create pseudo residuals and
working weights used in IRLS algorithm.
\item \code{devResids} -- A function that given the linear predictors
prior weights vector and response vector returns deviance residuals.
Not all family functions have these functions implemented yet.
\item \code{pointEst, popVar} -- Functions that given prior weights
linear predictors and in the latter case also estimate of
\mjseqn{\text{cov}(\hat{\boldsymbol{\beta}})} and \mjseqn{\boldsymbol{X_{vlm}}}
matrix return point estimate for population size and analytic estimation
of its variance.There is a additional boolean parameter \code{contr} in the
former function that if set to true returns contribution of each unit.
\item \code{etaNames} -- Names of linear predictors.
\item \code{densityFunction} -- A function that given linear predictors
returns value of PMF at values \code{x}. Additional argument \code{type}
specifies whether to return \mjseqn{\mathbb{P}(Y|Y>0)} or
\mjseqn{\mathbb{P}(Y)}.
\item \code{simulate} -- A function that generates values of a dependent
vector given linear predictors.
\item \code{getStart} -- An expression for generating starting points.
}
}
\description{
Package \code{singleRcapture} utilizes various family type
functions that specify variable parts of population size estimation,
regression, diagnostics and other necessary information that depends
on the model. These functions are used as \code{model} argument in
\code{estimatePopsize} function.
}
\details{
\loadmathjax

Most of these functions are based on some "base" distribution with
support \mjseqn{\mathbb{N}_{0}=\mathbb{N}\cup\lbrace 0\rbrace} that describe
distribution of \mjseqn{Y} before truncation. Currently they include:
\mjsdeqn{\mathbb{P}(Y=y|\lambda,\alpha)=\left\lbrace
\begin{array}{cc}
\frac{\lambda^{y}e^{-\lambda}}{y!}    & \text{Poisson distribution}  \cr
\frac{\Gamma(y+\alpha^{-1})}{\Gamma(\alpha^{-1})y!} 
\left(\frac{\alpha^{-1}}{\alpha^{-1}+\lambda}\right)^{\alpha^{-1}}
\left(\frac{\lambda}{\alpha^{-1}+\lambda}\right)^{y} & 
\text{negative binomial distribution} \cr
\frac{\lambda^{y}}{(1+\lambda)^{y+1}} & 
\text{geometric distribution}
\end{array}
\right.}
where \mjseqn{\lambda} is the Poisson parameter and
\mjseqn{\alpha} is the dispersion parameter. Geometric distribution
is a special case of negative binomial distribution when
\mjseqn{\alpha=1} it is included because negative binomial
distribution is quite troublesome numerical regression in fitting.
It is important to know that PMF of negative binomial distribution
approaches the PMF of Poisson distribution when
\mjseqn{\alpha\rightarrow 0^{+}}.

\strong{Note} in literature on single source capture recapture models
the dispersion parameter which introduces greater variability
in negative binomial distribution compared to Poisson distribution is
generally interpreted as explaining the \emph{unobserved} heterogeneity
i.e. presence of important unobserved independent variables.
All these methods for estimating population size are tied to Poisson
processes hence we use \mjseqn{\lambda} as parameter symbol
instead of \mjseqn{\mu} to emphasize this connection.
Also will not be hard to see that \strong{all} estimators derived from
modifying the "base" distribution are unbiased if assumptions
made by respective models are not violated.

The \strong{zero-truncated} models corresponding to "base" distributions are
characterized by relation:
\mjsdeqn{\mathbb{P}(Y=y|Y>0)=\left\lbrace
\begin{array}{cc}
\frac{\mathbb{P}(Y=y)}{1-\mathbb{P}(Y=0)} & \text{when }y\neq 0 \cr
0 & \text{when }y=0
\end{array}\right.}
which allows us to estimate parameter values using only observed part of
population. These models lead to the following estimates, respectively:
\mjsdeqn{
\begin{aligned}
\hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1}{1-\exp(-\lambda_{k})} &
\text{ For Poisson distribution} \cr
\hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1}{1-(1+\\alpha_{k}\lambda_{k})^{-\alpha_{k}^{-1}}} &
\text{ For negative binomial distribution} \cr
\hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1+\lambda_{k}}{\lambda_{k}} &
\text{ For geometric distribution}
\end{aligned}
}

One common way in which assumptions of zero-truncated models are violated is
presence of \strong{one-inflation} the presence of which is somewhat similar in
single source capture-recapture models to zero-inflation in usual count data
analysis. There are two ways in which one-inflation may be understood,
they relate to whether \mjseqn{\mathbb{P}(Y=0)} is
modified by inflation. The first approach is inflate
(\mjseqn{\omega} parameter) zero-truncated distribution as:
\mjsdeqn{
\mathbb{P}_{new}(Y=y|Y>0) = \left\lbrace\begin{array}{cc}
\omega + (1 - \omega)\mathbb{P}_{old}(Y=1|Y>0)& \text{when: } y = 1 \cr
(1 - \omega) \mathbb{P}_{old}(Y=y|Y>0) & \text{when: } y \neq 1
\end{array}\right.}
which corresponds to:
\mjsdeqn{
\mathbb{P}_{new}(Y=y) = \left\lbrace\begin{array}{cc}
\mathbb{P}_{old}(Y=0) & \text{when: } y = 0 \cr
\omega(1 - \mathbb{P}(Y=0)) + (1 - \omega)\mathbb{P}_{old}(Y=1) & \text{when: } y = 1 \cr
(1 - \omega) \mathbb{P}_{old}(Y=y) & \text{when: } y > 1
\end{array}\right.
}
before zero-truncation. Models that utilize this
approach are commonly referred to as \emph{one-inflated zero-truncated models} (\verb{oizt*()} family).
Another way of accommodating one-inflation in SSCR is by putting inflation
parameter on base distribution as:
\mjsdeqn{
\mathbb{P}_{new}(Y=y) = \left\lbrace\begin{array}{cc}
\omega + (1 - \omega)\mathbb{P}_{old}(Y=1)& \text{when: } y = 1 \cr
(1 - \omega) \mathbb{P}_{old}(Y=y) & \text{when: } y \neq 1
\end{array}\right.
}
which then becomes:
\mjsdeqn{
\mathbb{P}_{new}(Y=y|Y>0) = \left\lbrace\begin{array}{cc}
\frac{\omega}{1 - (1-\omega)\mathbb{P}_{old}(Y=0)} + \frac{(1 - \omega)}{1 - (1-\omega)\mathbb{P}_{old}(Y=0)}\mathbb{P}_{old}(Y=1)& \text{when: } y = 1 \cr
\frac{(1 - \omega)}{1 - (1-\omega)\mathbb{P}_{old}(Y=0)}\mathbb{P}_{old}(Y=y) & \text{when: } y > 1
\end{array}\right.
}
after truncation and are commonly referred to as \emph{zero-truncated one-inflated models} (\verb{ztoi*()} family).
It was shown by Böhning in 2022 paper that these approaches are equivalent
in terms of maximizing likelihoods if we do not put formula on
\mjseqn{\omega}. They can however lead to different
population size estimates.

For \emph{one-inflated zero-truncated models} the formula for population size
estimate \mjseqn{\hat{N}} does not change since
\mjseqn{\mathbb{P}(y=0)} remains the same but estimation of parameters
changes all calculations.

For \emph{zero-truncated one-inflated models} population size estimates are
expressed, respectively by:
\mjsdeqn{
\begin{aligned}
\hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1}{1-(1-\omega_{k})\exp(-\lambda_{k})} &\text{ For base Poisson distribution} \cr
\hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1}{1-(1-\omega_{k})(1+\\alpha_{k}\lambda_{k})^{-\alpha_{k}^{-1}}} &\text{ For base negative binomial distribution} \cr
\hat{N} &= \sum_{k=1}^{N_{obs}}\frac{1+\lambda_{k}}{\lambda_{k} + \omega_{k}} &\text{ For base geometric distribution}
\end{aligned}
}

\strong{Zero-one-truncated} models ignore one counts instead of accommodating
one-inflation by utilizing the identity
\mjsdeqn{
\ell_{\text{ztoi}}=\boldsymbol{f}_{1}\ln{\frac{\boldsymbol{f}_{1}}{N_{obs}}}
+(N_{obs}-\boldsymbol{f}_{1})\ln{\left(1-\frac{\boldsymbol{f}_{1}}{N_{obs}}
\right)} + \ell_{\text{zot}}
}
where \mjseqn{\ell_{\text{zot}}} is the log likelihood
of zero-one-truncated distribution characterized by probability mass function:
\mjsdeqn{\mathbb{P}(Y=y|Y>1)=\left\lbrace
\begin{array}{cc}
\frac{\mathbb{P}(Y=y)}{1-\mathbb{P}(Y=0)-\mathbb{P}(Y=1)} & \text{when }y > 1 \cr
0 & \text{when }y\in\lbrace 0, 1\rbrace
\end{array}\right.}
where \mjseqn{\mathbb{P}(Y)} is the probability mass function of
the "base" distribution. The identity above justifies use of zero-one-truncated,
unfortunately it was only proven for intercept only models, however
numerical simulations seem to indicate that even if the theorem cannot be
extended for (non trivial) regression population size estimation is still
possible.

For \emph{zero-one-truncated models} population size estimates are expressed by:
\mjsdeqn{
\begin{aligned}
\hat{N} &= \boldsymbol{f}_{1} + \sum_{k=1}^{N_{obs}}
\frac{1-\lambda_{k}\exp(-\lambda_{k})}{1-\exp(-\lambda_{k})-\lambda_{k}\exp(-\lambda_{k})} 
&\text{ For base Poisson distribution} \cr
\hat{N} &= \boldsymbol{f}_{1} + \sum_{k=1}^{N_{obs}}
\frac{1-\lambda_{k}(1+\alpha_{k}\lambda_{k})^{-1-\alpha_{k}^{-1}}}{
1-(1+\alpha_{k}\lambda_{k})^{-\alpha_{k}^{-1}}-\lambda_{k}(1+\alpha_{k}\lambda_{k})^{-1-\alpha_{k}^{-1}}} 
&\text{ For base negative binomial distribution} \cr
\hat{N} &= \boldsymbol{f}_{1} + \sum_{k=1}^{N_{obs}}
\frac{\lambda_{k}^{2}+\lambda_{k}+1}{\lambda_{k}^{2}} 
&\text{ For base geometric distribution}
\end{aligned}
}

Pseudo hurdle models are experimental and not yet described in literature.

Lastly there are \strong{chao} and \strong{zelterman} models which are based on
logistic regression on the dummy variable
\mjsdeqn{
Z = \left\lbrace\begin{array}{cc}
0     & \text{if }Y = 1  \cr
1     & \text{if }Y = 2
\end{array}\right.}
based on the equation:
\mjsdeqn{
\text{logit}(p_{k})=
\ln\left(\frac{\lambda_{k}}{2}\right)=
\boldsymbol{\beta}\mathbf{x}_{k}=\eta_{k}}
where \mjseqn{\lambda_{k}} is the Poisson parameter.

The \emph{zelterman} estimator of population size is expressed as:
\mjsdeqn{\hat{N}=\sum_{k=1}^{N_{obs}}{1-\exp\left(-\lambda_{k}\right)}}
and \emph{chao} estimator has the form:
\mjsdeqn{
\hat{N}=N_{obs}+\sum_{k=1}^{\boldsymbol{f}_{1}+\boldsymbol{f}_{2}}
\frac{1}{\lambda_{k}+ \frac{\lambda_{k}^{2}}{2}}
}
The \emph{oichao} estimator is the modified Chao analogue robust to one-inflation
and is based on the dummy variable
\mjsdeqn{
Z = \left\lbrace\begin{array}{cc}
0     & \text{if }Y = 2  \cr
1     & \text{if }Y = 3
\end{array}\right.}
with equation:
\mjsdeqn{
\text{logit}(p_{k})=
\ln\left(\frac{\lambda_{k}}{3}\right)=
\boldsymbol{\beta}\mathbf{x}_{k}=\eta_{k}}
and population size estimator:
\mjsdeqn{
\hat{N}=N_{obs}+\sum_{k=1}^{\boldsymbol{f}_{2}+\boldsymbol{f}_{3}}
\frac{1}{\frac{\lambda_{k}^{2}}{2}+ \frac{\lambda_{k}^{3}}{6}}
}
}
\seealso{
\code{\link[=estimatePopsize]{estimatePopsize()}}
}
\author{
Piotr Chlebicki, Maciej Beręsewicz

Cyprian Jurkowski, Piotr Chlebicki, Maciej Beręsewicz
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ratioReg.R
\name{plot.singleRRatioReg}
\alias{plot.singleRRatioReg}
\title{Diagnostic Plots for Ratio Regression Fits}
\usage{
\method{plot}{singleRRatioReg}(x, plotType = c("ratio", "bootHist", "freqFit"), ...)
}
\arguments{
\item{x}{object of class \code{singleRRatioReg}.}

\item{plotType}{one of \code{"ratio"}, \code{"bootHist"}, or \code{"freqFit"}.}

\item{...}{additional arguments passed to the underlying graphics calls.}
}
\value{
No return value, called for side effects.
}
\description{
Plot methods for \code{singleRRatioReg} objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/updates.R
\name{redoPopEstimation}
\alias{redoPopEstimation}
\alias{redoPopEstimation.singleRStaticCountData}
\title{Updating population size estimation results}
\usage{
redoPopEstimation(object, newdata, ...)

\method{redoPopEstimation}{singleRStaticCountData}(
  object,
  newdata,
  cov,
  weights,
  coef,
  control,
  popVar,
  offset,
  weightsAsCounts,
  ...
)
}
\arguments{
\item{object}{object for which update of population size estimation results will be done.}

\item{newdata}{optional \code{data.frame} with new data for pop size estimation.}

\item{...}{additional optional arguments, currently not used in \code{singleRStaticCountData} class method.}

\item{cov}{an updated covariance matrix estimate.}

\item{weights}{optional vector of weights to use in population size estimation.}

\item{coef}{optional vector of coefficients of regression on which to base
population size estimation. If missing it is set to \code{coef(object)}.}

\item{control}{similar to \code{controlPopVar} in \code{\link[=estimatePopsize]{estimatePopsize()}}.
If missing set to controls provided on call to \code{object}.}

\item{popVar}{similar to \code{popVar} in \code{\link[=estimatePopsize]{estimatePopsize()}}.
If missing set to \code{"analytic"}.}

\item{offset}{offset argument for new data}

\item{weightsAsCounts}{for \code{singleRStaticCountData} method used to specify
whether weights should be treated as number of occurrences for rows in data}
}
\value{
An object of class \code{popSizeEstResults} containing updated
population size estimation results.
}
\description{
A function that applies all post-hoc procedures that were taken
(such as heteroscedastic consistent covariance matrix estimation or bias
reduction) to population size estimation and standard error estimation.
}
\details{
\loadmathjax

Any non specified arguments will be inferred from the \code{object}
}
\examples{
# Create simple model
Model <- estimatePopsize(
  formula = capture ~ nation + gender, 
  data = netherlandsimmigrant, 
  model = ztpoisson, 
  method = "IRLS"
)
# Apply heteroscedasticity consistent covariance matrix estimation
require(sandwich)
cov <- vcovHC(Model, type = "HC3")
summary(Model, cov = cov,
popSizeEst = redoPopEstimation(Model, cov = cov))
# Compare to results with usual covariance matrix estimation
summary(Model)

## get confidence interval with larger significance level
redoPopEstimation(Model, control = controlPopVar(alpha = .000001))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{controlPopVar}
\alias{controlPopVar}
\title{Control parameters for population size estimation}
\usage{
controlPopVar(
  alpha = 0.05,
  bootType = c("parametric", "semiparametric", "nonparametric"),
  B = 500,
  confType = c("percentile", "normal", "basic"),
  keepbootStat = TRUE,
  traceBootstrapSize = FALSE,
  bootstrapVisualTrace = FALSE,
  fittingMethod = c("optim", "IRLS"),
  bootstrapFitcontrol = NULL,
  sd = c("sqrtVar", "normalMVUE"),
  covType = c("observedInform", "Fisher"),
  cores = 1L
)
}
\arguments{
\item{alpha}{a significance level, 0.05 used by default.}

\item{bootType}{the bootstrap type to be used. Default is \code{"parametric"},
other possible values are: \code{"semiparametric"} and \code{"nonparametric"}.}

\item{B}{a number of bootstrap samples to be performed (default 500).}

\item{confType}{a type of confidence interval for bootstrap confidence interval,
\code{"percentile"} by default.
Other possibilities: \code{"studentized"} and \code{"basic"}.}

\item{keepbootStat}{a boolean value indicating whether to keep a vector of
statistics produced by bootstrap.}

\item{traceBootstrapSize}{a boolean value indicating whether to print size of
bootstrapped sample after truncation for semi- and fully parametric bootstraps.}

\item{bootstrapVisualTrace}{a boolean value indicating whether to plot bootstrap
statistics in real time if \code{cores = 1} if \code{cores > 1} it instead
indicates whether to make progress bar.}

\item{fittingMethod}{a method used for fitting models from bootstrap samples.}

\item{bootstrapFitcontrol}{control parameters for each regression works exactly
like \code{controlMethod} but for fitting models from bootstrap samples.}

\item{sd}{a character indicating how to compute standard deviation of population
size estimator either as:
\mjsdeqn{\hat{\sigma}=\sqrt{\hat{\text{var}}(\hat{N})}}
for \code{sqrt} (which is slightly biased if \mjseqn{\hat{N}}
has a normal distribution) or for \code{normalMVUE} as the unbiased
minimal variance estimator for normal distribution:
\mjsdeqn{\hat{\sigma}=\sqrt{\hat{\text{var}}(\hat{N})}
\frac{\Gamma\left(\frac{N_{obs}-1}{2}\right)}{\Gamma\left(\frac{N_{obs}}{2}\right)}
\sqrt{\frac{N_{obs}}{2}}}
where the ration involving gamma functions is computed by log gamma function.}

\item{covType}{a type of covariance matrix for regression parameters by default
observed information matrix.}

\item{cores}{for bootstrap only, a number of processor cores to be used,
any number greater than 1 activates code designed with \code{doParallel},
\code{foreach} and \code{parallel} packages. Note that for now using parallel
computing makes tracing impossible so \code{traceBootstrapSize} and
\code{bootstrapVisualTrace} parameters are ignored in this case.}
}
\value{
A list with selected parameters, it is also possible to call list directly.
}
\description{
Creating control parameters for population size estimation and
respective standard error and variance estimation.
}
\seealso{
\code{\link[=estimatePopsize]{estimatePopsize()}} \code{\link[=controlModel]{controlModel()}} \code{\link[=controlMethod]{controlMethod()}}
}
\author{
Piotr Chlebicki, Maciej Beręsewicz
\loadmathjax
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal.R
\name{marginalFreq}
\alias{marginalFreq}
\title{Observed and fitted marginal frequencies}
\usage{
marginalFreq(
  object,
  includeones = TRUE,
  includezeros = TRUE,
  onecount = NULL,
  range,
  ...
)
}
\arguments{
\item{object}{object of \code{singleR} class.}

\item{includeones}{logical value indicating whether to include the estimated number of zero counts.}

\item{includezeros}{logical value indicating whether to include one counts in the zero-one truncated models.}

\item{onecount}{a numeric value indicating number of one counts if null \code{trcount} from object will be assumed to be a number one counts.}

\item{range}{optional argument specifying range of selected Y values.}

\item{...}{currently does nothing.}
}
\value{
A list with observed name of the fitted model family degrees of freedom and observed and fitted marginal frequencies.
}
\description{
A function that given a fitted \code{singleR} class object
computed marginal frequencies by as sum of probability density functions
for each unit in data at each point i.e. kth element of marginal frequency
table is given by \mjseqn{\sum_{j=1}^{N_{obs}}\mathbb{P}(Y_{j}=k|\eta_{j})}.
For k=0 only (if specified at call) they are computed as
\mjseqn{\hat{N}-N_{obs}} because
\mjseqn{\boldsymbol{f}_{0}} is assumed to
the unobserved part of the studied population.

These frequencies are useful in diagnostics for count data regression, such
as assessment of fit.
}
\seealso{
\code{\link[=estimatePopsize]{estimatePopsize()}} -- where example of usage is provided
}
\author{
Piotr Chlebicki
\loadmathjax
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{farmsubmission}
\alias{farmsubmission}
\title{The British farm submissions data}
\format{
Data frame with 12,036 rows and 4 columns.
\describe{
\item{\code{TOTAL_SUB}}{Number of submissions of animal material.}
\item{\code{log_size}}{Numerical value equal to logarithm of size of farm.}
\item{\code{log_distance}}{Numerical value equal to logarithm of distance to nearest AHVLA center.}
\item{\code{C_TYPE}}{Factor describing type of activity on farm that animals are
used for. Either \code{Dairy} or \code{Beef}}
}
}
\usage{
data("farmsubmission")
}
\description{
Data on British animal farms submissions to AHVLA. British farms
are able to submit samples to AHVLA if cause of death for an animal
cannot be determined and private veterinary surgeon decides to submit them,
unless there is notifiable disease suspected then such a submission is not
required.

This data set contains information about such farms.
All submissions from farms are included in this data frame
not only carcasses but also blood samples etc.
}
\references{
This data set and its description was provided in publication:
Böhning, D., Vidal Diez, A., Lerdsuwansri, R., Viwatwongkasem, C., and Arnold, M. (2013). "A generalization of Chao's estimator for covariate information". \emph{Biometrics}, 69(4), 1033-1042. doi:10.1111/biom.12082
}
\keyword{datasets}
